/*
FILE:     TblFileObj.h
*/
/*
VERSION:  1.700
*/
/*
DATE:     7/13/2005
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2005 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
/* 
  PURPOSE:    Class for tables management in persistent store
*/

#ifndef TBLFILEOBJ_H
#define TBLFILEOBJ_H

#include "ISTable.h"
#include "FileNavigator.h"
#include "FileNavigatorError.h"
#include "BTreeObj.h"

#define CASE_SENSITIVE 0
#define CASE_INSENSITIVE 1

class TableInfo {
  friend class TblFileObj;
 protected:
  char * _block;  // external block name
  char * _name;   // external table name
  char * _id;     // internal table name
  int _objectId;  // a zero indicates that the object is not on disk or not in memory
  // the object number (used to fetch from disk)
  int _memId;     // the object number (used to fetch from memory)
  int _numItems;  // the number of file navigator units used 
  int _modified;  // indicates table has been modified in append mode
  int _caseSense; // indicates case sensitivity of the internal table name

 public:
  TableInfo(int caseSense = CASE_SENSITIVE);
  TableInfo(const char *c, int caseSense = CASE_SENSITIVE);
  TableInfo(const TableInfo &t);
  ~TableInfo() ;
  char * GetBlock() {return _block;};  
  char * GetName() {return _name;};   
  char * GetId() {return _id;};
  int GetObjectId() {return _objectId;};
  int GetMemId() {return _memId;};
  int GetNumItems() {return _numItems;};
  int GetModified () {return _modified;};

  void SetBlock(const char *block) {if (_block) delete (_block);_block =(char *)malloc(strlen(block) +1);strcpy(_block,block);};  

  void SetName(const char *name) {if (_name) delete (_name);_name =(char *)malloc(strlen(name) +1); strcpy(_name, name);};   
  void SetId(const char *id) {if (_id) delete (_id);_id=(char *)malloc(strlen(id)+1); strcpy(_id,id);};
  void SetObjectId(const int objectId) { _objectId = objectId;};
  void SetMemId(const int memId) { _memId = memId;};
  void SetNumItems(const int numItems) { _numItems = numItems;};
  void SetModified (const int modified) { _modified = modified;};
  void DeleteElement();
  void Print();
  void Clear();
  void operator=(const TableInfo &t);
  int operator<(const TableInfo &t);
  int operator==(const TableInfo &t);
};

class BlockInfo {
  friend class TblFileObj;
 protected:
  char * _name;   // internal block name
  int _caseSense; // indicates case sensitivity of the block name

 public:
  BlockInfo(int caseSense = CASE_SENSITIVE);
  BlockInfo(const char *c, int caseSense = CASE_SENSITIVE);
  BlockInfo(const BlockInfo &t);
  ~BlockInfo() ;
  char * GetName() {return _name;}; 

  void SetName(const char *name) {if (_name) delete (_name);_name =(char *)malloc(strlen(name) +1); strcpy(_name, name);};
  void DeleteElement();
  void Print();
  void Clear();
  void operator=(const BlockInfo &b);
  int operator<(const BlockInfo &b);
  int operator==(const BlockInfo &b);
};


class TblFileObj {
 protected:
  static const int TABLES_INCR;// = 250; // the increment of allocated tables
  static const int BLOCKS_INCR;// = 100; // the increment of allocated data blocks

  FileNavigator * _f;
  char * _filename;
  int    _mode;
  char * _currentBlock;


  struct _fileIndexS {
    uWord numBlocks;             // the number of blocks
    uWord numTables;             // the number of table  object names
    uWord tablesAllocated;       // the number of tables allocated
    BTreeObj<TableInfo> tables;
    BTreeObj<BlockInfo> blocks;
  } _fileIndex;

  ISTable ** _tableObjs;
  int _lastTableInMem;       // index of the last table object in memory
  int _tablesAllocated;

  int _caseSense; // indicates case sensitivity of the table ids

  char *_BuildTableId(const char *blockName, const char *tableName, int flagBlockExternal=1);
  int _IsTablePresent(const char * blockName, const char * tableName);
  int _IsBlockPresent(const char * blockName);

  char **_GetTableIds(int & num);

  void _ManageTableSpace();

  int _GetIndexFromId(const char * tableId);
  int _GetIndexFromName(const char *blockName, const char * tableName);

  int _ShrinkFile(char *filename);

  void _AddDataBlock(const char *blockName);

  int  _ReadFileIndex();
  int  _WriteFileIndex();

  ISTable * _GetTablePtr(const char * tableId);
  ISTable * _GetTable(const char * tableId);
  int _DeleteTable(const char * tableId);

  char * _GetExternalBlockName(const char *blockId);
  char * _GetInternalBlockId(const char *blockName);
  char ** _GetTableBlockNames(int & num);
  int _WriteTable(ISTable * s, const char *blockName, const char *tableName, int saveTablePtr=1);

 public:
  static const int  MAKE_BIN;//      = 1;      // const for CloseFile
  static const int  DONT_MAKE_BIN;// = 0; // 

  TblFileObj(int caseSense = CASE_SENSITIVE) {
    Clear();
    _caseSense = caseSense;
  };

  TblFileObj(char * filename, int openMode, int caseSense = CASE_SENSITIVE);

  virtual ~TblFileObj() {
    CloseFile(0);
    Reset();
  }

  void Clear();
  void Reset();


  int  OpenFile(const char * filename, int openMode);
  void CloseFile(int makeBinFile = MAKE_BIN, int shrink = 0);

  int WriteTable(ISTable * s, const char *blockName, const char *tableName);
  inline int WriteTable(ISTable * s, const char *tableName){return(WriteTable(s,_currentBlock,tableName));};

  inline int IsBlockPresent(const char * blockName) {return(_IsBlockPresent(blockName));};

  CifString AddDataBlock(const char* blockName);

  inline int IsTablePresent(const char * blockName, const char * tableName) {
    return(_IsTablePresent(blockName, tableName));
  };

  inline int IsTablePresent(const char * tableName) {
    return(_IsTablePresent(_currentBlock, tableName));
  };

  ISTable * GetTablePtr(const char *blockName, const char * tableName);
  ISTable * GetTable(const char *blockName, const char * tableName);
  inline ISTable * GetTablePtr(const char * tableName){return(GetTablePtr(_currentBlock,tableName));};

  int  DeleteTable(const char *blockName, const char * tableName);
  inline int  DeleteTable(const char * tableName){return(DeleteTable(_currentBlock,tableName));};

  void PrintHeaderInfo();
  inline void DumpFile(void){if (_f) _f->DumpFile();};
  inline void PrintIndex(void){if (_f) _f->PrintIndex();};
  inline void PrintIndexPosition(int position){if (_f) _f->PrintIndexPosition(position);};

  char ** GetTableNames(int & num);
  char ** GetTableNames(const char *blockId, int & num);



  char ** GetBlockNames(int & num);
  inline int  GetNumBlocks(){ return _fileIndex.numBlocks;};
  char  * GetBlockName(int iOrgOrder);
  int GetBlockOrder(const char *blockName);
  int RenameBlock(const char *blockName, int iOrgOrder);
  void SetDataBlock(const char *blockId);
  inline const char *GetDataBlock(void) { return _currentBlock; };
  inline const char *GetFileName(void) { return _filename; };
  inline        int  GetFileMode(void) { return _mode; };
  inline int GetCaseSensitivity(void) {return _caseSense;};
};

#endif
