/*
FILE:     CifFileObjBase.C
*/
/*
VERSION:  1.700
*/
/*
DATE:     7/13/2005
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2005 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
/* 
  PURPOSE:    Base class for read/write cif files
*/

#include "CifFileObjBase.h"

//const int CifFileObjBase::TRUE = 1;
//const int CifFileObjBase::FALSE = 0;

CifFileObjBase::CifFileObjBase(const char * objFileName, int openMode, int maxLineLength, const char *nullString, int verbose, int caseSense) : TblFileObj(caseSense) {
  int ilen;
  Clear();
  _verbose = verbose;
  _maxCifLineLength = maxLineLength;
  if (nullString && (ilen=strlen(nullString)) > 0) {
    _nullString = new char[ilen+1];
    strcpy(_nullString,nullString);
  }
  OpenFile(objFileName, openMode);
}

void CifFileObjBase::Clear() {
  _nullString = NULL;
  _cifIn = NULL;
  _verbose=0;
}

void CifFileObjBase::Reset() {
  if (_nullString)      delete[] _nullString;
}




int CifFileObjBase::Write(const char *cifFileName, int sortFlag, int emptyTable) {
  int iret=0;
  ofstream cifo(cifFileName,ios::out|ios::trunc);
  iret =Write(cifo, sortFlag, emptyTable);
  cifo.close();
  return (iret);
}


int CifFileObjBase::Write(ostream &cifo, int sortFlag, int emptyTable) {
  int iret=0;
  if (sortFlag)
    iret = WriteSort(cifo, emptyTable);
  else
    iret = WriteUnSort(cifo, emptyTable);
  return (iret);
}


int CifFileObjBase::WriteSort(ostream &cifo, int emptyTable) {
  ISTable *tblP = NULL;
  
  ReVarCifArray<CifString> *colNames=NULL;
  ReVarCifArray<CifString> *rowValues=NULL;
  
  int ib, jt, ib0, i, j, k, l, linePos, numColumn, numRow, ilen;
  char *bId=NULL;
  int cwid, *cwidth=NULL;
  char *tid;
  int end=1;
  
  _fileIndex.blocks.GoFirst();
  ib=_fileIndex.blocks.Current();
  for (ib0=0; ib0 < (int)(_fileIndex.numBlocks); ib0++) {
    end=1;
    
    cifo << "data_" << (*_fileIndex.blocks.Data())[ib].GetName() << endl;
    
    tid=_BuildTableId((*_fileIndex.blocks.Data())[ib].GetName()," ");
    _fileIndex.tables.Seek(tid);
    if (tid) delete[] tid;
    jt=_fileIndex.tables.Current();
    bId = _GetExternalBlockName((*_fileIndex.tables.Data())[jt].GetBlock());
    while((!bId || strcmp(bId,(*_fileIndex.blocks.Data())[ib].GetName())==0)&&end!=-1){
      
      tblP = GetTablePtr(bId,(*_fileIndex.tables.Data())[jt].GetName());
      if (bId) delete[] bId; bId = NULL;
      numColumn = tblP->GetNumColumns();
      numRow    = tblP->GetNumRows();
      cifo << "# " << endl;
      if (numRow <= 0) {
	if (emptyTable) {
	  colNames  = tblP->GetColumnNames();
	  cwid = -1;
	  for (i=0; i< numColumn; i++) {
	    ilen = (*colNames)[i].Length();
	    if (  ilen > cwid) cwid=ilen;
	  }
	
	  for (i=0; i< numColumn; i++) {
	    _PrintItemName(cifo, (*_fileIndex.tables.Data())[jt].GetName(), &(*colNames)[i], linePos);
	    ilen = (*colNames)[i].Length();
	    for (k=0; k < 2+cwid-ilen; k++) cifo << " ";
	    linePos = cwid+2;
	  
	    cifo << _nullString; 
	    if (linePos != 0) cifo << endl;
	  }
	  if (colNames)  delete colNames;
	}
      }
      else if (numRow == 1) {
	colNames  = tblP->GetColumnNames();
	rowValues = tblP->GetRow(0);
	cwid = -1;
	for (i=0; i< numColumn; i++) {
	  ilen = (*colNames)[i].Length();
	  if (  ilen > cwid) cwid=ilen;
	}
	
	for (i=0; i< numColumn; i++) {
	  _PrintItemName(cifo, (*_fileIndex.tables.Data())[jt].GetName(), &(*colNames)[i], linePos);
	  ilen = (*colNames)[i].Length();
	  for (k=0; k < 2+cwid-ilen; k++) cifo << " ";
	  linePos = cwid+2;
	  
	  _PrintItemValue(cifo, &(*rowValues)[i], linePos);
	  if (linePos != 0) cifo << endl;
	}
	if (colNames)  delete colNames;
	if (rowValues) delete rowValues;
	
      } 
      else {
	
	cifo << "loop_" << endl;
	
	colNames  = tblP->GetColumnNames();
	for (i=0; i< numColumn; i++) {
	  _PrintItemName(cifo, (*_fileIndex.tables.Data())[jt].GetName(), &(*colNames)[i], linePos);
	  cifo << endl;
	}
	
	// ---
	cwidth  = new int[numColumn];
	
	for (i=0; i< numColumn; i++) { cwidth[i]=-1; }
	j = 0;
	l = 0;
	while (j < numRow) {
	  rowValues = tblP->GetRow(l);
	  if (rowValues != NULL) {
	    for (i=0; i< numColumn; i++) {
	      ilen = (*rowValues)[i].Length();
	      if ( _IsQuotableText(&(*rowValues)[i])) ilen+=2;
	      
	      if (  ilen > cwidth[i]) cwidth[i]=ilen;
	    }
	    if (rowValues) delete rowValues;
	    j++;
	  }
	  l++;
	}
	j = 0;
	l = 0;
	while (j < numRow) {
	  linePos = 0;
	  rowValues = tblP->GetRow(l);
	  if (rowValues != NULL) {
	    for (i=0; i< numColumn; i++) {
	      
	      ilen=_PrintItemValue(cifo, &(*rowValues)[i], linePos); 
	      if (linePos != 0) {
		if (linePos + cwidth[i]-ilen < _maxCifLineLength)
		  for (k=0; k < cwidth[i]-ilen; k++) cifo << " ";
		linePos +=    cwidth[i]-ilen;
	      }
	    }
	    if (linePos != 0) cifo << endl;
	    if (rowValues) delete rowValues;
	    j++;
	  }
	  l++;
	}
	if (cwidth) delete[] cwidth;
	if (colNames) delete colNames;
      }
      end=_fileIndex.tables.GoToNext();
      jt=_fileIndex.tables.Current();
      bId = _GetExternalBlockName((*_fileIndex.tables.Data())[jt].GetBlock());
    }
    cifo <<"# " << endl;
    _fileIndex.blocks.GoToNext();
    ib=_fileIndex.blocks.Current();
    if (bId) delete[] bId; bId = NULL;
  }
  return(1);
}



int CifFileObjBase::WriteUnSort(ostream &cifo, int emptyTable) {

ISTable *tblP = NULL;
  
  ReVarCifArray<CifString> *colNames=NULL;
  ReVarCifArray<CifString> *rowValues=NULL;
  
  int ib, jt, ib0, i, j, k, l, linePos, numColumn, numRow, ilen;
  char *bId=NULL;
  int cwid, *cwidth=NULL;
  int currIndex, currIndexend;
  int end;
  int irow;
  
  currIndex = 0;
  currIndexend = 0;
  for (ib0=0; ib0 < (int)(_fileIndex.numBlocks); ib0++) {
      ib = ib0;

#if DEBUG    
    cerr <<  "Writing data block " << (*_fileIndex.blocks.Data())[ib].GetName()<< "["<< ib << " of " 
	 << _fileIndex.numBlocks << "]"<< " Null is " << _nullString << endl;
#endif
    
    cifo << "data_" <<(*_fileIndex.blocks.Data())[ib].GetName() << endl;
    jt = currIndex;
    end = currIndexend;
    bId = _GetExternalBlockName((*_fileIndex.tables.Data())[jt].GetBlock());
    while((!bId || strcmp(bId,(*_fileIndex.blocks.Data())[ib].GetName())==0)&&end<(int)(_fileIndex.numTables)){
      if ((*_fileIndex.tables.Deleted())[jt] != 1){
	
	tblP = GetTablePtr(bId,(*_fileIndex.tables.Data())[jt].GetName());
	if (bId) delete[] bId; bId = NULL;
	numColumn = tblP->GetNumColumns();
	numRow    = tblP->GetNumRows();
#if DEBUG
	cerr << "Table " << (*_fileIndex.tables.Data())[jt].GetName()<< " Rows  " << numRow 
	     << " Columns " << numColumn << endl;
#endif
	cifo << "# " << endl;
	if (numRow <= 0) {
	  if (emptyTable) {
	    colNames  = tblP->GetColumnNames();
	    cwid = -1;
	    for (i=0; i< numColumn; i++) {
	      ilen = (*colNames)[i].Length();
	      if (  ilen > cwid) cwid=ilen;
	    }
	    
	    for (i=0; i< numColumn; i++) {
	      _PrintItemName(cifo, (*_fileIndex.tables.Data())[jt].GetName(), &(*colNames)[i], linePos);
	      ilen = (*colNames)[i].Length();
	      for (k=0; k < 2+cwid-ilen; k++) cifo << " ";
	      linePos = cwid+2;
	      
	      cifo << _nullString; 
	      if (linePos != 0) cifo << endl;
	    }
	    if (colNames)  delete colNames;
	  }
	}
	else if (numRow == 1) {
	  colNames  = tblP->GetColumnNames();
	  irow = 0;
	  while (tblP->IsDelete(irow))
	    irow++;
	  rowValues = tblP->GetRow(irow);
//	  rowValues = tblP->GetRow(0);
	  cwid = -1;
	  for (i=0; i< numColumn; i++) {
	    ilen = (*colNames)[i].Length();
	    if (  ilen > cwid) cwid=ilen;
	  }
	  
	  for (i=0; i< numColumn; i++) {
	    _PrintItemName(cifo, (*_fileIndex.tables.Data())[jt].GetName(), &(*colNames)[i], linePos);
	    ilen = (*colNames)[i].Length();
	    for (k=0; k < 2+cwid-ilen; k++) cifo << " ";
	    linePos = cwid+2;
	    
	    _PrintItemValue(cifo, &(*rowValues)[i], linePos);
	    if (linePos != 0) cifo << endl;
	  }
	  if (colNames)  delete colNames;
	  if (rowValues) delete rowValues;
	  
	} else {
	  
	  cifo << "loop_" << endl;
	  
	  colNames  = tblP->GetColumnNames();
	  for (i=0; i< numColumn; i++) {
	    _PrintItemName(cifo, (*_fileIndex.tables.Data())[jt].GetName(), &(*colNames)[i], linePos);
	    cifo << endl;
	  }
	  
	  // ---
	  cwidth  = new int[numColumn];
	  
	  for (i=0; i< numColumn; i++) { cwidth[i]=-1; }
	  j = 0;
	  l = 0;
	  while (j < numRow) {
	    rowValues = tblP->GetRow(l);
	    if (rowValues != NULL) {
	      for (i=0; i< numColumn; i++) {
		ilen = (*rowValues)[i].Length();
		if ( _IsQuotableText(&(*rowValues)[i])) ilen+=2;
		
		if (  ilen > cwidth[i]) cwidth[i]=ilen;
	      }
	      if (rowValues) delete rowValues;
	      j++;
	    }
	    l++;
	  }
#if DEBUG1
	  for (i=0; i< numColumn; i++) {
	    cerr << "Col " << i << " width " << cwidth[i] << endl;
	  }
#endif
	  j = 0;
	  l = 0;
	  while (j < numRow) {
	    linePos = 0;
	    rowValues = tblP->GetRow(l);
	    if (rowValues != NULL) {
	      for (i=0; i< numColumn; i++) {
	      
		ilen=_PrintItemValue(cifo, &(*rowValues)[i], linePos); 
#if DEBUG1
		cerr << "Row " << j <<  " Col " << i <<  " ilen " << ilen 
		     << " Length " << (*rowValues)[i].Length() 
		   << " linePos " << linePos 
		     << " Value ["<< (*rowValues)[i].Text() << "]" <<endl;
#endif
		if (linePos != 0) {
		  if (linePos + cwidth[i]-ilen < _maxCifLineLength)
		    for (k=0; k < cwidth[i]-ilen; k++) cifo << " ";
		  linePos +=    cwidth[i]-ilen;
		}
	      }
	      if (linePos != 0) cifo << endl;
	      if (rowValues) delete rowValues;
	      j++;
	    }
	    l++;
	  }
	  if (cwidth) delete[] cwidth;
	  if (colNames) delete colNames;
	}
	end++;
      }
      else
	if (bId) delete[] bId; bId = NULL;
      jt++;
      if (jt<(int)((*_fileIndex.tables.Data()).Length()))
	bId = _GetExternalBlockName((*_fileIndex.tables.Data())[jt].GetBlock());
    }
    cifo <<"# " << endl;
    jt++;
    if (bId) delete[] bId; bId = NULL;
    currIndex=jt-1;
    currIndexend = end;
  }
  return(1);
}


  
int CifFileObjBase::Write(const char *cifFileName, ReVarCifArray<CifString>& catOrder, int emptyTable) {
  int iret=0;
  ofstream cifo(cifFileName,ios::out|ios::trunc);
  iret =Write(cifo, catOrder, emptyTable);
  cifo.close();
  return (iret);
}

int CifFileObjBase::Write(ostream &cifo, ReVarCifArray<CifString>& catOrder,int emptyTable) {
  ISTable *tblP = NULL;
  
  ReVarCifArray<CifString> *colNames=NULL;
  ReVarCifArray<CifString> *rowValues=NULL;
  int currIndex, currIndexend;
  
  int ib, jt, ib0, i, j, k, m, linePos, numColumn, numRow, ilen;
  int irow;
  char *bId=NULL;
  int cwid, *cwidth=NULL;
  char *tid;
  int end=1;
  currIndex = 0;
  currIndexend = 0;

  int orderLen = catOrder.Length();
  int* order = (int*) calloc(sizeof(int),_fileIndex.numTables);
  int found;
  
  _fileIndex.blocks.GoFirst();
  ib=_fileIndex.blocks.Current();
  for (ib0=0; ib0 < (int)(_fileIndex.numBlocks); ib0++) {
    end=1;
    
    cifo << "data_" << (*_fileIndex.blocks.Data())[ib].GetName() << endl;
    
    for (int l=0; l<orderLen; l++) {
      tid=_BuildTableId((*_fileIndex.blocks.Data())[ib].GetName(),catOrder[l].Text());
      found = _fileIndex.tables.Seek(tid);
      if (tid) delete[] tid;
      if (found>=0) {
	jt=_fileIndex.tables.Current();
	order[jt] = 1;
	bId = _GetExternalBlockName((*_fileIndex.tables.Data())[jt].GetBlock());
//**********************
      tblP = GetTablePtr(bId,(*_fileIndex.tables.Data())[jt].GetName());
      if (bId) delete[] bId; bId = NULL;
      numColumn = tblP->GetNumColumns();
      numRow    = tblP->GetNumRows();
      cifo << "# " << endl;
      if (numRow <= 0) {
	if (emptyTable) {
	  colNames  = tblP->GetColumnNames();
	  cwid = -1;
	  for (i=0; i< numColumn; i++) {
	    ilen = (*colNames)[i].Length();
	    if (  ilen > cwid) cwid=ilen;
	  }
	
	  for (i=0; i< numColumn; i++) {
	    _PrintItemName(cifo, (*_fileIndex.tables.Data())[jt].GetName(), &(*colNames)[i], linePos);
	    ilen = (*colNames)[i].Length();
	    for (k=0; k < 2+cwid-ilen; k++) cifo << " ";
	    linePos = cwid+2;
	  
	    cifo << _nullString; 
	    if (linePos != 0) cifo << endl;
	  }
	  if (colNames)  delete colNames;
	}
      }
      else if (numRow == 1) {
	colNames  = tblP->GetColumnNames();
	  irow = 0;
	  while (tblP->IsDelete(irow))
	    irow++;
	  rowValues = tblP->GetRow(irow);
//	rowValues = tblP->GetRow(0);
	cwid = -1;
	for (i=0; i< numColumn; i++) {
	  ilen = (*colNames)[i].Length();
	  if (  ilen > cwid) cwid=ilen;
	}
	
	for (i=0; i< numColumn; i++) {
	  _PrintItemName(cifo, (*_fileIndex.tables.Data())[jt].GetName(), &(*colNames)[i], linePos);
	  ilen = (*colNames)[i].Length();
	  for (k=0; k < 2+cwid-ilen; k++) cifo << " ";
	  linePos = cwid+2;
	  
	  _PrintItemValue(cifo, &(*rowValues)[i], linePos);
	  if (linePos != 0) cifo << endl;
	}
	if (colNames)  delete colNames;
	if (rowValues) delete rowValues;
	
      } 
      else {
	
	cifo << "loop_" << endl;
	
	colNames  = tblP->GetColumnNames();
	for (i=0; i< numColumn; i++) {
	  _PrintItemName(cifo, (*_fileIndex.tables.Data())[jt].GetName(), &(*colNames)[i], linePos);
	  cifo << endl;
	}
	
	// ---
	cwidth  = new int[numColumn];
	
	for (i=0; i< numColumn; i++) { cwidth[i]=-1; }
	j = 0;
	m = 0;
	while (j < numRow) {
	  rowValues = tblP->GetRow(m);
	  if (rowValues != NULL) {
	    for (i=0; i< numColumn; i++) {
	      ilen = (*rowValues)[i].Length();
	      if ( _IsQuotableText(&(*rowValues)[i])) ilen+=2;
	      
	      if (  ilen > cwidth[i]) cwidth[i]=ilen;
	    }
	    if (rowValues) delete rowValues;
	    j++;
	  }
	  m++;
	}
	j = 0;
	m = 0;
	while (j < numRow) {
	  linePos = 0;
	  rowValues = tblP->GetRow(m);
	  if (rowValues != NULL) {
	    for (i=0; i< numColumn; i++) {
	      
	      ilen=_PrintItemValue(cifo, &(*rowValues)[i], linePos); 
	      if (linePos != 0) {
		if (linePos + cwidth[i]-ilen < _maxCifLineLength)
		  for (k=0; k < cwidth[i]-ilen; k++) cifo << " ";
		linePos +=    cwidth[i]-ilen;
	      }
	    }
	    if (linePos != 0) cifo << endl;
	    if (rowValues) delete rowValues;
	    j++;
	  }
	  m++;
	}
	if (cwidth) delete[] cwidth;
	if (colNames) delete colNames;
      }
//**************************	
      }
    }
    tid=_BuildTableId((*_fileIndex.blocks.Data())[ib].GetName()," ");
    _fileIndex.tables.Seek(tid);
    if (tid) delete[] tid;
    jt = currIndex;
    end = currIndexend;
    bId = _GetExternalBlockName((*_fileIndex.tables.Data())[jt].GetBlock());
    while((!bId || strcmp(bId,(*_fileIndex.blocks.Data())[ib].GetName())==0)&&end<(int)(_fileIndex.numTables)){
      if (order[jt]==0){
      tblP = GetTablePtr(bId,(*_fileIndex.tables.Data())[jt].GetName());
      if (bId) delete[] bId; bId = NULL;
      numColumn = tblP->GetNumColumns();
      numRow    = tblP->GetNumRows();
      cifo << "# " << endl;
      if (numRow <= 0) {
	if (emptyTable) {
	  colNames  = tblP->GetColumnNames();
	  cwid = -1;
	  for (i=0; i< numColumn; i++) {
	    ilen = (*colNames)[i].Length();
	    if (  ilen > cwid) cwid=ilen;
	  }
	
	  for (i=0; i< numColumn; i++) {
	    _PrintItemName(cifo, (*_fileIndex.tables.Data())[jt].GetName(), &(*colNames)[i], linePos);
	    ilen = (*colNames)[i].Length();
	    for (k=0; k < 2+cwid-ilen; k++) cifo << " ";
	    linePos = cwid+2;
	  
	    cifo << _nullString; 
	    if (linePos != 0) cifo << endl;
	  }
	  if (colNames)  delete colNames;
	}
      }
      else if (numRow == 1) {
	colNames  = tblP->GetColumnNames();
	  irow = 0;
	  while (tblP->IsDelete(irow))
	    irow++;
	  rowValues = tblP->GetRow(irow);
//	rowValues = tblP->GetRow(0);
	cwid = -1;
	for (i=0; i< numColumn; i++) {
	  ilen = (*colNames)[i].Length();
	  if (  ilen > cwid) cwid=ilen;
	}
	
	for (i=0; i< numColumn; i++) {
	  _PrintItemName(cifo, (*_fileIndex.tables.Data())[jt].GetName(), &(*colNames)[i], linePos);
	  ilen = (*colNames)[i].Length();
	  for (k=0; k < 2+cwid-ilen; k++) cifo << " ";
	  linePos = cwid+2;
	  
	  _PrintItemValue(cifo, &(*rowValues)[i], linePos);
	  if (linePos != 0) cifo << endl;
	}
	if (colNames)  delete colNames;
	if (rowValues) delete rowValues;
	
      } 
      else {
	
	cifo << "loop_" << endl;
	
	colNames  = tblP->GetColumnNames();
	for (i=0; i< numColumn; i++) {
	  _PrintItemName(cifo, (*_fileIndex.tables.Data())[jt].GetName(), &(*colNames)[i], linePos);
	  cifo << endl;
	}
	
	// ---
	cwidth  = new int[numColumn];
	
	for (i=0; i< numColumn; i++) { cwidth[i]=-1; }
	j = 0;
	m = 0;
	while (j < numRow) {
	  rowValues = tblP->GetRow(m);
	  if (rowValues != NULL) {
	    for (i=0; i< numColumn; i++) {
	      ilen = (*rowValues)[i].Length();
	      if ( _IsQuotableText(&(*rowValues)[i])) ilen+=2;
	      
	      if (  ilen > cwidth[i]) cwidth[i]=ilen;
	    }
	    if (rowValues) delete rowValues;
	    j++;
	  }
	  m++;
	}
	j = 0;
	m = 0;
	while (j < numRow) {
	  linePos = 0;
	  rowValues = tblP->GetRow(m);
	  if (rowValues != NULL) {
	    for (i=0; i< numColumn; i++) {
	      
	      ilen=_PrintItemValue(cifo, &(*rowValues)[i], linePos); 
	      if (linePos != 0) {
		if (linePos + cwidth[i]-ilen < _maxCifLineLength)
		for (k=0; k < cwidth[i]-ilen; k++) cifo << " ";
		linePos +=    cwidth[i]-ilen;
	      }
	    }
	    if (linePos != 0) cifo << endl;
	    if (rowValues) delete rowValues;
	    j++;
	  }
	  m++;
	}
	if (cwidth) delete[] cwidth;
	if (colNames) delete colNames;
      }
      }
      end++;
      jt++;
      if (jt<(int)((*_fileIndex.tables.Data()).Length()))
        bId = _GetExternalBlockName((*_fileIndex.tables.Data())[jt].GetBlock());
    }
    cifo <<"# " << endl;
    jt++;
    currIndex=jt-1;
    currIndexend = end;
    _fileIndex.blocks.GoToNext();
    ib=_fileIndex.blocks.Current();
    if (bId) delete[] bId; bId = NULL;
  }
  return(1);
}





void CifFileObjBase::_PrintItemName(ostream& cifo, const char *category, CifString *keyword, int& linePos)
{
  cifo << "_" << category << "." << keyword->Text() << " ";
  linePos = keyword->Length() + strlen(category) + 3;
}

void CifFileObjBase::_PrintItemNameTab(ostream& cifo, const char *category, CifString *keyword, int& linePos)
{
  int i;
  int start = 36;
  cifo << "    ";
  cifo << "_" << category << "." << keyword->Text() << " ";
  linePos = 4 + keyword->Length() + strlen(category) + 3;
  while (linePos>start)
    start = start + 10;
  for (i=0; i<start-linePos; i++)
    cifo<<" ";
  linePos=start;
}

int CifFileObjBase::_IsQuotableText(CifString  *itemValue) 
{
  if (!itemValue || (itemValue->Length() < 1)) return 0;

  if (itemValue->Text()[0] == '_')   return 1;

  for (int i=0; i< (int)(itemValue->Length()); i++) {
    if (itemValue->Text()[i] == ' ') 
      return 1;
    else if (itemValue->Text()[i] == '\n') 
      return 1;
    else if (itemValue->Text()[i] == '\'' || itemValue->Text()[i] == '\"') 
      return 1;

  }
  if (!strncasecmp(itemValue->Text(),"data_",5)) return 1;
  if (!strncasecmp(itemValue->Text(),"loop_",5)) return 1;
  if (!strncasecmp(itemValue->Text(),"save_",5)) return 1;
  return 0;
}


int CifFileObjBase::_PrintItemValue(ostream& cifo, CifString  *itemValue, int& linePos) 
{
  int i, str_len, multipleLine, multipleWord, embeddedQuotes;
  
  /* NULL VALUES */
  if (itemValue->IsNull() || (itemValue->Length() == 0)) { 
    if (linePos + 2  < _maxCifLineLength) {
      linePos += 2;
      cifo << _nullString << " ";
    } else {
      cifo << endl << _nullString << " ";
      linePos = 2;
    }
    return 1;
  }
    
  str_len = itemValue->Length();
  multipleLine = FALSE;
  multipleWord = FALSE;
  embeddedQuotes = FALSE;


  for (i=0; i< str_len; i++) {
    if (itemValue->Text()[i] == ' ') 
      multipleWord = TRUE;
    else if (itemValue->Text()[i] == '\n') 
      multipleLine =TRUE;
    else if (itemValue->Text()[i] == '\'' || itemValue->Text()[i] == '\"') 
      embeddedQuotes =TRUE;
    
  }
  if (itemValue->Text()[0] == '_')    multipleWord = TRUE;
  if (itemValue->Text()[0] == ';')    multipleWord = TRUE;

  if (!strncasecmp(itemValue->Text(),"data_",5))    multipleWord = TRUE;
  if (!strncasecmp(itemValue->Text(),"loop_",5))    multipleWord = TRUE;
  if (!strncasecmp(itemValue->Text(),"save_",5))    multipleWord = TRUE;

  
  if (embeddedQuotes && multipleWord) multipleLine = TRUE;
  if (embeddedQuotes)                 multipleWord = TRUE;

  
  if (str_len >= _maxCifLineLength || multipleLine) {
    if (linePos != 0)  cifo << endl;
    cifo << ";" << itemValue->Text();
    cifo << endl;
    cifo << ";" << endl;
    linePos = 0;
  }  else {
    if ((!multipleWord && str_len + 2 + linePos    > _maxCifLineLength) ||
	 (multipleWord && str_len + 4 + linePos    > _maxCifLineLength))  {
      cifo << endl; linePos = 0;
    }
    
    if (multipleWord) {
      cifo << "\'" << itemValue->Text() << "\' ";
      linePos += (str_len + 3);
    } else {
      cifo << itemValue->Text() << " ";
      linePos += str_len + 1;
    }
  }
  
  if (multipleWord || multipleLine || embeddedQuotes) 
    return (str_len+2);
  else
    return (str_len);
  
}


int CifFileObjBase::_PrintItemValueTab(ostream& cifo, CifString  *itemValue, int& linePos,int width) 
{
  int i, str_len, multipleLine, multipleWord, embeddedQuotes;
  int startcol=10;
  int N=0;

  if (linePos==0) {
    cifo << "          ";
    linePos = startcol;
  }
  width+=2;
  
  /* NULL VALUES */
  if (itemValue->IsNull() || (itemValue->Length() == 0)) { 
    if (linePos + 2  < _maxCifLineLength) {
      linePos += 2;
      N += 2;
      cifo << _nullString << " ";
    } else {
      cifo << endl << _nullString << " ";
      linePos = 2;
      N = 2;
    }
    return 1;
  }
  
  str_len = itemValue->Length();
  multipleLine = FALSE;
  multipleWord = FALSE;
  embeddedQuotes = FALSE;


  for (i=0; i< str_len; i++) {
    if (itemValue->Text()[i] == ' ') 
      multipleWord = TRUE;
    else if (itemValue->Text()[i] == '\n') 
      multipleLine =TRUE;
    else if (itemValue->Text()[i] == '\'' || itemValue->Text()[i] == '\"') 
      embeddedQuotes =TRUE;
  }
  if (itemValue->Text()[0] == '_')    multipleWord = TRUE;
  if (itemValue->Text()[0] == ';')    multipleWord = TRUE;

  if (!strncasecmp(itemValue->Text(),"data_",5))    multipleWord = TRUE;
  if (!strncasecmp(itemValue->Text(),"loop_",5))    multipleWord = TRUE;
  if (!strncasecmp(itemValue->Text(),"save_",5))    multipleWord = TRUE;

  
  if (embeddedQuotes && multipleWord) multipleLine = TRUE;
  if (embeddedQuotes)                 multipleWord = TRUE;

  
  if (str_len >= _maxCifLineLength || multipleLine) {
    if (linePos != 0)  cifo << endl;
    cifo << ";" << itemValue->Text();
    cifo << endl;
    cifo << ";" << endl;
    linePos = 0;
  }  else {

    if ((!multipleWord && str_len + 2 + linePos    > _maxCifLineLength) ||
	 (multipleWord && str_len + 4 + linePos    > _maxCifLineLength))  {
      cifo << endl; linePos = 0;
  if (linePos==0) {
    cifo << "          ";
    linePos = startcol;
  }
    }

    
    if (multipleWord) {
      cifo << "\'" << itemValue->Text() << "\' ";
      linePos += (str_len + 3);
      N += (str_len + 3);
    } else {
      cifo << itemValue->Text() << " ";
      linePos += str_len + 1;
      N += str_len + 1;
    }
  }
  
  if(linePos>startcol) {
    linePos=linePos+width-N;
    for (i=0; i<width-N; i++)
      cifo << " ";
  }

  if (multipleWord || multipleLine || embeddedQuotes) 
    return (str_len+2);
  else
    return (str_len);
  
}


int CifFileObjBase:: GetKeywordFromItemName(char keyword[MAX_CIF_NAME_LENGTH], const char *itemName) {
/* ----------------------------------------------------------------------
   Purpose: CifFileObjBase::GetKeywordFromItemName(keyword, itemName) 
   
            Get the keyword part of an item name (ie. _<category>.<keyword>)
	    Return 1 for success or 0 otherwise.

 * ---------------------------------------------------------------------- */

  int i,j,k,ilen;


  strcpy(keyword,"");
  if (!itemName ||   (ilen=strlen(itemName)) < 3 ) return(0);

  for (k=1; k < ilen; k++) {
    if (itemName[k] == '.') break;
  }
  if ((k==1) || (k == ilen)) return(0);
  j=0;
  for (i=k+1; i < ilen; i++) {
    keyword[j] = itemName[i];
    j++;
  }
  keyword[j]='\0';
  return(1);

}


int CifFileObjBase::GetCategoryNameFromItemName(char categoryName[MAX_CIF_NAME_LENGTH], 
					   const char *itemName) 
/* ----------------------------------------------------------------------
   Purpose: CifFileObjBase::GetCategoryNameFromItemName(categoryName, itemName) 
   
            Get the category part of an item name (ie. _<category>.<keyword>)
	    Return 1 for success or 0 otherwise.

 * ---------------------------------------------------------------------- */
{
  int i,j;
  strcpy(categoryName,"");
  if (!itemName) return(0);
  
  for (i=1, j=0; i < (int)(strlen(itemName)); i++) {
    if (itemName[i] != '.') {
      categoryName[j] = itemName[i];
      j++;
    } else {
      break;
    }
  }
  categoryName[j]='\0';
  return(1);
}

/******* validation ******/

int CifFileObjBase::MandatoryCategoryChecking(CifFileObjBase *in, CifFileObjBase *ref,ofstream &log){
  int ret = 1;
  int i,j;
  char ** BlockNames;
  int numBlocks;
  ISTable * Category;
  ReVarCifArray<CifString> list;
  ReVarCifArray<CifString> target;
  ReVarPCifArray<int> * OutList;
  int mandatoryIndex;
  int errCode;
  char ** refBlockNames;
  int refnumBlocks;
  CifString cell;



  refBlockNames = ref->GetBlockNames(refnumBlocks);
  if (refnumBlocks>1) log<<"WARNING - dictionary has "<<refnumBlocks<<" datablocks"<<endl;
  Category = ref->GetTablePtr(refBlockNames[0],"category");
  list.Clear();
  list.Add("mandatory_code");
  Category->CreateIndex("index0",list);
  target.Clear();
  target.Add("yes");
  OutList = Category->Search(target,list,errCode);
  if (errCode>=0) {
    mandatoryIndex = Category->GetColumnIndex("id");
    
    BlockNames = in->GetBlockNames(numBlocks);
    
    for (i=0; i<(int)(OutList->Length()); i++) { //do for all tables that are mandatory
      Category->GetCell(cell,mandatoryIndex,(*OutList)[i]);
      for (j=0; j<numBlocks; j++) { //do it for all datablocks
	if (in->IsTablePresent(BlockNames[j],cell.Text())==0) {
	  log<<"ERROR - table "<<cell.Text()<<" is mandatory, but is missing"<<endl;
	  ret = 0;
	}
      }
    }
    for (i=0; i<numBlocks; i++)
      if (BlockNames[i]) delete[] BlockNames[i];
    if (BlockNames) delete[] BlockNames;
  }
  if (OutList) delete (OutList);
    for (i=0; i<refnumBlocks; i++)
      if (refBlockNames[i]) delete[] refBlockNames[i];
    if (refBlockNames) delete[] refBlockNames;
  return ret;
}

// This method converts the string with embedded escape sequences into
// the real string. For example, a two character combination "\t" is
// converted into one character representing horizontal tab.
void CifFileObjBase::ConvertEscapedString(CifString& inString,
  CifString& outString)
{

    // Temporary string will hold the converted string during the conversion
    // process. Afterwards, it will be copied to the output CIF string.
    char* tmpStringP = NULL;


    // Allocate memory for the temporary string. The allocated size is equal
    // to the size of the input string. After converting the escape sequences
    // its size will be less than the size of the input string.
    tmpStringP = (char*)malloc(inString.Length() + 1);

    // Initialize current character position in the temporary string.
    int tmpPos = 0;

    // Go through the characters of the input CIF string
    for (int posI = 0; posI < (int)(inString.Length()); tmpPos++)
    {
        if ((inString.Text())[posI] == '\\')
        {
            switch ((inString.Text())[posI + 1])
            {
                case '\\':
                  tmpStringP[tmpPos] = '\\';
                  posI += 2;
                  break;
                case 't':
                  tmpStringP[tmpPos] = '\t';
                  posI += 2;
                  break;
                case 'n':
                  tmpStringP[tmpPos] = '\n';
                  posI += 2;
                  break;
                default:
                  tmpStringP[tmpPos] = (inString.Text())[posI];
                  posI++;
                  break;
            }
        }
        else
        {
            tmpStringP[tmpPos] = (inString.Text())[posI];
            posI++;
        }
    } // for (characters in the input CIF string)

    // End the converted string with the null.
    tmpStringP[tmpPos] = '\0';

    // Copy the converted string to the output CIF string.
    outString.Copy(tmpStringP);

    // Free the temporary string
    free(tmpStringP);

} // CifFileObjBase::ConvertEscapedString()



int CifFileObjBase::ItemsChecking(CifFileObjBase *in, CifFileObjBase *ref,ofstream &log) {
  int ret = 1;
  int i,j,k,l,m;
  char ** BlockNames;
  int numBlocks;
  char ** TableNames;
  int numTables;
  ISTable * Category;
  ReVarCifArray<CifString> * ItemNames;
  ISTable * ParentTbl = NULL;
  ReVarCifArray<CifString> ParentList;  ReVarCifArray<CifString> ParentTarget;
  ReVarPCifArray<int> * OutList;
  int numRows;

  ISTable * ParentChildTbl;
  char ParentTblName[MAX_CIF_NAME_LENGTH];
  char keyword[MAX_CIF_NAME_LENGTH];
  int ParentIndex;
  CifString ParentName;
  ReVarCifArray<CifString> list;
  ReVarCifArray<CifString> target;
  int iOut, errCode;
  CifString cell;
  CifString typeCode;
  int len;
  char* p;
  char* pm;
  
#define NS 10
  regex_t preg;
  regmatch_t pmatch[NS];
  int icmp = 0;
  int errCodeRegex;
  CifString patternString;
/***************************/


  /************** PARENT-CHILD **************/
  // For parent-chaild relationship checking we need table "item_linked"
  // index on "child_name" column, after search value from "parent_name" column
  ParentChildTbl = ref->GetTablePtr("item_linked");
  list.Clear();
  list.Add("child_name");
  int errCodeP_CH;
  // Creates an index on column child_name, table item_linked ddl/dictionary
  ParentChildTbl->CreateIndex("index0",list);
  ParentIndex = ParentChildTbl->GetColumnIndex("parent_name");


  /************** MANDATORY **************/
  // For mandatory checking we need tables "item" from ref
  // Table ref->Item - index on columns "catgory_id"+"mandatory_code",
  // after search value from "name" column
  // then checking for particular item in ctegory
  ReVarCifArray<CifString> nameList;

  ISTable * refItem;
  ReVarCifArray<CifString> refItemList;
  ReVarCifArray<CifString> refItemTarget;
  int refItemIndex;

  ReVarCifArray<CifString> inItemList;
  ReVarCifArray<CifString> inItemTarget;

  nameList.Clear();
  nameList.Add("name");

  refItem = ref->GetTablePtr("item");
  refItemList.Clear();
  refItemList.Add("category_id");
  refItemList.Add("mandatory_code");
  refItem->CreateIndex("index0",refItemList);
  refItemIndex = refItem->GetColumnIndex("name");

  /************** REGEX **************/
  // For regular expresson checking we need table "item_type"
  // index on column "name", after search value from "code" column
  // and table "item_type_list" (to find reg. exp.)
  // index on column "code, after search value from "construct" column

  ISTable * ItemType,*ItemTypeL;
  ReVarCifArray<CifString> ItemTypeLList;
  ReVarCifArray<CifString> ItemTypeLTarget;
  int ItemTypeIndex, ItemTypeLIndex;

  ItemType = ref->GetTablePtr("item_type");
  ItemTypeIndex = ItemType->GetColumnIndex("code");

  ItemTypeL = ref->GetTablePtr("item_type_list");
  ItemTypeLList.Clear();
  ItemTypeLList.Add("code");
  ItemTypeLIndex = ItemTypeL->GetColumnIndex("construct");

  
  /************** RANGE **************/
  // For range checking we need table "item_range"
  // index on column "name", after search value from "maximum" & "minimum" columns
  ISTable * ItemRange;
  int ItemRangeIndex1, ItemRangeIndex2;
  int errCodeRange;
  char **maxlist = NULL;
  char **minlist = NULL;
  int irange = 0;
  int matched;
  double dval, dmin, dmax;
  int ival, imin, imax;

  ItemRange = ref->GetTablePtr("item_range");
  ItemRangeIndex1 = ItemRange->GetColumnIndex("maximum");
  ItemRangeIndex2 = ItemRange->GetColumnIndex("minimum");


  /************** ENUMERATION **************/
  // For enumeration checking we need table "item_enumeration"
  // index on column "name", after search value from "value" column
  ISTable * ItemEnumeration;
  int ItemEnumerationIndex;
  int errCodeEnumeration;
  char **enumlist = NULL;
  int ienum = 0;
  int ie;
  double de;

  ItemEnumeration = ref->GetTablePtr("item_enumeration");
  ItemEnumerationIndex = ItemEnumeration->GetColumnIndex("value");

  /************** REDUNDANT ROWS - KEY **************/
  // For renduntant rows checking we need table "category_key"
  // index on column "id", after search value from "name" column
  ISTable * CategoryKey;
  int KeyIndex;
  ReVarCifArray<CifString> keyTarget;
  ReVarCifArray<CifString> keyList;
  keyList.Clear();
  keyList.Add("id");

  CategoryKey = ref->GetTablePtr("category_key");
  KeyIndex = CategoryKey->GetColumnIndex("name");
  ReVarCifArray<CifString> colNamesKey;


  BlockNames = in->GetBlockNames(numBlocks);
  
  for (i=0; i<numBlocks; i++) { //do for all datablocks
    TableNames = in->GetTableNames(BlockNames[i],numTables);
    for (j=0; j<numTables; j++) { //do for all tables
      Category = in->GetTablePtr(BlockNames[i], TableNames[j]);
      /************** REDUNDANT ROWS - KEY **************/
      keyTarget.Clear();
      keyTarget.Add(TableNames[j]);
      OutList = CategoryKey->Search(keyTarget,keyList,errCode);
      if (errCode>=0) {
	colNamesKey.Clear();
	for (k=0; k<(int)(OutList->Length()); k++) {
	  CategoryKey->GetCell(cell,KeyIndex,(*OutList)[k]);
	  CifFileObjBase::GetKeywordFromItemName(keyword,cell.Text());
	  colNamesKey.Add(keyword);
	}
	Category->FindRedundantRows(colNamesKey,log,1);
      }
      /**************************************************/
      /*************** MANDATORY **************/
      refItemTarget.Clear();
      refItemTarget.Add(TableNames[j]);
      refItemTarget.Add("yes");
      OutList = refItem->Search(refItemTarget,refItemList,errCode);
      if (errCode>=0) {
	for (k=0; k<(int)(OutList->Length()); k++) {
	  refItem->GetCell(cell,refItemIndex,(*OutList)[k]);
	  CifFileObjBase::GetKeywordFromItemName(keyword,cell.Text());
	  errCode = Category->GetColumnIndex(keyword);
	  if (errCode<0) {
	    log<<"ERROR - Mandatory item "<<cell.Text()<<" is not in datablock " << BlockNames[i] <<endl;
	    ret = 0;
	  }
	}
      }
      if (OutList) delete OutList;
      /****************************************/
      
      ItemNames = Category->GetItemNames();
      for (k=0; k<(int)(ItemNames->Length()); k++) { //all items

	target.Clear();
	target.Add((*ItemNames)[k]);

	/************** PARENT-CHILD **************/

	// Checking if item has a parent
	iOut = ParentChildTbl->FindFirst(target,list,errCodeP_CH);
	if (errCodeP_CH>=0) {
	  // if YES checking for "parent" table & value in "child" item
	  ParentChildTbl->GetCell(ParentName,ParentIndex,iOut);
	  CifFileObjBase::GetCategoryNameFromItemName(ParentTblName,ParentName.Text());
	  if (in->IsTablePresent(BlockNames[i],ParentTblName)==0) {
	    //table for parent item does not exist
	    log <<"WARNING - Item "<<(*ItemNames)[k].Text()<<" has parent item "<< ParentName.Text()<<" but table "
		<< ParentTblName <<" is NOT present in the file"<<endl;
            // Set indication that the parent table does not exist. This will
            // prevent step 2 of parent-child relationship to be executed.
            errCodeP_CH = NOT_FOUND;
	  }
	  else {
            // Set indication that the parent table exists.
            errCodeP_CH = NO_TABLE_ERROR;
	    ParentTbl = in->GetTablePtr(BlockNames[i],ParentTblName);
	    CifFileObjBase::GetKeywordFromItemName(keyword,ParentName.Text());
	    ParentList.Clear();
	    ParentList.Add(keyword);
	    if(ParentTbl->FindIndex(keyword)==-1) { 
	      ParentTbl->CreateIndex(keyword,ParentList);
	    }
	  }
	}

	/************** REGEX step 1**************/
	/* compile a regex */
	iOut = ItemType->FindFirst(target,nameList,errCodeRegex);
	if (errCodeRegex>=0) {
	  ItemType->GetCell(typeCode,ItemTypeIndex,iOut);
	  ItemTypeLTarget.Clear();
	  ItemTypeLTarget.Add(typeCode);
	  iOut = ItemTypeL->FindFirst(ItemTypeLTarget,ItemTypeLList,errCode);
	  ItemTypeL->GetCell(cell,ItemTypeLIndex,iOut);
          ConvertEscapedString(cell, patternString); 
	  icmp = regcomp(&preg,patternString.Text(),REG_EXTENDED);
          regfree(&preg);
	}
	/****************************************/
	/************** RANGE step 1**************/
	/* finding min & max */
	OutList = ItemRange->Search(target,nameList,errCodeRange);
	if (errCodeRange>=0) {
	  maxlist = (char**) malloc(sizeof(char*)*(OutList->Length()+1));
	  minlist = (char**) malloc(sizeof(char*)*(OutList->Length()+1));
	  irange = OutList->Length();
	  for (l=0;l<irange;l++) {
	    ItemRange->GetCell(cell,ItemRangeIndex1,(*OutList)[l]);
	    maxlist[l] = (char*) malloc(strlen(cell.Text())+1);
	    strcpy(maxlist[l],cell.Text());
	    ItemRange->GetCell(cell,ItemRangeIndex2,(*OutList)[l]);
	    minlist[l] = (char*) malloc(strlen(cell.Text())+1);
	    strcpy(minlist[l],cell.Text());
	  }
	}
	if (OutList) delete (OutList);
	/****************************************/
	/************ ENUMERATION step 1************/
	OutList = ItemEnumeration->Search(target,nameList,errCodeEnumeration);
	if (errCodeEnumeration>=0) {
	  enumlist = (char**) malloc(sizeof(char*)*(OutList->Length()+1));
	  ienum = OutList->Length();
	  for (l=0;l<ienum;l++) {
	    ItemEnumeration->GetCell(cell,ItemEnumerationIndex,(*OutList)[l]);
	    enumlist[l] = (char*) malloc(strlen(cell.Text())+1);
	    strcpy(enumlist[l],cell.Text());
	  }
	}
	if (OutList) delete (OutList);
	/****************************************/
	// loop for all rows
	numRows = Category->GetNumRows();
	for (l=0; l<numRows ;l++){

	  Category->GetCell(cell,k,l);
	  /*************** PARENT-CHILD step 2**************/
	  if (errCodeP_CH==0) {
	    if ((cell!=".") && (cell!="?") && (cell!="")) {
	      ParentTarget.Clear();
	      ParentTarget.Add(cell);
	      ParentTbl->CheckValue(ParentTarget,ParentList,errCode);
	      if (errCode<0) {
		log<<"ERROR - Value "<< cell.Text() <<" for "<<(*ItemNames)[k].Text()
		   <<" is NOT defined in parent item "<< ParentName.Text() <<endl;
		ret = 0;
	      }
	    }
	  }
	  /****************************************/
	  
	  /*************** REGEX step 2**************/
	  if (errCodeRegex>=0) {
	    if (icmp==0 && (cell!=".") && (cell!="?") && (cell!="")) {
	      regcomp(&preg,patternString.Text(),REG_EXTENDED);
	      ret = regexec(&preg,cell.Text(),NS,pmatch,0);
              regfree(&preg);
	      if (ret!=0) {
		log<<"ERROR - value "<<cell.Text()<<" for "<<(*ItemNames)[k].Text()<<" does not match the pattern (row "<<l<<")"<<endl;
		ret = 0;
	      }
	      else { 
		len=pmatch[0].rm_eo - pmatch[0].rm_so;
		if (len!=(int)(strlen(cell.Text()))&& strcmp(typeCode.Text(),"text")!=0) { 
		  p = (char*)malloc(pmatch[0].rm_eo+1);
		  strncpy(p, cell.Text(), pmatch[0].rm_eo); 
		  p[pmatch[0].rm_eo]='\0';
		  pm=p+pmatch[0].rm_so;
		  log<<"ERROR - value "<<cell.Text()<<" for "<<(*ItemNames)[k].Text()<<" does not match the pattern (row "<<l<<")"<<endl;
		  log<<"ERROR - for "<<(*ItemNames)[k].Text()<<" only "<<pm<<" is matched (row "<<l <<")"<<endl;
		  ret = 0;
		  free (p);
		}
	      }
	    }
	    /****************************************/
	    /************** RANGE step 2**************/
	    //	    Range(cell.Text(),maxlist,minlist,
	    if (errCodeRange>=0&& (cell!=".") && (cell!="?") && (cell!="")) {
	  m=0;
	  matched = 0;
	  if (strcmp(typeCode.Text(),"float")==0) {
	    dval = atof (cell.Text());
	    while (m<irange && !matched) {
	      if (((strcmp(minlist[m],"?")!=0) && (strcmp(minlist[m],".")!=0)
                && (strcmp(minlist[m],"")!=0)) && ((strcmp(maxlist[m],"?")!=0)
                && (strcmp(maxlist[m],".")!=0) && (strcmp(maxlist[m],"")!=0))) {
		dmin = atof(minlist[m]);
		dmax = atof(maxlist[m]);
		if (dmin==dmax) {
		  if (dval==dmax) matched = 1;
		}
		else
		  if (dval > dmin && dval < dmax) matched = 1;
	      }
	      else {
		if ((strcmp(minlist[m],"?")!=0 ) && 
                  (strcmp(minlist[m],".")!=0 ) && 
                  (strcmp(minlist[m],"")!=0 )) {
		  dmin = atof(minlist[m]);
		  if (dval > dmin) matched = 1;
		}
		else {
		  if ((strcmp(maxlist[m],"?")!=0) && 
                    (strcmp(maxlist[m],".")!=0) && 
                    (strcmp(maxlist[m],"")!=0)) {
		    dmax = atof(maxlist[m]);
		    if (dval < dmax) matched = 1;
		  }
		  else
		    matched = 1;
		}
	      }
	      m++;
	    }
	    if (matched==0) {
	      log<<"ERROR - for "<<(*ItemNames)[k].Text()<<" "<<cell.Text()<<" is NOT in a range (row "<<l <<")"<<endl;
	      ret = 0;
	    }
	  }
	  else {
	    m=0;
	    matched = 0;
	    if (strcmp(typeCode.Text(),"int")==0) {
	      ival = atoi(cell.Text());
	      while (m<irange && !matched) {
	      if (((strcmp(minlist[m],"?")!=0) && (strcmp(minlist[m],".")!=0) 
                && (strcmp(minlist[m],"")!=0)) && ((strcmp(maxlist[m],"?")!=0)
                && (strcmp(maxlist[m],".")!=0) && (strcmp(maxlist[m],"")!=0)))
              {
		imin = atoi(minlist[m]);
		imax = atoi(maxlist[m]);
		if (imin==imax) {
		  if (ival==imin) matched = 1;
		}
		else
		  if (ival > imin && ival < imax) matched = 1;
	      }
	      else {
		if ((strcmp(minlist[m],"?")!=0) && 
                  (strcmp(minlist[m],".")!=0) && 
                  (strcmp(minlist[m],"")!=0)) {
		  imin = atoi(minlist[m]);
		  if (ival > imin) matched = 1;
		}
		else {
		  if ((strcmp(maxlist[m],"?")!=0) && 
                    (strcmp(maxlist[m],".")!=0) && 
                    (strcmp(maxlist[m],"")!=0)){
		    imax = atoi(maxlist[m]);
		    if (ival < imax) matched = 1;
		  }
		  else
		    matched = 1;
		}
	      }
		m++;
	      }
	      if (matched==0) {
	      log<<"ERROR - for "<<(*ItemNames)[k].Text()<<" "<<cell.Text()<<" is NOT in a range (row "<<l <<")"<<endl;
	      ret = 0;
	      }
	    }
	  }
	}
	  /****************************************/
	  /************ ENUMERATION step 2************/
	  if (errCodeEnumeration>=0 && (cell!=".") && (cell!="?") && (cell!="")) {
	    m=0;
	    matched = 0;
	    if (strcmp(typeCode.Text(),"float")==0) {
	      dval = atof(cell.Text());
	      while (m<ienum && !matched) {
		de = atof(enumlist[m]);
		if (dval == de) matched = 1;
		m++;
	      }
	    }
	    else {
	      if (strcmp(typeCode.Text(),"int")==0) {
		ival = atoi(cell.Text());
		while (m<ienum && !matched) {
		  ie = atoi(enumlist[m]);
		  if (ival == ie) matched = 1;
		  m++;
		}
	      }
	      else {
		while (m<ienum && !matched) {
		if (strcmp(cell.Text(),enumlist[m])==0) matched = 1;
		m++;
		}
	      }
	    }
	    if (matched==0) {
	      log<<"ERROR - for "<<(*ItemNames)[k].Text()<<" "<<cell.Text()<<" is NOT defined in the enumeration list (row "<<l <<")"<<endl;
	      ret = 0;
	    }
	  }
	  /****************************************/
	  }
	} //for all rows
	
	if (errCodeRange>=0) {
	  for (l=0;l<irange;l++) {
	    if (maxlist[l]) free (maxlist[l]);
	    if (minlist[l]) free (minlist[l]);
	  }
	  
	  if (maxlist) free (maxlist);
	  if (minlist) free (minlist);
	}
	if (errCodeEnumeration>=0) {
	  for (l=0;l<ienum;l++) {
	    if (enumlist[l]) free (enumlist[l]);
	  }
	  if (enumlist) free (enumlist);
	}
      
      } //for all items
      
      if (ItemNames) delete ItemNames;
    }
    for (j=0; j<numTables; j++)
      if (TableNames[j]) delete[] TableNames[j];
    if (TableNames) delete[] TableNames;
  }
  
  for (i=0; i<numBlocks; i++)
    if (BlockNames[i]) delete[] BlockNames[i];
  if (BlockNames) delete[] BlockNames;

  return ret;
}



int CifFileObjBase::DataChecking(CifFileObjBase *ref) {
  int ret1, ret2;
  ofstream log;
  log.open("Diagnostics.log",ios::out|ios::app);
  ret1 = MandatoryCategoryChecking(this,ref,log);
  ret2 = ItemsChecking(this,ref,log);
  log.close();
  if (ret1==0) 
    return ret2;
  else
    return ret1;
}
