/*
FILE:     DICFileObj.C
*/
/*
VERSION:  1.700
*/
/*
DATE:     7/13/2005
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2005 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
/* 
  PURPOSE:    Class for read/write dictionary
*/

#include "DICFileObj.h"
#include "DDLFileObj.h"

void DICFileObj::Clear() {
  _cifFileName = NULL;
  _parser = NULL;
}

void DICFileObj::Reset() {
  if (_cifFileName)  delete[] _cifFileName;
  if (_parser)       delete _parser;
}


int DICFileObj::Read(char *cifFileName,  TblFileObj * ddl,ISTable *&format,char * &diagnostics) {
  int ilen;
  if (_cifFileName) delete[] _cifFileName;
  _cifFileName = new char[strlen(cifFileName)+1];
  strcpy(_cifFileName, cifFileName);

  if ((_cifIn = fopen(_cifFileName, "r")) == NULL )
  {
      ndb_log_message(NDB_MSG_DEBUG4,(char*)"Unable to open file %s",_cifFileName);
      exit(1);
  }

  _parser = new DICParser(_cifIn, this,ddl,format,_verbose);
  _parser->Parse();
  if ((ilen = _parser->errorLog.Length()) > 0) {
    if (diagnostics) delete diagnostics;
    diagnostics = new char[ilen+1];
    strcpy(diagnostics,_parser->errorLog.Text());
  }
  fclose(_cifIn);
  if (_parser) delete _parser; _parser = NULL;
  return(0);
}




int DICFileObj:: Write(const char * cifFileName,  TblFileObj * ddl,ISTable * format) {
  int iret=0;
  ofstream cifo(cifFileName,ios::out|ios::trunc);
  iret =Write(cifo,ddl,format);
  cifo.close();
  return (iret);
}

int DICFileObj::Write(ostream &cifo,  TblFileObj * ddl,ISTable * format){

  ISTable *tblP = NULL;
  ISTable *cattbl = NULL;
  ISTable *cattbl2 = NULL;
  ISTable *itemtbl = NULL;
  ISTable *cattblddl = NULL;
  ISTable *itemtblddl = NULL;
  
  ReVarCifArray<CifString> *colNames=NULL;
  ReVarCifArray<CifString> *rowValues=NULL;
  
  int ib, ib0, i, j, k, l, m, linePos, numColumn, numRow, ilen;
  int i2,i3,i4;
  char *bId=NULL;
  int cwid, *cwidth=NULL;
  CifString categoryName;
  CifString categoryName2;
  CifString catName;
  CifString itemName;
  CifString itemName2;
  char item2[MAX_CIF_NAME_LENGTH];
  int errCode;
  char ** TableNames;
  char * BlockName;
  int num;
  
  ReVarCifArray<CifString> list;
  ReVarCifArray<CifString> list2;
  ReVarCifArray<CifString> listcat;
  ReVarCifArray<CifString> listcat2;
  ReVarCifArray<CifString> listitem;

  ReVarPCifArray<int> *listOut = NULL;
  ReVarPCifArray<int> *listOutcat = NULL;
  ReVarPCifArray<int> *listOutcat2 = NULL;
  ReVarPCifArray<int> *listOutItem = NULL;
  ReVarPCifArray<int> *listOutItem2 = NULL;

  ReVarCifArray<CifString> target;
  ReVarCifArray<CifString> target2;
  ReVarCifArray<CifString> targetcat;
  ReVarCifArray<CifString> targetcat2;
  ReVarCifArray<CifString> targetitem;

  int colIndex, colIndexCat;
  int colIndexItem = 0;

  list.Add("dbName");
  list.Add("type");
  colIndex=format->GetColumnIndex("catName");
  format->CreateIndex("index0",list);
  /* 
     This is loop for all datablocks - we have to assume that dictionary might have more then one datablock
  */
  for (ib0=0; ib0 < (int)(_fileIndex.numBlocks); ib0++) {
    ib = ib0;
    
#if DEBUG    
    cerr <<  "Writing data block " << (*_fileIndex.blocks.Data())[ib].GetName() << "["<< ib << " of " 
	 << _fileIndex.numBlocks << "]"<< " Null is " << _nullString << endl;
#endif
    
    cifo << "data_" <<(*_fileIndex.blocks.Data())[ib].GetName() << endl;
    /*
      Table format has information wich tables have information about datablock, category save frame and item save frame. This table is creatid during reading dictionary, and have to be input in this method.
    */
    target.Add((*_fileIndex.blocks.Data())[ib].GetName());
    target.Add("data");
    listOut=format->Search("index0",target,list,errCode);
    if (errCode!=NOT_FOUND) {
    for (l=0; l<(int)(listOut->Length()); l++) {
      format->GetCell(categoryName, colIndex, (*listOut)[l]);
      tblP = GetTablePtr((*_fileIndex.blocks.Data())[ib].GetName(),categoryName.Text());
      if (bId) delete[] bId; bId = NULL;
      numColumn = tblP->GetNumColumns();
      numRow    = tblP->GetNumRows();
      cifo << "# " << endl;
      if (numRow <= 0) {
	if (TRUE /*emptyTable*/) {
	colNames  = tblP->GetColumnNames();
	cwid = -1;
	for (i=0; i< numColumn; i++) {
	  ilen = (*colNames)[i].Length();
	  if (  ilen > cwid) cwid=ilen;
	}
	
	for (i=0; i< numColumn; i++) {
	  _PrintItemName(cifo, categoryName.Text(), &(*colNames)[i], linePos);
	  ilen = (*colNames)[i].Length();
	  for (k=0; k < 2+cwid-ilen; k++) cifo << " ";
	  linePos = cwid+2;
	  
	  cifo << _nullString; 
	  if (linePos != 0) cifo << endl;
	}
	if (colNames)  delete colNames;
	}
      }
      else if (numRow == 1) {
	colNames  = tblP->GetColumnNames();
	rowValues = tblP->GetRow(0);
	cwid = -1;
	for (i=0; i< numColumn; i++) {
	ilen = (*colNames)[i].Length();
	if (  ilen > cwid) cwid=ilen;
	}
      
	for (i=0; i< numColumn; i++) {
	  _PrintItemName(cifo, categoryName.Text(), &(*colNames)[i], linePos);
	  ilen = (*colNames)[i].Length();
	  for (k=0; k < 2+cwid-ilen; k++) cifo << " ";
	  linePos = cwid+2;
	  
	  _PrintItemValue(cifo, &(*rowValues)[i], linePos);
	if (linePos != 0) cifo << endl;
	}
	if (colNames)  delete colNames;
	if (rowValues) delete rowValues;
	
      }
      else {
	
	cifo << "loop_" << endl;
	
	colNames  = tblP->GetColumnNames();
	for (i=0; i< numColumn; i++) {
	  _PrintItemName(cifo, categoryName.Text(), &(*colNames)[i], linePos);
	  cifo << endl;
	}
	
	// ---
	cwidth  = new int[numColumn];
      
	for (i=0; i< numColumn; i++) { cwidth[i]=-1; }
	j = 0;
	m = 0;
	while (j < numRow) {
	  rowValues = tblP->GetRow(m);
	  if (rowValues != NULL) {
	    for (i=0; i< numColumn; i++) {
	      ilen = (*rowValues)[i].Length();
	      if ( _IsQuotableText(&(*rowValues)[i])) ilen+=2;
	      
	      if (  ilen > cwidth[i]) cwidth[i]=ilen;
	    }
	    if (rowValues) delete rowValues;
	    j++;
	  }
	  m++;
	}
	j = 0;
	m = 0;
	while (j < numRow) {
	  linePos = 0;
	  rowValues = tblP->GetRow(m);
	  if (rowValues != NULL) {
	    for (i=0; i< numColumn; i++) {
	      
	      ilen=_PrintItemValue(cifo, &(*rowValues)[i], linePos); 
	      if (linePos != 0) {
		if (linePos + cwidth[i]-ilen < _maxCifLineLength)
		  for (k=0; k < cwidth[i]-ilen; k++) cifo << " ";
		linePos +=    cwidth[i]-ilen;
	      }
	    }
	    if (linePos != 0) cifo << endl;
	    if (rowValues) delete rowValues;
	    j++;
	  }
	  m++;
	}
	if (cwidth) delete[] cwidth;
	if (colNames) delete colNames;
      }
    }
  }
    // category save frames

    // target for format-seaching category for category save frame
    target.Clear();
    target.Add((*_fileIndex.blocks.Data())[ib].GetName());
    target.Add("category");

    // target for format-seaching category for item save frame
    target2.Clear();
    target2.Add((*_fileIndex.blocks.Data())[ib].GetName());
    target2.Add("item");

    // Creates index on table item, in ddl and dictionary, on column named name
    // Assumtion that ddl has table item, and that table hase column name
    itemtbl=GetTablePtr((*_fileIndex.blocks.Data())[ib].GetName(),"item");
    BlockName = ddl->GetBlockName(0);
    itemtblddl=ddl->GetTablePtr(BlockName,"item");
    listitem.Clear();
    listitem.Add("name");
    itemtbl->CreateIndex("index0",listitem);
    itemtblddl->CreateIndex("index0",listitem);


    // Creates index on table category, in ddl and dictionary, on column 
    // named id and index on columns id+implicit_key
    // Assumtion that ddl has table category, and that table hase column id
    // and implicit_key
    cattbl=GetTablePtr((*_fileIndex.blocks.Data())[ib].GetName(),"category");
    cattblddl=ddl->GetTablePtr(BlockName,"category");
    if (BlockName) delete[] BlockName;
    listcat.Clear();
    listcat.Add("id");
    cattbl->CreateIndex("index0",listcat);
    cattblddl->CreateIndex("index0",listcat);
    listcat.Clear();
    listcat.Add("implicit_key");
    cattbl->CreateIndex("index1",listcat);
    cattblddl->CreateIndex("index1",listcat);
    colIndexCat=cattbl->GetColumnIndex("id");

    // creating indices for all tables on column with mandatory_code=implicit
    // assumption: table item has columns category_id and mandatory_code
    TableNames=GetTableNames((*_fileIndex.blocks.Data())[ib].GetName(), num);
    listitem.Clear();
    listitem.Add("category_id");
    listitem.Add("mandatory_code");
    for (l=0; l<num; l++) {
      targetitem.Clear();
      targetitem.Add(TableNames[l]);
      targetitem.Add("implicit");
      if (listOutItem) delete listOutItem;
      // in table item in ddl searchs rows with value <TableName> in column
      // category_id and value "implicit" in column "mandatory_code"
      listOutItem = itemtblddl->Search(targetitem,listitem,errCode);
      if (errCode != NOT_FOUND) {
	colIndexItem = itemtblddl->GetColumnIndex("name");
	itemtblddl->GetCell(itemName,colIndexItem,(*listOutItem)[0]);
	cattbl2=GetTablePtr((*_fileIndex.blocks.Data())[ib].GetName(),TableNames[l]);
	GetKeywordFromItemName(item2, itemName.Text());
	listcat2.Clear();
	listcat2.Add(item2);
	cattbl2->CreateIndex("indexi",listcat2);
      }
    }
    for (l=0; l<num; l++) 
      if (TableNames[l]) delete[] TableNames[l];
    if (TableNames) delete[] TableNames;

    // loop for all categories in current datablock; print save frame
    targetcat.Clear();
    targetcat.Add((*_fileIndex.blocks.Data())[ib].GetName());
    if (listOutcat) delete listOutcat;
    //searchs all tables that belong to current datablock
    listOutcat=cattbl->Search(targetcat,listcat,errCode);
    if (errCode != NOT_FOUND) {
    for (i2=0; i2<(int)(listOutcat->Length());i2++) {
      cattbl->GetCell(categoryName,colIndexCat,(*listOutcat)[i2]);
      catName.Copy(categoryName);
      catName.ToUpper();
      cifo << "save_" <<catName.Text() << endl;
      if (listOut) delete listOut;
      // searchs in table format all row with <datablock_name> in column 
      // "dbName" and "category" in column "type"
      // that will find category, category_examples....
      // in all this tables are information wich have to be placed in
      // save frame for category
      // the following loop is for all this tables
      listOut=format->Search(target,list,errCode);
      if (errCode != NOT_FOUND) {
      for (i3=0; i3<(int)(listOut->Length());i3++) {
	format->GetCell(categoryName2,colIndex,(*listOut)[i3]);
	cattbl2=GetTablePtr((*_fileIndex.blocks.Data())[ib].GetName(),categoryName2.Text());
	// implicit value for "category" is name of datablock for column
	// "imlicit_key" and name of save frame for column "id", for all
	// other tables implicit value is name of save frame
	// assumption: in table item in ddl in column named implicit
	// is defined wich colum has implicit value
	// for table category, implicit value "imlicit_key" is name of 
	// datablock, "id"= name of save frame
	if (strcasecmp(categoryName2.Text(),"category")!=0) {
	  targetitem.Clear();
	  targetitem.Add(categoryName2);
	  targetitem.Add("implicit");
	  if (listOutItem) delete listOutItem;
      // in table item in ddl searchs rows with value <categoryName2> in column
      // category_id and value "implicit" in column "mandatory_code"
	  listOutItem=itemtblddl->Search(targetitem,listitem,errCode);
	  if (errCode != NOT_FOUND) {
	    colIndexItem  = itemtblddl->GetColumnIndex("name");
	    itemtblddl->GetCell(itemName,colIndexItem,(*listOutItem)[0]);
	    GetKeywordFromItemName(item2, itemName.Text());
	  }
	}
	else
	  strcpy(item2,"id");
	listcat2.Clear();
	listcat2.Add(item2);
	
	targetcat2.Clear();
	targetcat2.Add(categoryName);
	if (listOutcat2) delete listOutcat2;
	listOutcat2=cattbl2->Search(targetcat2,listcat2,errCode);
	// writes values for particular category like ATOM_SITE,SYMMETRY,
	// for all tables that describe category like category, category_key,
	// category_group
	if (errCode != NOT_FOUND) {
	if (listOutcat2->Length() == 1) {
	  // pair
	    numColumn = cattbl2->GetNumColumns();
	    colNames  = cattbl2->GetColumnNames();
	    rowValues = cattbl2->GetRow((*listOutcat2)[0]);
	    cwid = -1;
	    for (i=0; i< numColumn; i++) {
	      ilen = (*colNames)[i].Length();
	      if (  ilen > cwid) cwid=ilen;
	    }
	    
	    for (i=0; i< numColumn; i++) {
	      _PrintItemNameTab(cifo, categoryName2.Text(), &(*colNames)[i], linePos);
	      _PrintItemValue(cifo, &(*rowValues)[i], linePos);
	      if (linePos != 0) cifo << endl;
	    }
	    if (colNames)  delete colNames;
	    if (rowValues) delete rowValues;
	}
	if (listOutcat2->Length() > 1) {
	  // loop
	  cifo << "     loop_" << endl;
	  numColumn = cattbl2->GetNumColumns();
	  colNames  = cattbl2->GetColumnNames();
	  for (i=0; i< numColumn; i++) {
	    _PrintItemNameTab(cifo, categoryName2.Text(), &(*colNames)[i], linePos);
	    cifo << endl;
	  }
	
	  // ---
	  cwidth  = new int[numColumn];
	  
	  for (i=0; i< numColumn; i++) { cwidth[i]=-1; }
	  for (j=0; j < (int)(listOutcat2->Length()); j++) {
	    rowValues = cattbl2->GetRow((*listOutcat2)[j]);
	    for (i=0; i< numColumn-1; i++) {
	      ilen = (*rowValues)[i].Length();
	      if ( _IsQuotableText(&(*rowValues)[i])) ilen+=2;
	      
	      if (  ilen > cwidth[i]) cwidth[i]=ilen;
	    }
	    cwidth[numColumn-1] = 0;
	    if (rowValues) delete rowValues;
	  }
	  for (j=0; j < (int)(listOutcat2->Length()); j++) {
	    linePos = 0;
	    rowValues = cattbl2->GetRow((*listOutcat2)[j]);
	    for (i=0; i< numColumn; i++) {
	      
	      ilen=_PrintItemValueTab(cifo, &(*rowValues)[i], linePos,cwidth[i]); 
	    }
	    if (linePos != 0) cifo << endl;
	    if (rowValues) delete rowValues;
	  }

	  if (cwidth) delete[] cwidth;
	  if (colNames) delete colNames;
	}
	}
      }
    }
      cifo << "     save_" <<endl;
      cifo <<endl;
      // loop for all items in the category
      targetitem.Clear();
      targetitem.Add(categoryName);
      if (listOutItem) delete listOutItem;
      // Searchs in the table "item" all items with "category_id"
      // equall to particular category
      listOutItem=itemtbl->Search(targetitem,listitem,errCode);
    if (errCode != NOT_FOUND) {
      for (i4=0; i4<(int)(listOutItem->Length()); i4++) {
	itemtbl->GetCell(itemName,colIndexItem,(*listOutItem)[i4]);
	cifo << "save__" <<itemName.Text() << endl;
	if (listOut) delete listOut;
	listOut=format->Search(target2,list,errCode);
	
    if (errCode != NOT_FOUND) {
      // writes values for particular item like atom_site.aniso_B[1][1]
      //,symmetry.entry_id
      // for all tables that describe item like item, item_description
      // item_default....
	for (i3=0; i3<(int)(listOut->Length());i3++) {
	  format->GetCell(categoryName2,colIndex,(*listOut)[i3]);
	  cattbl2=GetTablePtr((*_fileIndex.blocks.Data())[ib].GetName(),categoryName2.Text());
	  targetitem.Clear();
	  targetitem.Add(categoryName2);
	  targetitem.Add("implicit");
	  if (listOutItem2) delete listOutItem2;
	  // searchs in table item in ddl all rows with <categoryName2> in
	  // column "category_id" and "implicit" in column "mandatory_code"
	  // assumption : there is only one row with this characteristic
	  // in column name is name of implicit item for this category
	  // later this category will be searched to find
	  // particular item i column name found in this search
	  listOutItem2=itemtblddl->Search(targetitem,listitem,errCode);
	  if (errCode>=0) {
	  colIndexItem  = itemtblddl->GetColumnIndex("name");
	  itemtblddl->GetCell(itemName2,colIndexItem,(*listOutItem2)[0]);
	  GetKeywordFromItemName(item2, itemName2.Text());
	  listcat2.Clear();
	  listcat2.Add(item2);
	
	  targetcat2.Clear();
	  targetcat2.Add(itemName);
	  if (listOutcat2) delete listOutcat2;
	  // search one of table related to item
	  listOutcat2=cattbl2->Search(targetcat2,listcat2,errCode);
	  // here is part to write value(s) for thiscategory
	  if (errCode>=0) {
	  if (listOutcat2->Length() == 1) {
	    // pair
	    numColumn = cattbl2->GetNumColumns();
	    colNames  = cattbl2->GetColumnNames();
	    rowValues = cattbl2->GetRow((*listOutcat2)[0]);
	    cwid = -1;
	    for (i=0; i< numColumn; i++) {
	      ilen = (*colNames)[i].Length();
	      if (  ilen > cwid) cwid=ilen;
	    }
	    
	    for (i=0; i< numColumn; i++) {
	      linePos=8; //**************
	      _PrintItemNameTab(cifo, categoryName2.Text(), &(*colNames)[i], linePos);
	      _PrintItemValue(cifo, &(*rowValues)[i], linePos);
	      if (linePos != 0) cifo << endl;
	    }
	    if (colNames)  delete colNames;
	    if (rowValues) delete rowValues;
	  }
	  
	  if (listOutcat2->Length() > 1) {
	    // loop
	    cifo << "     loop_" << endl;
	    numColumn = cattbl2->GetNumColumns();
	    colNames  = cattbl2->GetColumnNames();
	    for (i=0; i< numColumn; i++) {
	      _PrintItemNameTab(cifo, categoryName2.Text(), &(*colNames)[i], linePos);
	      cifo << endl;
	    }
	    
	    // ---
	    cwidth  = new int[numColumn];
	    
	    for (i=0; i< numColumn; i++) { cwidth[i]=-1; }
	    for (j=0; j < (int)(listOutcat2->Length()); j++) {
	      rowValues = cattbl2->GetRow((*listOutcat2)[j]);
	      for (i=0; i< numColumn-1; i++) {
		ilen = (*rowValues)[i].Length();
		if ( _IsQuotableText(&(*rowValues)[i])) ilen+=2;
		
		if (  ilen > cwidth[i]) cwidth[i]=ilen;
	      }
	      cwidth[numColumn-1] = 0;
	      if (rowValues) delete rowValues;
	    }
	    for (j=0; j < (int)(listOutcat2->Length()); j++) {
	      linePos = 0;
	      rowValues = cattbl2->GetRow((*listOutcat2)[j]);
	      for (i=0; i< numColumn; i++) {
		
		ilen=_PrintItemValueTab(cifo, &(*rowValues)[i], linePos,cwidth[i]); 
	      }
	      if (linePos != 0) cifo << endl;
	      if (rowValues) delete rowValues;
	    }
	    
	    if (cwidth) delete[] cwidth;
	    if (colNames) delete colNames;
	  }
	  }
	  }
	}
      }
	cifo << "     save_" <<endl;
	cifo <<endl;
      }
    }
    }
  }

  }
  if (listOut) delete listOut;
  if (listOutcat) delete listOutcat;
  if (listOutcat2) delete listOutcat2;
  if (listOutItem) delete listOutItem;
  if (listOutItem2) delete listOutItem2;
  return(1); 
}

int DICFileObj:: Write_new_format(const char * cifFileName,  TblFileObj * ddl,ISTable * format) {
  int iret=0;
  ofstream cifo(cifFileName,ios::out|ios::trunc);
  iret =Write_new_format(cifo,ddl,format);
  cifo.close();
  return (iret);
}

int DICFileObj::Write_new_format(ostream &cifo,  TblFileObj * ddl,ISTable * format){

  ISTable *tblP = NULL;
  ISTable *cattbl = NULL;
  ISTable *cattbl2 = NULL;
  ISTable *itemtbl = NULL;
  ISTable *cattblddl = NULL;
  ISTable *itemtblddl = NULL;
  
  ReVarCifArray<CifString> *colNames=NULL;
  ReVarCifArray<CifString> *rowValues=NULL;
  
  int ib, ib0, i, j, k, l, m, linePos, numColumn, numRow, ilen;
  int i2,i3,i4;
  char *bId=NULL;
  int cwid, *cwidth=NULL;
  CifString categoryName;
  CifString categoryName2;
  CifString catName;
  CifString itemName;
  CifString itemName2;
  char item2[MAX_CIF_NAME_LENGTH];
  int errCode;
  char ** TableNames;
  char * BlockName;
  int num;
  
  ReVarCifArray<CifString> list;
  ReVarCifArray<CifString> list2;
  ReVarCifArray<CifString> listcat;
  ReVarCifArray<CifString> listcat2;
  ReVarCifArray<CifString> listitem;

  ReVarPCifArray<int> *listOut = NULL;
  ReVarPCifArray<int> *listOutcat = NULL;
  ReVarPCifArray<int> *listOutcat2 = NULL;
  ReVarPCifArray<int> *listOutItem = NULL;
  ReVarPCifArray<int> *listOutItem2 = NULL;

  ReVarCifArray<CifString> target;
  ReVarCifArray<CifString> target2;
  ReVarCifArray<CifString> targetcat;
  ReVarCifArray<CifString> targetcat2;
  ReVarCifArray<CifString> targetitem;

  int colIndex, colIndexCat;
  int colIndexItem = 0;

  list.Add("dbName");
  list.Add("type");
  colIndex=format->GetColumnIndex("catName");
  format->CreateIndex("index0",list);
  /* 
     This is loop for all datablocks - we have to assume that dictionary might have more then one datablock
  */
  for (ib0=0; ib0 < (int)(_fileIndex.numBlocks); ib0++) {
    ib = ib0;
    
#if DEBUG    
    cerr <<  "Writing data block " << (*_fileIndex.blocks.Data())[ib].GetName() << "["<< ib << " of " 
	 << _fileIndex.numBlocks << "]"<< " Null is " << _nullString << endl;
#endif
    
    cifo << "data_" <<(*_fileIndex.blocks.Data())[ib].GetName() << endl;
    /*
      Table format has information wich tables have information about datablock, category save frame and item save frame. This table is creatid during reading dictionary, and have to be input in this method.
    */
    target.Add((*_fileIndex.blocks.Data())[ib].GetName());
    target.Add("data");
    listOut=format->Search("index0",target,list,errCode);
    for (l=0; l<(int)(listOut->Length()); l++) {
      format->GetCell(categoryName, colIndex, (*listOut)[l]);
      tblP = GetTablePtr((*_fileIndex.blocks.Data())[ib].GetName(),categoryName.Text());
      if (bId) delete[] bId; bId = NULL;
      numColumn = tblP->GetNumColumns();
      numRow    = tblP->GetNumRows();
      cifo << "# " << endl;
      if (numRow <= 0) {
	if (TRUE /*emptyTable*/) {
	colNames  = tblP->GetColumnNames();
	cwid = -1;
	for (i=0; i< numColumn; i++) {
	  ilen = (*colNames)[i].Length();
	  if (  ilen > cwid) cwid=ilen;
	}
	
	for (i=0; i< numColumn; i++) {
	  _PrintItemName(cifo, categoryName.Text(), &(*colNames)[i], linePos);
	  ilen = (*colNames)[i].Length();
	  for (k=0; k < 2+cwid-ilen; k++) cifo << " ";
	  linePos = cwid+2;
	  
	  cifo << _nullString; 
	  if (linePos != 0) cifo << endl;
	}
	if (colNames)  delete colNames;
	}
      }
      else if (numRow == 1) {
	colNames  = tblP->GetColumnNames();
	rowValues = tblP->GetRow(0);
	cwid = -1;
	for (i=0; i< numColumn; i++) {
	ilen = (*colNames)[i].Length();
	if (  ilen > cwid) cwid=ilen;
	}
      
	for (i=0; i< numColumn; i++) {
	  _PrintItemName(cifo, categoryName.Text(), &(*colNames)[i], linePos);
	  ilen = (*colNames)[i].Length();
	  for (k=0; k < 2+cwid-ilen; k++) cifo << " ";
	  linePos = cwid+2;
	  
	  _PrintItemValue(cifo, &(*rowValues)[i], linePos);
	if (linePos != 0) cifo << endl;
	}
	if (colNames)  delete colNames;
	if (rowValues) delete rowValues;
	
      }
      else {
	
	cifo << "loop_" << endl;
	
	colNames  = tblP->GetColumnNames();
	for (i=0; i< numColumn; i++) {
	  _PrintItemName(cifo, categoryName.Text(), &(*colNames)[i], linePos);
	  cifo << endl;
	}
	
	// ---
	cwidth  = new int[numColumn];
      
	for (i=0; i< numColumn; i++) { cwidth[i]=-1; }
	j = 0;
	m = 0;
	while (j < numRow) {
	  rowValues = tblP->GetRow(m);
	  if (rowValues != NULL) {
	    for (i=0; i< numColumn; i++) {
	      ilen = (*rowValues)[i].Length();
	      if ( _IsQuotableText(&(*rowValues)[i])) ilen+=2;
	      
	      if (  ilen > cwidth[i]) cwidth[i]=ilen;
	    }
	    if (rowValues) delete rowValues;
	    j++;
	  }
	  m++;
	}
	j = 0;
	m = 0;
	while (j < numRow) {
	  linePos = 0;
	  rowValues = tblP->GetRow(m);
	  if (rowValues != NULL) {
	    for (i=0; i< numColumn; i++) {
	      
	      ilen=_PrintItemValue(cifo, &(*rowValues)[i], linePos); 
	      if (linePos != 0) {
		if (linePos + cwidth[i]-ilen < _maxCifLineLength)
		  for (k=0; k < cwidth[i]-ilen; k++) cifo << " ";
		linePos +=    cwidth[i]-ilen;
	      }
	    }
	    if (linePos != 0) cifo << endl;
	    if (rowValues) delete rowValues;
	    j++;
	  }
	  m++;
	}
	if (cwidth) delete[] cwidth;
	if (colNames) delete colNames;
      }
    }
    // category save frames

    // target for format-seaching category for category save frame
    target.Clear();
    target.Add((*_fileIndex.blocks.Data())[ib].GetName());
    target.Add("category");

    // target for format-seaching category for item save frame
    target2.Clear();
    target2.Add((*_fileIndex.blocks.Data())[ib].GetName());
    target2.Add("item");

    // Creates index on table item, in ddl and dictionary, on column named name
    // Assumtion that ddl has table item, and that table hase column name
    itemtbl=GetTablePtr((*_fileIndex.blocks.Data())[ib].GetName(),"item");
    BlockName = ddl->GetBlockName(0);
    itemtblddl=ddl->GetTablePtr(BlockName,"item");
    listitem.Clear();
    listitem.Add("name");
    itemtbl->CreateIndex("index0",listitem);
    itemtblddl->CreateIndex("index0",listitem);


    // Creates index on table category, in ddl and dictionary, on column 
    // named id and index on columns id+implicit_key
    // Assumtion that ddl has table category, and that table hase column id
    // and implicit_key
    cattbl=GetTablePtr((*_fileIndex.blocks.Data())[ib].GetName(),"category");
    cattblddl=ddl->GetTablePtr(BlockName,"category");
    if (BlockName) delete[] BlockName;
    listcat.Clear();
    listcat.Add("id");
    cattbl->CreateIndex("index0",listcat);
    cattblddl->CreateIndex("index0",listcat);
    listcat.Clear();
    listcat.Add("implicit_key");
    cattbl->CreateIndex("index1",listcat);
    cattblddl->CreateIndex("index1",listcat);
    colIndexCat=cattbl->GetColumnIndex("id");

    // creating indices for all tables on column with mandatory_code=implicit
    // assumption: table item has columns category_id and mandatory_code
    TableNames=GetTableNames((*_fileIndex.blocks.Data())[ib].GetName(), num);
    listitem.Clear();
    listitem.Add("category_id");
    listitem.Add("mandatory_code");
    for (l=0; l<num; l++) {
      targetitem.Clear();
      targetitem.Add(TableNames[l]);
      targetitem.Add("implicit");
      if (listOutItem) delete listOutItem;
      // in table item in ddl searchs rows with value <TableName> in column
      // category_id and value "implicit" in column "mandatory_code"
      listOutItem = itemtblddl->Search(targetitem,listitem,errCode);
      if (errCode != NOT_FOUND) {
	colIndexItem = itemtblddl->GetColumnIndex("name");
	itemtblddl->GetCell(itemName,colIndexItem,(*listOutItem)[0]);
	cattbl2=GetTablePtr((*_fileIndex.blocks.Data())[ib].GetName(),TableNames[l]);
	GetKeywordFromItemName(item2, itemName.Text());
	listcat2.Clear();
	listcat2.Add(item2);
	cattbl2->CreateIndex("indexi",listcat2);
      }
    }
    for (l=0; l<num; l++) 
      if (TableNames[l]) delete[] TableNames[l];
    if (TableNames) delete[] TableNames;

    // loop for all categories in current datablock; print save frame
    targetcat.Clear();
    targetcat.Add((*_fileIndex.blocks.Data())[ib].GetName());
    if (listOutcat) delete listOutcat;
    //searchs all tables that belong to current datablock
    listOutcat=cattbl->Search(targetcat,listcat,errCode);
    if (errCode != NOT_FOUND) {
    for (i2=0; i2<(int)(listOutcat->Length());i2++) {
      cattbl->GetCell(categoryName,colIndexCat,(*listOutcat)[i2]);
      catName.Copy(categoryName);
      catName.ToUpper();
      if (listOut) delete listOut;
      // searchs in table format all row with <datablock_name> in column 
      // "dbName" and "category" in column "type"
      // that will find category, category_examples....
      // in all this tables are information wich have to be placed in
      // save frame for category
      // the following loop is for all this tables
      listOut=format->Search(target,list,errCode);
      if (errCode != NOT_FOUND) {
      for (i3=0; i3<(int)(listOut->Length());i3++) {
	format->GetCell(categoryName2,colIndex,(*listOut)[i3]);
	cattbl2=GetTablePtr((*_fileIndex.blocks.Data())[ib].GetName(),categoryName2.Text());
	// implicit value for "category" is name of datablock for column
	// "imlicit_key" and name of save frame for column "id", for all
	// other tables implicit value is name of save frame
	// assumption: in table item in ddl in column named implicit
	// is defined wich colum has implicit value
	// for table category, implicit value "imlicit_key" is name of 
	// datablock, "id"= name of save frame
	if (strcasecmp(categoryName2.Text(),"category")!=0) {
	  targetitem.Clear();
	  targetitem.Add(categoryName2);
	  targetitem.Add("implicit");
	  if (listOutItem) delete listOutItem;
      // in table item in ddl searchs rows with value <categoryName2> in column
      // category_id and value "implicit" in column "mandatory_code"
	  listOutItem=itemtblddl->Search(targetitem,listitem,errCode);
	  if (errCode != NOT_FOUND) {
	    colIndexItem  = itemtblddl->GetColumnIndex("name");
	    itemtblddl->GetCell(itemName,colIndexItem,(*listOutItem)[0]);
	    GetKeywordFromItemName(item2, itemName.Text());
	  }
	}
	else
	  strcpy(item2,"id");
	listcat2.Clear();
	listcat2.Add(item2);
	
	targetcat2.Clear();
	targetcat2.Add(categoryName);
	if (listOutcat2) delete listOutcat2;
	listOutcat2=cattbl2->Search(targetcat2,listcat2,errCode);
	// writes values for particular category like ATOM_SITE,SYMMETRY,
	// for all tables that describe category like category, category_key,
	// category_group
	if (errCode != NOT_FOUND) {
	if (listOutcat2->Length() == 1) {
	  // pair
	    numColumn = cattbl2->GetNumColumns();
	    colNames  = cattbl2->GetColumnNames();
	    rowValues = cattbl2->GetRow((*listOutcat2)[0]);
	    cwid = -1;
	    for (i=0; i< numColumn; i++) {
	      ilen = (*colNames)[i].Length();
	      if (  ilen > cwid) cwid=ilen;
	    }
	    
	    for (i=0; i< numColumn; i++) {
	      _PrintItemNameTab(cifo, categoryName2.Text(), &(*colNames)[i], linePos);
	      _PrintItemValue(cifo, &(*rowValues)[i], linePos);
	      if (linePos != 0) cifo << endl;
	    }
	    if (colNames)  delete colNames;
	    if (rowValues) delete rowValues;
	}
	if (listOutcat2->Length() > 1) {
	  // loop
	  cifo << "     loop_" << endl;
	  numColumn = cattbl2->GetNumColumns();
	  colNames  = cattbl2->GetColumnNames();
	  for (i=0; i< numColumn; i++) {
	    _PrintItemNameTab(cifo, categoryName2.Text(), &(*colNames)[i], linePos);
	    cifo << endl;
	  }
	
	  // ---
	  cwidth  = new int[numColumn];
	  
	  for (i=0; i< numColumn; i++) { cwidth[i]=-1; }
	  for (j=0; j < (int)(listOutcat2->Length()); j++) {
	    rowValues = cattbl2->GetRow((*listOutcat2)[j]);
	    for (i=0; i< numColumn-1; i++) {
	      ilen = (*rowValues)[i].Length();
	      if ( _IsQuotableText(&(*rowValues)[i])) ilen+=2;
	      
	      if (  ilen > cwidth[i]) cwidth[i]=ilen;
	    }
	    cwidth[numColumn-1] = 0;
	    if (rowValues) delete rowValues;
	  }
	  for (j=0; j < (int)(listOutcat2->Length()); j++) {
	    linePos = 0;
	    rowValues = cattbl2->GetRow((*listOutcat2)[j]);
	    for (i=0; i< numColumn; i++) {
	      
	      ilen=_PrintItemValueTab(cifo, &(*rowValues)[i], linePos,cwidth[i]); 
	    }
	    if (linePos != 0) cifo << endl;
	    if (rowValues) delete rowValues;
	  }

	  if (cwidth) delete[] cwidth;
	  if (colNames) delete colNames;
	}
	}
      }
    }
      cifo <<endl;
      // loop for all items in the category
      targetitem.Clear();
      targetitem.Add(categoryName);
      if (listOutItem) delete listOutItem;
      // Searchs in the table "item" all items with "category_name"
      // equall to particular category
      listOutItem=itemtbl->Search(targetitem,listitem,errCode);
    if (errCode != NOT_FOUND) {
      for (i4=0; i4<(int)(listOutItem->Length()); i4++) {
	itemtbl->GetCell(itemName,colIndexItem,(*listOutItem)[i4]);
	if (listOut) delete listOut;
	listOut=format->Search(target2,list,errCode);
	
    if (errCode != NOT_FOUND) {
      // writes values for particular item like atom_site.aniso_B[1][1]
      //,symmetry.entry_id
      // for all tables that describe item like item, item_description
      // item_default....
	for (i3=0; i3<(int)(listOut->Length());i3++) {
	  format->GetCell(categoryName2,colIndex,(*listOut)[i3]);
	  cattbl2=GetTablePtr((*_fileIndex.blocks.Data())[ib].GetName(),categoryName2.Text());
	  targetitem.Clear();
	  targetitem.Add(categoryName2);
	  targetitem.Add("implicit");
	  if (listOutItem2) delete listOutItem2;
	  // searchs in table item in ddl all rows with <categoryName2> in
	  // column "category_id" and "implicit" in column "mandatory_code"
	  // assumption : there is only one row with this characteristic
	  // in column name is name of implicit item for this category
	  // later this category will be searched to find
	  // particular item i column name found in this search
	  listOutItem2=itemtblddl->Search(targetitem,listitem,errCode);
	  if (errCode>=0) {
	  colIndexItem  = itemtblddl->GetColumnIndex("name");
	  itemtblddl->GetCell(itemName2,colIndexItem,(*listOutItem2)[0]);
	  GetKeywordFromItemName(item2, itemName2.Text());
	  listcat2.Clear();
	  listcat2.Add(item2);
	
	  targetcat2.Clear();
	  targetcat2.Add(itemName);
	  if (listOutcat2) delete listOutcat2;
	  // search one of table related to item
	  listOutcat2=cattbl2->Search(targetcat2,listcat2,errCode);
	  // here is part to write value(s) for thiscategory
	  if (errCode>=0) {
	  if (listOutcat2->Length() == 1) {
	    // pair
	    numColumn = cattbl2->GetNumColumns();
	    colNames  = cattbl2->GetColumnNames();
	    rowValues = cattbl2->GetRow((*listOutcat2)[0]);
	    cwid = -1;
	    for (i=0; i< numColumn; i++) {
	      ilen = (*colNames)[i].Length();
	      if (  ilen > cwid) cwid=ilen;
	    }
	    
	    for (i=0; i< numColumn; i++) {
	      linePos=8; //**************
	      _PrintItemNameTab(cifo, categoryName2.Text(), &(*colNames)[i], linePos);
	      _PrintItemValue(cifo, &(*rowValues)[i], linePos);
	      if (linePos != 0) cifo << endl;
	    }
	    if (colNames)  delete colNames;
	    if (rowValues) delete rowValues;
	  }
	  
	  if (listOutcat2->Length() > 1) {
	    // loop
	    cifo << "     loop_" << endl;
	    numColumn = cattbl2->GetNumColumns();
	    colNames  = cattbl2->GetColumnNames();
	    for (i=0; i< numColumn; i++) {
	      _PrintItemNameTab(cifo, categoryName2.Text(), &(*colNames)[i], linePos);
	      cifo << endl;
	    }
	    
	    // ---
	    cwidth  = new int[numColumn];
	    
	    for (i=0; i< numColumn; i++) { cwidth[i]=-1; }
	    for (j=0; j < (int)(listOutcat2->Length()); j++) {
	      rowValues = cattbl2->GetRow((*listOutcat2)[j]);
	      for (i=0; i< numColumn-1; i++) {
		ilen = (*rowValues)[i].Length();
		if ( _IsQuotableText(&(*rowValues)[i])) ilen+=2;
		
		if (  ilen > cwidth[i]) cwidth[i]=ilen;
	      }
	      cwidth[numColumn-1] = 0;
	      if (rowValues) delete rowValues;
	    }
	    for (j=0; j < (int)(listOutcat2->Length()); j++) {
	      linePos = 0;
	      rowValues = cattbl2->GetRow((*listOutcat2)[j]);
	      for (i=0; i< numColumn; i++) {
		
		ilen=_PrintItemValueTab(cifo, &(*rowValues)[i], linePos,cwidth[i]); 
	      }
	      if (linePos != 0) cifo << endl;
	      if (rowValues) delete rowValues;
	    }
	    
	    if (cwidth) delete[] cwidth;
	    if (colNames) delete colNames;
	  }
	  }
	  }
	}
      }

      }
    cifo << "      end" <<endl;
    cifo << "   end" <<endl;
	cifo <<endl;
    }
    }
    }
  }
  if (listOut) delete listOut;
  if (listOutcat) delete listOutcat;
  if (listOutcat2) delete listOutcat2;
  if (listOutItem) delete listOutItem;
  if (listOutItem2) delete listOutItem2;
  return(1); 
}


int DICFileObj:: PrintItemAliases(const char * cifFileName) {
  int iret=0;
  ofstream cifo(cifFileName,ios::out|ios::trunc);
  iret =PrintItemAliases(cifo);
  cifo.close();
  return (iret);
}

int DICFileObj::PrintItemAliases(ostream &cifo) {

  ISTable * tbl;
  int numRow;
  int colIndex1,colIndex2;
  CifString item;

  tbl=GetTablePtr((*_fileIndex.blocks.Data())[0].GetName(),"item_aliases");
  if (tbl!=NULL) {
    numRow    = tbl->GetNumRows();
    colIndex1  = tbl->GetColumnIndex("name");
    colIndex2  = tbl->GetColumnIndex("alias_name");
    for (int i=0; i<numRow;i++) {
      tbl->GetCell(item,colIndex1,i);
      cifo.setf(ios::left,ios::adjustfield);
      cifo.width(50);
      cifo<<item.Text();
      tbl->GetCell(item,colIndex2,i);
      cifo.setf(ios::left,ios::adjustfield);
      cifo.width(30);
      cifo<<item.Text();
      cifo << endl;
    }
    return(1);
  }
  else
    return(-1);
}





void DICFileObj::Compress(TblFileObj * ddl) {

  int ib, jt, i;
  ISTable * catkey;
  ISTable * tbl;
  ReVarCifArray<CifString> catList;
  ReVarCifArray<CifString> catList2;
  ReVarCifArray<CifString> catkeyList;

  ReVarPCifArray<int> *listOut2;

  ReVarCifArray<CifString> catTarget;
  ReVarCifArray<CifString> catkeyTarget;

  int errCode;
  int catkeyIndex;
  CifString cell;
  CifString categoryName;
  char itemName[MAX_CIF_NAME_LENGTH];
  char *bId=NULL;
  char *tid;
  int end=1;
  char * blockName;

  blockName = ddl->GetBlockName(0);

  for (ib=0; ib < (int)(_fileIndex.numBlocks); ib++) {

    catkey = ddl->GetTablePtr(blockName,"category_key");
    catkeyList.Clear();
    catkeyList.Add("id");
    catkey->CreateIndex("index0",catkeyList);
    catkeyIndex = catkey->GetColumnIndex("name");


    tid=_BuildTableId((*_fileIndex.blocks.Data())[ib].GetName()," ");
    _fileIndex.tables.Seek(tid);
    if (tid) delete[] tid;
    jt=_fileIndex.tables.Current();
    bId = _GetExternalBlockName((*_fileIndex.tables.Data())[jt].GetBlock());
    while((!bId || strcmp(bId,(*_fileIndex.blocks.Data())[ib].GetName())==0)&&end!=-1){
      tbl = GetTablePtr(bId,(*_fileIndex.tables.Data())[jt].GetName());
      catkeyTarget.Clear();
      catkeyTarget.Add((*_fileIndex.tables.Data())[jt].GetName());
      listOut2 = catkey->Search(catkeyTarget,catkeyList,errCode);
      if (errCode>=0){
	catList2.Clear();
	for (i=0; i<(int)(listOut2->Length());i++) {
	  catkey->GetCell(cell,catkeyIndex,(*listOut2)[i]);
	  GetKeywordFromItemName(itemName, cell.Text());
	  catList2.Add(itemName);
	}
	tbl->CreateIndex("index00",catList2,1);
	tbl->CompressTable();
	tbl->DeleteIndex("index00");
      }
      end=_fileIndex.tables.GoToNext();
      jt=_fileIndex.tables.Current();
      if(listOut2) delete listOut2;listOut2=NULL;
      if(bId) delete[] bId; bId=NULL;
      bId = _GetExternalBlockName((*_fileIndex.tables.Data())[jt].GetBlock());
    }
  }
  if(bId) delete[] bId; 
  if(blockName) delete[] blockName;
}





char* DICFileObj:: uppercase(char *str)
{
  char * ret;
  ret = (char *)calloc((strlen(str)+1),sizeof(char));
  int i=0;
  while (str[i] != '\0')
    {
      if (islower(str[i]))
	ret[i]=toupper(str[i]);
      else
	ret[i]=str[i];
      i++;
    }
  return ret;
}

char* DICFileObj:: lowercase(char *str)
{

  char * ret;
  ret = (char *)calloc((strlen(str)+1),sizeof(char));
  int i=0;

  while (str[i] !='\0')
    {
      if (isupper(str[i]))
	ret[i]=tolower(str[i]);
      else
	ret[i]=str[i];
      i++;
    }
  return ret;
}

