/*
FILE:     _connect.h
*/
/*
VERSION:  1.700
*/
/*
DATE:     7/13/2005
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2005 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#ifndef __CONNECT_
#define __CONNECT_

#include "_index.h"
#include "BTreeObj.h"
#include "FileNavigator.h"

class AtomFormat {
   private:
  static const int _CONN_INCR;// = 4;
        int  _nconect;
        int  _alloc_conect;
        char *_atomname;
        char *_atomtype;
        int  _charge;
        int  _ishydrogen;
        int  _align_pos;
        char *_coord_x;
        char *_coord_y;
        char *_coord_z;
        char **_connectto;
        char **_bond_type;
        void Clear();
   public:
        AtomFormat() { Clear(); }
        AtomFormat(const AtomFormat&);
        ~AtomFormat() { Reset(); }
        void Reset();
        int   nconect() { return _nconect; }
        char* atomname() { return _atomname; }
        char* atomtype() { return _atomtype; }
        int   charge() { return _charge; }
        int   ishydrogen() { return _ishydrogen; }
        int   align_pos() { return _align_pos; }
        char* coord_x() { return _coord_x; }
        char* coord_y() { return _coord_y; }
        char* coord_z() { return _coord_z; }
        char* connectto(const int pos);
        char* bond_type(const int pos);
        void  set_atomname(const char *name);
        void  set_atomtype(const char *type);
        void  set_ishydrogen(const int d) { _ishydrogen = d; }
        void  set_charge(const int d) { _charge = d; }
        void  set_align_pos(const int d) { _align_pos = d; }
        void  set_coord_x(const char *x);
        void  set_coord_y(const char *y);
        void  set_coord_z(const char *z);
        void  add_connect(const char *name, const char *type);
        void  set_connectto(const int pos, const char *name);
        void  set_bondtype(const int pos, const char *type);
        int   WriteObject(FileNavigator *fileNav);
        int   ReadObject(Word index, FileNavigator *fileNav);
};

class ConnectFormat {
   protected:
  static const int _ATOM_INCR;// = 10;
        int  _natoms;
        int  _nheavyatoms;
        int  _allocatoms;
        int  _ntypes;
        char _AtomType[7];
        int  _ndbAtomType;
        int  _typeModified;
        char _drugname[6];
        char *_type;
        char *_formula;
        char *_chemical_name;
        char *_het_name;
        char *_synonym;
        char *_parent_residue;
        char *_formal_charge;
        char *_class_1;
        char *_class_2;
        char _status[5];
        char _date[12];
        char _modified_date[12];
        AtomFormat **_atoms;
        int  _numatomTree;
        int  _numModification;
        char **_Modification;
        BTreeObj<_SIndex> _atomTree;
        BTreeObj<_IIndex> _OrderTree;
        void Clear();
        void Reset();
   public:
        ConnectFormat() { Clear(); }
        ConnectFormat(const ConnectFormat&);
        ~ConnectFormat() { Reset(); }
        int   natoms() { return _natoms; }
        int   nheavyatoms() { return _nheavyatoms; }
        int   ntypes() { return _ntypes; }
        char* AtomType() { return _AtomType; }
        int   ndbAtomType() { return _ndbAtomType; }
        int   typeModified() { return _typeModified; }
        char* drugname() { return _drugname; }
        char* het_name() { return _het_name; }
        char* type() { return _type; }
        char* formula() { return _formula; }
        char* chemical_name() { return _chemical_name; }
        char* synonym() { return _synonym; }
        char* parent_residue() { return _parent_residue; }
        char* formal_charge() { return _formal_charge; }
        char* class_1() { return _class_1; }
        char* class_2() { return _class_2; }
        char* status() { return _status; }
        char* date() { return _date; }
        char* modified_date() { return _modified_date; }
        AtomFormat* getFirstAtom();
        AtomFormat* getNextAtom();
        AtomFormat* atoms(const int pos);
        int   numModification() { return _numModification; }
        char* Modification(const int pos); 
        _SIndex* find_atom(const char *name);
        void  set_AtomType(const char *string);
        void  set_ndbAtomType(const int type) { _ndbAtomType = type; }
        void  set_typeModified(const int type) { _typeModified = type; }
        void  set_drugname(const char *string);
        void  set_type(const char *string);
        void  set_formula(const char *string);
        void  set_chemical_name(const char *string);
        void  set_het_name(const char *string);
        void  set_synonym(const char *string);
        void  set_parent_residue(const char *string);
        void  set_formal_charge(const char *string);
        void  set_class_1(const char *string);
        void  set_class_2(const char *string);
        void  set_status(const char *string);
        void  set_date(const char *string);
        void  set_modified_date(const char *string);
        void  set_atoms(AtomFormat **Atoms, const int Natoms);
        void  set_Modification(const char *string);
        void  add_atom(AtomFormat *atom);
        int   WriteObject(FileNavigator *fileNav);
        int   ReadObject(Word index, FileNavigator *fileNav);
        int   operator<(const ConnectFormat& c);
        int   operator>(const ConnectFormat& c);
        int   operator==(const ConnectFormat& c);
        int   FindAlignPos();
        void  ReorderAtoms();
        void  ReorderAtomBasedExistOrder(ConnectFormat*);
        void  Deoxynucleoside();
};

#endif
