/*
FILE:     adm_connect.C
*/
/*
VERSION:  1.700
*/
/*
DATE:     7/13/2005
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2005 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>
#include <sys/stat.h>

#include "ndb_misclib.h"
#include "adm_connect.h"
#include "connect_iolib.h"
#include "erf_util.h"

#define conectextfile         "conect.cif"
#define componentextfile      "component.cif"

static void output_ndb_conect_section(ConnectFormat *drug, FILE *fp);
static void output_ndb_conect_atom_section(ConnectFormat *drug, FILE *fp);
static void print_formatted_item(FILE *outfile, int *offset,
   const char *value1, const int newline_end, const int fieldLength);

void Connect_Adm::Clear_fp()
{
       if (_fp) {
            _fp->CloseFile(); delete _fp; _fp = NULL;
       }
}

int Connect_Adm::openFile(const char *path, const int _dic_type, const int persistent)
{
       char asciifile[300];
       struct stat statbin;

       Clear_fp();

       if (persistent) {
            _mode = WRITE_MODE;
            _fp = new FileNavigator();
            sprintf(asciifile, "%s/data/binary/%s", path, conectbinfile);
            int ret;
            if (((ret = _fp->OpenFile(asciifile, _mode)) < 0) ||
                ((ret = _fp->ReadFileHeader()) < 0)) {
                 Clear_fp();
                 return ret;
            }
       }

       if (_dic_type == DIC_DEFAULT_CONNECT) {
            sprintf(asciifile, "./%s", conectextfile);
            if (stat(asciifile, &statbin) == -1) {
                 Connect_Dic::OpenFile(path, DIC_DEFAULT_COMPONENT);
            } else ReadNdbConnect(asciifile);
       } else if (_dic_type == DIC_DEFAULT_COMPONENT) {
            sprintf(asciifile, "./%s", componentextfile);
            if (stat(asciifile, &statbin) == -1) {
                 Connect_Dic::OpenFile(path, DIC_DEFAULT_COMPONENT);
            } else ReadChemComp(asciifile);
       }

       return 0;
}

int Connect_Adm::openfile(const char *filename, const int _dic_type)
{
       struct stat statbin;

       if (stat(filename, &statbin) == -1) return (-1);

       Clear_fp();

       if (_dic_type == DIC_DEFAULT_CONNECT) {
            ReadNdbConnect(filename);
       } else if (_dic_type == DIC_DEFAULT_COMPONENT) {
            ReadChemComp(filename);
       }

       return 0;
}

void Connect_Adm::closeFile()
{
       if (!_fp) return;

       int i = 0, j = 0;
       int *drugLocs = new int[_numConnect];
       char **drugIds = new char*[_numConnect];
       for (i = 0, j = 0; j < _numConnect; i++) {
            if ((*_ConnTree.Deleted())[i] != 1) {
                 drugLocs[j] = (*_ConnTree.Data())[i].GetDrugId();
                 drugIds[j] = new char[strlen((*_ConnTree.Data())[i].GetId())+1];
                 strcpy(drugIds[j], (*_ConnTree.Data())[i].GetId());
                 j++;
            }
       }
       int ret = 0;
       _fp->WriteWords((Word *) drugLocs, _numConnect, (Word &) ret);
       _fp->WriteStrings(drugIds, _numConnect, (Word &) ret);
       for (i = 0; i < _numConnect; i++) delete [] drugIds[i]; 
       delete [] drugIds;
       delete [] drugLocs;

       _fp->CloseFile();
       delete _fp;
       _fp = NULL;
       _mode = -1;
}

int Connect_Adm::WriteChemComp(const char *filename, const int flag, const char *odb_file)
{
       int i, irow = 0;
       char tmp_file[200] = "TEMP.CIF", monmer[4][2] = { "A", "C", "G", "I" };
       CifFileObj *fobjR = NULL;
       CifString string;
       ConnectFormat *drug = NULL, *drug_tmp = NULL;
       ISTable *Table = NULL;

       if (_ConnTree.Root() < 0) return (-1);

       if (!flag) {
            strcpy(tmp_file, "component_new.odb");
            if (strcmp(odb_file, "")) strcpy(tmp_file, odb_file);
            Table = new ISTable("index");
            Table->AddColumn("id");
            Table->AddColumn("formula");
       }

       fobjR = new CifFileObj(tmp_file, WRITE_MODE);

       drug = GetFirstConnect();
       while (drug) {
            if (strlen(drug->drugname()) < 4) {
                 if (!flag || flag && (!strcmp(drug->status(), "REL") || !strcmp(drug->status(), "")))
                      Write_Chem_Comp(drug, fobjR, flag);
                 if (!flag) {
                      Table->AddRow();
                      string.Copy(drug->drugname());
                      Table->UpdateCell(string, 0, irow);
                      string.Copy("");
                      if (drug->formula() && strcmp(drug->formula(), ""))
                           string.Copy(drug->formula());
                      Table->UpdateCell(string, 1, irow);
                      irow++;
                 }
            }
            drug = GetNextConnect();
       }

       fobjR->Write(filename);
       if (!flag) {
            for (i = 0; i < 4; i++) {
                 drug = find_drug(monmer[i]);
                 if (!drug) continue;
                 drug_tmp = new ConnectFormat(*drug);
                 drug_tmp->Deoxynucleoside();
                 Write_Chem_Comp(drug_tmp, fobjR, flag);

                 Table->AddRow();
                 string.Copy(drug_tmp->drugname());
                 Table->UpdateCell(string, 0, irow);
                 string.Copy("");
                 if (drug_tmp->formula() && strcmp(drug_tmp->formula(), ""))
                      string.Copy(drug_tmp->formula());
                 Table->UpdateCell(string, 1, irow);
                 irow++;

                 delete drug_tmp;
            }
            fobjR->WriteTable(Table, "INDEX", "index");
            fobjR->CloseFile();
       }
       delete fobjR;
       return 0;
}

void Connect_Adm::add_drug(ConnectFormat *drug)
{
       if (!drug) return;

       int loc = -1;
       if (_fp) {
            loc = drug->WriteObject(_fp);
            if (loc < 0) return;
       }

       int index = _GetIndexFromId(drug->drugname());
       if (index >= 0) {
            DrugInfo *idx = _ConnTree.GetCurrDataPtr();
            idx->SetDrugId(loc);
            if (_drugs[idx->GetMemId()] != drug) {
                 drug->set_date(_drugs[idx->GetMemId()]->date());
                 drug->set_status(_drugs[idx->GetMemId()]->status());
                 delete _drugs[idx->GetMemId()];
            }
            _drugs[idx->GetMemId()] = drug;
       } else {
            _ManageSpace();
            _drugs[_n_drugs] = drug;
            DrugInfo idx(drug->drugname());
            idx.SetMemId(_n_drugs);
            idx.SetDrugId(loc);
            _ConnTree.Add(idx);
            _numConnect++;
            _n_drugs++;
       }
}

void Connect_Adm::add_drugs(ConnectFormat **drug, const int num)
{
       for (int i = 0; i < num; i++) add_drug(drug[i]);
}

int Connect_Adm::ReadChemComp(const char *filename)
{
       char *diags = NULL;

       CifFileObj *fobjR = new CifFileObj("TEMP.CIF", WRITE_MODE, 80, "?", 1);
       if (fobjR == NULL) return (-1);
       fobjR->Read(filename, diags);
       if (diags && strlen(diags)) {
            ndb_log_message(NDB_MSG_ERR, "Read cif file %s failed\n%s\n",
                    filename, diags);
            return (-1);
       }

       int num = 0;
       char **blockName = fobjR->GetBlockNames(num);
       ConnectFormat *drug = NULL;
       for (int n = 0; n < num; n++) {
            drug = new ConnectFormat();
            if (!Read_Chem_Comp(drug, fobjR, blockName[n], 0)) add_drug(drug);
       }

       for (int i = 0; i < num; i++) delete [] blockName[i];
       delete [] blockName;

       delete fobjR;
       return 0;
}

int Connect_Adm::ReadNdbConnect(const char *filename)
{
       char *diags = NULL;
       CifFileObj *fobjR = new CifFileObj("TEMP.CIF", WRITE_MODE, 80, "?", 1);
       if (fobjR == NULL) return (-1);
       fobjR->Read(filename, diags);
       if (diags && strlen(diags)) {
            ndb_log_message(NDB_MSG_ERR, "Read cif file %s failed\n%s\n",
                            filename, diags);
            exit (1);
       }
 
       if (!fobjR->IsTablePresent("NDB_RESIDUES", "ndb_connect")) {
            delete fobjR; return (-1);
       }
       ISTable *Table = fobjR->GetTablePtr("NDB_RESIDUES", "ndb_connect");
       int MxDrugs = Table->GetNumRows();
       int name_index = Table->GetColumnIndex("res_name");
       int het_index = Table->GetColumnIndex("hetgroup_name");
       int formul_index = Table->GetColumnIndex("formul");
       int chem_index = Table->GetColumnIndex("hetgroup_chemical_name");
       int synon_index = Table->GetColumnIndex("hetgroup_synonym_name");
       int parent_index = Table->GetColumnIndex("parent_residue");
       int charge_index = Table->GetColumnIndex("formal_charge");
       int class_1_index = Table->GetColumnIndex("class_1");
       int class_2_index = Table->GetColumnIndex("class_2");
       int type_index = Table->GetColumnIndex("type");
       int status_index = Table->GetColumnIndex("status");
       int date_index = Table->GetColumnIndex("date");
       int modified_index = Table->GetColumnIndex("modified_date");

       ConnectFormat **conect = new ConnectFormat*[MxDrugs];
       BTreeObj<_SIndex> drugTree;
       _SIndex sindex;
       CifString string;
       int irow = 0;
       for (irow = 0; irow < MxDrugs; irow++) {
            conect[irow] = new ConnectFormat();

            if (name_index >= 0 && !Table->GetCell(string, name_index, irow))
                 conect[irow]->set_drugname(string.Text());
            sindex.set_value(string.Text());
            sindex.set_pos(irow);
            drugTree.Add(sindex);

            if (het_index >= 0 && !Table->GetCell(string, het_index, irow)) {
                 if (string != "?" && string != ".")
                      conect[irow]->set_het_name(string.Text());
            }

            if (formul_index >= 0 && !Table->GetCell(string, formul_index, irow)) {
                 if (string != "?" && string != ".")
                      conect[irow]->set_formula(string.Text());
            }

            if (chem_index >= 0 && !Table->GetCell(string, chem_index, irow)) {
                 if (string != "?" && string != ".")
                      conect[irow]->set_chemical_name(string.Text());
            }

            if (synon_index >= 0 && !Table->GetCell(string, synon_index, irow)) {
                 if (string != "?" && string != ".")
                      conect[irow]->set_synonym(string.Text());
            }

            if (parent_index >= 0 && !Table->GetCell(string, parent_index, irow)) {
                 if (string != "?" && string != ".")
                      conect[irow]->set_parent_residue(string.Text());
            }

            if (charge_index >= 0 && !Table->GetCell(string, charge_index, irow)) {
                 if (string != "?" && string != ".")
                      conect[irow]->set_formal_charge(string.Text());
            }

            if (class_1_index >= 0 && !Table->GetCell(string, class_1_index, irow)) {
                 if (string != "?" && string != ".")
                      conect[irow]->set_class_1(string.Text());
            }

            if (class_2_index >= 0 && !Table->GetCell(string, class_2_index, irow)) {
                 if (string != "?" && string != ".")
                      conect[irow]->set_class_2(string.Text());
            }

            if (type_index >= 0 && !Table->GetCell(string, type_index, irow)) {
                 if (string != "?" && string != ".")
                      conect[irow]->set_type(string.Text());
            }

            if (status_index >= 0 && !Table->GetCell(string, status_index, irow)) {
                 if (string != "?" && string != ".")
                      conect[irow]->set_status(string.Text());
            }

            if (date_index >= 0 && !Table->GetCell(string, date_index, irow)) {
                 if (string != "?" && string != ".")
                      conect[irow]->set_date(string.Text());
            }

            if (modified_index >= 0 && !Table->GetCell(string, modified_index, irow)) {
                if (string != "?" && string != ".")
                     conect[irow]->set_modified_date(string.Text());
            }
       }

       if (!fobjR->IsTablePresent("NDB_RESIDUES", "ndb_connect_type")) {
            delete fobjR; return (-1);
       }
       Table = fobjR->GetTablePtr("NDB_RESIDUES", "ndb_connect_type");
       name_index = Table->GetColumnIndex("res_name");
       type_index = Table->GetColumnIndex("ndbTokenType");
       int modf_index = Table->GetColumnIndex("modified");
       ConnectFormat *drug = NULL;
       for (irow = 0; irow < Table->GetNumRows(); irow++) {
            Table->GetCell(string, name_index, irow);
            sindex.set_value(string.Text());
            if (drugTree.Seek(sindex) >= 0) {
                 sindex = drugTree.GetCurrData();
                 drug = conect[sindex.pos()];
                 if (!Table->GetCell(string, type_index, irow))
                      drug->set_AtomType(string.Text());
                 if (!Table->GetCell(string, modf_index, irow))
                      drug->set_typeModified(atoi(string.Text()));
            }
       }

       if (fobjR->IsTablePresent("NDB_RESIDUES", "ndb_connect_modification")) {
            Table = fobjR->GetTablePtr("NDB_RESIDUES", "ndb_connect_modification");
            name_index = Table->GetColumnIndex("res_name");
            type_index = Table->GetColumnIndex("modification");
            for (irow = 0; irow < Table->GetNumRows(); irow++) {
                 Table->GetCell(string, name_index, irow);
                 sindex.set_value(string.Text());
                 if (drugTree.Seek(sindex) >= 0) {
                      sindex = drugTree.GetCurrData();
                      drug = conect[sindex.pos()];
                      if (!Table->GetCell(string, type_index, irow))
                           drug->set_Modification(string.Text());
                 }
            }
       }

       if (!fobjR->IsTablePresent("NDB_RESIDUES", "ndb_connect_atom")) {
            delete fobjR; return (-1);
       }
       Table = fobjR->GetTablePtr("NDB_RESIDUES", "ndb_connect_atom");
       int res_index = Table->GetColumnIndex("res_name");
       int atom_index = Table->GetColumnIndex("atom_name");
       type_index = Table->GetColumnIndex("type_symbol");
       charge_index = Table->GetColumnIndex("charge");
       int conect_index = Table->GetColumnIndex("connect_to");
       int bond_index = Table->GetColumnIndex("bond_type");
       int align_pos = Table->GetColumnIndex("align_pos");

       CifString connecto, bondtype;
       CifString DrugName("XXX", 4);
       CifString AtomName("XXX", 4);
       CifString AtomType, charge, atomname;
       int subrow1 = 0, nfield = 0, i, nrow, iirow;
       AtomFormat **atoms = NULL;

       for (irow = 0; irow < Table->GetNumRows(); irow++) {
            Table->GetCell(string, res_index, irow);
            if (string != DrugName) {
                 DrugName = string;
                 nfield = 0;
                 i = irow;
                 AtomName.Clear();
                 AtomName.Copy("XXX");
                 do {
                      Table->GetCell(string, res_index, i);
                      if (string == DrugName) {
                           Table->GetCell(atomname, atom_index, i);
                           if (string == DrugName) {
                                if (atomname != AtomName) {
                                     AtomName = atomname;
                                     nfield++;
                                }
                           }
                           i++;
                      }
                 } while (i < Table->GetNumRows() && DrugName == string);

                 sindex.set_value(DrugName.Text());
                 if (drugTree.Seek(sindex) < 0) continue;
                 sindex = drugTree.GetCurrData();
                 drug = conect[sindex.pos()];

                 nrow = i;
                 atoms = new AtomFormat*[nfield];

                 subrow1 = 0;
                 Table->GetCell(AtomName, atom_index, irow);
                 Table->GetCell(AtomType, type_index, irow);
                 atoms[subrow1] = new AtomFormat();
                 atoms[subrow1]->set_atomname(AtomName.Text());
                 if (AtomType != "." && AtomType != "?") {
                      atoms[subrow1]->set_atomtype(AtomType.Text());
                      if (!strcmp(atoms[subrow1]->atomtype(), "H"))
                      atoms[subrow1]->set_ishydrogen(1);
                 }
                 charge.Clear();
                 Table->GetCell(charge, charge_index, irow);
                 if (charge != "." && charge != "?" && charge != "")
                      atoms[subrow1]->set_charge(atoi(charge.Text()));
                 charge.Clear();
                 Table->GetCell(charge, align_pos, irow);
                 if (charge != "." && charge != "?" && charge != "")
                      atoms[subrow1]->set_align_pos(atoi(charge.Text()));
                 for (iirow = irow; iirow < nrow; iirow++) {
                      if (!Table->GetCell(string, atom_index, iirow)) {
                           if (string != AtomName) {
                                subrow1++;
                                AtomName = string;
                                atoms[subrow1] = new AtomFormat();
                                atoms[subrow1]->set_atomname(AtomName.Text());
                                Table->GetCell(AtomType, type_index, iirow);
                                if (AtomType != "." && AtomType != "?") {
                                     atoms[subrow1]->set_atomtype(AtomType.Text());
                                     if (!strcmp(atoms[subrow1]->atomtype(), "H"))
                                     atoms[subrow1]->set_ishydrogen(1);
                                }
                                charge.Clear();
                                Table->GetCell(charge, charge_index, iirow);
                                if (charge != "." && charge != "?" && charge != "")
                                     atoms[subrow1]->set_charge(atoi(charge.Text()));
                                charge.Clear();
                                Table->GetCell(charge, align_pos, iirow);
                                if (charge != "." && charge != "?" && charge != "")
                                     atoms[subrow1]->set_align_pos(atoi(charge.Text()));
                           }
                      }
                      get_value(connecto, Table, conect_index, iirow);
                      if (connecto != "") {
                           get_value(bondtype, Table, bond_index, iirow);
                           atoms[subrow1]->add_connect(connecto.Text(), bondtype.Text());
                      }
                 }
                 drug->set_atoms(atoms, nfield);
                 delete [] atoms;
                 irow = nrow - 1;
            }
       }

       add_drugs(conect, MxDrugs);
       delete [] conect;
       delete fobjR;

       return 0;
}

int Connect_Adm::WriteNdbConnect(const char *filename)
{
       if (_ConnTree.Root() < 0) return (-1);
       FILE *fp = fopen(filename, "w");
       fprintf(fp, "data_NDB_RESIDUES\n");
       fprintf(fp, "loop_\n");
       fprintf(fp, "_ndb_connect.res_name\n");
       fprintf(fp, "_ndb_connect.hetgroup_name             #from HET\n");
       fprintf(fp, "_ndb_connect.formul                    #from FORMUL\n");
       fprintf(fp, "_ndb_connect.hetgroup_chemical_name    #from HETNAM\n");
       fprintf(fp, "_ndb_connect.hetgroup_synonym_name     #from HETSYN\n");
       fprintf(fp, "_ndb_connect.parent_residue\n");
       fprintf(fp, "_ndb_connect.formal_charge\n");
       fprintf(fp, "_ndb_connect.class_1\n");
       fprintf(fp, "_ndb_connect.class_2\n");
       fprintf(fp, "_ndb_connect.type\n");
       fprintf(fp, "_ndb_connect.status\n");
       fprintf(fp, "_ndb_connect.date\n");
       fprintf(fp, "_ndb_connect.modified_date\n");

       ConnectFormat *drug = GetFirstConnect();
       while (drug) {
            output_ndb_conect_section(drug, fp);
            drug = GetNextConnect();
       }

       fprintf(fp, "loop_\n");
       fprintf(fp, "_ndb_connect_type.res_name\n");
       fprintf(fp, "_ndb_connect_type.ndbTokenType\n");
       fprintf(fp, "_ndb_connect_type.modified\n");

       drug = GetFirstConnect();
       while (drug) {
            fprintf(fp, "%3s %6s %d\n", drug->drugname(),
                      drug->AtomType(), drug->typeModified());
            drug = GetNextConnect();
       }
       fprintf(fp, "\n");

       fprintf(fp, "loop_\n");
       fprintf(fp, "_ndb_connect_modification.res_name\n");
       fprintf(fp, "_ndb_connect_modification.modification\n");

       int i = 0;
       drug = GetFirstConnect();
       while (drug) {
            if (drug->numModification()) {
                 for (i = 0; i < drug->numModification(); i++)
                      fprintf(fp, "%3s '%s'\n", drug->drugname(),
                                  drug->Modification(i));
            }
            drug = GetNextConnect();
       }
       fprintf(fp, "\n");

       fprintf(fp, "loop_\n");
       fprintf(fp, "_ndb_connect_atom.res_name\n");
       fprintf(fp, "_ndb_connect_atom.atom_name\n");
       fprintf(fp, "_ndb_connect_atom.type_symbol\n");
       fprintf(fp, "_ndb_connect_atom.charge\n");
       fprintf(fp, "_ndb_connect_atom.connect_to\n");
       fprintf(fp, "_ndb_connect_atom.bond_type\n");
       fprintf(fp, "_ndb_connect_atom.align_pos\n");

       drug = GetFirstConnect();
       while (drug) {
            output_ndb_conect_atom_section(drug, fp);
            drug = GetNextConnect();
       }
       fclose (fp);
       return 0;
}

static void output_ndb_conect_section(ConnectFormat *drug, FILE *fp)
{
       int linePos = 0;
       int len = strlen(drug->drugname());
       print_formatted_item(fp, &linePos, drug->drugname(), 0, len);

       if (drug->het_name()) {
            len = strlen(drug->het_name());
            print_formatted_item(fp, &linePos, drug->het_name(), 0, len);
       } else print_formatted_item(fp, &linePos, "", 0, 1);

       if (drug->formula()) {
            len = strlen(drug->formula());
            print_formatted_item(fp, &linePos, drug->formula(), 0, len);
       } else print_formatted_item(fp, &linePos, "", 0, 1);

       if (drug->chemical_name()) {
            len = strlen(drug->chemical_name());
            print_formatted_item(fp, &linePos, drug->chemical_name(), 0, len);
       } else print_formatted_item(fp, &linePos, "", 0, 1);

       if (drug->synonym()) {
            len = strlen(drug->synonym());
            print_formatted_item(fp, &linePos, drug->synonym(), 0, len);
       } else print_formatted_item(fp, &linePos, "", 0, 1);

       if (drug->parent_residue()) {
            len = strlen(drug->parent_residue());
            print_formatted_item(fp, &linePos, drug->parent_residue(), 0, len);
       } else print_formatted_item(fp, &linePos, "", 0, 1);

       if (drug->formal_charge()) {
            len = strlen(drug->formal_charge());
            print_formatted_item(fp, &linePos, drug->formal_charge(), 0, len);
       } else print_formatted_item(fp, &linePos, "0", 0, 1);

       if (drug->class_1()) {
            len = strlen(drug->class_1());
            print_formatted_item(fp, &linePos, drug->class_1(), 0, len);
       } else print_formatted_item(fp, &linePos, "", 0, 1);

       if (drug->class_2()) {
            len = strlen(drug->class_2());
            print_formatted_item(fp, &linePos, drug->class_2(), 0, len);
       } else print_formatted_item(fp, &linePos, "", 0, 1);

       if (drug->type()) {
            len = strlen(drug->type());
            print_formatted_item(fp, &linePos, drug->type(), 0, len);
       } else print_formatted_item(fp, &linePos, "", 0, 1);

       len = strlen(drug->status());
       print_formatted_item(fp, &linePos, drug->status(), 0, len);

       len = strlen(drug->date());
       print_formatted_item(fp, &linePos, drug->date(), 0, len);

       len = strlen(drug->modified_date());
       print_formatted_item(fp, &linePos, drug->modified_date(), 1, len);

       fprintf(fp, "\n");
}

static void output_ndb_conect_atom_section(ConnectFormat *drug, FILE *fp)
{
       int linePos = 0, j, k;
       char charge[5], align[5];
       for (j = 0; j < drug->natoms(); j++) {
            linePos = 0;
            sprintf(charge, "%d", drug->atoms(j)->charge());
            sprintf(align, "%d", drug->atoms(j)->align_pos());
            if (drug->atoms(j)->nconect() == 0) {
                 print_formatted_item(fp, &linePos, drug->drugname(), 0, 3);
                 print_formatted_item(fp, &linePos,
                         drug->atoms(j)->atomname(), 0, 5);
                 if (drug->atoms(j)->atomtype() && 
                     strcmp(drug->atoms(j)->atomtype(), ""))
                      print_formatted_item(fp, &linePos,
                         drug->atoms(j)->atomtype(), 0, 3);
                 else print_formatted_item(fp, &linePos, ".", 0, 3);
                 print_formatted_item(fp, &linePos, charge, 0, 2);
                 print_formatted_item(fp, &linePos, "", 0, 5);
                 print_formatted_item(fp, &linePos, "", 0, 5);
                 print_formatted_item(fp, &linePos, align, 1, 2);
            }
            for (k = 0; k < drug->atoms(j)->nconect(); k++) {
                 linePos = 0;
                 print_formatted_item(fp, &linePos, drug->drugname(),
                         0, 3);
                 print_formatted_item(fp, &linePos,
                         drug->atoms(j)->atomname(), 0, 5);
                 if (drug->atoms(j)->atomtype() &&
                     strcmp(drug->atoms(j)->atomtype(), ""))
                      print_formatted_item(fp, &linePos,
                         drug->atoms(j)->atomtype(), 0, 3);
                 else print_formatted_item(fp, &linePos, ".", 0, 3);
                 print_formatted_item(fp, &linePos, charge, 0, 2);
                 if (drug->atoms(j)->connectto(k))
                      print_formatted_item(fp, &linePos,
                         drug->atoms(j)->connectto(k), 0, 5);
                 else print_formatted_item(fp, &linePos, ".", 0, 5);
                 if (strcmp(drug->atoms(j)->bond_type(k), ""))
                      print_formatted_item(fp, &linePos,
                        drug->atoms(j)->bond_type(k), 0, 5);
                 else print_formatted_item(fp, &linePos, ".", 0, 5);
                 print_formatted_item(fp, &linePos, align, 1, 2);
            }
       }
}

static void print_formatted_item(FILE *outfile, int *offset,
       const char *value1, const int newline_end, const int fieldLength)
{
       char value[1000], tmpstring[81], fieldFormat[81];

       if (value1 == NULL || !strcmp("", value1))
            strcpy(value, "?");
       else strcpy(value, value1);

       int str_len = strlen(value);
       int multipleLine = 0;
       int multipleWord = 0;
       int embeddedQuotes = 0;
       int i = 0, pos = 0, len;
       for (i = 0; i < str_len; i++) {
            if (value[i] == ' ') multipleWord = 1;
            else if (value[i] == '"' || value[i] == '\'') embeddedQuotes = 1;
       }
       if (embeddedQuotes && multipleWord) multipleLine = 1;
       if (embeddedQuotes) multipleWord = 1;

       if (str_len > 76 || multipleLine) {
            if (*offset != 0) fprintf(outfile, "\n");
            fprintf(outfile, ";");
            pos = 0;
            while (pos < str_len) {
                 memset(tmpstring, 0, 81);
                 strncpy(tmpstring, &value[pos], 80-1);
                 len = str_len - pos;
                 if ( str_len - pos > 80) len = 78;

                 for (i = len; i > 0; i--)
                      if (!isalnum(tmpstring[i]) && tmpstring[i] != '.' &&
                           tmpstring[i] != '(') break;
                 if (i == 0) i = len;

                 memset(tmpstring, 0, 81);
                 strncpy(tmpstring, &value[pos], i+1);
                 pos += (i+1);
                 fprintf(outfile, "%s\n", tmpstring);
            }
            fprintf(outfile, ";\n");
            *offset = 0;
       } else {
            if (str_len + (*offset) > 75 || fieldLength + (*offset) > 75)  {
                 fprintf(outfile, "\n");
                 *offset = 0;
            }

            if (!(multipleWord || embeddedQuotes)) {
                 if (*offset != 0)
                      sprintf(tmpstring, " %s", value);
                 else strcpy(tmpstring, value);
            } else {
                 if (*offset != 0)
                      sprintf(tmpstring, " \'%s\'", value);
                 else sprintf(tmpstring, "\'%s\'", value);
            }
            if (str_len < fieldLength) {
                 if ((fieldLength+3) > 80)
                      str_len +=3;
                 else str_len = fieldLength+3;
            } else str_len +=3;

            sprintf(fieldFormat,"%%-%ds",str_len);
            fprintf(outfile, fieldFormat, tmpstring);

            (*offset) += (str_len);
            if (newline_end) fprintf(outfile, "\n");
       }
}
