/*
FILE:     _fileobj.h
*/
/*
VERSION:  1.700
*/
/*
DATE:     7/13/2005
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2005 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#ifndef _FILEOBJ_H_
#define _FILEOBJ_H_

#include <stdio.h>

#define NDB_DATE_FIELD            7

#include "token_card.h"
#include "_molecule.h"
#include "CifFileObj.h"
#include "cry_util.h"

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#define EXPERIMENT_TYPE_BASIC       1
#define EXPERIMENT_TYPE_XRAY        2
#define EXPERIMENT_TYPE_NMR         4
#define EXPERIMENT_TYPE_MODEL       8
#define EXPERIMENT_TYPE_NEUTRON    16
#define EXPERIMENT_TYPE_ELECTRON   32
#define EXPERIMENT_TYPE_CRYO_EM    64
#define EXPERIMENT_TYPE_SOLN_SCT  128
#define EXPERIMENT_TYPE_FIBER     256

#define MAX_NUM_ERROR 50

#define FILE_STATUS_NDB_PLUS_MOL    1
#define FILE_STATUS_MMCIF           2
#define FILE_STATUS_MMCIF_PLUS_MOL  3

typedef struct {
        int  internal_order;
        int  order;
        int  Biol_ID;
        CifString Asym_ID;
        CifString NDB_Strand_ID;
        char PDB_Strand_ID[3];
        char PDB_Strand_ID_Flag[3];
        char Symmetry[7];
        int  bbrn;
        int  bern;
        int  abrn;
        int  aern;
        int  is_change;
} _STBIOL;

class _FileObj
{
   public:
      _FileObj(const int, const int);
      ~_FileObj();
      int   experiment_type() { return _experiment_type; }
      void  delete_Seqs();
      void  set_input_filename(const char*);
      void  set_output_filename(const char*);
      void  set_real_filename(const char*);
      void  set_StructureId(const char*);
      void  set_input_format(const int);
      void  set_output_format(const int);
      void  set_experiment_type(const int);
      void  set_file_status(const int);
      void  set_is_binary(const int);
      int   read_ndb_file(const int);
      int   read_ndb_pdb_file(const int, const int, const int);
      int   read_ndb_cif_file();
      int   read_mmcif_file(const int);
      int   read_mmcif_file(int&, const int, const int, const int, const int, const int, const int);
      void  print_general_cards(FILE*, const int, const int, int&, const int);
      void  write_ndb_file(const int, const int, const int, const int, const int);
      void  get_sequence_from_seqres();
      void  get_seqres_from_mapping(const int);
      void  seq_alignment(const int);
      void  find_missing_residues();
      void  get_seqifo(const int, const int);
      void  put_coordinates_into_molecules(const int);
      void  update_tokens(const int);
      void  convert_between_orthogonal_and_fractional(const int);
      void  add_chain_and_solvent();
      void  merge_waters();
   protected:
      typedef struct {
            char old_chn[5];
            char new_chn[5];
      } _CHAIN_CORR;
      void _ManageSpace();
      void _ManageSeqSpace();
      void _ManagePdbSpace();
      void _ManageResSpace(SEQ *seq);
      void get_temp_file();
      void get_one_line_info(FILE*, int&, int&, char*, int&, const int,
              char*, char FieldInfo[MxField][MxLen], const int,
              const int, const int, const int);
      void put_card(const int, const int, char FieldInfo[MxField][MxLen]); 
      void print_jrnl_cards(FILE*, const int, int&);
      void read_general_cards(ISTable*, const int, const int);
      void read_atom_records(ISTable*);
      void write_general_cards(const int, const int);
      void write_jrnl_cards();
      void print_molecules(FILE*, const int, const int, const int, const int);
      void print_molecule(_MOLECULE*, ISTable*, int&, FILE*, const int, const int, const int, const int);
      void get_connect_cards();
      void print_atom(ISTable*, FILE*, const int, int&, int&, const int, _ATOM*,
             char FieldInfo[MxField][MxLen], const int, int&, const int);
      void get_cell();
      void delete_null_cards();
      void get_sequence_types();
      void print_master_card(FILE*, const int);
      int  ndb_count_general_line_no(const int);
      int  ndb_count_atom_line_no();
      int  ndb_count_num_chain();
      void cifndb_process_category_correspond();
      void cifndb_process_asym_poly_seq_ndb(const int);
      void match_entity_poly_seq_to_seqres(const int);
      void insert_to_seqs(ISTable*, char*, const int, char**, int&, const int);
      void match_entity_poly_to_seqres(const int);
      void create_dbref_and_seqadv(const int, const int, int, char**, const int,
               char**, int&, const int);
      void get_symmetry_from_remark_350();

      char               _input_filename[200];
      char               _output_filename[200];
      char               _real_filename[200];
      char               _temp_file[200];
      char               _StructureId[100];
      int                _input_format;
      int                _output_format;
      int                _experiment_type;
      int                _file_status;
      int                _num_mol;
      int                _alloc_mol;
      int                _rep_mol;
      Cards_Info_Format* _CardInfo;
      Cards_Info_Format* _JrnlCardInfo;
      _MOLECULE**        _molecules;
      _MOLECULE*         _entity_molecule;
      CifFileObj*        _CifObj;
      int                _Alloc_Seqs;
      int                _Num_Seqs;
      SEQ**              _Seqs;
      int                _Alloc_NPdb;
      int                _Num_NPdb;
      SEQ**              _NPdb;
      NDBCELLCONST*      _cell;
      int                _n_stbiol;
      _STBIOL**          _stbiol;
      CifString          _error_message;
      CifString          _error_message_s;
      CifString          _unrecognized_residue;
      CifString          _Error_Check_Message[MAX_NUM_ERROR];
      int                _is_binary;
      int                _Error_Check[MAX_NUM_ERROR];
      _CHAIN_CORR*       _chain_corr;
      int                _n_chain_corr;
      int                _exist_pdb_chain_id;
};

extern void print_atom_card(FILE *fp, const int card_id, const int temp_id,
            int &atomSerialNo, int &WaterSerialNo, char Field[MxField][MxLen],
            const int format1, int align, const int is_iupac);
extern void get_pdb_id(ISTable *Table1, char *type);

#endif
