/*
FILE:     erf_iolib.h
*/
/*
VERSION:  1.700
*/
/*
DATE:     7/13/2005
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2005 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#ifndef _ERF_IOLIB_
#define _ERF_IOLIB_

#define AUX_ATOM                  2
#define SEQ_ATOM                  4

#define MxLen                  8000
#define MxStrLen                 84
#define TRUE                      1
#define FALSE                     0

typedef struct {
        char  *FieldLabel;          /* Field name */
        int   FieldType;            /* See above for type descriptions */
        int   FieldWidth;           /* Width of the field */
        int   FieldPrec;            /* Number of digit after decimal point */
        int   FieldStcol;           /* Starting column of the field */
        int   NdbOnly;              /* 1: for NDB used only, 0: used in both
                                     * NDB /PDB */
        int   MixedCase;            /* 1 consists capital and small cases
                                       0 for captial only */
} field_format;

typedef struct {
        char         TokenName[10];
        int          SeqField;     /* field of serial number */
        int          NumField;     /* number of fields */
        int          ConField;     /* mark for continuation indication */
        int          PdbConField;  /* Pdb continued field, for example
                                    * title consist paper TITLE and descriptor,
                                    * which is two parts for NDB files,
                                    * however, it is one thing for PDB files*/
        int          ContinuedF;   /* fields could be continued */
        int          MatrixType;   /* 0: non-matrix token, 3: matrix token */
        int          NdbOnly;      /* 1: for NDB used only, 0: used in both
                                    * NDB/PDB */
        int          DisplayOrder;
        field_format *FieldList;
} ndb_token_format;

typedef struct {
        int  NdbTokenId; /* TokenId corresponding to ndb_token_format */
        int  AuxAtom;    /* 1 to identify auxiliary atom, such as ANISOU, etc */
        int  Esd;        /* 1 to identify esd atom, such as SIGUIJ, SIGATM */
        int  SeqAtom;    /* 1 to identify atomserial number increment, TER is
                          * the only one in AuxAtom group with 1 value */
        char *PdbTitle;  /* TokenTitle used in PDB */
        char *TableName; /* table name in database */
} atom_format;

typedef struct{
        int  Data[15];
} SPACE_DATA;

typedef struct {
        char SpaceGroupName[MxStrLen];
        int  SpaceDataNo;
        SPACE_DATA *SpaceData;
} SPACE_GROUP;


extern int ALTSQN_TOKEN;
extern int ALTSQR_TOKEN;
extern int ANISOU_TOKEN;
extern int ATOMMN_TOKEN;
extern int ATOMMP_TOKEN;
extern int ATOMN_TOKEN;
extern int ATOMP_TOKEN;
extern int ATOMS_TOKEN;
extern int AUTHOR_TOKEN;
extern int BIOMT_TOKEN;
extern int BIOUNI_TOKEN;
extern int BSPAIR_TOKEN;
extern int COMPND_TOKEN;
extern int CRMETH_TOKEN;
extern int CRSOLU_TOKEN;
extern int CRYST1_TOKEN;
extern int DBREF_TOKEN;
extern int DCHEM_TOKEN;
extern int DTMEAS_TOKEN;
extern int DTMETH_TOKEN;
extern int DTTEMP_TOKEN;
extern int DTWAVE_TOKEN;
extern int ECNUM_TOKEN;
extern int ENTAIL_TOKEN;
extern int ENTDES_TOKEN;
extern int ENTITY_TOKEN;
extern int ENTMUT_TOKEN;
extern int ENTNAM_TOKEN;
extern int EXPDTA_TOKEN;
extern int FORMUL_TOKEN;
extern int FRAGMT_TOKEN;
extern int FREERF_TOKEN;
extern int HEADER_TOKEN;
extern int HELIX_TOKEN;
extern int HET_TOKEN;
extern int HETAC_TOKEN;
extern int HETAD_TOKEN;
extern int HETAI_TOKEN;
extern int HETAIN_TOKEN;
extern int HETAS_TOKEN;
extern int HETIC_TOKEN;
extern int HETNAM_TOKEN;
extern int HETSYN_TOKEN;
extern int HYDBND_TOKEN;
extern int INHIBT_TOKEN;
extern int JRNL_TOKEN;
extern int LINK_TOKEN;
extern int MASTER_TOKEN;
extern int MISMAT_TOKEN;
extern int MISRES_TOKEN;
extern int MODEL_TOKEN;
extern int MODIFB_TOKEN;
extern int MODIFP_TOKEN;
extern int MODIFS_TOKEN;
extern int MODRES_TOKEN;
extern int MTRIX_TOKEN;
extern int NCSGRO_TOKEN;
extern int NCSMOD_TOKEN;
extern int NDBFIL_TOKEN;
extern int NDBPRO_TOKEN;
extern int NDBSQN_TOKEN;
extern int NMRECD_TOKEN;
extern int NMREXP_TOKEN;
extern int NMRSFT_TOKEN;
extern int NMRSMP_TOKEN;
extern int NMRSPM_TOKEN;
extern int ORIGX_TOKEN;
extern int PCOMPN_TOKEN;
extern int PDBFIL_TOKEN;
extern int PDBSQN_TOKEN;
extern int PDBSQR_TOKEN;
extern int PHVAL_TOKEN;
extern int PREMRK_TOKEN;
extern int PROCES_TOKEN;
extern int PROTEI_TOKEN;
extern int PSOURC_TOKEN;
extern int RADIAT_TOKEN;
extern int RECVD1_TOKEN;
extern int REFMET_TOKEN;
extern int REMARK_TOKEN;
extern int RENTRY_TOKEN;
extern int REVDAT_TOKEN;
extern int RFACTR_TOKEN;
extern int SCALE_TOKEN;
extern int SEQIFO_TOKEN;
extern int SEQRES_TOKEN;
extern int SFTWAR_TOKEN;
extern int SHEET_TOKEN;
extern int SIGATM_TOKEN;
extern int SIGUIJ_TOKEN;
extern int SITE_TOKEN;
extern int SLTBRG_TOKEN;
extern int SOURCE_TOKEN;
extern int SSBOND_TOKEN;
extern int STASYM_TOKEN;
extern int STATUS_TOKEN;
extern int STBIOL_TOKEN;
extern int SYMMA_TOKEN;
extern int SYMTRA_TOKEN;
extern int TARGET_TOKEN;
extern int TER_TOKEN;
extern int TLSGRO_TOKEN;
extern int TURN_TOKEN;
extern int WAVLEN_TOKEN;
extern int XFILES_TOKEN;

extern int MxTokens;
extern ndb_token_format *NdbToken;
extern int MxJrnlTokens;
extern ndb_token_format *JrnlToken;
extern int MxAtomTokens;
extern atom_format *AtomInfo;

extern int find_card_id(const char *CardString);
extern int find_jrnl_id(const char *JRNLCard);
extern SPACE_GROUP *get_spacegroup_info(const char *spgname);
extern int is_atom_card(const int card_id);
extern int get_atom_card_id(const int card_id);
extern int read_ndb_tokens(void);

#endif
