/*
FILE:     _chain.C
*/
/*
VERSION:  1.700
*/
/*
DATE:     7/13/2005
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2005 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>
#include <stdlib.h>

#include "erf_connect.h"
#include "ndb_misclib.h"
#include "_chain.h"
#include "align_util.h"
#include "erf_util.h"
#include "util.h"

static SEQ *_seqa = NULL, *_seqb = NULL, *_pdba = NULL, *_pdbb = NULL;
static int score(const int k, const int l);
static int is_contact(_RESIDUE *res_a, _RESIDUE *res_b);

const int _CHAIN::_RES_INCR = 50;
void _CHAIN::Clear()
{
       _ChainID.Clear();
       memset(_PDB_ChainID, 0, 5);
       memset(_PUB_ChainID, 0, 5);
       memset(_PDB_ChainID_Flag, 0, 5);
       _has_sequence = 0;
       _order = 0;
       _internal_order = 0;
       _index = 0;
       _chain_type = 0;
       _na_type = -1;
       _is_linked = 0;
       _display = 1;
       _biol_id = 0;
       _icolor  = 0;
       _op = 1;
       _lx = 0;
       _ly = 0;
       _lz = 0;
       _isModified = 0;
       _entity_id = 0;
       _weight = 0;
       _ResidueNumbers = 0;
       _Num_Residue = 0;
       _Alloc_Residue = 0;
       _SeqLen = 0;
       _Alloc_SeqLen = 0;
       _alloc_helix = 0;
       _n_helix = 0;
       _alloc_turn = 0;
       _n_turn = 0;
       _residues = NULL;
       _currResidue = NULL;
       _SeqRes = NULL;
       _helix = NULL;
       _turn = NULL;
       _pdbIndex.Clear();
       _ndbIndex.Clear();
       _numIndex.Clear();
       _cifnumIndex.Clear();
       _resIndex.Clear();
       _ResNumIndex.Clear();
}

void _CHAIN::Reset()
{
       int i = 0;
       if (_residues) {
            for (i = 0; i < _Num_Residue; i++)
                 if (_residues[i]) delete _residues[i];
            delete [] _residues;
       }
       if (_SeqRes) {
            for (i = 0; i < _SeqLen; i++)
                 if (_SeqRes[i]) delete _SeqRes[i];
            delete [] _SeqRes;
       }
       if (_helix) {
            for (i = 0; i < _n_helix; i++) {
                 if (_helix[i]) {
                      if (_helix[i]->comment) delete [] _helix[i]->comment;
                      delete _helix[i];
                 }
            }
            delete [] _helix;
       }
       if (_turn) {
            for (i = 0; i < _n_turn; i++) {
                 if (_turn[i]) {
                      if (_turn[i]->comment) delete [] _turn[i]->comment;
                      delete _turn[i];
                 }
            }
            delete [] _turn;
       }
       Clear();
}

void _CHAIN::_ManageSpace()
{
       int i = 0;
       if (_SeqLen >= _Alloc_SeqLen) {
            _Alloc_SeqLen += _RES_INCR;
            _FIELD **tmp = _SeqRes;
            _SeqRes = new _FIELD*[_Alloc_SeqLen];
            for (i = 0; i < _SeqLen; i++) _SeqRes[i] = tmp[i];
            for (i = _SeqLen; i < _Alloc_SeqLen; i++) _SeqRes[i] = NULL;
            if (tmp) delete [] tmp;
       }
       if (_Num_Residue >= _Alloc_Residue) {
            _Alloc_Residue += _RES_INCR;
            _RESIDUE **tmp = _residues;
            _residues = new _RESIDUE*[_Alloc_Residue];
            for (i = 0; i < _Num_Residue; i++) _residues[i] = tmp[i];
            for (i = _Num_Residue; i < _Alloc_Residue; i++)
                 _residues[i] = NULL;
            if (tmp) delete [] tmp;
       }
}

void _CHAIN::Empty()
{
       delete [] _residues;
       _residues = NULL;
       _pdbIndex.Clear();
       _ndbIndex.Clear();
       _numIndex.Clear();
       _cifnumIndex.Clear();
       _resIndex.Clear();
       _ResNumIndex.Clear();
}

int _CHAIN::operator==(const _CHAIN& chn)
{
       if (_chain_type == HETAS_TOKEN && chn._chain_type == HETAS_TOKEN) {
            if (!strcmp(_SeqRes[0]->Field[0], chn._SeqRes[0]->Field[0]) &&
                !strcmp(_SeqRes[0]->Field[0], "HOH"))
                 return 1;
       }
/*
       if (_chain_type == ATOMS_TOKEN && chn._chain_type == ATOMS_TOKEN) {
            if (_SeqLen == chn._SeqLen && _SeqLen == 1 &&
               !strcmp(_SeqRes[0]->Field[0], chn._SeqRes[0]->Field[0]))
                 return 1;
            else return 0;
       }
*/       
       if (_chain_type != chn._chain_type || _SeqLen != chn._SeqLen) return 0;
       for (int i = 0; i < _SeqLen; i++) {
            if (_SeqRes[i]->Deleted && chn._SeqRes[i]->Deleted) continue;
            if (_SeqRes[i]->Deleted && !chn._SeqRes[i]->Deleted ||
                !_SeqRes[i]->Deleted && chn._SeqRes[i]->Deleted) return 0;
            if (strcmp(_SeqRes[i]->Field[0], chn._SeqRes[i]->Field[0])) return 0;
       }
       return 1;
}

int _CHAIN::is_same_chain(const _CHAIN& chn)
{
       if (_chain_type == HETAS_TOKEN && chn._chain_type == HETAS_TOKEN) {
            if (!strcmp(_SeqRes[0]->Field[0], chn._SeqRes[0]->Field[0]) &&
                !strcmp(_SeqRes[0]->Field[0], "HOH"))
                 return 1;
       }
       if (_chain_type != chn._chain_type || _SeqLen != chn._SeqLen) return 0;
       if (_chain_type == ATOMS_TOKEN && chn._chain_type == ATOMS_TOKEN) {
            if (_SeqLen > 1 && _ChainID != chn._ChainID) return 0;
       }
       for (int i = 0; i < _SeqLen; i++) {
            if (_SeqRes[i]->Deleted && chn._SeqRes[i]->Deleted) continue;
            if (_SeqRes[i]->Deleted && !chn._SeqRes[i]->Deleted ||
                !_SeqRes[i]->Deleted && chn._SeqRes[i]->Deleted) return 0;
            if (strcmp(_SeqRes[i]->Field[0], chn._SeqRes[i]->Field[0])) return 0;
       }
       return 1;
}

_RESIDUE* _CHAIN::GetFirstResidue()
{
       if (_resIndex.Root() >= 0) {
            _resIndex.GoFirst();
            _IIndex idx = _resIndex.GetCurrData();
            _currResidue = _SeqRes[idx.pos()]->Residue;
            return _currResidue;
       } else return NULL;
}

_RESIDUE* _CHAIN::GetLastResidue()
{
       if (_resIndex.Root() >= 0) {
            _resIndex.GoLast();
            _IIndex idx = _resIndex.GetCurrData();
            _currResidue = _SeqRes[idx.pos()]->Residue;
            return _currResidue;
       } else return NULL;
}

_RESIDUE* _CHAIN::GetPrevResidue()
{
       if (_resIndex.GoToPrev() >= 0) {
            _IIndex idx = _resIndex.GetCurrData();
            _currResidue = _SeqRes[idx.pos()]->Residue;
            return _currResidue;
       } else return NULL;
}

_RESIDUE* _CHAIN::GetNextResidue()
{
       if (_resIndex.GoToNext() >= 0) {
            _IIndex idx = _resIndex.GetCurrData();
            _currResidue = _SeqRes[idx.pos()]->Residue;
            return _currResidue;
       } else return NULL;
}

int _CHAIN::IsLastResidue()
{
       if (_resIndex.Root() >= 0) {
            _IIndex idx = _resIndex.GetCurrData();
            if (idx.pos() == (_SeqLen - 1)) return 1;
       }
       return 0;
}

void  _CHAIN::Insert_a_Helix(const char *id, const int initRes, const int endRes,
        const int helixClass, const char *comment)
{
       int i = 0;
       for (i = 0; i < _n_helix; i++) {
            if (_helix[i]->initRes == initRes &&
                _helix[i]->endRes == endRes) break;
       }
       if (i < _n_helix) return;

       if (_n_helix >= _alloc_helix) {
            _alloc_helix += 10;
            _HELIX **tmp = _helix;
            _helix = new _HELIX*[_alloc_helix];
            for (i = 0; i < _n_helix; i++) _helix[i] = tmp[i];
            for (i = _n_helix; i < _alloc_helix; i++) _helix[i] = NULL;
            if (tmp) delete [] tmp;
       }
       _helix[_n_helix] = new _HELIX;
       strcpy(_helix[_n_helix]->helixID, id);
       _helix[_n_helix]->initRes = initRes;
       _helix[_n_helix]->endRes = endRes;
       _helix[_n_helix]->helixClass = helixClass;
       if (comment) {
            _helix[_n_helix]->comment = new char[strlen(comment)+1];
            strcpy(_helix[_n_helix]->comment, comment);
       } else {
            _helix[_n_helix]->comment = new char;
            _helix[_n_helix]->comment[0] = '\0';
       }
       _n_helix++;
}

void  _CHAIN::Insert_a_Turn(const char *id, const int initRes, const int endRes,
        const char *comment)
{
       int i = 0;
       for (i = 0; i < _n_turn; i++) {
            if (_turn[i]->initRes == initRes &&
                _turn[i]->endRes == endRes) break;
       }
       if (i < _n_turn) return;

       if (_n_turn >= _alloc_turn) {
            _alloc_turn += 10;
            _TURN **tmp = _turn;
            _turn = new _TURN*[_alloc_turn];
            for (i = 0; i < _n_turn; i++) _turn[i] = tmp[i];
            for (i = _n_turn; i < _alloc_turn; i++) _turn[i] = NULL;
            if (tmp) delete [] tmp;
       }
       _turn[_n_turn] = new _TURN;
       strcpy(_turn[_n_turn]->turnID, id);
       _turn[_n_turn]->initRes = initRes;
       _turn[_n_turn]->endRes = endRes;
       if (comment) {
            _turn[_n_turn]->comment = new char[strlen(comment)+1];
            strcpy(_turn[_n_turn]->comment, comment);
       } else {
            _turn[_n_turn]->comment = new char;
            _turn[_n_turn]->comment[0] = '\0';
       }
       _n_turn++;
}

void _CHAIN::remove_helix()
{
       if (_helix) {
            for (int i = 0; i < _n_helix; i++) {
                 if (_helix[i]) {
                      if (_helix[i]->comment) delete [] _helix[i]->comment;
                      delete _helix[i];
                 }
            }
            delete [] _helix;
       }
       _n_helix = 0;
       _helix = NULL;
}

_FIELD* _CHAIN::SeqRes(const int pos)
{
       if (pos >= 0 && pos < _SeqLen)
            return (_SeqRes[pos]);
       else return NULL;
}

int _CHAIN::SeqResPos(const int index)
{
       _IIndex idx;
       idx.set_value(index);
       int ret = _ResNumIndex.Seek(idx);
       if (ret >= 0) {
            idx = _ResNumIndex.GetCurrData();
            return (idx.pos());
       }
       return (-1);
}

_FIELD* _CHAIN::SeqResIndex(const int index)
{
       _IIndex idx;
       idx.set_value(index);
       int ret = _ResNumIndex.Seek(idx);
       if (ret >= 0) {
            idx = _ResNumIndex.GetCurrData();
            return (_SeqRes[idx.pos()]);
       } 
       else return NULL;
}

_HELIX* _CHAIN::helix(const int pos)
{
       if (pos >= 0 && pos < _n_helix)
            return (_helix[pos]);
       else return NULL;
}

_TURN* _CHAIN::turn(const int pos)
{
       if (pos >= 0 && pos < _n_turn)
            return (_turn[pos]);
       else return NULL;
}

int _CHAIN::insert_a_residue(_RESIDUE *residue)
{
       char buffer[50];
       CifString index;
       _IIndex idx;
       _SIndex sindex;

       _ManageSpace();
       _residues[_Num_Residue] = residue;
       _currResidue = residue;
       _SeqRes[_SeqLen] = new _FIELD;
       _SeqRes[_SeqLen]->Deleted = 0;
       _SeqRes[_SeqLen]->ResIndex = _SeqLen;
       _SeqRes[_SeqLen]->Residue = residue;
       _ATOM *atom = residue->GetFirstAtom();
       strcpy(_SeqRes[_SeqLen]->InsCode, atom->ins_code());
       strcpy(_SeqRes[_SeqLen]->Field[0], atom->restype());
       strcpy(_SeqRes[_SeqLen]->Field[1], atom->restype());
       sprintf(_SeqRes[_SeqLen]->Field[2], "%d", atom->resnum());
       strcpy(_SeqRes[_SeqLen]->Field[3], atom->pdb_resnam());
       sprintf(_SeqRes[_SeqLen]->Field[4], "%d", atom->pdb_resnum());
       strcpy(_SeqRes[_SeqLen]->Field[5], atom->pub_resnam());
       sprintf(_SeqRes[_SeqLen]->Field[6], "%d", atom->pub_resnum());

       sprintf(buffer, "%s%s%s", _SeqRes[_SeqLen]->Field[4],
                    _SeqRes[_SeqLen]->Field[3], _SeqRes[_SeqLen]->InsCode);
       sindex.set_value(buffer);
       sindex.set_pos(_SeqLen);
       _pdbIndex.Add(sindex);       

       sprintf(buffer, "%s%s", _SeqRes[_SeqLen]->Field[2], _SeqRes[_SeqLen]->Field[0]);
       sindex.set_value(buffer);
       sindex.set_pos(_SeqLen);
       _ndbIndex.Add(sindex);

       index.Copy(_SeqRes[_SeqLen]->Field[4]);
       index += _SeqRes[_SeqLen]->InsCode;
       _numIndex.Add(index);

       index.Copy(_SeqRes[_SeqLen]->Field[2]);
       _cifnumIndex.Add(index);

       idx.set_value(_SeqLen);
       idx.set_pos(_SeqLen);
       _resIndex.Add(idx);

       _ResNumIndex.Add(idx);

       _ResidueNumbers++;
       _Num_Residue++;
       _SeqLen++;

       return 1;
}

int _CHAIN::insert_a_residue(_RESIDUE *residue, const int pos)
{
       int i = 0;
       char buffer[50], name[10], *p = NULL;
       CifString index;
       _IIndex idx;
       _SIndex sindex;

       if (pos >= _SeqLen) return insert_a_residue(residue);
    
       _ManageSpace();
       _residues[_Num_Residue] = residue;
       _currResidue = residue;
       for (i = _SeqLen - 1; i > pos; i--) _SeqRes[i+1] = _SeqRes[i];
       _SeqRes[pos+1] = new _FIELD;
       _SeqRes[pos+1]->Deleted = 0;
       _SeqRes[pos+1]->ResIndex = _SeqLen;
       _SeqRes[pos+1]->Residue = residue;
       _ATOM *atom = residue->GetFirstAtom();
       strcpy(_SeqRes[pos+1]->InsCode, atom->ins_code());
       strcpy(_SeqRes[pos+1]->Field[0], atom->restype());
       strcpy(_SeqRes[pos+1]->Field[1], atom->restype());
       sprintf(_SeqRes[pos+1]->Field[2], "%d", atom->resnum());
       strcpy(_SeqRes[pos+1]->Field[3], atom->pdb_resnam());
       sprintf(_SeqRes[pos+1]->Field[4], "%d", atom->pdb_resnum());
       strcpy(_SeqRes[pos+1]->Field[5], atom->pub_resnam());
       sprintf(_SeqRes[pos+1]->Field[6], "%d", atom->pub_resnum());
       _ResidueNumbers++;
       _Num_Residue++;
       _SeqLen++;

       idx.set_value(_SeqRes[pos+1]->ResIndex);
       idx.set_pos(pos+1);
       _ResNumIndex.Add(idx);

       _pdbIndex.Clear();
       _ndbIndex.Clear();
       _numIndex.Clear();
       _cifnumIndex.Clear();
       _resIndex.Clear();
       for (i = 0; i < _SeqLen; i++) {
            if (!_SeqRes[i]->Deleted) {
                 if (_SeqRes[i]->Residue && _SeqRes[i]->Residue->no_res_name() > 1) {
                      p = _SeqRes[i]->Residue->name_list().Text();
                      while (p) {
                           p = get_word(p, name);
                           if (!strcmp(name, "")) continue;

                           sprintf(buffer, "%s%s%s", _SeqRes[i]->Field[4], name, 
                                       _SeqRes[i]->InsCode);
                           sindex.set_value(buffer);
                           sindex.set_pos(i);
                           _pdbIndex.Add(sindex);

                           sprintf(buffer, "%s%s", _SeqRes[i]->Field[2], name);
                           sindex.set_value(buffer);
                           sindex.set_pos(i);
                           _ndbIndex.Add(sindex);
                      }
                 } else {
                      sprintf(buffer, "%s%s%s", _SeqRes[i]->Field[4],
                                _SeqRes[i]->Field[3], _SeqRes[i]->InsCode);
                      sindex.set_value(buffer);
                      sindex.set_pos(i);
                      _pdbIndex.Add(sindex);

                      sprintf(buffer, "%s%s", _SeqRes[i]->Field[2], _SeqRes[i]->Field[0]);
                      sindex.set_value(buffer);
                      sindex.set_pos(i);
                      _ndbIndex.Add(sindex);
                 }

                 index.Copy(_SeqRes[i]->Field[4]);
                 index += _SeqRes[i]->InsCode;
                 _numIndex.Add(index);

                 index.Copy(_SeqRes[i]->Field[2]);
                 _cifnumIndex.Add(index);

                 if (_SeqRes[i]->Residue) {
                      idx.set_value(i);
                      idx.set_pos(i);
                      _resIndex.Add(idx);
                 }
            } else {
                 index.Copy("");
                 _numIndex.Add(index);
                 _cifnumIndex.Add(index);
            }
       }
       return 1;
}

int _CHAIN::merge_a_residue(_RESIDUE *new_residue, _RESIDUE *exist_residue)
{
       int ret;
       char value[50];
       _SIndex sindex;

       if (!strcmp(new_residue->ResName(), exist_residue->ResName())) return 0;

       sprintf(value, "%d%s%s", exist_residue->pdb_res_no(), exist_residue->ResName(),
                 exist_residue->GetFirstAtom()->ins_code());
       sindex.set_value(value);
       if (_pdbIndex.Seek(sindex) >= 0) {
            sindex = _pdbIndex.GetCurrData();
            ret = sindex.pos();
            if (ret >= 0) {
                 sprintf(value, "%s%s%s", _SeqRes[ret]->Field[4],
                           new_residue->ResName(), _SeqRes[ret]->InsCode);
                 sindex.set_value(value);
                 sindex.set_pos(ret);
                 _pdbIndex.Add(sindex);

                 sprintf(value, "%s%s", _SeqRes[ret]->Field[2], new_residue->ResName());
                 sindex.set_value(value);
                 sindex.set_pos(ret);
                 _ndbIndex.Add(sindex);

                 return 1;
            }
       }
       return 0;
}

int _CHAIN::delete_a_residue(const char *resname, const int resnum, const char *ins_code)
{
       int ret;
       char value[50], name[10], *p = NULL;
       CifString index;
       _IIndex idx;
       _SIndex sindex;

       sprintf(value, "%d%s%s", resnum, resname, ins_code);
       sindex.set_value(value);
       if (_pdbIndex.Seek(sindex) >= 0) {
            sindex = _pdbIndex.GetCurrData();
            ret = sindex.pos();
            _SeqRes[ret]->Deleted = 1;
            _pdbIndex.Delete();
            if (_SeqRes[ret]->Residue && _SeqRes[ret]->Residue->no_res_name() > 1) {
                 p = _SeqRes[ret]->Residue->name_list().Text();
                 while (p) {
                      p = get_word(p, name);
                      if (!strcmp(name, "")) continue;

                      sprintf(value, "%s%s%s", _SeqRes[ret]->Field[4], name, 
                                       _SeqRes[ret]->InsCode);
                      sindex.set_value(value);
                      if (_pdbIndex.Seek(sindex) >= 0) {
                           _pdbIndex.Delete();
                      }			   
                      sprintf(value, "%s%s", _SeqRes[ret]->Field[2], name);
                      sindex.set_value(value);
                      if (_ndbIndex.Seek(sindex) >= 0) {
                           _ndbIndex.Delete();
                      }
                 }
            }

            sprintf(value, "%s%s", _SeqRes[ret]->Field[2], _SeqRes[ret]->Field[0]);
            sindex.set_value(value);
            if (_ndbIndex.Seek(sindex) >= 0) {
                 _ndbIndex.Delete();
            }

            _numIndex.GoToRecord(ret);
            _numIndex.Delete();

            _cifnumIndex.GoToRecord(ret);
            _cifnumIndex.Delete();

            idx.set_value(ret);
            if (_resIndex.Seek(idx) >= 0) {
                 _resIndex.Delete();
                 _ResidueNumbers--;
            }

            idx.set_value(_SeqRes[ret]->ResIndex);
            if (_ResNumIndex.Seek(idx) >= 0) _ResNumIndex.Delete();
            return 1;
       }
       return 0;
}

int _CHAIN::change_a_residue(_RESIDUE *residue)
{
       int ret;
       char value[50];
       _SIndex sindex;

       sprintf(value, "%d%s%s", residue->pdb_res_no(), residue->ResName(),
                 residue->GetFirstAtom()->ins_code());
       sindex.set_value(value);
       if (_pdbIndex.Seek(sindex) >= 0) {
            sindex = _pdbIndex.GetCurrData();
            ret = sindex.pos();
            if (ret >= 0) {
                 if (_SeqRes[ret]->Residue) delete _SeqRes[ret]->Residue;
                 _SeqRes[ret]->Residue = residue;
                 return 1;
            }
       }
       return 0;
}

void _CHAIN::clone_a_chain(_CHAIN *chn)
{
       int i, j;

       _SeqLen = chn->_SeqLen;
       _Alloc_SeqLen = _SeqLen;
       _SeqRes = new _FIELD*[_SeqLen];
       for (i = 0; i < _SeqLen; i++) {
            _SeqRes[i] = new _FIELD;
            _SeqRes[i]->Deleted = 0;
            _SeqRes[i]->ResIndex = i;
            _SeqRes[i]->Residue = NULL;
            for (j = 0; j <= 6; j++)
                 strcpy(_SeqRes[i]->Field[j], chn->_SeqRes[i]->Field[j]);
            strcpy(_SeqRes[i]->InsCode, chn->_SeqRes[i]->InsCode);
       }
       _chain_type = chn->_chain_type;
       _ChainID = chn->_ChainID;
}

_FIELD* _CHAIN::find_Residue(const int format, const char *resname, const int resnum,
                  const char *ins_code)
{
       if (format == NDB_FILE_FORMAT_NDB)
            return find_ndb_Residue(resname, resnum, ins_code);
       else if (format == NDB_FILE_FORMAT_PDB)
            return find_pdb_Residue(resname, resnum, ins_code);
       else return NULL;
}

_FIELD* _CHAIN::find_ndb_Residue(const char *resname, const int resnum, const char *ins_code)
{
       char value[50];
       _SIndex sindex;

       sprintf(value, "%d%s", resnum, resname);
       sindex.set_value(value);
       if (_ndbIndex.Seek(sindex) >= 0) {
            sindex = _ndbIndex.GetCurrData();
            return _SeqRes[sindex.pos()];
       } else return NULL;
}

_FIELD* _CHAIN::find_pdb_Residue(const char *resname, const int resnum, const char *ins_code)
{
       char value[50];
       _SIndex sindex;

       sprintf(value, "%d%s%s", resnum, resname, ins_code);
       sindex.set_value(value);
       if (_pdbIndex.Seek(sindex) >= 0) {
            sindex = _pdbIndex.GetCurrData();
            return _SeqRes[sindex.pos()];
       } else return NULL;
}

_FIELD* _CHAIN::find_PDB_residue(const char *resname, const int resnum, const char *ins_code,
               int &pos)
{
       char value[50];
       _SIndex sindex;

       pos = -1;
       sprintf(value, "%d%s%s", resnum, resname, ins_code);
       sindex.set_value(value);
       if (_pdbIndex.Seek(sindex) >= 0) {
            sindex = _pdbIndex.GetCurrData();
            pos = sindex.pos();
       }
       if (pos >= 0) return _SeqRes[pos];
       else return NULL;
}

_RESIDUE* _CHAIN::find_residue(const int format, const char *resname, const int resnum,
                const char *ins_code)
{
       if (format == NDB_FILE_FORMAT_NDB)
            return find_ndb_residue(resname, resnum, ins_code);
       else if (format == NDB_FILE_FORMAT_PDB)
            return find_pdb_residue(resname, resnum, ins_code);
       else return NULL;
}
           
_RESIDUE* _CHAIN::find_ndb_residue(const char *resname, const int resnum, const char *ins_code)
{
       _FIELD *field = find_ndb_Residue(resname, resnum, ins_code);
       if (field) return field->Residue;
       else return NULL;
}

_RESIDUE* _CHAIN::find_pdb_residue(const char *resname, const int resnum, const char *ins_code)
{
       _FIELD *field = find_pdb_Residue(resname, resnum, ins_code);
       if (field) return field->Residue;
       else return NULL;
}

_RESIDUE* _CHAIN::find_pdb_residue(const int resnum, const char *ins_code)
{

       char value[10];
       sprintf(value, "%d", resnum);
       CifString index(value);
       index += ins_code;
       int ret = _numIndex.Seek(index);
       if (ret >= 0) return _SeqRes[ret]->Residue;
       else return NULL;
}

_FIELD* _CHAIN::get_pdb_residue(const int resnum, const char *ins_code)
{

       char value[10];
       sprintf(value, "%d", resnum);
       CifString index(value);
       index += ins_code;
       int ret = _numIndex.Seek(index);
       if (ret >= 0) return _SeqRes[ret];
       else return NULL;
}

_RESIDUE* _CHAIN::find_prev_residue(const char *resname, const int resnum, const char *ins_code)
{
       int i, ret;
       char value[50];
       _SIndex sindex;

       sprintf(value, "%d%s%s", resnum, resname, ins_code);
       sindex.set_value(value);
       if (_pdbIndex.Seek(sindex) >= 0) {
            sindex = _pdbIndex.GetCurrData();
            ret = sindex.pos();
            for (i = ret - 1; i >= 0; i++) {
                 if (_SeqRes[i]->Residue) return _SeqRes[i]->Residue;
            }
       }

       return NULL; 
}

_ATOM** _CHAIN::find_atom(const int format, const char *atom_name, const char *resname,
           const int resnum, const char *ins_code, int &num)
{
       if (format == NDB_FILE_FORMAT_NDB)
            return find_ndb_atom(atom_name, resname, resnum, ins_code, num);
       else if (format == NDB_FILE_FORMAT_PDB)
            return find_pdb_atom(atom_name, resname, resnum, ins_code, num);
       else return NULL;
}

_ATOM** _CHAIN::find_ndb_atom(const char *atom_name, const char *resname, const int resnum,
         const char *ins_code, int &num)
{
       _RESIDUE *residue = find_ndb_residue(resname, resnum, ins_code);
       if (residue) return (residue->find_atom(atom_name, num));
       return NULL;
}

_ATOM** _CHAIN::find_pdb_atom(const char *atom_name, const char *resname, const int resnum,
         const char *ins_code, int &num)
{
       _RESIDUE *residue = find_pdb_residue(resname, resnum, ins_code);
       if (residue) return (residue->find_atom(atom_name, num));
       return NULL;
}

void _CHAIN::seq_alignment(SEQ *seq, SEQ *pdb)
{
       int i = 0, /* same_sequence = 1, */ repeat = 0, found = 0, found_begin = 0;
       int found_end = 0, begin = -1, end = -1, first, second, exist_pdbnum = 0;
       int found_gap = 0, k = 0, missing_sequence = 0, last = -1, seq_num, j;
       char name1[10], name2[10], buffer[50], *p = NULL;
       ALIGN *sa = NULL, *sb = NULL, *ss = NULL;
       CifString index;
       _IIndex idx;
       SEQ *exist = NULL;
       _SIndex sindex;

       _seqa = seq;
       if (_seqa == NULL || _chain_type != ATOMP_TOKEN &&
           _chain_type != ATOMN_TOKEN && _chain_type != ATOMS_TOKEN) {
            for (i = 0; i < _SeqLen; i++) {
                 strcpy(_SeqRes[i]->Field[0], _SeqRes[i]->Field[1]);
                 sprintf(_SeqRes[i]->Field[2], "%d", i + 1);
            }
            return;
       }

       _has_sequence = 1;
/*
       for (i = 1; i < _seqa->SeqLen; i++) {
            if (res_name_cmp(_seqa->res[i]->name, _seqa->res[0]->name)) {
                 same_sequence = 0;
                 break;
            }
       }
*/
       sa = new ALIGN;
       sa->AlignLength = _seqa->SeqLen;
       sa->aln = new ALN[sa->AlignLength];
       for (i = 0; i < sa->AlignLength; i++) sa->aln[i].pos[0] = i;

       sb = new ALIGN;
       sb->AlignLength = _SeqLen;
       sb->aln = new ALN[sb->AlignLength];
       for (i = 0; i < sb->AlignLength; i++) sb->aln[i].pos[0] = i;

       _seqb = new SEQ;
       _seqb->SeqLen = _SeqLen;
       _seqb->AllocLen = _SeqLen;
       _seqb->res = new RES*[_SeqLen];
       for (i = 0; i < _SeqLen; i++) {
            _seqb->res[i] = new RES;
            strcpy(_seqb->res[i]->name, _SeqRes[i]->Field[1]);
       }

       if (pdb && pdb->SeqLen == seq->SeqLen) {
            exist_pdbnum = 1;
            _pdba = pdb;
            _pdbb = new SEQ;
            _pdbb->SeqLen = _SeqLen;
            _pdbb->AllocLen = _SeqLen;
            _pdbb->res = new RES*[_SeqLen];
            for (i = 0; i < _SeqLen; i++) {
                 _pdbb->res[i] = new RES;
                 strcpy(_pdbb->res[i]->name, _SeqRes[i]->Field[4]);
            }
       }
  
       ss = alignment(sa, sb, 2, score);

       if (!exist_pdbnum) {
            name1[0] = name2[0] = '\0';
            if (_chain_type == ATOMP_TOKEN) {
                 strcpy(name1, "C");
                 strcpy(name2, "N");
            } else if (_chain_type == ATOMN_TOKEN) {
                 strcpy(name1, "O3*");
                 strcpy(name2, "P");
            }
            while (1) {
                 found = 0; found_begin = 0; found_end = 0;
                 begin = -1; end = -1;
                 for (i = 0; i < ss->AlignLength; i++) {
                      if (i == 0 && ss->aln[i].pos[0] >= 0 && ss->aln[i].pos[1] < 0) {
                           begin = -1; found_begin = 1;
                      } else if ((i+1) < ss->AlignLength) {
                           if (ss->aln[i].pos[0] >= 0 && ss->aln[i].pos[1] >= 0 &&
                               ss->aln[i+1].pos[0] >= 0 && ss->aln[i+1].pos[1] < 0) {
                                begin = i; found_begin = 1;
                           }
                      }
                      if (i && found_begin) {
                           if (ss->aln[i].pos[0] >= 0 && ss->aln[i].pos[1] >= 0 &&
                               ss->aln[i-1].pos[0] >= 0 && ss->aln[i-1].pos[1] < 0) {
                                end = i; found_end = 1;
                           }
                      }
                      if (found_begin && found_end) {
                           if (begin == -1) {
/*
                                if (!res_name_cmp(_seqa->res[ss->aln[begin+1].pos[0]]->name,
                                     _seqb->res[ss->aln[end].pos[1]]->name) && same_sequence) {
                                      ss->aln[begin+1].pos[1] = ss->aln[end].pos[1];
                                      ss->aln[end].pos[1] = -1;
                                      found = 1;
                                }
*/
                           } else if (!res_name_cmp(_seqa->res[ss->aln[begin+1].pos[0]]->name,
                                  _seqb->res[ss->aln[end].pos[1]]->name)) {
                                first = ss->aln[end].pos[1] - 1;
                                second = first + 1;
                                if (first >= 0 && second < _SeqLen) {
                                     if (is_connect(name1, _SeqRes[first]->Residue,
                                               name2, _SeqRes[second]->Residue)) {
                                          ss->aln[begin+1].pos[1] = ss->aln[end].pos[1];
                                          ss->aln[end].pos[1] = -1;
                                          found = 1;
                                     } else if (atoi(_SeqRes[second]->Field[4]) ==
                                               (atoi(_SeqRes[first]->Field[4]) + 1)) {
                                          ss->aln[begin+1].pos[1] = ss->aln[end].pos[1];
                                          ss->aln[end].pos[1] = -1;
                                          found = 1;
                                     }
                                }
                           } else if (!res_name_cmp(_seqb->res[ss->aln[begin].pos[1]]->name,
                                  _seqa->res[ss->aln[end-1].pos[0]]->name)) {
                                first = ss->aln[begin].pos[1];
                                second = first + 1;
                                if (first >= 0 && second < _SeqLen &&
                                    is_connect(name1, _SeqRes[first]->Residue,
                                               name2, _SeqRes[second]->Residue)) {
                                     ss->aln[end-1].pos[1] = ss->aln[begin].pos[1];
                                     ss->aln[begin].pos[1] = -1;
                                     found = 1;
                                }
                           }
                           found_begin = 0;
                           found_end = 0;
                      }
                 }
                 if (!found) break;
                 repeat++;
                 if (repeat > 20) break;
            }
       }

       if (ss->AlignLength > _Alloc_SeqLen) {
            _Alloc_SeqLen = ss->AlignLength;
            _FIELD **tmp = _SeqRes;
            _SeqRes = new _FIELD*[_Alloc_SeqLen];
            for (i = 0; i < _SeqLen; i++) _SeqRes[i] = tmp[i];
            for (i = _SeqLen; i < _Alloc_SeqLen; i++) _SeqRes[i] = NULL;
            delete [] tmp;
       }
       _SeqLen = ss->AlignLength;

       for (i = ss->AlignLength - 1; i >= 0; i--) {
            if (ss->aln[i].pos[1] >= 0) {
                 _SeqRes[i] = _SeqRes[ss->aln[i].pos[1]];
                 if (i != ss->aln[i].pos[1])
                      _SeqRes[ss->aln[i].pos[1]] = NULL;
            } else {
                 if (!_SeqRes[i]) _SeqRes[i] = new _FIELD;
                 _SeqRes[i]->Residue = NULL;
                 _SeqRes[i]->ResIndex = -1;
                 _SeqRes[i]->Deleted = 0;
                 for (k = 1; k < NUM_FIELD; k++) {
                      strcpy(_SeqRes[i]->Field[k], "?");
                 }
                 _SeqRes[i]->InsCode[0] = '\0';
                 found_gap = 1;
            }
            if (!_SeqRes[i]) {
                  _SeqRes[i] = new _FIELD;
                 _SeqRes[i]->Residue = NULL;
                 _SeqRes[i]->ResIndex = -1;
                 _SeqRes[i]->Deleted = 0;
                 for (k = 1; k < NUM_FIELD; k++) {
                      strcpy(_SeqRes[i]->Field[k], "?");
                 }
                 _SeqRes[i]->InsCode[0] = '\0';
            }
            if (ss->aln[i].pos[0] >= 0) {
                 strcpy(_SeqRes[i]->Field[0],
                         _seqa->res[ss->aln[i].pos[0]]->name);
            } else {
                 strcpy(_SeqRes[i]->Field[0], "?");
                 missing_sequence = 1;
            }
       }

       _seqa = NULL; _pdba = NULL;
       delete [] sa->aln; delete sa;
       delete [] sb->aln; delete sb;
       delete [] ss->aln; delete ss;
       for (i = 0; i < _seqb->SeqLen; i++) delete _seqb->res[i];
       delete [] _seqb->res; delete _seqb;
       _seqb = NULL;
       if (_pdbb) {
            for (i = 0; i < _pdbb->SeqLen; i++) delete _pdbb->res[i];
            delete [] _pdbb->res; delete _pdbb;
            _pdbb = NULL;
       }

       if (pdb && pdb->SeqLen == _SeqLen) exist = pdb;

       if (found_gap) {
            if (!missing_sequence && (_chain_type == ATOMN_TOKEN ||
                                      _chain_type == ATOMP_TOKEN)) {
                 if (_chain_type == ATOMP_TOKEN) {
                      strcpy(name1, "C");
                      strcpy(name2, "N");
                 } else if (_chain_type == ATOMN_TOKEN) {
                      strcpy(name1, "O");
                      strcpy(name2, "P");
                 }

                 int *single = new int[_SeqLen];
                 first = 1;
                 for (i = 0; i < _SeqLen; i++) {
                      single[i] = 0;
                      if (_SeqRes[i]->Residue) {
                           if (exist || !first) {
                                if (i == 0) {
                                     if (!_SeqRes[i+1]->Residue)
                                          single[i] = 1;
                                     else if (!_SeqRes[i]->Residue->is_single_atom() &&
                                              !_SeqRes[i+1]->Residue->is_single_atom() &&
                                              !is_connect(_SeqRes[i]->Residue,
                                              _SeqRes[i+1]->Residue, name1, name2))
                                          single[i] = 1;
                                } else if (i == (_SeqLen - 1)) {
                                     if (!_SeqRes[i-1]->Residue)
                                          single[i] = 1;
                                     else if (!_SeqRes[i-1]->Residue->is_single_atom() &&
                                              !_SeqRes[i]->Residue->is_single_atom() &&
                                              !is_connect(_SeqRes[i-1]->Residue,
                                              _SeqRes[i]->Residue, name1, name2))
                                          single[i] = 1;
                                } else {
                                     if (!_SeqRes[i-1]->Residue && !_SeqRes[i+1]->Residue)
                                          single[i] = 1;
                                     else if (!_SeqRes[i-1]->Residue &&
                                              !_SeqRes[i]->Residue->is_single_atom() &&
                                              !_SeqRes[i+1]->Residue->is_single_atom() &&
                                              !is_connect(_SeqRes[i]->Residue,
                                              _SeqRes[i+1]->Residue, name1, name2))
                                          single[i] = 1;
                                     else if (!_SeqRes[i+1]->Residue &&
                                              !_SeqRes[i-1]->Residue->is_single_atom() &&
                                              !_SeqRes[i]->Residue->is_single_atom() &&
                                              !is_connect(_SeqRes[i-1]->Residue,
                                              _SeqRes[i]->Residue, name1, name2))
                                          single[i] = 1;
                                }
                           }
                           first = 0;
                      }
                 }
 
                 for (i = 0; i < _SeqLen; i++) {
                      if (!single[i]) continue;
                      begin = i;
                      for (j = i - 1; j >= 0; j--) {
                           if (_SeqRes[j]->Residue) break;
                           begin = j;
                      }
                      end = i;
                      for (j = i + 1; j < _SeqLen; j++) {
                           if (_SeqRes[j]->Residue) break;
                           end = j;
                      }
                      if (begin == end) continue;
                      last = 99999999;
                      k = i;
                      for (j = begin; j <= end; j++) {
                           if (!strcmp(_SeqRes[j]->Field[0], _SeqRes[i]->Field[0])) {
                                if (exist) {
                                     if (abs(atoi(exist->res[j]->name) -
                                              _SeqRes[i]->Residue->pdb_res_no()) < last) {
                                          last = abs(atoi(exist->res[j]->name)
                                               - _SeqRes[i]->Residue->pdb_res_no());
                                          k = j;
                                     }
                                } else if (abs(atoi(_SeqRes[j]->Field[3]) -
                                              _SeqRes[i]->Residue->res_no()) < last) {
                                     last = abs(atoi(_SeqRes[j]->Field[3])
                                          - _SeqRes[i]->Residue->res_no());
                                     k = j;
                                }
                           }
                      }
                      if (k != i) {
                           *(_SeqRes[k]) = *(_SeqRes[i]);
                           sprintf(_SeqRes[k]->Field[4], "%d", _SeqRes[i]->Residue->pdb_res_no());
                           _SeqRes[i]->Residue = NULL;
                           _SeqRes[i]->ResIndex = -1;
                           _SeqRes[i]->Deleted = 0;
                           for (j = 1; j < NUM_FIELD; j++) {
                                strcpy(_SeqRes[i]->Field[j], "?");
                           }
                           _SeqRes[i]->InsCode[0] = '\0';
                      }
                 }
                 delete [] single;
            }

            found_gap = 0;
            last = -1;
            for (i = 0; i < _SeqLen; i++) {
                 if (!_SeqRes[i]->Residue) found_gap = 1;
                 else {
                      last = i;
                      if (found_gap) {
                           seq_num = atoi(_SeqRes[i]->Field[4]);
                           for (j = i - 1; j >= 0; j--) {
                                if (_SeqRes[j]->Residue) break;
                                seq_num--;
                                sprintf(_SeqRes[j]->Field[4], "%d", seq_num);
                           }
                           found_gap = 0;
                      }
                 }
            }
            if (/* found_gap && */ last >= 0) {
                 seq_num = atoi(_SeqRes[last]->Field[4]);
                 for (i = last + 1; i < _SeqLen; i++) {
                      if (_SeqRes[i]->Residue) break;
                      seq_num++;
                      sprintf(_SeqRes[i]->Field[4], "%d", seq_num);
                 }
            }
       }

       if (exist) {
            for (i = 0; i < _SeqLen; i++) {
                 strcpy(_SeqRes[i]->Field[4], exist->res[i]->name);
            }
       }

       _ndbIndex.Clear();
       _pdbIndex.Clear();
       _numIndex.Clear();
       _cifnumIndex.Clear();
       _resIndex.Clear();
       _ResNumIndex.Clear();
       for (i = 0; i < _SeqLen; i++) {
            if (!_SeqRes[i]->Deleted) {
                 if (_SeqRes[i]->Residue && _SeqRes[i]->Residue->no_res_name() > 1) {
                      p = _SeqRes[i]->Residue->name_list().Text();
                      while (p) {
                           p = get_word(p, name1);
                           if (!strcmp(name1, "")) continue;

                           sprintf(buffer, "%s%s%s", _SeqRes[i]->Field[4], name1, 
                                       _SeqRes[i]->InsCode);
                           sindex.set_value(buffer);
                           sindex.set_pos(i);
                           _pdbIndex.Add(sindex);

                           sprintf(buffer, "%s%s", _SeqRes[i]->Field[2], name1);
                           sindex.set_value(buffer);
                           sindex.set_pos(i);
                           _ndbIndex.Add(sindex);
                      }
                 } else {
                      sprintf(buffer, "%s%s", _SeqRes[i]->Field[2], _SeqRes[i]->Field[0]);
                      sindex.set_value(buffer);
                      sindex.set_pos(i);
                      _ndbIndex.Add(sindex);

                      sprintf(buffer, "%s%s%s", _SeqRes[i]->Field[4],
                                 _SeqRes[i]->Field[3], _SeqRes[i]->InsCode);
                      sindex.set_value(buffer);
                      sindex.set_pos(i);
                      _pdbIndex.Add(sindex);
                 }

                 index.Copy(_SeqRes[i]->Field[4]);
                 index += _SeqRes[i]->InsCode;
                 _numIndex.Add(index);
     
                 index.Copy(_SeqRes[i]->Field[2]);
                 _cifnumIndex.Add(index);
     
                 if (!_SeqRes[i]->Residue) continue;
                 idx.set_value(i);
                 idx.set_pos(i);
                 _resIndex.Add(idx);

                 idx.set_value(_SeqRes[i]->ResIndex);
                 _ResNumIndex.Add(idx);
            } else {
                 index.Copy("");
                 _numIndex.Add(index);
                 _cifnumIndex.Add(index);
            }
       }
}

int _CHAIN::print_mismatch(CifString &message, int &first, int &count, int &count1)
{
       int i = 0,  beg = -1, end = -1, mismatch = 0, m, n, num1 = 0, num2 = 0, fst = 1;
       char F_atom[5], F_atom1[5], S_atom[5], buffer[200];
       double dist;
       _ATOM /* *atom1 = NULL, *atom2 = NULL, */ **atm1 = NULL, **atm2 = NULL;

       if (_chain_type != ATOMN_TOKEN && _chain_type != ATOMP_TOKEN) return mismatch;

       if (_chain_type == ATOMN_TOKEN) {
            strcpy(F_atom1, "O3*");
            strcpy(F_atom, "O");
            strcpy(S_atom, "P");
       } else {
            strcpy(F_atom1, "C");
            strcpy(F_atom, "C");
            strcpy(S_atom, "N");
       }

       mismatch = 0;
       for (i = 0; i < _SeqLen; i++) {
            if (res_name_cmp(_SeqRes[i]->Field[0], _SeqRes[i]->Field[1]) &&
                strcmp(_SeqRes[i]->Field[1], "?")) {
                 if (first) {
                      message += "\n";
                      first = 0;
                 }
                 sprintf(buffer, "   MISMATCH: ChainID=%s ResNum=%s (%s <---> %s)\n",
                         _PDB_ChainID, _SeqRes[i]->Field[4],
                         _SeqRes[i]->Field[0], _SeqRes[i]->Field[1]);
                 message += buffer;
                 if (!strcmp(_SeqRes[i]->Field[1], "ALA") ||
                     !strcmp(_SeqRes[i]->Field[1], "GLY")) mismatch++;
                 count++;
            }
       }

       for (i = 0; i < _SeqLen - 1; i++) {
            if (_SeqRes[i]->Residue && _SeqRes[i+1]->Residue) {
/*
                 && !is_connect(F_atom, _SeqRes[i]->Residue,
                            S_atom, _SeqRes[i+1]->Residue)) {
                !_SeqRes[i]->Residue->is_single_atom() &&
                !_SeqRes[i+1]->Residue->is_single_atom() &&
                !is_connect(_SeqRes[i]->Residue, _SeqRes[i+1]->Residue,
                         F_atom, S_atom)) {
                 sprintf(buffer, "   SEQUENCE WARNING: Residue (%s %s %s %s) and",
                           _PDB_ChainID, _SeqRes[i]->Field[1],
                           _SeqRes[i]->Field[4], _SeqRes[i]->InsCode);
                 message += buffer;
                 sprintf(buffer, " Residue (%s %s %s %s) are not linked\n",
                           _PDB_ChainID, _SeqRes[i+1]->Field[1],
                           _SeqRes[i+1]->Field[4], _SeqRes[i+1]->InsCode);
                 message += buffer;
                 atom1 = _SeqRes[i]->Residue->find_atom(F_atom1);
                 atom2 = _SeqRes[i+1]->Residue->find_atom(S_atom);
                 if (atom1 && atom2) {
                      dist = cal_distance(atom1, atom2);
                      sprintf(buffer, "Distance of %s-%s bond is %.2f\n",
                          atom1->atmtype(), atom2->atmtype(), dist);
                      message += buffer;
                 }
                 count1++;
*/
                 atm1 = _SeqRes[i]->Residue->find_atom(F_atom1, num1);
                 atm2 = _SeqRes[i+1]->Residue->find_atom(S_atom, num2);
                 fst = 1;
                 if (atm1 && atm2) {
                      for (m = 0; m < num1; m++) {
                           for (n = 0; n < num2; n++) {
                                if (!strcmp(atm1[m]->alt_loc(), "") ||
                                    !strcmp(atm2[n]->alt_loc(), "") ||
                                    !strcmp(atm1[m]->alt_loc(),
                                            atm2[n]->alt_loc())) {
                                     if (!is_a_bond(atm1[m], atm2[n])) {
                                          if (fst) {
                                               fst = 0;
                                               sprintf(buffer, "   SEQUENCE WARNING: Residue (%s %s %s %s) and",
                                                         _PDB_ChainID, _SeqRes[i]->Field[1],
                                                         _SeqRes[i]->Field[4], _SeqRes[i]->InsCode);
                                               message += buffer;
                                               sprintf(buffer, " Residue (%s %s %s %s) are not linked\n",
                                                         _PDB_ChainID, _SeqRes[i+1]->Field[1],
                                                         _SeqRes[i+1]->Field[4], _SeqRes[i+1]->InsCode);
                                               message += buffer;
                                          }
                                          dist = cal_distance(atm1[m], atm2[n]);
                                          if (strcmp(atm1[m]->alt_loc(), "") && strcmp(atm2[n]->alt_loc(), ""))
                                               sprintf(buffer, "Distance of %s(%s)-%s(%s) bond is %.2f\n",
                                                     atm1[m]->atmtype(), atm1[m]->alt_loc(),
                                                     atm2[n]->atmtype(), atm2[n]->alt_loc(), dist);
                                          else if (strcmp(atm1[m]->alt_loc(), ""))
                                               sprintf(buffer, "Distance of %s(%s)-%s bond is %.2f\n",
                                                     atm1[m]->atmtype(), atm1[m]->alt_loc(),
                                                     atm2[n]->atmtype(), dist);
                                          else if (strcmp(atm2[n]->alt_loc(), ""))
                                               sprintf(buffer, "Distance of %s-%s(%s) bond is %.2f\n",
                                                     atm1[m]->atmtype(), atm2[n]->atmtype(),
                                                     atm2[n]->alt_loc(), dist);
                                          else sprintf(buffer, "Distance of %s-%s bond is %.2f\n",
                                                     atm1[m]->atmtype(), atm2[n]->atmtype(), dist);
                                          message += buffer;
                                          count1++;
                                     }
                                }
                           }
                      }
                 }
                 if (atm1) delete [] atm1;
                 if (atm2) delete [] atm2;
            }
       }

       for (i = 0; i < _SeqLen; i++) {
            if (strcmp(_SeqRes[i]->Field[0], "?") &&
                strcmp(_SeqRes[i]->Field[1], "?") && beg >= 0 && i &&
                strcmp(_SeqRes[i-1]->Field[0], "?") &&
               !strcmp(_SeqRes[i-1]->Field[1], "?")) end = i;
            if (beg >= 0 && end >= 0) {
                 if (_SeqRes[beg]->Residue && _SeqRes[end]->Residue &&
                     is_connect(_SeqRes[beg]->Residue, _SeqRes[end]->Residue, F_atom, S_atom)) {
                      sprintf(buffer, "   SEQUENCE WARNING: Residue (%s %s %s %s) and",
                                _PDB_ChainID, _SeqRes[beg]->Field[1],
                                _SeqRes[beg]->Field[4], _SeqRes[beg]->InsCode);
                      message += buffer;
                      sprintf(buffer, " Residue (%s %s %s %s) are linked together\n",
                                _PDB_ChainID, _SeqRes[end]->Field[1],
                                _SeqRes[end]->Field[4], _SeqRes[end]->InsCode);
                      message += buffer;
                      count1++;
                 }
                 beg = -1; end = -1;
            }
            if (strcmp(_SeqRes[i]->Field[0], "?") &&
                strcmp(_SeqRes[i]->Field[1], "?") && (i+1) < _SeqLen &&
                strcmp(_SeqRes[i+1]->Field[0], "?") &&
               !strcmp(_SeqRes[i+1]->Field[1], "?")) beg = i;
       }
       return mismatch;
}

void _CHAIN::print_alignment(CifString &message, const int num_per_line)
{
       int i = 0;
       char tmp_line[200], name[5], buffer[200];
       if (_chain_type != ATOMN_TOKEN && _chain_type != ATOMP_TOKEN) return;
       for (i = 0; i < _SeqLen; i++) {
            if (!strcmp(_SeqRes[i]->Field[0], "?") ||
                !strcmp(_SeqRes[i]->Field[1], "?")) break;
       }
       if (i == _SeqLen) return;
       
       sprintf(buffer, "\n   PDB Chain_ID: %s\n\n", _PDB_ChainID);
       message += buffer;


       int l = _SeqLen / num_per_line;
       int x = _SeqLen % num_per_line;
       int m = l;
       if (x == 0) m = l - 1;
       int num = 0, n, y;
       int first = -1, first_num = 0, last = -1, last_num = 0;
       for (i = 0; i <= m; i++) {
            n = num_per_line;
            if (i == l) n = x;

            message += "           ";
            first = -1;
            last = -1;
            for (y = 0; y < n; y++) {
                 if (strcmp(_SeqRes[i * num_per_line + y]->Field[0], "?")) {
                      num++;
                      if (first == -1) {
                           first = y;
                           first_num = num;
                      }
                      last = y;
                      last_num = num;
                 }
            }
            for (y = 0; y < n; y++) {
                 if (y == first)       sprintf(buffer, "%-4d", first_num);
                 else if (y == last)   sprintf(buffer, "%4d", last_num);
                 else if (y == last-1) sprintf(buffer, "   ");
                 else                  sprintf(buffer, "    ");
                 message += buffer;
            }
            message += "\n   SEQRES: ";
            for (y = 0; y < n; y++) {
                 sprintf(buffer, "%3s ", _SeqRes[i * num_per_line + y]->Field[0]);
                 message += buffer;
            }
            message += "\n   COORDS: ";
            int error = 0;
            strcpy(tmp_line, "           ");
            for (y = 0; y < n; y++) {
                 strcpy(name, _SeqRes[i * num_per_line + y]->Field[1]);
                 if (res_name_cmp(_SeqRes[i * num_per_line + y]->Field[0],
                                   _SeqRes[i * num_per_line + y]->Field[1]) &&
                     strcmp(_SeqRes[i * num_per_line + y]->Field[1], "?")) {
                      error = 1;
                      sprintf(&tmp_line[strlen(tmp_line)], "%3s ", "^^^");
                      ndb_set_string_to_lower(name);
                 } else sprintf(&tmp_line[strlen(tmp_line)], "%3s ", "   ");
                 sprintf(buffer, "%3s ", name);
                 message += buffer;
            }
            message += "\n           ";
            first = -1; last = -1;
            for (y = 0; y < n; y++) {
                 if (strcmp(_SeqRes[i * num_per_line + y]->Field[1], "?")) {
                      if (first == -1) first = y;
                      last = y;
                 }
            }
            for (y = 0; y < n; y++) {
                 if (y == first)
                      sprintf(buffer, "%-4s", _SeqRes[i * num_per_line + y]->Field[4]);
                 else if (y == last)
                      sprintf(buffer, "%4s", _SeqRes[i * num_per_line + y]->Field[4]);
                 else if (y == last - 1)
                      sprintf(buffer, "   ");
                 else sprintf(buffer, "    ");
                 message += buffer;
            }
            message += "\n";
            if (error) { message += tmp_line; message += "\n"; }
            message += "\n";
       }
}

void _CHAIN::update_residues_nomenclature(const int ndb_only)
{
       int i = 0, j = 0;
       char type[10], buffer[50], name[10], *p = NULL;
       CifString index;
       _IIndex idx;
       _RESIDUE *residue = NULL;
       _SIndex sindex;

       type[0] = '\0';
       if (_chain_type > 0) strcpy(type, NdbToken[_chain_type-1].TokenName);

       _ndbIndex.Clear();
       _cifnumIndex.Clear();
       if (!ndb_only) {
            _pdbIndex.Clear();
            _numIndex.Clear();
       }
       _resIndex.Clear();
       for (i = 0; i < _SeqLen; i++) {
            if (!_SeqRes[i]->Deleted) {
                 sprintf(_SeqRes[i]->Field[2], "%d", ++j);
                 if (_SeqRes[i]->Residue && _SeqRes[i]->Residue->no_res_name() > 1) {
                      p = _SeqRes[i]->Residue->name_list().Text();
                      while (p) {
                           p = get_word(p, name);
                           if (!strcmp(name, "")) continue;

                           sprintf(buffer, "%s%s", _SeqRes[i]->Field[2], name);
                           sindex.set_value(buffer);
                           sindex.set_pos(i);
                           _ndbIndex.Add(sindex);

                           if (!ndb_only) {
                                sprintf(buffer, "%s%s%s", _SeqRes[i]->Field[4], name, 
                                            _SeqRes[i]->InsCode);
                                sindex.set_value(buffer);
                                sindex.set_pos(i);
                                _pdbIndex.Add(sindex);
                           }
                      }
                 } else {
                      sprintf(buffer, "%s%s", _SeqRes[i]->Field[2], _SeqRes[i]->Field[0]);
                      sindex.set_value(buffer);
                      sindex.set_pos(i);
                      _ndbIndex.Add(sindex);

                      if (!ndb_only) {
                           sprintf(buffer, "%s%s%s", _SeqRes[i]->Field[4],
                                 _SeqRes[i]->Field[3], _SeqRes[i]->InsCode);
                           sindex.set_value(buffer);
                           sindex.set_pos(i);
                           _pdbIndex.Add(sindex);
                      }
                 }

                 if (!ndb_only) {
                      index.Copy(_SeqRes[i]->Field[4]);
                      index += _SeqRes[i]->InsCode;
                      _numIndex.Add(index);
                 }

                 index.Copy(_SeqRes[i]->Field[2]);
                 _cifnumIndex.Add(index);

                 residue = _SeqRes[i]->Residue;
                 if (!residue) continue;
                 if (ndb_only)
                      residue->update_nomenclature(type, _ChainID.Text(), atoi(_SeqRes[i]->Field[2]));
                 else residue->update_nomenclature(type, _ChainID.Text(), atoi(_SeqRes[i]->Field[2]),
                           _PDB_ChainID, atoi(_SeqRes[i]->Field[4]));
     
                 idx.set_value(i);
                 idx.set_pos(i);
                 _resIndex.Add(idx);
            } else {
                 index.Copy("");
		 _cifnumIndex.Add(index);
                 if (!ndb_only) {
                      _numIndex.Add(index);
                 }
            }
       }
}

void _CHAIN::update_residues_pdb_numbers(int &serial_no)
{
       int i = 0, j = 0;
       char type[10], buffer[50], name[10], *p = NULL;
       CifString index;
       _IIndex idx;
       _RESIDUE *residue = NULL;
       _SIndex sindex;

       type[0] = '\0';
       if (_chain_type > 0) strcpy(type, NdbToken[_chain_type-1].TokenName);

       _ndbIndex.Clear();
       _pdbIndex.Clear();
       _numIndex.Clear();
       _cifnumIndex.Clear();
       _resIndex.Clear();
       for (i = 0; i < _SeqLen; i++) {
            if (!_SeqRes[i]->Deleted) {
                 sprintf(_SeqRes[i]->Field[2], "%d", ++j);
                 serial_no++;
                 sprintf(_SeqRes[i]->Field[4], "%d", serial_no);
                 if (_SeqRes[i]->Residue && _SeqRes[i]->Residue->no_res_name() > 1) {
                      p = _SeqRes[i]->Residue->name_list().Text();
                      while (p) {
                           p = get_word(p, name);
                           if (!strcmp(name, "")) continue;

                           sprintf(buffer, "%s%s%s", _SeqRes[i]->Field[4], name, 
                                       _SeqRes[i]->InsCode);
                           sindex.set_value(buffer);
                           sindex.set_pos(i);
                           _pdbIndex.Add(sindex);

                           sprintf(buffer, "%s%s", _SeqRes[i]->Field[2], name);
                           sindex.set_value(buffer);
                           sindex.set_pos(i);
                           _ndbIndex.Add(sindex);
                      }
                 } else {
                      sprintf(buffer, "%s%s", _SeqRes[i]->Field[2], _SeqRes[i]->Field[0]);
                      sindex.set_value(buffer);
                      sindex.set_pos(i);
                      _ndbIndex.Add(sindex);

                      sprintf(buffer, "%s%s%s", _SeqRes[i]->Field[4],
                               _SeqRes[i]->Field[3], _SeqRes[i]->InsCode);				  
                      sindex.set_value(buffer);
                      sindex.set_pos(i);
                      _pdbIndex.Add(sindex);
                 }

                 index.Copy(_SeqRes[i]->Field[4]);
                 index += _SeqRes[i]->InsCode;
                 _numIndex.Add(index);
     
                 index.Copy(_SeqRes[i]->Field[2]);
                 _cifnumIndex.Add(index);
     
                 residue = _SeqRes[i]->Residue;
                 if (!residue) continue;
                 residue->update_nomenclature(type, _ChainID.Text(), atoi(_SeqRes[i]->Field[2]),
                           _PDB_ChainID, atoi(_SeqRes[i]->Field[4]));
     
                 idx.set_value(i);
                 idx.set_pos(i);
                 _resIndex.Add(idx);
            } else {
                 index.Copy("");
                 _numIndex.Add(index);
                 _cifnumIndex.Add(index);
            }
       }
}

void _CHAIN::correction_residues_name()
{
       int i = 0;
       char buffer[50];
       _RESIDUE *residue = NULL;
       _SIndex sindex;

       for (i = 0; i < _SeqLen; i++) {
            residue = _SeqRes[i]->Residue;
            if (strcmp(_SeqRes[i]->Field[0], _SeqRes[i]->Field[1]) && residue) {
                 sprintf(buffer, "%s%s%s", _SeqRes[i]->Field[4],
                       _SeqRes[i]->Field[3], _SeqRes[i]->InsCode);
                 sindex.set_value(buffer);
                 if (_pdbIndex.Seek(sindex) >= 0) {
                      _pdbIndex.Delete();
                 }
                 sprintf(buffer, "%s%s%s", _SeqRes[i]->Field[4],
                       _SeqRes[i]->Field[0], _SeqRes[i]->InsCode);
                 sindex.set_value(buffer);
                 sindex.set_pos(i);
                 _pdbIndex.Add(sindex);

                 strcpy(_SeqRes[i]->Field[1], _SeqRes[i]->Field[0]);
                 strcpy(_SeqRes[i]->Field[3], _SeqRes[i]->Field[0]);
                 residue->correction_name(_SeqRes[i]->Field[0]);
            }
       }
}

void _CHAIN::get_weight()
{
       static float /* Oxygen = 15.9994, */ Water = 18.0152, PO2 = 62.97256;
    
       _weight = 0.0;
       if (_chain_type == ATOMN_TOKEN ||
           _chain_type == ATOMP_TOKEN ||
           _chain_type == ATOMS_TOKEN) {
            for (int i = 0; i < _SeqLen; i++) {
                 if (!strcmp(_SeqRes[i]->Field[0], "?")) continue;

                 _weight += get_molecule_weight(_SeqRes[i]->Field[0]);
/*
                 if (_residues[_SeqRes[i].index].type() == MONOMER_TYPE_DNA)
                      _weight -= Oxygen;
*/
            }
            _weight -= (double) (_SeqLen - 1) * Water;
            if (_chain_type == ATOMN_TOKEN) _weight -= PO2;
       } else {
            if (!strcmp(_SeqRes[0]->Field[0], "HOH") ||
                !strcmp(_SeqRes[0]->Field[0], "DOD"))
                 _weight = Water;
            else _weight = get_molecule_weight(_SeqRes[0]->Field[0]);
       }
}

int _CHAIN::TotalResidueNumbers()
{
       int k, count = 0;

       for (k = 0; k < _SeqLen; k++) {
            if (_SeqRes[k]->Deleted) continue;
            count++;
       }
       return count; 
}

void _CHAIN::get_seqres(const int format, const int token_id, Cards_Info_Format *CardInfo)
{
       int k, seqField = 4, count = 0;
       One_Card_Info_Format *pCard = NULL;

       if (!_has_sequence && _chain_type != ATOMN_TOKEN &&
            _chain_type != ATOMP_TOKEN) return;
       if (_chain_type == ATOMP_TOKEN && _SeqLen < 3 ||
           _chain_type == ATOMN_TOKEN && _SeqLen < 2) return;

       for (k = 0; k < _SeqLen; k++) {
            if (_SeqRes[k]->Deleted) continue;
            count++;
       }

       pCard = CardInfo->add_a_new_card(NdbToken[token_id], 0);
       if (format == NDB_FILE_FORMAT_PDB)
            pCard->AddElementToField(_PDB_ChainID, 2);
       else pCard->AddElementToField(_ChainID.Text(), 2);
       pCard->AddElementToField(count, 3);
       for (k = 0; k < _SeqLen; k++) {
            if (_SeqRes[k]->Deleted) continue;
            if (seqField == NdbToken[token_id].NumField) {
                 pCard = CardInfo->add_a_new_card(NdbToken[token_id], 0);
                 if (format == NDB_FILE_FORMAT_PDB)
                      pCard->AddElementToField(_PDB_ChainID, 2);
                 else pCard->AddElementToField(_ChainID.Text(), 2);
                 pCard->AddElementToField(count, 3);
                 seqField = 4;
            }
            pCard->AddElementToField(_SeqRes[k]->Field[0], seqField);
            seqField++;
       }
}

void _CHAIN::create_chain_type(const int Num_Seqres, SEQ **SeqRes)
{
       int exist_unknown = 0, type, found, i, k, l, types[7], max_num = -1;
       int index = -1, exist_DNA = 0, exist_RNA = 0, exist_TRNA = 0, num;
       _ATOM **atom = NULL;
       _RESIDUE *residue = NULL;

       if (_SeqLen == 0) return;

       if (_SeqLen == 1) {
            type = find_residue_type(_SeqRes[0]->Field[0], exist_unknown);
            if (type == ATOMP_TOKEN || type == ATOMN_TOKEN) {
                 found = 0;
                 for (k = 0; k < Num_Seqres; k++) {
                      if (_ChainID != SeqRes[k]->ChainId) continue;
                      // if (SeqRes[k]->SeqLen < 10) { /* FOR SINGLE RESIDUE */
                           for (l = 0; l < SeqRes[k]->SeqLen; l++) {
                                if (!strcmp(SeqRes[k]->res[l]->name,
                                     _SeqRes[0]->Field[0])) {
                                     found = 1;
                                     break;
                                }
                           }
                      // }
                 }
                 if (found) _chain_type = type;
                 else _chain_type = HETAIN_TOKEN;
            } else _chain_type = type;
            return;
       }

       for (i = 0; i < 7; i++) types[i] = 0;
       for (i = 0; i < _SeqLen; i++) {
            type = find_residue_type(_SeqRes[i]->Field[0], exist_unknown);
            if (type == ATOMN_TOKEN)      types[0]++;
            else if (type == ATOMP_TOKEN) types[1]++;
            else if (type == ATOMS_TOKEN) types[2]++;
            else if (type ==HETAIN_TOKEN) types[3]++;
            else if (type == HETAS_TOKEN) types[4]++;
            else if (type == HETAI_TOKEN) types[5]++;
            else if (type == HETAD_TOKEN) types[6]++;
       }

       for (i = 0; i < 7; i++) {
            if (types[i] > max_num) {
                 max_num = types[i];
                 index = i;
            }
       }
       _chain_type = HETAIN_TOKEN;
       if (index >= 0) {
            if (index == 0)      _chain_type = ATOMN_TOKEN;
            else if (index == 1) _chain_type = ATOMP_TOKEN;
            else if (index == 2) _chain_type = ATOMS_TOKEN;
            else if (index == 3) _chain_type = HETAIN_TOKEN;
            else if (index == 4) _chain_type = HETAS_TOKEN;
            else if (index == 5) _chain_type = HETAI_TOKEN;
            else if (index == 6) _chain_type = HETAD_TOKEN;
       }

       if ((_chain_type == ATOMN_TOKEN || _chain_type == ATOMP_TOKEN) &&
           Num_Seqres) {
            for (k = 0; k < Num_Seqres; k++) {
                 if (_ChainID == SeqRes[k]->ChainId) break;
            }
            if (k == Num_Seqres) _chain_type = HETAIN_TOKEN;
       }

       if (_chain_type == ATOMP_TOKEN) {
            for (i = 0; i < _SeqLen; i++) {
                 residue = _SeqRes[i]->Residue;
                 if (!residue) continue;
                 residue->set_type(MONOMER_TYPE_A_A);
            }
       } else if (_chain_type == ATOMN_TOKEN) {
            for (i = 0; i < _SeqLen; i++) {
                 residue = _SeqRes[i]->Residue;
                 if (!residue) continue;
                 atom = residue->find_atom("O2*", num);
                 if (atom) {
                      exist_RNA = 1;
                      residue->set_type(MONOMER_TYPE_RNA);
                      if (is_a_nonstandard_trna(residue->ResName()))
                           exist_TRNA = 1;
                      delete [] atom;
                 } else {
                      if (residue->AtomNumbers() == 1 &&
                          !strcmp(residue->GetFirstAtom()->atmtype(), "P")) {}
                      else {
                           residue->set_type(MONOMER_TYPE_DNA);
                           exist_DNA = 1;
                      }
                 }
            }
            if (exist_TRNA)                  _na_type = ATOMN_TYPE_TRNA;
            else if (exist_DNA && exist_RNA) _na_type = ATOMN_TYPE_DNA_RNA;
            else if (exist_RNA)              _na_type = ATOMN_TYPE_RNA_ONLY;
            else if (exist_DNA)              _na_type = ATOMN_TYPE_DNA_ONLY;
       }
}

void _CHAIN::get_descriptor(char *descriptor)
{
       int j, k, len, num;
       char sugar[20], prev_sugar[20];
       _ATOM **atom = NULL;
       _RESIDUE *residue = NULL;

       if (_chain_type == ATOMS_TOKEN && _Num_Residue > 1) {
            sprintf(descriptor, "%d-MER", _Num_Residue);
       } else if (_chain_type  == ATOMN_TOKEN) {
            if (_Num_Residue > 24) {
                 sprintf(descriptor, "%d-MER", _Num_Residue);
                 return;
            }
            strcpy(descriptor, "5\'-");
            len = strlen(descriptor);

            strcpy(prev_sugar, "XXX");
            for (j = 0; j < _SeqLen; j++) {
                 if (_SeqRes[j]->Deleted) continue;
                 if (!_SeqRes[j]->Residue) {
                      if (j == 0) {
                           for (k = 1; k < _SeqLen; k++) {
                                if (!_SeqRes[k]->Deleted &&
                                     _SeqRes[k]->Residue) break;
                           }
                           residue = _SeqRes[k]->Residue;
                           if (residue->type() == MONOMER_TYPE_RNA)
                                strcpy(sugar, "R");
                           else strcpy(sugar, "D");
                           if (strcmp(sugar, prev_sugar)) {
                                len = strlen(descriptor);
                                sprintf(&descriptor[len], "%s(", sugar);
                                strcpy(prev_sugar, sugar);
                           }
                           strcat(descriptor, "*");
                      } else strcat(descriptor, "P*");
                      len = strlen(descriptor);
                      if (is_na_residue(_SeqRes[j]->Field[0]))
                           sprintf(&descriptor[len], "%s", _SeqRes[j]->Field[0]);
                      else sprintf(&descriptor[len], "(%s)", _SeqRes[j]->Field[0]);
                      continue;
                 }

                 residue = _SeqRes[j]->Residue;
                 if (residue->type() == MONOMER_TYPE_RNA)
                      strcpy(sugar, "R");
                 else strcpy(sugar, "D");
                 if (strcmp(sugar, prev_sugar)) {
                      if (j) strcat(descriptor, ")-");
                      len = strlen(descriptor);
                      sprintf(&descriptor[len], "%s(", sugar);
                      strcpy(prev_sugar, sugar);
                 }
                 atom = residue->find_atom("P", num);
                 if (atom) {
                      strcat(descriptor, "P");
                      delete [] atom;
                 }
                 strcat(descriptor, "*");
                 len = strlen(descriptor);
                 if (is_na_residue(residue->ResName()))
                      sprintf(&descriptor[len], "%s", residue->ResName());
                 else sprintf(&descriptor[len], "(%s)", residue->ResName());
            }
            strcat(descriptor, ")-3\'");
            len = strlen(descriptor);
            descriptor[len] = '\0';
       } else {
            ConnectFormat *drug = find_drug(_SeqRes[0]->Field[0]);
            if (drug && drug->chemical_name() && strcmp(drug->chemical_name(), "")) {
                 strcpy(descriptor, drug->chemical_name());
            } else if (!strcmp(_SeqRes[0]->Field[0], "UNL")) {
                 strcpy(descriptor, "UNKNOWN LIGAND");
            }
       }
}

void _CHAIN::convert_between_orthogonal_and_fractional(const NDBCELLCONST *cell, const int iflag)
{
       for (int k = 0; k < _SeqLen; k++) {
            _RESIDUE *residue = _SeqRes[k]->Residue;
            if (residue) residue->convert_between_orthogonal_and_fractional(cell, iflag);
       }
}

void _CHAIN::symmetry_related_chain(const _CHAIN& chain, const int lx, const int ly,
             const int lz, NDBSYMMETRY* op, const int align)
{
       int i = 0;
       _RESIDUE *residue = NULL;
       _lx = lx; _ly = ly; _lz = lz;
       _chain_type = chain._chain_type;
       _entity_id = chain._entity_id;
       for (i = 0; i < chain._SeqLen; i++) {
            if (!chain._SeqRes[i]->Residue) continue;
            residue = new _RESIDUE;
            *residue = *(chain._SeqRes[i]->Residue);
            residue->symmetry_operations(lx, ly, lz, op, 0);
            insert_a_residue(residue);
       }

       if (align) {
            SEQ *seq = new SEQ;
            seq->SeqLen = chain._SeqLen;
            seq->res = new RES*[seq->SeqLen];
            for (i = 0; i < chain._SeqLen; i++) {
                 seq->res[i] = new RES;
                 strcpy(seq->res[i]->name, chain._SeqRes[i]->Field[0]);
            }
            seq_alignment(seq, NULL);
            for (i = 0; i < chain._SeqLen; i++)
                 if (seq->res[i]) delete seq->res[i];
            delete [] seq->res;
            delete seq;
       }
}

void _CHAIN::do_symmetry_translation(const int lx, const int ly, const int lz)
{
       int k = 0;
       _ATOM *atom = NULL;
       _RESIDUE *residue = NULL;
       _lx += lx; _ly += ly; _lz += lz;
       for (k = 0; k < _SeqLen; k++) {
            residue = _SeqRes[k]->Residue;
            if (residue) {
                 atom = residue->GetFirstAtom();
                 while (atom) {
                      COORD coord;
                      coord.x = atom->orig().x + (float) lx;
                      coord.y = atom->orig().y + (float) ly;
                      coord.z = atom->orig().z + (float) lz;
                      atom->set_orig(coord);
                      atom = residue->GetNextAtom();
                 }
            }
       }
}

void _CHAIN::symmetry_related_chain(const _CHAIN& chain, double m[3][4])
{
       int i = 0;
       _RESIDUE *residue = NULL;
       _chain_type = chain._chain_type;
       _entity_id = chain._entity_id;
       for (i = 0; i < chain._SeqLen; i++) {
            if (!chain._SeqRes[i]->Residue) continue;
            residue = new _RESIDUE;
            *residue = *(chain._SeqRes[i]->Residue);
            residue->transformation(m);
            insert_a_residue(residue);
       }
       if (!chain._has_sequence) return;

       SEQ *seq = new SEQ;
       seq->SeqLen = chain._SeqLen;
       seq->res = new RES*[seq->SeqLen];
       for (i = 0; i < chain._SeqLen; i++) {
            seq->res[i] = new RES;
            strcpy(seq->res[i]->name, chain._SeqRes[i]->Field[0]);
       }
       seq_alignment(seq, NULL);
       for (i = 0; i < chain._SeqLen; i++)
            if (seq->res[i]) delete seq->res[i];
       delete [] seq->res;
       delete seq;
}

void _CHAIN::change_coordinates(const NDBCELLCONST *cell, const int op,
                     const int lx, const int ly, const int lz)
{
       int i = 0;
       _RESIDUE *residue = NULL;
       for (i = 0; i < _SeqLen; i++) {
            residue = _SeqRes[i]->Residue;
            if (residue) {
                 residue->convert_between_orthogonal_and_fractional(cell, NDB_TRANS_ORTHO_TO_FRAC);
                 residue->symmetry_operations(lx, ly, lz, &(cell->symops[op]), 0);
                 residue->convert_between_orthogonal_and_fractional(cell, NDB_TRANS_FRAC_TO_ORTHO);
            }
       }
}

void _CHAIN::change_water_coordinates(const _CHAIN& chain, const NDBCELLCONST *cell,
                 const int op, const int lx, const int ly, const int lz)
{
       int i = 0, j = 0;
       _RESIDUE *residue = NULL, *res = NULL;
       for (i = 0; i < _SeqLen; i++) {
            residue = _SeqRes[i]->Residue;
            if (residue) {
                 for (j = 0; j < chain._SeqLen; j++) {
                      res = chain._SeqRes[j]->Residue;
                      if (res && is_contact(residue, res)) {
                           residue->convert_between_orthogonal_and_fractional(cell,
                                       NDB_TRANS_ORTHO_TO_FRAC);
                           residue->symmetry_operations(lx, ly, lz, &(cell->symops[op]), 0);
                           residue->convert_between_orthogonal_and_fractional(cell,
                                       NDB_TRANS_FRAC_TO_ORTHO);
                      }
                 }
            }
       }
}

void _CHAIN::change_coordinates(double m[3][4])
{
       for (int i = 0; i < _SeqLen; i++) {
            if (_SeqRes[i]->Residue)
                 _SeqRes[i]->Residue->transformation(m);
       }
}

void _CHAIN::labeling_iupac_hydrogen()
{
       if (!_SeqLen || !_SeqRes) return;
       if (_SeqRes[0]->Residue)
            _SeqRes[0]->Residue->labeling_iupac_hydrogen_first();
}

static int score(const int k, const int l)
{
       if (res_name_cmp(_seqa->res[k]->name, _seqb->res[l]->name))
            return 0;
       else {
            if (_pdba && _pdbb && !strcmp(_pdba->res[k]->name, _pdbb->res[l]->name))
                 return 7;
	    else return 5;
       }
}

static int is_contact(_RESIDUE *res_a, _RESIDUE *res_b)
{
       _ATOM *atom_i = res_a->GetFirstAtom();
       while (atom_i) {
            _ATOM *atom_j = res_b->GetFirstAtom();
            while (atom_j) {
                 if (cal_distance(atom_i, atom_j) <= 7.0) return 1;
                 atom_j = res_b->GetNextAtom();
            }
            atom_i = res_a->GetNextAtom();
       }
       return 0;
}

