/*
FILE:     _fileobj_ioutil.C
*/
/*
VERSION:  1.700
*/
/*
DATE:     7/13/2005
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2005 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "_fileobj_ioutil.h"
#include "erf_iolib.h"
#include "ndb_misclib.h"
#include "util.h"

#define NUM_SPACE_GROUP  95

typedef struct {
        char fake_name[15];
        char real_name[15];
} SPACE_GROUP_NAME;

static SPACE_GROUP_NAME space_group_name[NUM_SPACE_GROUP] = {
       { "A1",         "A 1" },
       { "A121",       "A 1 2 1" },
       { "A2",         "A 2" },
       { "B112",       "B 1 1 2" },
       { "B2",         "B 2" },
       { "B2212",      "B 2 21 2" },
       { "C121",       "C 1 2 1" },
       { "C1211",      "C 1 21 1" },
       { "C2",         "C 2" },
       { "C2(A112)",   "C 2(A 112)" },
       { "C21",        "C 21" },
       { "C222",       "C 2 2 2" },
       { "C2221",      "C 2 2 21" },
       { "C4212",      "C 4 21 2" },
       { "F222",       "F 2 2 2" },
       { "F23",        "F 2 3" },
       { "F4132",      "F 41 3 2" },
       { "F422",       "F 4 2 2" },
       { "F432",       "F 4 3 2" },
       { "H3",         "H 3" },
       { "H32",        "H 3 2" },
       { "I121",       "I 1 2 1" },
       { "I1211",      "I 1 21 1" },
       { "I2",         "I 2" },
       { "I21",        "I 21" },
       { "I212121",    "I 21 21 21" },
       { "I213",       "I 21 3" },
       { "I222",       "I 2 2 2" },
       { "I23",        "I 2 3" },
       { "I4",         "I 4" },
       { "I41",        "I 41" },
       { "I4122",      "I 41 2 2" },
       { "I4132",      "I 41 3 2" },
       { "I422",       "I 4 2 2" },
       { "I432",       "I 4 3 2" },
       { "P-1",        "P 1-" },
       { "P1",         "P 1" },
       { "P1-",        "P 1-" },
       { "P112",       "P 1 1 2" },
       { "P1121",      "P 1 1 21" },
       { "P1211",      "P 1 21 1" },
       { "P121",       "P 1 2 1" },
       { "P2",         "P 2" },
       { "P21",        "P 21" },
       { "P21(C)",     "P 21(C)" },
       { "P21212",     "P 21 21 2" },
       { "P212121",    "P 21 21 21" },
       { "P21212A",    "P 21 21 2 A" },
       { "P21221",     "P 21 2 21" },
       { "P213",       "P 21 3" },
       { "P22121",     "P 2 21 21" },
       { "P222",       "P 2 2 2" },
       { "P2221",      "P 2 2 21" },
       { "P23",        "P 2 3" },
       { "P3",         "P 3" },
       { "P31",        "P 31" },
       { "P3112",      "P 31 1 2" },
       { "P312",       "P 3 1 2" },
       { "P3121",      "P 31 2 1" },
       { "P32",        "P 32" },
       { "P321",       "P 3 2 1" },
       { "P3212",      "P 32 1 2" },
       { "P3221",      "P 32 2 1" },
       { "P4",         "P 4" },
       { "P41",        "P 41" },
       { "P41212",     "P 41 21 2" },
       { "P4122",      "P 41 2 2" },
       { "P4132",      "P 41 3 2" },
       { "P42",        "P 42" },
       { "P4212",      "P 4 21 2" },
       { "P422",       "P 4 2 2" },
       { "P42212",     "P 42 21 2" },
       { "P4222",      "P 42 2 2" },
       { "P4232",      "P 42 3 2" },
       { "P43",        "P 43" },
       { "P432",       "P 4 3 2" },
       { "P43212",     "P 43 21 2" },
       { "P4322",      "P 43 2 2" },
       { "P4332",      "P 43 3 2" },
       { "P6",         "P 6" },
       { "P61",        "P 61" },
       { "P6122",      "P 61 2 2" },
       { "P62",        "P 62" },
       { "P622",       "P 6 2 2" },
       { "P6222",      "P 62 2 2" },
       { "P63",        "P 63" },
       { "P6322",      "P 63 2 2" },
       { "P64",        "P 64" },
       { "P6422",      "P 64 2 2" },
       { "P65",        "P 65" },
       { "P6522",      "P 65 2 2" },
       { "R3",         "R 3" },
       { "R3(RHOM)",   "R 3(RHOM)" },
       { "R32",        "R 3 2" },
       { "R32(RHOM)",  "R 3 2(RHOM)" }
};

static void ndb_get_correct_labels(char Field[MxField][MxLen], const int format,
                 const int is_iupac);
#ifdef SUN_OS
#ifdef __cplusplus
extern "C" {
#endif
#endif
static int real_name_compare(const void *group1, const void *group2);
static int fake_name_compare(const void *group1, const void *group2);
#ifdef SUN_OS
#ifdef __cplusplus
}
#endif
#endif

int get_card(FILE *infile, char *CardString, int &lineno)
{
       int i, temp_c, card_id, str_len;
       lineno++;

       temp_c = fscanf(infile, "%s", CardString);
       if (temp_c != EOF) {
            str_len = strlen(CardString);
            if (str_len < 6 ) {
                for (i = str_len; i < 6 && temp_c != '\n'; i++)
                     temp_c = fgetc(infile); 
                if (temp_c == '\n') fseek(infile, (long) -1, 1);
            } else if (str_len > 6) {
                for (i = str_len; i> 6 ; i--)
                     fseek(infile, (long) -1, 1);
                CardString[i] = 0;
            }
       } else return -2;

       card_id = find_card_id(CardString);
       if (card_id == -1) {
            if (!strcmp(CardString, "ATOM")) 
                 card_id = ATOMN_TOKEN -1;
            else if (!strcmp(CardString, "ATOMM")) 
                 card_id = ATOMMN_TOKEN -1;
            else if (strcmp(CardString, "HETATM") &&
                     strcmp(CardString, "BEFORE") &&
                     strcmp(CardString, "AFTER")) {
                 ndb_log_message(NDB_MSG_ERR, 
                    "The record %s at line %d is unrecognized", CardString, lineno);
            }
       }
       return(card_id);
}

int get_ndbfile_field(FILE *infile, const int card_id, const int field_no, 
		      char *FieldAnswer, int &length)
{
       char field_value[MxLen];
       int temp_c = 1, i, j, str_len, gap;

       length = 0;
       for (i = 0; i <( NdbToken[card_id].FieldList[field_no].FieldStcol -
	       (NdbToken[card_id].FieldList[field_no -1].FieldWidth +
		NdbToken[card_id].FieldList[field_no-1].FieldStcol)); i++) {
            temp_c = fgetc(infile);
            length++;
            if (temp_c == '\n') break;
       }

       if (temp_c != '\n' && temp_c != '\0') {
            for (str_len = 0; str_len < NdbToken[card_id].FieldList[field_no].FieldWidth;
                             str_len++) {
                 field_value[str_len] = fgetc(infile);
                 length++;
                 if (field_value[str_len] == '\n' || field_value[str_len] == '\0') { 
                      str_len++; break;
                 }
            }
            if (field_value[str_len-1] != '\n' && field_value[str_len-1] != '\0' &&
                field_no != NdbToken[card_id].NumField-1) {
                 gap = NdbToken[card_id].FieldList[field_no+1].FieldStcol -
                       NdbToken[card_id].FieldList[field_no].FieldWidth -
                       NdbToken[card_id].FieldList[field_no].FieldStcol;
                 for (i = 0; i < gap; i++) {
                     field_value[str_len] = fgetc(infile);
                     if (field_value[str_len] == ' ' || field_value[str_len] == '\n' || 
                         field_value[str_len] == '\t') {
                          i++; break;
                     }
                     length++;
                     str_len++;
                 }
                 if (i != 0) fseek(infile, (long) (-1*i), 1);
            }
            field_value[str_len] = '\0';
            if (field_value[str_len] != '\0') field_value[str_len] = '\0';
            /* Deleting spaces before field value */
            if (card_id != REMARK_TOKEN-1 && card_id != PREMRK_TOKEN-1 &&
                card_id != PCOMPN_TOKEN-1 && card_id != PSOURC_TOKEN-1 ||
                field_no != 2) {
                 j = 0;
                 for (i = 0; i < str_len; i++) {
                      if (field_value[i] == ' ' || field_value[i] == '\t') j++;
                      else break;
                 }
                 if (i == str_len) {
                     str_len =0;
                     strcpy(field_value, "");
                 } else for (i = j; i <= str_len; i++) field_value[i-j] = field_value[i];
            }
            /* Deleting spaces after field value */
            str_len = strlen(field_value);
            for (i = str_len-1; i >= 0; i--) {
                 if (field_value[i] == ' ' || field_value[i] == '\t')
                     field_value[i] = '\0';
                 else break;
            }
            // ndb_strip_trailing_blanks(field_value);
       } else {
            sprintf(field_value, "%c", temp_c);
            field_value[1] = '\0';
       }
       str_len = strlen(field_value);
  
       /* found a \n means it is the end of line */
       if (field_no == NdbToken[card_id].NumField -1 &&  
          (str_len == 0 || field_value[str_len-1] != '\n')) {
            field_value[str_len] = '\n';
            field_value[str_len+1] = '\0';
            str_len++;
            length += end_line(infile);
       }
       if (str_len != 0 && field_value[str_len-1] == '\n') {
            field_value[str_len-1] = '\0';
            str_len = strlen(field_value);
            for (i = str_len-1; i >=0; i--) {
                 if (field_value[i] == ' ' || field_value[i] == '\t')
        	      field_value[i] = '\0';
                 else break;
            }
            strcpy(FieldAnswer, field_value);
            return(0);
       } else {
            strcpy(FieldAnswer, field_value);
            return(1);
       }
}

int get_jrnl_token(FILE *infile, char *JrnlString)
{
       int i, j, temp_c, str_len;

       for (i = 0; i < (JrnlToken[0].FieldList[2].FieldStcol -
                   (JrnlToken[0].FieldList[1].FieldWidth +
                   JrnlToken[0].FieldList[1].FieldStcol)); i++){
            temp_c = fgetc(infile);
            if (temp_c == '\n'){
                 fseek(infile, -1L, 1);
                 return (EOF);
            } else if (temp_c != ' ') {
                 fseek(infile,-1L, 1);
                 break;
            } 
       }
       temp_c = fscanf(infile, "%4s", JrnlString);
       str_len = strlen(JrnlString);
       if (temp_c != EOF) {
            if (str_len != 4) {
              for (i = str_len; i <4  && temp_c != EOF && temp_c != '\n'; i++)
                   temp_c = fgetc(infile);
            }
            if (temp_c == '\n') fseek(infile, -1L, 1);
            for (i = str_len-1; i >= 0 && JrnlString[i] == ' '; i--)
                 if (JrnlString[i] == ' ') JrnlString[i] = '\0';
            str_len = strlen(JrnlString);
            for (i = 0; i < str_len && JrnlString[i] == ' '; i++);
            if (i != 0) {
                 for (j = i; j < str_len; j++)
                      JrnlString[j-i] = JrnlString[j];
                 JrnlString[j-i] = '\0';
            }
       }
       for (i = 0; i < MxJrnlTokens && 
                   strcmp(JrnlToken[i].TokenName, JrnlString) != 0; i++);
       if (i != MxJrnlTokens) return (i);
       else return (-1);
}

int get_jrnl_field(FILE *infile, const int jrnl_id, const int field_no,
            char *FieldAnswer)
{
       int i, j, temp_c = 1, str_len;
       char field_value[MxStrLen];

       memset(field_value, 0, MxStrLen);
       for (i = 0; i < (JrnlToken[jrnl_id].FieldList[field_no].FieldStcol -
                   (JrnlToken[jrnl_id].FieldList[field_no -1].FieldWidth +
                   JrnlToken[jrnl_id].FieldList[field_no-1].FieldStcol)); i++) {
            temp_c = fgetc(infile);
            if (temp_c != ' ') {
                 if (temp_c != '\n') fseek(infile,-1L, 1);
                 break;
            }
       }
       if (temp_c != '\n' && temp_c != '\0') {
            str_len =0;
            for (i = 0; i < JrnlToken[jrnl_id].FieldList[field_no].FieldWidth; i++){
                 field_value[i] = fgetc(infile);
                 str_len ++;
                 if (field_value[i] == '\n' || field_value[i] == '\0') break;
            }
            if (field_value[i] != '\0') field_value[str_len] = '\0';
            
            j = 0;
            for (i = 0; i < str_len; i++) {
                 if (field_value[i] == ' ' || field_value[i] == '\t') j++;
                 else break;
            }
            if (i == str_len) strcpy(field_value, "");
            else for (i = j; i <= str_len; i++) field_value[i-j] = field_value[i];
            str_len = strlen(field_value);
            for (i = str_len-1; i >= 0; i--) {
                 if (field_value[i]==' ' || field_value[i] == '\t') 
                      field_value[i] = '\0';
                 else break;
            }
       } else {
            sprintf(field_value, "%c", temp_c);
            field_value[1] = '\0';
       }
       str_len = strlen(field_value);
       if (field_no == JrnlToken[jrnl_id].NumField - 1 &&
          (str_len == 0 || field_value[str_len-1] != '\n')) {
            field_value[str_len] = '\n';
            str_len++;
            end_line(infile); 
       }
       if (str_len != 0 && field_value[str_len-1] == '\n') {
            field_value[str_len-1] = '\0';
            str_len = strlen(field_value);
            for (i = str_len-1; i >= 0; i--) {
                 if (field_value[i] == ' ' || field_value[i] == '\t') 
                      field_value[i] = '\0';
                 else break;
            }
            strcpy(FieldAnswer, field_value);
            return(0);
       } else {
            strcpy(FieldAnswer, field_value);
            return(1);
       }
}

int get_one_line(char *line, FILE *infile)
{
       int temp_c, length;

       *line = '\0';
       length = 0;
       do {
            temp_c = fgetc(infile);
            *line = temp_c;
            line++;
            length++;
       } while (temp_c != '\n' && temp_c != EOF);
       *line = '\0';
       return length;
}

int end_line(FILE *infile)
{
       int temp_c;
       int i = 0;
       do {
            temp_c = fgetc(infile);
            i++;
       } while (temp_c != '\n' && temp_c != EOF );
       return i;
}

void print_atom_card(char *temp_line, const int card_id, const int temp_id,
            int &atomSerialNo, int &WaterSerialNo, char Field[MxField][MxLen],
            const int format1, int align, const int is_iupac)
{
       char FieldFormat[MxStrLen], *p = NULL;
       int atomNameField = 2, residueNameField = 4, pos = 6, atom_id, format;
       int tmp_field_no, field_no, align_pos, n;

       format = format1;
       if (!(is_atom_card(temp_id) & AUX_ATOM) && format == NDB_FILE_FORMAT_PDB) {
            atom_id = get_atom_card_id(temp_id);
            if (!strcmp(AtomInfo[atom_id].PdbTitle, "ATOM")) {
                 if (is_a_standard_residue(Field[16]))
                      sprintf(temp_line, "%-6s", AtomInfo[atom_id].PdbTitle);
                 else sprintf(temp_line, "%-6s", "HETATM");
            } else sprintf(temp_line, "%-6s", AtomInfo[atom_id].PdbTitle);
       } else sprintf(temp_line, "%-6s", NdbToken[temp_id].TokenName);
       if (format == NDB_FILE_FORMAT_MIX) {
            sprintf(temp_line, "%-6s", Field[0]);
            format = NDB_FILE_FORMAT_NDB;
       }

       if (!(is_atom_card(temp_id) & AUX_ATOM)) atomSerialNo++;

       ndb_get_correct_labels(Field, format, is_iupac);
       for (field_no = 1; field_no < NdbToken[temp_id].NumField; field_no++) {
            if ((format == NDB_FILE_FORMAT_PDB || format == NDB_FILE_FORMAT_SHORT)
                 && NdbToken[temp_id].FieldList[field_no].NdbOnly) break;
            tmp_field_no = get_field_no(temp_id, field_no, format);
            if ((field_no == 11 || field_no == 12) &&
                !strcmp("", Field[field_no])) {
                  if (field_no == 11) strcpy(Field[field_no], "1.0");
                  else strcpy(Field[field_no], "10.0");
            }
            if (!NdbToken[temp_id].FieldList[field_no].MixedCase)
                 ndb_set_string_to_upper(Field[field_no]);
            print_space_between_field(temp_id, field_no, temp_line, pos, 0);
            get_field_format(temp_id, field_no, FieldFormat, 0, format);
            if (field_no == 1) {
                 sprintf(Field[1], "%d", atomSerialNo);
                 sprintf(&temp_line[pos], FieldFormat, (double) atomSerialNo);
            } else if (strcmp("", Field[field_no])) {
                 if (field_no == atomNameField) {
                      if (is_iupac && !strcmp(Field[23], "H")) {
                           p = strstr(Field[field_no], Field[23]);
                           if (p) {
                                n = p - Field[field_no];
                                if (n) align = 0;
                                else   align = 1;
                           }
                      }
                      print_atom_atomName_field(Field[23], pos, temp_line,
                          Field[field_no], Field[residueNameField], align);
                 } else if (NdbToken[temp_id].FieldList[tmp_field_no].FieldType <= 2) {
                      double temp_double = atof(Field[field_no]);
                      sprintf(&temp_line[pos], FieldFormat, temp_double);
                 } else sprintf(&temp_line[pos], FieldFormat, Field[field_no]);
            } else {
                 sprintf(FieldFormat, "%%%-ds",
                          NdbToken[temp_id].FieldList[field_no].FieldWidth);
                 sprintf(&temp_line[pos], FieldFormat, "");
            }
            pos += NdbToken[temp_id].FieldList[field_no].FieldWidth;
       }

       if (format == NDB_FILE_FORMAT_PDB) {
            sprintf(FieldFormat, "%%%ds", 
                 78 - NdbToken[temp_id].FieldList[field_no].FieldStcol + 1);
            sprintf(&temp_line[pos], FieldFormat, "");
            pos = strlen(temp_line) - 1;
            align_pos = 1;
            if (strlen(Field[23]) == 2) align_pos = 0;
            print_atom_atomName_field(Field[23], pos, temp_line, Field[23],
                    Field[residueNameField], align_pos);
       }
}

void print_atom_card(FILE *fp, const int card_id, const int temp_id,
            int &atomSerialNo, int &WaterSerialNo, char Field[MxField][MxLen],
            const int format1, int align, const int is_iupac)
{
       char temp_line[MxLen];
       print_atom_card(temp_line, card_id, temp_id, atomSerialNo, WaterSerialNo, Field,
                 format1, align, is_iupac);
       fprintf(fp, "%s\n", temp_line);
}

void print_space_between_field(const int card_id, const int field_no,
        char *temp_line, int &LinePos, const int IsJRNL)
{
       char temp_string[MxStrLen];
       if (IsJRNL == 0) {
            if (NdbToken[card_id].FieldList[field_no].FieldStcol != LinePos+1) {
                 sprintf(temp_string, "%%-%ds",
                         NdbToken[card_id].FieldList[field_no].FieldStcol -
                         NdbToken[card_id].FieldList[field_no-1].FieldStcol -
                         NdbToken[card_id].FieldList[field_no-1].FieldWidth);
                 sprintf(&temp_line[LinePos], temp_string, "");
                 LinePos = NdbToken[card_id].FieldList[field_no].FieldStcol -1;
            }
       } else {
            if (JrnlToken[card_id].FieldList[field_no].FieldStcol != LinePos+1){
                 sprintf(temp_string, "%%-%ds",
                       JrnlToken[card_id].FieldList[field_no].FieldStcol -
                       JrnlToken[card_id].FieldList[field_no-1].FieldStcol -
                       JrnlToken[card_id].FieldList[field_no-1].FieldWidth);
                 sprintf(&temp_line[LinePos], temp_string, "");
                 LinePos = JrnlToken[card_id].FieldList[field_no].FieldStcol-1;
            }
       }
}

void print_ter_card(FILE *fp, int &atomSerialNo, const int format, char Field[MxField][MxLen])
{
       char FieldFormat[MxStrLen], temp_line[MxLen];
       int  residueNoField = 4, pos = 6, field_no, temp_id;

       temp_id = find_card_id("TER");
       sprintf(temp_line, "%-6s","TER");
       for (field_no = 1; field_no < NdbToken[temp_id].NumField; field_no++) {
            if (!NdbToken[temp_id].FieldList[field_no].MixedCase)
                 ndb_set_string_to_upper(Field[field_no]);
            print_space_between_field(temp_id, field_no, temp_line, pos, 0);
            get_field_format(temp_id, field_no, FieldFormat, 0, format);
            if (format != NDB_FILE_FORMAT_NDB && field_no == residueNoField)
                 sprintf(FieldFormat, "%%%ds",
                      NdbToken[temp_id].FieldList[field_no].FieldWidth);
        
            if (field_no == 1) {
                 atomSerialNo++;
                 sprintf(&temp_line[pos], FieldFormat, (double) atomSerialNo);
            } else if (strcmp(Field[field_no+2], "") == 0){
                 sprintf(FieldFormat, "%%%ds",
                      NdbToken[temp_id].FieldList[field_no].FieldWidth);
                 sprintf(&temp_line[pos], FieldFormat, "");
            } else if (NdbToken[temp_id].FieldList[field_no].FieldType > 2 || 
               field_no == residueNoField && format != NDB_FILE_FORMAT_NDB)
                 sprintf(&temp_line[pos], FieldFormat, Field[field_no+2]);
            else {
                 double temp_double = atof(Field[field_no+2]);
                 sprintf(&temp_line[pos], FieldFormat, temp_double);
            }
            pos += NdbToken[temp_id].FieldList[field_no].FieldWidth;
       }
       if (format == NDB_FILE_FORMAT_PDB)
            fprintf(fp,"%s                                                     \n", temp_line);
       else fprintf(fp,"%s\n",temp_line);
}

void print_ndb_continuedf_field(const int card_id, One_Card_Info_Format *pCard,
         const int pos, int &start, char *temp_line, const char *FieldFormat,
         const int ContNo, const int PdbContNo)
{
       int i, str_len, max_len;
       char temp[200] /* , buffer[4000] */ ;
       
       if (PdbContNo == 1) 
            max_len = NdbToken[card_id].FieldList[NdbToken[card_id].ContinuedF-1].FieldWidth;
       else max_len = NdbToken[card_id].FieldList[NdbToken[card_id].ContinuedF-1].FieldWidth-1;
     
       if (card_id == PCOMPN_TOKEN - 1 || card_id == PSOURC_TOKEN - 1 ||
           card_id == find_card_id("KEYWDS"))
            pCard->CleanStringField(NdbToken[card_id].ContinuedF-1);
       str_len = strlen(pCard->Field(NdbToken[card_id].ContinuedF-1)) - start;
     
       if (str_len > max_len) {
            for (i = max_len-1; i>=0; i--) {
                 if (pCard->Field(NdbToken[card_id].ContinuedF-1)[i+start] == ' ' ||
                     pCard->Field(NdbToken[card_id].ContinuedF-1)[i+start] == ',' ||
                     pCard->Field(NdbToken[card_id].ContinuedF-1)[i+start] == '-')
                      break;
                 if (pCard->Field(NdbToken[card_id].ContinuedF-1)[i+start] == ')') {
                      if ((i < (max_len - 1)) &&
                           pCard->Field(NdbToken[card_id].ContinuedF-1)[i+start+1] != ';')
                      break;
                 }
            }
            if (i < max_len / 5) i =  max_len;
            else i = i + 1;
       } else i = str_len;
     
       if (PdbContNo == 1 && NdbToken[card_id].PdbConField != 0 ||
           card_id == PCOMPN_TOKEN -1 || card_id == PSOURC_TOKEN-1 || 
           NdbToken[card_id].PdbConField == 0 && ContNo == 1) {
            strncpy(temp, &(pCard->Field(NdbToken[card_id].ContinuedF-1)[start]), i);
            temp[i] = '\0';
       } else {
            temp[0] = ' ';
            strncpy(&temp[1], &(pCard->Field(NdbToken[card_id].ContinuedF-1)[start]), i);
            temp[i+1] = '\0';
       }
       if (pCard->Field(NdbToken[card_id].ContinuedF-1)[i+start] == ' ') i++;
       start += i;
       sprintf(&temp_line[pos], FieldFormat, temp);
       str_len = strlen(temp_line);
       for (i = 0; i < str_len; i++)
            if (temp_line[i] == '\n') temp_line[i] = ' ';
}

void print_jrnl_continuedf_field(const int jrnl_id, One_Card_Info_Format *pJrnlCard,
            const int pos, int &start, char *temp_line, const char *FieldFormat)
{
       int i, continueField, str_len;
       char temp[MxStrLen] /* , buffer[2000] */;
       
       continueField = JrnlToken[jrnl_id].ContinuedF-1;
       pJrnlCard->CleanStringField(continueField);
       str_len = strlen(pJrnlCard->Field(continueField)) - start;
       i = str_len;
       if (str_len > JrnlToken[jrnl_id].FieldList[continueField].FieldWidth) {
            if (strcmp(JrnlToken[jrnl_id].TokenName, "AUTH") !=0) {
                 for (i = JrnlToken[jrnl_id].FieldList[continueField].FieldWidth-1; i>=0; i--) {
                      if (pJrnlCard->Field(JrnlToken[jrnl_id].ContinuedF-1)[i+start] == ' ') break;
                 }
            }
            if (i < JrnlToken[jrnl_id].FieldList[continueField].FieldWidth / 5 ||
                strcmp(JrnlToken[jrnl_id].TokenName, "AUTH") == 0){
                 for (i = JrnlToken[jrnl_id].FieldList[continueField].FieldWidth-1; i>=0; i--) {
                      if (pJrnlCard->Field(continueField)[i+start] == ',' ||
                          pJrnlCard->Field(continueField)[i+start] == ')') break;
                      if (strcmp(JrnlToken[jrnl_id].TokenName,"AUTH") != 0 &&
                          pJrnlCard->Field(continueField)[i+start] == '-') break;
                      if (!strcmp(JrnlToken[jrnl_id].TokenName,"REF") &&
                          pJrnlCard->Field(continueField)[i+start] == '.') break;
                 }
            }
	    if (i < 0 && strcmp(JrnlToken[jrnl_id].TokenName, "AUTH") == 0) {
                 for (i = JrnlToken[jrnl_id].FieldList[continueField].FieldWidth-1; i>=0; i--) {
                      if (pJrnlCard->Field(JrnlToken[jrnl_id].ContinuedF-1)[i+start] == ' ') break;
                 }
            }
            if (i < JrnlToken[jrnl_id].FieldList[continueField].FieldWidth / 5) {
                 if (strcmp(JrnlToken[jrnl_id].TokenName, "AUTH") == 0) {
                      for (i = JrnlToken[jrnl_id].FieldList[continueField].FieldWidth-1; i>=0; i--) {
                           if (pJrnlCard->Field(JrnlToken[jrnl_id].ContinuedF-1)[i+start] == ' ') break;
                      }
		      i = i + 1;
                 } else i = JrnlToken[jrnl_id].FieldList[continueField].FieldWidth-1;
	    } else i = i + 1;
       } else i = str_len;
       strncpy(temp, &(pJrnlCard->Field(continueField)[start]), i);
       temp[i] = '\0';
       if (pJrnlCard->Field(continueField)[i+start] == ' ') i++;
       start += i;
/*
       strcpy(buffer, &(pJrnlCard->Field(continueField)[i]));
       str_len = str_len - i ;
       if (str_len >= 0) buffer[str_len] = '\0';
       else buffer[0] = '\0';
       pJrnlCard->AddElementToField(buffer, continueField);
*/
       sprintf(&temp_line[pos],FieldFormat, temp);
}

int check_space_group(const char *old_name, char *new_name)
{
       strcpy(new_name, old_name);
       qsort((void *) space_group_name, NUM_SPACE_GROUP,
             sizeof(SPACE_GROUP_NAME), real_name_compare);
       SPACE_GROUP_NAME tmp;
       strcpy(tmp.real_name, old_name);
       SPACE_GROUP_NAME *exist = (SPACE_GROUP_NAME *) bsearch(&tmp,
            space_group_name, NUM_SPACE_GROUP, sizeof(SPACE_GROUP_NAME),
            real_name_compare);
       if (exist) return 1;
       strcpy(tmp.fake_name, old_name);
       int stringlen = strlen(tmp.fake_name);
       for (int i = 0; i < stringlen; i++) {
            if (tmp.fake_name[i] == ' ') {
                 strcpy(&(tmp.fake_name[i]), &(tmp.fake_name[i+1]));
                 stringlen--; i--;
            }
       }
       qsort((void *) space_group_name, NUM_SPACE_GROUP,
             sizeof(SPACE_GROUP_NAME), fake_name_compare);
       exist = (SPACE_GROUP_NAME *) bsearch(&tmp, space_group_name,
           NUM_SPACE_GROUP, sizeof(SPACE_GROUP_NAME), fake_name_compare);
       if (exist) {
            strcpy(new_name, exist->real_name);
            return 1;
       }
       return 0;
}

static void ndb_get_correct_labels(char Field[MxField][MxLen], const int format,
                 const int is_iupac)
{
       int  atomNameField = 2, residueNameField = 4, chainIdField = 5;
       int  residueNoField = 6, pdbResidueNoField = 15;
       int  pdbResidueNameField = 16, pdbChainIdField = 17;
       int  pdbAtomNameField = 18, altResidueNoField = 19;
       int  altResidueNameField = 20, altChainIdField = 21;
       int  altAtomNameField = 22;

       if (format == NDB_FILE_FORMAT_PDB && strcmp(Field[pdbResidueNoField], ""))
            strcpy(Field[residueNoField], Field[pdbResidueNoField]);
       else if (format == NDB_FILE_FORMAT_ALIAS && strcmp(Field[altResidueNoField], ""))
            strcpy(Field[residueNoField], Field[altResidueNoField]);
 
       if (format == NDB_FILE_FORMAT_PDB && strcmp(Field[pdbResidueNameField], ""))
            strcpy(Field[residueNameField], Field[pdbResidueNameField]);
       else if (format == NDB_FILE_FORMAT_ALIAS &&
                strcmp(Field[altResidueNameField], ""))
            strcpy(Field[residueNameField], Field[altResidueNameField]);

       if (format == NDB_FILE_FORMAT_PDB)
            strcpy(Field[chainIdField], Field[pdbChainIdField]);
       else if (format == NDB_FILE_FORMAT_ALIAS)
            strcpy(Field[chainIdField], Field[altChainIdField]);

       if (format == NDB_FILE_FORMAT_PDB && strcmp(Field[pdbAtomNameField], "") &&
          (strcmp(Field[residueNameField], "HOH") || Field[atomNameField][0] != 'O')) {
            if (!is_iupac) strcpy(Field[atomNameField], Field[pdbAtomNameField]);
       } else if (format == NDB_FILE_FORMAT_ALIAS && strcmp(Field[altAtomNameField], ""))
            strcpy(Field[atomNameField], Field[altAtomNameField]);
}

#ifdef SUN_OS
#ifdef __cplusplus
extern "C" {
#endif
#endif

static int real_name_compare(const void *group1, const void *group2)
{
       return (strcmp(((SPACE_GROUP_NAME *) group1)->real_name,
                      ((SPACE_GROUP_NAME *) group2)->real_name));
}

static int fake_name_compare(const void *group1, const void *group2)
{
       return (strcmp(((SPACE_GROUP_NAME *) group1)->fake_name,
                      ((SPACE_GROUP_NAME *) group2)->fake_name));
}

#ifdef SUN_OS
#ifdef __cplusplus
}
#endif
#endif
