/*
FILE:     _fileobj_ndb_cif_iolib.C
*/
/*
VERSION:  1.700
*/
/*
DATE:     7/13/2005
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2005 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <math.h>
#include <string.h>

#include "ndb_misclib.h"
#include "dict_exchange.h"
#include "_fileobj.h"
#include "_fileobj_ioutil.h"
#include "align_util.h"
#include "erf_iolib.h"
#include "erf_util.h"
#include "util.h"

typedef struct {
        int atom_id;
        int con_num;
        int con_to[20];
} CONECT;

typedef struct {
        char new_chainid[10];
        char old_chainid[10];
} _CHAINID_CORR;

typedef struct {
        const char  *category;
        const char  *item1;
        const char  *item2;
        const char  *item3;
} _category_corr;

#define NUM_CATAGORY_CORR   7
#define N_NCS_KEYWORDS      6
#define N_LIM_KEYWORDS      7
#define N_LIM_MISSING      12

static _category_corr Category_CORR[NUM_CATAGORY_CORR] = {
        { "ebi_validate_angle",         "auth_asym_id_1",   "auth_asym_id_2", "auth_asym_id_3" },
        { "ebi_validate_bond",          "auth_asym_id_1",   "auth_asym_id_2", ""               },
        // { "ebi_validate_contact",       "auth_asym_id_1",   "auth_asym_id_2", ""               },
        { "ebi_validate_peptide_omega", "auth_asym_id_1",   "auth_asym_id_2", ""               },
        { "ebi_validate_torsion",       "auth_asym_id",     "",               ""               },
        { "rcsb_missing_residue_list",  "pdb_chain_id",     "",               ""               },
        { "struct_ref_seq",             "ndb_chain_id",     "",               ""               },
        { "struct_ref_seq_dif",         "ndb_pdb_chain_id", "",               ""               }
};

static const char *ncs_keywords[N_NCS_KEYWORDS] = {
        "dom_id",
       	"pdbx_type",
       	"pdbx_asym_id",
       	"pdbx_number",
        "rms_dev_position",
       	"weight_position"
};

static const char *ccp4_ncs_keywords[N_NCS_KEYWORDS] = {
        "dom_id",
       	"type",
       	"auth_asym_id",
       	"number",
        "rms_dev",
       	"weight"
};

static const char *lim_keywords[N_LIM_KEYWORDS] = {
        "dom_id",
       	"pdbx_component_id",
       	"pdbx_refine_code",
       	"beg_auth_asym_id",
       	"beg_auth_seq_id",
        "end_auth_asym_id",
       	"end_auth_seq_id"
};

static const char *lim_missing[N_LIM_MISSING] = {
       	"beg_auth_asym_id",
       	"beg_auth_seq_id",
        "end_auth_asym_id",
       	"end_auth_seq_id",
        "beg_label_asym_id",
       	"beg_label_comp_id",
       	"beg_label_seq_id",
       	"beg_label_alt_id",
        "end_label_asym_id",
       	"end_label_comp_id",
       	"end_label_seq_id",
       	"end_label_alt_id"
};

static const char *ccp4_lim_keywords[N_LIM_KEYWORDS] = {
        "dom_id",
       	"component_id",
       	"refine_code",
       	"beg_auth_asym_id",
       	"beg_auth_seq_id",
        "end_auth_asym_id",
       	"end_auth_seq_id"
};

static CONECT *connect = NULL;
static int Alloc_connect = 0;
static int Num_connect = 0;

static void add_connect_record(ConnectFormat *drug, _RESIDUE *residue);
static void insert_atom_id(const int atom_id);
static void insert_connect_atom_id(const int atom_id, const int atnum);

#ifdef SUN_OS
#ifdef __cplusplus
extern "C" {
#endif
#endif
static int connect_compare(const void *connect1, const void *connect2);
static int int_compare(const void *int1, const void *int2);
#ifdef SUN_OS
#ifdef __cplusplus
}
#endif
#endif

static void convert_remark_350_to_symmetry_operation(BTreeObj<_SSIndex>& _symmetry_chain,
        BTreeObj<_SSIndex>& _biol_unit, NDBCELLCONST *cell, char *remark_350,
        CifString &message, const char *structureID);
static char *get_line_from_string(char *line, char *string);
static char *get_chain_ids(const int entity_id, ISTable *Table);

int _FileObj::read_ndb_cif_file()
{
       int numBlock, numCategory, i, IsJRNL, card_id, ndb_token[5], serial_no;
       char *diags = NULL, **blockName = NULL, **categoryName = NULL, tokenName[MxStrLen];
       ISTable *Table = NULL;
       One_Card_Info_Format *pCard = NULL;

       if (_CifObj) { delete _CifObj; }
       get_temp_file();
       _CifObj = new CifFileObj(_temp_file, WRITE_MODE);
       _CifObj->Read(_real_filename, diags);
       if (diags && strlen(diags)) {
            ndb_log_message(NDB_MSG_ERR, "%s: Read cif file failed\n%s\n",
                    _input_filename, diags);
            exit (1);
       }

       blockName = _CifObj->GetBlockNames(numBlock);
       categoryName = _CifObj->GetTableNames(numCategory);

       for (i = 0; i < numCategory; i++) {
            if (!strcmp(categoryName[i], "NDB_atom_site")) continue;
            Table = _CifObj->GetTablePtr(blockName[0], categoryName[i]);
            IsJRNL = FALSE;
            if (!strncmp(categoryName[i], "NDB_", 4))
                  strcpy(tokenName, &categoryName[i][4]);
            else  strcpy(tokenName, categoryName[i]);

            if (!strcmp(tokenName, "AUTH") ||
                !strcmp(tokenName, "TITL") ||
                !strcmp(tokenName, "REF") ||
                !strcmp(tokenName, "PUBL") ||
                !strcmp(tokenName, "REFN") ||
                !strcmp(tokenName, "EDIT")) IsJRNL = TRUE;
            card_id = -1;
            if (IsJRNL) card_id = find_jrnl_id(tokenName);
            else        card_id = find_card_id(tokenName);
            if (card_id != -1)
                 read_general_cards(Table, card_id, IsJRNL);
       }
       pCard = _CardInfo[NDBFIL_TOKEN-1].Head();
       if (!strcmp(_StructureId, "") && pCard && strcmp(pCard->Field(1), ""))
            strcpy(_StructureId, pCard->Field(1));

       get_sequence_from_seqres();

       ndb_token[0] = SYMMA_TOKEN-1;
       ndb_token[1] = ORIGX_TOKEN-1;
       ndb_token[2] = SCALE_TOKEN-1;
       ndb_token[3] = MTRIX_TOKEN-1;
       ndb_token[4] = BIOMT_TOKEN-1;
       for (i = 0; i < 5; i++) {
            pCard = _CardInfo[ndb_token[i]].Head();
            serial_no = 1;
            while (pCard) {
                 sprintf(tokenName, "%s%d", pCard->Field(0), serial_no);
                 pCard->AddElementToField(tokenName, 0);
                 serial_no++;
                 if (serial_no > 3) serial_no = 1;
                 pCard = pCard->Next();
            }
       }

       if (_CifObj->IsTablePresent(blockName[0], "NDB_atom_site")) {
            Table = _CifObj->GetTablePtr(blockName[0], "NDB_atom_site");
            read_atom_records(Table);
       }

       pCard = _CardInfo[EXPDTA_TOKEN-1].Head();
       _experiment_type = EXPERIMENT_TYPE_XRAY;
       if (pCard) {
            CifString string;
            string.Copy(pCard->Field(2));
            ndb_set_string_to_upper(string.Text());
            if (strcmp(string.Text(), ""))
                 _experiment_type = EXPERIMENT_TYPE_BASIC;
            if (strstr(string.Text(), "X-RAY DIFFRACTION") ||
                strstr(string.Text(), "X-RAY POWDER DIFFRACTION"))
                 _experiment_type = EXPERIMENT_TYPE_XRAY;
            else if (strstr(string.Text(), "NMR"))
                 _experiment_type = EXPERIMENT_TYPE_NMR;
            else if (strstr(string.Text(), "SOLUTION SCATTERING"))
                 _experiment_type = EXPERIMENT_TYPE_SOLN_SCT;
            else if (strstr(string.Text(), "THEORETICAL MODEL"))
                 _experiment_type = EXPERIMENT_TYPE_MODEL;
            else if (strstr(string.Text(), "NEUTRON DIFFRACTION"))
                 _experiment_type = EXPERIMENT_TYPE_NEUTRON;
            else if (strstr(string.Text(), "ELECTRON DIFFRACTION"))
                 _experiment_type = EXPERIMENT_TYPE_ELECTRON;
            else if (strstr(string.Text(), "CRYO-ELECTRON MICROSCOPY") ||
                     strstr(string.Text(), "ELECTRON TOMOGRAPHY") ||
                     strstr(string.Text(), "ELECTRON MICROSCOPY"))
                 _experiment_type = EXPERIMENT_TYPE_CRYO_EM;
	    else if (strstr(string.Text(), "FIBER DIFFRACTION"))
                 _experiment_type = EXPERIMENT_TYPE_FIBER;
       }
       get_cell();

       for (i = 0; i < numCategory; i++) delete categoryName[i];
       delete categoryName;
       for (i = 0; i < numBlock; i++) delete blockName[i];
       delete blockName;
       if (_CifObj) { delete _CifObj; }
       _CifObj = NULL;

       return TRUE;
}

int _FileObj::read_mmcif_file(const int exchange)
{
       int numBlock, i, j, index, nCols, nRows, is_refmac5 = 0, ncs_index[N_NCS_KEYWORDS];
       int t_index, j_index, f_index, l_index, v_index, y_index, lim_index[N_LIM_KEYWORDS];
       int beg_label_asym_id, beg_label_seq_id, end_label_asym_id, end_label_seq_id;
       int beg_auth_asym_id, beg_auth_seq_id, end_auth_asym_id, end_auth_seq_id;
       int beg_label_asym_yes, beg_label_seq_yes, end_label_asym_yes, end_label_seq_yes;
       int beg_auth_asym_no, beg_auth_seq_no, end_auth_asym_no, end_auth_seq_no;
       char *diags = NULL, **blockName = NULL, buffer[100];
       CifString cifstring, cifstring1, title, journal_abbrev, page_first, page_last, year;
       CifString journal_volume;
       ISTable *Table = NULL, *Table1 = NULL;
       ReVarCifArray<CifString> *ColumnNames = NULL;

       if (_CifObj) { delete _CifObj; }
       if (_is_binary == 1) _CifObj = new CifFileObj(_real_filename, UPDATE_MODE);
       else {
            if (_is_binary)
                 _CifObj = new CifFileObj(_output_filename, WRITE_MODE);
            else {
                 get_temp_file();
                 _CifObj = new CifFileObj(_temp_file, WRITE_MODE);
            }
            _CifObj->Read(_real_filename, diags);
            if (diags && strlen(diags)) {
                 ndb_log_message(NDB_MSG_ERR, "%s: Read cif file failed\n%s\n",
                         _input_filename, diags);
                 exit (1);
            }
       }

       blockName = _CifObj->GetBlockNames(numBlock);
       strcpy(_StructureId, blockName[0]);
       for (i = 0; i < numBlock; i++) delete [] blockName[i];
       delete [] blockName;

       Table = _CifObj->GetTablePtr(_StructureId, "computing");
       if (!is_empty_table(Table)) {
            index = Table->GetColumnIndex("structure_refinement");
            Table->GetCell(cifstring, index, 0);
            strcpy(buffer, cifstring.Text());
            ndb_set_string_to_upper(buffer);
            if (strstr(buffer, "REFMAC 5") || strstr(buffer, "REFMAC5")) is_refmac5 = 1;
       }

       if (exchange) {
            set_exchange_status(EXCHANGE_TO_INTERNAL);
            _CifObj = dictionary_exchange(_CifObj, _temp_file);
       } else cif_dictionary_patch(_StructureId, _CifObj);

       if (is_refmac5) {
            Table = _CifObj->GetTablePtr(_StructureId, "ccp4_refine_ls_restr_ncs");
            Table1 = _CifObj->GetTablePtr(_StructureId, "refine_ls_restr_ncs");
            if (!is_empty_table(Table) && is_empty_table(Table1)) {
                 _CifObj->DeleteTable(_StructureId, "refine_ls_restr_ncs");
                 Table1 = new ISTable("refine_ls_restr_ncs");
                 for (i = 0; i < N_NCS_KEYWORDS; i++) {
                      ncs_index[i] = -1;
                      ncs_index[i] = Table->GetColumnIndex(ccp4_ncs_keywords[i]);
                      Table1->AddColumn(ncs_keywords[i]);
                 }
                 nRows = Table->GetNumRows();
                 for (i = 0; i < nRows; i++) {
                      Table1->AddRow();
		      for (j = 0; j < N_NCS_KEYWORDS; j++) {
                           if (ncs_index[j] >= 0) {
                                Table->GetCell(cifstring, ncs_index[j], i);
				Table1->UpdateCell(cifstring, j, i);
                           }
                      }
		 }
                 _CifObj->DeleteTable(_StructureId, "ccp4_refine_ls_restr_ncs");
		 _CifObj->WriteTable(Table1, _StructureId, "refine_ls_restr_ncs");
	    } else if (!is_empty_table(Table) && !is_empty_table(Table1)) {
                 if (Table->GetNumRows() == Table1->GetNumRows()) {
                      _CifObj->DeleteTable(_StructureId, "ccp4_refine_ls_restr_ncs");
                 }
            }

            Table = _CifObj->GetTablePtr(_StructureId, "ccp4_struct_ncs_dom_lim");
            Table1 = _CifObj->GetTablePtr(_StructureId, "struct_ncs_dom_lim");
            if (!is_empty_table(Table1)) {
                 nRows = Table1->GetNumRows();
                 beg_label_asym_id = Table1->GetColumnIndex("beg_label_asym_id");
                 beg_label_seq_id = Table1->GetColumnIndex("beg_label_seq_id");
                 end_label_asym_id = Table1->GetColumnIndex("end_label_asym_id");
                 end_label_seq_id = Table1->GetColumnIndex("end_label_seq_id");
                 beg_auth_asym_id = Table1->GetColumnIndex("beg_auth_asym_id");
                 beg_auth_seq_id = Table1->GetColumnIndex("beg_auth_seq_id");
                 end_auth_asym_id = Table1->GetColumnIndex("end_auth_asym_id");
                 end_auth_seq_id = Table1->GetColumnIndex("end_auth_seq_id");
		 if (beg_label_asym_id >= 0 && beg_label_seq_id >= 0 &&
                     end_label_asym_id >= 0 && end_label_seq_id >= 0 &&
		     beg_auth_asym_id  >= 0 && beg_auth_seq_id  >= 0 &&
		     end_auth_asym_id  >= 0 && end_auth_seq_id  >= 0) {
                      beg_label_asym_yes = beg_label_seq_yes =
                      end_label_asym_yes = end_label_seq_yes = 
                      beg_auth_asym_no = beg_auth_seq_no =
		      end_auth_asym_no = end_auth_seq_no = 1;
                      for (i = 0; i < nRows; i++) {
                           Table1->GetCell(cifstring, beg_label_asym_id, i);
                           if (cifstring == "" || cifstring == "?" || cifstring == ".")
                                beg_label_asym_yes = 0;

                           Table1->GetCell(cifstring, beg_label_seq_id, i);
                           if (cifstring == "" || cifstring == "?" || cifstring == ".")
                                beg_label_seq_yes = 0;

                           Table1->GetCell(cifstring, end_label_asym_id, i);
                           if (cifstring == "" || cifstring == "?" || cifstring == ".")
                                end_label_asym_yes = 0;

                           Table1->GetCell(cifstring, end_label_seq_id, i);
                           if (cifstring == "" || cifstring == "?" || cifstring == ".")
                                end_label_seq_yes = 0;

                           Table1->GetCell(cifstring, beg_auth_asym_id, i);
                           if (cifstring != "" && cifstring != "?" && cifstring != ".")
                                beg_auth_asym_no = 0;

                           Table1->GetCell(cifstring, beg_auth_seq_id, i);
                           if (cifstring != "" && cifstring != "?" && cifstring != ".")
                                beg_auth_seq_no = 0;

                           Table1->GetCell(cifstring, end_auth_asym_id, i);
                           if (cifstring != "" && cifstring != "?" && cifstring != ".")
                                end_auth_asym_no = 0;

                           Table1->GetCell(cifstring, end_auth_seq_id, i);
                           if (cifstring != "" && cifstring != "?" && cifstring != ".")
                                end_auth_seq_no = 0;
                      }

                      if (beg_label_asym_yes && beg_label_seq_yes && end_label_asym_yes &&
                          end_label_seq_yes && beg_auth_asym_no && beg_auth_seq_no &&
                          end_auth_asym_no && end_auth_seq_no) {
                           for (i = 0; i < nRows; i++) {
                                Table1->GetCell(cifstring, beg_label_asym_id, i);
                                Table1->UpdateCell(cifstring, beg_auth_asym_id, i);
                                Table1->GetCell(cifstring, beg_label_seq_id, i);
                                Table1->UpdateCell(cifstring, beg_auth_seq_id, i);
                                Table1->GetCell(cifstring, end_label_asym_id, i);
                                Table1->UpdateCell(cifstring, end_auth_asym_id, i);
                                Table1->GetCell(cifstring, end_label_seq_id, i);
                                Table1->UpdateCell(cifstring, end_auth_seq_id, i);
                           }
		           _CifObj->WriteTable(Table1, _StructureId, "struct_ncs_dom_lim");
                      }
                 } else if (beg_label_asym_id >= 0 && beg_label_seq_id >= 0 &&
                            end_label_asym_id >= 0 && end_label_seq_id >= 0 &&
		            beg_auth_asym_id  <  0 && beg_auth_seq_id  <  0 &&
		            end_auth_asym_id  <  0 && end_auth_seq_id  <  0) {
                      Table1->RenameColumn("beg_label_asym_id", "beg_auth_asym_id");
                      Table1->RenameColumn("beg_label_seq_id", "beg_auth_seq_id");
                      Table1->RenameColumn("end_label_asym_id", "end_auth_asym_id");
                      Table1->RenameColumn("end_label_seq_id", "end_auth_seq_id");
		      _CifObj->WriteTable(Table1, _StructureId, "struct_ncs_dom_lim");
                 }
	    } 

            if (!is_empty_table(Table) && is_empty_table(Table1)) {
                 _CifObj->DeleteTable(_StructureId, "struct_ncs_dom_lim");
                 Table1 = new ISTable("struct_ncs_dom_lim");
                 for (i = 0; i < N_LIM_KEYWORDS; i++) {
                      lim_index[i] = -1;
                      lim_index[i] = Table->GetColumnIndex(ccp4_lim_keywords[i]);
                      Table1->AddColumn(lim_keywords[i]);
                 }
                 nRows = Table->GetNumRows();
                 for (i = 0; i < nRows; i++) {
                      Table1->AddRow();
		      for (j = 0; j < N_LIM_KEYWORDS; j++) {
                           if (lim_index[j] >= 0) {
                                Table->GetCell(cifstring, lim_index[j], i);
				Table1->UpdateCell(cifstring, j, i);
                           }
                      }
		 }
                 _CifObj->DeleteTable(_StructureId, "ccp4_struct_ncs_dom_lim");
		 _CifObj->WriteTable(Table1, _StructureId, "struct_ncs_dom_lim");
	    } else if (!is_empty_table(Table) && !is_empty_table(Table1)) {
                 if (Table->GetNumRows() == Table1->GetNumRows()) {
                      _CifObj->DeleteTable(_StructureId, "ccp4_struct_ncs_dom_lim");
                 }
            }

	    Table = _CifObj->GetTablePtr(_StructureId, "struct_ncs_dom_lim");
	    if (!is_empty_table(Table)) {
                 nRows = Table->GetNumRows();
		 nCols = Table->GetNumColumns();
                 ColumnNames = new ReVarCifArray<CifString>;
		 cifstring.Copy(".");
                 for (i = 0; i < nRows; i++) ColumnNames->Add(cifstring);
		 for (i = 0; i < N_LIM_MISSING; i++) {
                      beg_label_asym_id = Table->GetColumnIndex(lim_missing[i]);
		      if (beg_label_asym_id < 0) {
                           Table->AddColumn(lim_missing[i]);
			   Table->FillColumn(*ColumnNames, nCols);
			   nCols++;
		      } 
                 }
		 _CifObj->WriteTable(Table1, _StructureId, "struct_ncs_dom_lim");
                 delete ColumnNames; ColumnNames = NULL;
            }

            Table = _CifObj->GetTablePtr(_StructureId, "ccp4_refine_tls_group");
	    if (!is_empty_table(Table)) {
                 nRows = Table->GetNumRows();
		 nCols = Table->GetNumColumns();
                 ColumnNames = new ReVarCifArray<CifString>;
		 cifstring.Copy(".");
                 for (i = 0; i < nRows; i++) ColumnNames->Add(cifstring);

                 beg_label_asym_id = Table->GetColumnIndex("beg_label_asym_id");
                 if (beg_label_asym_id < 0) {
                      Table->AddColumn("beg_label_asym_id");
                      Table->FillColumn(*ColumnNames, nCols);
                      nCols++;
                      beg_label_asym_id = Table->GetColumnIndex("beg_label_asym_id");
                 }
                 beg_label_seq_id = Table->GetColumnIndex("beg_label_seq_id");
                 if (beg_label_seq_id < 0) {
                      Table->AddColumn("beg_label_seq_id");
                      Table->FillColumn(*ColumnNames, nCols);
                      nCols++;
                      beg_label_seq_id = Table->GetColumnIndex("beg_label_seq_id");
                 }
                 end_label_asym_id = Table->GetColumnIndex("end_label_asym_id");
                 if (end_label_asym_id < 0) {
                      Table->AddColumn("end_label_asym_id");
                      Table->FillColumn(*ColumnNames, nCols);
                      nCols++;
                      end_label_asym_id = Table->GetColumnIndex("end_label_asym_id");
                 }
                 end_label_seq_id = Table->GetColumnIndex("end_label_seq_id");
                 if (end_label_seq_id < 0) {
                      Table->AddColumn("end_label_seq_id");
                      Table->FillColumn(*ColumnNames, nCols);
                      nCols++;
                      end_label_seq_id = Table->GetColumnIndex("end_label_seq_id");
                 }
                 beg_auth_asym_id = Table->GetColumnIndex("beg_auth_asym_id");
                 if (beg_auth_asym_id < 0) {
                      Table->AddColumn("beg_auth_asym_id");
                      Table->FillColumn(*ColumnNames, nCols);
                      nCols++;
                      beg_auth_asym_id = Table->GetColumnIndex("beg_auth_asym_id");
                 }
                 beg_auth_seq_id = Table->GetColumnIndex("beg_auth_seq_id");
                 if (beg_auth_seq_id < 0) {
                      Table->AddColumn("beg_auth_seq_id");
                      Table->FillColumn(*ColumnNames, nCols);
                      nCols++;
                      beg_auth_seq_id = Table->GetColumnIndex("beg_auth_seq_id");
                 }
                 end_auth_asym_id = Table->GetColumnIndex("end_auth_asym_id");
                 if (end_auth_asym_id < 0) {
                      Table->AddColumn("end_auth_asym_id");
                      Table->FillColumn(*ColumnNames, nCols);
                      nCols++;
                      end_auth_asym_id = Table->GetColumnIndex("end_auth_asym_id");
                 }
                 end_auth_seq_id = Table->GetColumnIndex("end_auth_seq_id");
                 if (end_auth_seq_id < 0) {
                      Table->AddColumn("end_auth_seq_id");
                      Table->FillColumn(*ColumnNames, nCols);
                      nCols++;
                      end_auth_seq_id = Table->GetColumnIndex("end_auth_seq_id");
                 }
                 delete ColumnNames; ColumnNames = NULL;

                 beg_label_asym_yes = beg_label_seq_yes = 0;
                 end_label_asym_yes = end_label_seq_yes = 0;
                 beg_auth_asym_no = beg_auth_seq_no = 1;
                 end_auth_asym_no = end_auth_seq_no = 1;
                 for (i = 0; i < nRows; i++) {
/*
                      Table->GetCell(cifstring, beg_label_asym_id, i);
                      if (cifstring == "" || cifstring == "?" || cifstring == ".")
                           beg_label_asym_yes = 0;

                      Table->GetCell(cifstring, beg_label_seq_id, i);
                      if (cifstring == "" || cifstring == "?" || cifstring == ".")
                           beg_label_seq_yes = 0;

                      Table->GetCell(cifstring, end_label_asym_id, i);
                      if (cifstring == "" || cifstring == "?" || cifstring == ".")
                           end_label_asym_yes = 0;

                      Table->GetCell(cifstring, end_label_seq_id, i);
                      if (cifstring == "" || cifstring == "?" || cifstring == ".")
                           end_label_seq_yes = 0;
*/
                      Table->GetCell(cifstring, beg_label_asym_id, i);
                      if (cifstring != "" && cifstring != "?" && cifstring != ".")
                           beg_label_asym_yes = 1;

                      Table->GetCell(cifstring, beg_label_seq_id, i);
                      if (cifstring != "" && cifstring != "?" && cifstring != ".")
                           beg_label_seq_yes = 1;

                      Table->GetCell(cifstring, end_label_asym_id, i);
                      if (cifstring != "" && cifstring != "?" && cifstring != ".")
                           end_label_asym_yes = 1;

                      Table->GetCell(cifstring, end_label_seq_id, i);
                      if (cifstring != "" && cifstring != "?" && cifstring != ".")
                           end_label_seq_yes = 1;

                      Table->GetCell(cifstring, beg_auth_asym_id, i);
                      if (cifstring != "" && cifstring != "?" && cifstring != ".")
                           beg_auth_asym_no = 0;

                      Table->GetCell(cifstring, beg_auth_seq_id, i);
                      if (cifstring != "" && cifstring != "?" && cifstring != ".")
                           beg_auth_seq_no = 0;

                      Table->GetCell(cifstring, end_auth_asym_id, i);
                      if (cifstring != "" && cifstring != "?" && cifstring != ".")
                           end_auth_asym_no = 0;

                      Table->GetCell(cifstring, end_auth_seq_id, i);
                      if (cifstring != "" && cifstring != "?" && cifstring != ".")
                           end_auth_seq_no = 0;
                 }

                 if (beg_label_asym_yes && beg_label_seq_yes && end_label_asym_yes &&
                     end_label_seq_yes && beg_auth_asym_no && beg_auth_seq_no &&
                     end_auth_asym_no && end_auth_seq_no) {
                      cifstring1.Copy(".");
                      for (i = 0; i < nRows; i++) {
                           Table->GetCell(cifstring, beg_label_asym_id, i);
                           Table->UpdateCell(cifstring1, beg_label_asym_id, i);
                           Table->UpdateCell(cifstring, beg_auth_asym_id, i);
                           Table->GetCell(cifstring, beg_label_seq_id, i);
			   Table->UpdateCell(cifstring1, beg_label_seq_id, i);
                           Table->UpdateCell(cifstring, beg_auth_seq_id, i);
                           Table->GetCell(cifstring, end_label_asym_id, i);
			   Table->UpdateCell(cifstring1, end_label_asym_id, i);
                           Table->UpdateCell(cifstring, end_auth_asym_id, i);
                           Table->GetCell(cifstring, end_label_seq_id, i);
			   Table->UpdateCell(cifstring1, end_label_seq_id, i);
                           Table->UpdateCell(cifstring, end_auth_seq_id, i);
                      }
                      _CifObj->WriteTable(Table, _StructureId, "ccp4_refine_tls_group");
                 }
            }
       }

       Table = _CifObj->GetTablePtr(_StructureId, "citation");
       if (!is_empty_table(Table)) {
            t_index = Table->GetColumnIndex("title");
            j_index = Table->GetColumnIndex("journal_abbrev");
            v_index = Table->GetColumnIndex("journal_volume");
            f_index = Table->GetColumnIndex("page_first");
            l_index = Table->GetColumnIndex("page_last");
            y_index = Table->GetColumnIndex("year");
            get_value(title, Table, t_index, 0);
            get_value(journal_abbrev, Table, j_index, 0);
            get_value(journal_volume, Table, v_index, 0);
            get_value(page_first, Table, f_index, 0);
            get_value(page_last, Table, l_index, 0);
            get_value(year, Table, y_index, 0);
            if (title != "" && journal_abbrev == "" && journal_volume == "" &&
                page_first == "" && page_last == "" && year == "") {
                 journal_abbrev.Copy("TO BE PUBLISHED");
                 Table->UpdateCell(journal_abbrev, j_index, 0);
                 _CifObj->WriteTable(Table, _StructureId, "citation");
            }
       }

       return TRUE;
}

int _FileObj::read_mmcif_file(int &encapsulated_coordinates_format, const int special,
       const int exchange, const int short_version, const int get_symmetry_flag,
       const int no_dbref, const int adit_view)
{
       int input_format = NDB_FILE_FORMAT_PDB, num_index = -1, release_index = -1, method_index = -1;
       char Table_Name[100], buffer[300];
       CifString string, buff;
       ISTable *Table = NULL, *Table1 = NULL, *rev_table = NULL, *ndb_table = NULL;
       FILE *fp = NULL;

       read_mmcif_file(exchange);

       _file_status = FILE_STATUS_MMCIF;
       encapsulated_coordinates_format = 0;

       if (_CifObj->IsTablePresent(_StructureId, "ndb_original_pdb_coordinates")) {
            _file_status = FILE_STATUS_MMCIF_PLUS_MOL;
            encapsulated_coordinates_format = NDB_FILE_FORMAT_PDB;
            input_format = NDB_FILE_FORMAT_PDB;
            Table = _CifObj->GetTablePtr(_StructureId, "ndb_original_pdb_coordinates");
            strcpy(Table_Name, "ndb_original_pdb_coordinates");
       } else if (_CifObj->IsTablePresent(_StructureId, "ndb_original_ndb_coordinates")) {
            _file_status = FILE_STATUS_MMCIF_PLUS_MOL;
            encapsulated_coordinates_format = NDB_FILE_FORMAT_MIX;
            input_format = NDB_FILE_FORMAT_MIX;
            Table = _CifObj->GetTablePtr(_StructureId, "ndb_original_ndb_coordinates");
            strcpy(Table_Name, "ndb_original_ndb_coordinates");
       }
       get_cell();
       if (get_symmetry_flag) get_symmetry_from_remark_350();

       cifndb_process_category_correspond();
       cifndb_process_asym_poly_seq_ndb(no_dbref);
       if (!_Num_Seqs) match_entity_poly_seq_to_seqres(no_dbref);
       if (!_Num_Seqs) match_entity_poly_to_seqres(no_dbref);

       if (!short_version) {
            if (Table) {
                 Table->GetCell(string, 0, 0);
                 strcpy(buffer, _real_filename);
                 strcpy(_real_filename, "special_cif_output_file");
                 fp = fopen("special_cif_output_file", "w");
                 fprintf(fp, "%s\n", string.Text());
                 fclose(fp);
     
                 read_ndb_pdb_file(input_format, 0, adit_view);
                 system("rm special_cif_output_file");
                 _CifObj->DeleteTable(_StructureId, Table_Name);
                 strcpy(_real_filename, buffer);
            } else put_coordinates_into_molecules(special);
       }

       if (!_Num_Seqs) {
            get_seqres_from_mapping(NDB_FILE_FORMAT_NDB);
            get_sequence_from_seqres();
       }

       if (_CifObj->IsTablePresent(_StructureId, "exptl")) {
            Table = _CifObj->GetTablePtr(_StructureId, "exptl");
            method_index = Table->GetColumnIndex("method");
            string.Copy("");
            Table->GetCell(string, method_index, 0);
            ndb_set_string_to_upper(string.Text());
            if (strcmp(string.Text(), ""))
                 _experiment_type = EXPERIMENT_TYPE_BASIC;
            if (strstr(string.Text(), "X-RAY DIFFRACTION") ||
                strstr(string.Text(), "X-RAY POWDER DIFFRACTION"))
                 _experiment_type = EXPERIMENT_TYPE_XRAY;
            else if (strstr(string.Text(), "NMR"))
                 _experiment_type = EXPERIMENT_TYPE_NMR;
            else if (strstr(string.Text(), "SOLUTION SCATTERING"))
                 _experiment_type = EXPERIMENT_TYPE_SOLN_SCT;
            else if (strstr(string.Text(), "THEORETICAL MODEL"))
                 _experiment_type = EXPERIMENT_TYPE_MODEL;
            else if (strstr(string.Text(), "NEUTRON DIFFRACTION"))
                 _experiment_type = EXPERIMENT_TYPE_NEUTRON;
            else if (strstr(string.Text(), "ELECTRON DIFFRACTION"))
                 _experiment_type = EXPERIMENT_TYPE_ELECTRON;
            else if (strstr(string.Text(), "CRYO-ELECTRON MICROSCOPY") ||
                     strstr(string.Text(), "ELECTRON TOMOGRAPHY") ||
                     strstr(string.Text(), "ELECTRON MICROSCOPY"))
                 _experiment_type = EXPERIMENT_TYPE_CRYO_EM;
	    else if (strstr(string.Text(), "FIBER DIFFRACTION"))
                 _experiment_type = EXPERIMENT_TYPE_FIBER;
       } else if (adit_view) {
            if (adit_view == 1) _experiment_type = EXPERIMENT_TYPE_XRAY;
	    else if (adit_view == 2)  _experiment_type = EXPERIMENT_TYPE_NMR;
            else if (adit_view == 4) _experiment_type = EXPERIMENT_TYPE_CRYO_EM;
       } else _experiment_type = EXPERIMENT_TYPE_XRAY;

       ndb_table = _CifObj->GetTablePtr(_StructureId, "ndb_database_status");
       if (!is_empty_table(ndb_table)) {
            release_index = ndb_table->GetColumnIndex("date_of_NDB_release");
            method_index = ndb_table->GetColumnIndex("status_code");
            if (release_index >= 0 && method_index >= 0) {
                 ndb_table->GetCell(string, method_index, 0);
                 get_value(buff, ndb_table, release_index, 0);
                 if (buff != "" && string == "REL") {
                      rev_table = _CifObj->GetTablePtr(_StructureId, "database_PDB_rev");
                      if (!rev_table) {
                           rev_table = new ISTable("database_PDB_rev");
                           rev_table->AddColumn("num");
                           rev_table->AddColumn("date");
                           rev_table->AddColumn("date_original");
                           rev_table->AddColumn("status");
                           rev_table->AddColumn("replaces");
                           rev_table->AddColumn("mod_type");
                           rev_table->AddRow();
                      }
                      num_index = rev_table->GetColumnIndex("num");
                      string.Copy("1");
                      rev_table->UpdateCell(string, num_index, 0);
                      num_index = rev_table->GetColumnIndex("date");
                      rev_table->UpdateCell(buff, num_index, 0);
                      num_index = rev_table->GetColumnIndex("mod_type");
                      string.Copy("0");
                      rev_table->UpdateCell(string, num_index, 0);

                      buffer[0] = '\0';
                      Table1 = _CifObj->GetTablePtr(_StructureId, "database_2");
                      if (!is_empty_table(Table1)) get_pdb_id(Table1, buffer);
                      num_index = rev_table->GetColumnIndex("replaces");
                      string.Copy(buffer);
                      rev_table->UpdateCell(string, num_index, 0);
                      _CifObj->WriteTable(rev_table, _StructureId, "database_PDB_rev");
                 }
            }
       }

       return TRUE;
}

void _FileObj::write_ndb_file(const int format, const int remove_hydrogen,
              const int create_seqres, const int is_iupac, const int short_version)
{
       int numRemark = 0, i = 0, displayOrder = 0, card_id = 0;
       FILE *fp = NULL;
       
       _output_format = format;
       if (_output_format == NDB_FILE_FORMAT_PDB) {
            for (i = 0; i < _num_mol; i++) {
                 _molecules[i]->reorder_chain_order();
            }
       }

       if (create_seqres) {
            if (!short_version) get_seqres_from_mapping(_output_format);
       }

       if (format == NDB_FILE_FORMAT_NDBCIF) {
            if (_CifObj) delete _CifObj;
            get_temp_file();
            _CifObj = new CifFileObj(_temp_file, WRITE_MODE);
       } else fp = fopen(_output_filename, "w");

       for (displayOrder = 0; displayOrder < MxTokens; displayOrder++) {
            for (card_id = 0; card_id < MxTokens; card_id++) 
                 if (displayOrder == NdbToken[card_id].DisplayOrder) break;

            if (card_id == ATOMN_TOKEN-1) {
                 if (!short_version) print_molecules(fp, format, remove_hydrogen, create_seqres, is_iupac);
            }

            if (card_id == MxTokens || card_id + 1 != JRNL_TOKEN &&
                _CardInfo[card_id].Head() == NULL || card_id == NDBSQN_TOKEN-1 ||
                card_id == SYMTRA_TOKEN-1 || format == NDB_FILE_FORMAT_PDB &&
               (NdbToken[card_id].NdbOnly || card_id + 1 == REMARK_TOKEN ||
                card_id == MASTER_TOKEN-1)) continue;

            if (card_id +1 != JRNL_TOKEN) {
                 if (format == NDB_FILE_FORMAT_NDBCIF)
                      write_general_cards(card_id, FALSE);
                 else print_general_cards(fp, card_id, format, numRemark, 0);
            } else {
                 if (format == NDB_FILE_FORMAT_NDBCIF) write_jrnl_cards();
                 else {
                      print_jrnl_cards(fp, format, numRemark);
                      if (format == NDB_FILE_FORMAT_PDB)
                           print_general_cards(fp, REMARK_TOKEN - 1, format, numRemark, 0);
                 }
            }
       }
       if (format == NDB_FILE_FORMAT_NDBCIF) {
            _CifObj->Write(_output_filename);
       } else {
            if (format == NDB_FILE_FORMAT_PDB) {
                 print_master_card(fp, numRemark);
                 fprintf(fp, "END");
                 for (i = 0; i < 77; i++) fprintf(fp, " ");
                 fprintf(fp, "\n");
            } else fprintf(fp, "END\n");
            fclose (fp);
       }
}

void _FileObj::read_general_cards(ISTable *Table, const int card_id, const int IsJRNL)
{
       int i = 0, j = 0, rowNo, colNo, irow, icol;
       int cols[200]; char value[2000];
       CifString string;
       ndb_token_format *Tokens = NULL;
       One_Card_Info_Format *pCard = NULL;
       ReVarCifArray<CifString> *ColumnNames = NULL;

       if (is_empty_table(Table)) return;

       Tokens = NdbToken;
       if (IsJRNL) Tokens = JrnlToken;
       rowNo = Table->GetNumRows();
       colNo = Table->GetNumColumns();

       ColumnNames = Table->GetColumnNames();
       for (i = 0; i < colNo; i++) {
            for (j = 1; j < Tokens[card_id].NumField; j++) {
                 if (!strcasecmp(Tokens[card_id].FieldList[j].FieldLabel,
                       (*ColumnNames)[i].Text())) break;
                 if (IsJRNL && !strcasecmp(Tokens[card_id].FieldList[j].FieldLabel,
                         "Serial_No") && strstr((*ColumnNames)[i].Text(), "Serial_No"))
                      break;
            }
            if (j == Tokens[card_id].NumField) cols[i] = -1;
            else                               cols[i] = j;
       }
       if (ColumnNames) delete ColumnNames;

       for (irow = 0; irow < rowNo; irow++) {
            pCard = NULL;
            if (IsJRNL) pCard = _JrnlCardInfo[card_id].add_a_new_card(JrnlToken[card_id], 1);
            else pCard = _CardInfo[card_id].add_a_new_card(NdbToken[card_id], 0);

            for (icol = 0; icol < colNo; icol++) {
                 j = cols[icol];
                 if (j < 0 || j >= Tokens[card_id].NumField) continue;
                 Table->GetCell(string, icol, irow);
                 if (string == "?" || string == "." || string == "") continue;
                 strcpy(value, string.Text());
                 if (card_id == REMARK_TOKEN-1 || card_id == PREMRK_TOKEN-1 ||
                     card_id == PCOMPN_TOKEN-1 || card_id == PSOURC_TOKEN-1) {
                      if (value[0] == '\n') strcpy(value, &value[1]);
                 } else ndb_clean_string(value);
                 pCard->AddElementToField(value, j);
            }
       }
}

void _FileObj::read_atom_records(ISTable *Table)
{
       int model_pos = -1, prev_model = -1, i, j, irow, icol, model = 1;
       int rowNo, colNo, cols[200], card_id;
       char FieldInfo[MxField][MxLen];
       char prevChainId[MxStrLen], prevResidueNo[MxStrLen];
       char prevResidueName[MxStrLen], prevInsCode[MxStrLen];
       CifString buffer;
       _RESIDUE *residue = NULL;
       ReVarCifArray<CifString> *ColumnNames = NULL;

       if (is_empty_table(Table)) return;

       rowNo = Table->GetNumRows();
       colNo = Table->GetNumColumns();
       ColumnNames = Table->GetColumnNames();
       for (i = 0; i < colNo; i++) {
            if (!strcmp((*ColumnNames)[i].Text(), "ndb_model")) model_pos = i;
            for (j = 0; j < NdbToken[ATOMN_TOKEN-1].NumField; j++)
                 if (!strcmp(NdbToken[ATOMN_TOKEN-1].FieldList[j].FieldLabel,
                        (*ColumnNames)[i].Text())) break;
            if (j == NdbToken[ATOMN_TOKEN-1].NumField)
                 cols[i] = -1;
            else cols[i] = j;
       }
       if (ColumnNames) delete ColumnNames;

       strcpy(prevChainId, "XXX");
       strcpy(prevResidueNo, "XXX");
       strcpy(prevResidueName, "XXX");
       strcpy(prevInsCode, "XXX");

       for (irow = 0; irow < rowNo; irow++) {
            for (i = 0; i < NdbToken[ATOMN_TOKEN-1].NumField; i++)
                 FieldInfo[i][0] = '\0';

            for (icol = 0; icol < colNo; icol++) {
                 if (icol == model_pos) {
                      get_value(buffer, Table, icol, irow);
                      model = atoi(buffer.Text());
                      continue;
                 } else {
                      j = cols[icol];
                      if (j < 0 || j >= NdbToken[ATOMN_TOKEN-1].NumField) continue;
                      get_value(buffer, Table, icol, irow);
                      strcpy(FieldInfo[j], buffer.Text());
                 }
            }

            if (strcmp(FieldInfo[5], prevChainId) ||
                strcmp(FieldInfo[6], prevResidueNo) ||
                strcmp(FieldInfo[4], prevResidueName) ||
                strcmp(FieldInfo[7], prevInsCode)) {
                 if (residue) {
                      _molecules[_num_mol-1]->insert_a_residue(0, residue,
                            _error_message, _unrecognized_residue);
                      residue = NULL;
                 }
                 residue = new _RESIDUE;
                 residue->set_ResName(FieldInfo[4]);
                 residue->set_res_no(atoi(FieldInfo[6]));
                 if (strcmp(FieldInfo[15], ""))
                      residue->set_pdb_res_no(atoi(FieldInfo[15]));
                 else residue->set_pdb_res_no(residue->res_no());
            }
            card_id = find_card_id(FieldInfo[0]);
            residue->insert_a_atom(card_id, FieldInfo, 0, _error_message);

            strcpy(prevChainId, FieldInfo[5]);
            strcpy(prevResidueNo, FieldInfo[6]);
            strcpy(prevResidueName, FieldInfo[4]);
            strcpy(prevInsCode, FieldInfo[7]);

            if (model != prev_model) {
                 prev_model = model;
                 _ManageSpace();
                 _molecules[_num_mol-1]->set_Mol_ID(model);
            }
       }

       if (residue) {
            _molecules[_num_mol-1]->insert_a_residue(0, residue,
                     _error_message, _unrecognized_residue);
            residue = NULL;
       }

       for (i = 0; i < _num_mol; i++)
            _molecules[i]->check_chain_type(_Num_Seqs, _Seqs);
}

void _FileObj::write_general_cards(const int card_id, const int IsJRNL)
{
       int irow = 0, SerialNo = 0, NumField = 0, i;
       char category[MxStrLen];
       CifString string;
       ISTable *Table = NULL;
       One_Card_Info_Format *pCard = NULL;

       if (IsJRNL) pCard = _JrnlCardInfo[card_id].Head();
       else        pCard = _CardInfo[card_id].Head();
       if (pCard == NULL) return;

       if (IsJRNL) sprintf(category, "NDB_%s", JrnlToken[card_id].TokenName);
       else        sprintf(category, "NDB_%s", NdbToken[card_id].TokenName);
       Table = new ISTable(category);
       
       if (IsJRNL) {
            for (i = 1; i < JrnlToken[card_id].NumField; i++)
                 Table->AddColumn(JrnlToken[card_id].FieldList[i].FieldLabel, FALSE);
       } else {
            for (i = 1; i < NdbToken[card_id].NumField; i++)
                 Table->AddColumn(NdbToken[card_id].FieldList[i].FieldLabel, FALSE);
       }

       while (pCard) {
            if (!IsJRNL && NdbToken[card_id].MatrixType == 0 &&
                 NdbToken[card_id].SeqField != 0) {
                 SerialNo++;
                 if (!strcmp(pCard->Field(NdbToken[card_id].SeqField-1), "")) {
                      pCard->AddElementToField(SerialNo, NdbToken[card_id].SeqField-1);
                 }
            }
            Table->AddRow();

            NumField = 0;
            if (IsJRNL) NumField = JrnlToken[card_id].NumField;
            else        NumField = NdbToken[card_id].NumField;
            for (i = 1; i < NumField; i++) {
                 if (!IsJRNL && !NdbToken[card_id].FieldList[i].MixedCase)
                      pCard->UpperCaseField(i);
                 pCard->StripTrailingBlanks(i);
                 string.Copy(pCard->Field(i));
                 Table->UpdateCell(string, i-1, irow);
            }
            irow++;
            pCard = pCard->Next();
       }
       _CifObj->WriteTable(Table, _StructureId, category);
}

void _FileObj::write_jrnl_cards()
{
       int vField = 5, astmField = 3, issnField = 6, jrnl_id, displayOrder, card_id;
       One_Card_Info_Format *pJrnlCard = NULL;

       if (_JrnlCardInfo == NULL) return;

       jrnl_id = find_jrnl_id("REF");
       pJrnlCard = _JrnlCardInfo[jrnl_id].Head();
       while (pJrnlCard != NULL) {
            if (strcmp(pJrnlCard->Field(vField+1), "") &&
               !strcmp(pJrnlCard->Field(vField), ""))
                 pJrnlCard->AddElementToField("V.", vField);
            else if (!strcmp(pJrnlCard->Field(vField+1), ""))
                 pJrnlCard->AddElementToField("", vField);
            pJrnlCard = pJrnlCard->Next();
       }

       jrnl_id = find_jrnl_id("REFN");
       pJrnlCard = _JrnlCardInfo[jrnl_id].Head();
       while (pJrnlCard != NULL) {
            if (strcmp(pJrnlCard->Field(astmField+1), "") &&
               !strcmp(pJrnlCard->Field(astmField), ""))
                 pJrnlCard->AddElementToField("ASTM", astmField);
            else if (!strcmp(pJrnlCard->Field(astmField+1), ""))
                 pJrnlCard->AddElementToField("", astmField);

            if (strcmp(pJrnlCard->Field(issnField+1), "") &&
               !strcmp(pJrnlCard->Field(issnField), ""))
                 pJrnlCard->AddElementToField("ISSN", issnField);
            else if (!strcmp(pJrnlCard->Field(issnField+1), ""))
                 pJrnlCard->AddElementToField("", issnField);

            pJrnlCard = pJrnlCard->Next();
       }

       for (displayOrder = 0; displayOrder < MxJrnlTokens; displayOrder++) {
            for (card_id = 0; card_id < MxJrnlTokens; card_id++)
                 if (displayOrder == JrnlToken[card_id].DisplayOrder) break;

            if (card_id == MxJrnlTokens) continue;
            write_general_cards(card_id, TRUE);
       }
}

void _FileObj::print_molecules(FILE *fp, const int format, const int remove_hydrogen,
                const int add_connect, const int is_iupac)
{
       int irow = 0, l, num_mol_display = 0;
       ISTable *Table = NULL;

       num_mol_display = 0;
       for (l = 0; l < _num_mol; l++) {
            if (_molecules[l]->display()) num_mol_display++;
       }
       if (!num_mol_display) return;
          
       if (format == NDB_FILE_FORMAT_NDBCIF) {
            Table = new ISTable("NDB_atom_site");
            Table->AddColumn("ndb_model", FALSE);
            for (l = 0; l < NdbToken[ATOMN_TOKEN-1].NumField; l++)
                 Table->AddColumn(NdbToken[ATOMN_TOKEN-1].FieldList[l].FieldLabel, FALSE);
       }

       connect = NULL;
       Alloc_connect = 0;
       Num_connect = 0;

       for (l = 0; l < _num_mol; l++) {
            if (!_molecules[l]->display()) continue;
            if (format == NDB_FILE_FORMAT_PDB) _molecules[l]->reorder_chain_order();

            if (num_mol_display > 1 && format != NDB_FILE_FORMAT_NDBCIF) {
                  if (format == NDB_FILE_FORMAT_PDB) {
                       fprintf(fp, "MODEL     %4d                          ",
                                   _molecules[l]->Mol_ID());
                       fprintf(fp, "                                        \n");
                  } else fprintf(fp, "MODEL     %4d\n", _molecules[l]->Mol_ID());
            }

            if (add_connect && format == NDB_FILE_FORMAT_PDB && !l)
                 print_molecule(_molecules[l], Table, irow, fp, format,
                       remove_hydrogen, 1, is_iupac);
            else print_molecule(_molecules[l], Table, irow, fp, format,
                       remove_hydrogen, 0, is_iupac);

            if (num_mol_display > 1 && format != NDB_FILE_FORMAT_NDBCIF) {
                 if (format == NDB_FILE_FORMAT_PDB) {
                      fprintf(fp, "ENDMDL                                  ");
                      fprintf(fp, "                                        \n");
                 } else fprintf(fp, "ENDMDL\n");
            }
       }

       if (format == NDB_FILE_FORMAT_PDB && add_connect) get_connect_cards();

       if (format == NDB_FILE_FORMAT_NDBCIF) {
            _CifObj->WriteTable(Table, _StructureId, "NDB_atom_site");
       }
}

void _FileObj::print_molecule(_MOLECULE *mol, ISTable *Table, int &irow, FILE *fp,
          const int format, const int remove_hydrogen, const int add_connect,
          const int is_iupac)
{
       int atomSerialNo, WaterSerialNo, card_id;
       char FieldInfo[MxField][MxLen], buffer[10], *p = NULL;
       _ATOM *atom = NULL;
       _RESIDUE *residue = NULL;
       _CHAIN *chain = NULL;
       ConnectFormat *drug = NULL;

       atomSerialNo = 0;
       WaterSerialNo = 0;
       chain = mol->GetFirstChain();
       while (chain) {
            if (!chain->display()) {
                 chain = mol->GetNextChain();
                 continue;
            }
            card_id = chain->chain_type();
            residue = chain->GetFirstResidue();
            while (residue) {
                 atom = residue->GetFirstAtom();
                 while (atom) {
                      if (atom->is_hydrogen() && remove_hydrogen) continue;
                      print_atom(Table, fp, card_id-1, atomSerialNo,
                                 WaterSerialNo, format, atom, FieldInfo,
                                 mol->Mol_ID(), irow, is_iupac);
                      atom = residue->GetNextAtom();
                 }
                 if (add_connect) {
                      p = residue->name_list().Text();
                      while (p) {
                           p = get_word(p, buffer);
                           if (!strcmp(buffer, "") || is_a_standard_residue(buffer) ||
                               !strcmp(buffer, "HOH") || !strcmp(buffer, "DOD")) continue;
                           drug = find_drug(buffer);
                           if (drug) add_connect_record(drug, residue);
                      }
                 }
                 residue = chain->GetNextResidue();
            }

            if ((card_id == ATOMN_TOKEN || card_id == ATOMP_TOKEN) &&
                 format != NDB_FILE_FORMAT_NDBCIF) {
                 print_ter_card(fp, atomSerialNo, format, FieldInfo);
            } else if (card_id == ATOMS_TOKEN-1 && format == NDB_FILE_FORMAT_NDB)
                 print_ter_card(fp, atomSerialNo, format, FieldInfo);

            chain = mol->GetNextChain();
       }
}

void _FileObj::get_connect_cards()
{
       int i, j, CONECT_TOKEN, field, num1 = 0, num2 = 0;
       _ATOM **atom1 = NULL, **atom2 = NULL;
       double dist = 10.0;
       _MOLECULE *molecule = NULL;
       One_Card_Info_Format *pCard = NULL;

       for (i = 0; i < _num_mol; i++) {
            if (_molecules[i]->display()) {
                 molecule = _molecules[i];
                 break;
            }
       }
       if (molecule) {
            pCard = _CardInfo[LINK_TOKEN-1].Head();
            while (pCard) {
                 if ((!strcmp(pCard->Field(13), "") || !strcmp(pCard->Field(13), "1_555")) &&
                     (!strcmp(pCard->Field(14), "") || !strcmp(pCard->Field(14), "1_555"))) {
                      atom1 = molecule->find_pdb_atom(pCard->Field(1),
                            pCard->Field(4), pCard->Field(3), atoi(pCard->Field(5)),
                            pCard->Field(6), num1);
                      atom2 = molecule->find_pdb_atom(pCard->Field(7),
                            pCard->Field(10), pCard->Field(9), atoi(pCard->Field(11)),
                            pCard->Field(12), num2);
                      if (atom1 && atom2) {
                           for (i = 0; i < num1; i++) {
                                if (!strcmp(atom1[i]->alt_loc(), pCard->Field(2))) break;
                           }
                           for (j = 0; j < num2; j++) {
                                if (!strcmp(atom2[j]->alt_loc(), pCard->Field(8))) break;
                           }
                           if (i < num1 && j < num2) {  
                                insert_connect_atom_id(atom1[i]->atnum(), atom2[j]->atnum());
                                insert_connect_atom_id(atom2[j]->atnum(), atom1[i]->atnum());
                           }
                      }
                      if (atom1) delete [] atom1;
                      if (atom2) delete [] atom2;
                 }
                 pCard = pCard->Next();
            }
     
            pCard = _CardInfo[SSBOND_TOKEN-1].Head();
            while (pCard) {
                 atom1 = molecule->find_pdb_atom("SG",
                       pCard->Field(3), pCard->Field(2), atoi(pCard->Field(4)),
                       pCard->Field(5), num1);
                 atom2 = molecule->find_pdb_atom("SG", 
                       pCard->Field(7), pCard->Field(6), atoi(pCard->Field(8)),
                       pCard->Field(9), num2);
                 if (atom1 && atom2) {
                      for (i = 0; i < num1; i++) {
                           for (j = 0; j < num2; j++) {
                                if (!strcmp(atom1[i]->alt_loc(), atom2[j]->alt_loc()) ||
                                    !strcmp(atom1[i]->alt_loc(), "") ||
                                    !strcmp(atom2[j]->alt_loc(), "")) {
                                     dist = cal_distance(atom1[i], atom2[j]);
                                     if (dist < 3.0) {
                                          insert_connect_atom_id(atom1[i]->atnum(), atom2[j]->atnum());
                                          insert_connect_atom_id(atom2[j]->atnum(), atom1[i]->atnum());
                                     }
                                }
                           }
                      }
                 }
                 if (atom1) delete [] atom1;
                 if (atom2) delete [] atom2;
                 pCard = pCard->Next();
            }
       }

       CONECT_TOKEN = find_card_id("CONECT");
       pCard = _CardInfo[CONECT_TOKEN].Head();
       while (pCard) pCard = _CardInfo[CONECT_TOKEN].delete_a_card(pCard);
       for (i = 0; i < Num_connect; i++) {
            if (connect[i].con_num == 0) continue;

            field = 0;
            pCard = _CardInfo[CONECT_TOKEN].add_a_new_card(NdbToken[CONECT_TOKEN], 0);
            pCard->AddElementToField(connect[i].atom_id, 1);
            for (j = 0; j < connect[i].con_num; j++) {
                 if (field == 4) {
                      pCard = _CardInfo[CONECT_TOKEN].
                                 add_a_new_card(NdbToken[CONECT_TOKEN], 0);
                      pCard->AddElementToField(connect[i].atom_id, 1);
                      field = 0;
                 }
                 pCard->AddElementToField(connect[i].con_to[j], 2 + field);
                 field++;
            }
       }
       if (connect) free ((void *) connect);
       Num_connect = 0;
}

void _FileObj::print_atom(ISTable *Table, FILE *fp, const int card_id, int &atomSerialNo,
       int &WaterSerialNo, const int format, _ATOM *atom, char FieldInfo[MxField][MxLen],
       const int n_mol, int &irow, const int is_iupac)
{
       int temp_id, i;
       char value[10];
       CifString string;

       if (format == NDB_FILE_FORMAT_NDBCIF) {
            atomSerialNo++;
            atom->set_atnum(atomSerialNo);
       }
       strcpy(FieldInfo[0], NdbToken[card_id].TokenName);
       sprintf(FieldInfo[1], "%d", atom->atnum());
       strcpy(FieldInfo[2], atom->atmtype());
       strcpy(FieldInfo[3], atom->alt_loc());
       strcpy(FieldInfo[4], atom->restype());
       strcpy(FieldInfo[5], atom->chnid());
       sprintf(FieldInfo[6], "%d", atom->resnum());
       strcpy(FieldInfo[7], atom->ins_code());
       sprintf(FieldInfo[8], "%.4f", atom->orig().x);
       sprintf(FieldInfo[9], "%.4f", atom->orig().y);
       sprintf(FieldInfo[10], "%.4f", atom->orig().z);
       sprintf(FieldInfo[11], "%6.2f", atom->occ());
       sprintf(FieldInfo[12], "%6.2f", atom->t_fct());
       if (atom->footnote() != -10000)
            sprintf(FieldInfo[13], "%d", atom->footnote());
       else strcpy(FieldInfo[13], "");
       strcpy(FieldInfo[14], atom->not_in_asym());
       sprintf(FieldInfo[15], "%d", atom->pdb_resnum());
       strcpy(FieldInfo[16], atom->pdb_resnam());
       strcpy(FieldInfo[17], atom->pdb_chnid());
       strcpy(FieldInfo[18], atom->pdb_atmnam());
       sprintf(FieldInfo[19], "%d", atom->pub_resnum());
       strcpy(FieldInfo[20], atom->pub_resnam());
       strcpy(FieldInfo[21], atom->pub_chnid());
       strcpy(FieldInfo[22], atom->pub_atmnam());
       strcpy(FieldInfo[23], atom->atom_type());

       if (format == NDB_FILE_FORMAT_NDBCIF) {
            sprintf(value, "%d", n_mol);
            string.Copy(value);
            Table->AddRow();
            Table->UpdateCell(string, 0, irow);
            for (i = 0; i < 23; i++) {
                 string.Copy(FieldInfo[i]);
                 Table->UpdateCell(string, i+1, irow);
            }
            irow++;
       } else {
            print_atom_card(fp, card_id, card_id, atomSerialNo,
                  WaterSerialNo, FieldInfo, format, atom->align(), is_iupac);
            atom->set_atnum(atomSerialNo);
       }

       if (atom->is_sigatm()) {
            strcpy(FieldInfo[0], "SIGATM");
            temp_id = find_card_id(FieldInfo[0]);

            sprintf(FieldInfo[8],  "%8.3f", atom->sigatm().sig_x);
            sprintf(FieldInfo[9],  "%8.3f", atom->sigatm().sig_y);
            sprintf(FieldInfo[10], "%8.3f", atom->sigatm().sig_z);
            sprintf(FieldInfo[11], "%6.2f", atom->sigatm().sig_o);
            sprintf(FieldInfo[12], "%6.2f", atom->sigatm().sig_t);

            if (format == NDB_FILE_FORMAT_NDBCIF) {
                 string.Copy(value);
                 Table->AddRow();
                 Table->UpdateCell(string, 0, irow);
                 for (i = 0; i < 23; i++) {
                      string.Copy(FieldInfo[i]);
                      Table->UpdateCell(string, i+1, irow);
                 }
                 irow++;
            } else print_atom_card(fp, card_id, temp_id, atomSerialNo, 
                WaterSerialNo, FieldInfo, format, atom->align(), is_iupac);
       } 

       if (atom->is_anisou()) {
            strcpy(FieldInfo[0], "ANISOU");
            temp_id = find_card_id(FieldInfo[0]);

            sprintf(FieldInfo[8],  "%d", atom->anisou().u11);
            sprintf(FieldInfo[9],  "%d", atom->anisou().u22);
            sprintf(FieldInfo[10], "%d", atom->anisou().u33);
            sprintf(FieldInfo[11], "%d", atom->anisou().u12);
            sprintf(FieldInfo[12], "%d", atom->anisou().u13);
            sprintf(FieldInfo[13], "%d", atom->anisou().u23);

            if (format == NDB_FILE_FORMAT_NDBCIF) {
                 string.Copy(value);
                 Table->AddRow();
                 Table->UpdateCell(string, 0, irow);
                 for (i = 0; i < 23; i++) {
                      string.Copy(FieldInfo[i]);
                      Table->UpdateCell(string, i+1, irow);
                 }
                 irow++;
            } else print_atom_card(fp, card_id, temp_id, atomSerialNo,
                WaterSerialNo, FieldInfo, format, atom->align(), is_iupac);
       }

       if (atom->is_siguij()) {
            strcpy(FieldInfo[0], "SIGUIJ");
            temp_id = find_card_id(FieldInfo[0]);

            sprintf(FieldInfo[8],  "%d", atom->siguij().sig_u11);
            sprintf(FieldInfo[9],  "%d", atom->siguij().sig_u22);
            sprintf(FieldInfo[10], "%d", atom->siguij().sig_u33);
            sprintf(FieldInfo[11], "%d", atom->siguij().sig_u12);
            sprintf(FieldInfo[12], "%d", atom->siguij().sig_u13);
            sprintf(FieldInfo[13], "%d", atom->siguij().sig_u23);

            if (format == NDB_FILE_FORMAT_NDBCIF) {
                 string.Copy(value);
                 Table->AddRow();
                 Table->UpdateCell(string, 0, irow);
                 for (i = 0; i < 23; i++) {
                      string.Copy(FieldInfo[i]);
                      Table->UpdateCell(string, i+1, irow);
                 }
                 irow++;
            } else print_atom_card(fp, card_id, temp_id, atomSerialNo,
                WaterSerialNo, FieldInfo, format, atom->align(), is_iupac);
       }
}

void _FileObj::put_coordinates_into_molecules(const int special)
{
       static const char *keywords[] = { "ndb_model", "group_PDB", "id", "type_symbol",
             "label_atom_id", "label_alt_id", "label_comp_id", "label_asym_id",
             "label_entity_id", "label_seq_id", "ndb_label_seq_num", "ndb_ins_code",
             "Cartn_x", "Cartn_y", "Cartn_z", "occupancy", "B_iso_or_equiv",
             "Cartn_x_esd", "Cartn_y_esd", "Cartn_z_esd", "occupancy_esd",
             "B_iso_or_equiv_esd", "footnote_id", "ndb_not_in_asym" /* , "ndb_PDB_residue_no",
             "ndb_PDB_residue_name", "ndb_PDB_strand_id", "ndb_PDB_atom_name",
             "auth_seq_id" */ };
       int n_keywords = 24, colNo, rowNo, i, j, format, prev_model, prevResidueNo;
       int model, is_sigatm, irow, id, is_anisou, is_siguij, total_atoms;
       int *models = NULL, exist_PDB = 0, exist_auth = 0, exist_ndb_auth = 0;
       int ndb_PDB_residue_no, ndb_PDB_residue_name, ndb_PDB_strand_id;
       int ndb_PDB_atom_name, ndb_auth_seq_id, ndb_auth_comp_id, ndb_auth_asym_id;
       int ndb_auth_atom_name, auth_seq_id, auth_comp_id, auth_asym_id;
       int auth_atom_id, no_copy_pdb_chainid = 0, str_len = 0;
       int exist_entity_poly = 0, exist_scheme = 0, is_exchange_cif = 0;
       int is_public_cif = 0, index;
       char prevChainId[MxStrLen], prevResidueName[MxStrLen];
       char prevInsCode[MxStrLen], prevPdbResName[MxStrLen], value[MxStrLen];
       CifString string;
       COORD coord;
       SIGATM sigatm;
       ANISOU anisou;
       SIGUIJ siguij;
       _ATOM **atoms = NULL, *atom = NULL;
       _RESIDUE *residue = NULL;
       ISTable *Table = NULL;
       ReVarCifArray<CifString> *ColumnNames = NULL;
       _IIndex idx;
       BTreeObj<_IIndex> _AtomNumbers;

       sigatm.sig_x = sigatm.sig_y = sigatm.sig_z = sigatm.sig_o = sigatm.sig_t = 0;

       Table = _CifObj->GetTablePtr(_StructureId, "ndb_poly_seq_scheme");
       if (!is_empty_table(Table)) exist_scheme = 1;

       Table = _CifObj->GetTablePtr(_StructureId, "database_PDB_rev");
       if (!is_empty_table(Table)) {
            index = Table->GetColumnIndex("date");
            string.Clear();
            Table->GetCell(string, index, 0);
            if (string != "" && string != "." && string != "?")
                 is_public_cif = 1;
       }

       Table = _CifObj->GetTablePtr(_StructureId, "entity_poly");
       if (!is_empty_table(Table)) exist_entity_poly = 1;
       Table = _CifObj->GetTablePtr(_StructureId, "atom_site");
       if (is_empty_table(Table)) return;

       colNo = Table->GetNumColumns();
       rowNo = Table->GetNumRows();

       ndb_PDB_residue_no = Table->GetColumnIndex("ndb_auth_seq_id");
       ndb_PDB_residue_name = Table->GetColumnIndex("ndb_auth_comp_id"); 
       ndb_PDB_strand_id = Table->GetColumnIndex("ndb_auth_asym_id");
       ndb_PDB_atom_name = Table->GetColumnIndex("ndb_auth_atom_name");
       if (exist_entity_poly && !exist_scheme && ndb_PDB_residue_no < 0 &&
           ndb_PDB_residue_name < 0 && ndb_PDB_strand_id < 0 && ndb_PDB_atom_name < 0) {
            is_exchange_cif = 1;
       }
       if (is_exchange_cif && !is_public_cif) _file_status = FILE_STATUS_MMCIF_PLUS_MOL;

       ndb_PDB_residue_no = Table->GetColumnIndex("ndb_PDB_residue_no");
       ndb_PDB_residue_name = Table->GetColumnIndex("ndb_PDB_residue_name");
       ndb_PDB_strand_id = Table->GetColumnIndex("ndb_PDB_strand_id");
       ndb_PDB_atom_name = Table->GetColumnIndex("ndb_PDB_atom_name");
       if (ndb_PDB_residue_no >= 0 && ndb_PDB_residue_name >= 0 &&
           ndb_PDB_strand_id >= 0 && ndb_PDB_atom_name >= 0) exist_PDB = 1;

       int label_atom_id = Table->GetColumnIndex("label_atom_id");
       int label_comp_id = Table->GetColumnIndex("label_comp_id");
       int label_asym_id = Table->GetColumnIndex("label_asym_id");
       int label_seq_id = Table->GetColumnIndex("label_seq_id");

       ndb_auth_seq_id = Table->GetColumnIndex("ndb_auth_seq_id");
       ndb_auth_comp_id = Table->GetColumnIndex("ndb_auth_comp_id");
       ndb_auth_asym_id = Table->GetColumnIndex("ndb_auth_asym_id");
       ndb_auth_atom_name = Table->GetColumnIndex("ndb_auth_atom_name");

       auth_seq_id = Table->GetColumnIndex("auth_seq_id");
       auth_comp_id = Table->GetColumnIndex("auth_comp_id");
       auth_asym_id = Table->GetColumnIndex("auth_asym_id");
       auth_atom_id = Table->GetColumnIndex("auth_atom_id");

       if (label_asym_id >= 0 && auth_asym_id >= 0) no_copy_pdb_chainid = 1;

       if (exist_PDB) {
            if (auth_seq_id >= 0 && (ndb_auth_comp_id >= 0 || auth_comp_id >= 0) &&
               (ndb_auth_asym_id >= 0 || auth_asym_id >= 0) &&
                ndb_auth_atom_name >= 0) exist_ndb_auth = 1;
       } else {
            if (auth_seq_id >= 0 && auth_comp_id >= 0 && auth_asym_id >= 0 &&
                auth_atom_id >= 0) exist_auth = 1;
            if (ndb_auth_seq_id >= 0 && ndb_auth_comp_id >= 0 &&
                ndb_auth_asym_id >= 0 && ndb_auth_atom_name >= 0) exist_ndb_auth = 1;
       }

       ColumnNames = Table->GetColumnNames();
       format = NDB_FILE_FORMAT_PDB;
       if (colNo == (n_keywords + 8)) {
            for (i = 0; i < n_keywords; i++) {
                 for (j = 0; j < colNo; j++) {
                      if (!strcasecmp(keywords[i], (*ColumnNames)[j].Text())) break;
                 }
                 if (j == colNo) break;
            }
            if (i == n_keywords && (exist_PDB || exist_auth) && exist_ndb_auth)
                 format = NDB_FILE_FORMAT_NDB;
       }

       total_atoms = 0;
       atoms = new _ATOM*[rowNo];
       models = new int[rowNo];
       for (irow = 0; irow < rowNo; irow++) models[irow] = 1;
       _AtomNumbers.Clear();
       for (irow = 0; irow < rowNo; irow++) {
            atom = new _ATOM;
            atom->set_type("ATOM");
            coord.x = coord.y = coord.z = 0.0;
            is_sigatm = 0;
            for (i = 0; i < colNo; i++) {
                 Table->GetCell(string, i, irow);
                 if (string == "" || string == "?") continue;
                 if (string == "." && strcasecmp((*ColumnNames)[i].Text(), "type_symbol"))
                      continue;
     
                 strcpy(value, string.Text());
                 ndb_clean_string(value);
                 if (!strcmp((*ColumnNames)[i].Text(), "ndb_model"))
                      models[total_atoms] = atoi(value);
                 else if (!strcasecmp((*ColumnNames)[i].Text(), "group_PDB"))
                      atom->set_type(value);
                 else if (!strcasecmp((*ColumnNames)[i].Text(), "id")) {
                      atom->set_atnum(atoi(value));
                 } else if (!strcasecmp((*ColumnNames)[i].Text(), "type_symbol"))
                      atom->set_atom_type(value);
                 else if (!strcasecmp((*ColumnNames)[i].Text(), "label_atom_id"))
                      atom->set_atmtype(value);
                 else if (!strcasecmp((*ColumnNames)[i].Text(), "label_alt_id"))
                      atom->set_alt_loc(value);
                 else if (!strcasecmp((*ColumnNames)[i].Text(), "label_comp_id"))
                      atom->set_restype(value);
                 else if (!strcasecmp((*ColumnNames)[i].Text(), "label_asym_id"))
                      atom->set_chnid(value);
                 else if (!strcasecmp((*ColumnNames)[i].Text(), "label_seq_id") &&
                          format == NDB_FILE_FORMAT_PDB)
                      atom->set_resnum(atoi(value));
                 else if (!strcasecmp((*ColumnNames)[i].Text(), "ndb_label_seq_num"))
                      atom->set_resnum(atoi(value));
                 else if (!strcasecmp((*ColumnNames)[i].Text(), "ndb_ins_code"))
                      atom->set_ins_code(value);
                 else if (!strcasecmp((*ColumnNames)[i].Text(), "Cartn_x"))
                      coord.x = atof(value);
                 else if (!strcasecmp((*ColumnNames)[i].Text(), "Cartn_y"))
                      coord.y = atof(value);
                 else if (!strcasecmp((*ColumnNames)[i].Text(), "Cartn_z"))
                      coord.z = atof(value);
                 else if (!strcasecmp((*ColumnNames)[i].Text(), "occupancy"))
                      atom->set_occ(atof(value));
                 else if (!strcasecmp((*ColumnNames)[i].Text(), "B_iso_or_equiv"))
                      atom->set_t_fct(atof(value));
                 else if (!strcasecmp((*ColumnNames)[i].Text(), "Cartn_x_esd")) {
                      is_sigatm = 1;
                      sigatm.sig_x = atof(value);
                 } else if (!strcasecmp((*ColumnNames)[i].Text(), "Cartn_y_esd")) {
                      is_sigatm = 1;
                      sigatm.sig_y = atof(value);
                 } else if (!strcasecmp((*ColumnNames)[i].Text(), "Cartn_z_esd")) { 
                      is_sigatm = 1;
                      sigatm.sig_z = atof(value);
                 } else if (!strcasecmp((*ColumnNames)[i].Text(), "occupancy_esd")) {
                      is_sigatm = 1;
                      sigatm.sig_o = atof(value);
                 } else if (!strcasecmp((*ColumnNames)[i].Text(), "B_iso_or_equiv_esd")) {
                      is_sigatm = 1;
                      sigatm.sig_t = atof(value);
                 } else if (!strcasecmp((*ColumnNames)[i].Text(), "footnote_id"))
                      atom->set_footnote(atoi(value));
                 else if (!strcasecmp((*ColumnNames)[i].Text(), "ndb_not_in_asym"))
                      atom->set_not_in_asym(value);
                 else if (!strcasecmp((*ColumnNames)[i].Text(), "ndb_PDB_residue_no") && exist_PDB)
                      atom->set_pdb_resnum(atoi(value));
                 else if (!strcasecmp((*ColumnNames)[i].Text(), "ndb_PDB_residue_name") && exist_PDB)
                      atom->set_pdb_resnam(value);
                 else if (!strcasecmp((*ColumnNames)[i].Text(), "ndb_PDB_strand_id") && exist_PDB)
                      atom->set_pdb_chnid(value);
                 else if (!strcasecmp((*ColumnNames)[i].Text(), "ndb_PDB_atom_name") && exist_PDB)
                      atom->set_pdb_atmnam(value);
                 else if (!strcasecmp((*ColumnNames)[i].Text(), "auth_seq_id")) {
                      if (exist_auth)
                           atom->set_pdb_resnum(atoi(value));
                      else if (exist_ndb_auth)
                           atom->set_pub_resnum(atoi(value));
                      else if (!exist_auth && !exist_ndb_auth)
                           atom->set_pdb_resnum(atoi(value));
                      str_len = strlen(value);
                      for (j = 0; j < str_len; j++) {
                           if (isalpha(value[j])) break;
                      }
                      if (j < str_len) atom->set_ins_code(&value[j]);
                 } else if (!strcasecmp((*ColumnNames)[i].Text(), "ndb_auth_seq_id"))
                      atom->set_pub_resnum(atoi(value));
                 else if (!strcasecmp((*ColumnNames)[i].Text(), "auth_comp_id")) {
                      if (exist_auth)
                           atom->set_pdb_resnam(value);
                      else if (exist_ndb_auth)
                           atom->set_pub_resnam(value);
                      else if (!exist_auth && !exist_ndb_auth)
                           atom->set_pdb_resnam(value);
                 } else if (!strcasecmp((*ColumnNames)[i].Text(), "ndb_auth_comp_id"))
                      atom->set_pub_resnam(value);
                 else if (!strcasecmp((*ColumnNames)[i].Text(), "auth_asym_id")) {
                      if (exist_auth)
                           atom->set_pdb_chnid(value);
                      else if (exist_ndb_auth)
                           atom->set_pub_chnid(value);
                      else if (!exist_auth && !exist_ndb_auth)
                           atom->set_pdb_chnid(value);
                 } else if (!strcasecmp((*ColumnNames)[i].Text(), "ndb_auth_asym_id"))
                      atom->set_pub_chnid(value);
                 else if (!strcasecmp((*ColumnNames)[i].Text(), "auth_atom_id"))
                      atom->set_pdb_atmnam(value);
                 else if (!strcasecmp((*ColumnNames)[i].Text(), "ndb_auth_atom_name"))
                      atom->set_pub_atmnam(value);
            }

            if (special && !strcmp(atom->atom_type(), "Q")) {
                 delete atom;
                 continue;
            }

            if (auth_seq_id < 0 && label_seq_id >= 0) {
                 Table->GetCell(string, label_seq_id, irow);
                 if (string != "" && string != "?" && string != ".")
                      atom->set_pdb_resnum(atoi(string.Text()));
            }

            if (auth_comp_id < 0 && label_comp_id >= 0) {
                 Table->GetCell(string, label_comp_id, irow);
                 if (string != "" && string != "?" && string != ".")
                      atom->set_pdb_resnam(string.Text());
            }

            if (auth_asym_id < 0 && label_asym_id >= 0) {
                 Table->GetCell(string, label_asym_id, irow);
                 if (string != "" && string != "?" && string != ".")
                      atom->set_pdb_chnid(string.Text());
            }

            if (auth_atom_id < 0 && label_atom_id >= 0) {
                 Table->GetCell(string, label_atom_id, irow);
                 if (string != "" && string != "?" && string != ".")
                      atom->set_pdb_atmnam(string.Text());
            }

            atom->set_orig(coord);

            if (atom->resnum() == (-99999)) {
                 if (atom->pdb_resnum() == (-99999)) {
                      if (atom->pub_resnum() != (-99999))
                           atom->set_resnum(atom->pub_resnum());
                 } else atom->set_resnum(atom->pdb_resnum());
            }

            if (is_sigatm) {
                 atom->set_is_sigatm(is_sigatm);
                 atom->set_sigatm(sigatm);
            }

            if (!strcmp(atom->restype(), "")) {
                 if (strcmp(atom->pdb_resnam(), ""))
                      atom->set_restype(atom->pdb_resnam());
                 else if (strcmp(atom->pub_resnam(), ""))
                      atom->set_restype(atom->pub_resnam());
            }

            atoms[total_atoms] = atom;
            idx.set_value(atom->atnum());
            idx.set_pos(total_atoms);
            _AtomNumbers.Add(idx);
            total_atoms++;
       }
       if (ColumnNames) delete ColumnNames;

       Table = _CifObj->GetTablePtr(_StructureId, "atom_site_anisotrop");
       if (!is_empty_table(Table)) {
            colNo = Table->GetNumColumns();
            rowNo = Table->GetNumRows();
            ColumnNames = Table->GetColumnNames();
            for (irow = 0; irow < rowNo; irow++) {
                 id = 0;
                 is_anisou = 0;
                 is_siguij = 0;
                 anisou.u11 = anisou.u12 = anisou.u13 =
                 anisou.u22 = anisou.u23 = anisou.u33 = 0;
                 siguij.sig_u11 = siguij.sig_u12 = siguij.sig_u13 =
                 siguij.sig_u22 = siguij.sig_u23 = siguij.sig_u33 = 0;
                 for (i = 0; i < colNo; i++) {
                      Table->GetCell(string, i, irow);
                      if (string == "" || string == "?" || string == ".") continue;
     
                      strcpy(value, string.Text());
                      ndb_clean_string(value);
                      if (!strcasecmp((*ColumnNames)[i].Text(), "id"))
                           id = atoi(value);
                      else if (!strcasecmp((*ColumnNames)[i].Text(), "U[1][1]")) {
                           is_anisou = 1;
                           anisou.u11 = (int) (rint(10000.0 * atof(value)));
                      } else if (!strcasecmp((*ColumnNames)[i].Text(), "U[1][2]")) {
                           is_anisou = 1;
                           anisou.u12 = (int) (rint(10000.0 * atof(value)));
                      } else if (!strcasecmp((*ColumnNames)[i].Text(), "U[1][3]")) {
                           is_anisou = 1;
                           anisou.u13 = (int) (rint(10000.0 * atof(value)));
                      } else if (!strcasecmp((*ColumnNames)[i].Text(), "U[2][2]")) {
                           is_anisou = 1;
                           anisou.u22 = (int) (rint(10000.0 * atof(value)));
                      } else if (!strcasecmp((*ColumnNames)[i].Text(), "U[2][3]")) {
                           is_anisou = 1;
                           anisou.u23 = (int) (rint(10000.0 * atof(value)));
                      } else if (!strcasecmp((*ColumnNames)[i].Text(), "U[3][3]")) {
                           is_anisou = 1;
                           anisou.u33 = (int) (rint(10000.0 * atof(value)));
                      } else if (!strcasecmp((*ColumnNames)[i].Text(), "U[1][1]_esd")) {
                           is_siguij = 1;
                           siguij.sig_u11 = (int) (rint(10000.0 * atof(value)));
                      } else if (!strcasecmp((*ColumnNames)[i].Text(), "U[1][2]_esd")) {
                           is_siguij = 1;
                           siguij.sig_u12 = (int) (rint(10000.0 * atof(value)));
                      } else if (!strcasecmp((*ColumnNames)[i].Text(), "U[1][3]_esd")) {
                           is_siguij = 1;
                           siguij.sig_u13 = (int) (rint(10000.0 * atof(value)));
                      } else if (!strcasecmp((*ColumnNames)[i].Text(), "U[2][2]_esd")) {
                           is_siguij = 1;
                           siguij.sig_u22 = (int) (rint(10000.0 * atof(value)));
                      } else if (!strcasecmp((*ColumnNames)[i].Text(), "U[2][3]_esd")) {
                           is_siguij = 1;
                           siguij.sig_u23 = (int) (rint(10000.0 * atof(value)));
                      } else if (!strcasecmp((*ColumnNames)[i].Text(), "U[3][3]_esd")) {
                           is_siguij = 1;
                           siguij.sig_u33 = (int) (rint(10000.0 * atof(value)));
                      }
                 }
                 idx.set_value(id);
                 i = _AtomNumbers.Seek(idx);
                 if (i >= 0) {
                      idx = _AtomNumbers.GetCurrData();
                      i = idx.pos();
                      if (is_anisou) {
                           atoms[i]->set_is_anisou(is_anisou);
                           atoms[i]->set_anisou(anisou);
                      }
                      if (is_siguij) {
                           atoms[i]->set_is_siguij(is_siguij);
                           atoms[i]->set_siguij(siguij);
                      }
                 }
            }
            if (ColumnNames) delete ColumnNames;
       }

       if (_molecules) {
            for (i = 0; i < _num_mol; i++)
                 if (_molecules[i]) delete _molecules[i];
            delete [] _molecules;
       }
       _num_mol = 0;
       _alloc_mol = 0;
       _molecules = NULL;

       prev_model = -1;
       model = 1;
       prevResidueNo = -10000;
       strcpy(prevChainId, "XXX");
       strcpy(prevResidueName, "XXX");
       strcpy(prevInsCode, "XXX");
       strcpy(prevPdbResName, "XXX");
/*
       if (_AtomNumbers.Root() >= 0) {
            j = 1;
            _AtomNumbers.GoFirst();
            while (j >= 0) {
                 idx = _AtomNumbers.GetCurrData();
                 i = idx.pos();
                 j = _AtomNumbers.GoToNext();
*/
            for (i = 0; i < total_atoms; i++) {
                 model = models[i];
                 strcpy(value, atoms[i]->chnid());
                 if (is_exchange_cif) strcpy(value, atoms[i]->pdb_chnid());
                 // if (strcmp(prevChainId, atoms[i]->chnid()) ||
                 if (strcmp(prevChainId, value) ||
                     strcmp(prevResidueName, atoms[i]->restype()) ||
                     strcmp(prevPdbResName, atoms[i]->pdb_resnam()) ||
                     strcmp(prevInsCode, atoms[i]->ins_code()) ||
                     prevResidueNo != atoms[i]->resnum()) {
                      if (residue) {
                           if (format == NDB_FILE_FORMAT_PDB) {
                                if (is_public_cif)
                                      residue->change_nomenclature(NDB_FILE_FORMAT_NDB, 1, exist_auth, no_copy_pdb_chainid);
				else  residue->change_nomenclature(format, 1, exist_auth, no_copy_pdb_chainid);
                                if (!is_a_standard_residue(residue->ResName()))
                                     residue->correction_atom_name(_error_message_s);
                           } else if (special && !is_a_standard_residue(residue->ResName()))
                                residue->correction_atom_name(_error_message_s);
                           _molecules[_num_mol-1]->insert_a_residue(0, residue,
                                  _error_message, _unrecognized_residue);
                           residue = NULL;
                      }
                      residue = new _RESIDUE;
                      residue->set_ResName(atoms[i]->restype());
                      residue->set_res_no(atoms[i]->resnum());
                      if (atoms[i]->pdb_resnum() != -99999)
                           residue->set_pdb_res_no(atoms[i]->pdb_resnum());
                      else residue->set_pdb_res_no(residue->res_no());
                 }

                 if (model != prev_model) {
                      prev_model = model;
                      _ManageSpace();
                      _molecules[_num_mol-1]->set_Mol_ID(model);
                 }

                 if (strcmp(atoms[i]->atmtype(), "")) {
                      if (is_exchange_cif) atoms[i]->set_chnid(atoms[i]->pdb_chnid());
                      residue->insert_a_atom(atoms[i], _error_message);
                 }

                 strcpy(prevChainId, atoms[i]->chnid());
                 strcpy(prevResidueName, atoms[i]->restype());
                 strcpy(prevPdbResName, atoms[i]->pdb_resnam());
                 strcpy(prevInsCode, atoms[i]->ins_code());
                 prevResidueNo = atoms[i]->resnum();
            }
/*
       }
*/

       if (residue) {
            if (format == NDB_FILE_FORMAT_PDB) {
                 if (is_public_cif)
                      residue->change_nomenclature(NDB_FILE_FORMAT_NDB, 1, exist_auth, no_copy_pdb_chainid);
		 else residue->change_nomenclature(format, 1, exist_auth, no_copy_pdb_chainid);
                 if (!is_a_standard_residue(residue->ResName()))
                      residue->correction_atom_name(_error_message_s);
            } else if (special && !is_a_standard_residue(residue->ResName()))
                 residue->correction_atom_name(_error_message_s);
            _molecules[_num_mol-1]->insert_a_residue(0, residue,
                       _error_message, _unrecognized_residue);
            residue = NULL;
       }

       for (i = 0; i < _num_mol; i++) {
            _molecules[i]->check_chain_type(_Num_Seqs, _Seqs);
       }
       delete [] atoms;
       delete [] models;
}

void _FileObj::cifndb_process_category_correspond()
{
       int i, j, k, id_index, mon_id, pdb_id, row, rowNo, num = 0, index[3];
       int auth_asym_id, auth_comp_id, auth_seq_id, idx[2][3], label_asym_id;
       int type, ret;
       CifString chainId, new_chainId, cifstring;
       _CHAINID_CORR *chain_corr = NULL;
       ISTable *Table = NULL, *AtomTable = NULL;
       TblIndexObj AtomTableIndex;

       Table = _CifObj->GetTablePtr(_StructureId, "struct_asym");
       if (is_empty_table(Table)) return;
       Table = _CifObj->GetTablePtr(_StructureId, "struct_biol_gen");
       if (is_empty_table(Table)) return;
       Table = _CifObj->GetTablePtr(_StructureId, "atom_site");
       if (is_empty_table(Table)) return;

       Table = _CifObj->GetTablePtr(_StructureId, "struct_asym");
       rowNo = Table->GetNumRows();
       id_index = Table->GetColumnIndex("id");
       pdb_id   = Table->GetColumnIndex("ndb_pdb_id");
       type = Table->GetColumnIndex("ndb_type");

       num = 0;
       chain_corr = new _CHAINID_CORR[rowNo];
       for (i = 0; i < rowNo; i++) {
            Table->GetCell(chainId, id_index, i);
            Table->GetCell(new_chainId, pdb_id, i);
            Table->GetCell(cifstring, type, i);
            if (cifstring != "ATOMN" && cifstring != "ATOMP" &&
                cifstring != "NA" && cifstring != "PR") continue;

            if (chainId == "?" || chainId == ".") chainId.Copy("");
            if (new_chainId == "?" || new_chainId == ".") new_chainId.Copy("");
            strcpy(chain_corr[num].new_chainid, chainId.Text());
            strcpy(chain_corr[num].old_chainid, new_chainId.Text());
            num++;
       }

       Table = _CifObj->GetTablePtr(_StructureId, "struct_biol_gen");
       rowNo = Table->GetNumRows();
       id_index = Table->GetColumnIndex("asym_id");
       pdb_id   = Table->GetColumnIndex("ndb_new_pdb_asym_id");
       mon_id   = Table->GetColumnIndex("symmetry");
       for (i = 0; i < num; i++) {
            for (row = 0; row < rowNo; row++) {
                 Table->GetCell(chainId, id_index, row);
                 Table->GetCell(cifstring, mon_id, row);
                 if (chainId == chain_corr[i].new_chainid &&
                     cifstring == "1_555") {
                      Table->GetCell(new_chainId, pdb_id, row);
                      if (new_chainId == "?" || new_chainId == ".") new_chainId.Copy("");
                      strcpy(chain_corr[i].new_chainid, new_chainId.Text());
                      break;
                 }
            }
            if (!strcmp(chain_corr[i].new_chainid, chain_corr[i].old_chainid) ||
                row == rowNo) {
                 for (j = i; j < num - 1; j++) {
                      strcpy(chain_corr[j].new_chainid, chain_corr[j+1].new_chainid);
                      strcpy(chain_corr[j].old_chainid, chain_corr[j+1].old_chainid);
                 }
                 i--; num--;
            }
       }
       if (num) {
            for (i = 0; i < NUM_CATAGORY_CORR; i++) {
                 Table = _CifObj->GetTablePtr(_StructureId, Category_CORR[i].category);
                 if (is_empty_table(Table)) continue;
                 rowNo = Table->GetNumRows();
                 for (j = 0; j < 3; j++) index[j] = -1;
                 index[0] = Table->GetColumnIndex(Category_CORR[i].item1);
                 if (strcmp(Category_CORR[i].item2, ""))
                      index[1] = Table->GetColumnIndex(Category_CORR[i].item2);
                 if (strcmp(Category_CORR[i].item3, ""))
                      index[2] = Table->GetColumnIndex(Category_CORR[i].item3);
                 for (row = 0; row < rowNo; row++) {
                      for (j = 0; j < 3; j++) {
                           if (index[j] < 0) continue;
                           Table->GetCell(new_chainId, index[j], row);
                           if (new_chainId == "?" || new_chainId == ".") new_chainId.Copy("");
                           for (k = 0; k < num; k++) {
                                if (new_chainId == chain_corr[k].old_chainid) {
                                     cifstring.Copy(chain_corr[k].new_chainid);
                                     Table->UpdateCell(cifstring, index[j], row);
                                     break;
                                }
                           }
                      }
                 }
                 _CifObj->WriteTable(Table, _StructureId, Category_CORR[i].category);
            }
       }

       if (chain_corr) delete [] chain_corr;

       Table = _CifObj->GetTablePtr(_StructureId, "ebi_validate_contact");
       if (is_empty_table(Table)) return;
  
       Table = _CifObj->GetTablePtr(_StructureId, "struct_biol_gen");
       rowNo = Table->GetNumRows();
       id_index = Table->GetColumnIndex("asym_id");
       pdb_id   = Table->GetColumnIndex("ndb_new_pdb_asym_id");
       mon_id   = Table->GetColumnIndex("symmetry");
       num = 0;
       chain_corr = new _CHAINID_CORR[rowNo];
       for (row = 0; row < rowNo; row++) {
            Table->GetCell(cifstring, mon_id, row);
            if (cifstring != "1_555") continue;

            Table->GetCell(chainId, id_index, row);
            Table->GetCell(new_chainId, pdb_id, row);
            strcpy(chain_corr[num].old_chainid, chainId.Text());
            strcpy(chain_corr[num].new_chainid, new_chainId.Text());
            num++;
       }

       AtomTable = _CifObj->GetTablePtr(_StructureId, "atom_site");
       rowNo = AtomTable->GetNumRows();
       auth_asym_id = AtomTable->GetColumnIndex("auth_asym_id");
       auth_comp_id = AtomTable->GetColumnIndex("auth_comp_id");
       auth_seq_id  = AtomTable->GetColumnIndex("auth_seq_id");
       label_asym_id  = AtomTable->GetColumnIndex("label_asym_id");
       int ndb_model = AtomTable->GetColumnIndex("ndb_model");
       AtomTableIndex.Clear();
       for (i = 0; i < rowNo; i++) {
            if (ndb_model >= 0) {
                 AtomTable->GetCell(cifstring, ndb_model, i);
                 if (atoi(cifstring.Text()) > 1) break;
            }
            AtomTable->GetCell(chainId, auth_asym_id, i);
            AtomTable->GetCell(cifstring, auth_comp_id, i);
            AtomTable->GetCell(new_chainId, auth_seq_id, i);
            chainId += "_";
            chainId += cifstring;
            chainId += "_";
            chainId += new_chainId;
            AtomTableIndex.Add(chainId);
       }

       Table = _CifObj->GetTablePtr(_StructureId, "ebi_validate_contact");
       rowNo = Table->GetNumRows();
       idx[0][0] = Table->GetColumnIndex("auth_asym_id_1");
       idx[0][1] = Table->GetColumnIndex("auth_comp_id_1");
       idx[0][2] = Table->GetColumnIndex("auth_seq_id_1");
       idx[1][0] = Table->GetColumnIndex("auth_asym_id_2");
       idx[1][1] = Table->GetColumnIndex("auth_comp_id_2");
       idx[1][2] = Table->GetColumnIndex("auth_seq_id_2");
       for (i = 0; i < rowNo; i++) {
            for (j = 0; j < 2; j++) {
                 cifstring.Clear();
                 for (k = 0; k < 3; k++) {
                      Table->GetCell(chainId, idx[j][k], i);
                      cifstring += chainId;
                      if (k < 2) cifstring += "_";
                 }
                 ret = AtomTableIndex.Seek(cifstring);
                 if (ret >= 0) {
                      AtomTable->GetCell(cifstring, label_asym_id, ret);
                      for (k = 0; k < num; k++) {
                           if (cifstring == chain_corr[k].old_chainid) {
                                cifstring.Copy(chain_corr[k].new_chainid);
                                Table->UpdateCell(cifstring, idx[j][0], i);
                                break;
                           }
                      }
                 }
            }
       }
       _CifObj->WriteTable(Table, _StructureId, "ebi_validate_contact");

       if (chain_corr) delete [] chain_corr;
}

void _FileObj::cifndb_process_asym_poly_seq_ndb(const int no_dbref)
{
       int id_index, mon_id, pdb_num, rowNo, row, need_dbref = 0, first = 1;
       int i, j, *entity_id = NULL, entity_idx, pdb_idx, pdb_start_number;
       int is_protein = 1;
       char buffer[10], **pdb_chain_ids = NULL, **seqs = NULL;
       CifString value, chainId, new_chainId;
       ISTable *Table = NULL;

       Table = _CifObj->GetTablePtr(_StructureId, "struct_ref_seq");
       if (is_empty_table(Table)) need_dbref = 1;
       if (no_dbref) need_dbref = 0;

       Table = _CifObj->GetTablePtr(_StructureId, "ndb_poly_seq_scheme");
       if (is_empty_table(Table)) return;

       id_index = Table->GetColumnIndex("id");
       mon_id   = Table->GetColumnIndex("mon_id");
       pdb_num  = Table->GetColumnIndex("pdb_num");
       entity_idx = Table->GetColumnIndex("entity_id");
       pdb_idx = Table->GetColumnIndex("pdb_id");
       if (id_index < 0 || mon_id < 0) return;

       delete_Seqs();
       rowNo = Table->GetNumRows();
       chainId.Copy("XXX");
       for (row = 0; row < rowNo; row++) {
            Table->GetCell(new_chainId, id_index, row);
            if (new_chainId == "?" || new_chainId == ".") new_chainId.Copy(""); 
            if (new_chainId != chainId) {
                 chainId.Copy(new_chainId);
                 _ManageSeqSpace();
                 _Seqs[_Num_Seqs-1]->ChainId.Copy(new_chainId);
                 if (pdb_num >= 0) {
                      _ManagePdbSpace();
                      _NPdb[_Num_NPdb-1]->ChainId.Copy(new_chainId);
                 }
            }
            get_value(value, Table, mon_id, row);
            _ManageResSpace(_Seqs[_Num_Seqs-1]);
            _Seqs[_Num_Seqs-1]->res[_Seqs[_Num_Seqs-1]->SeqLen] = new RES;
            if (value == "")
                 strcpy(_Seqs[_Num_Seqs-1]->res[_Seqs[_Num_Seqs-1]->SeqLen]->name, "?");
	    else strcpy(_Seqs[_Num_Seqs-1]->res[_Seqs[_Num_Seqs-1]->SeqLen]->name, value.Text());
            _Seqs[_Num_Seqs-1]->SeqLen++;
            if (pdb_num >= 0) {
                 get_value(value, Table, pdb_num, row);
                 _ManageResSpace(_NPdb[_Num_NPdb-1]);
                 _NPdb[_Num_NPdb-1]->res[_NPdb[_Num_NPdb-1]->SeqLen] = new RES;
                 sprintf(buffer, "%d", atoi(value.Text()));
                 strcpy(_NPdb[_Num_NPdb-1]->res[_NPdb[_Num_NPdb-1]->SeqLen]->name,
                           buffer);
                 _NPdb[_Num_NPdb-1]->SeqLen++;
            }
       }
       get_sequence_types();
       if (need_dbref && entity_idx >= 0 && pdb_idx >= 0) {
            entity_id = new int[_Num_Seqs];
            pdb_chain_ids = new char*[_Num_Seqs];
            for (i = 0; i < _Num_Seqs; i++) pdb_chain_ids[i] = new char[4];    
            chainId.Copy("XXX");
            i = 0;
            for (row = 0; row < rowNo; row++) {
                 Table->GetCell(new_chainId, id_index, row);
                 if (new_chainId == "?" || new_chainId == ".") new_chainId.Copy("");
                 if (new_chainId != chainId) {
                      chainId.Copy(new_chainId);
                      get_value(value, Table, entity_idx, row);
                      entity_id[i] = atoi(value.Text());
                      get_value(value, Table, pdb_idx, row);
                      strcpy(pdb_chain_ids[i], value.Text());
                      i++;
                 }
            }
            for (i = 0; i < _Num_Seqs; i++) {
                 seqs = new char*[_Seqs[i]->SeqLen];
                 for (j = 0; j < _Seqs[i]->SeqLen; j++) {
                      seqs[j] = new char[4];
                      strcpy(seqs[j], _Seqs[i]->res[j]->name);
                 }
                 pdb_start_number = 1;
                 if (pdb_num >= 0) pdb_start_number = atoi(_NPdb[i]->res[0]->name);
                 is_protein = 0;
                 if (_Seqs[i]->chain_type == ATOMP_TOKEN) is_protein = 1;
                 create_dbref_and_seqadv(entity_id[i], _Seqs[i]->SeqLen, pdb_start_number,
                        seqs, 1, &pdb_chain_ids[i], first, is_protein);
                 for (j = 0; j < _Seqs[i]->SeqLen; j++) delete [] seqs[j];
                 delete [] seqs;
            }
            delete [] entity_id;
            for (i = 0; i < _Num_Seqs; i++) delete [] pdb_chain_ids[i];
            delete [] pdb_chain_ids;
       }
}

void _FileObj::match_entity_poly_seq_to_seqres(const int no_dbref)
{
       int entity_id, mon_index, entity_id1, id_index, rowNo, SeqLen, irow;
       int first = 1, need_dbref = 0;
       char prev_entity_id[5], **seqs = NULL;
       CifString value;
       ISTable *Table = NULL, *Table1 = NULL;

       Table = _CifObj->GetTablePtr(_StructureId, "struct_ref_seq");
       if (is_empty_table(Table)) need_dbref = 1;
       if (no_dbref) need_dbref = 0;

       Table = _CifObj->GetTablePtr(_StructureId, "entity_poly_seq");
       Table1 = _CifObj->GetTablePtr(_StructureId, "struct_asym");
       if (is_empty_table(Table) || is_empty_table(Table1)) return;

       entity_id = Table->GetColumnIndex("entity_id");
       mon_index = Table->GetColumnIndex("mon_id");
       entity_id1 = Table1->GetColumnIndex("entity_id");
       id_index = Table1->GetColumnIndex("id");
       if (entity_id < 0 || mon_index < 0 || entity_id1 < 0 || id_index < 0)
            return;

       if (need_dbref) {
            id_index = Table1->GetColumnIndex("ndb_pdb_id");
            if (id_index < 0) need_dbref = 0;
       }
       rowNo = Table->GetNumRows();
       strcpy(prev_entity_id, "-1");
       SeqLen = 0;
       seqs = new char*[rowNo];
       for (irow = 0; irow < rowNo; irow++) seqs[irow] = new char[4];
       for (irow = 0; irow < rowNo; irow++) {
            get_value(value, Table, entity_id, irow);
            if (value != prev_entity_id) {
                 if (SeqLen) insert_to_seqs(Table1, prev_entity_id, SeqLen, seqs,
                                first, need_dbref);
                 strcpy(prev_entity_id, value.Text());
                 SeqLen = 0;
            }
            get_value(value, Table, mon_index, irow);
            if (value == "")
                 strcpy(seqs[SeqLen], "?");
	    else strcpy(seqs[SeqLen], value.Text());
            SeqLen++;
       }
       if (SeqLen) insert_to_seqs(Table1, prev_entity_id, SeqLen, seqs, first, need_dbref);
       for (irow = 0; irow < rowNo; irow++) delete [] seqs[irow];
       delete [] seqs;
       get_sequence_types();
}

void _FileObj::insert_to_seqs(ISTable *Table, char *entity_id, const int SeqLen, char **seqs,
                 int &first, const int need_dbref)
{
       int entity_index, id_index, pdb_index, rowNo, irow, i, entity, is_protein = 1;
       char **chain_ids = NULL;
       CifString string, value;

       entity = atoi(entity_id);
       entity_index = Table->GetColumnIndex("entity_id");
       id_index = Table->GetColumnIndex("id");
       pdb_index = Table->GetColumnIndex("ndb_pdb_id");
       rowNo = Table->GetNumRows();
       if (need_dbref) {
            chain_ids = new char*[1];
            chain_ids[0] = new char[4];
            for (i = 0; i < SeqLen; i++) {
                 if (!res_name_cmp(seqs[i], "A") || !res_name_cmp(seqs[i], "C") ||
                     !res_name_cmp(seqs[i], "G") || !res_name_cmp(seqs[i], "I") ||
                     !res_name_cmp(seqs[i], "T") || !res_name_cmp(seqs[i], "U")) {
                      is_protein = 0;
                      break;
                 }
            }
       }
       for (irow = 0; irow < rowNo; irow++) {
            get_value(value, Table, entity_index, irow);
            if (value == entity_id) {
                 Table->GetCell(string, id_index, irow);
                 if (string == "?" || string == ".") string.Copy("");
                 _ManageSeqSpace();
                 _Seqs[_Num_Seqs-1]->ChainId.Copy(string);
                 for (i = 0; i < SeqLen; i++) {
                      _ManageResSpace(_Seqs[_Num_Seqs-1]);
                      _Seqs[_Num_Seqs-1]->res[_Seqs[_Num_Seqs-1]->SeqLen] = new RES;
                      strcpy(_Seqs[_Num_Seqs-1]->res[_Seqs[_Num_Seqs-1]->SeqLen]->
                               name, seqs[i]);
                      _Seqs[_Num_Seqs-1]->SeqLen++;
                 }
                 if (need_dbref) {
                      Table->GetCell(string, pdb_index, irow);
                      strcpy(chain_ids[0], string.Text());
                      create_dbref_and_seqadv(entity, SeqLen, 1, seqs, 1, chain_ids, first, is_protein);
                 }
            }
       }
       if (chain_ids) {
            delete [] chain_ids[0];
            delete [] chain_ids;
       }
}

void _FileObj::match_entity_poly_to_seqres(const int no_dbref)
{
       int first = 1, entity_index, seq_index, chain_id, three_index, start_index;
       int i, j, rowNo, irow, num, alloc_num, is_protein, is_nucleic_acid;
       int alloc_chain, num_chain, new_chain, entity_id, start_number, type_index; 
       int need_dbref = 1;
       char **chain, buff[4], *p, *value = NULL, **seqs = NULL;
       CifString string;
       ISTable *Table = NULL, *Table1 = NULL;

       Table = _CifObj->GetTablePtr(_StructureId, "entity_poly");
       Table1 = _CifObj->GetTablePtr(_StructureId, "struct_asym");
       if (is_empty_table(Table)) return;

       entity_index = Table->GetColumnIndex("entity_id");
       seq_index = Table->GetColumnIndex("ndb_seq_one_letter_code");
       chain_id  = Table->GetColumnIndex("ndb_chain_id");
       three_index = Table->GetColumnIndex("ndb_seq_three_letter_code");
       start_index = Table->GetColumnIndex("rcsb_seq_align_begin");
       type_index = Table->GetColumnIndex("type");
       if (three_index < 0 && seq_index < 0) return;

       if (no_dbref) need_dbref = 0;

       rowNo = Table->GetNumRows();
       for (irow = 0; irow < rowNo; irow++) {
            string.Clear();
            value = NULL;
            seqs = NULL;
            num = 0;
            alloc_num = 0;
            Table->GetCell(string, seq_index, irow);
            is_protein = 1;
            if (string == "" || string == "." || string == "?" || string == "@") {
                 if (three_index >= 0) {
                      Table->GetCell(string, three_index, irow);
                      if (string != "" && string != "." && string != "?") {
                           value = new char[string.Length()+1];
                           strcpy(value, string.Text());
                           ndb_clean_string(value);
                           alloc_num = strlen(value) / 4 + 1;
                           seqs = new char*[alloc_num];
                           for (i = 0; i < alloc_num; i++) seqs[i] = new char[4];
                           p = value;
                           while (p) {
                                p = get_word(p, buff);
                                ndb_set_string_to_upper(buff);
                                strcpy(seqs[num], buff);
                                num++;
                           }
                      }
                 }
            } else {
                 value = new char[string.Length()+1];
                 strcpy(value, string.Text());
                 ndb_clean_string(value);
                 alloc_num = strlen(value);
                 seqs = new char*[alloc_num];
                 for (i = 0; i < alloc_num; i++) seqs[i] = new char[4];
                 for (i = 0; i < alloc_num; i++) {
                      if (value[i] == ' ') continue;
     
                      if (value[i] != '(') {
                           sprintf(seqs[num], "%c", value[i]);
                           num++;
                      } else {
                           i++;
                           j = 0;
                           while (value[i] != ')') {
                                seqs[num][j] = value[i];
                                i++; j++;
                           }
                           seqs[num][j] = '\0';
                           num++;
                      }
                 }

                 string.Clear();
                 if (type_index >= 0) Table->GetCell(string, type_index, irow);
                 if (string != "" && string != "?" && string != ".") {
                      if (!strncasecmp(string.Text(), "polypeptide", 11) ||
                          !strncasecmp(string.Text(), "PROTEIN", 7))
                           is_protein = 1;
                      else is_protein = 0;
                 } else {
                      is_nucleic_acid = 0;
                      for (i = 0; i < num; i++) {
                           ndb_set_string_to_upper(seqs[i]);
                           if (strlen(seqs[i]) == 1 && (seqs[i][0] == 'A' ||
                               seqs[i][0] == 'C' || seqs[i][0] == 'G' ||
                               seqs[i][0] == 'I' || seqs[i][0] == 'T' ||
                               seqs[i][0] == 'U' || seqs[i][0] == 'N')) {
                                is_nucleic_acid++;
                           }
                      }
                      if ((((float) is_nucleic_acid) / ((float) num)) > 0.7)
                           is_nucleic_acid = 1;
                      else is_nucleic_acid = 0;
                      if (is_nucleic_acid) is_protein = 0;
                 }
                 if (is_protein) {
                      for (i = 0; i < num; i++) {
                           if (strlen(seqs[i]) == 1) {
                                strcpy(buff, get_threeletter_code(seqs[i][0]));
                                ndb_set_string_to_upper(buff);
                                strcpy(seqs[i], buff);
                           }
                      }
                 }
            }
            if (value == NULL) continue;
            delete [] value;

            entity_id = -1;
            get_value(string, Table, entity_index, irow);
            if (string != "") entity_id = atoi(string.Text());

            if (chain_id < 0)
                 value = get_chain_ids(entity_id, Table1);
            else {
                 // delete [] value;
                 get_value(string, Table, chain_id, irow);
                 value = new char[string.Length()+1];
                 strcpy(value, string.Text());
            }
            ndb_log_message(NDB_MSG_WARN, "WARNING: PDB Chain_ID(s) = %s\n", value);

            num_chain = 0;
            new_chain = 0;
            alloc_chain = strlen(value) + 1;
            chain = new char*[alloc_chain];
            for (i = 0; i < alloc_chain; i++) chain[i] = NULL;
            j = 0;
            // for (i = 0; i < strlen(value); i++) {
            for (i = 0; i < alloc_chain - 1; i++) {
                 chain[i] = new char[10];
                 if (value[i] != ' ' && value[i] != ',') {
                      chain[num_chain][j++] = value[i];
                      new_chain = 1;
                 } else if (new_chain) {
                      chain[num_chain][j] = '\0';
                      j = 0;
                      num_chain++;
                      new_chain = 0;
                 }
            }
            if (!chain[num_chain]) chain[num_chain] = new char[1];
            chain[num_chain][j] = '\0';
            num_chain++;

            start_number = -9999;
            get_value(string, Table, start_index, irow);
            if (string != "") start_number = atoi(string.Text());
            if (need_dbref) create_dbref_and_seqadv(entity_id, num, start_number, seqs,
                               num_chain, chain, first, is_protein);

            for (i = 0; i < num_chain; i++) {
                 _ManageSeqSpace();
                 _Seqs[_Num_Seqs-1]->ChainId.Copy(chain[i]);
                 for (j = 0; j < num; j++) {
                      _ManageResSpace(_Seqs[_Num_Seqs-1]);
                      _Seqs[_Num_Seqs-1]->res[_Seqs[_Num_Seqs-1]->SeqLen] = new RES;
                      strcpy(_Seqs[_Num_Seqs-1]->res[_Seqs[_Num_Seqs-1]->SeqLen]->
                             name, seqs[j]);
                      _Seqs[_Num_Seqs-1]->SeqLen++;
                 }
            }
            for (i = 0; i < alloc_num; i++) delete [] seqs[i];
            delete [] seqs;
            for (i = 0; i < alloc_chain; i++) if (chain[i]) delete [] chain[i];
            delete [] chain;
            delete [] value;
       }
       get_sequence_types();
}

void _FileObj::get_symmetry_from_remark_350()
{
       int biol_id, colNo, i, id_idx, j, remark_idx, ret;
       int row, rowNo, sym_idx;
       BTreeObj<_SSIndex> _symmetry_chain, _biol_unit;
       char *p, chain_id[5], buffer[10];
       CifString cifstring, remark, symmetry;
       ISTable *Table = NULL;
       _SSIndex index;

       Table = _CifObj->GetTablePtr(_StructureId, "database_PDB_remark");
       if (is_empty_table(Table)) return;

       Table = _CifObj->GetTablePtr(_StructureId, "struct_biol_gen");
       if (is_empty_table(Table)) return;

       Table = _CifObj->GetTablePtr(_StructureId, "database_PDB_remark");
       rowNo = Table->GetNumRows();
       id_idx = Table->GetColumnIndex("id");
       remark_idx = Table->GetColumnIndex("text");
       remark.Clear();
       for (i = 0; i < rowNo; i++) {
            Table->GetCell(cifstring, id_idx, i);
            if (cifstring == "350") {
                 Table->GetCell(remark, remark_idx, i);
                 if (remark == "." || remark == "?") remark.Clear();
                 break;
            }
       }
       if (remark == "") return;

       convert_remark_350_to_symmetry_operation(_symmetry_chain, _biol_unit,
                    _cell, remark.Text(), _error_message, _StructureId);
       if (_symmetry_chain.Root() < 0 && _biol_unit.Root() < 0) return;

       Table = _CifObj->GetTablePtr(_StructureId, "struct_biol_gen");
       rowNo = Table->GetNumRows();
       colNo = Table->GetNumColumns();
       biol_id = Table->GetColumnIndex("biol_id");
       id_idx = Table->GetColumnIndex("ndb_new_pdb_asym_id");
       sym_idx = Table->GetColumnIndex("symmetry");
       remark_idx = Table->GetColumnIndex("rcsb_pdb_order");

       if (_biol_unit.Root() >= 0) {
            _biol_unit.GoFirst();
            ret = 1;
            while (ret >= 0) {
                 index = _biol_unit.GetCurrData();
                 p = index.ivalue();
                 while (p) {
                      p = get_word(p, chain_id);
                      for (i = 0; i < rowNo; i++) {
                           Table->GetCell(cifstring, id_idx, i);
                           if (cifstring == chain_id) {
                                cifstring.Copy(index.svalue());
                                Table->UpdateCell(cifstring, biol_id, i);
                           }
                      }
                 }
                 ret = _biol_unit.GoToNext();
            }
            _CifObj->WriteTable(Table, _StructureId, "struct_biol_gen");
       }

       if (_symmetry_chain.Root() < 0) return;

       _symmetry_chain.GoFirst();
       ret = 1;
       while (ret >= 0) {
            index = _symmetry_chain.GetCurrData();
            p = index.ivalue();
            while (p) {
                 p = get_word(p, chain_id);
                 for (i = 0; i < rowNo; i++) {
                      Table->GetCell(cifstring, id_idx, i);
                      Table->GetCell(symmetry, sym_idx, i);
                      if (cifstring == chain_id && symmetry == index.svalue()) break;
                 }
                 if (i < rowNo) continue;
                 for (i = 0; i < rowNo; i++) {
                      Table->GetCell(cifstring, id_idx, i);
                      Table->GetCell(symmetry, sym_idx, i);
                      if (cifstring == chain_id && symmetry == "1_555") {
                           row = Table->GetNumRows();
                           Table->AddRow();
                           for (j = 0; j < colNo; j++) {
                                Table->GetCell(cifstring, j, i);
                                if (j == sym_idx)
                                     cifstring.Copy(index.svalue());
                                else if (j == remark_idx) {
                                     sprintf(buffer, "%d", row + 1);
                                     cifstring.Copy(buffer);
                                } else if (j == id_idx) {
                                     strcpy(buffer, cifstring.Text());
                                     ndb_set_string_to_lower(buffer);
                                     cifstring.Copy(buffer);
                                }
                                Table->UpdateCell(cifstring, j, row);
                           }
                      }
                 }
            }
            ret = _symmetry_chain.GoToNext();
       }
       _CifObj->WriteTable(Table, _StructureId, "struct_biol_gen");
}

static char **_seqa = NULL, **_seqb = NULL;

static int score(const int k, const int l)
{
       if (res_name_cmp(_seqa[k], _seqb[l]))
            return 0;
       else return 4;
}

void _FileObj::create_dbref_and_seqadv(const int entity_id, const int length,
          int pdb_start_number, char **pdb_seqs, const int num_chain,
          char **pdb_chain_ids, int &first, const int is_protein)
{
       int id_index, name_index, accession_index, entity_index;
       int seq_index, start_index, rowNo, irow, entity_ID, alloc_num, i, j, num;
       int start_number, p_start, p_end, s_start, s_end, begin, end, align_id;
       int row_no, exist_seqadv;
       char buff[4], pdb_code[5], *value = NULL, **seqs = NULL;
       CifString string, id, accession, name;
       ALIGN *sa = NULL, *sb = NULL, *ss = NULL;
       ISTable *Table = NULL, *Table1 = NULL, *Dbref = NULL, *Seqadv = NULL;

       if (entity_id < 0) return;
       Table = _CifObj->GetTablePtr(_StructureId, "struct_ref");
       if (is_empty_table(Table)) return;

       id_index = Table->GetColumnIndex("id");
       name_index = Table->GetColumnIndex("db_name");
       accession_index = Table->GetColumnIndex("rcsb_db_accession");
       entity_index = Table->GetColumnIndex("entity_id");
       seq_index = Table->GetColumnIndex("rcsb_seq_one_letter_code");
       start_index = Table->GetColumnIndex("rcsb_align_begin");
       if (id_index < 0 || name_index < 0 || entity_index < 0 || seq_index < 0) return;

       if (pdb_start_number == -9999) pdb_start_number = 1;


       pdb_code[0] = '\0';
       Table1 = _CifObj->GetTablePtr(_StructureId, "database_2");
       if (!is_empty_table(Table1)) get_pdb_id(Table1, pdb_code);

       rowNo = Table->GetNumRows();
       for (irow = 0; irow < rowNo; irow++) {
            entity_ID = -1;
            get_value(id, Table, entity_index, irow);
            if (id != "") entity_ID = atoi(id.Text());
            if (entity_ID == entity_id) {
                 Table->GetCell(string, seq_index, irow);
                 if (string == "" || string == "." || string == "?") break;
                 value = new char[string.Length()+1];
                 strcpy(value, string.Text());
                 ndb_clean_string(value);
                 ndb_set_string_to_upper(value);
                 alloc_num = strlen(value);
                 seqs = new char*[alloc_num];
                 for (i = 0; i < alloc_num; i++) seqs[i] = new char[4];
                 num = 0;
                 for (i = 0; i < alloc_num; i++) {
                      if (value[i] == ' ') continue;

                      if (value[i] != '(') {
                           sprintf(seqs[num], "%c", value[i]);
                           num++;
                      } else {
                           i++;
                           j = 0;
                           while (value[i] != ')') {
                                seqs[num][j] = value[i];
                                i++; j++;
                           }
                           seqs[num][j] = '\0';
                           num++;
                      }
                 }
                 if (is_protein) {
                      for (i = 0; i < num; i++) {
                           if (strlen(seqs[i]) == 1) {
                                strcpy(buff, get_threeletter_code(seqs[i][0]));
                                ndb_set_string_to_upper(buff);
                                strcpy(seqs[i], buff);
                           }
                      }
                 }
                 start_number = 1;
                 get_value(id, Table, start_index, irow);
                 if (id != "") start_number = atoi(id.Text());
                 get_value(id, Table, id_index, irow);
                 get_value(name, Table, name_index, irow);
                 get_value(accession, Table, accession_index, irow);
                 
                 sa = new ALIGN;
                 sa->AlignLength = num;
                 sa->aln = new ALN[sa->AlignLength];
                 for (i = 0; i < sa->AlignLength; i++) sa->aln[i].pos[0] = i;
                 sb = new ALIGN;
                 sb->AlignLength = length;
                 sb->aln = new ALN[sb->AlignLength];
                 for (i = 0; i < sb->AlignLength; i++) sb->aln[i].pos[0] = i;
                 _seqa = seqs;
                 _seqb = pdb_seqs;
                 ss = alignment(sa, sb, 2, score);

                 Dbref = NULL;
                 Seqadv = NULL;
                 if (first) {
                      first = 0;
                      if (_CifObj->IsTablePresent(_StructureId, "struct_ref_seq"))
                           _CifObj->DeleteTable(_StructureId, "struct_ref_seq");
                      Dbref = new ISTable("struct_ref_seq");
                      Dbref->AddColumn("align_id");
                      Dbref->AddColumn("ref_id");
                      Dbref->AddColumn("ndb_pdb_id_code");
                      Dbref->AddColumn("ndb_chain_id");
                      Dbref->AddColumn("seq_align_beg");
                      Dbref->AddColumn("ndb_seq_align_beg_ins_code");
                      Dbref->AddColumn("seq_align_end");
                      Dbref->AddColumn("ndb_seq_align_end_ins_code");
                      Dbref->AddColumn("ndb_db_accession");
                      Dbref->AddColumn("db_align_beg");
                      Dbref->AddColumn("ndb_db_align_beg_ins_code");
                      Dbref->AddColumn("db_align_end");
                      Dbref->AddColumn("ndb_db_align_end_ins_code");

                      if (_CifObj->IsTablePresent(_StructureId, "struct_ref_seq_dif"))
                           _CifObj->DeleteTable(_StructureId, "struct_ref_seq_dif");
                      Seqadv = new ISTable("struct_ref_seq_dif");
                      Seqadv->AddColumn("align_id");
                      Seqadv->AddColumn("ndb_pdb_id_code");
                      Seqadv->AddColumn("mon_id");
                      Seqadv->AddColumn("ndb_pdb_chain_id");
                      Seqadv->AddColumn("seq_num");
                      Seqadv->AddColumn("ndb_seq_ins_code");
                      Seqadv->AddColumn("ndb_seq_db_name");
                      Seqadv->AddColumn("ndb_seq_db_accession_code");
                      Seqadv->AddColumn("db_mon_id");
                      Seqadv->AddColumn("ndb_seq_db_seq_num");
                      Seqadv->AddColumn("details");
                 } else {
                      Dbref = _CifObj->GetTablePtr(_StructureId, "struct_ref_seq");
                      Seqadv = _CifObj->GetTablePtr(_StructureId, "struct_ref_seq_dif");
                 }

                 p_start = p_end = s_start = s_end = -9999;
                 begin = -1; end = -1;
                 for (i = 0; i < ss->AlignLength; i++) {
                      if (ss->aln[i].pos[0] >= 0 && ss->aln[i].pos[1] >= 0) {
                           if (p_start == -9999) {
                                s_start = start_number + ss->aln[i].pos[0];
                                p_start = pdb_start_number + ss->aln[i].pos[1];
                           }
                           s_end = start_number + ss->aln[i].pos[0];
                           p_end = pdb_start_number + ss->aln[i].pos[1];
                      }
                      if (ss->aln[i].pos[1] >= 0) {
                           if (begin < 0) begin = i;
                           end = i;
                      }
                 }
                 if (p_start != -9999 && p_end != -9999 && s_start != -9999 &&
                     s_end != -9999 && begin >= 0 && end >= 0) {
                      align_id = Dbref->GetNumRows();
                      for (i = 0; i < num_chain; i++) {
                           Dbref->AddRow();
                           sprintf(buff, "%d", align_id+1);
                           string.Copy(buff);
                           Dbref->UpdateCell(string, 0, align_id);
                           Dbref->UpdateCell(id, 1, align_id);
                           string.Copy(pdb_code);
                           Dbref->UpdateCell(string, 2, align_id);
                           string.Copy(pdb_chain_ids[i]);
                           Dbref->UpdateCell(string, 3, align_id);
                           sprintf(buff, "%d", p_start);
                           string.Copy(buff);
                           Dbref->UpdateCell(string, 4, align_id);
                           string.Copy("");
                           Dbref->UpdateCell(string, 5, align_id);
                           Dbref->UpdateCell(string, 7, align_id);
                           Dbref->UpdateCell(string, 10, align_id);
                           Dbref->UpdateCell(string, 12, align_id);
                           sprintf(buff, "%d", p_end);
                           string.Copy(buff);
                           Dbref->UpdateCell(string, 6, align_id);
                           Dbref->UpdateCell(accession, 8, align_id);
                           sprintf(buff, "%d", s_start);
                           string.Copy(buff);
                           Dbref->UpdateCell(string, 9, align_id);
                           sprintf(buff, "%d", s_end);
                           string.Copy(buff);
                           Dbref->UpdateCell(string, 11, align_id);
                           row_no = Seqadv->GetNumRows();
                           for (j = begin; j <= end; j++) {
                                exist_seqadv = 0;
                                if (ss->aln[j].pos[0] >= 0 && ss->aln[j].pos[1] >= 0) {
                                     if (res_name_cmp(_seqa[ss->aln[j].pos[0]],
                                             _seqb[ss->aln[j].pos[1]])) {
                                          exist_seqadv = 1;
                                          Seqadv->AddRow();
                                          string.Copy(_seqb[ss->aln[j].pos[1]]);
                                          Seqadv->UpdateCell(string, 2, row_no);
                                          sprintf(buff, "%d", pdb_start_number + ss->aln[j].pos[1]);
                                          string.Copy(buff);
                                          Seqadv->UpdateCell(string, 4, row_no);
                                          string.Copy(_seqa[ss->aln[j].pos[0]]);
                                          Seqadv->UpdateCell(string, 8, row_no);
                                          sprintf(buff, "%d", start_number + ss->aln[j].pos[0]);
                                          string.Copy(buff);
                                          Seqadv->UpdateCell(string, 9, row_no);
                                          string.Copy("CONFLICT");
                                          Seqadv->UpdateCell(string, 10, row_no);
                                     }
                                } else if (ss->aln[j].pos[0] >= 0) {
                                     exist_seqadv = 1;
                                     Seqadv->AddRow();
                                     string.Copy("");
                                     Seqadv->UpdateCell(string, 2, row_no);
                                     Seqadv->UpdateCell(string, 4, row_no);
                                     string.Copy(_seqa[ss->aln[j].pos[0]]);
                                     Seqadv->UpdateCell(string, 8, row_no);
                                     sprintf(buff, "%d", start_number + ss->aln[j].pos[0]);
                                     string.Copy(buff);
                                     Seqadv->UpdateCell(string, 9, row_no);
                                     string.Copy("DELETION");
                                     Seqadv->UpdateCell(string, 10, row_no);
                                } else if (ss->aln[j].pos[1] >= 0) {
                                     exist_seqadv = 1;
                                     Seqadv->AddRow();
                                     string.Copy(_seqb[ss->aln[j].pos[1]]);
                                     Seqadv->UpdateCell(string, 2, row_no);
                                     sprintf(buff, "%d", pdb_start_number + ss->aln[j].pos[1]);
                                     string.Copy(buff);
                                     Seqadv->UpdateCell(string, 4, row_no);
                                     string.Copy("");
                                     Seqadv->UpdateCell(string, 8, row_no);
                                     Seqadv->UpdateCell(string, 9, row_no);
                                     string.Copy("INSERTION");
                                     Seqadv->UpdateCell(string, 10, row_no);
                                }
                                if (exist_seqadv) {
                                     sprintf(buff, "%d", align_id+1);
                                     string.Copy(buff);
                                     Seqadv->UpdateCell(string, 0, row_no);
                                     string.Copy(pdb_code);
                                     Seqadv->UpdateCell(string, 1, row_no);
                                     string.Copy(pdb_chain_ids[i]);
                                     Seqadv->UpdateCell(string, 3, row_no);
                                     string.Copy("");
                                     Seqadv->UpdateCell(string, 5, row_no);
                                     string.Copy(name);
                                     Seqadv->UpdateCell(string, 6, row_no);
                                     Seqadv->UpdateCell(accession, 7, row_no);
                                     row_no++;
                                }
                           }
                           align_id++;
                      }
                 }
                 _CifObj->WriteTable(Dbref, _StructureId, "struct_ref_seq");
                 _CifObj->WriteTable(Seqadv, _StructureId, "struct_ref_seq_dif");

                 _seqa = NULL;
                 _seqb = NULL;
                 delete [] sa->aln; delete sa;
                 delete [] sb->aln; delete sb;
                 delete [] ss->aln; delete ss;
                 for (i = 0; i < alloc_num; i++) delete [] seqs[i];
                 delete [] seqs;
                 delete [] value;
                 break;
            }
       }
}

static void add_connect_record(ConnectFormat *drug, _RESIDUE *residue)
{
       int i = 0, j = 0, k = 0, num1 = 0, num2 = 0, atomNo;
       _ATOM **atom1 = NULL, **atom2 = NULL;

       if (drug->natoms() <= 1 || residue->AtomNumbers() <= 1) return;
       for (atomNo = 0; atomNo < drug->natoms(); atomNo++) {
            if (!drug->atoms(atomNo)->nconect()) continue; 
        
            atom1 = residue->find_atom(drug->atoms(atomNo)->atomname(), num1);
            if (!atom1) continue;

            for (i = 0; i < drug->atoms(atomNo)->nconect(); i++) {
                 atom2 = residue->find_atom(drug->atoms(atomNo)->connectto(i), num2);
                 if (!atom2) continue;
                 for (j = 0; j < num1; j++) {
                      if (strcmp(atom1[j]->restype(), drug->drugname())) continue;
                      for (k = 0; k < num2; k++) {
                           if (strcmp(atom2[k]->restype(), drug->drugname())) continue;
                           if (!strcmp(atom1[j]->alt_loc(), "") ||
                               !strcmp(atom2[k]->alt_loc(), "") ||
                               !strcmp(atom1[j]->alt_loc(),
                                       atom2[k]->alt_loc())) {
                                insert_atom_id(atom1[j]->atnum());
                                insert_connect_atom_id(atom1[j]->atnum(),
                                                       atom2[k]->atnum());
                           }
                      }
                 }
                 delete [] atom2;
            }
            delete [] atom1;
       }
}

static void insert_atom_id(const int atom_id)
{
       int i;
       CONECT *found = NULL, tmp;
       
       if (Alloc_connect == 0) {
            Alloc_connect = 50;
            connect = (CONECT *) calloc(Alloc_connect, sizeof(CONECT));
            for (i = 0; i < Alloc_connect; i++) {
                 connect[i].atom_id = 0;
                 connect[i].con_num = 0;
            }
       }

       if (Num_connect) {
            tmp.atom_id = atom_id;
            found = (CONECT *) bsearch(&tmp, connect, Num_connect,
                     sizeof(CONECT), connect_compare);
       }
       if (found) return;
   
       Num_connect++;
       if (Num_connect > Alloc_connect) {
            connect = (CONECT *) realloc(connect, (Alloc_connect + 50)
                      * sizeof(CONECT));
            for (i = Alloc_connect; i < (Alloc_connect + 50); i++) {
                 connect[i].atom_id = 0;
                 connect[i].con_num = 0;
            }
            Alloc_connect += 50;
       }
       connect[Num_connect-1].atom_id = atom_id;
       connect[Num_connect-1].con_num = 0;
       if (Num_connect > 1) qsort((void *) connect, Num_connect,
                     sizeof(CONECT), connect_compare);
}

static void insert_connect_atom_id(const int atom_id, const int atnum)
{
       int connect_to = atnum;
       int *num = NULL;
       CONECT tmp, *found = NULL;

       if (Num_connect == 0) insert_atom_id(atom_id);

       tmp.atom_id = atom_id;
       found = (CONECT *) bsearch(&tmp, connect, Num_connect,
                   sizeof(CONECT), connect_compare);

       if (found == NULL) {
            insert_atom_id(atom_id);
            found = (CONECT *) bsearch(&tmp, connect, Num_connect,
                    sizeof(CONECT), connect_compare);
       }

       if (found->con_num) {
            num = (int *) bsearch(&connect_to, found->con_to,
                    found->con_num, sizeof(int), int_compare); 
       }
       if (num) return;

       found->con_to[found->con_num] = atnum;
       found->con_num++;
       if (found->con_num > 1) qsort((void *) found->con_to,
                      found->con_num, sizeof(int), int_compare);
}

#ifdef SUN_OS
#ifdef __cplusplus
extern "C" {
#endif
#endif
static int connect_compare(const void *connect1, const void *connect2)
{
       return (((CONECT *) connect1)->atom_id - 
               ((CONECT *) connect2)->atom_id);
}

static int int_compare(const void *int1, const void *int2)
{
       return (*((int *) int1) - *((int *) int2));
}
#ifdef SUN_OS
#ifdef __cplusplus
}
#endif
#endif

void get_pdb_id(ISTable *Table1, char *type)
{
       int index_1, index_2, rowNo, i;
       CifString id;

       *type = '\0';
       index_1 = Table1->GetColumnIndex("database_id");
       index_2 = Table1->GetColumnIndex("database_code");
       rowNo = Table1->GetNumRows();
       for (i = 0; i < rowNo; i++) {
            get_value(id, Table1, index_1, i);
            if (!strcasecmp(id.Text(), "PDB")) {
                 get_value(id, Table1, index_2, i);
                 strcpy(type, id.Text());
                 break;
            }
       }
}

static void convert_remark_350_to_symmetry_operation(BTreeObj<_SSIndex>& _symmetry_chain,
        BTreeObj<_SSIndex>& _biol_unit, NDBCELLCONST *cell, char *remark_350,
        CifString &message, const char *structureID)
{
       int  i, is_chain_id_line = 0, ret = 1, str_len = 0;
       char buff[100], buffer[100], line[100], symmetry[20], *p, *q, *r, unit[10];
       double matrix[3][3], vector[3];
       CifString chain_ids;
       _SSIndex SymIdx, *SymPtr = NULL;

       chain_ids.Clear();
       _symmetry_chain.Clear();
       _biol_unit.Clear();
       memset(unit, '\0', 10);
       p = remark_350;
       while (p) {
            p = get_line_from_string(p, line);
            if ((q = strstr(line, "BIOMOLECULE:"))) {
                 q += 12;
                 sscanf(q, "%s", unit);
            } else if ((q = strstr(line, "APPLY THE FOLLOWING TO CHAINS:"))) {
                 is_chain_id_line = 1;
                 q += 30;
                 chain_ids.Clear();
                 if (!strstr(q, "NULL")) {
                      str_len = strlen(q);
                      for (i = 0; i < str_len; i++) {
                           if (q[i] == ',') q[i] = ' ';
                      }
                      while (q) {
                           q = get_word(q, buffer);
                           if (strcmp(buffer, "")) {
                                if (chain_ids != "") chain_ids += " ";
                                chain_ids += buffer;
                           }
                      }
                 }
            } else if (strstr(line, "BIOMT1")) {
                 if (strcmp(unit, "")) {
                      if (_biol_unit.Root() >= 0) {
                           ret = 1;
                           while (ret >= 0) {
                                SymIdx = _biol_unit.GetCurrData();
                                r = SymIdx.ivalue();
                                while (r) {
                                     r = get_word(r, buffer);
                                     q = chain_ids.Text();
                                     while (q) {
                                          q = get_word(q, buff);
                                          if (!strcmp(buff, buffer)) {
                                               message += structureID;
                                               message += ": Different biol unit had same chain(s):\n";
                                               message += "Unit ";
                                               message += SymIdx.svalue();
                                               message += ": ";
                                               message += SymIdx.ivalue();
                                               message += "\n";
                                               message += "Unit ";
                                               message += unit;
                                               message += ": ";
                                               message += chain_ids;
                                               message += "\n";
                                               _symmetry_chain.Clear();
                                               _biol_unit.Clear();
                                               return;
                                          }
                                     }
                                }
                                ret = _biol_unit.GoToNext();
                           }
                      }
                      
                      SymIdx.set_svalue(unit);
                      SymIdx.set_ivalue(chain_ids.Text());
                      _biol_unit.Add(SymIdx);
                      memset(unit, '\0', 10);
                 }

                 is_chain_id_line = 0;
                 sscanf(line, "%s %d %lf %lf %lf %lf", buff, &i, &matrix[0][0],
                                 &matrix[0][1], &matrix[0][2], &vector[0]);
            } else if (strstr(line, "BIOMT2")) {
                 is_chain_id_line = 0;
                 sscanf(line, "%s %d %lf %lf %lf %lf", buff, &i, &matrix[1][0],
                                 &matrix[1][1], &matrix[1][2], &vector[1]);
            } else if (strstr(line, "BIOMT3")) {
                 is_chain_id_line = 0;
                 sscanf(line, "%s %d %lf %lf %lf %lf", buff, &i, &matrix[2][0],
                                 &matrix[2][1], &matrix[2][2], &vector[2]);

                 ndb_get_symmetry_operation(cell, matrix, vector, symmetry);

                 if (!strcmp(symmetry, "")) {
                      sprintf(buffer, "%s: No symmetry operator space group=%s\n",
                                          structureID, cell->space_group);
                      message += buffer;
                      for (i = 0; i < 3; i++) {
                           sprintf(buffer, "	%8.5f %8.5f %8.5f   %8.5f\n", matrix[i][0],
                                    matrix[i][1], matrix[i][2], vector[i]);
                           message += buffer;
                      }
                      _symmetry_chain.Clear();
                      _biol_unit.Clear();
                      return;
                 } else if (strcmp(symmetry, "1_555")) {
                      SymIdx.set_svalue(symmetry);
                      if (_symmetry_chain.Seek(SymIdx) >= 0) {
                            SymPtr = _symmetry_chain.GetCurrDataPtr();
                            if (strcmp(SymPtr->ivalue(), "")) {
                                 if (chain_ids != "") chain_ids += " ";
                                 chain_ids += SymPtr->ivalue();
                            }
                            SymPtr->set_ivalue(chain_ids.Text());
                      } else {
                            SymIdx.set_ivalue(chain_ids.Text());
                            _symmetry_chain.Add(SymIdx);
                      }
                 }
            } else if (is_chain_id_line) {
                 if (!strstr(line, "NULL")) {
                      q = line;
                      str_len = strlen(q);
                      for (i = 0; i < str_len; i++) {
                           if (q[i] == ',') q[i] = ' ';
                      }
                      while (q) {
                           q = get_word(q, buffer);
                           if (strcmp(buffer, "")) {
                                if (chain_ids != "") chain_ids += " ";
                                chain_ids += buffer;
                           }
                      }
                 }
            }
       }
}

static char *get_line_from_string(char *line, char *string)
{
       *string = '\0';
       while ((*line == ' ' || *line == '\n') && *line != '\0') line++;
       if (*line == '\0') return NULL;
       while (*line != '\n' && *line != '\0') {
            *string = *line;
            string++; line++;
       }
       *string = '\0';
       while ((*line == ' ' || *line == '\n') && *line != '\0') line++;
       if (*line == '\0') return NULL;
       return line;
}

static char *get_chain_ids(const int entity_id, ISTable *Table)
{
       int i, entity_index, id_index, rowNo;
       char *value = NULL;
       CifString cifstring, chainids;

       chainids.Clear();
       if (!is_empty_table(Table)) {
            entity_index = Table->GetColumnIndex("entity_id");
            id_index = Table->GetColumnIndex("id");
            rowNo = Table->GetNumRows();
            for (i = 0; i < rowNo; i++) {
                 get_value(cifstring, Table, entity_index, i);
                 if (cifstring != "" && atoi(cifstring.Text()) == entity_id) {
                      get_value(cifstring, Table, id_index, i);
                      if (cifstring != "") {
                           if (chainids != "") chainids += ", ";
                           chainids += cifstring;
                      }
                 }
            }
       }
       value = new char[strlen(chainids.Text())+1];
       strcpy(value, chainids.Text());
       return value;
}

