/*
FILE:     _fileobj_ndb_pdb_iolib.C
*/
/*
VERSION:  1.700
*/
/*
DATE:     7/13/2005
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2005 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include "_fileobj.h"
#include "_fileobj_ioutil.h"
#include "erf_iolib.h"
#include "erf_util.h"
#include "ndb_misclib.h"
#include "util.h"

typedef struct {
        char chnid[5];
        int  entity_id;
        CifString seq;
} A1_SEQ;

static void change_to_one_letter_code_sequences(char *line, int &num_seq, A1_SEQ ***seq);

int _FileObj::read_ndb_file(const int adit_view)
{
       char line[512];
       FILE *infile = fopen(_real_filename, "r");
       if (infile == (FILE *) NULL) {
            ndb_log_message(NDB_MSG_ERR, "Can not read file %s!\n", _input_filename);
            return (FALSE);
       }

       int format = NDB_FILE_FORMAT_NDB;
       while (!feof(infile)) {
            fgets(line, 511, infile);
            if (feof(infile)) break;
            ndb_clean_string(line);
            if (line[0] == '\0') continue;
            else if (line[0] == '#' ||
               !strncmp(line, "data_", 5)) {
                 format = NDB_FILE_FORMAT_NDBCIF;
                 break;
            } else if (!strncmp(line, "HEADER", 6) ||
                 !strncmp(line, "TITLE", 5) ||
                 !strncmp(line, "COMPND", 6) ||
                 !strncmp(line, "NDBFIL", 6)) break;
       }
       fclose (infile);

       if (format == NDB_FILE_FORMAT_NDB)
            return read_ndb_pdb_file(format, 0, adit_view);
       else return read_ndb_cif_file();
}

int _FileObj::read_ndb_pdb_file(const int input_format, const int is_encapsulated, const int adit_view)
{
       int num_model = 0, model[10000], i, j, found_hydrogen_atoms = FALSE;
       int PrevTokenId = -1, PrevJrnlId = -1, JrnlId = -1;
       int num_seq = 0, lineno = 0, JrnlSerialNo = -1, prevJrnlSerialNo = -1;
       int first_atom = 1, new_chain = 0, TokenId, isatomCard, is_new, t;
       int length, ret, entity_id, row, *entity = NULL, card_id;
       int atomSerialNo, WaterSerialNo, is_pdb_format;
       long backward;
       char CardString[MxStrLen], PrevSeqNo[MxStrLen], SeqNo[MxStrLen];
       char FieldAnswer[MxStrLen], FieldInfo[MxField][MxLen];
       char prevResidueNo[MxStrLen], prevChainId[MxStrLen];
       char prevResidueName[MxStrLen], prevInsCode[MxStrLen];
       char line[200], buffer[210], *p;
       CifString string, compnd, string1;
       _RESIDUE *residue = NULL;
       A1_SEQ **seq = NULL;
       ISTable *entity_poly = NULL, *table = NULL;
       One_Card_Info_Format *pCard = NULL;

       FILE *infile = fopen(_real_filename, "r");
       if (infile == (FILE *) NULL) {
            ndb_log_message(NDB_MSG_ERR, "Can not read file %s!\n", _input_filename);
            return (FALSE);
       }
       for (i = 0; i < MxField; i++) memset(FieldInfo[i], 0, MxLen);

       strcpy(PrevSeqNo, "xxx");
       strcpy(SeqNo, "");
       strcpy(prevResidueNo, "xxx");
       strcpy(prevChainId, "xxx");
       strcpy(prevResidueName, "xxx");
       strcpy(prevInsCode, "xxx");

       is_pdb_format = 0;
       if (input_format == NDB_FILE_FORMAT_PDB ||
           input_format == NDB_FILE_FORMAT_MIX) is_pdb_format = 1;

       string.Clear();
       compnd.Clear();
       if (_molecules) {
            for (i = 0; i < _num_mol; i++)
                 if (_molecules[i]) delete _molecules[i];
            delete [] _molecules;
       }
       _num_mol = 0;
       _alloc_mol = 0;
       _molecules = NULL;
       do {
            TokenId = get_card(infile, CardString, lineno);
            if (TokenId == -2) break;

            if (is_encapsulated) {
                 if (!strcmp(CardString, "ATOM") ||
                     !strcmp(CardString, "HETATM") ||
                     !strcmp(CardString, "MODEL")) {
                      put_card(PrevTokenId, PrevJrnlId, FieldInfo);
                      break;
                 } else if (!strcmp(CardString, "SEQRES") ||
                            !strcmp(CardString, "COMPND") ||
                            !strcmp(CardString, "SOURCE") ||
                            !strcmp(CardString, "HELIX") ||
                            !strcmp(CardString, "SHEET") ||
                            !strcmp(CardString, "TURN") ||
                            !strcmp(CardString, "LINK") ||
                            !strcmp(CardString, "HYDBND") ||
                            !strcmp(CardString, "MODRES") ||
                            !strcmp(CardString, "SSBOND") ||
                            !strcmp(CardString, "SITE") ||
                            !strcmp(CardString, "SLTBRG")) {
                      backward = get_one_line(line, infile);
                      line[66] = '\0';
                      if (!strcmp(CardString, "SEQRES")) {
                           if (is_encapsulated == 2) {
                                pCard = _CardInfo[SEQRES_TOKEN-1].add_a_new_card(NdbToken[SEQRES_TOKEN-1], 0);
                                for (i = 0; i < 2; i++) buffer[i] = line[2 + i];
                                buffer[i] = '\0'; ndb_clean_string(buffer);
                                pCard->AddElementToField(buffer, 1);
                                for (i = 0; i < 1; i++) buffer[i] = line[5 + i];
                                buffer[i] = '\0'; ndb_clean_string(buffer);
                                pCard->AddElementToField(buffer, 2);
                                for (i = 0; i < 4; i++) buffer[i] = line[7 + i];
                                buffer[i] = '\0'; ndb_clean_string(buffer);
                                pCard->AddElementToField(buffer, 3);
                                for (j = 0; j < 13; j++) {
                                     for (i = 0; i < 4; i++) buffer[i] = line[13 + j * 4 + i];
                                     buffer[i] = '\0'; ndb_clean_string(buffer);
                                     pCard->AddElementToField(buffer, 4 + j);
                                }
                           }
                           change_to_one_letter_code_sequences(line, num_seq, &seq);
                           continue;
                      } else if (!strcmp(CardString, "COMPND") ||
                                 !strcmp(CardString, "SOURCE")) {
                           sprintf(buffer, "%-6s%s\n", CardString, line);
                           compnd += buffer;
                           backward = (-1) * backward;
                           fseek(infile, backward, 1);
                      } else {
                           sprintf(buffer, "%-6s%s\n", CardString, line);
                           string += buffer;
                           continue;
                      }
                 }
            }

            isatomCard = 0;
            if (is_atom_card(TokenId) || !strcmp(CardString, "ATOM") ||
                !strcmp(CardString, "HETATM")) {
                 if (TokenId == -1) TokenId = ATOMN_TOKEN - 1;
                 isatomCard = 1;
                 if (first_atom) {
                      first_atom = 0;
                      if (!_molecules) _ManageSpace();
                 }
            }

            if (PrevTokenId != TokenId) strcpy(PrevSeqNo,"xxx");

            if (!strcmp(CardString, "MODEL")) {
                 put_card(PrevTokenId, PrevJrnlId, FieldInfo);
                 is_new = 0;
                 get_one_line_info(infile, TokenId, PrevTokenId, PrevSeqNo,
                      JrnlId, PrevJrnlId, CardString, FieldInfo, JrnlSerialNo,
                      prevJrnlSerialNo, input_format, is_new);

                 _ManageSpace();
                 _molecules[_num_mol-1]->set_Mol_ID(atoi(FieldInfo[1]));
                 for (t = 0; t < num_model; t++) {
                      if (model[t] == _molecules[_num_mol-1]->Mol_ID()) break;
                 }
                 if (t == num_model) {
                      model[t] = _molecules[_num_mol-1]->Mol_ID();
                      num_model++;
                 } else ndb_log_message(NDB_MSG_ERR, "%s: duplicate MODEL number %d\n",
                            _input_filename, _molecules[_num_mol-1]->Mol_ID());

                 strcpy(prevResidueNo, "xxx");
                 strcpy(prevChainId, "xxx");
                 strcpy(prevResidueName, "xxx");
                 strcpy(prevInsCode, "xxx");
                 PrevTokenId = TokenId;
                 continue;
            } else if (strcmp(CardString,"TER") == 0) {
                 if (residue) {
                      if (input_format == NDB_FILE_FORMAT_PDB ||
                          input_format == NDB_FILE_FORMAT_MIX) {
                           if (residue->change_nomenclature(input_format, 1, 0, 0))
                                found_hydrogen_atoms = TRUE;
                           if (!is_a_standard_residue(residue->ResName()))
                                residue->correction_atom_name(_error_message_s);
                      }
                      _molecules[_num_mol-1]->insert_a_residue(new_chain, residue,
                                           _error_message, _unrecognized_residue);
                      residue = NULL;
                 }
                 new_chain = 1;
                 end_line(infile);
                 continue;
            } else if (strcmp(CardString,"ENDMDL") == 0) {
                 if (residue) {
                      if (input_format == NDB_FILE_FORMAT_PDB ||
                          input_format == NDB_FILE_FORMAT_MIX) {
                           if (residue->change_nomenclature(input_format, 1, 0, 0))
                                found_hydrogen_atoms = TRUE;
                           if (!is_a_standard_residue(residue->ResName()))
                                residue->correction_atom_name(_error_message_s);
                      }
                      _molecules[_num_mol-1]->insert_a_residue(new_chain, residue,
                                           _error_message, _unrecognized_residue);
                      residue = NULL;
                 }
                 end_line(infile);
                 continue;
            } else if (TokenId == -1) {
                 end_line(infile);
                 continue;
            }
     
            /* 
             * This part to identity REMARK 1 ==> JRNL 
             */

            if (input_format == NDB_FILE_FORMAT_PDB) {
                 if (PrevTokenId == JRNL_TOKEN -1 && (TokenId != 
                     REMARK_TOKEN-1 && TokenId != JRNL_TOKEN-1))
     	              JrnlSerialNo = -1;
                 if (strcmp(CardString, "REMARK") == 0 && 
                     TokenId == REMARK_TOKEN-1) {
     	              ret = get_ndbfile_field(infile, TokenId, 1, 
                               FieldAnswer, length);
     	              backward = length;
     	              if (ret == TRUE && strcmp(FieldAnswer, "1") == 0) {
     	                   ret = get_ndbfile_field(infile, TokenId, 2, 
                                    FieldAnswer, length); 
     	                   if (strcmp(FieldAnswer, ""))
                                backward += length;
                           else backward = 1;
     	                   if (strncmp("REFE", FieldAnswer, 4) == 0) {
                                if (JrnlSerialNo == -1) {
                                     for (i = 0; i < MxJrnlTokens; i++) {
                                          if (_JrnlCardInfo[i].Head()) break;
                                     }
                                     if (i < MxJrnlTokens) JrnlSerialNo = 1;
                                     else JrnlSerialNo = 0;
                                } else JrnlSerialNo++;
     	                   }
     	                   TokenId = JRNL_TOKEN - 1;
     	                   strcpy(CardString, "JRNL");
     	              } else if (ret == FALSE) backward = 1;
     	              backward = (-1) * backward;
     	              fseek(infile, backward, 1);
                 }
            }
     
            /* 
             * Token of previous line <> Token of current line
             */

            is_new = FALSE;
            if (feof(infile) || PrevTokenId == REMARK_TOKEN-1 || 
                TokenId == PREMRK_TOKEN-1 || TokenId == PCOMPN_TOKEN-1 ||
                TokenId == PSOURC_TOKEN-1 || PrevTokenId != -1 && 
                PrevTokenId != JRNL_TOKEN-1 && TokenId != PrevTokenId ||
                PrevTokenId == JRNL_TOKEN-1 && TokenId != JRNL_TOKEN-1) {
                 put_card(PrevTokenId, PrevJrnlId, FieldInfo);
                 is_new = TRUE;
                 strcpy(FieldInfo[0], CardString); 
            }

            if (feof(infile)) break;

            get_one_line_info(infile, TokenId, PrevTokenId, PrevSeqNo, 
                   JrnlId, PrevJrnlId, CardString, FieldInfo, JrnlSerialNo,
                   prevJrnlSerialNo, input_format, is_new);

            if (isatomCard) {
                 if (strcmp(FieldInfo[5], prevChainId) ||
                     strcmp(FieldInfo[6], prevResidueNo) ||
                     strcmp(FieldInfo[4], prevResidueName) ||
                     strcmp(FieldInfo[7], prevInsCode)) {
                      if (is_atom_card(TokenId) & AUX_ATOM) {
                           atomSerialNo = atoi(FieldInfo[1]);
                           print_atom_card(buffer, TokenId, TokenId, atomSerialNo,
                                WaterSerialNo, FieldInfo, NDB_FILE_FORMAT_PDB, 0, 0);
                           ndb_log_message(NDB_MSG_ERR,
                              "No coordinate record for following auxiliary record:\n%s\n",
                              buffer);
                           exit (1);
                      }
                      if (residue) {
                           if (input_format == NDB_FILE_FORMAT_PDB ||
                               input_format == NDB_FILE_FORMAT_MIX) {
                                residue->change_nomenclature(input_format, 1, 0, 0);
                                if (!is_a_standard_residue(residue->ResName()))
                                     residue->correction_atom_name(_error_message_s);
                           }
                           _molecules[_num_mol-1]->insert_a_residue(new_chain,
                                      residue, _error_message, _unrecognized_residue);
                           residue = NULL;
                           new_chain = 0;
                      }
                      residue = new _RESIDUE;
                      residue->set_ResName(FieldInfo[4]);
                      residue->set_res_no(atoi(FieldInfo[6]));
                      if (strcmp(FieldInfo[15], ""))
                           residue->set_pdb_res_no(atoi(FieldInfo[15]));
                      else residue->set_pdb_res_no(residue->res_no());
                 }
                 if (residue) residue->insert_a_atom(TokenId , FieldInfo, is_pdb_format, _error_message);

                 strcpy(prevChainId, FieldInfo[5]);
                 strcpy(prevResidueNo, FieldInfo[6]);
                 strcpy(prevResidueName, FieldInfo[4]);
                 strcpy(prevInsCode, FieldInfo[7]);

                 for (i = 0; i < 23; i++) FieldInfo[i][0] = '\0';
            }
    
            if (TokenId == EOF || !strcmp(CardString, "END")) {
                 if (residue) {
                      if (input_format == NDB_FILE_FORMAT_PDB ||
                          input_format == NDB_FILE_FORMAT_MIX) {
                           if (residue->change_nomenclature(input_format, 1, 0, 0))
                                found_hydrogen_atoms = TRUE;
                           if (!is_a_standard_residue(residue->ResName()))
                                residue->correction_atom_name(_error_message_s);
                      }
                      _molecules[_num_mol-1]->insert_a_residue(new_chain, residue,
                                     _error_message, _unrecognized_residue);
                      residue = NULL;
                 }
                 break;
            }
    
            if (TokenId + 1 == JRNL_TOKEN) PrevJrnlId = JrnlId;
            else if (NdbToken[TokenId].SeqField &&
                     strcmp(FieldInfo[NdbToken[TokenId].SeqField-1], ""))
                 strcpy(PrevSeqNo, FieldInfo[NdbToken[TokenId].SeqField-1]);
            PrevTokenId = TokenId;
            prevJrnlSerialNo = JrnlSerialNo;
       } while (!feof(infile) && strcmp(CardString, "END") != 0);

       pCard = _CardInfo[NDBFIL_TOKEN-1].Head();
       if (!strcmp(_StructureId, "") && pCard && strcmp(pCard->Field(1), ""))
            strcpy(_StructureId, pCard->Field(1));

       if (!strcmp(_StructureId, "")) {
            for (i = strlen(_input_filename) - 1; i >= 0; i--) {
                 if (_input_filename[i] == '/') break;
            }
            strcpy(_StructureId, &_input_filename[i+1]);
            p = strstr(_StructureId, ".");
            if (p) *p = '\0';
            ndb_set_string_to_upper(_StructureId);
       }

       if (residue) {
            if (input_format == NDB_FILE_FORMAT_PDB ||
                 input_format == NDB_FILE_FORMAT_MIX) {
                 if (residue->change_nomenclature(input_format, 1, 0, 0))
                      found_hydrogen_atoms = TRUE;
                 if (!is_a_standard_residue(residue->ResName()))
                      residue->correction_atom_name(_error_message_s);
            }
            _molecules[_num_mol-1]->insert_a_residue(new_chain, residue,
                   _error_message, _unrecognized_residue);
            residue = NULL;
       }

       if (!is_encapsulated) {
            get_sequence_from_seqres();
            for (i = 0; i < _num_mol; i++)
                 _molecules[i]->check_chain_type(_Num_Seqs, _Seqs);
            if (!_Num_Seqs) {
                 get_seqres_from_mapping(NDB_FILE_FORMAT_NDB);
                 get_sequence_from_seqres();
            }
       }

       if (is_encapsulated) {
            if (num_seq) {
                 compnd.Clear();
                 entity_id = 1;
                 seq[0]->entity_id = entity_id;
                 for (i = 1; i < num_seq; i++) {
                      for (j = 0; j < i; j++) {
                           if (seq[i]->seq == seq[j]->seq) {
                                seq[i]->entity_id = seq[j]->entity_id;
                                break;
                           }
                      }
                      if (j == i) {
                           entity_id++;
                           seq[i]->entity_id = entity_id;
                      }
                 }
                 entity_poly = new ISTable("entity_poly");
                 entity_poly->AddColumn("entity_id");
                 entity_poly->AddColumn("ndb_seq_one_letter_code");
                 entity_poly->AddColumn("ndb_chain_id");
                 row = 0;
                 entity = new int[num_seq];
                 for (i = 0; i < num_seq; i++) entity[i] = 0;
                 entity_id = 0;
                 for (i = 0; i < num_seq; i++) {
                      for (j = 0; j < entity_id; j++) {
                           if (entity[j] == seq[i]->entity_id) break;
                      }
                      if (j == entity_id) {
                           entity[j] = seq[i]->entity_id;
                           entity_id++;
                           entity_poly->AddRow();
                           sprintf(buffer, "%d", entity_id);
                           string1.Copy(buffer);
                           entity_poly->UpdateCell(string1, 0, row);
                           entity_poly->UpdateCell(seq[i]->seq, 1, row);
                           string1.Copy(seq[i]->chnid);
                           entity_poly->UpdateCell(string1, 2, row);
                           row++;
                      } else {
                           entity_poly->GetCell(string1, 2, j);
                           sprintf(buffer, "%s,%s", string1.Text(), seq[i]->chnid);
                           string1.Copy(buffer);
                           entity_poly->UpdateCell(string1, 2, j);
                      }
                 }
                 delete entity;
                 if (!_CifObj) {
                      get_temp_file();
                      _CifObj = new CifFileObj(_temp_file, WRITE_MODE);
                 }
                 if (entity_poly) _CifObj->WriteTable(entity_poly, _StructureId, "entity_poly");
            } else {
                 pCard = _CardInfo[COMPND_TOKEN-1].Head();
                 while (pCard) pCard = _CardInfo[COMPND_TOKEN-1].delete_a_card(pCard);
                 pCard = _CardInfo[SOURCE_TOKEN-1].Head();
                 while (pCard) pCard = _CardInfo[SOURCE_TOKEN-1].delete_a_card(pCard);
            }
            if (is_encapsulated != 2) {
                 TokenId = -1; atomSerialNo = 0; WaterSerialNo = 0;
                 do {
                      if (!strcmp(CardString, "ATOM") ||
                          !strcmp(CardString, "HETATM") ||
                          !strcmp(CardString, "SIGATM") ||
                          !strcmp(CardString, "ANISOU") ||
                          !strcmp(CardString, "SIGUIJ")) {
                           if (TokenId == -1) TokenId = ATOMN_TOKEN-1;
                           for (i = 1; i < NdbToken[TokenId].NumField; i++) {
                                FieldAnswer[0] = '\0';
                                if (NdbToken[TokenId].FieldList[i].FieldStcol >= 72) {
                                     end_line(infile);
                                     break;
                                } else ret = get_ndbfile_field(infile, TokenId, i,
                                                FieldAnswer, length);
                                strcpy(FieldInfo[i], FieldAnswer);
                                if (ret == FALSE) break;
                           }
                           strcpy(FieldInfo[0], CardString);
                           strcpy(FieldInfo[15], FieldInfo[6]);
                           strcpy(FieldInfo[19], FieldInfo[6]);
                           strcpy(FieldInfo[16], FieldInfo[4]);
                           strcpy(FieldInfo[20], FieldInfo[4]);
                           strcpy(FieldInfo[17], FieldInfo[5]);
                           strcpy(FieldInfo[21], FieldInfo[5]);
                           strcpy(FieldInfo[18], FieldInfo[2]);
                           strcpy(FieldInfo[22], FieldInfo[2]);
                           print_atom_card(buffer, ATOMN_TOKEN-1, TokenId, atomSerialNo,
                               WaterSerialNo, FieldInfo, NDB_FILE_FORMAT_MIX, 0, 0);
                           string += buffer;
                           string += "\n";
                      } else {
                           get_one_line(line, infile);
                           sprintf(buffer, "%-6s%s\n", CardString, line);
                           string += buffer;
                      }
                      TokenId = get_card(infile, CardString, lineno);
                      if (TokenId == -2) break;
                 } while (!feof(infile) && strcmp(CardString, "END") != 0);
                 if (!_CifObj) {
                      get_temp_file();
                      _CifObj = new CifFileObj(_temp_file, WRITE_MODE);
                 }
                 table = new ISTable("ndb_original_ndb_coordinates");
                 table->AddColumn("coord_section");
                 table->AddRow();
                 if (compnd != "") {
                      compnd += string;
                      table->AddElementToRow(compnd, 0);
                 } else table->AddElementToRow(string, 0);
                 _CifObj->WriteTable(table, _StructureId, "ndb_original_ndb_coordinates");
            }
       }
       fclose (infile);

       if (found_hydrogen_atoms) {
            card_id = find_card_id("COORDH");
            pCard = _CardInfo[card_id].Head();
            if (pCard == NULL)
                 pCard = _CardInfo[card_id].add_a_new_card(NdbToken[card_id], 0);
            pCard->AddElementToField("Y", 1);
       }

       pCard = _CardInfo[EXPDTA_TOKEN-1].Head();
       if (pCard) {
            string.Copy(pCard->Field(2));
            ndb_set_string_to_upper(string.Text());
            if (strcmp(string.Text(), ""))
                 _experiment_type = EXPERIMENT_TYPE_BASIC;
            if (strstr(string.Text(), "X-RAY DIFFRACTION") ||
                strstr(string.Text(), "X-RAY POWDER DIFFRACTION"))
                 _experiment_type = EXPERIMENT_TYPE_XRAY;
            else if (strstr(string.Text(), "NMR"))
                 _experiment_type = EXPERIMENT_TYPE_NMR;
            else if (strstr(string.Text(), "SOLUTION SCATTERING"))
                 _experiment_type = EXPERIMENT_TYPE_SOLN_SCT;
            else if (strstr(string.Text(), "THEORETICAL MODEL"))
                 _experiment_type = EXPERIMENT_TYPE_MODEL;
            else if (strstr(string.Text(), "NEUTRON DIFFRACTION"))
                 _experiment_type = EXPERIMENT_TYPE_NEUTRON;
            else if (strstr(string.Text(), "ELECTRON DIFFRACTION"))
                 _experiment_type = EXPERIMENT_TYPE_ELECTRON;
            else if (strstr(string.Text(), "CRYO-ELECTRON MICROSCOPY") ||
                     strstr(string.Text(), "ELECTRON TOMOGRAPHY") ||
                     strstr(string.Text(), "ELECTRON MICROSCOPY"))
                 _experiment_type = EXPERIMENT_TYPE_CRYO_EM;
            else if (strstr(string.Text(), "FIBER DIFFRACTION"))
		 _experiment_type = EXPERIMENT_TYPE_FIBER;
       } else if (adit_view) {
            if (adit_view == 1) _experiment_type = EXPERIMENT_TYPE_XRAY;
            else if (adit_view == 2) _experiment_type = EXPERIMENT_TYPE_NMR;
            else if (adit_view == 4) _experiment_type = EXPERIMENT_TYPE_CRYO_EM;
       } else _experiment_type = EXPERIMENT_TYPE_XRAY;
       get_cell();

       for (i = 0; i < num_seq; i++)
            if (seq[i]) delete seq[i];
       if (seq) delete [] seq;

       return 1;
}

void _FileObj::print_general_cards(FILE *fp, const int card_id, const int format,
                  int &numRemark, const int is_star)
{
       int IsJRNL = FALSE, prevSeqNo = 0, ContNo = 1, PdbContNo = 1, pos, field_no;
       int /* remark_no = 0, */ resField, width, str_len, Stcol, i, found, start = 0, len;
       double temp_double;
       char temp_line[MxLen], FieldFormat[MxStrLen], temp_line2[MxLen];

       if (format == NDB_FILE_FORMAT_PDB && NdbToken[card_id].NdbOnly) return;

       One_Card_Info_Format *pCard = _CardInfo[card_id].Head();
       while (pCard) {
            if (ContNo != 1) strcpy(temp_line2, temp_line);
            else sprintf(temp_line, "%-6s", pCard->Field(0));

            pos = 6;
            // remark_no = 0;
            // if (card_id == REMARK_TOKEN-1) remark_no = atoi(pCard->Field(1));
            for (field_no = 1; field_no < NdbToken[card_id].NumField; field_no++) {

                 if (NdbToken[card_id].FieldList[field_no].NdbOnly &&
                     format == NDB_FILE_FORMAT_PDB) continue;

                 width = NdbToken[card_id].FieldList[field_no].FieldWidth;
                 print_space_between_field(card_id, field_no, temp_line, pos, IsJRNL);
                 get_field_format(card_id, field_no, FieldFormat,IsJRNL, format);
                 str_len = strlen(NdbToken[card_id].FieldList[field_no].FieldLabel);

                 if (format == NDB_FILE_FORMAT_PDB && 
                            NdbToken[card_id].FieldList[field_no].NdbOnly) {
                      sprintf(FieldFormat, "%%%ds", width); 
                      sprintf(&temp_line[pos], FieldFormat, "");
                 } else if (ContNo != 1 && NdbToken[card_id].ConField != field_no + 1 
                      && NdbToken[card_id].ContinuedF != field_no +1 &&
                         NdbToken[card_id].PdbConField != field_no +1) {
                      Stcol = NdbToken[card_id].FieldList[field_no].FieldStcol;
                      for (i = 0; i < width; i++)
                           temp_line[Stcol + i - 1] = temp_line2[Stcol + i - 1];
                      temp_line[Stcol + i - 1] = '\0';
                      pos = Stcol + width - 1;
                      continue;
                 }
                 if (pCard && pCard->Field(field_no) && (format != NDB_FILE_FORMAT_PDB || 
                    !NdbToken[card_id].FieldList[field_no].NdbOnly)) {
                      if (/* remark_no != 350 && remark_no != 400 && remark_no != 465 &&
                          remark_no != 470 && remark_no != 500 &&
                         (*/ !NdbToken[card_id].FieldList[field_no].MixedCase ||
                           format == NDB_FILE_FORMAT_PDB && card_id == AUTHOR_TOKEN-1 /* ) */ )
                           pCard->UpperCaseField(field_no);
               
                      if (!strncmp(NdbToken[card_id].FieldList[field_no].FieldLabel,
                            "Atom_Name", 9) ||
                          strstr(NdbToken[card_id].FieldList[field_no].FieldLabel, "atom_id")) {
                           resField = -1;
                           if (card_id == LINK_TOKEN-1 || card_id == SLTBRG_TOKEN-1) {
                                if (field_no == 1) resField = 3;
                                else if (field_no == 7) resField = 9;
                           } else if (card_id == HYDBND_TOKEN-1) {
                                if (field_no == 1) resField = 3;
                                else if (field_no == 12) resField = 14;
                           }
                           if (resField > 0)
                                print_atom_atomName_field("", pos, temp_line,
                                      pCard->Field(field_no), pCard->Field(resField), 1);
                           else print_atom_atomName_field("", pos, temp_line,
                                      pCard->Field(field_no), NULL, 1);
                      } else if (NdbToken[card_id].ConField == field_no + 1) {
                           if (ContNo != 1) 
                                sprintf(&temp_line[pos],FieldFormat, (double) ContNo);
                           else {
                                sprintf(FieldFormat, "%%%ds", width);
                                sprintf(&temp_line[pos], FieldFormat, "");
                           } 
                      } else if (NdbToken[card_id].PdbConField == field_no +1) {
                           if (PdbContNo != 1) 
                                sprintf(&temp_line[pos],FieldFormat,(double)PdbContNo);
                           else {
                                sprintf(FieldFormat, "%%%ds", width);
                                sprintf(&temp_line[pos], FieldFormat,"");
                           } 
                      } else if (NdbToken[card_id].ContinuedF == field_no + 1) 
                           print_ndb_continuedf_field(card_id, pCard, pos, start,
                                   temp_line, FieldFormat, ContNo, PdbContNo);
                      else if (!strcmp(pCard->Field(field_no), "") || ContNo != 1) {
                           sprintf(FieldFormat, "%%%ds", width);
                           sprintf(&temp_line[pos], FieldFormat, "");
                      } else if (NdbToken[card_id].FieldList[field_no].FieldType<=2) {
                           temp_double = atof(pCard->Field(field_no));
                           sprintf(&temp_line[pos], FieldFormat, temp_double);
                           found = FALSE;
                           len = strlen(pCard->Field(field_no));
                           for (i = 0; i < len; i++) {
                                if (isalpha(pCard->Field(field_no)[i])) {
                                     found = TRUE; break;
                                }
                           }
                           if (found) {
                                if (NdbToken[card_id].FieldList[field_no].FieldType==1)
                                     ndb_log_message(NDB_MSG_ERR, 
                                       "Token %s Field %d suppose to be integer, but got %s",
                                       NdbToken[card_id].TokenName,
                                       field_no, pCard->Field(field_no));
                                else if(NdbToken[card_id].FieldList[field_no].FieldType==2)
                                     ndb_log_message(NDB_MSG_ERR,
                                       "Token %s Field %d suppose to be float, but got %s",
                                       NdbToken[card_id].TokenName,
                                       field_no, pCard->Field(field_no));
                                for (i = 0; i < width; i++) temp_line[pos+i] = ' ';
                           }
                      } else {
                           if (NdbToken[card_id].FieldList[field_no].FieldWidth == 1 &&
                               strlen(pCard->Field(field_no)) > 1 &&
                               pCard->Field(field_no)[0] == '+')
                                sprintf(&temp_line[pos], FieldFormat,
                                    &pCard->Field(field_no)[1]);
                           else sprintf(&temp_line[pos], FieldFormat,
                                    pCard->Field(field_no));
                      }
                      pos += NdbToken[card_id].FieldList[field_no].FieldWidth;
                      if (field_no == NdbToken[card_id].SeqField - 1) {
                           if ((card_id == REMARK_TOKEN-1) && 
                               prevSeqNo != atoi(pCard->Field(field_no))) {
                                if (pCard->Field(NdbToken[card_id].ContinuedF-1) &&
                                  strcmp(pCard->Field(NdbToken[card_id].ContinuedF-1), "") !=0) {
                                     fprintf(fp, "%-11s", temp_line);
                                     if (format == NDB_FILE_FORMAT_PDB) {
                                          fprintf(fp, "                    ");
                                          fprintf(fp, "                    ");
                                          fprintf(fp, "                             \n");
                                     } else fprintf(fp, "\n");
                                     numRemark++;
                                }
                                prevSeqNo = atoi(pCard->Field(field_no));
                           } else if (card_id != REMARK_TOKEN-1 && 
                                      card_id != PREMRK_TOKEN-1 &&
                                      card_id != PCOMPN_TOKEN-1 &&
                                      card_id != PSOURC_TOKEN-1)
                                prevSeqNo++;
                      }
                 } else {
                      if (field_no == NdbToken[card_id].SeqField -1 && 
                          card_id != REMARK_TOKEN-1 && card_id != PREMRK_TOKEN-1 &&
                          card_id != PCOMPN_TOKEN-1 && card_id != PSOURC_TOKEN-1)
                           prevSeqNo++;
                      sprintf(FieldFormat,"%%%-ds", width);
                      print_space_between_field(card_id, field_no, temp_line, pos,
                                         IsJRNL);
                      sprintf(&temp_line[pos], FieldFormat, "");
                      pos += width;
                 }
            }
            if (NdbToken[card_id].PdbConField != 0) PdbContNo++;
            str_len = strlen(temp_line);
            if (format == NDB_FILE_FORMAT_PDB && str_len < 80) {
                 for (i = 0; i < 80 - str_len; i++)
                      strcat(temp_line, " ");
            }
            if (is_star) fprintf(fp, "*%s\n", temp_line);
            else fprintf(fp, "%s\n", temp_line);
            fflush(fp);
            if (card_id == REMARK_TOKEN-1) numRemark++;
            if (NdbToken[card_id].ContinuedF != 0 && 
               pCard->Field(NdbToken[card_id].ContinuedF-1) &&
	       start < (int) strlen(pCard->Field(NdbToken[card_id].ContinuedF-1))) {
                 ContNo++;
                 continue;
            } else ContNo = 1;
        
            pCard = pCard->Next();
            start = 0;
       }
}

void _FileObj::print_jrnl_cards(FILE *fp, const int format, int &numRemark)
{
       int vField = 5, astmField = 3, issnField = 6, IsJRNL = TRUE, jrnl_id;
       int remark_1_flag = 1, remark_printed = FALSE, JrnlSerialNo = 0, ContNo;
       int oldJrnlSerialNo = 0, maxJrnlSerialNo = 0, i, j, displayOrder, pos;
       int str_len, field_no, start, is_unpublished = 0;
       double temp_double;
       char temp_line[MxLen], FieldFormat[MxStrLen];
       One_Card_Info_Format *pJrnlCard = NULL, *HeadHold[10];

       if (_JrnlCardInfo == NULL) return;
       jrnl_id = find_jrnl_id("REF");
       pJrnlCard = _JrnlCardInfo[jrnl_id].Head();
       if (pJrnlCard && !strcasecmp(pJrnlCard->Field(4), "Unpublished")) is_unpublished = 1;
       while (pJrnlCard) {
            if (strcmp(pJrnlCard->Field(vField+1), "") &&
               !strcmp(pJrnlCard->Field(vField), ""))
                 pJrnlCard->AddElementToField("V.", vField);
            else if (!strcmp(pJrnlCard->Field(vField+1), ""))
                 pJrnlCard->AddElementToField("", vField);
            pJrnlCard = pJrnlCard->Next();
       }

       jrnl_id = find_jrnl_id("REFN");
       pJrnlCard = _JrnlCardInfo[jrnl_id].Head();
       while (pJrnlCard) {
            if (strcmp(pJrnlCard->Field(astmField+1), "") &&
               !strcmp(pJrnlCard->Field(astmField), ""))
                 pJrnlCard->AddElementToField("ASTM", astmField);
            else if (!strcmp(pJrnlCard->Field(astmField+1), ""))
                 pJrnlCard->AddElementToField("", astmField);

            if (strcmp(pJrnlCard->Field(issnField+1), "") &&
               !strcmp(pJrnlCard->Field(issnField), ""))
                 pJrnlCard->AddElementToField("ISSN", issnField);
            else if (!strcmp(pJrnlCard->Field(issnField+1), ""))
                 pJrnlCard->AddElementToField("", issnField);
            pJrnlCard = pJrnlCard->Next();
       }

       for (i = 0; i < MxJrnlTokens; i++) {
            pJrnlCard = _JrnlCardInfo[i].Head();
            while (pJrnlCard) {
                 if (atoi(pJrnlCard->Field(1)) > maxJrnlSerialNo)
                      maxJrnlSerialNo = atoi(pJrnlCard->Field(1));
                 pJrnlCard = pJrnlCard->Next();
            }
       }
       for (i = 0; i < MxJrnlTokens; i++)
            HeadHold[i] = _JrnlCardInfo[i].Head();

       while (TRUE) {
            for (displayOrder = 0; displayOrder < MxJrnlTokens; displayOrder++) {
                 for (j = 0; j < MxJrnlTokens; j++) 
                        if (displayOrder == JrnlToken[j].DisplayOrder) break;
                 if (_JrnlCardInfo[j].Head()) break; 
            }

            if (displayOrder == MxJrnlTokens || JrnlSerialNo > maxJrnlSerialNo) break;
     
            ContNo = 1;
            start = 0;
            for (displayOrder = 0; displayOrder < MxJrnlTokens; displayOrder++) {
                 for (j = 0; j < MxJrnlTokens; j++) 
                      if (displayOrder == JrnlToken[j].DisplayOrder) break;
                 if (j == MxJrnlTokens) continue;

                 pJrnlCard = _JrnlCardInfo[j].Head();
                 if (pJrnlCard == NULL || atoi(pJrnlCard->Field(1)) != JrnlSerialNo)
                      continue;

                 if (format == NDB_FILE_FORMAT_PDB && ContNo == 1 &&
                     JrnlSerialNo != oldJrnlSerialNo) {
                      oldJrnlSerialNo = JrnlSerialNo;
                      remark_1_flag = 0;
                      sprintf(temp_line, "%-6s", "REMARK");
                      pos = 6;
                      get_field_format(REMARK_TOKEN-1, 1, FieldFormat, FALSE, format);
                      print_space_between_field(REMARK_TOKEN-1, 1, temp_line, pos,
                                    FALSE);
                      sprintf(&temp_line[pos], FieldFormat, (double) 1);
                      if (remark_printed == FALSE) {
                           str_len = strlen(temp_line);
                           if (str_len < 80) {
                                for (i = 0; i < 80 - str_len; i++)
                                     strcat(temp_line, " ");
                           }
                           fprintf(fp, "%s\n", temp_line);
                           numRemark++;
                           remark_printed = TRUE;
                      }
                      pos += NdbToken[REMARK_TOKEN-1].FieldList[1].FieldWidth;
            
                      print_space_between_field(REMARK_TOKEN-1, 2, temp_line, pos,
                                       FALSE);
                      get_field_format(REMARK_TOKEN-1, 2, FieldFormat, FALSE, format);
                      sprintf(&temp_line[pos], "REFERENCE %d", JrnlSerialNo);
                      str_len = strlen(temp_line);
                      if (str_len < 80) {
                           for (i = 0; i < 80 - str_len; i++)
                                strcat(temp_line, " ");
                      }
                      fprintf(fp, "%s\n", temp_line);
                      numRemark++;
                 } else if (format == NDB_FILE_FORMAT_PDB && JrnlSerialNo != 0)
                      sprintf(temp_line, "%-6s", "REMARK");
                 else sprintf(temp_line, "%-6s",pJrnlCard->Field(0));

                 pos = 6;
                 for (field_no = 1; field_no < JrnlToken[j].NumField; field_no++) {
                      if (JrnlToken[j].FieldList[field_no].NdbOnly &&
                                format == NDB_FILE_FORMAT_PDB) continue;

                      if (format == NDB_FILE_FORMAT_PDB)
                           pJrnlCard->UpperCaseField(field_no);
                      print_space_between_field(j, field_no, temp_line, pos, IsJRNL);
                      get_field_format(j, field_no, FieldFormat, IsJRNL, format);
                      if (field_no == 1) {
                           if (JrnlSerialNo != 0) {
                                if (format == NDB_FILE_FORMAT_PDB)
                                     sprintf(&temp_line[pos], FieldFormat, 1);
                                else sprintf(&temp_line[pos], FieldFormat, JrnlSerialNo);
                           } else {
                                sprintf(FieldFormat, "%%%ds", 
                                       JrnlToken[j].FieldList[field_no].FieldWidth);
                                sprintf(&temp_line[pos], FieldFormat, "");
                           }
                      } else if (JrnlToken[j].ConField == field_no + 1) {
                           if (ContNo != 1) 
                                sprintf(&temp_line[pos], FieldFormat, ContNo);
                           else {
                                sprintf(FieldFormat, "%%%ds", 
                                       JrnlToken[j].FieldList[field_no].FieldWidth);
                                sprintf(&temp_line[pos], FieldFormat, "");
                           }
                      } else if (JrnlToken[j].ContinuedF == field_no + 1) 
                           print_jrnl_continuedf_field(j, pJrnlCard, pos, start,
                                                temp_line, FieldFormat);
                      else if (strcmp(pJrnlCard->Field(field_no), "") == 0 ||
                              format == NDB_FILE_FORMAT_PDB && 
                              JrnlToken[j].FieldList[field_no].NdbOnly ||
                              ContNo != 1 && field_no != 2) {
                           sprintf(FieldFormat, "%%%ds", 
                                  JrnlToken[j].FieldList[field_no].FieldWidth);
                           sprintf(&temp_line[pos], FieldFormat,"");
                      } else if (JrnlToken[j].FieldList[field_no].FieldType <= 2) {
                           temp_double = atof(pJrnlCard->Field(field_no));
                           sprintf(&temp_line[pos], FieldFormat, (int) temp_double);
                      } else {
                           sprintf(&temp_line[pos], FieldFormat, pJrnlCard->Field(field_no));
                      }
                      pos += JrnlToken[j].FieldList[field_no].FieldWidth;
/*
                      if (JrnlToken[j].ContinuedF != 0 && 
                          field_no != JrnlToken[j].ContinuedF - 1 && field_no > 3)
                           pJrnlCard->AddElementToField("", field_no);
*/
                 }
                 str_len = strlen(temp_line);
                 if (format == NDB_FILE_FORMAT_PDB && str_len < 80) {
                      for (i = 0; i < 80 - str_len; i++)
                           strcat(temp_line, " ");
                 }
                 if (format == NDB_FILE_FORMAT_PDB && is_unpublished &&
                     !strncmp(temp_line, "JRNL", 4)) ;
                 else {
                      fprintf(fp, "%s\n", temp_line);
                      fflush(fp);
                 }
                 if (!strncmp(temp_line, "REMARK", 6)) numRemark++;

                 if (JrnlToken[j].ContinuedF != 0 && start <
                         (int) strlen(pJrnlCard->Field(JrnlToken[j].ContinuedF-1))) {
                      ContNo++;
                      displayOrder--;
                      continue;
                 } else ContNo = 1;
                 _JrnlCardInfo[j].MoveHead(pJrnlCard->Next());
                 start = 0;
            }
            oldJrnlSerialNo = JrnlSerialNo;
            JrnlSerialNo++;
       }

       for (i = 0; i < MxJrnlTokens; i++) {
            _JrnlCardInfo[i].MoveHead(HeadHold[i]);
       }

       if (format == NDB_FILE_FORMAT_PDB && remark_1_flag) {
            fprintf(fp, "REMARK   1");
            fprintf(fp, "                                                                      \n");
            fflush(fp);
       }
}

void _FileObj::update_tokens(const int format)
{
       int card_id, MatrixNo, SerialNo, len, field_no, width;
       char buffer[20];
       One_Card_Info_Format *pCard = NULL;
       for (card_id = 0; card_id < MxTokens; card_id++) {
            MatrixNo = 1;
            SerialNo = 0;
            pCard = _CardInfo[card_id].Head();
            if (NdbToken[card_id].MatrixType == 0 && NdbToken[card_id].SeqField != 0) {
                 while (pCard) {
                      SerialNo++;
                      if (!strcmp(pCard->Field(NdbToken[card_id].SeqField-1), "")) {
                           pCard->AddElementToField(SerialNo, NdbToken[card_id].SeqField-1);
                      }
                      pCard = pCard->Next();
                 }
            } else if (NdbToken[card_id].MatrixType) {
                 while (pCard) {
                      len = strlen(pCard->Field(0));
		      if (!isdigit(pCard->Field(0)[len-1])) {
			   sprintf(buffer, "%s%d", pCard->Field(0), MatrixNo);
			   pCard->AddElementToField(buffer, 0);
		      }
                      pCard = pCard->Next();
                      if (MatrixNo % 3 == 0) MatrixNo = 1;
                      else MatrixNo++;
                 }
            }
            pCard = _CardInfo[card_id].Head();
            while (pCard) {
                 for (field_no = 1; field_no < NdbToken[card_id].NumField; field_no++) {
                      if (NdbToken[card_id].FieldList[field_no].FieldType == NDB_DATE_FIELD &&
                          strcmp(pCard->Field(field_no), "")) {
                           width = NdbToken[card_id].FieldList[field_no].FieldWidth;
                           if (card_id == DTMEAS_TOKEN-1 && format == NDB_FILE_FORMAT_PDB)
                                width++;
                           ndb_convert_time_format(format, pCard->Field(field_no),
                                     buffer, width);
                           pCard->AddElementToField(buffer, field_no);
                      }
                 }
                 pCard = pCard->Next();
            }
       }
}

static void change_to_one_letter_code_sequences(char *line, int &num_seq, A1_SEQ ***seq)
{
       static int num = 0;
       int i, j;
       char chnid[5], buffer[5];
       A1_SEQ **tmp = NULL;

       sprintf(chnid, "%c", line[5]);
       if (chnid[0] == ' ') chnid[0] = '\0';
       for (i = 0; i < num_seq; i++) {
            if (!strcmp((*seq)[i]->chnid, chnid)) break;
       }
       if (i == num_seq) {
            tmp = *seq;
            num_seq++;
            *seq = new A1_SEQ*[num_seq];
            for (j = 0; j < num_seq - 1; j++)
                 (*seq)[j] = tmp[j];
            if (tmp) delete [] tmp;
            (*seq)[i] = new A1_SEQ;
            (*seq)[i]->entity_id = 0;
            (*seq)[i]->seq.Copy("");
            strcpy((*seq)[i]->chnid, chnid);
            num = 0;
       }
       char *p = &line[13];
       ndb_clean_string(p);
       while (p) {
            p = get_word(p, buffer);
            char *code = get_oneletter_code(buffer);
            int len1 = strlen(code);
            if (len1 == 1) {
                 (*seq)[i]->seq += code;
                 num++;
            } else {
                 (*seq)[i]->seq += '(';
                 (*seq)[i]->seq += code;
                 (*seq)[i]->seq += ')';
                 num += (2 + len1);
            }
            if (num > 70) {
                 (*seq)[i]->seq += '\n';
                 num = 0;
            }
       }
}

