/*
FILE:     _residue.C
*/
/*
VERSION:  1.700
*/
/*
DATE:     7/13/2005
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2005 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "_atom_change.h"
#include "_residue.h"
#include "erf_connect.h"
#include "erf_iolib.h"
#include "erf_util.h"
#include "graph_container.h"
#include "util.h"

typedef struct {
        char atom_name[5];
        char atom_type[5];
        int  changed;
} _NAME_TYPE;

#define FORMAT "\tATOM  %5d %4s%c%-3s %s%4d%c   %8.3f%8.3f%8.3f%6.2f%6.2f\n"

const int _RESIDUE::_ATOM_INCR = 10;
static CHANGE_ATOM_NAME IU_FIRST = {
        1,
         { { "H", "H1" }}
};

static CHANGE_ATOM_NAME IU_LAST = {
        1,
         { { "OXT", "O''" }}
};

static int get_align(char *atomName, char *atom_type);
static void write_error_message(_ATOM *new_atom, _ATOM *old_atom, CifString& error, const int flag);
static void write_occupancy_message(CifString& error, _ATOM *atom, const char *name, const float occupancy);

void _RESIDUE::Clear()
{
       memset(_ResName, 0, 5);
       memset(_alt_loc, 0, 10);
       _res_no = 0;
       _pdb_res_no = 0;
       _Num_Atom = 0;
       _Alloc_Atom = 0;
       _atoms = NULL;
       _deleted = NULL;
       _type = MONOMER_TYPE_UNK;
       _chain_break = 0;
       _no_res_name = 1;
       _wrong_connectivity = 0;
       _atomNames.Clear();
       _atomNameAlts.Clear();
       _atomOrder.Clear();
       _name_list.Copy("");
       _missing.Copy("");
       _extras.Copy("");
       _is_single_atom = 0;
}

void _RESIDUE::Reset()
{
       int i;
       if (_atoms) {
            for (i = 0; i < _Num_Atom; i++)
                 if (_atoms[i]) delete _atoms[i];
            delete [] _atoms;
       }
       if (_deleted) delete [] _deleted;
       Clear();
}

void _RESIDUE::_UpdateIndices()
{
       int i;
       CifString index;
       for (i = 0; i < _Num_Atom; i++) {
            if (_deleted[i]) continue;
            index.Copy(_atoms[i]->restype());
            index += _atoms[i]->pdb_atmnam();
            _atomNames.GoToRecord(i);
            _atomNames.ReplData(index);
            index += "_";
            index += _atoms[i]->alt_loc();
            _atomNameAlts.GoToRecord(i);
            _atomNameAlts.ReplData(index);
       }
}

void _RESIDUE::operator=(const _RESIDUE& a)
{
       Reset();
       strcpy(_ResName, a._ResName);
       strcpy(_alt_loc, a._alt_loc);
       _res_no = a._res_no;
       _pdb_res_no = a._pdb_res_no;
       _type = a._type;
       _name_list = a._name_list;
       _missing = a._missing;
       _extras = a._extras;
       if (a._Num_Atom) {
            _ATOM **atom = new _ATOM*[a._Num_Atom];
            CifString error;
            for (int i = 0; i < a._Num_Atom; i++) {
                 atom[i] = new _ATOM;
                 *(atom[i]) = *(a._atoms[i]);
                 insert_a_atom(atom[i], error);
            }
            delete [] atom;
       }
}

char*     _RESIDUE::ResName()            { return _ResName; }
char*     _RESIDUE::alt_loc()            { return _alt_loc; }
int       _RESIDUE::res_no()             { return _res_no; }
int       _RESIDUE::pdb_res_no()         { return _pdb_res_no; }
int       _RESIDUE::type()               { return _type; }
int       _RESIDUE::chain_break()        { return _chain_break; }
int       _RESIDUE::no_res_name()        { return _no_res_name; }
int       _RESIDUE::wrong_connectivity() { return _wrong_connectivity; }
CifString _RESIDUE::missing()            { return _missing; }
CifString _RESIDUE::extras()             { return _extras; }
CifString _RESIDUE::name_list()          { return _name_list; }
int       _RESIDUE::is_single_atom()     { return _is_single_atom; }
int       _RESIDUE::AtomNumbers()        { return (_atomNames.NumElem()); }

void  _RESIDUE::set_ResName(const char* a)
{ 
       strcpy(_ResName, a);
       if (!strstr(_name_list.Text(), _ResName))
            _name_list.Copy(_ResName);
}
void  _RESIDUE::UpdateIndices()              { _UpdateIndices(); }
void  _RESIDUE::set_alt_loc(const char* a)   { strcpy(_alt_loc, a); }
void  _RESIDUE::set_res_no(const int a)      { _res_no = a; }
void  _RESIDUE::set_pdb_res_no(const int a)  { _pdb_res_no = a; }
void  _RESIDUE::set_type(const int a)        { _type = a; }
void  _RESIDUE::set_chain_break(const int a) { _chain_break = a; }

_ATOM* _RESIDUE::GetFirstAtom()
{
       if (_atomOrder.Root() >= 0) {
            _atomOrder.GoFirst();
            _ISIndex idx = _atomOrder.GetCurrData();
            return _atoms[idx.pos()];
       } else return NULL;
}

_ATOM* _RESIDUE::GetLastAtom()
{
       if (_atomOrder.Root() >= 0) {
            _atomOrder.GoLast();
            _ISIndex idx = _atomOrder.GetCurrData();
            return _atoms[idx.pos()];
       } else return NULL;
}

_ATOM* _RESIDUE::GetPrevAtom()
{
       if (_atomOrder.GoToPrev() >= 0) {
            _ISIndex idx = _atomOrder.GetCurrData();
            return _atoms[idx.pos()];
       } else return NULL;
}

_ATOM* _RESIDUE::GetNextAtom()
{
       if (_atomOrder.GoToNext() >= 0)  {
            _ISIndex idx = _atomOrder.GetCurrData();
            return _atoms[idx.pos()];
       } else return NULL;
}

void _RESIDUE::GetAllBonds(int &num, _ATOM **f_bond, _ATOM **e_bond)
{
       num = 0;
       int i = 0, j = 0;
       for (i = 0; i < _Num_Atom - 1; i++) {
            if (_deleted[i]) continue;
            for (j = i + 1; j < _Num_Atom; j++) {
                 if (_deleted[j]) continue;
                 if ((!strcmp(_atoms[i]->alt_loc(), "") ||
                      !strcmp(_atoms[j]->alt_loc(), "") ||
                      !strcmp(_atoms[i]->alt_loc(),
                              _atoms[j]->alt_loc())) &&
                      is_a_bond(_atoms[i], _atoms[j])) {
                      f_bond[num] = _atoms[i];
                      e_bond[num] = _atoms[j];
                      num++;
                 }
            }
       }
}

void _RESIDUE::insert_a_atom(_ATOM *atom, CifString& error)
{
       if ((!strcmp(_ResName, "HOH") || !strcmp(_ResName, "DOD")) &&
            atom->atmtype()[0] == 'O') {
            atom->set_atmtype("O");
            atom->set_pdb_atmnam("O");
            // atom->set_pub_atmnam("O");
       }
       CifString index(atom->restype());
       index += atom->pdb_atmnam();
       index += "_";
       index += atom->alt_loc();
       int i = _atomNameAlts.Seek(index);
       if (i >= 0) write_error_message(atom, _atoms[i], error, 1);

       _atomNameAlts.Add(index);
       index.Copy(atom->restype());
       index += atom->pdb_atmnam();

       i = _atomNames.Seek(index);
       if (i >= 0) {
            if (strcmp(atom->alt_loc(), "") && !strcmp(_atoms[i]->alt_loc(), "") ||
                !strcmp(atom->alt_loc(), "") && strcmp(_atoms[i]->alt_loc(), ""))
                 write_error_message(atom, _atoms[i], error, 1);
       }
       _atomNames.Add(index);

       _ISIndex idx;
       idx.set_ivalue(_Num_Atom);
       idx.set_svalue(atom->alt_loc());
       idx.set_pos(_Num_Atom);
       _atomOrder.Add(idx);

       if (_Num_Atom >= _Alloc_Atom) {
            _Alloc_Atom += _ATOM_INCR;
            _ATOM **tmp_atom = _atoms;
            int *tmp_deleted = _deleted;
            _atoms = new _ATOM*[_Alloc_Atom];
            _deleted = new int[_Alloc_Atom];
            for (i = 0; i < _Num_Atom; i++) {
                 _atoms[i] = tmp_atom[i];
                 _deleted[i] = tmp_deleted[i];
            }
            for (i = _Num_Atom; i < _Alloc_Atom; i++) _atoms[i] = NULL;
            delete [] tmp_atom;
            delete [] tmp_deleted;
       }
       _atoms[_Num_Atom] = atom;
       _deleted[_Num_Atom] = 0;
       _Num_Atom++;
       if (strcmp(atom->alt_loc(), "")) {
           if (!strstr(_alt_loc, atom->alt_loc()))
                strcat(_alt_loc, atom->alt_loc());
       }
}

void _RESIDUE::insert_a_atom(const int card_id, char FieldInfo[MxField][MxLen],
                 const int is_pdb_format, CifString& error_message)
{
       _ATOM *atom = NULL;
       COORD orig;
       SIGATM sigatm;
       ANISOU anisou;
       SIGUIJ siguij;

       if (!(is_atom_card(card_id) & AUX_ATOM)) {
            if (is_aa_residue(FieldInfo[4])) {
                 if (!strcmp(FieldInfo[2], "OT") || !strcmp(FieldInfo[2], "OT2"))
                      strcpy(FieldInfo[2], "OXT");
                 else if (!strcmp(FieldInfo[2], "OT1"))
                      strcpy(FieldInfo[2], "O");
            }
            atom = new _ATOM;
            atom->set_type(FieldInfo[0]);
            atom->set_atnum(atoi(FieldInfo[1]));
            atom->set_atmtype(FieldInfo[2]);
            atom->set_is_hydrogen(0);
            atom->set_alt_loc(FieldInfo[3]);
            atom->set_restype(FieldInfo[4]);
            atom->set_chnid(FieldInfo[5]);
            atom->set_resnum(atoi(FieldInfo[6]));
            atom->set_ins_code(FieldInfo[7]);
            orig.x = atof(FieldInfo[8]);
            orig.y = atof(FieldInfo[9]);
            orig.z = atof(FieldInfo[10]);
            atom->set_orig(orig);
            atom->set_occ(atof(FieldInfo[11]));
            atom->set_t_fct(atof(FieldInfo[12]));
            if (strcmp(FieldInfo[13], ""))
                 atom->set_footnote(atoi(FieldInfo[13]));
            else atom->set_footnote(-10000);
            atom->set_not_in_asym(FieldInfo[14]);
            if (is_pdb_format) {
                 atom->set_pdb_resnum(atoi(FieldInfo[6]));
                 atom->set_pdb_resnam(FieldInfo[4]);
                 atom->set_pdb_chnid(FieldInfo[5]);
                 atom->set_pdb_atmnam(FieldInfo[2]);
                 if (strcmp(FieldInfo[19], "") &&
                     strcmp(FieldInfo[20], "") &&
                     strcmp(FieldInfo[22], "")) {
                      atom->set_pub_resnum(atoi(FieldInfo[19]));
                      atom->set_pub_resnam(FieldInfo[20]);
                      atom->set_pub_chnid(FieldInfo[21]);
                      atom->set_pub_atmnam(FieldInfo[22]);
                 } else {
                      atom->set_pub_resnum(atoi(FieldInfo[6]));
                      atom->set_pub_resnam(FieldInfo[4]);
                      atom->set_pub_chnid(FieldInfo[5]);
                      atom->set_pub_atmnam(FieldInfo[2]);
                 }
            } else {
                 if (strcmp(FieldInfo[15], ""))
                      atom->set_pdb_resnum(atoi(FieldInfo[15]));
                 else atom->set_pdb_resnum(-99999);
                 atom->set_pdb_resnam(FieldInfo[16]);
                 atom->set_pdb_chnid(FieldInfo[17]);
                 atom->set_pdb_atmnam(FieldInfo[18]);
                 if (strcmp(FieldInfo[19], ""))
                      atom->set_pub_resnum(atoi(FieldInfo[19]));
                 else atom->set_pub_resnum(-99999);
                 atom->set_pub_resnam(FieldInfo[20]);
                 atom->set_pub_chnid(FieldInfo[21]);
                 atom->set_pub_atmnam(FieldInfo[22]);
            }
            atom->set_is_sigatm(0);
            atom->set_is_anisou(0);
            atom->set_is_siguij(0);
            insert_a_atom(atom, error_message);
       } else {
            atom = GetLastAtom();
            if (atom) {
                 if (card_id == SIGATM_TOKEN - 1) {
                      sigatm.sig_x = atof(FieldInfo[8]);
                      sigatm.sig_y = atof(FieldInfo[9]);
                      sigatm.sig_z = atof(FieldInfo[10]);
                      sigatm.sig_o = atof(FieldInfo[11]);
                      sigatm.sig_t = atof(FieldInfo[12]);
                      atom->set_sigatm(sigatm);
                      atom->set_is_sigatm(1);
                 } else if (card_id == ANISOU_TOKEN - 1) {
                      anisou.u11 = atoi(FieldInfo[8]);
                      anisou.u22 = atoi(FieldInfo[9]);
                      anisou.u33 = atoi(FieldInfo[10]);
                      anisou.u12 = atoi(FieldInfo[11]);
                      anisou.u13 = atoi(FieldInfo[12]);
                      anisou.u23 = atoi(FieldInfo[13]);
                      atom->set_anisou(anisou);
                      atom->set_is_anisou(1);
                 } else if (card_id == SIGUIJ_TOKEN - 1) {
                      siguij.sig_u11 = atoi(FieldInfo[8]);
                      siguij.sig_u22 = atoi(FieldInfo[9]);
                      siguij.sig_u33 = atoi(FieldInfo[10]);
                      siguij.sig_u12 = atoi(FieldInfo[11]);
                      siguij.sig_u13 = atoi(FieldInfo[12]);
                      siguij.sig_u23 = atoi(FieldInfo[13]);
                      atom->set_siguij(siguij);
                      atom->set_is_siguij(1);
                 }
            }
       }
}

void _RESIDUE::delete_a_atom(const char* name)
{
       int i, ret;
       CifString index(_ResName);
       _ISIndex isindex;

       index += name;
       ret = _atomNames.Seek(index);
       while (ret >= 0) {
            _deleted[ret] = 1;
            _atomNames.Delete();
            _atomNameAlts.GoToRecord(ret);
            _atomNameAlts.Delete();
            ret = _atomNames.GoToNext();
            if (ret >= 0) {
                 if (_atomNames.GetCurrData() != index) break;
            }
       }
       _atomOrder.Clear();
       for (i = 0; i < _Num_Atom; i++) {
            if (_deleted[i]) continue;
            isindex.set_ivalue(i);
            isindex.set_svalue(_atoms[i]->alt_loc());
            isindex.set_pos(i);
            _atomOrder.Add(isindex);
       }
}

void _RESIDUE::change_a_atom(_ATOM *atom)
{
       if ((!strcmp(_ResName, "HOH") || !strcmp(_ResName, "DOD")) &&
            atom->atmtype()[0] == 'O') {
            atom->set_atmtype("O");
            atom->set_pdb_atmnam("O");
            atom->set_pub_atmnam("O");
       }
       CifString index(atom->restype());
       index += atom->pdb_atmnam();
       index += "_";
       index += atom->alt_loc();
       int ret = _atomNameAlts.Seek(index);
       if (ret >= 0) {
            if (_atoms[ret]) delete _atoms[ret];
            _atoms[ret] = atom;
       }
}

_ATOM** _RESIDUE::find_atom(const char* name, int &num)
{
       int alloc = 0, i, ret;
       char buffer[10], *p = NULL;
       _ATOM **atom = NULL, **tmp_atom = NULL;
       CifString index;

       num = 0;
       p = _name_list.Text();
       while (p) {
            p = get_word(p, buffer);
            if (!strcmp(buffer, "")) continue;

            index.Copy(buffer);
            index += name;
            ret = _atomNames.Seek(index);
            while (ret >= 0) {
                 if (num >= alloc) {
                      alloc += 5;
                      tmp_atom = atom;
                      atom = new _ATOM*[alloc];
                      for (i = 0; i < num; i++) atom[i] = tmp_atom[i];
                      for (i = num; i < alloc; i++) atom[i] = NULL;
                      if (tmp_atom) delete [] tmp_atom;
                 }
                 atom[num] = _atoms[ret];
                 num++;
                 ret = _atomNames.GoToNext();
                 if (ret >= 0) {
                      if (_atomNames.GetCurrData() != index) break;
                 }
            }
       }
       return atom;
}

_ATOM* _RESIDUE::find_atom(const char* name)
{
       CifString index(_ResName);
       index += name;
       int ret = _atomNames.Seek(index);
       if (ret >= 0) return _atoms[ret];
       else return NULL;
}

void _RESIDUE::add_a_residue(_RESIDUE *residue, CifString& error)
{
       if (!strstr(_name_list.Text(), residue->_ResName)) {
            _name_list += " ";
            _name_list += residue->_ResName;
            _no_res_name++;
       }
       for (int i = 0; i < residue->_Num_Atom; i++) {
            if (residue->_deleted[i]) continue;
            insert_a_atom(residue->_atoms[i], error);
            residue->_atoms[i] = NULL;
       }
       delete residue;
}

void _RESIDUE::replace_residue(_RESIDUE *residue)
{
       int i;
       CifString index;

       if (_atoms) {
            for (i = 0; i < _Num_Atom; i++)
                 if (_atoms[i]) delete _atoms[i];
            delete [] _atoms;
       }
       if (_deleted) delete [] _deleted;
       _Num_Atom = residue->_Num_Atom;
       _Alloc_Atom = residue->_Alloc_Atom;
       _atoms = residue->_atoms;
       _deleted = residue->_deleted;

       residue->_Num_Atom = 0;
       residue->_Alloc_Atom = 0;
       residue->_atoms = NULL;
       residue->_deleted = NULL;
 
       _atomNames.Clear();
       _atomNameAlts.Clear();
       for (i = 0; i < _Num_Atom; i++) {
            index.Copy(_atoms[i]->restype());
            index += _atoms[i]->pdb_atmnam();
            _atomNames.Add(index);
            index += "_";
            index += _atoms[i]->alt_loc();
            _atomNameAlts.Add(index);
       }
       reorder_atoms();
}

void _RESIDUE::get_atom_type()
{
       int i, j, align, count = 0, is_aa = 0, n_name_type = 0;
       int n_n = 0, n_c = 0, n_ac = 0, n_op = 0;
       char atom_name[5], type[5], *p = NULL;
       _SIndex *index = NULL;
       ConnectFormat *drug = NULL;
       _NAME_TYPE *name_type = NULL;

       if (!_Num_Atom) return;

       name_type = new _NAME_TYPE[_Num_Atom];
       n_name_type = 0;
       drug = find_drug(_ResName);
       if (!drug) {
            for (i = 0; i < _Num_Atom; i++) {
                 if (_deleted[i]) continue;
                 count++;
                 _atoms[i]->set_is_hydrogen(0);
                 strcpy(atom_name, _atoms[i]->pdb_atmnam());
                 if (!strcmp(_atoms[i]->atom_type(), "")) {
                      ndb_get_atom_symbol(atom_name, type);
                      if (!strcmp(type, "CA") || !strcmp(type, "CD") ||
                          !strcmp(type, "CE") || !strcmp(type, "N") ||
                          !strcmp(type, "AC") || !strcmp(type, "NP") ||
                          !strcmp(type, "NO")) {
                           for (j = 0; j < n_name_type; j++) {
                                if (!strcmp(atom_name, name_type[j].atom_name)) break;
                           }
                           if (j == n_name_type) {
                                strcpy(name_type[j].atom_name, atom_name);
                                strcpy(name_type[j].atom_type, type);
                                name_type[j].changed = 0;
                                n_name_type++;
                           }
                      }
                      _atoms[i]->set_atom_type(type);
                      align = get_align(atom_name, _atoms[i]->atom_type());
                      _atoms[i]->set_align(align);
                 }
            }
       } else {
            if (drug->ndbAtomType() == ATOMP_TOKEN) is_aa = 1;
            for (i = 0; i < _Num_Atom; i++) {
                 if (_deleted[i]) continue;
                 _atoms[i]->set_is_hydrogen(0);
                 count++;
                 strcpy(atom_name, _atoms[i]->pdb_atmnam());
                 if ((!strcmp(_atoms[i]->pdb_resnam(), "HOH") ||
                      !strcmp(_atoms[i]->pdb_resnam(), "DOD")) &&
                       atom_name[0] == 'O') strcpy(atom_name, "O");
                 index = drug->find_atom(atom_name);
                 if (index) {
                      if (!strcmp(_atoms[i]->atom_type(), ""))
                           _atoms[i]->set_atom_type(drug->atoms(index->pos())
                                    ->atomtype());
                      _atoms[i]->set_align(drug->atoms(index->pos())->align_pos());
                 } else if (!strcmp(_atoms[i]->pdb_atmnam(), "1H") ||
                            !strcmp(_atoms[i]->pdb_atmnam(), "2H") ||
                            !strcmp(_atoms[i]->pdb_atmnam(), "3H")) {
                      _atoms[i]->set_atom_type("H");
                 } else if (!strcmp(_atoms[i]->pdb_atmnam(), "1D") ||
                            !strcmp(_atoms[i]->pdb_atmnam(), "2D") ||
                            !strcmp(_atoms[i]->pdb_atmnam(), "3D")) {
                      _atoms[i]->set_atom_type("D");
                 } else {
                      if (!strcmp(_atoms[i]->atom_type(), "")) {
                           ndb_get_atom_symbol(atom_name, type);
                           if ((!strcmp(type, "CA") || !strcmp(type, "CD") ||
                                !strcmp(type, "CE") || !strcmp(type, "NA") ||
                                !strcmp(type, "NB") || !strcmp(type, "PA") ||
                                !strcmp(type, "PB")) &&
                                drug->ndbAtomType() != HETAI_TOKEN &&
                                drug->ndbAtomType() != HETIC_TOKEN)
                                type[1] = '\0';
                           if (!strcmp(type, "CA") || !strcmp(type, "CD") ||
                               !strcmp(type, "CE") || !strcmp(type, "N") ||
                               !strcmp(type, "AC") || !strcmp(type, "NP") ||
                               !strcmp(type, "NO")) {
                                for (j = 0; j < n_name_type; j++) {
                                     if (!strcmp(atom_name, name_type[j].atom_name)) break;
                                }
                                if (j == n_name_type) {
                                     strcpy(name_type[j].atom_name, atom_name);
                                     strcpy(name_type[j].atom_type, type);
                                     name_type[j].changed = 0;
                                     n_name_type++;
                                }
                           }
                           _atoms[i]->set_atom_type(type);
                      }
                      align = get_align(atom_name, _atoms[i]->atom_type());
                      _atoms[i]->set_align(align);
                 }
                 if (is_aa && (!strcmp(_atoms[i]->atom_type(), "HE") ||
                     !strcmp(_atoms[i]->atom_type(), "HF") ||
                     !strcmp(_atoms[i]->atom_type(), "HG") ||
                     !strcmp(_atoms[i]->atom_type(), "HO")))
                      _atoms[i]->set_atom_type("H");
                 if (!strcmp(_atoms[i]->atom_type(), "H") ||
                     !strcmp(_atoms[i]->atom_type(), "D"))
                      _atoms[i]->set_is_hydrogen(1);
            }
       }
       for (i = 0; i < n_name_type; i++) {
            if (!strcmp(name_type[i].atom_type, "CA") ||
                !strcmp(name_type[i].atom_type, "CD") ||
                !strcmp(name_type[i].atom_type, "CE"))
                 n_c++;
            if (!strcmp(name_type[i].atom_type, "N")) {
                 strcpy(atom_name, name_type[i].atom_name);
                 p = strstr(atom_name, name_type[i].atom_type);
                 if (p) {
                      p++; p[1] = '\0';
                      if (isalpha(*p) && ndb_get_atom_number(p) > 0) n_n++;
                 }
            }
            if (!strcmp(name_type[i].atom_type, "AC"))
                 n_ac++;
            if (!strcmp(name_type[i].atom_type, "NP") ||
                !strcmp(name_type[i].atom_type, "NO"))
                 n_op++;
       }
       if (n_c >= 2) {
            for (i = 0; i < n_name_type; i++) {
                 if (!strcmp(name_type[i].atom_type, "CA") ||
                     !strcmp(name_type[i].atom_type, "CD") ||
                     !strcmp(name_type[i].atom_type, "CE")) {
                      strcpy(name_type[i].atom_type, "C");
                      name_type[i].changed = 1;
                 }
            }
       }
       if (n_n >= 4) {
            for (i = 0; i < n_name_type; i++) {
                 if (!strcmp(name_type[i].atom_type, "N")) {
                      strcpy(atom_name, name_type[i].atom_name);
                      p = strstr(atom_name, name_type[i].atom_type);
                      if (p) {
                           p++; p[1] = '\0';
                           if (isalpha(*p) && ndb_get_atom_number(p) > 0) {
                                strcpy(name_type[i].atom_type, p);
                                name_type[i].changed = 1;
                           }
                      }
                 }
            }
       }
       if (n_ac >= 4) {
            for (i = 0; i < n_name_type; i++) {
                 if (!strcmp(name_type[i].atom_type, "AC")) {
                      strcpy(name_type[i].atom_type, &name_type[i].atom_type[1]);
                      name_type[i].changed = 1;
                 }
            }
       }
       if (n_op >= 4) {
            for (i = 0; i < n_name_type; i++) {
                 if (!strcmp(name_type[i].atom_type, "NP") ||
                     !strcmp(name_type[i].atom_type, "NO")) {
                      strcpy(name_type[i].atom_type, &name_type[i].atom_type[1]);
                      name_type[i].changed = 1;
                 }
            }
       }
       for (j = 0; j < n_name_type; j++) {
            if (!name_type[j].changed) continue;
            for (i = 0; i < _Num_Atom; i++) {
                 if (_deleted[i]) continue;
                 if (!strcmp(_atoms[i]->pdb_atmnam(), name_type[j].atom_name)) {
                      _atoms[i]->set_atom_type(name_type[j].atom_type);
                      align = get_align(_atoms[i]->pdb_atmnam(), _atoms[i]->atom_type());
                      _atoms[i]->set_align(align);
                 }
            }
       }
       if (count == 1) _is_single_atom = 1;
       delete [] name_type;
}

void _RESIDUE::reorder_atoms()
{
       int i, j, k, len, num_name_list = 0, start[10], heavy_atom[10], extra_atom[10];
       char name_list[10][5], atom_name[5], /* type[5], */ *p;
       _SIndex *index = NULL;
       _ISIndex isindex;
       ConnectFormat *drug[10];

       if (!_Num_Atom || !strcmp(_ResName, "UNL")) return;

       p = _name_list.Text();
       while (p) {
            p = get_word(p, name_list[num_name_list]);
            if (!strcmp(name_list[num_name_list], "")) break;
            num_name_list++;
       }
       if (!num_name_list) return;
       start[0] = 0;
       for (i = 0; i < num_name_list; i++) {
            drug[i] = find_drug(name_list[i]);
            if (!drug[i]) break;
            heavy_atom[i] = drug[i]->nheavyatoms();
            extra_atom[i] = drug[i]->natoms() + 4;
            if (i) start[i] = 4 * extra_atom[i-1];
       }
       if (i < num_name_list) return;

       _atomOrder.Clear();
       for (i = 0; i < _Num_Atom; i++) {
            if (_deleted[i]) continue;
            strcpy(atom_name, _atoms[i]->pdb_atmnam());
            if ((!strcmp(_atoms[i]->pdb_resnam(), "HOH") ||
                 !strcmp(_atoms[i]->pdb_resnam(), "DOD")) &&
                  atom_name[0] == 'O') strcpy(atom_name, "O");
            for (j = 0; j < num_name_list; j++)
                 if (!strcmp(_atoms[i]->restype(), name_list[j])) break;
            if (j == num_name_list) continue;
            index = drug[j]->find_atom(atom_name);
            if (!index && _atoms[i]->is_hydrogen()) {
                 len = strlen(atom_name);
                 for (k = 0; k < len; k++) {
                      if (atom_name[k] == 'D') {
                           atom_name[k] = 'H';
                           break;
                      }
                 }
                 index = drug[j]->find_atom(atom_name);
            }
            if (index) {
                 if (!strcmp(_atoms[i]->atom_type(), "H") ||
                     !strcmp(_atoms[i]->atom_type(), "D")) {
                      isindex.set_ivalue(start[j] + index->pos() + 3);
                      // _atoms[i]->set_is_hydrogen(1);
                 } else isindex.set_ivalue(start[j] + index->pos());
            } else if (!strcmp(_atoms[i]->pdb_atmnam(), "1H") ||
                       !strcmp(_atoms[i]->pdb_atmnam(), "1D") ||
                       !strcmp(_atoms[i]->pdb_atmnam(), "2H") ||
                       !strcmp(_atoms[i]->pdb_atmnam(), "2D") ||
                       !strcmp(_atoms[i]->pdb_atmnam(), "3H") ||
                       !strcmp(_atoms[i]->pdb_atmnam(), "3D")) {
                 if (!strcmp(_atoms[i]->pdb_atmnam(), "1H") ||
                     !strcmp(_atoms[i]->pdb_atmnam(), "1D"))
                      isindex.set_ivalue(start[j]+heavy_atom[j]+1);
                 if (!strcmp(_atoms[i]->pdb_atmnam(), "2H") ||
                     !strcmp(_atoms[i]->pdb_atmnam(), "2D"))
                      isindex.set_ivalue(start[j]+heavy_atom[j]+2);
                 if (!strcmp(_atoms[i]->pdb_atmnam(), "3H") ||
                     !strcmp(_atoms[i]->pdb_atmnam(), "3D"))
                      isindex.set_ivalue(start[j]+heavy_atom[j]+3);
            } else {
                 isindex.set_ivalue(start[j]+extra_atom[j]++);
            }
            isindex.set_svalue(_atoms[i]->alt_loc());
            isindex.set_pos(i);
            _atomOrder.Add(isindex);
       }
}

void _RESIDUE::switch_hydrogen_labeling()
{
       int i, num_loc = 0, *loc = NULL, ret;
       CifString index;
       CHANGE_ATOM_NAME *exist = find_change(_ResName);
       if (!exist || !exist->num) return;

       loc = new int[exist->num * 10];
       for (i = 0; i < exist->num; i++) {
            index.Copy(_ResName);
            index += exist->atom_list[i].old_name;
            ret = _atomNames.Seek(index);
            while (ret >= 0) {
                 _atoms[ret]->set_atmtype(exist->atom_list[i].new_name);
                 _atoms[ret]->set_pdb_atmnam(exist->atom_list[i].new_name);
                 loc[num_loc] = ret;
                 num_loc++;
                 ret = _atomNames.GoToNext();
                 if (ret >= 0) {
                      if (_atomNames.GetCurrData() != index) break;
                 }
            }
       }
       for (i = 0; i < num_loc; i++) {
            index.Copy(_atoms[loc[i]]->restype());
            index += _atoms[loc[i]]->pdb_atmnam();
            _atomNames.GoToRecord(loc[i]);
            _atomNames.ReplData(index);
            index += "_";
            index += _atoms[loc[i]]->alt_loc();
            _atomNameAlts.GoToRecord(loc[i]);
            _atomNameAlts.ReplData(index);
       }
       delete [] loc;
}

void _RESIDUE::labeling_iupac_hydrogen_first()
{
       int i, ret;
       CifString index;

       for (i = 0; i < IU_FIRST.num; i++) {
            index.Copy(_ResName);
            index += IU_FIRST.atom_list[i].old_name;
            ret = _atomNames.Seek(index);
            while (ret >= 0) {
                 _atoms[ret]->set_atmtype(IU_FIRST.atom_list[i].new_name);
                 ret = _atomNames.GoToNext();
                 if (ret >= 0) {
                      if (_atomNames.GetCurrData() != index) break;
                 }
            }
       }
}

void _RESIDUE::labeling_iupac_hydrogen()
{
       int i, ret;
       CifString index;
       CHANGE_ATOM_NAME *exist = find_iupac(_ResName);
       if (!exist || !exist->num) return;

       for (i = 0; i < exist->num; i++) {
            index.Copy(_ResName);
            index += exist->atom_list[i].old_name;
            ret = _atomNames.Seek(index);
            while (ret >= 0) {
                 _atoms[ret]->set_atmtype(exist->atom_list[i].new_name);
                 ret = _atomNames.GoToNext();
                 if (ret >= 0) {
                      if (_atomNames.GetCurrData() != index) break;
                 }
            }
       }

       for (i = 0; i < IU_LAST.num; i++) {
            index.Copy(_ResName);
            index += IU_LAST.atom_list[i].old_name;
            ret = _atomNames.Seek(index);
            while (ret >= 0) {
                 _atoms[ret]->set_atmtype(IU_LAST.atom_list[i].new_name);
                 ret = _atomNames.GoToNext();
                 if (ret >= 0) {
                      if (_atomNames.GetCurrData() != index) break;
                 }
            }
       }
}

void _RESIDUE::find_extra_H_atom()
{
       int ret = -1, ret1 = -1, ret2 = -1, ret3 = -1;
       CifString atom_index;

       atom_index.Copy(_ResName);
       atom_index += "H";
       ret = _atomNames.Seek(atom_index);

       atom_index.Copy(_ResName);
       atom_index += "1H";
       ret1 = _atomNames.Seek(atom_index);

       atom_index.Copy(_ResName);
       atom_index += "2H";
       ret2 = _atomNames.Seek(atom_index);

       atom_index.Copy(_ResName);
       atom_index += "3H";
       ret3 = _atomNames.Seek(atom_index);

       if (ret >= 0) {
/*
            _extras += "H ";
            if (ret1 >= 0 && ret2 >= 0 && ret3 >= 0)
                 _extras += "(has H, 1H, 2H, 3H) ";
            else if (ret1 >= 0 && ret2 >= 0)
                 _extras += "(has H, 1H, 2H) ";
            else if (ret1 >= 0 && ret3 >= 0)
                 _extras += "(has H, 1H, 3H) ";
            else if (ret2 >= 0 && ret3 >= 0)
                 _extras += "(has H, 2H, 3H) ";
            else if (ret1 >= 0)
                 _extras += "(has H, 1H) ";
            else if (ret2 >= 0)
                 _extras += "(has H, 2H) ";
            else if (ret3 >= 0)
                 _extras += "(has H, 3H) ";
*/
            if (ret1 >= 0 || ret2 >= 0 || ret3 >= 0)
                 _extras += "H is extra ";
	    else _extras += "H ";
            if (ret1 >= 0 && ret2 >= 0 && ret3 >= 0)
                 _extras += "(1H, 2H, 3H already exist) ";
            else if (ret1 >= 0 && ret2 >= 0)
                 _extras += "(1H, 2H already exist) ";
            else if (ret1 >= 0 && ret3 >= 0)
                 _extras += "(1H, 3H already exist) ";
            else if (ret2 >= 0 && ret3 >= 0)
                 _extras += "(2H, 3H already exist) ";
            else if (ret1 >= 0)
                 _extras += "(1H already exist) ";
            else if (ret2 >= 0)
                 _extras += "(2H already exist) ";
            else if (ret3 >= 0)
                 _extras += "(3H already exist) ";
       }
}

void _RESIDUE::find_missing_or_extra_atoms(const int chain_type,
        const int is_first, const int is_last, const int need_hydrogen)
{
       int i, j, len, no_oxygen = 0, *index = NULL, ret, found;
       char atomname[5];
       CifString atom_index;
       _SIndex *sidx = NULL;
       ConnectFormat *drug = NULL;

       if (!strcmp(_ResName, "UNK") || !strcmp(_ResName, "UNL") ||
           !strcmp(_ResName, "N")) return;
       if (strlen(_name_list.Text()) > 3) return;

       if ((!strcmp(_ResName, "A") || !strcmp(_ResName, "+A") ||
            !strcmp(_ResName, "C") || !strcmp(_ResName, "+C") ||
            !strcmp(_ResName, "G") || !strcmp(_ResName, "+G") ||
            !strcmp(_ResName, "I") || !strcmp(_ResName, "+I")) &&
            !find_atom("O2*"))
            no_oxygen = 1;
           
       drug = find_drug(_ResName, no_oxygen);
       if (drug == NULL) return;

       index = new int[_Num_Atom];
       for (i = 0; i < _Num_Atom; i++) index[i] = 1;
       for (i = 0; i < drug->natoms(); i++) {
            if (!strcmp(drug->atoms(i)->atomname(), "OXT") &&
                 chain_type == ATOMP_TOKEN) continue;
            atom_index.Copy(_ResName);
            atom_index += drug->atoms(i)->atomname();
            ret = _atomNames.Seek(atom_index);
            if (ret < 0 && !strcmp(drug->atoms(i)->atomtype(), "H")) {
                 strcpy(atomname, drug->atoms(i)->atomname());
                 len = strlen(atomname);
                 for (j = 0; j < len; j++) {
                      if (atomname[j] == 'H') {
                           atomname[j] = 'D';
                           break;
                      }
                 }
                 atom_index.Copy(_ResName);
                 atom_index += atomname;
                 ret = _atomNames.Seek(atom_index);
            }
            if (ret >= 0) {
                 while (ret >= 0) {
                      index[ret] = 0;
                      ret = _atomNames.GoToNext();
                      if (ret >= 0) {
                           if (_atomNames.GetCurrData() != atom_index) break;
                      }
                 }
            } else if (drug->atoms(i)->ishydrogen() == 1 && !need_hydrogen ||
               (!strcmp(drug->atoms(i)->atomname(), "O2*") ||
                !strcmp(drug->atoms(i)->atomname(), "2HO*") ||
                !strcmp(drug->atoms(i)->atomname(), "H2*")) &&
                 _type == MONOMER_TYPE_DNA)
                 continue;
            else if (!strcmp(drug->atoms(i)->atomname(), "O3P") ||
                     !strcmp(drug->atoms(i)->atomname(), "OXT") ||
                     !strcmp(drug->atoms(i)->atomname(), "O1") &&
                      chain_type == ATOMS_TOKEN ||
                      chain_type == ATOMN_TOKEN && is_first &&
                    (!strcmp(drug->atoms(i)->atomname(), "P") ||
                     !strcmp(drug->atoms(i)->atomname(), "O1P") ||
                     !strcmp(drug->atoms(i)->atomname(), "O2P")) ||
                     is_aa_residue(_ResName) &&
                    (!strcmp(drug->atoms(i)->atomname(), "HN2") ||
                     !strcmp(drug->atoms(i)->atomname(), "HXT") ||
                     is_first && !strcmp(drug->atoms(i)->atomname(), "H")) ||
                    (!strcmp(drug->atoms(i)->atomname(), "HT1") ||
                     !strcmp(drug->atoms(i)->atomname(), "HT2")) &&
                    (!strcmp(_ResName, "PRO") ||
                     !strcmp(_ResName, "DPR")) || 
                     is_na_residue(_ResName) &&
                    (!strcmp(drug->atoms(i)->atomname(), "3HOP") ||
                     !strcmp(drug->atoms(i)->atomname(), "2HOP") ||
                     !strcmp(drug->atoms(i)->atomname(), "H3T")) ||
                     !strcmp(_ResName, "HAC") &&
                     !strcmp(drug->atoms(i)->atomname(), "O2")) 
                 continue;
            else if ((!strcmp(_ResName, "HOH") || !strcmp(_ResName, "DOD")) &&
                     !strcmp(drug->atoms(i)->atomname(), "O")) {
                 found = 0;
                 for (j = 0; j < _Num_Atom; j++) {
                      if (_atoms[j]->atmtype()[0] == 'O') {
                           index[j] = 0;
                           found = 1;
                      }
                 }
                 if (!found) {
                      _missing += drug->atoms(i)->atomname();
                      _missing += " ";
                 }
            } else if (!strcmp(_ResName, "HOH") && drug->atoms(i)->ishydrogen())
                 continue;
            else {
                 _missing += drug->atoms(i)->atomname();
                 _missing += " ";
            }
       }

       for (i = 0; i < _Num_Atom; i++) {
            if (_deleted[i] || !index[i]) continue;
            strcpy(atomname, _atoms[i]->pdb_atmnam());
            sidx = drug->find_atom(atomname);
            if (sidx) {
                 if (!strcmp(_atoms[i]->pdb_atmnam(), "OXT") &&
                      chain_type == ATOMP_TOKEN) {}
                 else continue;
            }
            if (_atoms[i]->is_hydrogen()) {
                 len = strlen(atomname); 
                 for (j = 0; j < len; j++) {
                      if (atomname[j] == 'D') {
                           atomname[j] = 'H';
                           break;
                      }
                 }
                 sidx = drug->find_atom(atomname);
                 if (sidx) continue;
            }
            if (!strcmp(_atoms[i]->pdb_atmnam(), "OXT") &&
                is_last && chain_type == ATOMP_TOKEN ||
                is_first && chain_type == ATOMP_TOKEN &&
               (!strcmp(_atoms[i]->pdb_atmnam(), "1H") ||
                !strcmp(_atoms[i]->pdb_atmnam(), "2H") ||
                !strcmp(_atoms[i]->pdb_atmnam(), "3H") ||
                !strcmp(_atoms[i]->pdb_atmnam(), "1D") ||
                !strcmp(_atoms[i]->pdb_atmnam(), "2D") ||
                !strcmp(_atoms[i]->pdb_atmnam(), "3D")) ||
                is_first && chain_type == ATOMN_TOKEN &&
                !strcmp(_atoms[i]->pdb_atmnam(), "H5T") ||
                 _type == MONOMER_TYPE_DNA &&
               (!strcmp(_atoms[i]->pdb_atmnam(), "1H2*") ||
                !strcmp(_atoms[i]->pdb_atmnam(), "2H2*"))) continue;
            _extras += _atoms[i]->pdb_atmnam();
            _extras += " ";
       }

       delete [] index;
}

void _RESIDUE::find_atom_type_mismatch(const int Mol_ID, CifString& error)
{
       int i, no_oxygen = 0;
       char atomname[5], Chain_ID[5], buffer[80];
       _SIndex *sidx = NULL;
       ConnectFormat *drug = NULL;

       if (!strcmp(_ResName, "UNK") || !strcmp(_ResName, "UNL") ||
           !strcmp(_ResName, "N")) return;
       if (strlen(_name_list.Text()) > 3) return;

       if ((!strcmp(_ResName, "A") || !strcmp(_ResName, "+A") ||
            !strcmp(_ResName, "C") || !strcmp(_ResName, "+C") ||
            !strcmp(_ResName, "G") || !strcmp(_ResName, "+G") ||
            !strcmp(_ResName, "I") || !strcmp(_ResName, "+I")) &&
            !find_atom("O2*"))
            no_oxygen = 1;
           
       drug = find_drug(_ResName, no_oxygen);
       if (drug == NULL) return;

       for (i = 0; i < _Num_Atom; i++) {
            if (_deleted[i]) continue;
            strcpy(atomname, _atoms[i]->pdb_atmnam());
            sidx = drug->find_atom(atomname);
            if (!sidx) continue;
            if (strcmp(_atoms[i]->atom_type(), drug->atoms(sidx->pos())->atomtype())) {
                 strcpy(Chain_ID, " ");
                 if (strcmp(_atoms[i]->pdb_chnid(), ""))
                      strcpy(Chain_ID, _atoms[i]->pdb_chnid());
                 if (Mol_ID < 0) {
                      sprintf(buffer, " %7s(    %c%4d)         %-4s      %-2s      %-2s\n",
                             _ResName, Chain_ID[0], _atoms[i]->pdb_resnum(),
                             _atoms[i]->pdb_atmnam(), _atoms[i]->atom_type(),
                             drug->atoms(sidx->pos())->atomtype());
                 } else {
                      sprintf(buffer, " %7s(%3d %c%4d)         %-4s      %-2s      %-2s\n",
                             _ResName, Mol_ID, Chain_ID[0], _atoms[i]->pdb_resnum(), 
                             _atoms[i]->pdb_atmnam(), _atoms[i]->atom_type(),
                             drug->atoms(sidx->pos())->atomtype());
                 }
                 error += buffer;
            }
       }
}

void _RESIDUE::cal_hydrogen_bond_distance(const int Mol_ID, CifString& error)
{
       int i, j, k, no_oxygen = 0, num, is_cys = 0, is_hg = 0;
       char buffer[80];
       double distance = 0.0;
       _ATOM **atoms = NULL;
       _SIndex *sidx = NULL;
       ConnectFormat *drug = NULL;

       if (!strcmp(_ResName, "UNK") || !strcmp(_ResName, "UNL") ||
           !strcmp(_ResName, "N")) return;
       if (strlen(_name_list.Text()) > 3) return;

       if ((!strcmp(_ResName, "A") || !strcmp(_ResName, "+A") ||
            !strcmp(_ResName, "C") || !strcmp(_ResName, "+C") ||
            !strcmp(_ResName, "G") || !strcmp(_ResName, "+G") ||
            !strcmp(_ResName, "I") || !strcmp(_ResName, "+I")) &&
            !find_atom("O2*"))
            no_oxygen = 1;
           
       drug = find_drug(_ResName, no_oxygen);
       if (drug == NULL) return;

       if (!strcmp(_ResName, "CYS")) is_cys = 1;

       for (i = 0; i < _Num_Atom; i++) {
            if (_deleted[i]) continue;
            if (!_atoms[i]->is_hydrogen()) continue;

            is_hg = 0;
            if (is_cys && !strcmp(_atoms[i]->pdb_atmnam(), "HG")) is_hg = 1;
 
            sidx = drug->find_atom(_atoms[i]->pdb_atmnam());
            if (!sidx) continue;
            for (j = 0; j < drug->atoms(sidx->pos())->nconect(); j++) {
                 atoms = find_atom(drug->atoms(sidx->pos())->connectto(j), num);
                 for (k = 0; k < num; k++) {
                      if (!strcmp(_atoms[i]->alt_loc(), "") ||
                          !strcmp(atoms[k]->alt_loc(), "") ||
                          !strcmp(_atoms[i]->alt_loc(), atoms[k]->alt_loc())) {
                           distance = cal_distance(_atoms[i], atoms[k]);
                           if (!is_hg && distance > 1.2 || is_hg && distance > 1.4) {
                                _wrong_connectivity = 1;
                                if (Mol_ID < 0) {
                                     sprintf(buffer, "    The bond distance is %.3f between '%s %s %s %d %s' and '%s %s %s %d %s'\n",
                                          distance, _atoms[i]->atmtype(), _atoms[i]->pdb_resnam(),
                                          _atoms[i]->pdb_chnid(), _atoms[i]->pdb_resnum(),
                                          _atoms[i]->ins_code(), atoms[k]->atmtype(),
                                          atoms[k]->pdb_resnam(), atoms[k]->pdb_chnid(),
                                          atoms[k]->pdb_resnum(), atoms[k]->ins_code());
                                } else {
                                     sprintf(buffer, "    The bond distance is %.3f between '%s %s %s %d %s' and '%s %s %s %d %s' of Model %d\n",
                                          distance, _atoms[i]->atmtype(), _atoms[i]->pdb_resnam(),
                                          _atoms[i]->pdb_chnid(), _atoms[i]->pdb_resnum(),
                                          _atoms[i]->ins_code(), atoms[k]->atmtype(),
                                          atoms[k]->pdb_resnam(), atoms[k]->pdb_chnid(),
                                          atoms[k]->pdb_resnum(), atoms[k]->ins_code(), Mol_ID);
                                }
                                error += buffer;
                           }
                      }
                 }
                 if (atoms) delete [] atoms;
                 atoms = NULL;
            }
       }
}

void _RESIDUE::find_wrong_alt_id(CifString& error)
{
       int i;
       char Chain_ID[5], buffer[200];

       for (i = 0; i < _Num_Atom; i++) {
            if (_deleted[i]) continue;
            if (!strcmp(_atoms[i]->alt_loc(), "")) continue;
            if (_atoms[i]->alt_loc()[0] >= 'A' &&
                _atoms[i]->alt_loc()[0] <= 'Z' ||
                _atoms[i]->alt_loc()[0] >= 'a' &&
                _atoms[i]->alt_loc()[0] <= 'z') continue;
            strcpy(Chain_ID, " ");
            if (strcmp(_atoms[i]->pdb_chnid(), ""))
                 strcpy(Chain_ID, _atoms[i]->pdb_chnid());
            sprintf(buffer, " Atom %-4s %-4s %c %4d (%8.3f%8.3f%8.3f) has alt ID '%s'\n",
                     _atoms[i]->pdb_atmnam(), _ResName, Chain_ID[0], _atoms[i]->pdb_resnum(),
                     _atoms[i]->orig().x, _atoms[i]->orig().y, _atoms[i]->orig().z,
                     _atoms[i]->alt_loc());
            error += buffer;
       }
}

void _RESIDUE::check_same_coordinates(CifString& error)
{
       int i, j;
       float dx, dy, dz;

       for (i = 0; i < _Num_Atom - 1; i++) {
            if (_deleted[i]) continue;
            for (j = i + 1; j < _Num_Atom; j++) {
                 if (_deleted[j]) continue;
                 if (!strcmp(_atoms[i]->alt_loc(), "") ||
                     !strcmp(_atoms[j]->alt_loc(), "") ||
                     !strcmp(_atoms[i]->alt_loc(),
                            _atoms[j]->alt_loc())) {
                      dx = fabs(_atoms[i]->orig().x - _atoms[j]->orig().x);
                      dy = fabs(_atoms[i]->orig().y - _atoms[j]->orig().y);
                      dz = fabs(_atoms[i]->orig().z - _atoms[j]->orig().z);
                      if (dx < 0.002 && dy < 0.002 && dz < 0.002)
                           write_error_message(_atoms[i], _atoms[j], error, 0);
                 }
            }
       }
}

void _RESIDUE::check_occupancy(CifString& error)
{
       int i;
       char name[5];
       float occupancy = 0;

       strcpy(name, "XXXX");
       for (i = 0; i < _Num_Atom; i++) {
            if (_deleted[i]) continue;
            if (strcmp(name, _atoms[i]->pdb_atmnam())) {
                 if (occupancy > 1.0) {
                      write_occupancy_message(error, _atoms[0], name, occupancy);
                 }
                 strcpy(name, _atoms[i]->pdb_atmnam());
                 occupancy = 0;
            }
            occupancy += _atoms[i]->occ();
       }
       if (occupancy > 1.0) {
            write_occupancy_message(error, _atoms[0], name, occupancy);
       }
}

void _RESIDUE::check_insertion_code(CifString& error)
{
       int i;
       char buffer[80], Chain_ID[5];

       for (i = 0; i < _Num_Atom; i++) {
            if (_deleted[i]) continue;
            if (strcmp(_atoms[i]->ins_code(), "") &&
               (_atoms[i]->ins_code()[0] < 'A' || _atoms[i]->ins_code()[0] > 'Z' &&
                _atoms[i]->ins_code()[0] < 'a' || _atoms[i]->ins_code()[0] > 'z')) {
                 strcpy(Chain_ID, " ");
                 if (strcmp(_atoms[i]->pdb_chnid(), ""))
                      strcpy(Chain_ID, _atoms[i]->pdb_chnid());
                 sprintf(buffer, "Residue %-4s %c %4d has wrong insertion code '%s'\n",
                    _ResName, Chain_ID[0], _atoms[i]->pdb_resnum(), _atoms[i]->ins_code());
                 error += buffer;
                 break;
            }
       }
}

int is_connect(const char *atm_name_a, _RESIDUE *res_a,
              const char *atm_name_b, _RESIDUE *res_b)
{
       int num1 = 0, num2 = 0, m, n;
       _ATOM **atom_a = res_a->find_atom(atm_name_a, num1);
       _ATOM **atom_b = res_b->find_atom(atm_name_b, num2);
       if (atom_a && atom_b) {
            for (m = 0; m < num1; m++) {
                 for (n = 0; n < num2; n++) {
                      if (!strcmp(atom_a[m]->alt_loc(), "") ||
                          !strcmp(atom_b[n]->alt_loc(), "") ||
                          !strcmp(atom_a[m]->alt_loc(),
                                  atom_b[n]->alt_loc())) {
                           if (is_a_bond(atom_a[m], atom_b[n])) {
                                delete [] atom_a;
                                delete [] atom_b;
                                return 1;
                           }
                      }
                 }
            }
       }
       if (atom_a) delete [] atom_a;
       if (atom_b) delete [] atom_b;
       return 0;
}

int is_connect(_RESIDUE *res_a, _RESIDUE *res_b, const char *atm_type_a,
              const char *atm_type_b)
{
       int i, j;
       char type1[5], type2[5];
       for (i = 0; i < res_a->_Num_Atom; i++) {
            if (res_a->_deleted[i]) continue;
            for (j = 0; j < res_b->_Num_Atom; j++) {
                 if (res_b->_deleted[j]) continue;
                 strcpy(type1, res_a->_atoms[i]->atom_type());
                 strcpy(type2, res_b->_atoms[j]->atom_type());
                 if (!strcmp(type1, ""))
                      ndb_get_atom_symbol(res_a->_atoms[i]->pdb_atmnam(), type1);
                 if (!strcmp(type2, ""))
                      ndb_get_atom_symbol(res_b->_atoms[j]->pdb_atmnam(), type2);
                 if (!strcmp(type1, atm_type_a) && !strcmp(type2, atm_type_b) ||
                     !strcmp(type1, atm_type_b) && !strcmp(type2, atm_type_a)) {
                      if (!strcmp(res_a->_atoms[i]->alt_loc(), "") ||
                          !strcmp(res_b->_atoms[j]->alt_loc(), "") ||
                          !strcmp(res_a->_atoms[i]->alt_loc(),
                                  res_b->_atoms[j]->alt_loc())) {
                           if (is_a_bond(res_a->_atoms[i], res_b->_atoms[j]))
                                return 1;
                      }
                 }
            }
       }
       return 0;
}

int is_connect(_RESIDUE *res_a, _RESIDUE *res_b)
{
       int i, j;
       for (i = 0; i < res_a->_Num_Atom; i++) {
            for (j = 0; j < res_b->_Num_Atom; j++) {
                 if (!strcmp(res_a->_atoms[i]->alt_loc(), "") ||
                     !strcmp(res_b->_atoms[j]->alt_loc(), "") ||
                     !strcmp(res_a->_atoms[i]->alt_loc(),
                             res_b->_atoms[j]->alt_loc())) {
                      if (is_a_bond(res_a->_atoms[i], res_b->_atoms[j]))
                           return 1;
                 }
            }
       }
       return 0;
}

int _RESIDUE::change_nomenclature(const int format, const int include_hydrogen,
       const int exclude_pdb, const int no_copy_pdb_chainid) 
{
       char p[5], atomtype[5];
       int found_hydrogen_atoms = 0, changed = 0, k, l;

       if (/* !strcmp(_ResName, "W")   || */ !strcmp(_ResName, "H2O") ||
           !strcmp(_ResName, "WAT") || !strncmp(_ResName, "TIP", 3) ||
           !strcmp(_ResName, "WTR") || !strcmp(_ResName, "OH2")) {
            strcpy(_ResName, "HOH");
            changed = 1;
       }

       for (k = 0; k < _Num_Atom; k++) {
            if (format == NDB_FILE_FORMAT_MIX) {
                 _atoms[k]->set_pdb_resnum(_atoms[k]->resnum());
                 _atoms[k]->set_pdb_chnid(_atoms[k]->chnid());
                 _atoms[k]->set_pdb_resnam(_ResName);
                 _atoms[k]->set_pdb_atmnam(_atoms[k]->atmtype());
            } else {
                 if (!exclude_pdb) {
                      if (_atoms[k]->pdb_resnum() == -99999) {
                           if (_atoms[k]->pub_resnum() != -99999)
                                _atoms[k]->set_pdb_resnum(_atoms[k]->pub_resnum());
                           else _atoms[k]->set_pdb_resnum(_atoms[k]->resnum());
                      }
                      if (!strcmp(_atoms[k]->pdb_resnam(), ""))
                           _atoms[k]->set_pdb_resnam(_ResName);
                      if (!strcmp(_atoms[k]->pdb_chnid(), "") && !no_copy_pdb_chainid) {
                           if (strcmp(_atoms[k]->pub_chnid(), ""))
                                _atoms[k]->set_pdb_chnid(_atoms[k]->pub_chnid());
                           else _atoms[k]->set_pdb_chnid(_atoms[k]->chnid());
                      }
                      if (!strcmp(_atoms[k]->pdb_atmnam(), ""))
                           _atoms[k]->set_pdb_atmnam(_atoms[k]->atmtype());
                 }
     
                 if (_atoms[k]->pub_resnum() == -99999) {
                      if (_atoms[k]->pdb_resnum() != -99999)
                           _atoms[k]->set_pub_resnum(_atoms[k]->pdb_resnum());
                      else _atoms[k]->set_pub_resnum(_atoms[k]->resnum());
                 }
                 if (!strcmp(_atoms[k]->pub_resnam(), "")) {
                      if (strcmp(_atoms[k]->pdb_resnam(), ""))
                           _atoms[k]->set_pub_resnam(_atoms[k]->pdb_resnam());
                      else _atoms[k]->set_pub_resnam(_atoms[k]->restype());
                 }
                 if (!strcmp(_atoms[k]->pub_chnid(), "")) {
                      if (no_copy_pdb_chainid)
                           _atoms[k]->set_pub_chnid(_atoms[k]->pdb_chnid());
                      else {
                           if (strcmp(_atoms[k]->pdb_chnid(), ""))
                                _atoms[k]->set_pub_chnid(_atoms[k]->pdb_chnid());
                           else _atoms[k]->set_pub_chnid(_atoms[k]->chnid());
                      }
                 }
                 if (!strcmp(_atoms[k]->pub_atmnam(), "")) {
                      if (strcmp(_atoms[k]->pdb_atmnam(), ""))
                           _atoms[k]->set_pub_atmnam(_atoms[k]->pdb_atmnam());
                      else _atoms[k]->set_pub_atmnam(_atoms[k]->atmtype());
                 }
            }
            _atoms[k]->set_restype(_ResName);
            if (!strcmp(_ResName, "HOH")) _atoms[k]->set_pdb_resnam(_ResName);

            strcpy(p, _atoms[k]->atmtype());
            if (!res_name_cmp("A", _ResName) || !res_name_cmp("C", _ResName) ||
                !res_name_cmp("G", _ResName) || !res_name_cmp("I", _ResName) ||
                !res_name_cmp("T", _ResName) || !res_name_cmp("U", _ResName) ||
                 is_a_nonstandard_trna(_ResName)) {
                 for (l = 0; l < 5 && p[l] != '\0'; l++) {

                      if (p[l] == '\'') {
                           p[l] = '*';
                           changed = 1;
                      }
                 }
                 if (!strcmp(p, "O1*")) {
                      strcpy(p, "O4*"); changed = 1;
                 } else if (!strcmp(p, "O5T")) {
                      strcpy(p, "O3P"); changed = 1;
                 } else if ((!strcmp(p, "CM5") || !strcmp(p, "C5A") ||
                    !strcmp(p, "M5")) && !res_name_cmp("T", _ResName)) {
                      strcpy(p, "C5M"); changed = 1;
                 }
                 if (!res_name_cmp("T", _ResName)) {
                      if (!strcmp(p, "C7")) {
                           strcpy(p, "C5M"); changed = 1;
                      } else if (!strcmp(p, "1H7M") ||
                                 !strcmp(p, "1H7")) {
                           strcpy(p, "1H5M"); changed = 1;
                      } else if (!strcmp(p, "2H7M") ||
                                 !strcmp(p, "2H7")) {
                           strcpy(p, "2H5M"); changed = 1;
                      } else if (!strcmp(p, "3H7M") ||
                                 !strcmp(p, "3H7")) {
                           strcpy(p, "3H5M"); changed = 1;
                      }
                 }
            } else if (!strcmp(_ResName, "HOH")) {
                 if (!strcmp(p, "OW") || !strcmp(p, "WAT") ||
                     !strcmp(p, "OH2") || !strcmp(p, "WTR") ||
                     !strcmp(p, "W") || p[0] == 'O') {
                      strcpy(p, "O"); changed = 1;
                 } else if (!strcmp(p, "H1")) {
                      strcpy(p, "1H"); changed = 1;
                 } else if (!strcmp(p, "H2")) {
                      strcpy(p, "2H"); changed = 1;
                 }
            }
            if (format == NDB_FILE_FORMAT_MIX || !exclude_pdb ||
                format == NDB_FILE_FORMAT_PDB) {
                 _atoms[k]->set_pdb_atmnam(p);
            }
            _atoms[k]->set_atmtype(p);

            ndb_get_atom_symbol(p, atomtype);
            // _atoms[k]->set_atom_type(atomtype);
            if (ndb_get_atom_number(atomtype) == 1) {
                 _atoms[k]->set_is_hydrogen(1);
                 found_hydrogen_atoms = 1;
            }
       }
       if (changed) _UpdateIndices();

       if ((is_aa_residue(_ResName) || !strcmp(_ResName, "A") ||
           !strcmp(_ResName, "C") || !strcmp(_ResName, "G") ||
           !strcmp(_ResName, "T") || !strcmp(_ResName, "U")) &&
           include_hydrogen) change_hydrogen_name();

       return (found_hydrogen_atoms);
}

void _RESIDUE::change_hydrogen_name()
{
       int num_loc = 0, i = 0, ret1, ret2, *loc = NULL;
       CifString index1, index2;
       CHANGE_ATOM_NAME *exist = NULL;

       loc = new int[_Num_Atom];
       exist = find_multiple(_ResName);
       if (exist) {
            for (i = 0; i < exist->num / 2; i++) {
                 index1.Copy(_ResName);
                 index1 += exist->atom_list[2 * i].old_name;
                 index2.Copy(_ResName);
                 index2 += exist->atom_list[2*i+1].old_name;

                 ret1 = _atomNames.Seek(index1);
                 ret2 = _atomNames.Seek(index2);
                 if (ret1 < 0 || ret2 < 0) continue;
                 while (ret1 >= 0) {
                      _atoms[ret1]->set_atmtype(exist->atom_list[2 * i].new_name);
                      _atoms[ret1]->set_pdb_atmnam(exist->atom_list[2 * i].new_name);
                      loc[num_loc] = ret1;
                      num_loc++;
                      ret1 = _atomNames.GoToNext();
                      if (ret1 >= 0) {
                           if (_atomNames.GetCurrData() != index1) break;
                      }
                 }
                 while (ret2 >= 0) {
                      _atoms[ret2]->set_atmtype(exist->atom_list[2*i+1].new_name);
                      _atoms[ret2]->set_pdb_atmnam(exist->atom_list[2*i+1].new_name);
                      loc[num_loc] = ret2;
                      num_loc++;
                      ret2 = _atomNames.GoToNext();
                      if (ret2 >= 0) {
                           if (_atomNames.GetCurrData() != index2) break;
                      }
                 }
            }
       }
       for (i = 0; i < num_loc; i++) {
            index1.Copy(_atoms[loc[i]]->restype());
            index1 += _atoms[loc[i]]->pdb_atmnam();
            _atomNames.GoToRecord(loc[i]);
            _atomNames.ReplData(index1);
            index1 += "_";
            index1 += _atoms[loc[i]]->alt_loc();
            _atomNameAlts.GoToRecord(loc[i]);
            _atomNameAlts.ReplData(index1);
       }

       num_loc = 0;
       exist = find_single(_ResName);
       if (exist) {
            for (i = 0; i < exist->num; i++) {
                 index1.Copy(_ResName);
                 index1 += exist->atom_list[i].old_name;
                 ret1 = _atomNames.Seek(index1);
                 while (ret1 >= 0) {
                      _atoms[ret1]->set_atmtype(exist->atom_list[i].new_name);
                      _atoms[ret1]->set_pdb_atmnam(exist->atom_list[i].new_name);
                      loc[num_loc] = ret1;
                      num_loc++;
                      ret1 = _atomNames.GoToNext();
                      if (ret1 >= 0) {
                           if (_atomNames.GetCurrData() != index1) break;
                      }
                 }
            }
       }
       for (i = 0; i < num_loc; i++) {
            index1.Copy(_atoms[loc[i]]->restype());
            index1 += _atoms[loc[i]]->pdb_atmnam();
            _atomNames.GoToRecord(loc[i]);
            _atomNames.ReplData(index1);
            index1 += "_";
            index1 += _atoms[loc[i]]->alt_loc();
            _atomNameAlts.GoToRecord(loc[i]);
            _atomNameAlts.ReplData(index1);
       }
       delete [] loc;
}

void _RESIDUE::correction_atom_name(CifString& error)
{
       int i = 0, j = 0, no_oxygen = 0, num = 0, n_list = 0;
       char value[15], alt_loc[5];
       _SIndex _index;
       CifString idx;
       BTreeObj<_SIndex> atomTree, bondTree;
       DiGraph *graph = NULL, *graph1 = NULL;
       _MATCH_LIST *list = NULL;
       ConnectFormat *drug = NULL;

       if (!strcmp(_ResName, "UNL")) return;

       if (!strcmp(_ResName, "DML") || !strcmp(_ResName, "TML")) {
            for (i = 0; i < _Num_Atom; i++) {
                 if (_deleted[i]) continue;
                 if (!strcmp(_atoms[i]->atmtype(), "CH1")) {
                      _atoms[i]->set_atmtype("CM1");
                      _atoms[i]->set_pdb_atmnam("CM1");
                 }
                 if (!strcmp(_atoms[i]->atmtype(), "CH2")) {
                      _atoms[i]->set_atmtype("CM2");
                      _atoms[i]->set_pdb_atmnam("CM2");
                 }
                 if (!strcmp(_atoms[i]->atmtype(), "CH3")) {
                      _atoms[i]->set_atmtype("CM3");
                      _atoms[i]->set_pdb_atmnam("CM3");
                 }
            }
            _UpdateIndices();
            return;
       }
       if (_atomNames.NumElem() == 1) {
            drug = find_drug(_ResName);
            if (drug && drug->natoms() == 1) {
                 for (i = 0; i < _Num_Atom; i++) {
                      if (_deleted[i]) continue;
                      _atoms[i]->set_atmtype(drug->atoms(0)->atomname());
                      _atoms[i]->set_pdb_atmnam(drug->atoms(0)->atomname());
                 }
                 _UpdateIndices();
                 return;
            }
       }
       if (_Num_Atom < 2) return;

       if ((!strcmp(_ResName, "A") || !strcmp(_ResName, "+A") ||
            !strcmp(_ResName, "C") || !strcmp(_ResName, "+C") ||
            !strcmp(_ResName, "G") || !strcmp(_ResName, "+G") ||
            !strcmp(_ResName, "I") || !strcmp(_ResName, "+I")) &&
            !find_atom("O2*"))
            no_oxygen = 1;

       graph1 = Graphs->find_graph(_ResName, no_oxygen);
       if (!graph1) {
            error += "\nCan't create Graph for ";
            error += _ResName;
            error += "\n";
            return;
       }

       get_atom_type();

       graph = new DiGraph(_Num_Atom);
       atomTree.Clear();
       memset(alt_loc, 0, 5);
       for (i = 0; i < _Num_Atom; i++) {
            if (_deleted[i]) continue;
            if (strcmp(_atoms[i]->alt_loc(), "")) {
                 if (!strcmp(alt_loc, "")) {
                      strcpy(alt_loc, _atoms[i]->alt_loc());
                 }
            }
            if (!strcmp(_atoms[i]->alt_loc(), "") ||
                !strcmp(_atoms[i]->alt_loc(), alt_loc)) {
                 _index.set_value(_atoms[i]->atmtype());
                 _index.set_pos(num);
                 if (atomTree.Seek(_index) >= 0) continue;
                 atomTree.Add(_index);
                 strcpy(value, _atoms[i]->atom_type());
                 if (!strcmp(value, ""))
                      ndb_get_atom_symbol(_atoms[i]->atmtype(), value);
                 graph->add_vertex(num, value, _atoms[i]->atmtype());
                 num++;
            }
       }
       if (num < 2) {
            delete graph;
            return;
       }

       bondTree.Clear();
       for (i = 0; i < _Num_Atom - 1; i++) {
            if (_deleted[i]) continue;
            if (!strcmp(_atoms[i]->alt_loc(), "") ||
                !strcmp(_atoms[i]->alt_loc(), alt_loc)) {
                 for (j = i + 1; j < _Num_Atom; j++) {
                      if (_deleted[j]) continue;
                      if (!strcmp(_atoms[i]->atom_type(), "H") &&
                          !strcmp(_atoms[j]->atom_type(), "H"))
                           continue;
/*
                      if ((!strcmp(_atoms[i]->alt_loc(), "") ||
                           !strcmp(_atoms[j]->alt_loc(), "") ||
                           !strcmp(_atoms[i]->alt_loc(),
                                   _atoms[j]->alt_loc())) &&
*/
                      if ((!strcmp(_atoms[j]->alt_loc(), "") ||
                           !strcmp(_atoms[j]->alt_loc(), alt_loc)) &&
                           is_a_bond(_atoms[i], _atoms[j])) {
                           if (strcmp(_atoms[i]->atmtype(),
                                      _atoms[j]->atmtype()) < 0)
                                sprintf(value, "%s%s", _atoms[i]->atmtype(),
                                      _atoms[j]->atmtype());
                           else sprintf(value, "%s%s", _atoms[j]->atmtype(),
                                      _atoms[i]->atmtype());
                           _index.set_value(value);
                           if (bondTree.Seek(_index) >= 0) continue;
                           bondTree.Add(_index);
                           graph->add_adjacent_vrtx(_atoms[i]->atmtype(),
                                     _atoms[j]->atmtype(), 1, 0);
                           graph->add_adjacent_vrtx(_atoms[j]->atmtype(),
                                     _atoms[i]->atmtype(), 1, 0);
                      }
                 }
            }
       }

       if (!graph->isconnect()) {
            error += "\nResidue (";
            error += GetFirstAtom()->pdb_chnid();
            error += " ";
            error += GetFirstAtom()->pdb_resnam();
            error += " ";
            error += GetFirstAtom()->pdb_resnum();
            error += ") is not connected\n";
            delete graph;
            return;
       }

       list = graph1->find_match(*graph, n_list);
       // if (n_list == num) {
       if (list) {
            for (i = 0; i < n_list; i++) {
                 if (strcmp(list[i].source, list[i].target)) {
                      if (!strcmp(list[i].source, "1H") ||
                          !strcmp(list[i].source, "2H") ||
                          !strcmp(list[i].source, "3H")) continue;
                      break;
                 }
            }
            if (i < n_list) {
                 for (i = 0; i < n_list; i++) {
                      idx.Copy(_ResName);
                      idx += list[i].source;
                      j = _atomNames.Seek(idx);
                      while (j >= 0) {
                           _atoms[j]->set_atmtype(list[i].target);
                           _atoms[j]->set_pdb_atmnam(list[i].target);
                           j = _atomNames.GoToNext();
                           if (j >= 0) {
                                if (_atomNames.GetCurrData() != idx) break;
                           }
                      }
                 }
                 _UpdateIndices();
            }
       } else {
            error += "\nCan't find match for residue (";
            error += GetFirstAtom()->pdb_chnid();
            error += " ";
            error += GetFirstAtom()->pdb_resnam();
            error += " ";
            error += GetFirstAtom()->pdb_resnum();
            error += ")\n";
       }
       if (list) delete [] list;
       delete graph;
}

void _RESIDUE::recover_pdb_residue_name()
{
       int k;
       if (_Num_Atom == 0) return;
       strcpy(_ResName, _atoms[0]->pdb_resnam());
       for (k = 0; k < _Num_Atom; k++) {
            _atoms[k]->set_restype(_atoms[k]->pdb_resnam());
            _atoms[k]->set_atmtype(_atoms[k]->pdb_atmnam());
       }
}

void _RESIDUE::update_nomenclature(const char *type, const char *chain_id, const int res_num)
{
       int i;
       _res_no = res_num;
       for (i = 0; i < _Num_Atom; i++) {
            if (strcmp(type, ""))
                 _atoms[i]->set_type(type);
            _atoms[i]->set_chnid(chain_id);
            _atoms[i]->set_resnum(res_num);
       }
}

void _RESIDUE::update_nomenclature(const char *type, const char *chain_id, const int res_num,
                 const char *pdb_chainid, const int pdb_res_num)
{
       int i;
       _res_no = res_num;
       _pdb_res_no = pdb_res_num;
       for (i = 0; i < _Num_Atom; i++) {
            if (strcmp(type, ""))
                 _atoms[i]->set_type(type);
            _atoms[i]->set_chnid(chain_id);
            _atoms[i]->set_resnum(res_num);
            _atoms[i]->set_pdb_chnid(pdb_chainid);
            _atoms[i]->set_pdb_resnum(pdb_res_num);
       }
}

void _RESIDUE::correction_name(const char *name)
{
       strcpy(_ResName, name);
       for (int i = 0; i < _Num_Atom; i++) {
            _atoms[i]->set_restype(name);
            _atoms[i]->set_pdb_resnam(name);
       }
}

void _RESIDUE::convert_between_orthogonal_and_fractional(const NDBCELLCONST *cell, const int iflag)
{
       int k;
       COORD coord;
       SIGATM sigatm;
       double x, y, z;
       for (k = 0; k < _Num_Atom; k++) {
            if (_deleted[k]) continue;
            coord = _atoms[k]->orig();
            ndb_trans_coord_3by3(coord, cell, iflag);
            _atoms[k]->set_orig(coord);
            if (_atoms[k]->is_sigatm()) {
                 sigatm = _atoms[k]->sigatm();
                 x = sigatm.sig_x;
                 y = sigatm.sig_y;
                 z = sigatm.sig_z;
                 ndb_trans_coord_3by3(&x, &y, &z, cell, iflag);
                 sigatm.sig_x = x;
                 sigatm.sig_y = y;
                 sigatm.sig_z = z;
                 _atoms[k]->set_sigatm(sigatm);
            }
       }
}

void _RESIDUE::symmetry_operations(const int lx, const int ly, const int lz, NDBSYMMETRY* op,
                const int not_moving)
{
       int k = 0, exist_anisou = 0;
       COORD coord;
       for (k = 0; k < _Num_Atom; k++) {
            if (_deleted[k]) continue;
            if (_atoms[k]->is_anisou()) {
                 exist_anisou = 1;
                 break;
            }
       }
       if (exist_anisou && not_moving) return;

       for (k = 0; k < _Num_Atom; k++) {
            if (_deleted[k]) continue;
            do_symmetry_operation(coord, _atoms[k]->orig(), op);
            coord.x += (float) lx;
            coord.y += (float) ly;
            coord.z += (float) lz;
            _atoms[k]->set_orig(coord);
       }
}

void _RESIDUE::transformation(double m[3][4])
{
       int k;
       COORD coord;
       for (k = 0; k < _Num_Atom; k++) {
            if (_deleted[k]) continue;
            coord_transformation(coord, _atoms[k]->orig(), m);
            _atoms[k]->set_orig(coord);
       }
}

static int get_align(char *atomName, char *atom_type)
{
       int n, len;
       char *p;

       if (strlen(atomName) == 4) return 0;

       p = strstr(atomName, atom_type);
       n = p - atomName;
       len = strlen(atom_type);
       if (n == 1 || len == 2 || len == 1 && atomName[1] == atom_type[0] &&
           atom_type[0] != 'D' && atom_type[0] != 'H' && atom_type[0] != 'C')
            return 0;
       else return 1;
}

static void write_error_message(_ATOM *new_atom, _ATOM *old_atom, CifString& error, const int flag)
{
       char buffer[80], chn_id1[5], inscode1[5], alt_loc1[5], chn_id2[5], inscode2[5], alt_loc2[5];
       strcpy(chn_id1, new_atom->pdb_chnid());
       strcpy(inscode1, new_atom->ins_code());
       strcpy(alt_loc1, new_atom->alt_loc());
       strcpy(chn_id2, old_atom->pdb_chnid());
       strcpy(inscode2, old_atom->ins_code());
       strcpy(alt_loc2, old_atom->alt_loc());
       if (!strcmp(chn_id1, "")) strcpy(chn_id1, " ");
       if (!strcmp(inscode1, "")) strcpy(inscode1, " ");
       if (!strcmp(alt_loc1, "")) strcpy(alt_loc1, " ");
       if (!strcmp(chn_id2, "")) strcpy(chn_id2, " ");
       if (!strcmp(inscode2, "")) strcpy(inscode2, " ");
       if (!strcmp(alt_loc2, "")) strcpy(alt_loc2, " ");
       if (flag) error += "\nMissing Alternate Conformation ID:\n";
       else error += "\nAtoms with same coordinates:\n";
       sprintf(buffer, FORMAT, new_atom->atnum(), new_atom->pdb_atmnam(),
            alt_loc1[0], new_atom->pdb_resnam(), chn_id1,
            new_atom->pdb_resnum(), inscode1[0], new_atom->orig().x,
            new_atom->orig().y, new_atom->orig().z, new_atom->occ(),
            new_atom->t_fct());
       error += buffer;
       sprintf(buffer, FORMAT, old_atom->atnum(), old_atom->pdb_atmnam(),
            alt_loc2[0], old_atom->pdb_resnam(), chn_id2,
            old_atom->pdb_resnum(), inscode2[0], old_atom->orig().x,
            old_atom->orig().y, old_atom->orig().z, old_atom->occ(),
            old_atom->t_fct());
       error += buffer;
}

static void write_occupancy_message(CifString& error, _ATOM *atom, const char *name, const float occupancy)
{
       char buffer[80];
       if (strcmp(atom->ins_code(), "") && strcmp(atom->pdb_chnid(), ""))
            sprintf(buffer, "Atom '%s %s %d%s %s' has occupancy %.3f\n", name,
                 atom->pdb_resnam(), atom->pdb_resnum(), atom->ins_code(),
                 atom->pdb_chnid(), occupancy);
       else if (strcmp(atom->ins_code(), ""))
            sprintf(buffer, "\nAtom '%s %s %d%s' has occupancy %.3f\n", name,
                 atom->pdb_resnam(), atom->pdb_resnum(), atom->ins_code(), occupancy);
       else if (strcmp(atom->pdb_chnid(), ""))
            sprintf(buffer, "\nAtom '%s %s %d %s' has occupancy %.3f\n", name,
                 atom->pdb_resnam(), atom->pdb_resnum(), atom->pdb_chnid(), occupancy);
       else sprintf(buffer, "\nAtom '%s %s %d' has occupancy %.3f\n", name,
                 atom->pdb_resnam(), atom->pdb_resnum(), occupancy);
       error += buffer;
}
