/*
FILE:     cry_util.C
*/
/*
VERSION:  1.700
*/
/*
DATE:     7/13/2005
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2005 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "ndb_misclib.h"
#include "erf_iolib.h"
#include "cry_util.h"

static void ndb_sym_rot_op(const int *rot, char *sym_name);
static void mult_matrixes(const double matrix1[3][3], const double matrix2[3][3],
             double result[3][3]);
static void mult_matrix_vector(const double matrix[3][3], const double vector[3],
             double result[3]);

void symmetry_operation(COORD &coord, const NDBCELLCONST *cell, const int op,
                 const int lx, const int ly, const int lz)
{
       ndb_trans_coord_3by3(coord, cell, NDB_TRANS_ORTHO_TO_FRAC);
       do_symmetry_operation(coord, coord, &(cell->symops[op]));
       coord.x += (double) lx;
       coord.y += (double) ly;
       coord.z += (double) lz;
       ndb_trans_coord_3by3(coord, cell, NDB_TRANS_FRAC_TO_ORTHO);
}

void symmetry_operation(double *x, double *y, double *z,
        const NDBCELLCONST *cell, const int op, const int lx,
        const int ly, const int lz)
{
       double x1, y1, z1;
       ndb_trans_coord_3by3(x, y, z, cell, NDB_TRANS_ORTHO_TO_FRAC);

       x1 = *x; y1 = *y; z1 = *z;
       *x = x1 * cell->symops[op].rot[0][0] +
            y1 * cell->symops[op].rot[0][1] +
            z1 * cell->symops[op].rot[0][2] +
            cell->symops[op].trans[0] + (double) lx;
       *y = x1 * cell->symops[op].rot[1][0] +
            y1 * cell->symops[op].rot[1][1] +
            z1 * cell->symops[op].rot[1][2] +
            cell->symops[op].trans[1] + (double) ly;
       *z = x1 * cell->symops[op].rot[2][0] +
            y1 * cell->symops[op].rot[2][1] +
            z1 * cell->symops[op].rot[2][2] +
            cell->symops[op].trans[2] + (double) lz;

       ndb_trans_coord_3by3(x, y, z, cell, NDB_TRANS_FRAC_TO_ORTHO);
}

void do_symmetry_operation(COORD &new_coord, COORD old_coord, NDBSYMMETRY *op)
{
       new_coord.x = old_coord.x * op->rot[0][0] 
                   + old_coord.y * op->rot[0][1]
                   + old_coord.z * op->rot[0][2] + op->trans[0];
       new_coord.y = old_coord.x * op->rot[1][0]
                   + old_coord.y * op->rot[1][1]
                   + old_coord.z * op->rot[1][2] + op->trans[1];
       new_coord.z = old_coord.x * op->rot[2][0]
                   + old_coord.y * op->rot[2][1]
                   + old_coord.z * op->rot[2][2] + op->trans[2];
}

void ndb_trans_coord_3by3(double *x, double *y, double *z,
      const NDBCELLCONST *cell, int iflag)
{
       int i, j;
       double x0[3], t[3];

       x0[0] = *x;
       x0[1] = *y;
       x0[2] = *z;

       if (iflag == NDB_TRANS_ORTHO_TO_FRAC) {
            for (i = 0; i < 3; i++) {
                t[i] = 0.0;
                for (j = 0; j < 3; j++) {
                    t[i] = t[i] + cell->o_to_f[i][j]*x0[j];
                }
            }
            *x = t[0] + cell->o_to_f_v[0];
            *y = t[1] + cell->o_to_f_v[1];
            *z = t[2] + cell->o_to_f_v[2];
       } else if (iflag == NDB_TRANS_FRAC_TO_ORTHO) {
            for (i = 0; i < 3; i++) {
                 x0[i] -= cell->o_to_f_v[i];
            }
            for (i = 0; i < 3; i++) {
                t[i] = 0.0;
                for (j = 0; j < 3; j++) {
                    t[i] = t[i] + cell->f_to_o[i][j]*x0[j];
                }
            }
            *x = t[0] /* - cell->o_to_f_v[0] */ ;
            *y = t[1] /* - cell->o_to_f_v[1] */ ;
            *z = t[2] /* - cell->o_to_f_v[2] */ ;
       }
}

void ndb_trans_coord_3by3(COORD &coord, const NDBCELLCONST *cell, int iflag)
{
       int i, j;
       double x0[3], t[3];

       x0[0] = coord.x;
       x0[1] = coord.y;
       x0[2] = coord.z;

       if (iflag == NDB_TRANS_ORTHO_TO_FRAC) {
            for (i = 0; i < 3; i++) {
                t[i] = 0.0;
                for (j = 0; j < 3; j++) {
                    t[i] = t[i] + cell->o_to_f[i][j] * x0[j];
                }
            }
            coord.x = t[0] + cell->o_to_f_v[0];
            coord.y = t[1] + cell->o_to_f_v[1];
            coord.z = t[2] + cell->o_to_f_v[2];
       } else if (iflag == NDB_TRANS_FRAC_TO_ORTHO) {
            for (i = 0; i < 3; i++) {
                 x0[i] -= cell->o_to_f_v[i];
            }
            for (i = 0; i < 3; i++) {
                t[i] = 0.0;
                for (j = 0; j < 3; j++) {
                    t[i] = t[i] + cell->f_to_o[i][j] * x0[j];
                }
            }
            coord.x = t[0] /* - cell->o_to_f_v[0] */ ;
            coord.y = t[1] /* - cell->o_to_f_v[1] */ ;
            coord.z = t[2] /* - cell->o_to_f_v[2] */ ;
       }
}

void ndb_get_symmetry_operation_name(const NDBCELLCONST *cell, const int op,
          const int lx, const int ly, const int lz, char *sym_name)
{
       int i, j, len, t[3], rot[3][3], tn[3], td[3];
       char buffer[80];

       sym_name[0] = '\0';
       if (cell == NULL) return;

       t[0] = lx; t[1] = ly; t[2] = lz;
       for (i = 0; i < 3; i++) {
            for (j = 0; j < 3; j++)
                 rot[i][j] = (int) cell->symops[op].rot[i][j];
            tn[i] = cell->symops[op].tn[i];
            td[i] = cell->symops[op].td[i];
            if (td[i] != 0) tn[i] += t[i] * td[i];
            else tn[i] = t[i];
            buffer[0] = '\0';
            ndb_sym_rot_op(&rot[i][0], buffer);
            len = strlen(sym_name);
            if (tn[i] != 0) {
                 if (td[i]) {
                      if (tn[i] / td[i] * td[i] == tn[i]) {
                           tn[i] = tn[i] / td[i];
                           sprintf(&sym_name[len], "%d", tn[i]);
                      } else sprintf(&sym_name[len], "%d/%d", tn[i], td[i]);
                 } else sprintf(&sym_name[len], "%d", tn[i]);
                 len = strlen(sym_name);
                 if (buffer[0] == '-')
                      sprintf(&sym_name[len], "%s", buffer);
                 else sprintf(&sym_name[len], "+%s", buffer);
/*
            } else if (t[i] != 0) {
                 sprintf(&sym_name[len], "%d", t[i]);
                 len = strlen(sym_name);
                 if (buffer[0] == '-')
                      sprintf(&sym_name[len], "%s", buffer);
                 else sprintf(&sym_name[len], "+%s", buffer);
*/
            } else sprintf(&sym_name[len], "%s", buffer);
            len = strlen(sym_name);
            if (i < 2) sprintf(&sym_name[len], ",");
            // else       sprintf(&sym_name[len], "");
            else sym_name[len] = '\0';
       }
}

void rcsb_get_symmetry_operation_name(const NDBCELLCONST *cell, char *sym_op, char *sym_name)
{
       int j, k, op, len, t[3];

       sym_name[0] = '\0';
       if (cell == NULL) return;

       op = atoi(sym_op) - 1;
       t[0] = t[1] = t[2] = 0;
       len = strlen(sym_op);
       for (j = 0; j < len; j++) {
            if (sym_op[j] == '_') break;
       }
       if (j + 3 < len) {
            for (k = 0; k < 3; k++) {
                 t[k] = sym_op[j + k + 1] - '5';
            }
       }
       ndb_get_symmetry_operation_name(cell, op, t[0], t[1], t[2], sym_name);
}

void ndb_get_symmetry_matrix(const NDBCELLCONST *cell, const int op,
       const int lx, const int ly, const int lz, double matrix[3][3],
       double vector[3])
{
       int i, j;
       double matrix1[3][3], tt[3], tr1[3], tr2[3], tr3[3];

       tt[0] = (double) lx + cell->symops[op].trans[0];
       tt[1] = (double) ly + cell->symops[op].trans[1];
       tt[2] = (double) lz + cell->symops[op].trans[2];

       mult_matrixes(cell->f_to_o, cell->symops[op].rot, matrix1);
       mult_matrixes(matrix1, cell->o_to_f, matrix);

       for (i = 0; i < 3; i++)
            for (j = 0; j < 3; j++) {
                 if (fabs(matrix[i][j]) < 0.0001) matrix[i][j] = 0.0;
                 if (fabs(1.0 - fabs(matrix[i][j])) < 0.0001)
                      matrix[i][j] = matrix[i][j] / fabs(matrix[i][j]);
       }

       mult_matrix_vector(matrix1, cell->symops[op].trans, tr1);
       mult_matrix_vector(cell->f_to_o, tt, tr2);
       mult_matrix_vector(cell->f_to_o, cell->symops[op].trans, tr3);

       for (i = 0; i < 3; i++) {
            vector[i] = tr2[i];
            if (fabs(vector[i]) < 0.0001) vector[i] = 0.0;
       }
}

void ndb_get_symmetry_operation(const NDBCELLCONST *cell, double matrix[3][3],
       double vector[3], char *symmetry)
{
       double matrix1[3][3], rot[3][3], tt[3];

       *symmetry = '\0';
       mult_matrixes(cell->o_to_f, matrix, matrix1);
       mult_matrixes(matrix1, cell->f_to_o, rot);
       mult_matrix_vector(cell->o_to_f, vector, tt);
       ndb_get_symmetry_operator(cell, rot, tt, symmetry);
}

void ndb_get_symmetry_operator(const NDBCELLCONST *cell, double matrix[3][3],
       double vector[3], char *symmetry)
{
       int i, j, k, found, x, y, op = 1, lx = 0, ly = 0, lz = 0;
       double r;

       *symmetry = '\0';
       for (i = 0; i < cell->nop; i++) {
            found = 1;
            for (j = 0; j < 3; j++) {
                 for (k = 0; k < 3; k++) {
                      x = int(rint(matrix[j][k]));
                      y = int(rint(cell->symops[i].rot[j][k]));
                      if (x != y) {
                           found = 0;
                           break;
                      }
                 }
                 if (!found) break;
            }
            if (found) {
                 r = vector[0] - cell->symops[i].trans[0];
                 lx = int(rint(r));
                 if (fabs(r - (double) lx) > 0.1) continue;
                 r = vector[1] - cell->symops[i].trans[1];
                 lx = int(rint(r));
                 if (fabs(r - (double) lx) > 0.1) continue;
                 r = vector[2] - cell->symops[i].trans[2];
                 lx = int(rint(r));
                 if (fabs(r - (double) lx) > 0.1) continue;
                 op = i + 1;
                 break;
            }
       }
       if (i == cell->nop) return;

       lx = int(rint(vector[0] - cell->symops[op-1].trans[0]));
       ly = int(rint(vector[1] - cell->symops[op-1].trans[1]));
       lz = int(rint(vector[2] - cell->symops[op-1].trans[2]));
       sprintf(symmetry, "%d_%d%d%d", op, lx + 5, ly + 5, lz + 5);
}

void ndb_ortho_to_frac_matrix(NDBCELLCONST *cell)
{
       int i;
       double rad, va, aa[3], ang[3], alphr[3], cosa[3], sina[3];

       aa[0] = cell->a;
       aa[1] = cell->b;
       aa[2] = cell->c;

       ang[0] = cell->alpha;
       ang[1] = cell->beta;
       ang[2] = cell->gamma;

       rad = acos(-1.0)/180.0;
       for (i = 0; i < 3; i++) {
            alphr[i] = ang[i] * rad;
            cosa[i]  = cos(alphr[i]);
            sina[i]  = sin(alphr[i]);
            if (fabs(ang[i] - 90.0) < 0.001) {
                 cosa[i] = 0.0;
                 sina[i] = 1.0;
            }
       }
       va = sqrt(1.00 + 2.0 * cosa[0] * cosa[1] * cosa[2] - cosa[0] * cosa[0]
                 - cosa[1] * cosa[1] - cosa[2] * cosa[2]);
     
       cell->o_to_fc[0][0] = 1.0 / aa[0];
       cell->o_to_fc[0][1] = -cosa[2] / (aa[0] * sina[2]);
       cell->o_to_fc[0][2] = (cosa[2] * cosa[0]-cosa[1]) / (aa[0] * va * sina[2]);
       cell->o_to_fc[1][0] = 0.0;
       cell->o_to_fc[1][1] = 1.0 / (aa[1] * sina[2]);
       cell->o_to_fc[1][2] = (cosa[1] * cosa[2]-cosa[0]) / (aa[1] * va * sina[2]);
       cell->o_to_fc[2][0] = 0.0;
       cell->o_to_fc[2][1] = 0.0;
       cell->o_to_fc[2][2] = sina[2] / (aa[2] * va);

       cell->o_to_f[0][0] = 1.0 / aa[0];
       cell->o_to_f[0][1] = -cosa[2] / (aa[0] * sina[2]);
       cell->o_to_f[0][2] = (cosa[2] * cosa[0]-cosa[1]) / (aa[0] * va * sina[2]);
       cell->o_to_f[1][0] = 0.0;
       cell->o_to_f[1][1] = 1.0 / (aa[1] * sina[2]);
       cell->o_to_f[1][2] = (cosa[1] * cosa[2]-cosa[0]) / (aa[1] * va * sina[2]);
       cell->o_to_f[2][0] = 0.0;
       cell->o_to_f[2][1] = 0.0;
       cell->o_to_f[2][2] = sina[2] / (aa[2] * va);

       cell->o_to_f_v[0] = cell->o_to_f_v[1] = cell->o_to_f_v[2] = 0.0;
}

void ndb_frac_to_ortho_matrix(NDBCELLCONST *cell)
{
       int i;
       double rad, va, v, aa[3], ang[3], alphr[3], cosa[3], sina[3];

       aa[0] = cell->a;
       aa[1] = cell->b;
       aa[2] = cell->c;

       ang[0] = cell->alpha;
       ang[1] = cell->beta;
       ang[2] = cell->gamma;

       rad = acos(-1.0) / 180.0;
       for (i = 0; i < 3; i++) {
            alphr[i] = ang[i] * rad;
            cosa[i]  = cos(alphr[i]);
            sina[i]  = sin(alphr[i]);
            if (fabs(ang[i] - 90.0) < 0.001) {
                 cosa[i]=0.0;
                 sina[i]=1.0;
            }
       }
       va = sqrt(1.0 + 2.0 * cosa[0] * cosa[1] * cosa[2] - cosa[0] * cosa[0]
                 - cosa[1] * cosa[1] - cosa[2] * cosa[2]);
       v = aa[0] * aa[1] * aa[2] * va;

       cell->f_to_o[0][0] = aa[0];
       cell->f_to_o[0][1] = aa[1] * cosa[2];
       cell->f_to_o[0][2] = aa[2] * cosa[1];
       cell->f_to_o[1][0] = 0.0;
       cell->f_to_o[1][1] = aa[1] * sina[2];
       cell->f_to_o[1][2] = aa[2] * (cosa[0] - cosa[1] * cosa[2]) / sina[2];
       cell->f_to_o[2][0] = 0.0;
       cell->f_to_o[2][1] = 0.0;
       cell->f_to_o[2][2] = v / (aa[0] * aa[1] * sina[2]);
}

int ndb_get_sym_data(NDBCELLCONST *cell)
{
       int i, j, k, not_found = 0;
       ndb_clean_string(cell->space_group);
       SPACE_GROUP *spacegroup = get_spacegroup_info(cell->space_group);
       if (!spacegroup) {
            spacegroup = get_spacegroup_info("P 1");
            not_found = 1;
       }
       if (spacegroup) {
            cell->nop = spacegroup->SpaceDataNo;
            cell->symops = new NDBSYMMETRY[cell->nop];
            for (i = 0; i < cell->nop; i++) {
                 for (j = 0; j < 3; j++) {
                      for (k = 0; k < 3; k++) {
                           cell->symops[i].rot[j][k] = 
                              (double) spacegroup->SpaceData[i].Data[j*3+k];
                      }
                      cell->symops[i].tn[j] = spacegroup->SpaceData[i].Data[9+j*2];
                      cell->symops[i].td[j] = spacegroup->SpaceData[i].Data[9+j*2+1];
                      cell->symops[i].trans[j] = 0.0;
                      if (cell->symops[i].td[j] != 0)
                           cell->symops[i].trans[j] = (double) cell->symops[i].tn[j]
                                                    / (double) cell->symops[i].td[j];
                 }
            }
       } else {
            cell->nop = 0;
            cell->symops = NULL;
       }
       return not_found;
}

int check_scale_token(NDBCELLCONST *cell, double scale[3][3], double vec[3])
{
       int ierror = FALSE, wrong_scale = 0, i, j;
       double v = 0;

       for (i = 0; i < 3; i++) {
            for (j = 0; j < 3; j++) {
                 if (fabs(scale[i][j]) > 0.00001) {
                      if (fabs(scale[i][j]) > 0.8) wrong_scale = 1; 
                      ierror = TRUE; break;
                 }
            }
            if (ierror) break;
       }
       if (!ierror || wrong_scale) return 1;

       ierror = FALSE;
       for (i = 0; i < 3; i++) cell->o_to_f_v[i] = vec[i];
       for (i = 0; i < 3; i++) {
            for (j = 0; j < 3; j++) {
                 if (fabs(scale[i][j] - cell->o_to_f[i][j]) > 0.00005)
                      ierror = TRUE;
            }
       }
       if (!ierror) return 0;

       for (i = 0; i < 3; i++) {
            for (j = 0; j < 3; j++) {
                 cell->o_to_f[i][j] = scale[i][j];
            }
       }

       v = scale[0][0] * scale[1][1] * scale[2][2] +
           scale[1][0] * scale[2][1] * scale[0][2] +
           scale[2][0] * scale[1][2] * scale[0][1] -
          (scale[0][2] * scale[1][1] * scale[2][0] +
           scale[0][0] * scale[1][2] * scale[2][1] +
           scale[1][0] * scale[2][2] * scale[0][1]);

       if (fabs(v) > 0.0) {
            cell->f_to_o[0][0] = (scale[1][1] * scale[2][2] - 
                                  scale[1][2] * scale[2][1]) / v;
            cell->f_to_o[0][1] = (scale[2][1] * scale[0][2] -
                                  scale[2][2] * scale[0][1]) / v;
            cell->f_to_o[0][2] = (scale[0][1] * scale[1][2] -
                                  scale[1][1] * scale[0][2] ) / v;
            cell->f_to_o[1][0] = (scale[1][2] * scale[2][0] -
                                  scale[2][2] * scale[1][0] ) / v;
            cell->f_to_o[1][1] = (scale[0][0] * scale[2][2] -
                                  scale[0][2] * scale[2][0] ) / v;
            cell->f_to_o[1][2] = (scale[0][2] * scale[1][0] -
                                  scale[1][2] * scale[0][0] ) / v;
            cell->f_to_o[2][0] = (scale[1][0] * scale[2][1] -
                                  scale[2][0] * scale[1][1] ) / v;
            cell->f_to_o[2][1] = (scale[2][0] * scale[0][1] -
                                  scale[2][1] * scale[0][0] ) / v;
            cell->f_to_o[2][2] = (scale[0][0] * scale[1][1] -
                                  scale[1][0] * scale[0][1] ) / v;
       }
       return 0;
}

void coord_transformation(COORD &new_coord, COORD old_coord, double m[3][4])
{
       new_coord.x = old_coord.x * m[0][0] + old_coord.y * m[0][1]
                   + old_coord.z * m[0][2] + m[0][3];
       new_coord.y = old_coord.x * m[1][0] + old_coord.y * m[1][1]
                   + old_coord.z * m[1][2] + m[1][3];
       new_coord.z = old_coord.x * m[2][0] + old_coord.y * m[2][1]
                   + old_coord.z * m[2][2] + m[2][3];
}

static void ndb_sym_rot_op(const int *rot, char *sym_name)
{
       int i, len;
       for (i = 0; i < 3; i++) {
            len = strlen(sym_name);
            if (rot[i] > 0) {
                 if (i == 0) sprintf(&sym_name[len], "%c", 120+i);
                 else sprintf(&sym_name[len], "+%c", 120+i);
            } else if (rot[i] < 0) {
                 sprintf(&sym_name[len], "-%c", 120+i);
            }
       }
       if (sym_name[0] == '+') {
            len = strlen(sym_name);
            for (i = 0; i < len; i++)
                 sym_name[i] = sym_name[i+1];
       }
}

static void mult_matrixes(const double matrix1[3][3], const double matrix2[3][3],
             double result[3][3])
{
       int i, j, k;
       for (i = 0; i < 3; i++) {
            for (j = 0; j < 3; j++) {
                 result[i][j] = 0.0;
                 for (k = 0; k < 3; k++)
                      result[i][j] += (matrix1[i][k] * matrix2[k][j]);
            }
       }
}

static void mult_matrix_vector(const double matrix[3][3], const double vector[3],
             double result[3])
{
       int i, j;
       for (i = 0; i < 3; i++) {
            result[i] = 0.0;
            for (j = 0; j < 3; j++)
                 result[i] += (matrix[i][j] * vector[j]);
       }
}

