
/*
FILE:     dict_exchange.C
*/
/*
VERSION:  1.700
*/
/*
DATE:     7/13/2005
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2005 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <sys/stat.h>

#include "dict_exchange.h"
#include "erf_util.h"

#define NUM_PATCH    3
#define NUM_PAIR     7

typedef struct {
       const char *old_item;
       const char *new_item;
} _ITEM_PAIR;

typedef struct {
       const char *category;
       int num_pair;
       _ITEM_PAIR item_pair[NUM_PAIR];
} _CATEGORY_PATCH;

static const _CATEGORY_PATCH _category_patch[NUM_PATCH] = {
       { "refine", 7,
         { { "solvent_vdw_probe_radii", "ccp4_solvent_vdw_probe_radii" },
           { "solvent_ion_probe_radii", "ccp4_solvent_ion_probe_radii" },
           { "solvent_shrinkage_radii", "ccp4_solvent_shrinkage_radii" },
           { "sovent_shrinkage_radii",  "ccp4_solvent_shrinkage_radii" },
           { "rcsb_overall_ESU_R",      "ebi_overall_ESU_R"            },
           { "ebi_overall_ESU_ML",      "overall_SU_ML"                },
           { "ebi_overall_ESU_B",       "overall_SU_B"                 } }
       },
       { "refine_hist", 2,
         { { "rcsb_number_atoms_solvent", "number_atoms_solvent" },
           { "rcsb_number_atoms_total",   "number_atoms_total"   } }
       },
       { "struct_ncs_dom_lim", 1,
         { { "pdbx_component", "pdbx_component_id" } }
       }
};

static int exchange_status = 0;

static DICFileObj *dict = NULL;
static ISTable *aliases = NULL;
static ISTable *items = NULL;
static ReVarCifArray<CifString> queryCol1;
static ReVarCifArray<CifString> queryCol2;
static int iCol1, colIndex;

static int read_binary_dictionary();
static int GetCategoryNameFromItemName(char categoryName[MAX_CIF_NAME_LENGTH],
                                       const char *itemName);

static int GetKeywordFromItemName(char keyword[MAX_CIF_NAME_LENGTH],
                                  const char *itemName);

void set_exchange_status(const int status)
{
       int ret = 1;

       if (status != INTERNAL_TO_EXCHANGE &&
           status != EXCHANGE_TO_INTERNAL) return;

       if (!dict) ret = read_binary_dictionary();
       if (!ret || !aliases || !items) return;

       if (exchange_status == status) return;

       exchange_status = status;

       queryCol1.Clear();
       if (status == INTERNAL_TO_EXCHANGE) {
            colIndex = aliases->GetColumnIndex("name");
            queryCol1.Add("alias_namelc");
       } else if (status == EXCHANGE_TO_INTERNAL) {
            colIndex = aliases->GetColumnIndex("alias_name");
            queryCol1.Add("namelc");
       }
       queryCol1.Add("dictionary");
       queryCol1.Add("version");
       aliases->CreateIndex("ind0", queryCol1);

       queryCol2.Clear();
       queryCol2.Add("namelc");
       items->CreateIndex("ind1", queryCol2);
}

CifFileObj *dictionary_exchange(CifFileObj *CifObj, const char *temp_file)
{
       int i = 0, numBlock = 0, errCode, nTables = 0, nCols = 0, it = 0;
       int ic = 0, iAliasCol = 0, iOrgCol = 0, iExistTab = 0, nRows = 0;
       int remove_pdb_id_field = 0, index = -1;
       char **blockName = NULL, StructureId[100];
       char rColName[MAX_CIF_NAME_LENGTH], rCatName[MAX_CIF_NAME_LENGTH];
       char **tableList = NULL;
       CifFileObj *fobjOut = NULL;
       CifString cs, cs1, aliasName, itemName;
       ISTable *t = NULL, *tAlias = NULL;
       ReVarCifArray<CifString> queryTarget1, queryTarget2;
       ReVarCifArray<CifString> *colNames = NULL, *colData = NULL;
       ReVarPCifArray<int> *queryResult1 = NULL, *queryResult2 = NULL;

       if (!exchange_status) return CifObj;

       blockName = CifObj->GetBlockNames(numBlock);
       strcpy(StructureId, blockName[0]);
       for (i = 0; i < numBlock; i++) delete [] blockName[i];
       delete [] blockName;

       if (exchange_status == EXCHANGE_TO_INTERNAL) {
            t = CifObj->GetTablePtr(StructureId, "struct_biol_gen");
            if (t) {
                 index = t->GetColumnIndex("ndb_new_pdb_asym_id");
                 if (index < 0) remove_pdb_id_field = 1;
                 else {
                      nRows = t->GetNumRows();
                      for (i = 0; i < nRows; i++) {
                           t->GetCell(cs, index, i);
                           if (cs != "." && cs != "?" && cs != "") break;
                      }
                      if (i == nRows) remove_pdb_id_field = 1;
                 }
            }
       }

       fobjOut = new CifFileObj(temp_file, WRITE_MODE);

       tableList = CifObj->GetTableNames(StructureId, nTables);
       for (it = 0; it < nTables; it++) {
            t = CifObj->GetTablePtr(StructureId, tableList[it]);
            nCols = t->GetNumColumns();
            colNames = t->GetColumnNames();
            for (ic = 0; ic < nCols; ic++) {
                 cs.Clear(); cs.Copy("_");
                 cs += tableList[it]; cs += "."; cs += (*colNames)[ic].Text();
                 cs1.Copy(cs.Text());
                 cs.ToLower();
                 queryTarget1.Clear();
                 queryTarget1.Add(cs);
                 queryTarget1.Add("cif_rcsb.dic");
                 queryTarget1.Add("1.1");
                 queryResult1 = aliases->Search(queryTarget1, queryCol1, errCode);
                 if (errCode < 0) {
                      queryTarget2.Clear();
                      queryTarget2.Add(cs);
                      queryResult2 = items->Search(queryTarget2,queryCol2,errCode);
                      if (errCode < 0) {
                        // cout << "Warning skipping item "<< cs.Text() 
                        //      << " is NOT in dictionary." << endl;
                        if (exchange_status == INTERNAL_TO_EXCHANGE) {
                             if (queryResult1) delete queryResult1;
                             if (queryResult2) delete queryResult2;
                             continue;
                        }
                        GetKeywordFromItemName(rColName, cs1.Text());
                        GetCategoryNameFromItemName(rCatName, cs1.Text());
                      } else {
                           itemName.Clear();
                           items->GetCell(itemName, iCol1, (*queryResult2)[0]);
                           GetKeywordFromItemName(rColName,itemName.Text());
                           GetCategoryNameFromItemName(rCatName,itemName.Text());
                      }
                      if (queryResult2) delete queryResult2;
                 } else {
                      aliasName.Clear();
                      aliases->GetCell(aliasName,colIndex,(*queryResult1)[0]);
                      GetKeywordFromItemName(rColName,aliasName.Text());
                      GetCategoryNameFromItemName(rCatName,aliasName.Text());
                 }

                 if (queryResult1) delete queryResult1;

                 if (!strcmp(rCatName, "")) continue;

                 iExistTab = fobjOut->IsTablePresent(StructureId, rCatName);
                 if (iExistTab) {
                      tAlias = fobjOut->GetTablePtr(StructureId, rCatName);
                      iAliasCol = tAlias->GetColumnIndex(rColName);
                      if (iAliasCol >= 0) {
/*
                           cerr << "Alias " << rCatName << "." << rColName
                              << " maps to existing column - exiting" << endl;
                           exit(1);
*/
                           ndb_log_message(NDB_MSG_ERROR, "Alias %s.%s maps to existing column - exiting\n",
                                   rCatName, rColName);
                      } else {  // Copy column to existing table
                           tAlias->AddColumn(rColName);
                           iAliasCol = tAlias->GetColumnIndex(rColName);
                           iOrgCol = t->GetColumnIndex((*colNames)[ic].Text());
                           colData = t->GetColumn(iOrgCol);
                           tAlias->FillColumn(*colData, iAliasCol);
                           fobjOut->WriteTable(tAlias, StructureId, rCatName);
                           if (colData) delete colData;
                      }
                 } else {
                      tAlias = new ISTable(rCatName);
                      tAlias->AddColumn(rColName);
                      iAliasCol = tAlias->GetColumnIndex(rColName);
                      iOrgCol = t->GetColumnIndex((*colNames)[ic].Text());
                      colData = t->GetColumn(iOrgCol);
                      tAlias->FillColumn(*colData, iAliasCol);
                      fobjOut->WriteTable(tAlias, StructureId, rCatName);
                      if (colData) delete colData;
                 }
            }
/* 
 * temporal fix for pdbx_refine_ls_restr_ncs and pdbx_struct_ncs_dom_lim
 */
            if (exchange_status == EXCHANGE_TO_INTERNAL &&
               (!strcasecmp(tableList[it], "pdbx_refine_ls_restr_ncs") ||
                !strcasecmp(tableList[it], "pdbx_struct_ncs_dom_lim"))) {
                 if (!strcasecmp(tableList[it], "pdbx_refine_ls_restr_ncs"))
                      tAlias = new ISTable("ccp4_refine_ls_restr_ncs");
                 else tAlias = new ISTable("ccp4_struct_ncs_dom_lim");
                 for (ic = 0; ic < nCols; ic++) {
                      tAlias->AddColumn((*colNames)[ic].Text());
                 }
                 nRows = t->GetNumRows();
                 for (i = 0; i < nRows; i++) {
                      tAlias->AddRow();
                      for (ic = 0; ic < nCols; ic++) {
                           t->GetCell(cs, ic, i);
                           tAlias->UpdateCell(cs, ic, i);
                      }
                 }
                 if (!strcasecmp(tableList[it], "pdbx_refine_ls_restr_ncs"))
                      fobjOut->WriteTable(tAlias, StructureId, "ccp4_refine_ls_restr_ncs");
                 else fobjOut->WriteTable(tAlias, StructureId, "ccp4_struct_ncs_dom_lim"); 

            }
 
            if (tableList[it]) delete [] tableList[it];
            if (colNames) delete (colNames);
       }
       if (tableList) delete [] tableList;
       delete CifObj;
       if (remove_pdb_id_field) {
            t = fobjOut->GetTablePtr(StructureId, "struct_biol_gen");
            index = t->GetColumnIndex("ndb_new_pdb_asym_id");
            if (index >= 0) {
                 t->DeleteColumn(index);
                 t->RemoveColumn(index);
            }
       }

       if (exchange_status == EXCHANGE_TO_INTERNAL)
            cif_dictionary_patch(StructureId, fobjOut);

       return fobjOut;
}

void cif_dictionary_patch(const char *StructureId, CifFileObj *CifObj)
{
       int i, j, k, index, index1, nRows;
       char buffer[10];
       CifString cs;
       ISTable *t = NULL;
       ReVarCifArray<CifString> colData;

       t = CifObj->GetTablePtr(StructureId, "pdbx_SG_project");
       if (t) {
            index = t->GetColumnIndex("entry_id");
            if (index >= 0) t->RemoveColumn(index);
            index = t->GetColumnIndex("id");
            if (index < 0) {
                 t->AddColumn("id");
                 index = t->GetColumnIndex("id");
                 nRows = t->GetNumRows();
                 colData.Clear();
                 for (i = 0; i < nRows; i++) {
                      sprintf(buffer, "%d", i + 1);
                      cs.Copy(buffer);
                      colData.Add(cs);
                 }
                 t->FillColumn(colData, index);
                 CifObj->WriteTable(t, StructureId, "pdbx_SG_project");
            }
       }
 
       for (i = 0; i < NUM_PATCH; i++) {
            t = CifObj->GetTablePtr(StructureId, _category_patch[i].category);
            if (!t) continue;

            nRows = t->GetNumRows();
            for (j = 0; j < _category_patch[i].num_pair; j++) {
                 index = t->GetColumnIndex(_category_patch[i].item_pair[j].old_item);
                 index1= t->GetColumnIndex(_category_patch[i].item_pair[j].new_item);
                 if (index >= 0 && index1 < 0) {
                      t->RenameColumn(_category_patch[i].item_pair[j].old_item,
                                      _category_patch[i].item_pair[j].new_item);
                 } else if (index >= 0 && index1 >= 0) {
                      for (k = 0; k < nRows; k++) {
                           t->GetCell(cs, index1, k);
                           if (cs != "." && cs != "?" && cs != "") break;
                      }
                      if (k == nRows) {
                           t->DeleteColumn(index1);
                           t->RemoveColumn(index1);
                           t->RenameColumn(_category_patch[i].item_pair[j].old_item,
                                           _category_patch[i].item_pair[j].new_item);
                      } 
                 }
            }
            CifObj->WriteTable(t, StructureId, _category_patch[i].category);
       }
}

static int read_binary_dictionary()
{
       unsigned int i;
       int iCol2, iCol3, iCol4, iCol5, iCol6;
       ReVarCifArray<CifString> *col0 = NULL, *col1 = NULL, *col2 = NULL;
       char binaryfile[300], *BlockName = NULL;
       struct stat stat1;

       sprintf(binaryfile, "%s/data/binary/%s", root_dir, "mmcif_pdbx.odb");
       if (!stat(binaryfile, &stat1)) {
            dict = new DICFileObj(binaryfile, READ_MODE);
            BlockName = dict->GetBlockName(0);
            aliases = dict->GetTablePtr(BlockName, "item_aliases");
            items   = dict->GetTablePtr(BlockName, "item");
            if (BlockName) delete [] BlockName;

            iCol1 = items->GetColumnIndex("name");
            col0 = items->GetColumn(iCol1);
            for (i = 0; i < col0->Length(); i++) {
              (*col0)[i].ToLower();
            }
            items->AddColumn("namelc");
            iCol2 = items->GetColumnIndex("namelc");
            items->FillColumn(*col0, iCol2);

            iCol3 = aliases->GetColumnIndex("name");
            col1 = aliases->GetColumn(iCol3);
            for (i = 0; i < col1->Length(); i++) {
              (*col1)[i].ToLower();
            }
            aliases->AddColumn("namelc");
            iCol4 = aliases->GetColumnIndex("namelc");
            aliases->FillColumn(*col1, iCol4);

            iCol5 = aliases->GetColumnIndex("alias_name");
            col2 = aliases->GetColumn(iCol5);
            for (i = 0; i < col2->Length(); i++) {
              (*col2)[i].ToLower();
            }
            aliases->AddColumn("alias_namelc");
            iCol6 = aliases->GetColumnIndex("alias_namelc");
            aliases->FillColumn(*col2, iCol6);
            if (col0) delete col0;
            if (col1) delete col1;
            if (col2) delete col2;
            return 1;
       } else return 0;
}

static int GetCategoryNameFromItemName(char categoryName[MAX_CIF_NAME_LENGTH],
                                       const char *itemName)
/* ----------------------------------------------------------------------
   Purpose: GetCategoryNameFromItemName(categoryName, itemName)
  
            Get the category part of an item name (ie. _<category>.<keyword>)
            Return 1 for success or 0 otherwise.

 * ---------------------------------------------------------------------- */
{
  unsigned int i,j;
  strcpy(categoryName,"");
  if (!itemName) return(0);

  for (i=1, j=0; i < strlen(itemName); i++) {
    if (itemName[i] != '.') {
      categoryName[j] = itemName[i];
      j++;
    } else {
      break;
    }
  }
  categoryName[j]='\0';
  return(1);
}

static int GetKeywordFromItemName(char keyword[MAX_CIF_NAME_LENGTH],
                                  const char *itemName)
/* ----------------------------------------------------------------------
   Purpose: GetKeywordFromItemName(keyword, itemName)

            Get the keyword part of an item name (ie. _<category>.<keyword>)
            Return 1 for success or 0 otherwise.

 * ---------------------------------------------------------------------- */
{
  int i,j,k,ilen;

  strcpy(keyword,"");
  if (!itemName ||   (ilen=strlen(itemName)) < 3 ) return(0);

  for (k=1; k < ilen; k++) {
    if (itemName[k] == '.') break;
  }
  if ((k==1) || (k == ilen)) return(0);

  j=0;
  for (i=k+1; i < ilen; i++) {
    keyword[j] = itemName[i];
    j++;
  }
  keyword[j]='\0';
  return(1);

}

