#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "ndb_misclib.h"
#include "erf_util.h"
#include "util.h"

#define NUM_STAND        34
#define NUM_CODE         28
#define NUM_SPACE_GROUP  95

typedef struct {
        char a1let;
        const char *a3let;
} CODES;

typedef struct {
        char chnid;
        char seq[3000];
} SEQRES;

typedef struct {
        char fake_name[15];
        char real_name[15];
} SPACE_GROUP_NAME;

static char standard_code[NUM_STAND][4] = {
          "A",   "C",   "G",   "I",   "T",   "U",   "Ade", "Cyt", "Gua", "Ino",
          "Thy", "Ura", "Uri", "Ala", "Arg", "Asn", "Asp", "Cys", "Gln", "Glu",
          "Gly", "His", "Ile", "Leu", "Lys", "Met", "Phe", "Pro", "Ser", "Thr",
          "Trp", "Tyr", "Val", "Hyp"
};

static CODES codes[NUM_CODE] = {
       { 'A',  "Ade" },
       { 'C',  "Cyt" },
       { 'G',  "Gua" },
       { 'I',  "Ino" },
       { 'T',  "Thy" },
       { 'U',  "Ura" },
       { 'U',  "Uri" },
       { 'A',  "Ala" },
       { 'R',  "Arg" },
       { 'N',  "Asn" },
       { 'D',  "Asp" },
       { 'C',  "Cys" },
       { 'Q',  "Gln" },
       { 'E',  "Glu" },
       { 'G',  "Gly" },
       { 'H',  "His" },
       { 'I',  "Ile" },
       { 'L',  "Leu" },
       { 'K',  "Lys" },
       { 'M',  "Met" },
       { 'F',  "Phe" },
       { 'P',  "Pro" },
       { 'S',  "Ser" },
       { 'T',  "Thr" },
       { 'W',  "Trp" },
       { 'Y',  "Tyr" },
       { 'V',  "Val" },
       { 'O',  "Hyp" }
};

static SPACE_GROUP_NAME space_group_name[NUM_SPACE_GROUP] = {
       { "A1",         "A 1" },
       { "A121",       "A 1 2 1" },
       { "A2",         "A 2" },
       { "B112",       "B 1 1 2" },
       { "B2",         "B 2" },
       { "B2212",      "B 2 21 2" },
       { "C121",       "C 1 2 1" },
       { "C1211",      "C 1 21 1" },
       { "C2",         "C 2" },
       { "C2(A112)",   "C 2(A 112)" },
       { "C21",        "C 21" },
       { "C222",       "C 2 2 2" },
       { "C2221",      "C 2 2 21" },
       { "C4212",      "C 4 21 2" },
       { "F222",       "F 2 2 2" },
       { "F23",        "F 2 3" },
       { "F4132",      "F 41 3 2" },
       { "F422",       "F 4 2 2" },
       { "F432",       "F 4 3 2" },
       { "H3",         "H 3" },
       { "H32",        "H 3 2" },
       { "I121",       "I 1 2 1" },
       { "I1211",      "I 1 21 1" },
       { "I2",         "I 2" },
       { "I21",        "I 21" },
       { "I212121",    "I 21 21 21" },
       { "I213",       "I 21 3" },
       { "I222",       "I 2 2 2" },
       { "I23",        "I 2 3" },
       { "I4",         "I 4" },
       { "I41",        "I 41" },
       { "I4122",      "I 41 2 2" },
       { "I4132",      "I 41 3 2" },
       { "I422",       "I 4 2 2" },
       { "I432",       "I 4 3 2" },
       { "P-1",        "P 1-" },
       { "P1",         "P 1" },
       { "P1-",        "P 1-" },
       { "P112",       "P 1 1 2" },
       { "P1121",      "P 1 1 21" },
       { "P1211",      "P 1 21 1" },
       { "P121",       "P 1 2 1" },
       { "P2",         "P 2" },
       { "P21",        "P 21" },
       { "P21(C)",     "P 21(C)" },
       { "P21212",     "P 21 21 2" },
       { "P212121",    "P 21 21 21" },
       { "P21212A",    "P 21 21 2 A" },
       { "P21221",     "P 21 2 21" },
       { "P213",       "P 21 3" },
       { "P22121",     "P 2 21 21" },
       { "P222",       "P 2 2 2" },
       { "P2221",      "P 2 2 21" },
       { "P23",        "P 2 3" },
       { "P3",         "P 3" },
       { "P31",        "P 31" },
       { "P3112",      "P 31 1 2" },
       { "P312",       "P 3 1 2" },
       { "P3121",      "P 31 2 1" },
       { "P32",        "P 32" },
       { "P321",       "P 3 2 1" },
       { "P3212",      "P 32 1 2" },
       { "P3221",      "P 32 2 1" },
       { "P4",         "P 4" },
       { "P41",        "P 41" },
       { "P41212",     "P 41 21 2" },
       { "P4122",      "P 41 2 2" },
       { "P4132",      "P 41 3 2" },
       { "P42",        "P 42" },
       { "P4212",      "P 4 21 2" },
       { "P422",       "P 4 2 2" },
       { "P42212",     "P 42 21 2" },
       { "P4222",      "P 42 2 2" },
       { "P4232",      "P 42 3 2" },
       { "P43",        "P 43" },
       { "P432",       "P 4 3 2" },
       { "P43212",     "P 43 21 2" },
       { "P4322",      "P 43 2 2" },
       { "P4332",      "P 43 3 2" },
       { "P6",         "P 6" },
       { "P61",        "P 61" },
       { "P6122",      "P 61 2 2" },
       { "P62",        "P 62" },
       { "P622",       "P 6 2 2" },
       { "P6222",      "P 62 2 2" },
       { "P63",        "P 63" },
       { "P6322",      "P 63 2 2" },
       { "P64",        "P 64" },
       { "P6422",      "P 64 2 2" },
       { "P65",        "P 65" },
       { "P6522",      "P 65 2 2" },
       { "R3",         "R 3" },
       { "R3(RHOM)",   "R 3(RHOM)" },
       { "R32",        "R 3 2" },
       { "R32(RHOM)",  "R 3 2(RHOM)" }
};

static char defalut_dir[] = "/ndbapps";

static SEQRES seqres[50];

static char one_letter_code(char *res);
static int is_standard(const char *res);
static int check_space_group(char *old_name, char *new_name);
#ifdef SUN_OS
#ifdef __cplusplus
extern "C" {
#endif
#endif
static int real_name_compare(const void *group1, const void *group2);
static int fake_name_compare(const void *group1, const void *group2);
#ifdef SUN_OS
#ifdef __cplusplus
}
#endif
#endif

static int setErrState(int iErrCount);

#define ERR_FILE_NAME "filter.err"
#define OUT_FILE_NAME "filter.out"

int setErrState(int iErrCount) {
  if (!iErrCount) 
    remove(ERR_FILE_NAME);
  else
    remove(OUT_FILE_NAME);
  return(iErrCount);
}

int main(int argc, char *argv[])
{
       char filename[200], line[200], res[5];
       int nErr=0;
       filename[0] = '\0';
       int is_nmr = 0, i, j, len;

       root_dir = getenv("RCSBROOT");
       if (root_dir == NULL) root_dir = defalut_dir;

       for (i = 1; i < argc; i++) {
            if (strcmp(argv[i], "-i") == 0) {
                 i++;
                 strcpy(filename,argv[i]);
            } else if (strcmp(argv[i], "-nmr") == 0) {
                 is_nmr = 1;
            }
       }

       FILE *outfile = fopen(OUT_FILE_NAME, "w");
       FILE *errfile = fopen(ERR_FILE_NAME, "w");

       if (filename[0] == '\0') {
	    nErr++;
            fprintf(errfile, "<li>ERROR: No input PDB file\n");
            fclose (errfile);
            fclose (outfile);
	    setErrState(nErr);
            return (1);
       }

       int level = NDB_MSG_INFO;
       ndb_set_debug_level(level);
       ndb_open_log("prefilter", "prefilter.err");
       if (!ndb_filterlib_init(root_dir)) {
            printf("Initial filter libaray failed\n");
            return -1;
       }

       FILE *infile = fopen(filename, "r");
       if (infile == NULL) {
	    nErr++;
            fprintf(errfile, "<li>ERROR: Can not open file: %s\n", filename);
            fclose (errfile);
            fclose (outfile);
	    setErrState(nErr);
            return (1);
       }

       int exist_ter = 0;
       int exist_chnid = 1;
       int misplace = 0;
       int no_chnid = 0;
       int nAtoms = 0;
       int exist_standard = 0;
       int exist_cryst1 = 0;

       char seq_chainid[50];
       seq_chainid[0] = '\0';

       char prev_chain = 0;
       int num_end = 0;

       while (!feof(infile)) {
            fgets(line, 199, infile);
            if (feof(infile)) break;
            len = strlen(line);
            for (i = len - 1; i >= 0; i--)
                 if (line[i] == 10 || line[i] == 13) line[i] = 0;

            if (!strncasecmp(line, "TER", 3)) {
                 exist_ter = 1;
                 exist_standard = 0;
                 if (no_chnid) exist_chnid = 0;
            }

            if (!strncasecmp(line, "END", 3) &&
                 strncmp(line, "ENDMDL", 6)) {
                 num_end++;
            }

            if (!strncasecmp(line, "MODEL", 5)) {
                  len = strlen(line); 
                  for (i = len - 1; i >= 0; i--) {
                       if (line[i] == ' ') line[i] = '\0';
                       if (isdigit(line[i])) break;
                  }
                  len = strlen(line);
                  if (i <= 0 || len < 6) {
                       fprintf(errfile, "<li>ERROR: The format of MODEL record ");
                       fprintf(errfile, "is not in compliance with PDB V2 format.\n");
                       fprintf(errfile, "The MODEL number must be between column 11 and 14.\n");
                       nErr++;
                  }
                  for (i = 5; i < len; i++) {
                       if (isdigit(line[i]) && (i < 9 || i > 13)) {
                            fprintf(errfile, "<li>ERROR: The format of MODEL record ");
                            fprintf(errfile, "is not in compliance with PDB V2 format.\n");
                            fprintf(errfile, "The MODEL number must be between column 11 and 14.\n");
                            nErr++;
                            break;
                       }
                  }
            }

            if (!strncmp(line, "SEQRES", 6) && line[11] != ' ') {
                 len = strlen(seq_chainid);
                 for (i = 0; i < len; i++)
                      if (seq_chainid[i] == line[11]) break;
                 if (i == len) {
                      seq_chainid[i] = line[11];
                      seq_chainid[i+1] = '\0';
                 }
            }

            if (!strncmp(line, "CRYST1", 6) && !is_nmr) {
                 exist_cryst1 = 1;
                 char a[10], b[10], c[10], alpha[10], beta[10], gamma[10], name[12], real_name[12];
                 for (i = 0; i < 9; i++) {
                      a[i] = line[6+i];
                      b[i] = line[15+i];
                      c[i] = line[24+i];
                 }
                 a[i] = b[i] = c[i] = '\0';
                 ndb_clean_string(a);
                 ndb_clean_string(b);
                 ndb_clean_string(c);

                 for (i = 0; i < 7; i++) {
                      alpha[i] = line[33+i];
                      beta[i]  = line[40+i];
                      gamma[i] = line[47+i];
                 }
                 alpha[i] = beta[i] = gamma[i] = '\0';
                 ndb_clean_string(alpha);
                 ndb_clean_string(beta);
                 ndb_clean_string(gamma);

                 for (i = 0; i < 11; i++) name[i] = line[55+i];
                 name[i] = '\0';
                 ndb_clean_string(name);
                 fprintf(outfile, "Cell Unit Parameters: a=%s b=%s c=%s alpha=%s beta=%s gamma=%s\n",
                          a, b, c, alpha, beta, gamma);
                 int exist = check_space_group(name, real_name);
                 fprintf(outfile, "Space Group: %s\n", real_name);
                 if (!exist) fprintf(outfile,
                      "WARNING: Space Group \"%s\" did not exist in the dictionary\n", name);
/*
                 if (!strcmp(name, "C 1 2 1"))       strcpy(name, "C 2");
                 else if (!strcmp(name, "R 3 2"))    strcpy(name, "R 3 2(RHOM)");
                 else if (!strcmp(name, "H 3 2"))    strcpy(name, "R 3 2");
                 else if (!strcmp(name, "H 3"))      strcpy(name, "R 3");
                 else if (!strcmp(name, "P -1"))     strcpy(name, "P 1-");
                 else if (!strcmp(name, "P 1 1 21")) strcpy(name, "P 21(C)");
                 else if (!strcmp(name, "P 1 21 1")) strcpy(name, "P 21");
                 else if (!strcmp(name, "P 1 2 1"))  strcpy(name, "P 2");
                 else if (!strcmp(name, "B 2"))      strcpy(name, "B 1 1 2");

                 SPACE_GROUP spacegroup;
                 spacegroup.SpaceData = NULL;
                 if (get_spacegroup_info(name, &spacegroup, NULL)) {
                      free(spacegroup.SpaceData);
                 } else {
                      fprintf(outfile, "WARNING: Space Group \"%s\" did not exist in the dictionary\n",
                               name);
                 }
*/
                 fprintf(outfile, "\n");
            }

            if (strncmp(line, "ATOM", 4) &&
                strncmp(line, "HETATM", 6)) continue;
	    nAtoms++;

            if (line[11] != ' ' || line[20] != ' ' ||
                line[27] != ' ' || line[28] != ' ' ||
                line[29] != ' ' || isdigit(line[66]) /* ||
                line[65] == ' ' */ ) {
                 misplace = 1;
                 nErr++;
                 fprintf(errfile, "<li>ERROR: The format of coordinate section ");
                 fprintf(errfile, "is not in compliance with PDB V2 format\n");
                 if (line[11] != ' ') 
                      fprintf(errfile, "12th column should be space instead of %c\n",
                                line[11]);
                 else if (line[20] != ' ')
                      fprintf(errfile, "21th column should be space instead of %c\n",
                                line[20]);
                 else if (line[27] != ' ')
                      fprintf(errfile, "28th column should be space instead of %c\n",
                                line[27]);
                 else if (line[28] != ' ')
                      fprintf(errfile, "29th column should be space instead of %c\n",
                                line[28]);
                 else if (line[29] != ' ')
                      fprintf(errfile, "30th column should be space instead of %c\n",
                                line[29]);
                 else if (line[66] != ' ')
                      fprintf(errfile, "67th column should be space instead of %c\n",
                                line[66]);
                 fprintf(errfile, "%s\n", line);
            }

	    if (!strncmp(line, "ATOM", 4)) {
                 if (line[21] == ' ') no_chnid = 1;
                 else {
                      if (line[21] != prev_chain) {
                           len = strlen(seq_chainid);
                           for (i = 0; i < len; i++) {
                                if (seq_chainid[i] == line[21]) break;
                           }
                           if (i == len && len) {
                                nErr++;
                                fprintf(errfile, "<li>ERROR: Chain ID '%c' has no corresponding SEQRES records: ", line[21]);
                                for (j = 0; j < len; j++) {
                                     fprintf(errfile, "'%c' ", seq_chainid[j]);
                                }
                                fprintf(errfile, "\n");
                           }
                      }
                      prev_chain = line[21];
                 }
                 if (!exist_standard) {
                      for (i = 0; i < 3; i++) res[i] = line[17 + i];
                      res[i] = '\0';
                      ndb_clean_string(res);
                      exist_standard = is_standard(res);
                      if (exist_standard && line[21] == ' ')
                           exist_chnid = 0; 
                 }
            }
/*
            if (!exist_standard) {
                 for (i = 0; i < 3; i++) res[i] = line[17 + i];
                 res[i] = '\0';
                 ndb_clean_string(res);
                 exist_standard = is_standard(res);
                 if (exist_standard && line[21] == ' ')
                      exist_chnid = 0; 
            }
*/
       }

       if (num_end < 1) {
            nErr++;
            fprintf(errfile, "<li>ERROR: file has no 'END' tokens.\n");
       }

       if (num_end > 1) {
            nErr++;
            fprintf(errfile, "<li>ERROR: file has %d 'END' tokens.\n", num_end);
       }

       if (!exist_chnid && seq_chainid[0] != '\0') {
            nErr++;
            fprintf(errfile, "<li>ERROR: Chain ID ' ' has no corresponding SEQRES records: ");
            len = strlen(seq_chainid);
            for (j = 0; j < len; j++) {
                 fprintf(errfile, "'%c' ", seq_chainid[j]);
            }
            fprintf(errfile, "\n");
       }

       if (!exist_cryst1 && !is_nmr) {
            fprintf(outfile, "WARNING: No Cell Unit Parameters and Space Group Information\n\n");
       }

       if (!exist_ter || !exist_chnid || exist_standard) {
            if (!exist_chnid && nAtoms) {
 	         // nErr++;
                 fprintf(errfile, "<li>WARNING: Polymer atom records missing chain ID field\n\n");
                 fprintf(outfile, "WARNING: Polymer atom records missing chain ID field. For structures\n");
                 fprintf(outfile, "with one chain, validation results may be correct. However, structures\n");
                 fprintf(outfile, "with more than one chain and no chain IDs may produce incorrect\n");
                 fprintf(outfile, "validation results.\n");
	    }
            if ((!exist_ter || exist_standard) && nAtoms) {
	         // nErr++;
                 fprintf(errfile, "<li>WARNING: No \"TER\" records after each polymer\n\n");
                 fprintf(outfile, "WARNING: No \"TER\" records after each polymer\n\n");
	    }
       }

       if (num_end < 1 || num_end > 1 || misplace || !nAtoms) {
            if (misplace) {
	         nErr++;
                 fprintf(errfile, "<li>ERROR: The format of coordinate section ");
                 fprintf(errfile, "is not in compliance with PDB V2 format\n\n");
            }
	    if (!nAtoms) {
	      nErr++;
	      fprintf(errfile, "<li>ERROR: No ATOM/HETATM records found.\n\n");
	    }

            fclose (errfile);
            fclose (outfile);
            fclose (infile);
	    setErrState(nErr);
            return (1);
       }

       rewind (infile);

       for (i = 0; i < 50; i++) seqres[i].seq[0] = '\0';

       char prevRes[5], prevNum[10], resNum[10], prevline[200];
       strcpy(prevRes, "XXX");
       strcpy(prevNum, "XXX");
       int num = 0;
       while (!feof(infile)) {
            fgets(line, 199, infile);
            if (feof(infile)) break;
            if (!strncasecmp(line, "TER", 3)) {
                 num++;
                 strcpy(prevRes, "XXX");
                 strcpy(prevNum, "XXX");
            }

            if (!strncmp(line, "ENDMDL", 6)) break;

            if (strncmp(line, "ATOM", 4) &&
                strncmp(line, "HETATM", 6)) continue;

            line[strlen(line) - 1] = '\0';
            for (i = strlen(line) - 1; i >= 0; i--) {
                 if (line[i] == ' ' || line[i] == '\t' ||
                     line[i] == 13  || line[i] == '\n')
                      line[i] = '\0';
                 else break;
            }

            for (i = 0; i < 3; i++) res[i] = line[17 + i];
            res[i] = '\0';
            ndb_clean_string(res);
            for (i = 0; i < 6; i++) resNum[i] = line[22 + i];
            resNum[i] = '\0';
            ndb_clean_string(resNum);

            if (strcmp(res, "W") && strcmp(res, "H2O") &&
                strcmp(res, "WAT") && strncmp(res, "TIP", 3) &&
                strcmp(res, "WTR") && strcmp(res, "OH2") &&
                strcmp(res, "HOH")) {
                 if (strcmp(res, "CYT") && strcmp(res, "GUA") &&
                     strcmp(res, "THY") && strcmp(res, "ADE") &&
                     strcmp(res, "INO") && strcmp(res, "URA") &&
                     strcmp(res, "URI")) {
                      ConnectFormat *drug = find_drug(res);
                      if (drug == NULL) {
                           fprintf(errfile,
                            "<li>WARNING: Unrecognized residue name %s in record:\n\n", res);
                           fprintf(errfile, "<br>%s\n\n", line);
                           fprintf(outfile,
                            "WARNING: Unrecognized residue name %s in record:\n\n", res);
                           fprintf(outfile, "%s\n\n", line);
                      }
                 }

                 len = strlen(seqres[num].seq);
/*
                 if (len && seqres[num].chnid != line[21]) {
	              nErr++;
                      fprintf(errfile, "<li>ERROR: There are different chain IDs for one polymer\n\n");
                      fprintf(errfile, "<br>%s\n", prevline);
                      fprintf(errfile, "<br>%s\n", line);
                 }
*/
                 if ((strcmp(resNum, prevNum) || strcmp(res, prevRes)) &&
                      !strncmp(line, "ATOM", 4) /* This will delete non-standard residue */ ) {
                      seqres[num].chnid = line[21];
                      sprintf(&(seqres[num].seq[len]), "%c", one_letter_code(res));
                 }
            }
            strcpy(prevRes, res);
            strcpy(prevNum, resNum);
            strcpy(prevline, line);
       }

       fprintf(outfile, "One letter code sequence extracted from coordinate section:\n\n");
       for (i = 0; i < num; i++) {
            fprintf(outfile,
        "                 10        20        30        40        50        60\n");
            int length = 0;
            fprintf(outfile, "CHAIN %c: ", seqres[i].chnid);
            len = strlen(seqres[i].seq);
            for (j = 0; j < len; j++) {
                 fprintf(outfile, "%c", seqres[i].seq[j]);
                 length++;
                 if (length == 60) {
                      fprintf(outfile, "\n         ");
                      length = 0;
                 }
            }
            fprintf(outfile, "\n");
            fprintf(outfile, "\nTotal %d residues.\n\n", (int) strlen(seqres[i].seq));
       }

       fclose (errfile);
       fclose (outfile);
       fclose (infile);
       setErrState(nErr);
       ndb_filterlib_close();
       ndb_close_log();
       return (0);
}

static char one_letter_code(char *res)
{
       if (strlen(res) == 1) return (res[0]);

       for (int i = 0; i < NUM_CODE; i++) {
            if (!strncasecmp(codes[i].a3let, res, 3))
                 return codes[i].a1let;
       }
       return ('?');
}

static int is_standard(const char *res)
{
       char Name[10];
       if (res[0] == '+')
            strcpy(Name, &res[1]);
       else strcpy(Name, res);

       for (int i = 0; i < NUM_STAND; i++) {
            if (!strcasecmp(Name, standard_code[i])) return 1;
       }
       return 0;
}

static int check_space_group(char *old_name, char *new_name)
{
       strcpy(new_name, old_name);
       qsort((void *) space_group_name, NUM_SPACE_GROUP,
             sizeof(SPACE_GROUP_NAME), real_name_compare);
       SPACE_GROUP_NAME tmp;
       strcpy(tmp.real_name, old_name);
       SPACE_GROUP_NAME *exist = (SPACE_GROUP_NAME *) bsearch(&tmp,
            space_group_name, NUM_SPACE_GROUP, sizeof(SPACE_GROUP_NAME),
            real_name_compare);
       if (exist) return 1;
       strcpy(tmp.fake_name, old_name);
       int stringlen = strlen(tmp.fake_name);
       for (int i = 0; i < stringlen; i++) {
            if (tmp.fake_name[i] == ' ') {
                 strcpy(&(tmp.fake_name[i]), &(tmp.fake_name[i+1]));
                 stringlen--; i--;
            }
       }
       qsort((void *) space_group_name, NUM_SPACE_GROUP,
             sizeof(SPACE_GROUP_NAME), fake_name_compare);
       exist = (SPACE_GROUP_NAME *) bsearch(&tmp, space_group_name,
           NUM_SPACE_GROUP, sizeof(SPACE_GROUP_NAME), fake_name_compare);
       if (exist) {
            strcpy(new_name, exist->real_name);
            return 1;
       }
       return 0;
}

static int real_name_compare(const void *group1, const void *group2)
{
       return (strcmp(((SPACE_GROUP_NAME *) group1)->real_name,
                      ((SPACE_GROUP_NAME *) group2)->real_name));
}

static int fake_name_compare(const void *group1, const void *group2)
{
       return (strcmp(((SPACE_GROUP_NAME *) group1)->fake_name,
                      ((SPACE_GROUP_NAME *) group2)->fake_name));
}
