/*
FILE:     promotif.C
*/
/*
VERSION:  1.700
*/
/*
DATE:     7/13/2005
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2005 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <math.h>

#include "_maxit.h"
#include "erf_util.h"
#include "misc_util.h"

#define AA_BACKBONE  3
#define NITROGEN     0
#define CARBON       1
#define OXYGEN       2
#define HYDROGEN     3

static const char *aa_atoms[AA_BACKBONE] = {"N", "C", "O"};
static const char stchap[27] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
static const char stchpa[27] = "abcdefghijklmnopqrstuvwxyz";

static void engset(const double energy, _NRESIDUE **residues, const int donres,
                   const int accres);
static int xeqy(const double x, const double y);
static int engpl(const double energy, double *honde);
static void rejbnd(const int rejptr, _NRESIDUE *residue, const int start);
static void stbrg(int *strcd, int *brdg, const int stsht, const int finsht, _NRESIDUE **allresidues);
static void whstrd(int *strcd, const int stsht, const int finsht, int &stres, int &stpl,
              const int lststd, int &bstval);
static int strdpl(const int stpl, const int finpl, const int thcode, int *strand);
static void fdshus(const int stpt, _NRESIDUE **residues, int *shtcod, int &stsht,
               int &finsht, const int seqlen);
static void setsht(const int refpl, const int shtsum, int *shtcod, _NRESIDUE **residues,
               const int seqlen);
static int fdshps(const int stpt, const int shtsum, int *shtcod, _NRESIDUE **residues,
               const int seqlen);
static void hlxset(const int seqlen, _NRESIDUE **residues, const int hlxpos,
               const int hlxsz, const char hlxch, const char altch);
static void hlxres(const int seqlen, _NRESIDUE **residues, const char hlxch,
               const char altch, const char turnch, const char alturn, const int hlxsz);
static void get_helix_cards(_CHAIN *chain, _RESIDUE *strR, _RESIDUE *endR,
                   const char type, const int count);

void _MAXIT::get_create_secondary_structures_using_promotif(const char *support_file)
{
       int found, RMKSKP_TOKEN, need_helix = 1, need_sheet = 1, need_ssbond = 1;
       int i, j, k, l, m, n_allresidues = 0, alloc_allresidues = 0;
       double maxeng = -0.5, englim = -9.9, ksq1 = 0.42, ksq2 = 0.2;
       double ksf = 332, maxcad = 8.0, cadist, don, doh, dch, dcn, energy;
       _ATOM *atom[AA_BACKBONE];
       _RESIDUE *residue = NULL;
       _CHAIN *chain = NULL;
       _NRESIDUE **allresidues = NULL, **tmp = NULL;
       CifString string;
       COORD coord, coord1;
       ISTable *Table = NULL;

       if (_num_mol == 0) return;
       if (_exist_helix && _exist_sheet && _exist_ssbond) return;

       RMKSKP_TOKEN = find_card_id("RMKSKP");
       if (_CardInfo[RMKSKP_TOKEN].Head()) {
            if (strstr(_CardInfo[RMKSKP_TOKEN].Head()->Field(1), "HELIX") ||
                strstr(_CardInfo[RMKSKP_TOKEN].Head()->Field(1), "helix"))
                 need_helix = 0;
            if (strstr(_CardInfo[RMKSKP_TOKEN].Head()->Field(1), "SHEET") ||
                strstr(_CardInfo[RMKSKP_TOKEN].Head()->Field(1), "sheet"))
                 need_sheet = 0;
            if (strstr(_CardInfo[RMKSKP_TOKEN].Head()->Field(1), "SSBOND") ||
                strstr(_CardInfo[RMKSKP_TOKEN].Head()->Field(1), "ssbond"))
                 need_ssbond = 0;
       }
       if (_CifObj) {
            Table = _CifObj->GetTablePtr(_StructureId, "ndb_database_status");
            if (!is_empty_table(Table)) {
                 i = Table->GetColumnIndex("skip_PDB_REMARK");
                 if (i >= 0) {
                      Table->GetCell(string, i, 0);
                      if (strstr(string.Text(), "HELIX") ||
                          strstr(string.Text(), "helix"))
                           need_helix = 0;
                      if (strstr(string.Text(), "SHEET") ||
                          strstr(string.Text(), "sheet"))
                           need_sheet = 0;
                      if (strstr(string.Text(), "SSBOND") ||
                          strstr(string.Text(), "ssbond"))
                           need_ssbond = 0;
                 }
            }
       }
       if (!need_helix && !need_sheet && !need_ssbond) return;

       chain = _molecules[0]->GetFirstChain();
       while (chain) {
            if (chain->chain_type() != ATOMP_TOKEN) {
                 chain = _molecules[0]->GetNextChain();
                 continue;
            }
            alloc_allresidues = n_allresidues + chain->TotalResidueNumbers();
            tmp = allresidues;
            allresidues = new _NRESIDUE*[alloc_allresidues];
            for (i = 0; i < n_allresidues; i++) {
                 allresidues[i] = tmp[i];
            }
            if (tmp) delete [] tmp;
            for (i = n_allresidues; i < alloc_allresidues; i++)
                 allresidues[i] = NULL;
            if (n_allresidues) {
                 allresidues[n_allresidues-1]->chnbreak = 1;
            }
            residue = chain->GetFirstResidue();
            while (residue) {
                 found = 0;
                 for (i = 0; i < AA_BACKBONE; i++) {
                      atom[i] = residue->find_atom(aa_atoms[i]);
                      if (atom[i]) found = 1;
                 }
                 if (found) {
                      allresidues[n_allresidues] = new _NRESIDUE;
                      allresidues[n_allresidues]->res = residue;
                      allresidues[n_allresidues]->chn = chain;
                      allresidues[n_allresidues]->chnbreak = 0;
                      for (i = 0; i < 4; i++)
                           allresidues[n_allresidues]->exist[i] = 0;
                      for (i = 0; i < AA_BACKBONE; i++) {
                           if (atom[i]) {
                                allresidues[n_allresidues]->exist[i] = 1;
                                allresidues[n_allresidues]->coord[i] = atom[i]->orig();
                           }
                      }
                      if (n_allresidues && !allresidues[n_allresidues-1]->chnbreak) {
                           allresidues[n_allresidues-1]->chnbreak =
                             !is_connect("C", allresidues[n_allresidues-1]->res,
                                         "N", allresidues[n_allresidues]->res);
                           if (!allresidues[n_allresidues-1]->chnbreak &&
                               allresidues[n_allresidues-1]->exist[CARBON] &&
                               allresidues[n_allresidues-1]->exist[OXYGEN] &&
                               allresidues[n_allresidues]->exist[NITROGEN]) {
                                vector_difference(coord,
                                    allresidues[n_allresidues-1]->coord[CARBON],
                                    allresidues[n_allresidues-1]->coord[OXYGEN]);
                                vector_normalize(coord);
                                vector_sum(allresidues[n_allresidues]->coord[HYDROGEN],
                                      allresidues[n_allresidues]->coord[NITROGEN], coord);
                                allresidues[n_allresidues]->exist[HYDROGEN] = 1;
                           }
                      }
                      n_allresidues++;
                 }
                 residue = chain->GetNextResidue();
            }
            chain = _molecules[0]->GetNextChain();
       }

       if (!n_allresidues) {
            if (allresidues) delete [] allresidues;
            return;
       }

       for (i = 0; i < n_allresidues; i++) {
            for (j = 0; j < 4; j++) {
                 allresidues[i]->energy[j] = 0;
                 allresidues[i]->hbond[j] = 0;
            }
            for (j = 0; j < 11; j++)
                 allresidues[i]->ssbond[j] = ' ';
            allresidues[i]->ssbond[j] = '\0';
       }

       for (i = 0; i < n_allresidues; i++) {
            if (!allresidues[i]->exist[NITROGEN] || !allresidues[i]->exist[HYDROGEN] ||
                !strcmp(allresidues[i]->res->ResName(), "PRO")) continue;
            for (j = 0; j < n_allresidues; j++) {
                 if ((abs(i - j) == 1 && allresidues[i]->chn != allresidues[j]->chn) ||
                     abs(i - j) >= 2) {
                      atom[0] = allresidues[i]->res->find_atom("CA");
                      atom[1] = allresidues[j]->res->find_atom("CA");
                      if (atom[0] && atom[1]) {
                           coord = atom[0]->orig();
                           coord1 = atom[1]->orig();
                           cadist = vector_distance(coord, coord1);
                           if (cadist < maxcad && allresidues[j]->exist[CARBON] &&
                               allresidues[j]->exist[OXYGEN]) {
                                don = vector_distance(allresidues[j]->coord[OXYGEN],
                                                      allresidues[i]->coord[NITROGEN]);
                                doh = vector_distance(allresidues[j]->coord[OXYGEN],
                                                      allresidues[i]->coord[HYDROGEN]);
                                dch = vector_distance(allresidues[j]->coord[CARBON],
                                                      allresidues[i]->coord[HYDROGEN]);
                                dcn = vector_distance(allresidues[j]->coord[CARBON],
                                                      allresidues[i]->coord[NITROGEN]);
                                if (!xeqy(don, 0.0) && !xeqy(doh, 0.0) &&
                                    !xeqy(dch, 0.0) && !xeqy(dcn, 0.0)) {
                                     energy = ksq1 * ksq2 * ksf * (1.0 / don
                                            + 1.0 / dch - 1.0 / doh - 1.0 / dcn);
                                     if (energy < englim) energy = englim;
                                     if (energy < maxeng)
                                          engset(energy, allresidues, i, j);
                                }
                           }
                      }
                 }
            }
       }

       int turnpl[3] = {6, 0, 7};
       int turnsz[3] = {3, 4, 5};
       const char turnch[4] = "345";

       for (i = 0; i < n_allresidues; i++) {
            for (j = 0; j < 2; j++) {
                 for (k = 0; k < 3; k++) {
                      if (allresidues[i]->hbond[j]) {
                           l = allresidues[i]->hbond[j] - 1;
                           if ((l - i) == turnsz[k] &&
                               allresidues[l]->chn == allresidues[i]->chn) {
                                if (allresidues[i]->ssbond[turnpl[k]] == '<')
                                     allresidues[i]->ssbond[turnpl[k]] = 'X';
                                else allresidues[i]->ssbond[turnpl[k]] = '>';
                                allresidues[l]->ssbond[turnpl[k]] = '<';
                                for (l = i + 1; l < i + turnsz[k]; l++) {
                                     if (allresidues[l]->ssbond[turnpl[k]] == ' ')
                                          allresidues[l]->ssbond[turnpl[k]] = turnch[k];
                                }
                                for (l = i; l < i + turnsz[k]; l++)
                                     allresidues[l]->ssbond[8] = 'T';
                           }
                      }
                 }
            }
       }

       int bsrng[3][2] = { { 2, -1 }, { 1, 0 }, { 2, -1 } };
       int bsofst[3][4] = { { -1, 0, 0, 1 }, { 0, 0, 0, 0 }, { -1, -1, -2, 1 } };
       int pbrdg[3] = { 1, -1, -1 };
       int exrule[2][3] = { { -1, 1, -1 }, { 1, 1, -1 } };
       int *brdg = NULL, bndptr, brdgpt, brpl, brpla;

       brdg = new int[n_allresidues * 4];
       for (i = 0; i < n_allresidues * 4; i++) brdg[i] = 0;

       for (i = 0; i < 3; i++) {
            for (j = bsrng[i][0] - 1; j < n_allresidues + bsrng[i][1]; j++) {
                 for (k = 0; k < 2; k++) {
                      bndptr = allresidues[j + bsofst[i][0]]->hbond[k] - 1;
                      if (!i || bndptr > j) {
                           brdgpt = bndptr + bsofst[i][1];
                           if ((bndptr+1) > -bsofst[i][2] && abs(j - brdgpt) > 2) {
                                for (l = 0; l < 2; l++) {
                                     if (allresidues[bndptr + bsofst[i][2]]->hbond[l] ==
                                        (j + bsofst[i][3] + 1)) {
                                          brpl = 1;
                                          if (brdg[4 * j + brpl - 1]) brpl = 2;
                                          brdg[4 * j + brpl - 1] = pbrdg[i] * (brdgpt + 1);
                                          brdg[4 * j + brpl + 1] = exrule[0][i];
                                          brpla = 1;
                                          if (brdg[4 * brdgpt + brpla - 1]) brpla = 2;
                                          brdg[4 * brdgpt + brpla - 1] = pbrdg[i] * (j + 1);
                                          brdg[4 * brdgpt + brpla + 1] = exrule[1][i];
                                     }
                                }
                           }
                      }
                 }
            }
       }

       int bkdszl = 5, blgszs = 2, brptr, brdir, brptpl, newptr, nwptpl, stsht, finsht;
       int nxstrn = 0;
       char wkch;
       int *shtcod = new int[n_allresidues];
       int *strcd = new int[2 * n_allresidues];
       for (i = 0; i < n_allresidues; i++) {
            shtcod[i] = 0;
            strcd[2 * i] = 0;
            strcd[2*i+1] = 0;
            allresidues[i]->bridpt[0] = 0;
            allresidues[i]->bridpt[1] = 0;
       }

       for (i = 0; i < n_allresidues; i++) {
            for (k = 0; k < 2; k++) {
                 if (brdg[i * 4 + k] && abs(brdg[i * 4 + k]) > (i + 1)) {
                      brptr = abs(brdg[i * 4 + k]);
                      brdir = brdg[i * 4 + k] / brptr;
                      brptpl = 1;
                      if (abs(brdg[(brptr - 1) * 4]) == (i + 1)) brptpl = 0;
                      brpla = n_allresidues;
                      if ((i + bkdszl + 1) < brpla) brpla = i + bkdszl + 1;
                      found = 0;
                      for (j = i + 1; j < brpla; j++) {
                           for (l = 0; l < 2; l++) {
                                if (brdg[j * 4 + l] * brdir > 0) {
                                     newptr = abs(brdg[j * 4 + l]);
                                     if (abs(newptr - brptr) && ((newptr - brptr) /
                                         abs(newptr - brptr)) == brdir &&
                                        ((j - i) > blgszs && abs(newptr - brptr) <= blgszs ||
                                         (j - i) <= blgszs && abs(newptr - brptr) <= bkdszl)) {
                                          nwptpl = 1;
                                          if (abs(brdg[(newptr - 1) * 4]) == (j + 1)) nwptpl = 0;
                                          for (m = i; m <= j; m++) {
                                               if (allresidues[m]->ssbond[2] == ' ')
                                                    allresidues[m]->ssbond[2] = 'E';
                                          }
                                          if (brdg[i * 4 + k + 2] < 0)
                                               allresidues[i]->ssbond[2] = 'e';
                                          if (brdg[j * 4 + l + 2] < 0)
                                               allresidues[j]->ssbond[2] = 'e';
                                          stsht = brptr - 1;
                                          finsht   = newptr - 1;
                                          if (stsht > finsht) {
                                               stsht = newptr - 1;
                                               finsht = brptr - 1;
                                          }
                                          for (m = stsht; m <= finsht; m++) {
                                               if (allresidues[m]->ssbond[2] == ' ')
                                                    allresidues[m]->ssbond[2] = 'E';
                                          }
                                          if (brdg[(brptr - 1) * 4 + brptpl + 2] < 0)
                                               allresidues[brptr - 1]->ssbond[2] = 'e';
                                          if (brdg[(newptr - 1) * 4 + nwptpl + 2] < 0)
                                               allresidues[newptr - 1]->ssbond[2] = 'e';
                                          if (!strcd[2 * i + k]) {
                                               nxstrn++;
                                               strcd[2 * i + k] = nxstrn * brdir;
                                               strcd[2 * (brptr - 1) + brptpl] = nxstrn * brdir;
                                          }
                                          strcd[2 * j + l] = strcd[2 * i + k];
                                          strcd[2*(newptr-1)+nwptpl] = strcd[2*(brptr-1)+brptpl];
                                          found = 1;
                                          break;
                                     }
                                }
                           }
                           if (found) break;
                      }
                      wkch = 'b';
                      if (brdir < 0) wkch = 'B';
                      allresidues[i]->ssbond[3] = wkch;
                      allresidues[brptr - 1]->ssbond[3] = wkch;
                 }
            }
       }

       char strdch;
       int nstchs = 26, bridg1 = 4, bridg2 = 5;
       int chcode, stchpl;
       stsht = -1;
       finsht = -1;
       for (i = 0; i < n_allresidues; i++) {
            if (allresidues[i]->ssbond[2] == ' ') {
                 if (stsht >= 0 && finsht >= 0) {
                      finsht++;
                      stbrg(strcd, brdg, stsht, finsht, allresidues);
                 }
                 stsht = -1;
                 finsht = -1;
            } else {
                 if (stsht < 0) stsht = i;
                 finsht = i;
            }
       }
       if (stsht >= 0 && finsht >= 0) {
            finsht++;
            stbrg(strcd, brdg, stsht, finsht, allresidues);
       }

       int shtsum = 0;
       for (i = 0; i < n_allresidues; i++) {
            fdshus(i, allresidues, shtcod, stsht, finsht, n_allresidues);
            if (stsht < 0) break;

            shtsum++;
            for (j = stsht; j < finsht; j++) {
                 shtcod[j] = shtsum;
                 for (l = 0; l < 2; l++) {
                      if (allresidues[j]->bridpt[l]) {
                           if (!shtcod[allresidues[j]->bridpt[l] - 1]) {
                                setsht(allresidues[j]->bridpt[l] - 1, shtsum, shtcod,
                                        allresidues, n_allresidues);
                           }
                      }
                 }
            }
            found = 1;
            while (found) {
                 found = fdshps(finsht, shtsum, shtcod, allresidues, n_allresidues);
            }
            i = finsht;
       }

       for (i = 0; i < n_allresidues; i++) {
            if (shtcod[i])
                 allresidues[i]->ssbond1 = shtcod[i];
            else allresidues[i]->ssbond1 = 0;
       }

       shtsum = 0;
       for (i = 0; i < n_allresidues; i++) {
            for (j = 0; j < 2; j++) {
                 if (brdg[i * 4 + j] && !strcd[i * 2 + j] &&
                     abs(brdg[i * 4 + j]) > (i + 1)) {
                      shtsum++;
                      brptr = abs(brdg[i * 4 + j]);
                      chcode = shtsum % nstchs;
                      if (!chcode) chcode = nstchs;
                      if (brdg[i * 4 + j] < 0)
                           strdch = stchap[chcode - 1];
                      else strdch = stchpa[chcode - 1];
                      stchpl = bridg1;
                      if (allresidues[i]->ssbond[bridg1] != ' ')
                           stchpl = bridg2;
                      allresidues[i]->ssbond[stchpl] = strdch;
                      allresidues[i]->bridpt[stchpl - bridg1] = brptr;
                      stchpl = bridg1;
                      if (allresidues[brptr-1]->ssbond[bridg1] != ' ')
                           stchpl = bridg2;
                      allresidues[brptr-1]->ssbond[stchpl] = strdch;
                      allresidues[brptr-1]->bridpt[stchpl - bridg1] = i + 1;
                 }
            }
       }

       delete [] strcd;
       delete [] shtcod;
       delete [] brdg;

       hlxset(n_allresidues, allresidues, 0, 4, 'H', 'h');
       for (i = 0; i < n_allresidues; i++) {
            if (allresidues[i]->ssbond[2] != ' ') {
                 if (allresidues[i]->ssbond[9] == ' ')
                      allresidues[i]->ssbond[9] = 'E';
                 if (allresidues[i]->ssbond[10] == ' ' ||
                     allresidues[i]->ssbond[10] == 'h' ||
                     allresidues[i]->ssbond[10] == 'e')
                      allresidues[i]->ssbond[10] = 'E';
                 if (allresidues[i]->ssbond[2] == 'e') {
                      if (i && allresidues[i-1]->ssbond[10] == ' ' ||
                          allresidues[i-1]->ssbond[10] == 'B')
                           allresidues[i-1]->ssbond[10] = 'e';
                      if ((i+1) < n_allresidues && i &&
                          allresidues[i+1]->ssbond[10] == ' ' ||
                          allresidues[i-1]->ssbond[10] == 'B')
                           allresidues[i+1]->ssbond[10] = 'e';
                 }
                 if (allresidues[i]->ssbond[3] != ' ') {
                      if (allresidues[i]->ssbond[9] == ' ')
                           allresidues[i]->ssbond[9] = 'B';
                      if (allresidues[i]->ssbond[10] == ' ')
                           allresidues[i]->ssbond[10] = 'B';
                 }
            }
       }
       hlxset(n_allresidues, allresidues, 6, 3, 'G', 'g');
       hlxset(n_allresidues, allresidues, 7, 5, 'I', 'i');
       hlxres(n_allresidues, allresidues, 'G', 'g', 'T', 't', 3);
       hlxres(n_allresidues, allresidues, 'I', 'i', 'T', 't', 5);

       if (!_exist_helix && need_helix) find_helices(n_allresidues, allresidues);
       if (!_exist_sheet && need_sheet) find_sheets(n_allresidues, allresidues, support_file);
       if (!_exist_ssbond && need_ssbond) find_ssbonds(n_allresidues, allresidues);

       for (i = 0; i < n_allresidues; i++) {
            if (allresidues[i]) delete allresidues[i];
       }
       delete [] allresidues;
}

void _MAXIT::find_helices(const int seqlen, _NRESIDUE **residues)
{
       int i, j, n, *start = NULL, *stop = NULL, stop_count = 0, start_count = 0;

       n = seqlen/4+1;
       start = new int[n];
       stop = new int[n];
       for (i = 0; i < seqlen; i++) {
            if (i == 0 || i && residues[i-1]->chnbreak) {
                 if (residues[i]->ssbond[10] == 'H' ||
                     residues[i]->ssbond[10] == 'h' ||
                     residues[i]->ssbond[10] == 'G' ||
                     residues[i]->ssbond[10] == 'g')
                      start[start_count++] = i;
            } else {
                 if (residues[i]->ssbond[10] == 'H' &&
                     residues[i-1]->ssbond[10] != 'h' &&
                     residues[i-1]->ssbond[10] != 'H' ||
                     residues[i]->ssbond[10] == 'h' &&
                     residues[i-1]->ssbond[10] != 'H' ||
                     residues[i]->ssbond[10] == 'G' &&
                     residues[i-1]->ssbond[10] != 'g' &&
                     residues[i-1]->ssbond[10] != 'G' ||
                     residues[i]->ssbond[10] == 'g' &&
                     residues[i-1]->ssbond[10] != 'G') {
                      if (start_count) {
                           if (start[start_count-1] != i)
                                start[start_count++] = i;
                      } else start[start_count++] = i;
                 } else if (i < (seqlen - 1) &&
                     residues[i]->ssbond[10] == 'h' &&
                     residues[i+1]->ssbond[10] == 'H') {
                      if (start_count) {
                           if (start[start_count-1] != i)
                                start[start_count++] = i;
                      } else start[start_count++] = i;
                 }
            }
       }
       for (i = 0; i < seqlen; i++) {
            if (i == (seqlen - 1) || residues[i]->chnbreak) {
                 if (residues[i]->ssbond[10] == 'H' ||
                     residues[i]->ssbond[10] == 'h' ||
                     residues[i]->ssbond[10] == 'G' ||
                     residues[i]->ssbond[10] == 'g') {
                      if (stop_count) {
                           if (stop[stop_count-1] != i)
                                stop[stop_count++] = i;
                      } else stop[stop_count++] = i;
                 }
            } else {
                 if (residues[i]->ssbond[10] == 'H' &&
                     residues[i+1]->ssbond[10] != 'h' &&
                     residues[i+1]->ssbond[10] != 'H' ||
                     residues[i]->ssbond[10] == 'h' &&
                     residues[i+1]->ssbond[10] != 'H' ||
                     residues[i]->ssbond[10] == 'G' &&
                     residues[i+1]->ssbond[10] != 'g' &&
                     residues[i+1]->ssbond[10] != 'G' ||
                     residues[i]->ssbond[10] == 'g' &&
                     residues[i+1]->ssbond[10] != 'G') {
                      if (stop_count) {
                           if (stop[stop_count-1] != i)
                                stop[stop_count++] = i;
                      } else stop[stop_count++] = i;
                 } else if (i && residues[i-1]->ssbond[10] == 'H' &&
                     residues[i]->ssbond[10] == 'h') {
                      if (stop_count) {
                           if (stop[stop_count-1] != i)
                                stop[stop_count++] = i;
                      } else stop[stop_count++] = i;
                 }
            }
       }
       if (start_count != stop_count) {
            printf("start_count=%d stop_count=%d\n", start_count, stop_count);
            delete [] start;
            delete [] stop;
            return;
       }

       for (i = 1; i < start_count; i++) {
            if (start[i] == stop[i] && (stop[i-1]+1) == start[i] &&
                !residues[stop[i-1]]->chnbreak) {
                 if ((residues[start[i]]->ssbond[10] == 'G' ||
                      residues[start[i]]->ssbond[10] == 'g') &&
                     (residues[start[i-1]]->ssbond[10] == 'G' ||
                      residues[start[i-1]]->ssbond[10] == 'g') ||
                     (residues[start[i]]->ssbond[10] == 'H' ||
                      residues[start[i]]->ssbond[10] == 'h') &&
                     (residues[start[i-1]]->ssbond[10] == 'H' ||
                      residues[start[i-1]]->ssbond[10] == 'h')) {
                      stop[i-1] = stop[i];
                      for (j = i; j < start_count - 1; j++) {
                           start[j] = start[j + 1];
                           stop[j] = stop[j + 1];
                      }
                      start_count--;
                 }
            }
       }

       for (i = 0; i < start_count - 1; i++) {
            if (start[i] == stop[i] && !residues[start[i]]->chnbreak &&
                start[i+1] == (stop[i]+1)) {
                 if ((residues[start[i]]->ssbond[10] == 'G' ||
                      residues[start[i]]->ssbond[10] == 'g') &&
                     (residues[start[i+1]]->ssbond[10] == 'G' ||
                      residues[start[i+1]]->ssbond[10] == 'g') ||
                     (residues[start[i]]->ssbond[10] == 'H' ||
                      residues[start[i]]->ssbond[10] == 'h') &&
                     (residues[start[i+1]]->ssbond[10] == 'H' ||
                      residues[start[i+1]]->ssbond[10] == 'h')) {
                      stop[i] = stop[i+1];
                      for (j = i + 1; j < start_count - 1; j++) {
                           start[j] = start[j + 1];
                           stop[j] = stop[j + 1];
                      }
                      start_count--;
                 }
            }
       }

       for (i = 0; i < start_count; i++) {
            get_helix_cards(residues[start[i]]->chn, residues[start[i]]->res,
                residues[stop[i]]->res, residues[stop[i]]->ssbond[10], i+1);
       }
       delete [] start;
       delete [] stop;
}

void _MAXIT::find_ssbonds(const int seqlen, _NRESIDUE **residues)
{
       int i, j, partpl, initIndex, initRes, endIndex, endRes;
       int k, l, num_i = 0, num_j = 0;
       _ATOM **atomi = NULL, **atomj = NULL;
       COORD coordi, coordj;
       double partds, dist;
       _RESIDUE *res_i = NULL, *res_j = NULL;

       for (i = 0; i < seqlen - 1; i++) {
            if (strcmp(residues[i]->res->ResName(), "CYS")) continue;
            atomi = residues[i]->res->find_atom("SG", num_i);
            if (!atomi) continue;

            partpl = 0;
            partds = 10.0;
            for (j = i + 1; j < seqlen; j++) {
                 if (strcmp(residues[j]->res->ResName(), "CYS")) continue;
                 atomj = residues[j]->res->find_atom("SG", num_j);
                 if (!atomj) continue;
                 for (k = 0; k < num_i; k++) {
                      for (l = 0; l < num_j; l++) {
                           if (!strcmp(atomi[k]->alt_loc(), atomj[l]->alt_loc()) ||
                               !strcmp(atomi[k]->alt_loc(), "") ||
                               !strcmp(atomj[l]->alt_loc(), "")) {
                                coordi = atomi[k]->orig();
                                coordj = atomj[l]->orig();
                                dist = vector_distance(coordi, coordj);
                                if (dist < 3.0) {
                                     if (dist < partds) {
                                          partds = dist;
                                          partpl = j;
                                     }
                                }
                           }
                      }
                 }
                 if (atomj) delete [] atomj; atomj = NULL; 
            }
            if (atomi) delete [] atomi; atomi = NULL;

            if (partpl) {
                 res_i = residues[i]->res;
                 res_j = residues[partpl]->res;
                 if (_molecules[0]->find_pdb_residue(res_i->GetFirstAtom()->pdb_chnid(),
                      res_i->GetFirstAtom()->pdb_resnam(), res_i->GetFirstAtom()->pdb_resnum(),
                      res_i->GetFirstAtom()->ins_code(), initIndex, initRes) &&
                     _molecules[0]->find_pdb_residue(res_j->GetFirstAtom()->pdb_chnid(),
                      res_j->GetFirstAtom()->pdb_resnam(), res_j->GetFirstAtom()->pdb_resnum(),
                      res_j->GetFirstAtom()->ins_code(), endIndex, endRes)) {
                      Insert_a_SSbond(res_i->GetFirstAtom()->pdb_resnam(), initIndex, initRes,
                             res_j->GetFirstAtom()->pdb_resnam(), endIndex, endRes, "", "");
                 }
            }
       }
}

static void engset(const double energy, _NRESIDUE **residues, const int donres,
                   const int accres)
{
       int acpl, donpl, rejptr;

       donpl = engpl(energy, &(residues[donres]->energy[2]));
       acpl  = engpl(energy, &(residues[accres]->energy[0]));
       if (donpl && acpl) {
            if (residues[donres]->hbond[3]) {
                 rejptr = residues[donres]->hbond[3];
                 rejbnd(donres, residues[rejptr - 1], 2);
            }
            if (residues[accres]->hbond[1]) {
                 rejptr = residues[accres]->hbond[1];
                 rejbnd(accres, residues[rejptr - 1], 0);
            }
            if (acpl == 1) {
                 residues[accres]->energy[1] = residues[accres]->energy[0];
                 residues[accres]->hbond[1]  = residues[accres]->hbond[0];
            }
            if (donpl == 1) {
                 residues[donres]->energy[3] = residues[donres]->energy[2];
                 residues[donres]->hbond[3]  = residues[donres]->hbond[2];
            }
            residues[accres]->energy[acpl -1] = energy;
            residues[accres]->hbond[acpl - 1] = donres + 1;
            residues[donres]->energy[donpl+1] = energy;
            residues[donres]->hbond[donpl+ 1] = accres + 1;
       }
}

static int xeqy(const double x, const double y)
{
       double acurcy = 0.0001;
       return (fabs(x - y) < acurcy);
}

static int engpl(const double energy, double *honde)
{
       int index = 0;
       if (energy < honde[0]) {
            index = 1;
       } else {
            if (energy < honde[1]) index = 2;
       }
       return index;
}

static void rejbnd(const int rejptr, _NRESIDUE *residue, const int start)
{
       int rejpl;

       rejpl = 1;
       if (residue->hbond[start] != (rejptr + 1)) rejpl++;
       if (rejpl == 1) {
            residue->hbond[start] = residue->hbond[start + 1];
            residue->energy[start] = residue->energy[start + 1];
       }
       residue->hbond[start + 1] = 0;
       residue->energy[start + 1] = 0;
}

static void stbrg(int *strcd, int *brdg, const int stsht, const int finsht, _NRESIDUE **allresidues)
{
       int lststd = 0, bridg1 = 4, bridg2 = 5, nstchs = 26;
       int j = 0, l = 0, m, brptpl, bstval, chcode, stchpl, thstpl, nxstpl;
       char strdch;

       while (1) {
            whstrd(strcd, stsht, finsht, j, l, lststd, bstval);
            if (bstval) {
                 lststd = abs(strcd[2 * j + l]);
                 chcode = lststd % nstchs;
                 if (!chcode) chcode = nstchs;
                 if (strcd[2 * j + l] < 0)
                      strdch = stchap[chcode - 1];
                 else strdch = stchpa[chcode - 1];
                 stchpl = bridg1;
                 if (allresidues[j]->ssbond[bridg1] != ' ') stchpl = bridg2;
                 if (stchpl != bridg2) {
                      thstpl = j;
                      do {
                           nxstpl = strdpl(thstpl, finsht, strcd[2 * j + l], strcd);
                           if (nxstpl < 0) break;
                           if (allresidues[nxstpl]->ssbond[bridg1] != ' ')
                                stchpl = bridg2;
                           thstpl = nxstpl;
                      } while (stchpl != bridg2);
                 }
                 allresidues[j]->ssbond[stchpl] = strdch;
                 allresidues[j]->bridpt[stchpl - bridg1] = abs(brdg[j * 4 + l]);
                 thstpl = j;
                 while (1) {
                      nxstpl = strdpl(thstpl, finsht, strcd[2 * j + l], strcd);
                      if (nxstpl < 0) break;
                      for (m = thstpl + 1; m < nxstpl; m++)
                           allresidues[m]->ssbond[stchpl] = '*';
                      allresidues[nxstpl]->ssbond[stchpl] = strdch;
                      brptpl = 0;
                      if (strcd[2 * nxstpl] != strcd[2 * j + l]) brptpl = 1;
                      allresidues[nxstpl]->bridpt[stchpl - bridg1] =
                                       abs(brdg[nxstpl * 4 + brptpl]);
                      thstpl = nxstpl;
                 }
            } else break;
       }
}

static void whstrd(int *strcd, const int stsht, const int finsht, int &stres, int &stpl,
              const int lststd, int &bstval)
{
       int i, j;
       bstval = 0;
       for (i = stsht; i < finsht; i++) {
            for (j = 0; j < 2; j++) {
                 if (abs(strcd[2 * i + j]) > lststd) {
                      if (!bstval) {
                           bstval = abs(strcd[2 * i + j]);
                           stres = i;
                           stpl = j;
                      } else {
                           if (abs(strcd[2 * i + j]) < bstval) {
                                bstval = abs(strcd[2 * i + j]);
                                stres = i;
                                stpl = j;
                           }
                      }
                 }
            }
       }
}

static int strdpl(const int stpl, const int finpl, const int thcode, int *strand)
{
       int i;
       for (i = stpl + 1; i < finpl; i++) {
            if (strand[2 * i] == thcode) break;
            if (strand[2*i+1] == thcode) break;
       }
       if (i < finpl)
            return i;
       else return -1;
}

static void fdshus(const int stpt, _NRESIDUE **residues, int *shtcod, int &stsht,
               int &finsht, const int seqlen)
{
       int i;
       stsht = -1;
       for (i = stpt; i < seqlen; i++) {
            if (residues[i]->ssbond[2] != ' ' && !shtcod[i]) break;
       }
       if (i == seqlen) return;
       stsht = i;
       for (; i < seqlen; i++) {
            if (residues[i]->ssbond[2] == ' ') break;
            finsht = i;
       }
       finsht++;
}

static void setsht(const int refpl, const int shtsum, int *shtcod, _NRESIDUE **residues,
               const int seqlen)
{
       int i, stsht = 0, finsht = -1;

       for (i = refpl; i < seqlen; i++) {
            if (residues[i]->ssbond[2] == ' ') break;
            finsht = i;
       }
       for (i = refpl; i >= 0; i--) {
            if (residues[i]->ssbond[2] == ' ') break;
            stsht = i;
       }
       for (i = stsht; i <= finsht; i++)
            shtcod[i] = -shtsum;
}

static int fdshps(const int stpt, const int shtsum, int *shtcod, _NRESIDUE **residues,
               const int seqlen)
{
       int i, j, found = 0, stsht = -1, finsht = -1;

       for (i = stpt; i < seqlen; i++) {
            if (shtcod[i] == (-shtsum)) break;
       }
       if (i == seqlen) return found;

       found = 1;
       stsht = i;
       for (; i < seqlen; i++) {
            if (shtcod[i] != (-shtsum)) break;
            finsht = i;
       }
       for (i = stsht; i <= finsht; i++) {
            shtcod[i] = shtsum;
            for (j = 0; j < 2; j++) {
                 if (residues[i]->bridpt[j]) {
                      if (!shtcod[residues[i]->bridpt[j] - 1]) {
                           setsht(residues[i]->bridpt[j] - 1, shtsum, shtcod,
                                     residues, seqlen);
                      }
                 }
            }
       }
       return found;
}

static void hlxset(const int seqlen, _NRESIDUE **residues, const int hlxpos,
               const int hlxsz, const char hlxch, const char altch)
{
       int i, j;

       for (i = 1; i < seqlen - hlxsz; i++) {
            if (residues[i]->ssbond[hlxpos] == '>' ||
                residues[i]->ssbond[hlxpos] == 'X') {
                 if (residues[i-1]->ssbond[hlxpos] == '>' ||
                     residues[i-1]->ssbond[hlxpos] == 'X') {
                      for (j = 0; j < hlxsz; j++) {
                           if (residues[i+j]->ssbond[9] == ' ')
                                residues[i+j]->ssbond[9] = hlxch;
                           if (residues[i+j]->ssbond[10] == ' ' ||
                               residues[i+j]->ssbond[10] == altch)
                                residues[i+j]->ssbond[10] = hlxch;
                      }
                      if (residues[i-1]->ssbond[10] == ' ')
                           residues[i-1]->ssbond[10] = altch;
                      if (residues[i+hlxsz]->ssbond[10] == ' ')
                           residues[i+hlxsz]->ssbond[10] = altch;
                 }
            }
       }
}

static void hlxres(const int seqlen, _NRESIDUE **residues, const char hlxch,
               const char altch, const char turnch, const char alturn, const int hlxsz)
{
       int index[2], i, j, k, stpos;
  
       index[0] = 9; index[1] = 10;
       for (i = 0; i < 2; i++) {
            stpos = -1;
            for (j = 0; j < seqlen; j++) {
                 if (residues[j]->ssbond[index[i]] == hlxch ||
                     residues[j]->ssbond[index[i]] == altch) {
                      if (stpos < 0) stpos = j;
                 } else {
                      if (stpos >= 0) {
                           if ((j - stpos) < hlxsz) {
                                for (k = stpos; k < j; k++) {
                                     if (residues[k]->ssbond[index[i]] == hlxch)
                                          residues[k]->ssbond[index[i]] = turnch;
                                     else residues[k]->ssbond[index[i]] = alturn;
                                }
                           }
                           stpos = -1;
                      }
                 }
            }
            if (stpos >= 0) {
                 if ((seqlen - stpos) < hlxsz) {
                      for (k = stpos; k < seqlen; k++) {
                           if (residues[k]->ssbond[index[i]] == hlxch)
                                residues[k]->ssbond[index[i]] = turnch;
                           else residues[k]->ssbond[index[i]] = alturn;
                      }
                 }
            }
       }
}

static void get_helix_cards(_CHAIN *chain, _RESIDUE *strR, _RESIDUE *endR,
                   const char type, const int count)
{
       int helix_type = 1;
       char code[10];
       _FIELD *init = NULL, *end = NULL;

       if (type == 'G' || type == 'g') helix_type = 5;
       sprintf(code, "%d", count);
       init = chain->find_pdb_Residue(strR->GetFirstAtom()->pdb_resnam(),
                       strR->GetFirstAtom()->pdb_resnum(),
                       strR->GetFirstAtom()->ins_code());
       end = chain->find_pdb_Residue(endR->GetFirstAtom()->pdb_resnam(),
                       endR->GetFirstAtom()->pdb_resnum(),
                       endR->GetFirstAtom()->ins_code());
       if (init && end)
            chain->Insert_a_Helix(code, init->ResIndex, end->ResIndex,
                      helix_type, "");
}

