/*
FILE:     stereochemistry.h
*/
/*
VERSION:  1.700
*/
/*
DATE:     7/13/2005
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2005 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#ifndef _STEREOCHEM_H_
#define _STEREOCHEM_H_

#include "TblIndexObj.h"

#define NUM_OF_CHIRALITY  26
#define NUM_EZ_CHECK      12
#define NUM_MP_CHECK      11
#define NUM_ATOM           6
#define FIRST              0
#define SECOND             1
#define THIRD              2
#define FOURTH             3
#define FIFTH              4
#define SIXTH              5

typedef struct {
        char old_name[5];
        char new_name[5];
} CHANGE_LIST;

typedef struct {
        char  center_atom[5];
        char  first_atom[5];
        char  second_atom[5];
        char  check_atom[5];
        char  change_to[5];
        int   isHydrogen;
        float correct_value;
        int   n_change;
        CHANGE_LIST change_list[3];
} CHIRALITY;

typedef struct {
        int   num;
        int   isPhosphorus;
        CHIRALITY chiral[2];
} CHIRALITYS;

typedef struct {
        const char *residue_name;
        int  num_residues;
        CHIRALITYS *lists;
} CHIRALITYS_POOL;

static CHIRALITYS chirality_na_a_g[11] = {
       { 1, 0, {{"C1*", "O4*", "C2*", "N9",  "N9",  0,  1.0, 0}} },
       { 1, 0, {{"C1*", "O4*", "C2*", "H1*", "H1*", 1, -1.0, 0}} },
       { 1, 0, {{"C2*", "C1*", "C3*", "O2*", "O2*", 0, -1.0, 0}} },
       { 1, 0, {{"C2*", "C1*", "C3*", "H2*", "H2*", 1,  1.0, 0}} },
       { 2, 0, {{"C2*", "C1*", "C3*", "1H2*","2H2*",1,  1.0, 0},
                {"C2*", "C1*", "C3*", "2H2*","1H2*",1, -1.0, 0}} },
       { 1, 0, {{"C3*", "C2*", "C4*", "O3*", "H3*", 0, -1.0, 0}} },
       { 1, 0, {{"C3*", "C2*", "C4*", "H3*", "O3*", 1,  1.0, 0}} },
       { 1, 0, {{"C4*", "C3*", "O4*", "C5*", "H4*", 0,  1.0, 0}} },
       { 1, 0, {{"C4*", "C3*", "O4*", "H4*", "C5*", 1, -1.0, 0}} },
       { 2, 0, {{"C5*", "O5*", "C4*", "1H5*","2H5*",1,  1.0, 0},
                {"C5*", "O5*", "C4*", "2H5*","1H5*",1, -1.0, 0}} },
       { 2, 1, {{"P",   "O5*", "O3*", "O2P", "O1P", 0,  1.0, 0},
                {"P",   "O5*", "O3*", "O1P", "O2P", 0, -1.0, 0}} }
};

static CHIRALITYS chirality_na_other[11] = {
       { 1, 0, {{"C1*", "O4*", "C2*", "N1",  "N1",  0,  1.0, 0}} },
       { 1, 0, {{"C1*", "O4*", "C2*", "H1*", "H1*", 1, -1.0, 0}} },
       { 1, 0, {{"C2*", "C1*", "C3*", "O2*", "O2*", 0, -1.0, 0}} },
       { 1, 0, {{"C2*", "C1*", "C3*", "H2*", "H2*", 1,  1.0, 0}} },
       { 2, 0, {{"C2*", "C1*", "C3*", "1H2*","2H2*",1,  1.0, 0},
                {"C2*", "C1*", "C3*", "2H2*","1H2*",1, -1.0, 0}} },
       { 1, 0, {{"C3*", "C2*", "C4*", "O3*", "H3*", 0, -1.0, 0}} },
       { 1, 0, {{"C3*", "C2*", "C4*", "H3*", "O3*", 1,  1.0, 0}} },
       { 1, 0, {{"C4*", "C3*", "O4*", "C5*", "H4*", 0,  1.0, 0}} },
       { 1, 0, {{"C4*", "C3*", "O4*", "H4*", "C5*", 1, -1.0, 0}} },
       { 2, 0, {{"C5*", "O5*", "C4*", "1H5*","2H5*",1,  1.0, 0},
                {"C5*", "O5*", "C4*", "2H5*","1H5*",1, -1.0, 0}} },
       { 2, 1, {{"P",   "O5*", "O3*", "O2P", "O1P", 0,  1.0, 0},
                {"P",   "O5*", "O3*", "O1P", "O2P", 0, -1.0, 0}} }
};

static CHIRALITYS chirality_ALA[3] = {
       { 1, 0, {{"CA", "N",  "C",   "CB",  "CB",  0,  1.0, 0}} },
       { 1, 0, {{"CA", "N",  "CB",  "C",   "C",   0, -1.0, 0}} },
       { 1, 0, {{"CA", "N",  "C",   "HA",  "HA",  1, -1.0, 0}} }
};

static CHIRALITYS chirality_ARG[6] = {
       { 1, 0, {{"CA", "N",  "C",   "CB",  "CB",  0,  1.0, 0}} },
       { 1, 0, {{"CA", "N",  "CB",  "C",   "C",   0, -1.0, 0}} },
       { 1, 0, {{"CA", "N",  "C",   "HA",  "HA",  1, -1.0, 0}} },
       { 2, 0, {{"CB", "CA", "CG",  "1HB", "2HB", 1, -1.0, 0},
                {"CB", "CA", "CG",  "2HB", "1HB", 1,  1.0, 0}} },
       { 2, 0, {{"CG", "CB", "CD",  "1HG", "2HG", 1, -1.0, 0},
                {"CG", "CB", "CD",  "2HG", "1HG", 1,  1.0, 0}} },
       { 2, 0, {{"CD", "CG", "NE",  "1HD", "2HD", 1, -1.0, 0},
                   {"CD", "CG", "NE",  "2HD", "1HD", 1,  1.0, 0}} }
};

static CHIRALITYS chirality_ASN[4] = {
       { 1, 0, {{"CA", "N",  "C",   "CB",  "CB",  0,  1.0, 0}} },
       { 1, 0, {{"CA", "N",  "CB",  "C",   "C",   0, -1.0, 0}} },
       { 1, 0, {{"CA", "N",  "C",   "HA",  "HA",  1, -1.0, 0}} },
       { 2, 0, {{"CB", "CA", "CG",  "1HB", "2HB", 1, -1.0, 0},
                {"CB", "CA", "CG",  "2HB", "1HB", 1,  1.0, 0}} }
};

static CHIRALITYS chirality_CYS[4] = {
       { 1, 0, {{"CA", "N",  "C",   "CB",  "CB",  0,  1.0, 0}} },
       { 1, 0, {{"CA", "N",  "CB",  "C",   "C",   0, -1.0, 0}} },
       { 1, 0, {{"CA", "N",  "C",   "HA",  "HA",  1, -1.0, 0}} },
       { 2, 0, {{"CB", "CA", "SG",  "1HB", "2HB", 1, -1.0, 0},
                {"CB", "CA", "SG",  "2HB", "1HB", 1,  1.0, 0}} }
};

static CHIRALITYS chirality_GLN[5] = {
       { 1, 0, {{"CA", "N",  "C",   "CB",  "CB",  0,  1.0, 0}} },
       { 1, 0, {{"CA", "N",  "CB",  "C",   "C",   0, -1.0, 0}} },
       { 1, 0, {{"CA", "N",  "C",   "HA",  "HA",  1, -1.0, 0}} },
       { 2, 0, {{"CB", "CA", "CG",  "1HB", "2HB", 1, -1.0, 0},
                {"CB", "CA", "CG",  "2HB", "1HB", 1,  1.0, 0}} },
       { 2, 0, {{"CG", "CB", "CD",  "1HG", "2HG", 1, -1.0, 0},
                {"CG", "CB", "CD",  "2HG", "1HG", 1,  1.0, 0}} }
};

static CHIRALITYS chirality_GLY[4] = {
       { 1, 0, {{"CA", "N",  "C",   "CB",  "CB",  0,  1.0, 0}} },
       { 1, 0, {{"CA", "N",  "CB",  "C",   "C",   0, -1.0, 0}} },
       { 1, 0, {{"CA", "N",  "C",   "HA",  "HA",  1, -1.0, 0}} },
       { 2, 0, {{"CA", "N",  "C",   "1HA", "2HA", 1, -1.0, 0},
                {"CA", "N",  "C",   "2HA", "1HA", 1,  1.0, 0}} }
};

static CHIRALITYS chirality_ILE[7] = {
       { 1, 0, {{"CA", "N",  "C",   "CB",  "CB",  0,  1.0, 0}} },
       { 1, 0, {{"CA", "N",  "CB",  "C",   "C",   0, -1.0, 0}} },
       { 1, 0, {{"CA", "N",  "C",   "HA",  "HA",  1, -1.0, 0}} },
       { 1, 0, {{"CB", "CA", "CG1", "CG2", "CG2", 0,  1.0, 0}} },
       { 1, 0, {{"CB", "CA", "CG2", "CG1", "CG1", 0, -1.0, 0}} },
       { 1, 0, {{"CB", "CA", "CG1", "HB",  "HB",  1, -1.0, 0}} },
       { 2, 0, {{"CG1","CB", "CD1", "1HG1","2HG1",1, -1.0, 0},
                {"CG1","CB", "CD1", "2HG1","1HG1",1,  1.0, 0}} }
};

static CHIRALITYS chirality_LEU[6] = {
       { 1, 0, {{"CA", "N",  "C",   "CB",  "CB",  0,  1.0, 0}} },
       { 1, 0, {{"CA", "N",  "CB",  "C",   "C",   0, -1.0, 0}} },
       { 1, 0, {{"CA", "N",  "C",   "HA",  "HA",  1, -1.0, 0}} },
       { 2, 0, {{"CB", "CA", "CG",  "1HB", "2HB", 1, -1.0, 0},
                {"CB", "CA", "CG",  "2HB", "1HB", 1,  1.0, 0}} },
       { 2, 0, {{"CG", "CB", "CD1", "CD2", "CD1", 0, -1.0, 3,
               {{"1HD2","1HD1"},{"2HD2","2HD1"},{"3HD2","3HD1"}}}, 
                {"CG", "CB", "CD2", "CD1", "CD2", 0,  1.0, 3,
               {{"1HD1","1HD2"},{"2HD1","2HD2"},{"3HD1","3HD2"}}}}},
       { 1, 0, {{"CG", "CB", "CD2", "HG",  "HG",  1, -1.0, 0}} }
};

static CHIRALITYS chirality_LYS[7] = {
       { 1, 0, {{"CA", "N",  "C",   "CB",  "CB",  0,  1.0, 0}} },
       { 1, 0, {{"CA", "N",  "CB",  "C",   "C",   0, -1.0, 0}} },
       { 1, 0, {{"CA", "N",  "C",   "HA",  "HA",  1, -1.0, 0}} },
       { 2, 0, {{"CB", "CA", "CG",  "1HB", "2HB", 1, -1.0, 0},
                {"CB", "CA", "CG",  "2HB", "1HB", 1,  1.0, 0}} },
       { 2, 0, {{"CG", "CB", "CD",  "1HG", "2HG", 1, -1.0, 0},
                {"CG", "CB", "CD",  "2HG", "1HG", 1,  1.0, 0}} },
       { 2, 0, {{"CD", "CG", "CE",  "1HD", "2HD", 1, -1.0, 0},
                {"CD", "CG", "CE",  "2HD", "1HD", 1,  1.0, 0}} },
       { 2, 0, {{"CE", "CD", "NZ",  "1HE", "2HE", 1, -1.0, 0},
                {"CE", "CD", "NZ",  "2HE", "1HE", 1,  1.0, 0}} }
};

static CHIRALITYS chirality_MET[5] = {
       { 1, 0, {{"CA", "N",  "C",   "CB",  "CB",  0,  1.0, 0}} },
       { 1, 0, {{"CA", "N",  "CB",  "C",   "C",   0, -1.0, 0}} },
       { 1, 0, {{"CA", "N",  "C",   "HA",  "HA",  1, -1.0, 0}} },
       { 2, 0, {{"CB", "CA", "CG",  "1HB", "2HB", 1, -1.0, 0},
                {"CB", "CA", "CG",  "2HB", "1HB", 1,  1.0, 0}} },
       { 2, 0, {{"CG", "CB", "SD",  "1HG", "2HG", 1, -1.0, 0},
                {"CG", "CB", "SD",  "2HG", "1HG", 1,  1.0, 0}} }
};

static CHIRALITYS chirality_PRO[7] = {
       { 1, 0, {{"CA", "N",  "C",   "CB",  "CB",  0,  1.0, 0}} },
       { 1, 0, {{"CA", "N",  "CB",  "C",   "C",   0, -1.0, 0}} },
       { 1, 0, {{"CA", "N",  "C",   "HA",  "HA",  1, -1.0, 0}} },
       { 2, 0, {{"N",  "CA", "CD",  "HT1", "HT2", 1, -1.0, 0},
                {"N",  "CA", "CD",  "HT2", "HT1", 1,  1.0, 0}} },
       { 2, 0, {{"CB", "CA", "CG",  "1HB", "2HB", 1, -1.0, 0},
                {"CB", "CA", "CG",  "2HB", "1HB", 1,  1.0, 0}} },
       { 2, 0, {{"CG", "CB", "CD",  "1HG", "2HG", 1, -1.0, 0},
                {"CG", "CB", "CD",  "2HG", "1HG", 1,  1.0, 0}} },
       { 2, 0, {{"CD", "CG", "N",   "1HD", "2HD", 1, -1.0, 0},
                {"CD", "CG", "N",   "2HD", "1HD", 1,  1.0, 0}} }
};

static CHIRALITYS chirality_SER[4] = {
       { 1, 0, {{"CA", "N",  "C",   "CB",  "CB",  0,  1.0, 0}} },
       { 1, 0, {{"CA", "N",  "CB",  "C",   "C",   0, -1.0, 0}} },
       { 1, 0, {{"CA", "N",  "C",   "HA",  "HA",  1, -1.0, 0}} },
       { 2, 0, {{"CB", "CA", "OG",  "1HB", "2HB", 1, -1.0, 0},
                {"CB", "CA", "OG",  "2HB", "1HB", 1,  1.0, 0}} }
};

static CHIRALITYS chirality_THR[6] = {
       { 1, 0, {{"CA", "N",  "C",   "CB",  "CB",  0,  1.0, 0}} },
       { 1, 0, {{"CA", "N",  "CB",  "C",   "C",   0, -1.0, 0}} },
       { 1, 0, {{"CA", "N",  "C",   "HA",  "HA",  1, -1.0, 0}} },
       { 1, 0, {{"CB", "CA", "OG1", "CG2", "CG2", 0,  1.0, 0}} },
       { 1, 0, {{"CB", "CA", "CG2", "OG1", "OG1", 0, -1.0, 0}} },
       { 1, 0, {{"CB", "CA", "CG2", "HB",  "HB",  1,  1.0, 0}} }
};

static CHIRALITYS chirality_VAL[5] = {
       { 1, 0, {{"CA", "N",  "C",   "CB",  "CB",  0,  1.0, 0}} },
       { 1, 0, {{"CA", "N",  "CB",  "C",   "C",   0, -1.0, 0}} },
       { 1, 0, {{"CA", "N",  "C",   "HA",  "HA",  1, -1.0, 0}} },
       { 2, 0, {{"CB", "CA", "CG1", "CG2", "CG1", 0, -1.0, 3,
               {{"1HG2","1HG1"},{"2HG2","2HG1"},{"3HG2","3HG1"}}},
                {"CB", "CA", "CG2", "CG1", "CG2", 0,  1.0, 3,
               {{"1HG1","1HG2"},{"2HG1","2HG2"},{"3HG1","3HG2"}}}}},
       { 1, 0, {{"CB", "CA", "CG2", "HB",  "HB",  1, -1.0, 0}} }
};

static CHIRALITYS_POOL total_list[26] = {
       { "A",   11, chirality_na_a_g },
       { "ALA",  3, chirality_ALA },
       { "ARG",  6, chirality_ARG },
       { "ASN",  4, chirality_ASN },
       { "ASP",  4, chirality_ASN },
       { "C",   11, chirality_na_other },
       { "CYS",  4, chirality_CYS },
       { "G",   11, chirality_na_a_g },
       { "GLN",  5, chirality_GLN },
       { "GLU",  5, chirality_GLN },
       { "GLY",  4, chirality_GLY },
       { "HIS",  4, chirality_ASN },
       { "I",   11, chirality_na_a_g },
       { "ILE",  7, chirality_ILE },
       { "LEU",  6, chirality_LEU },
       { "LYS",  7, chirality_LYS },
       { "MET",  5, chirality_MET },
       { "PHE",  4, chirality_ASN },
       { "PRO",  7, chirality_PRO },
       { "SER",  4, chirality_SER },
       { "T",   11, chirality_na_other },
       { "THR",  6, chirality_THR },
       { "TRP",  4, chirality_ASN },
       { "TYR",  4, chirality_ASN },
       { "U",   11, chirality_na_other },
       { "VAL",  5, chirality_VAL }
};

typedef struct {
        char atom[5];
} CHECK_LIST;

typedef struct {
        char residue[5];
        int n_check;
        CHECK_LIST check_list[4];
        char atom1[5];
        char atom2[5];
        char atom3[5];
        char atom4[5];
        char no_atom[5];
        int n_change;
        CHANGE_LIST change_list[5];
} EZ_CHECK;

typedef struct {
        char residue[5];
        CHECK_LIST check_list[NUM_ATOM];
} MP_CHECK;

typedef struct {
        double torsion;
        int    old_index;
        int    new_index;
} MP_TORSION;

static EZ_CHECK ez_check[NUM_EZ_CHECK] = {
       { "A", 2, { {"1H6"}, {"2H6"} },
         "N1", "C6", "N6", "1H6", "", 1,
         {{ "1H6", "2H6" }}
       },
       { "ARG", 2, { {"NH1"}, {"NH2"} /* , "1HH1", "2HH1", "1HH2", "2HH2" */ },
         "CD", "NE", "CZ", "NH1", "", 3,
         {{ "NH1",  "NH2" }, { "1HH1", "1HH2" }, { "2HH1", "2HH2" }}
       },
       { "ARG", 2, { {"1HH1"}, {"2HH1"} },
         "NE", "CZ", "NH1", "1HH1", "", 1,
         {{ "1HH1", "2HH1" }}
       },
       { "ARG", 2, { {"1HH2"}, {"2HH2"} },
         "NE", "CZ", "NH2", "1HH2", "", 1,
         {{ "1HH2", "2HH2" }}
       },
       { "ASN", 2, { {"1HD2"}, {"2HD2"} },
         "CB", "CG", "ND2", "1HD2", "", 1,
         {{ "1HD2", "2HD2" }}
       },
       { "ASP", 2, { {"OD1"}, {"OD2"} },
         "CA", "CB", "CG", "OD1", "HD2", 1,
         {{ "OD1", "OD2" }}
       },
       { "C", 2, { {"1H4"}, {"2H4"} },
         "N3", "C4", "N4", "1H4", "", 1,
         {{ "1H4", "2H4" }}
       },
       { "G", 2, { {"1H2"}, {"2H2"} },
         "N1", "C2", "N2", "1H2", "", 1,
         {{ "1H2", "2H2" }}
       },
       { "GLN", 2, { {"1HE2"}, {"2HE2"} },
         "CG", "CD", "NE2", "1HE2", "", 1,
         {{ "1HE2", "2HE2" }}
       },
       { "GLU", 2, { {"OE1"}, {"OE2"} },
         "CB", "CG", "CD", "OE1", "HE2", 1,
         {{ "OE1", "OE2" }}
       },
       { "PHE", 2, { {"CD1"}, {"CD2"} },
         "CA", "CB", "CG", "CD1", "", 4,
         {{ "CD1", "CD2" }, { "CE1", "CE2" }, { "HD1", "HD2" }, { "HE1", "HE2" }}
       },
       { "TYR", 2, { {"CD1"}, {"CD2"} },
         "CA", "CB", "CG", "CD1", "", 4,
         {{ "CD1", "CD2" }, { "CE1", "CE2" }, { "HD1", "HD2" }, { "HE1", "HE2" }} 
       } 
};
/*
static EZ_CHECK _C_TERMINAL = {
       "", 2, { "O", "OXT" },
       "N", "CA", "C", "O", "HXT", 1,
       {{ "O", "OXT" }}
};

static MP_CHECK mp_check[NUM_MP_CHECK] = {
       { "ALA", { {"N"},  {"CA"},  {"CB"},  {"1HB"},  {"2HB"},  {"3HB"  }}},
       { "ILE", { {"CA"}, {"CB"},  {"CG2"}, {"1HG2"}, {"2HG2"}, {"3HG2" }}},
       { "ILE", { {"CB"}, {"CG1"}, {"CD1"}, {"1HD1"}, {"2HD1"}, {"3HD1" }}},
       { "LEU", { {"CB"}, {"CG"},  {"CD1"}, {"1HD1"}, {"2HD1"}, {"3HD1" }}},
       { "LEU", { {"CB"}, {"CG"},  {"CD2"}, {"1HD2"}, {"2HD2"}, {"3HD2" }}},
       { "LYS", { {"CD"}, {"CE"},  {"NZ"},  {"1HZ"},  {"2HZ"},  {"3HZ"  }}},
       { "MET", { {"CG"}, {"SD"},  {"CE"},  {"1HE"},  {"2HE"},  {"3HE"  }}},
       { "T",   { {"C4"}, {"C5"},  {"C5M"}, {"1H5M"}, {"2H5M"}, {"3H5M" }}},
       { "THR", { {"CA"}, {"CB"},  {"CG2"}, {"1HG2"}, {"2HG2"}, {"3HG2" }}},
       { "VAL", { {"CA"}, {"CB"},  {"CG1"}, {"1HG1"}, {"2HG1"}, {"3HG1" }}},
       { "VAL", { {"CA"}, {"CB"},  {"CG2"}, {"1HG2"}, {"2HG2"}, {"3HG2" }}}
};
*/
static TblIndexObj _chiralIndex;

static void _chiralIndex_init()
{
       CifString index;
       _chiralIndex.Clear();
       for (int i = 0; i < NUM_OF_CHIRALITY; i++) {
            index.Copy(total_list[i].residue_name);
            _chiralIndex.Add(index);
       }
}

static CHIRALITYS_POOL *find_chiral(const char *resname)
{
       CifString index(resname);
       int ret = _chiralIndex.Seek(index);
       if (ret >= 0) return (&total_list[ret]);
       return NULL;
}

#endif

