/*
FILE:     misc_util.C
*/
/*
VERSION:  1.700
*/
/*
DATE:     7/13/2005
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2005 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "misc_util.h"


void arb_rotation(COORD rot_mtx[3], COORD &va, double ang_deg)
{
       double vlen, c, s, dc;
       vlen = vector_length(va);
       if (vlen < XEPS) {
            rot_mtx[0].x = rot_mtx[1].y = rot_mtx[2].z = 1.0;
            rot_mtx[1].x = rot_mtx[2].x = rot_mtx[0].y =
            rot_mtx[2].y = rot_mtx[0].z = rot_mtx[1].z = 0.0;
       } else {
            vector_scale(va, 1.0 / vlen);
            ang_deg = deg2rad(ang_deg);
            c = cos(ang_deg);
            s = sin(ang_deg);
            dc = 1 - c;
            rot_mtx[0].x = c + dc * va.x * va.x;
            rot_mtx[1].x = va.x * va.y * dc - va.z * s;
            rot_mtx[2].x = va.x * va.z * dc + va.y * s;
            rot_mtx[0].y = va.x * va.y * dc + va.z * s;
            rot_mtx[1].y = c + dc * va.y * va.y;
            rot_mtx[2].y = va.y * va.z * dc - va.x * s;
            rot_mtx[0].z = va.x * va.z * dc - va.y * s;
            rot_mtx[1].z = va.y * va.z * dc + va.x * s;
            rot_mtx[2].z = c + dc * va.z * va.z;
       }
}

double deg2rad(const double ang)
{
       return ang * PI / 180.0;
}

double dot2ang(const double dotval)
{
       double ang_deg;
       if (dotval >= 1.0)
            ang_deg = 0.0;
       else if (dotval <= -1.0)
            ang_deg = 180.0;
       else ang_deg = rad2deg(acos(dotval));
       return ang_deg;
}

void get_vector(COORD &va, COORD &vref, double deg_ang, COORD &vo)
{
       COORD temp[3];
       if (vector_dot_product(va, vref) > XEPS) {
            fprintf(stderr, "va and vref should be perpendicular\n");
            exit (1);
       }
       arb_rotation(temp, vref, deg_ang);
       vo.x = temp[0].x * va.x + temp[1].x * va.y + temp[2].x * va.z;
       vo.y = temp[0].y * va.x + temp[1].y * va.y + temp[2].y * va.z;
       vo.z = temp[0].z * va.x + temp[1].z * va.y + temp[2].z * va.z;
       vector_normalize(vo);
}

double magang(COORD &va, COORD &vb)
{
       double ang_deg, temp;
       if (vector_length(va) < XEPS || vector_length(vb) < XEPS) ang_deg = 0.0;
       else {
            vector_normalize(va);
            vector_normalize(vb);
            temp = vector_dot_product(va, vb);
            ang_deg = dot2ang(temp);
            // ang_deg = (temp >= 1.0) ? 0.0 : rad2deg(acos(temp));
       }
       return ang_deg;
}

void multi_matrix(COORD o[3], const COORD a[3], const COORD b[3])
{
       int i;
       for (i = 0; i < 3; i++) {
            o[i].x = a[0].x * b[i].x + a[1].x * b[i].y + a[2].x * b[i].z;
            o[i].y = a[0].y * b[i].x + a[1].y * b[i].y + a[2].y * b[i].z;
            o[i].z = a[0].z * b[i].x + a[1].z * b[i].y + a[2].z * b[i].z;
       }
}

double rad2deg(const double ang)
{
       return ang * 180.0 / PI;
}

void vector_init(COORD &dest, const double x, const double y, const double z)
{
       dest.x = x; dest.y = y; dest.z = z;
}

void vector_copy(COORD &dest, const COORD src)
{
       dest.x = src.x; dest.y = src.y; dest.z = src.z;
}

void vector_sum(COORD &dest, const COORD v1, const COORD v2)
{
       dest.x = v1.x + v2.x; dest.y = v1.y + v2.y; dest.z = v1.z + v2.z;
}

void vector_difference(COORD &dest, const COORD v1, const COORD v2)
{
       dest.x = v1.x - v2.x; dest.y = v1.y - v2.y; dest.z = v1.z - v2.z;
}

void vector_scaled(COORD &dest, const double s, const COORD src)
{
       dest.x = s * src.x; dest.y = s * src.y; dest.z = s * src.z;
}

void vector_add(COORD &dest, const COORD v)
{
       dest.x += v.x; dest.y += v.y; dest.z += v.z;
}

void vector_subtract(COORD &dest, const COORD v)
{
       dest.x -= v.x; dest.y -= v.y; dest.z -= v.z;
}

void vector_scale(COORD &dest, const double s)
{
       dest.x *= s; dest.y *= s; dest.z *= s;
}

void vector_invert(COORD &v)
{
       v.x = 1.0 / v.x; v.y = 1.0 / v.y; v.z = 1.0 / v.z;
}

void vector_reverse(COORD &v)
{
       v.x = - v.x; v.y = - v.y; v.z = - v.z;
}

void vector_sum_scaled(COORD &dest, const COORD v1, const double s, const COORD v2)
{
       dest.x = v1.x + s * v2.x; dest.y = v1.y + s * v2.y;
       dest.z = v1.z + s * v2.z;
}

void vector_add_scaled(COORD &dest, const double s, const COORD v)
{
       dest.x += s * v.x; dest.y += s * v.y; dest.z += s * v.z;
}

double vector_dot_product(const COORD v1, const COORD v2)
{
       return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z;
}

void vector_cross_product(COORD &vz, const COORD vx, const COORD vy)
{
       vz.x = vx.y * vy.z - vx.z * vy.y;
       vz.y = vx.z * vy.x - vx.x * vy.z;
       vz.z = vx.x * vy.y - vx.y * vy.x;
}

void vector_normalize(COORD &v)
{
       register double ir;

       ir = 1.0 / sqrt (v.x * v.x + v.y * v.y + v.z * v.z);
       v.x *= ir; v.y *= ir; v.z *= ir;
}

void vector_middle(COORD &dest, const COORD v1, const COORD v2)
{
       dest.x = 0.5 * (v1.x + v2.x);
       dest.y = 0.5 * (v1.y + v2.y);
       dest.z = 0.5 * (v1.z + v2.z);
}

double vector_distance(const COORD v1, const COORD v2)
{
       double xdiff = v1.x - v2.x;
       double ydiff = v1.y - v2.y;
       double zdiff = v1.z - v2.z;
       return (sqrt(xdiff * xdiff + ydiff * ydiff + zdiff * zdiff));
}

double vector_length(const COORD v)
{
       return (sqrt(v.x * v.x + v.y * v.y + v.z * v.z));
}

double vector_angle(const COORD va, const COORD vb, const COORD vref)
{
       double ang_deg;
       COORD vc, va_cp, vb_cp, vref_cp;
       va_cp = va;
       vb_cp = vb;
       vref_cp = vref;
       vector_orthogonal(va_cp, vref_cp);
       vector_orthogonal(vb_cp, vref_cp);
       ang_deg = magang(va_cp, vb_cp);
       vector_cross_product(vc, va_cp, vb_cp);
       if (vector_dot_product(vc, vref_cp) < 0) ang_deg = -ang_deg;
       return ang_deg;
}

void vector_orthogonal(COORD &va, COORD &vref)
{
       double d;
       vector_normalize(vref);
       d = vector_dot_product(va, vref);
       vector_add_scaled(va, -d, vref);
       vector_normalize(va);
}

void matrix_init(double m[4][4])
{
       m[0][0] = 1.0;
       m[0][1] = 0.0;
       m[0][2] = 0.0;
       m[0][3] = 0.0;
       m[1][0] = 0.0;
       m[1][1] = 1.0;
       m[1][2] = 0.0;
       m[1][3] = 0.0;
       m[2][0] = 0.0;
       m[2][1] = 0.0;
       m[2][2] = 1.0;
       m[2][3] = 0.0;
       m[3][0] = 0.0;
       m[3][1] = 0.0;
       m[3][2] = 0.0;
       m[3][3] = 1.0;
}

void matrix_concatenate(double dest[4][4], const double src[4][4])
{
       double t[4][4];
     
       memcpy (t, dest, 16 * sizeof (double));
     
       for (int i = 0; i < 4; i++) {
            for (int j = 0; j < 4; j++) {
                 dest[i][j] = t[i][0] * src[0][j] +
           	              t[i][1] * src[1][j] +
           	              t[i][2] * src[2][j] +
           	              t[i][3] * src[3][j];
            }
       }
}

void matrix_x_rotation(double m[4][4], const double radians)
{
       matrix_init(m);
     
       double d = cos(radians);
       m[1][1] = d;
       m[2][2] = d;
       d = sin(radians);
       m[1][2] = d;
       m[2][1] = -d;
}

void matrix_y_rotation(double m[4][4], const double radians)
{
       matrix_init(m);
     
       double d = cos(radians);
       m[0][0] = d;
       m[2][2] = d;
       d = sin(radians);
       m[2][0] = d;
       m[0][2] = -d;
}

void matrix_z_rotation(double m[4][4], const double radians)
{
       matrix_init(m);
     
       double d = cos(radians);
       m[0][0] = d;
       m[1][1] = d;
       d = sin(radians);
       m[0][1] = d;
       m[1][0] = -d;
}

void matrix_translation(double m[4][4], const double x, const double y, const double z)
{
       matrix_init(m);
     
       m[3][0] = x;
       m[3][1] = y;
       m[3][2] = z;
}

void matrix_scale(double m[4][4], const double sx, const double sy, const double sz)
{
       matrix_init(m);
     
       m[0][0] = sx;
       m[1][1] = sy;
       m[2][2] = sz;
}

void matrix_orthographic_projection(double m[4][4], const double left,
       const double right, const double bottom, const double top,
       const double near, const double far)
{
       matrix_init(m);
     
       m[0][0] = 2.0 / (right - left);
       m[3][0] = - (right + left) / (right - left);
       m[1][1] = 2.0 / (top - bottom);
       m[3][1] = - (top + bottom) / (top - bottom);
       m[2][2] = -2.0 / (far - near);
       m[3][2] = - (far + near) / (far - near);
}

void matrix_transform(COORD &v, const double m[4][4])
{
       register double x, y, z, w;
     
       x = v.x;
       y = v.y;
       z = v.z;
       w = x * m[0][3] + y * m[1][3] + z * m[2][3] + m[3][3];
     
       v.x = (x * m[0][0] + y * m[1][0] + z * m[2][0] + m[3][0]) / w;
       v.y = (x * m[0][1] + y * m[1][1] + z * m[2][1] + m[3][1]) / w;
       v.z = (x * m[0][2] + y * m[1][2] + z * m[2][2] + m[3][2]) / w;
}

void matrix_rotate(COORD &v, const double m[4][4])
{
       register double x, y, z;
     
       x = v.x;
       y = v.y;
       z = v.z;
       v.x = x * m[0][0] + y * m[1][0] + z * m[2][0];
       v.y = x * m[0][1] + y * m[1][1] + z * m[2][1];
       v.z = x * m[0][2] + y * m[1][2] + z * m[2][2];
}

void matrix_rotate_translate(COORD &v, const double m[4][4])
{
       register double x, y, z;
     
       x = v.x;
       y = v.y;
       z = v.z;
       v.x = x * m[0][0] + y * m[1][0] + z * m[2][0] + m[3][0];
       v.y = x * m[0][1] + y * m[1][1] + z * m[2][1] + m[3][1];
       v.z = x * m[0][2] + y * m[1][2] + z * m[2][2] + m[3][2];
}

void superposition_matrix(const int lt, double *seqf1, double *seqf2, double *r,
            double *xm, double *ym)
{
       int    i, j, i1, l, iq, ip;
       double s0, s1, r1, r2, r3, r4, evmax, /* xm[3], ym[3], */ xn[3] ,yn[3];
       double c[3][3], a[4][4], s[4][4], g, v1, v2, v3, u, st, ct, s2, s3;
       double eps;

       eps = 1.0e-5;
       for (i = 0; i < 3; i++) {
	    xm[i] = 0.0;
	    ym[i] = 0.0;
	    for (j = 0; j < 3; j++) c[i][j] = 0.0;
       }

       for (i = 0; i < lt; i++)
	    for (l = 0; l < 3; l++) {
		 xm[l] = xm[l] + seqf1[3 * i + l] / (double) lt;
		 ym[l] = ym[l] + seqf2[3 * i + l] / (double) lt;
       }

       for (i1 = 0; i1 < lt; i1++) {
	    for (j = 0; j < 3; j++) {
		 xn[j] = seqf1[3 * i1 + j] - xm[j];
		 yn[j] = seqf2[3 * i1 + j] - ym[j];
	    }
	    for (i = 0; i < 3; i++)
		 for (j = 0; j < 3; j++)
		      c[i][j] = xn[i] * yn[j] + c[i][j];
       }

       a[0][0] = -c[0][0]+c[1][1]-c[2][2];
       a[0][1] = -c[0][1]-c[1][0];
       a[0][2] = -c[1][2]-c[2][1];
       a[0][3] = c[0][2]-c[2][0];
       a[1][1] = c[0][0]-c[1][1]-c[2][2];
       a[1][2] = c[0][2]+c[2][0];
       a[1][3] = c[1][2]-c[2][1];
       a[2][2] = -c[0][0]-c[1][1]+c[2][2];
       a[2][3] = c[0][1]-c[1][0];
       a[3][3] = c[0][0]+c[1][1]+c[2][2];
       for (i = 1; i < 4; i++)
	    for (j = 0; j < i;j++)
		 a[i][j] = a[j][i];
       for (i = 0; i < 4; i++)
	    for (j = 0; j < 4; j++)
	 	 s[i][j] = ((j == i)? 1.0:0.0);
	

       s0 = 0.0;
       for (i = 1; i < 4; i++)
	    for (j = 0; j < i; j++)
		 s0 = s0 + 2.0 * a[i][j] * a[i][j];

       s1 = sqrt(s0);
       s2 = eps / 4.0 * s1;
       s3 = s1;
       do {
	    s3 = s3 / 4.0;
	    do {
		 l = 0;
		 for (iq = 1; iq < 4; iq++) {
		      for (ip = 0; ip < iq; ip++) {
			   if (fabs(a[ip][iq]) < s3) continue;
			   l = 1;
			   v1 = a[ip][ip];
			   v2 = a[ip][iq];
			   v3 = a[iq][iq];
			   u = 0.5 * (v1 - v3);
			   if (fabs(u) >= 1.0e-10) 
                                g = -v2 / sqrt(v2 * v2 + u * u) * u / fabs(u);
                           else g = 1.0;
			   st = g / sqrt(2.0*(1.0+sqrt(1.0-g*g)));
			   ct = sqrt(1.0 - st * st);
			   for (i = 0; i < 4; i++) {
			        g = a[i][ip] * ct - a[i][iq] * st;
			        a[i][iq] = a[i][ip] * st + a[i][iq] * ct;
			        a[i][ip] = g;
			        g = s[i][ip] * ct - s[i][iq] * st;
			        s[i][iq] = s[i][ip] * st + s[i][iq] * ct;
			        s[i][ip] = g;
			   }
			   for (i = 0; i < 4; i++) {
			        a[ip][i] = a[i][ip];
			        a[iq][i] = a[i][iq];
			   }
			   g = 2.0 * v2 * st * ct;
			   a[ip][ip] = v1 * ct * ct + v3 * st * st - g;
			   a[iq][iq] = v1 * st * st + v3 * ct * ct + g;
			   a[ip][iq] = (v1 - v3) * st * ct + v2*(ct*ct-st*st);
			   a[iq][ip] = a[ip][iq];
		      }
		 }
	    } while(l == 1);
       } while(s3 > s2);

       l = 0;
       evmax = a[0][0];
       for (i = 1; i < 4; i++)
	    if (a[i][i] > evmax) {
	         l = i;
	         evmax = a[i][i];
	    }

       r1 = s[0][l];
       r2 = s[1][l];
       r3 = s[2][l];
       r4 = s[3][l];
       r[0] = -r1 * r1 + r2 * r2 - r3 * r3 + r4 * r4;
       r[1] = 2.0 * (r3 * r4 - r1 * r2);
       r[2] = 2.0 * (r2 * r3 + r1 * r4);
       r[3] = -2.0 * (r1 * r2 + r3 * r4);
       r[4] = r1 * r1 - r2 * r2 - r3 * r3 + r4 * r4;
       r[5] = 2.0 * (r2 * r4 - r1 * r3);
       r[6] = 2.0 * (r2 * r3 - r1 * r4);
       r[7] = -2.0 * (r1 * r3 + r2 * r4);
       r[8] = -r1 * r1 - r2 * r2 + r3 * r3 + r4 * r4;

       for (i = 0; i < 3; i++) {
	    s0 = 0.0;
	    for (j = 0; j < 3; j++)
		 s0 = r[j * 3 + i] * xm[j] + s0;
	    r[9 + i] = ym[i] - s0;
       }
}

