/*
FILE:     na_classification.C
*/
/*
VERSION:  1.700
*/
/*
DATE:     7/13/2005
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2005 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "_validation.h"
#include "erf_util.h"
#include "util.h"

typedef struct {
        char base;
        int  chn_idx;
        int  res_idx;
        char sym[10];
        int  n_base_pair;
        int  base_pair[4];
        int  wc_bspair[4];
        int  is_break;
} BASE_PAIR_INFO;

typedef struct {
        int  chain_index;
        char sym[10];
        int  count;
} CHAIN_COUNT;

typedef struct {
        char label;
        int  first;
        int  last;
        int  index;
} HELIX_INFO;

static char stchpa[27] = "abcdefghijklmnopqrstuvwxyz";
static char stchap[27] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";

static HELIX_INFO *na_sec_cls(const int n_pair_info, BASE_PAIR_INFO *pair_info,
              int &classification, int &n_helix_info);
static int  whstrd(int **strcd, const int stsht, const int finsht,
              int &nxstr, int &nxpl, const int lststd);
static int  strdpl(const int stpl, const int finpl, const int thcode, int **strand);
static void fdshus(const int stpt, char **code, int *shtcod, int &stsht,
              int &finsht, const int n_pair_info);
static int  fdshps(const int stpt, const int shtnum, char **code,
              int *shtcod, int **bridpt, const int n_pair_info);
static void setsht(const int refpl, const int shtnum, int *shtcod, char **code,
              const int n_pair_info);
static int  is_same_strand(const char *code_a, const char *code_b);
static int junction(int &count, const char label, const int fst,
              HELIX_INFO *helix_info, const int n_helix_info);
static int is_tetraloop(const char a1, const char a2, const char a3, const char a4);
#ifdef SUN_OS
#ifdef __cplusplus
extern "C" {
#endif
#endif
static int chain_compare(const void *chain1, const void *chain2);
#ifdef SUN_OS
#ifdef __cplusplus
}
#endif
#endif
static int non_watson_crick_pair(const char a, const char b);

void _VALIDATION::na_classification()
{
       int m = 0, i, j, k, count, n_count1, n_count2, fst_index, n_pair_info;
       int pos1, pos2, pos11 = 0, pos22 = 0, chain_index1 = 0, chain_index2 = 0, t, t1;
       int n_helix_info = 0, irow = 0, found = 0;
       char base[4];
       CifString index, entry_id, feature;
       TblIndexObj _BaseIndex;
       _BSPAIR *base_info = NULL;
       CHAIN_COUNT count1[10], count2[10];
       _CHAIN *chain[10];
       BASE_PAIR_INFO *pair_info = NULL;
       HELIX_INFO *helix_info = NULL;
       ISTable *Table = NULL;

       if (!_n_bspair) return;

       base_info = new _BSPAIR[_n_bspair];
       for (i = 0; i < _n_bspair; i++) {
            index.Copy(_bspair[i]->Asym_or_Gen);
            index += "_";
            index += _bspair[i]->FstIndex;
            index += "_";
            index += _bspair[i]->FstRes;
            index += "_";
            index += _bspair[i]->SndIndex;
            index += "_";
            index += _bspair[i]->SndRes;
            if (_BaseIndex.Seek(index) < 0) {
                 base_info[m].FstIndex = _bspair[i]->FstIndex;
                 base_info[m].FstRes   = _bspair[i]->FstRes;
                 base_info[m].SndIndex = _bspair[i]->SndIndex;
                 base_info[m].SndRes   = _bspair[i]->SndRes;
                 strcpy(base_info[m].Asym_or_Gen, _bspair[i]->Asym_or_Gen);
                 strcpy(base_info[m].FstAtom, "");
                 m++;
                 _BaseIndex.Add(index);
            }
       }
       if (m == 0) {
            delete [] base_info;
            return;
       }

       _MOLECULE *mol = _molecules[0];
       while (1) {
            n_count1 = 0;
            n_count2 = 0;
            fst_index = -1;
            for (i = 0; i < m; i++) {
                 if (!strcmp(base_info[i].FstAtom, "used")) continue;
                 if (base_info[i].FstIndex != fst_index && fst_index == -1) {
                      strcpy(base_info[i].FstAtom, "proc");
                      fst_index = base_info[i].FstIndex;
                      for (j = 0; j < n_count1; j++)
                           if (count1[j].chain_index == fst_index) break;
                      if (j == n_count1) {
                           n_count1++;
                           count1[j].chain_index = fst_index;
                           count1[j].sym[0] = '\0';
                      }
                      for (j = 0; j < n_count1; j++)
                           if (count1[j].chain_index == base_info[i].SndIndex &&
                               !strcmp(count1[j].sym, base_info[i].Asym_or_Gen))
                                break;
                      if (j == n_count1) {
                           n_count1++;
                           count1[j].chain_index = base_info[i].SndIndex;
                           strcpy(count1[j].sym, base_info[i].Asym_or_Gen);
                      }
                      for (j = 0; j < n_count2; j++)
                           if (count2[j].chain_index == base_info[i].SndIndex &&
                               !strcmp(count2[j].sym, base_info[i].Asym_or_Gen))
                                break;
                      if (j == n_count2) {
                           n_count2++;
                           count2[j].chain_index = base_info[i].SndIndex;
                           strcpy(count2[j].sym, base_info[i].Asym_or_Gen);
                           count2[j].count = 0;
                      }
                      count2[j].count++;
                 } else if (base_info[i].FstIndex == fst_index) {
                      strcpy(base_info[i].FstAtom, "proc");
                      for (j = 0; j < n_count1; j++)
                           if (count1[j].chain_index == base_info[i].SndIndex &&
                               !strcmp(count1[j].sym, base_info[i].Asym_or_Gen))
                                break;
                      if (j == n_count1) {
                           n_count1++;
                           count1[j].chain_index = base_info[i].SndIndex;
                           strcpy(count1[j].sym, base_info[i].Asym_or_Gen);
                      }
                      for (j = 0; j < n_count2; j++)
                           if (count2[j].chain_index == base_info[i].SndIndex &&
                               !strcmp(count2[j].sym, base_info[i].Asym_or_Gen))
                                break;
                      if (j == n_count2) {
                           n_count2++;
                           count2[j].chain_index = base_info[i].SndIndex;
                           strcpy(count2[j].sym, base_info[i].Asym_or_Gen);
                           count2[j].count = 0;
                      }
                      count2[j].count++;
                 }
            }

            if (n_count2 > 1) {
                 count = 0;
                 for (j = 0; j < n_count2; j++)
                      count += count2[j].count;
                 for (j = 0; j < n_count2; j++) {
                      if (((float) count2[j].count / (float) count) < 0.1)
                           count2[j].count = 0;

                 }
            }
            for (i = 0; i < m; i++) {
                 if (!strcmp(base_info[i].FstAtom, "used")) continue;
                 if (!strcmp(base_info[i].FstAtom, "proc")) {
                      for (j = 0; j < n_count2; j++) {
                           if (count2[j].chain_index == base_info[i].SndIndex &&
                               !strcmp(count2[j].sym, base_info[i].Asym_or_Gen)) {
                                if (!count2[j].count)
                                     strcpy(base_info[i].FstAtom, "used");
                                break;
                           }
                      }
                 } else {
                      for (j = 0; j < n_count2; j++) {
                           if (!count2[j].count) continue;
                           if (base_info[i].FstIndex == count2[j].chain_index ||
                               base_info[i].SndIndex == count2[j].chain_index &&
                               !strcmp(base_info[i].Asym_or_Gen, "")) {
                                strcpy(base_info[i].FstAtom, "proc");
                                for (k = 0; k < n_count1; k++)
                                     if (count1[k].chain_index == base_info[i].FstIndex) break;
                                if (k == n_count1) {
                                     n_count1++;
                                     count1[k].chain_index = base_info[i].FstIndex;
                                     count1[k].sym[0] = '\0';
                                }
                                for (k = 0; k < n_count1; k++)
                                     if (count1[k].chain_index == base_info[i].SndIndex &&
                                         !strcmp(count1[k].sym, base_info[i].Asym_or_Gen))
                                if (k == n_count1) {
                                     n_count1++;
                                     count1[k].chain_index = base_info[i].SndIndex;
                                     strcpy(count1[k].sym, base_info[i].Asym_or_Gen);
                                }
                           }
                      }
                 }
            }
            if (n_count1 > 1) qsort((void *) count1, n_count1, sizeof(CHAIN_COUNT),
                                     chain_compare);

            n_pair_info = 0;
            for (i = 0; i < n_count1; i++) {
                 chain[i] = mol->GetIndexChain(count1[i].chain_index);
                 count1[i].count = n_pair_info;
                 n_pair_info += chain[i]->TotalResidueNumbers();
            }
            if (!n_pair_info) break;
            pair_info = new BASE_PAIR_INFO[n_pair_info];
            for (i = 0; i < n_pair_info; i++) {
                 pair_info[i].chn_idx = -1;
                 pair_info[i].res_idx = -1;
                 pair_info[i].sym[0] = '\0';
                 pair_info[i].n_base_pair = 0;
                 pair_info[i].is_break = 0;
            }
            for (i = 0; i < n_count1; i++) {
                 if (i && count1[i].count < n_pair_info)
                      pair_info[count1[i].count].is_break = 1;
                 for (j = 0; j < chain[i]->SeqLen(); j++) {
                      if (chain[i]->SeqRes(j)->Deleted) continue;
                      strcpy(base, chain[i]->SeqRes(j)->Field[0]);
                      get_parent_name(base);
                      if (base[0] == '+')
                           pair_info[count1[i].count+j].base = base[1];
                      else pair_info[count1[i].count+j].base = base[0];
                 }
            }
            found = 0;
            for (i = 0; i < m; i++) {
                 if (!strcmp(base_info[i].FstAtom, "used")) continue;
                 if (!strcmp(base_info[i].FstAtom, "proc")) {
                      strcpy(base_info[i].FstAtom, "used");
                      pos1 = -1;
                      pos2 = -1;
                      for (j = 0; j < n_count1; j++) {
                           if (count1[j].chain_index == base_info[i].FstIndex &&
                               !strcmp(count1[j].sym, "")) {
                                pos1 = chain[j]->SeqResPos(base_info[i].FstRes);
                                chain_index1 = mol->ChainIndex(count1[j].chain_index);
                                pos11 = pos1 + count1[j].count;
                           }
                           if (count1[j].chain_index == base_info[i].SndIndex &&
                               !strcmp(count1[j].sym, base_info[i].Asym_or_Gen)) {
                                pos2 = chain[j]->SeqResPos(base_info[i].SndRes);
                                chain_index2 = mol->ChainIndex(count1[j].chain_index);
                                pos22 = pos2 + count1[j].count;
                           }
                      }
                      if (pos1 >= 0 && pos2 >= 0) {
                           found = 1;
                           pair_info[pos11].chn_idx = chain_index1;
                           pair_info[pos11].res_idx = pos1;
                           pair_info[pos22].chn_idx = chain_index2;
                           pair_info[pos22].res_idx = pos2;
                           pair_info[pos11].base_pair[pair_info[pos11].n_base_pair] = pos22;
                           pair_info[pos22].base_pair[pair_info[pos22].n_base_pair] = pos11;
                           strcpy(pair_info[pos22].sym, base_info[i].Asym_or_Gen);
                           if (!strcmp(base_info[i].Type, "WATSON-CRICK") ||
                               !strcmp(base_info[i].Type, "A-T PAIR") ||
                               !strcmp(base_info[i].Type, "T-A PAIR") ||
                               !strcmp(base_info[i].Type, "G-C PAIR") ||
                               !strcmp(base_info[i].Type, "C-G PAIR")) {
                                pair_info[pos11].wc_bspair[pair_info[pos11].n_base_pair] = 1;
                                pair_info[pos22].wc_bspair[pair_info[pos22].n_base_pair] = 1;
                           } else {
                                pair_info[pos11].wc_bspair[pair_info[pos11].n_base_pair] = 0;
                                pair_info[pos22].wc_bspair[pair_info[pos22].n_base_pair] = 0;
                           }
                           pair_info[pos11].n_base_pair++;
                           pair_info[pos22].n_base_pair++;
                      }
                 }
            }
            n_helix_info = 0;
            helix_info = na_sec_cls(n_pair_info, pair_info,
                          _classification, n_helix_info);
            if (n_helix_info) {
                 for (i = 0; i < n_helix_info; i++) {
                      if (helix_info[i].label == ' ') continue;
                      for (j = i + 1; j < n_helix_info; j++) {
                           if (helix_info[j].label != helix_info[i].label) continue;
                           helix_info[i].label = ' ';
                           helix_info[j].label = ' ';
                           for (k = 0; k < helix_info[i].last - helix_info[i].first + 1; k++) {
                                t = helix_info[i].first + k;
                                t1 = pair_info[t].base_pair[0];
                                if (non_watson_crick_pair(pair_info[t].base, pair_info[t1].base))
                                     _classification |= MISMATCH;
                           }
                           break;
                      }
                 }
            }
            if (helix_info) delete [] helix_info;
            delete [] pair_info;
            if (!found) break;
       }

       if (_CifObj) {
            Table = _CifObj->GetTablePtr(_StructureId, "ndb_struct_conf_na");
            if (is_empty_table(Table)) {
                 if (_CifObj->IsTablePresent(_StructureId, "ndb_struct_conf_na"))
                      _CifObj->DeleteTable(_StructureId, "ndb_struct_conf_na");

                 Table = new ISTable("ndb_struct_conf_na");
                 Table->AddColumn("entry_id");
                 Table->AddColumn("feature");
                 _CifObj->WriteTable(Table, _StructureId, "ndb_struct_conf_na");
                 entry_id.Copy(_StructureId);
                 irow = 0;
                 if (_classification & DOUBLE_HELIX) {
                      feature.Copy("double helix");
                      Table->AddRow();
                      Table->UpdateCell(entry_id, 0, irow);
                      Table->UpdateCell(feature, 1, irow);
                      irow++;
                 }
                 if (_classification & A_DOUBLE_HELIX) {
                      feature.Copy("a-form double helix");
                      Table->AddRow();
                      Table->UpdateCell(entry_id, 0, irow);
                      Table->UpdateCell(feature, 1, irow);
                      irow++;
                 }
                 if (_classification & B_DOUBLE_HELIX) {
                      feature.Copy("b-form double helix");
                      Table->AddRow();
                      Table->UpdateCell(entry_id, 0, irow);
                      Table->UpdateCell(feature, 1, irow);
                      irow++;
                 }
                 if (_classification & Z_DOUBLE_HELIX) {
                      feature.Copy("z-form double helix");
                      Table->AddRow();
                      Table->UpdateCell(entry_id, 0, irow);
                      Table->UpdateCell(feature, 1, irow);
                      irow++;
                 }
                 if (_classification & PARALLEL_HELIX) {
                      feature.Copy("parallel strands");
                      Table->AddRow();
                      Table->UpdateCell(entry_id, 0, irow);
                      Table->UpdateCell(feature, 1, irow);
                      irow++;
                 }
                 if (_classification & HAIRPIN_LOOP) {
                      feature.Copy("hairpin loop");
                      Table->AddRow();
                      Table->UpdateCell(entry_id, 0, irow);
                      Table->UpdateCell(feature, 1, irow);
                      irow++;
                 }
                 if (_classification & TETRA_LOOP) {
                      feature.Copy("tetraloop");
                      Table->AddRow();
                      Table->UpdateCell(entry_id, 0, irow);
                      Table->UpdateCell(feature, 1, irow);
                      irow++;
                 }
                 if (_classification & BULGE_LOOP) {
                      feature.Copy("bulge loop");
                      Table->AddRow();
                      Table->UpdateCell(entry_id, 0, irow);
                      Table->UpdateCell(feature, 1, irow);
                      irow++;
                 }
                 if (_classification & MISMATCH) {
                      feature.Copy("mismatched base pair");
                      Table->AddRow();
                      Table->UpdateCell(entry_id, 0, irow);
                      Table->UpdateCell(feature, 1, irow);
                      irow++;
                 }
                 if (_classification & INTERNAL_LOOP) {
                      feature.Copy("internal loop");
                      Table->AddRow();
                      Table->UpdateCell(entry_id, 0, irow);
                      Table->UpdateCell(feature, 1, irow);
                      irow++;
                 }
                 if (_classification & TRIPLE_HELIX) {
                      feature.Copy("triple helix");
                      Table->AddRow();
                      Table->UpdateCell(entry_id, 0, irow);
                      Table->UpdateCell(feature, 1, irow);
                      irow++;
                 }
                 if (_classification & QUADRUPLE_HELIX) {
                      feature.Copy("quadruple helix");
                      Table->AddRow();
                      Table->UpdateCell(entry_id, 0, irow);
                      Table->UpdateCell(feature, 1, irow);
                      irow++;
                 }
                 if (_classification & THREE_STEM) {
                      feature.Copy("three-way junction");
                      Table->AddRow();
                      Table->UpdateCell(entry_id, 0, irow);
                      Table->UpdateCell(feature, 1, irow);
                      irow++;
                 }
                 if (_classification & FOUR_STEM) {
                      feature.Copy("four-way junction");
                      Table->AddRow();
                      Table->UpdateCell(entry_id, 0, irow);
                      Table->UpdateCell(feature, 1, irow);
                      irow++;
                 }
            }
       }
       
       delete [] base_info;
}

static HELIX_INFO *na_sec_cls(const int n_pair_info, BASE_PAIR_INFO *pair_info,
              int &classification, int &n_helix_info)
{
       int i, nxstrn, j, k, l, m, j1, dir, nxstpl, strpl, shtnum, stsht, finsht;
       int lststd, bstval, chcode, stchpl, nxstr = 0, nxpl = 0, thstpl, num_strand, len;
       int count, first, last, last_j, ij, diff_i, diff_j, start, end, index;
       int *shtcod = NULL, **strcd = NULL, **bridpt = NULL;
       char **code = NULL, strdch, label, **helix = NULL;
       HELIX_INFO *helix_info = NULL;

       n_helix_info = 0;
       for (i = 0; i < n_pair_info; i++)
            if (pair_info[i].n_base_pair) break;
       if (i == n_pair_info) return NULL;

       shtcod = new int[n_pair_info];
       strcd = new int*[n_pair_info];
       code = new char*[n_pair_info];
       bridpt = new int*[n_pair_info];
       for (i = 0; i < n_pair_info; i++) {
            shtcod[i] = 0;
            strcd[i] = new int[4];
            for (j = 0; j < 4; j++) strcd[i][j] = 0;
            code[i] = new char[7];
            strcpy(code[i], "      ");
            bridpt[i] = new int[4];
            for (j = 0; j < 4; j++) bridpt[i][j] = -1;
       }

       nxstrn = 0;
       for (i = 0; i < n_pair_info - 1; i++) {
            for (k = 0; k < pair_info[i].n_base_pair; k++) {
                 j = pair_info[i].base_pair[k];
                 if (j < i) continue;
                 if (pair_info[i+1].is_break) continue;

                 for (l = 0; l < pair_info[i+1].n_base_pair; l++) {
                      j1 = pair_info[i+1].base_pair[l];
                      dir = 0;
                      if (j1 == (j+1)) dir = 1;
                      else if (j1 == (j-1)) dir = -1;
                      if (dir) {
                           if (!strcd[i][k]) {
                                nxstrn++;
                                strcd[i][k] = nxstrn * dir;
                           }
                           strcd[i+1][l] = strcd[i][k];
                           for (m = 0; m < pair_info[j].n_base_pair; m++) {
                                if (pair_info[j].base_pair[m] == i) {
                                     strcd[j][m] = strcd[i][k];
                                     break;
                                }
                           }
                           for (m = 0; m < pair_info[j1].n_base_pair; m++) {
                                if (pair_info[j1].base_pair[m] == (i+1)) {
                                     strcd[j1][m] = strcd[i+1][l];
                                     break;
                                }
                           }
                           code[i][0] = 'H';
                           code[i+1][0] = 'H';
                           code[j][0] = 'H';
                           code[j1][0] = 'H';
                      }
                 }
            }
       }

       i = 0;
       while (i < n_pair_info) {
            while (i < n_pair_info && code[i][0] != 'H') i++;
            if (i >= n_pair_info) break;
            stsht = i;
            i++;
            while (i < n_pair_info && code[i][0] == 'H') i++;
            finsht = i;

            lststd = 0;
            while (1) {
                 bstval = whstrd(strcd, stsht, finsht, nxstr, nxpl, lststd);
                 if (!bstval) break;
                 lststd = bstval;
                 chcode = lststd % 26;
                 if (chcode == 0) chcode = 26;
                 strdch = stchpa[chcode-1];
                 if (strcd[nxstr][nxpl] < 0)
                      strdch = stchap[chcode-1];
                 stchpl = 1;
                 while (stchpl < 4 && code[nxstr][stchpl] != ' ') stchpl++;
                 if (stchpl == 1) {
                      thstpl = nxstr;
                      while (1) {
                           nxstpl = strdpl(thstpl, finsht, strcd[nxstr][nxpl], strcd);
                           if (nxstpl < 0) break;
                           while (stchpl < 4 && code[nxstpl][stchpl] != ' ') stchpl++;
                           thstpl = nxstpl;
                           if (stchpl > 1) break;
                      }
                 }
                 code[nxstr][stchpl] = strdch;
                 bridpt[nxstr][stchpl-1] = pair_info[nxstr].base_pair[nxpl];
                 thstpl = nxstr;
                 while (1) {
                      nxstpl = strdpl(thstpl, finsht, strcd[nxstr][nxpl], strcd);
                      if (nxstpl < 0) break;
                      code[nxstpl][stchpl] = strdch;
                      strpl = 0;
                      while (strpl < 4 && strcd[nxstr][nxpl] != strcd[nxstpl][strpl]) strpl++;
                      bridpt[nxstpl][stchpl-1] = pair_info[nxstpl].base_pair[strpl];
                      thstpl = nxstpl;
                 }
            }
       }

       shtnum = 0;
       i = 0;
       while (i < n_pair_info) {
            fdshus(i, code, shtcod, stsht, finsht, n_pair_info);
            if (stsht < 0) break;
            shtnum++;
            for (j = stsht; j < finsht; j++) {
                 shtcod[j] = shtnum;
                 for (k = 0; k < 4; k++) {
                      if (bridpt[j][k] >= 0 && !shtcod[bridpt[j][k]])
                           setsht(bridpt[j][k], shtnum, shtcod, code,
                               n_pair_info);
                 }
            }
            while (fdshps(finsht, shtnum, code, shtcod, bridpt, n_pair_info)) ;
            i = finsht;
       }

       if (shtnum) {
            helix = new char*[shtnum];
            for (i = 0; i < shtnum; i++) {
                 helix[i] = new char[50];
                 memset(helix[i], 0, 50);
                 for (j = 0; j < n_pair_info; j++) {
                      if (shtcod[j] != (i+1)) continue;
                      for (k = 1; k < 5; k++) {
                           if (code[j][k] == ' ') continue;
                           len = strlen(helix[i]);
                           for (l = 0; l < len; l++)
                                if (code[j][k] == helix[i][l]) break;
                           if (l == len)
                                helix[i][l] = code[j][k];
                      }
                 }
            }

            for (i = 0; i < shtnum; i++) {
                 if (strlen(helix[i]) == 3 || strlen(helix[i]) == 4)
                      classification |= QUADRUPLE_HELIX;
                 else if (strlen(helix[i]) == 2)
                      classification |= TRIPLE_HELIX;
                 else if (strlen(helix[i]) == 1 && helix[i][0] >= 'a' &&
                          helix[i][0] <= 'z')
                      classification |= PARALLEL_HELIX;
            }

            last = -1;
            label = ' ';
            num_strand = 0;
/*
printf("n_pair_info=%d\n",n_pair_info);
for(i=0;i<n_pair_info;i++) {
printf("i=%d base=%c chn_idx=%d res_idx=%d sym=%s n_base_pair=%d is_break=%d\n",
i,pair_info[i].base,pair_info[i].chn_idx,pair_info[i].res_idx,pair_info[i].sym,
pair_info[i].n_base_pair,pair_info[i].is_break);
for (j=0; j<pair_info[i].n_base_pair;j++)
printf("\t%d %d\n",pair_info[i].base_pair[j],pair_info[i].wc_bspair[j]);
}
*/
            for (i = 0; i < n_pair_info; i++) {
                 if (pair_info[i].is_break) {
                      last = -1; label = ' ';
                 }
                 if (code[i][1] == ' ') {
                      label = ' '; continue;
                 }
                 if (code[i][2] != ' ' || code[i][3] != ' ' ||
                     code[i][4] != ' ') continue;

                 if (code[i][1] != label) {
                      num_strand++;
                      if (last >= 0) {
                           last_j = pair_info[last].base_pair[0];
                           ij = pair_info[i].base_pair[0];
                           if ((last_j >= last || last_j >= i) && (ij >= last ||
                               ij >= i) && pair_info[last_j].chn_idx ==
                               pair_info[ij].chn_idx && !strcmp(pair_info[last_j].sym,
                               pair_info[ij].sym)) {
                                if (last_j == i && ij == last) {
                                     if ((i - last) == 3 && is_tetraloop(pair_info[i-3].base,
                                          pair_info[i-2].base, pair_info[i-1].base,
                                          pair_info[i].base))
                                          classification |= TETRA_LOOP;
                                     else classification |= HAIRPIN_LOOP;
                                } else {
                                     diff_i = i - last;
                                     diff_j = abs(last_j - ij);
                                     if (diff_i == 1 || diff_j == 1)
                                          classification |= BULGE_LOOP;
                                     else if (diff_i == 2 && diff_j == 2)
                                          classification |= MISMATCH;
                                     else {
                                          start = last_j;
                                          end = ij;
                                          if (start > end) {
                                               start = ij;
                                               end = last_j;
                                          }
                                          for (j = start + 1; j < end; j++)
                                               if (code[j][0] == 'H') break;
                                          if (j == end) classification |= INTERNAL_LOOP;
                                     }
                                }
                           }
                      }
                      label = code[i][1];
                      last = -1;
                 } else last = i;
            }

            if (num_strand) {
                 helix_info = new HELIX_INFO[num_strand];
                 n_helix_info = 0;
                 index = 1;
                 label = ' ';
                 first = -1;
                 last = -1;
                 for (i = 0; i < n_pair_info; i++) {
                      if (pair_info[i].is_break || code[i][1] == ' ') {
                           if (first >= 0 && last >= 0 &&
                              (last-first+1) >= 3) {
                                helix_info[n_helix_info].label = label;
                                helix_info[n_helix_info].first = first;
                                helix_info[n_helix_info].last  = last;
                                helix_info[n_helix_info].index = index;
                                n_helix_info++;
                           }
                           first = -1; last = -1; label = ' ';
                           if (pair_info[i].is_break) index++;
                           if (code[i][1] == ' ') continue;
                      }
                      if (code[i][2] != ' ' || code[i][3] != ' ' ||
                          code[i][4] != ' ') continue;
     
                      if (code[i][1] != label) {
                           if (first >= 0 && last >= 0 &&
                              (last-first+1) >= 3) {
                                helix_info[n_helix_info].label = label;
                                helix_info[n_helix_info].first = first;
                                helix_info[n_helix_info].last  = last;
                                helix_info[n_helix_info].index = index;
                                n_helix_info++;
                           }
                           first = last = i;
                           label = code[i][1];
                      } else last = i;
                 }
                 if (first >= 0 && last >= 0 && (last-first+1) >= 3) {
                      helix_info[n_helix_info].label = label;
                      helix_info[n_helix_info].first = first;
                      helix_info[n_helix_info].last  = last;
                      helix_info[n_helix_info].index = index;
                      n_helix_info++;
                 }
     
                 for (i = 0; i < n_helix_info - 1; i++) {
                      count = 0;
                      if (junction(count, helix_info[i].label, i,
                          helix_info, n_helix_info)) {
                           if (count == 6) classification |= THREE_STEM;
                           else if (count == 8) classification |= FOUR_STEM;
                      }
                 }
            }

            for (i = 0; i < shtnum; i++) delete [] helix[i];
            delete [] helix;
       }

       for (i = 0; i < n_pair_info; i++) {
            if (shtcod[i] <= 0) continue;
            code[i][5] = stchap[shtcod[i]-1];
       }

       delete [] shtcod;
       for (i = 0; i < n_pair_info; i++) {
            delete [] strcd[i];
            delete [] code[i];
            delete [] bridpt[i];
       }
       delete [] strcd;
       delete [] code;
       delete [] bridpt;

       return helix_info;
}

static int whstrd(int **strcd, const int stsht, const int finsht,
             int &nxstr, int &nxpl, const int lststd)
{
       int bstval = 0, i, j;
       for (i = stsht; i < finsht; i++) {
            for (j = 0; j < 4; j++) {
                 if (abs(strcd[i][j]) > lststd) {
                      if (bstval == 0) {
                           bstval = abs(strcd[i][j]);
                           nxstr = i;
                           nxpl = j;
                      } else if (abs(strcd[i][j]) < bstval) {
                           bstval = abs(strcd[i][j]);
                           nxstr = i;
                           nxpl = j;
                      }
                 }
            }
       }
       return bstval;
}

static int strdpl(const int stpl, const int finpl, const int thcode, int **strand)
{
       int nxpl = -1, i, j;
       for (i = stpl + 1; i < finpl; i++) {
            for (j = 0; j < 4; j++) {
                 if (strand[i][j] == thcode) {
                      nxpl = i;
                      break;
                 }
            }
            if (j < 4) break;
       }
       return nxpl;
}

static void fdshus(const int stpt, char **code, int *shtcod, int &stsht,
                int &finsht, const int n_pair_info)
{
       int nxres = stpt;
       stsht = -1;
       while (nxres < n_pair_info && (code[nxres][0] == ' ' ||
               shtcod[nxres] != 0)) nxres++;
       if (nxres >= n_pair_info) return;
       stsht = nxres;
       // while (nxres < n_pair_info && code[nxres][0] != ' ') nxres++;
       while (nxres < n_pair_info && is_same_strand(code[stsht], code[nxres]))
            nxres++;
       finsht = nxres;
}

static int fdshps(const int stpt, const int shtnum, char **code,
                int *shtcod, int **bridpt, const int n_pair_info)
{
       int nxres, stsht, finsht, j, k;
       nxres = stpt;
       while (nxres < n_pair_info && shtcod[nxres] != (-shtnum)) nxres++;
       if (nxres >= n_pair_info) return 0;
       stsht = nxres;
       while (nxres < n_pair_info && shtcod[nxres] == (-shtnum)) nxres++;
       finsht = nxres;
       for (j = stsht; j < finsht; j++) {
            shtcod[j] = shtnum;
            for (k = 0; k < 4; k++) {
                 if (bridpt[j][k] >= 0 && !shtcod[bridpt[j][k]])
                      setsht(bridpt[j][k], shtnum, shtcod, code, n_pair_info);
            }
       }
       return 1;
}

static void setsht(const int refpl, const int shtnum, int *shtcod, char **code,
                 const int n_pair_info)
{
       int nxres = refpl;
       // while (nxres < n_pair_info && code[nxres][0] != ' ') nxres++;
       while (nxres < n_pair_info && is_same_strand(code[refpl], code[nxres]))
            nxres++;
       int finsht = nxres;
       nxres = refpl;
       // while (nxres >= 0 && code[nxres][0] != ' ') nxres--;
       while (nxres >= 0 && is_same_strand(code[refpl], code[nxres]))
            nxres--;
       int stsht = nxres + 1;
       for (int i = stsht; i < finsht; i++) shtcod[i] = -shtnum;
}

static int is_same_strand(const char *code_a, const char *code_b)
{
       int is_same = 0;
       for (int i = 1; i < 5; i++) {
            if (code_a[i] != ' ' && code_a[i] == code_b[i]) {
                 is_same = 1;
                 break;
            }
       }
       return is_same;
}

static int junction(int &count, const char label, const int fst,
             HELIX_INFO *helix_info, const int n_helix_info)
{
       int i;
       if ((fst + 1) >= n_helix_info) return 0;
       if (helix_info[fst].index != helix_info[fst+1].index)
            return 0;
       count += 2;
       if (helix_info[fst+1].label == label) return 1;

       for (i = fst + 2; i < n_helix_info; i++) {
            if (helix_info[fst+1].label == helix_info[i].label)
                 break;
       }
       if (i == n_helix_info) return 0;
       return (junction(count, label, i, helix_info, n_helix_info));
}

static int is_tetraloop(const char a1, const char a2, const char a3, const char a4)
{
       int found = 0;
       if (a1 == 'G' && a4 == 'A' && (a3 == 'A' || a3 == 'G')) found = 1;
       if (a1 == 'U' && a4 == 'G' && a3 == 'C') found = 1;
       if (a1 == 'C' && a4 == 'G' && a2 == 'U') found = 1;
       return found;
}

#ifdef SUN_OS
#ifdef __cplusplus
extern "C" {
#endif
#endif
static int chain_compare(const void *chain1, const void *chain2)
{
       return (((CHAIN_COUNT *) chain1)->chain_index -
               ((CHAIN_COUNT *) chain2)->chain_index);
}
#ifdef SUN_OS
#ifdef __cplusplus
}
#endif
#endif

static int non_watson_crick_pair(const char a, const char b)
{
       if (a == 'A' && (b == 'T' || b == 'U')) return 0;
       if ((a == 'T' || a == 'U') && b == 'A') return 0;
       if (a == 'G' && b == 'C' || a == 'C' && b == 'G') return 0;
       return 1;
}
