/*
FILE:     ndb_atlas_util.C
*/
/*
VERSION:  1.700
*/
/*
DATE:     7/13/2005
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2005 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/

/* ----------------------------------------------------------------
   Original by :  Lester Clowney
   Purpose: The functions in this file are used for preparing an atlas of NDB 
   structures. Two different versions may be produced, HTML for online viewing
   and LaTeX for a printed copy.
   ----------------------------------------------------------------
*/

/*         
 *  Small functions for building an atlas
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "ndb_misclib.h"
#include "ndb_atlas.h"

void html_atlas_preamble(FILE *fp, const char *title)
{
    fprintf(fp, "<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">\n");
    fprintf(fp, "<HTML>\n");
    fprintf(fp, "<HEAD>\n");
    html_title(fp, title);
    fprintf(fp, "</HEAD>\n");
    fprintf(fp, "<BODY BGCOLOR=\"#FFFFFF\">\n");	
}

void html_atlas_trailer(FILE *fp)
{
    fprintf(fp, "</BODY>\n");
    fprintf(fp, "</HTML>\n");

}

void html_copyright(FILE *fp) 
{
  html_line(fp, "<pre>\n\n\n</pre>\n");
  html_line(fp, "<hr><p align=center><i>\n");
  html_line(fp, "&copy;1995,1996,1997 The Nucleic Acid Database Project<br>\n");
  html_line(fp, "Rutgers, The State University of New Jersey\n");
  html_line(fp, "</i>\n");
}

void html_title(FILE* fp, const char *st)
{
    fprintf(fp,"<title> %s </title>\n", st);
}

void html_line(FILE *fp, const char *st)
{ 
  if (st) fprintf(fp, "%s",st);
}

void html_par(FILE *fp)
{
  fprintf(fp, "<p>\n");
}

void html_begin(FILE *fp, const char *st)
{
  fprintf(fp, "<%s>\n",st);
}

void html_end(FILE *fp, const char *st)
{
  fprintf(fp, "</%s>\n",st);
}

void html_head(FILE* fp, const int level, const char *st)
{
  int tmp;
  if (level < 0) 
      tmp = -1 * level;
  else 
      tmp = level;
  fprintf(fp,"<h%d> %s </h%d>\n", tmp, st, tmp);
}

void html_bold(FILE* fp, const char *st)
{
  fprintf(fp,"<b>%s</b>", st);
}

void html_it(FILE* fp, const char *st)
{
  fprintf(fp,"<i>%s</i>\n", st);
}

void atlas_file(FILE *fp, const char *s1, const char *fname, const char *s3)
{
   fprintf(fp, "%s <a href=\"%s\">%s</a>", s1,fname,s3);
}

void atlas_title(FILE* fp, const char *st)
{
   html_head(fp, 1, st);
}

void atlas_head(FILE* fp, const int level, const char *st)
{
   html_head(fp, level, st);
}

void atlas_line(FILE *fp, const char *st)
{ 
  if (st) fprintf(fp, "%s",st);
}

void atlas_par(FILE *fp)
{
   html_par(fp);
}

void atlas_begin(FILE *fp, const char *st)
{	
  if (!strcmp(st,"itemize")) {
    html_begin(fp, "ul");
  } else {
    html_begin(fp,st);
  }
}

void atlas_end(FILE *fp, const char *st)
{	
  if (!strncmp(st,"itemize", 7)) {
    html_end(fp, "ul");
  } else {
    html_end(fp,st);
  }
}

void atlas_item(FILE *fp, const char *itst, const char *item) 
{
  fprintf(fp, "<li> %s %s\n", itst ? itst : "", item);
}

void atlas_bold(FILE* fp, const char *st)
{
  html_bold(fp,st);
}

void atlas_it(FILE* fp, const char *st)
{
  html_it(fp,st);
}

void atlas_line_break(FILE *fp) 
{
  fprintf(fp, "<br>");
}

void detex(char *st)
/**************************************************************************
 * Underscore characters are often used in the compound name strings. If 
 * an '_' is found and it is not in math mode then escape it. This is 
 * potentially quite inefficient and should be rewritten.  lsc
 ***************************************************************************/
{
  char *st_pt = NULL;
  char *buff;
  int i = 0;
  buff = (char*)malloc(sizeof(char) * 2 * strlen(st));
  st_pt = st;
  while (*st_pt != '\0') {
    if (*st_pt == '_') {
      if (st_pt == st) {
	strncpy(buff + i, "\\_", 2);
	i++;
      }
      else if (*(st_pt - 1) == '$') 
	buff[i] = '_';
      else {
	strncpy(buff + i, "\\_", 2);
	i++;
      }
    }
    else buff[i] = *st_pt;
    i++;
    st_pt++;
  }
  buff[i] = '\0';
  strcpy(st, buff);
  free(buff);
}

/* Destructively change stringcase. */

char *lowercased(char *st)
{
  char *st1 = st;  
  if (st == NULL) return st;  
  while (*st != '\0') {
    *st = (char)tolower((int)*st);
    st++;
  }
  return st1; 
}

char *uppercased(char *st)
{
  char *st1 = st;  
  if (st == NULL) return st;  
  while (*st != '\0') {
    *st = (char)toupper((int)*st);
    st++;
  }
  return st1; 
}

int get_ncommas(const char *stpt)
{
  int len, j, commas = 0; 
  len = strlen(stpt); 
  for (j = 0; j < len; j++)
    if (stpt[j] == ',') commas++;
  return commas;
}

int repl_subst_one(char *st, const char *newword, const char *old)
{ 
/**************************************************************************
 * 
 * Replace the first occurrence of <old> with <newword>.
 * Return a 1 if substitution made, else 0.
 * 
 ***************************************************************************/ 
    unsigned int len_old;
    char *st_pt = NULL;
    char *buff;
    int hit = 0;
    st_pt = st;
    len_old = strlen(old);
    buff = (char*)malloc(sizeof(char) * (strlen(st) + strlen(newword)));
    while (*st_pt != '\0') {
	if (!strncmp(st_pt, old, len_old)) {
	    hit++;
	    if (strlen(st_pt) > len_old)
		strcpy(buff, st_pt + len_old);
	    else buff[0] = '\0';
	    strcpy(st_pt, newword);
	    strcat(st, buff);
	    if (strlen(st) >= BIGSTRING) {
		ndb_log_message(NDB_MSG_DEBUG4, "Error: string %s too short in repl_subst_one.\n", st);
		exit(1);
	    }
	    goto finish;
	}
	st_pt++;
    }
  finish:
    free(buff);
    return hit;
}

int repl_subst(char st[BIGSTRING], const char *newword, const char *old)
/*
 * 
 * Replace substring <old> with substring <newword>.
 * Return the number of substitutions made.
 * 
 */
{ 
    unsigned int len_old;
    char *st_pt = NULL;
    char *buff;
    int hit = 0;
    st_pt = st;
    len_old = strlen(old);
    buff = (char*)malloc(sizeof(char) * (strlen(st) + strlen(newword)));
    while (*st_pt != '\0') {
	if (!strncmp(st_pt, old, len_old)) {
	    hit++;
	    if (strlen(st_pt) > len_old)
		strcpy(buff, st_pt + len_old);
	    else buff[0] = '\0';
	    strcpy(st_pt, newword);
	    strcat(st, buff);
	    if (strlen(st) >= BIGSTRING) {
		ndb_log_message(NDB_MSG_DEBUG4, "Error: string %s too long in repl_subst.\n", st);
		exit(1);
	    }
	    st_pt += strlen(newword);
	}
	else 
	    st_pt++;
    }
    free(buff);
    return hit;
}

int repl_subst_last(char st[BIGSTRING], const char *newword, const char *old)
/*
 * 
 * Replace substring <old> with substring <newword> if it is in the last postion.
 * Return 1 if substitution made.
 * 
 */
{ 
    unsigned  int len_old, ilen;
    char *st_pt = NULL;
    char *buff = NULL;
    int hit = 0;
    buff = NULL;
    st_pt = st;
    len_old = strlen(old);
    ilen = strlen(st) + strlen(newword) + 1;
    buff = (char*)malloc(sizeof(char) * ilen);
    memset(buff, 0, ilen);
    while (*st_pt != '\0') {
	if ((!strncmp(st_pt, old, len_old)) && (strlen(st_pt) == strlen(old))) {
	    hit++;
	    if (strlen(st_pt) > len_old)
		strcpy(buff, st_pt + len_old);
	    else buff[0] = '\0';
	    strcpy(st_pt, newword);
	    strcat(st, buff);
	    if (strlen(st) >= BIGSTRING) {
		ndb_log_message(NDB_MSG_DEBUG4, "Error: string %s too long in repl_subst_last.\n", st);
		exit(1);
	    }
	    goto finish;
	}
	else 
	    st_pt++;
    }
  finish:
    free(buff);
    return hit;
}

int repl_ext(char st[BIGSTRING], const char *newword, const char *old)
/*
 * 
 * Replace the extension of file <st> (<old>) with extension <newword>.
 * 
 */
{ 
  return repl_subst_last(st, newword, old);
}

char *gif_to_ps(const char *fname, const int ppi) 
/*
 * 
 * Convert a gif-format to a Postscript-format file.
 * 
 */
{
    char buff[160] = "fromgif ";
    static char new_fname[160];
    char new_fname_Z[160];
    FILE *fp;

    strcpy(new_fname, fname);
    repl_ext(new_fname, ".ps", ".gif");
    /* If the ps file exists then just use it. */
    if ((fp = fopen(new_fname, "r"))){
	ndb_log_message(NDB_MSG_DEBUG4,"Using preexisting Postscript file %s.\n", new_fname);
	fclose(fp);
    }
    else {
	strcpy(new_fname_Z, new_fname);
	strcat(new_fname_Z, ".Z");
	if ((fp = fopen(new_fname_Z, "r"))){
	    ndb_log_message(NDB_MSG_DEBUG4,
			    "Using preexisting compressed Postscript file %s.\n", new_fname_Z);
	    fclose(fp);
	}
	else {
	    /* Convert the gif file to a Postscript file */ 
	    strcat(buff, fname);
	    strcat(buff, " tmp.rgb");
	    system(buff);
	    strcpy(buff, "tops tmp.rgb  -rgb -eps ");
	    strcat(buff, " > ");
	    strcat(buff, new_fname);
	    system(buff);
	}
    }
    return new_fname;
}

int nullst_or_comment(char *st)
{
  while (isspace(*st) && (*st != '\0')) {
    st++;
  }
  return((*st == '#')|| (*st == '\0'));
}

char *itoa(int i)
{
  static char buff[160];
  sprintf(buff, "%d", i);
  return buff;
}

int ndb_get_sequence_homogeneity(char *seqstr, int length)
/*
 *   Returns TRUE if any NA residue accounts for more than 50%
 *   of the sequence ... 
 */
{
    int	i, len, nc, ng, na, nt, ni, nu, halflength; 

    halflength = length/2;
    if (length <= 0 || halflength <= 0 ) return(0);
    lowercased(seqstr);

    nc = ng = na = nt = ni = nu = 0;
    len = strlen(seqstr);
    for (i = 0; i < len; i++) {
	if (seqstr[i] == 'c') nc++;
	if (seqstr[i] == 'g') ng++;
	if (seqstr[i] == 'a') na++;
	if (seqstr[i] == 't') nt++;
	if (seqstr[i] == 'i') ni++;
	if (seqstr[i] == 'u') nu++;
    }


    if (nc > halflength ||
	ng > halflength ||
	na > halflength ||
	nt > halflength ||
	ni > halflength ||
	nu > halflength) 
	return(1);
    else
	return(0);
}
