/*
FILE:     ndb_picture.C
*/
/*
VERSION:  1.700
*/
/*
DATE:     7/13/2005
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2005 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <iostream.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "_validation.h"
#include "erf_util.h"
#include "ndb_misclib.h"
#include "ndb_atlas.h"
#include "misc_util.h"
#include "ps_util.h"
#include "util.h"

#define MAX_LENGTH_A            0
#define MAX_LENGTH_B            1
#define MAX_LENGTH_C            2
#define MAX_PICTURE           100

#define CMD_PS2GIF     "ps2gifdir1"
#define CMD_PS2PDF     "ps2pdfdir1"

static int  num_pic = 0;
static char pic_legend[MAX_PICTURE][100], pic_postfix[MAX_PICTURE][20];

static void draw_crystal_axes(_plot &plot, const double xll, const double yll, const double xur,
             const double yur, COORD &abc, const char *label);
static void get_rotation_matrix(const int view, double m[4][4]);
static void get_rotation_matrix(_MOLECULE *mol, const NDBCELLCONST *cell,
              double m[4][4], const int packopt, const int n,
              _BSPAIR **base_pair, int &max_length, COORD *xyz);
static void register_to_html(const char *postfix, const char *legend);
static void append_to_html(const char *sid, const int public_flag);
static void add_color_to_chains(_MOLECULE *mol);
static void select_biol_unit_to_display(_MOLECULE *mol, const int biol_id);

void _VALIDATION::write_picture_part(_plot &plot, const char *File_ID, const char *sid,
                         const int separate)
{
       int num_biold = 0, is_fractional = 0, len, fst, snd, i, j, k, order;
       int op, lx, ly, lz, biol_id;
       char buffer[100], string[200];
       _CHAIN *chain = NULL;

       if (!_num_mol || !_molecules) return;
       if (!_molecules[0]->ClearNonAsymAndWaterChains()) return;

       num_pic = 0;
       _MOLECULE *mol = _molecules[0];
       if (!_cell->is_artifical) {
            for (i = 0; i < _n_stbiol; i++) {
                 if (!_stbiol[i]) continue;
                 if (_stbiol[i]->Biol_ID > num_biold) num_biold = _stbiol[i]->Biol_ID;
                 if (strcmp(_stbiol[i]->Symmetry, "1_555")) {
                      if (!is_fractional) {
                           mol->convert_between_orthogonal_and_fractional(_cell,
                                 NDB_TRANS_ORTHO_TO_FRAC);
                           is_fractional = 1;
                      }
                      order = mol->Num_Chain() + 1;
                      op = atoi(_stbiol[i]->Symmetry) - 1;
                      len = strlen(_stbiol[i]->Symmetry);
                      for (k = 0; k < len; k++)
                           if (_stbiol[i]->Symmetry[k] == '_') break;
                      lx = _stbiol[i]->Symmetry[k+1] - '5';
                      ly = _stbiol[i]->Symmetry[k+2] - '5';
                      lz = _stbiol[i]->Symmetry[k+3] - '5';
                      mol->symmetry_related_chain(order, order, _stbiol[i]->Biol_ID,
                           _stbiol[i]->Asym_ID.Text(), _stbiol[i]->Asym_ID.Text(),
                           _stbiol[i]->PDB_Strand_ID, _cell, op, lx, ly, lz, 0);
                 }
            }
       }
       if (is_fractional) {
            mol->convert_between_orthogonal_and_fractional(_cell, NDB_TRANS_FRAC_TO_ORTHO);
       } else if (num_biold < 2) num_biold = 0;

       for (i = 0; i < _n_bspair; i++) {
            fst = mol->ChainIndex(_bspair[i]->FstIndex);
            snd = mol->ChainIndex(_bspair[i]->SndIndex);
            strcpy(_bspair[i]->FstAtom, mol->Chains(fst)->ChainID());
            _bspair[i]->FstIndex = fst;
            if (strcmp(_bspair[i]->Asym_or_Gen, "")) {
                 strcpy(_bspair[i]->SndAtom, "");
                 op = atoi(_bspair[i]->Asym_or_Gen);
                 len = strlen(_bspair[i]->Asym_or_Gen);
                 for (j = 0; j < len; j++)
                      if (_bspair[i]->Asym_or_Gen[j] == '_') break;
                 lx = _bspair[i]->Asym_or_Gen[j+1] - '5';
                 ly = _bspair[i]->Asym_or_Gen[j+2] - '5';
                 lz = _bspair[i]->Asym_or_Gen[j+3] - '5';
                 chain = mol->GetFirstChain();
                 while (chain) {
                      if (!strcmp(chain->ChainID(), mol->Chains(snd)->ChainID()) &&
                           chain->op() == op && chain->lx() == lx &&
                           chain->ly() == ly && chain->lz() == lz) {
                           strcpy(_bspair[i]->SndAtom, chain->ChainID());
                           _bspair[i]->SndIndex = mol->ChainIndex(chain->index());
                           break;
                      }
                      chain = mol->GetNextChain();
                 }
            } else {
                 strcpy(_bspair[i]->SndAtom, mol->Chains(snd)->ChainID());
                 _bspair[i]->SndIndex = snd;
            }
       }

       chain = mol->GetFirstChain();
       while (chain) {
            chain->set_display(0);
            if (chain->op() == 1) chain->set_display(1);
            chain = mol->GetNextChain();
       }
       add_color_to_chains(mol);
       Clear();
       make_ribbon(mol, COLOR_BY_RESIDUE, 1);

       string[0] = '\0';
       if ((_experiment_type & EXPERIMENT_TYPE_NMR) ||
           (_experiment_type & EXPERIMENT_TYPE_MODEL))
            sprintf(string, "View of the Structure %s", sid);
       else sprintf(string, "View of the Asymmetric Unit for Structure %s", sid);
       if (separate) {
            sprintf(buffer, "%s-asym.ps", File_ID);
            _plot plot1(buffer, 1);
            draw_pictures(plot1, string, 0);
            plot1.close();
            if ((_experiment_type & EXPERIMENT_TYPE_NMR) ||
                (_experiment_type & EXPERIMENT_TYPE_MODEL))
                 strcpy(string, "One View of the Structure");
            else strcpy(string, "View of the Asymmetric Unit");
            register_to_html("-asym", string);
       } else {
            plot.newpage(1);
            draw_pictures(plot, string, 0);
       }

       biol_id = 1;
       while (num_biold >= biol_id) {
            select_biol_unit_to_display(mol, biol_id);
            add_color_to_chains(mol);
            Clear();
            make_ribbon(mol, COLOR_BY_RESIDUE, 1);
            string[0] = '\0';
            if (num_biold > 1) 
                 sprintf(string, "View of the Biological Unit %d for Structure %s",
                                  biol_id, sid);
            else sprintf(string, "View of the Biological Unit for Structure %s", sid);

            if (separate) {
                 sprintf(buffer, "%s-biol-%d.ps", File_ID, biol_id);
                 _plot plot1(buffer, 1);
                 draw_pictures(plot1, string, 0);
                 plot1.close();
                 if (num_biold > 1)
                      sprintf(string, "View of the Biological Unit %d", biol_id);
                 else strcpy(string, "View of the Biological Unit");
                 sprintf(buffer, "-biol-%d", biol_id);
                 register_to_html(buffer, string);
            } else {
                 plot.newpage(1);
                 draw_pictures(plot, string, 0);
            }
            biol_id++;
       }
       _molecules[0]->ClearNonAsymAndWaterChains();
}

void _VALIDATION::write_packing_picture_part(_plot &plot, const char *File_ID, const char *sid,
                  const int separate, const int public_flag)
{
       if (_experiment_type & EXPERIMENT_TYPE_MODEL) return;

       char buffer[100], string[200];

       if (strcmp(_cell->space_group, "P 1") && !_cell->is_artifical) {
            _molecules[0]->crystal_packing(_cell);
            select_biol_unit_to_display(_molecules[0], 0);
            Clear();
            make_ribbon(_molecules[0], COLOR_BY_CHAIN, 0);
            string[0] = '\0';
            sprintf(string, "Crystal Packing of Structure %s", sid);
            if (separate) {
                 sprintf(buffer, "%s-packing.ps", File_ID);
                 _plot plot1(buffer, 1);
                 draw_pictures(plot1, string, 1);
                 plot1.close();
                 strcpy(string, "View of Crystal Packing");
                 register_to_html("-packing", string);
            } else {
                 plot.newpage(1);
                 draw_pictures(plot, string, 1);
            }
            _molecules[0]->ClearNonAsymAndWaterChains();
       } else if ((_experiment_type & EXPERIMENT_TYPE_NMR) && (_num_mol > 1)) {
            Clear();
            for (int i = 0; i < _num_mol; i++) {
                 _CHAIN *chain = _molecules[i]->GetFirstChain();
                 while (chain) {
                      if (chain->chain_type() == ATOMN_TOKEN ||
                          chain->chain_type() == ATOMP_TOKEN)
                           chain->set_display(1);
                      else chain->set_display(0);
                      chain->set_icolor(i);
                      if (chain->icolor() > MAXMOLECULETYPE) {
                           int icolor = chain->icolor() - MAXMOLECULETYPE;
                           chain->set_icolor(icolor);
                      }
                      if (chain->chain_type() == ATOMP_TOKEN)
                           chain->set_icolor(-1);
                      chain = _molecules[i]->GetNextChain();
                 }
                 make_ribbon(_molecules[i], COLOR_BY_MOLECULE, 0);
            }

            if (separate) {
                 sprintf(buffer, "%s-ensemble.ps", File_ID);
                 _plot plot1(buffer, 1);
                 draw_nmr_pictures(plot1, sid);
                 plot1.close();
                 string[0] = '\0';
                 strcpy(string, "View of Ensemble");
                 register_to_html("-ensemble", string);
            } else {
                 plot.newpage(1);
                 draw_nmr_pictures(plot, sid);
            }
       }
       if (separate) append_to_html(File_ID, public_flag);
}

void _VALIDATION::draw_pictures(_plot &plot, const char *buffer, const int packopt)
{
       // int color_type = COLOR_BY_RESIDUE;
       // if (packopt) color_type = COLOR_BY_CHAIN;
       double x = (plot.page_xmin() + plot.page_xmax()) / 2.0;
       double y = plot.page_ymax() - 70;
       plot.print_string(buffer, FONT_HELVETICA, 20, x, y, CENTERX);
       plot.setcolor(0.0, 0.0, 0.0);
       plot.setline_width(0.2);

       double xll = 50;
       double yll = 100;
       double xur = 550;
       double yur = 600;
       double area[4], rot[4][4];
       area[0] = xll; area[1] = yll; area[2] = xur; area[3] = yur;
       COORD abc[3], base[8];
       int max_length = MAX_LENGTH_C, k;
       get_rotation_matrix(_molecules[0], _cell, rot, packopt, _n_bspair,
                _bspair, max_length, &abc[0]);

       int is_draw_crystal = FALSE;
       if (strcmp(_cell->space_group, "P 1") && !_cell->is_artifical) {
            for (k = 0; k < 3; k++) {
                 ndb_trans_coord_3by3(abc[k], _cell, NDB_TRANS_FRAC_TO_ORTHO);
                 matrix_rotate(abc[k], rot);
            }
            if (packopt) {
                 base[0].x = 0; base[0].y = 0; base[0].z = 0;
                 base[1].x = 1; base[1].y = 0; base[1].z = 0;
                 base[2].x = 1; base[2].y = 1; base[2].z = 0;
                 base[3].x = 0; base[3].y = 1; base[3].z = 0;
                 base[4].x = 0; base[4].y = 0; base[4].z = 1;
                 base[5].x = 1; base[5].y = 0; base[5].z = 1;
                 base[6].x = 1; base[6].y = 1; base[6].z = 1;
                 base[7].x = 0; base[7].y = 1; base[7].z = 1;
                 for (k = 0; k < 8; k++) {
                      ndb_trans_coord_3by3(base[k], _cell,
                           NDB_TRANS_FRAC_TO_ORTHO);
                      matrix_rotate(base[k], rot);
                 }
                 is_draw_crystal = TRUE;
            }
       }

       if (is_draw_crystal) {
            draw_ps_picture(plot, rot, &base[0], area, abc);
       } else {
            draw_ps_picture(plot, rot, NULL, area, abc);
       }

       if (packopt) {
            plot.setline_width(0.5);
            if (max_length == MAX_LENGTH_A) {
                 draw_crystal_axes(plot, xll, yll, xur, yur, abc[1], "b");
                 draw_crystal_axes(plot, xll, yll, xur, yur, abc[2], "c");
                 draw_crystal_axes(plot, xll, yll, xur, yur, abc[0], "a");
            } else if (max_length == MAX_LENGTH_B) {
                 draw_crystal_axes(plot, xll, yll, xur, yur, abc[2], "c");
                 draw_crystal_axes(plot, xll, yll, xur, yur, abc[0], "a");
                 draw_crystal_axes(plot, xll, yll, xur, yur, abc[1], "b");
            } else {
                 draw_crystal_axes(plot, xll, yll, xur, yur, abc[0], "a");
                 draw_crystal_axes(plot, xll, yll, xur, yur, abc[1], "b");
                 draw_crystal_axes(plot, xll, yll, xur, yur, abc[2], "c");
            }
       } else {
            _CHAIN *chain = _molecules[0]->GetFirstChain();
            while (chain) {
                 if (chain->display() && chain->chain_type() == ATOMN_TOKEN)
                      break;
                 chain = _molecules[0]->GetNextChain();
            }
            if (chain) {
                 char buff[300];
                 x = plot.page_xmin() + 35.0;
                 y = plot.page_ymin() + 20.0;
                 strcpy(buff, "The nucleic acid is colored by residue. C, G, A, T, ");
                 strcat(buff, "U, and I are yellow, green, red, blue, cyan and dark ");
                 strcat(buff, "respectively.");
                 plot.print_string(buff, FONT_HELVETICA, 10, x, y+15, LJUSTIFY);
/*
                 plot.print_string("Other residues and drugs are black.",
                         FONT_HELVETICA, 10, x, y, LJUSTIFY);
*/
            }
       }
}

void _VALIDATION::draw_nmr_pictures(_plot &plot, const char *sid)
{
       char buffer[200];
       // int color_type = COLOR_BY_MOLECULE;

       sprintf(buffer, "Ensemble of Structure %s", sid);
       double x = (plot.page_xmin() + plot.page_xmax()) / 2.0;
       double y = plot.page_ymax() - 70;
       plot.print_string(buffer, FONT_HELVETICA, 20, x, y, CENTERX);
       plot.setcolor(0.0, 0.0, 0.0);
       plot.setline_width(0.2);

       double area[4], rot[4][4];
       area[0] = 50; area[1] = 100; area[2] = 550; area[3] = 600;
       int max_length = MAX_LENGTH_C;
       COORD abc[3];
       get_rotation_matrix(_molecules[0], NULL, rot, 0, _n_bspair, _bspair, max_length,
              &abc[0]);

       draw_ps_picture(plot, rot, NULL, area, abc);
}

static void draw_crystal_axes(_plot &plot, const double xll, const double yll, const double xur,
             const double yur, COORD &abc, const char *label)
{
       double char_size = 8;
       double width = xur - xll;
       double height = yur - yll;
       double length = MIN(width / 30.0, height / 30.0);
       double x1 = xll + width / 10.0 * 9.0;
       double y1 = yll + height / 10.0 * 9.0;
       double z1 = x1;

       plot.setcolor(0.0, 0.0, 0.0);
       plot.line_style(LINE_SOLID);
       plot.setfont(FONT_HELVETICA_OBL, 12, 0);

       double x = abc.x - (xur + xll) / 2.0;
       double y = abc.y - (yur + yll) / 2.0;
       double r = sqrt(x * x + y * y);
       x = length * x / r;
       y = length * y / r;
       double x2 = x1 + x;
       double y2 = y1 + y;
       double z2 = z1 + abc.z;
       plot.line(x1, y1, x2, y2);
       if (x2 > x1 && y2 > y1)
            plot.wrt_litstr_xy(label, x2 + char_size, y2 + char_size, CENTERXY);
       else if (y2 > y1)
            plot.wrt_litstr_xy(label, x2, y2 + char_size, CENTERXY);
       else if (x2 > x1 )
            plot.wrt_litstr_xy(label, x2 + char_size, y2, CENTERXY);
       else if (x2 < x1 && y2 < y1)
            plot.wrt_litstr_xy(label, x2 - char_size, y2 - char_size, CENTERXY);
       else if (y2 < y1 )
            plot.wrt_litstr_xy(label, x2, y2 - char_size, CENTERXY);
       else if (x2 < x1)
            plot.wrt_litstr_xy(label, x2 - char_size, y2, CENTERXY);
       else if (z1 >= z2)
            plot.wrt_litstr_xy(label, x2 + char_size, y2 + char_size, CENTERXY);
       else plot.wrt_litstr_xy(label, x2 - char_size, y2 - char_size, CENTERXY);
}

static void get_rotation_matrix(const int view, double m[4][4])
{
       static double x_angle[4] = {10, -100, -10, 45};
       static double y_angle[4] = {-10, 10, 100, 45};

       double mtmp[4][4];

       double torad = acos(-1.0) / 180.0;
       if (view == 2) {
            matrix_y_rotation(m, y_angle[view] * torad);
            matrix_x_rotation(mtmp, x_angle[view] * torad);
       } else {
            matrix_x_rotation(m, x_angle[view] * torad);
            matrix_y_rotation(mtmp, y_angle[view] * torad);
       }
       matrix_concatenate(m, mtmp);
}

static void get_rotation_matrix(_MOLECULE *mol, const NDBCELLCONST *cell,
         double m[4][4], const int packopt, const int n,
         _BSPAIR **base_pair, int &max_length, COORD *xyz)
{
       typedef struct {
               char chainID[5];
               int  op, lx, ly, lz;
       } SYM_CHAIN;
       int i, j, k;
       double target[9], source[9], rt[12], xm[3], ym[3];

       matrix_init(m);

       if (packopt) {
            double length = cell->c;
            if (cell->a > length) {
                 max_length = MAX_LENGTH_A;
                 length = cell->a;
            }
            if (cell->b > length) max_length = MAX_LENGTH_B;

            COORD abc[3];
            for (k = 0; k < 3; k++) {
                 abc[k].x = abc[k].y = abc[k].z = 0.0;
                 xyz[k].x = xyz[k].y = xyz[k].z = 0.0;
            }
            xyz[0].x = abc[0].x = 1.1 / cell->b;
            xyz[1].y = abc[1].y = 1.1 / cell->c;
            xyz[2].z = abc[2].z = 1.1 / cell->a;

            for (k = 0; k < 3; k++) {
                 ndb_trans_coord_3by3(abc[k], cell, NDB_TRANS_FRAC_TO_ORTHO);
                 vector_normalize(abc[k]);
            }
            vector_cross_product(abc[2], abc[0], abc[1]);
            vector_normalize(abc[2]);
            vector_cross_product(abc[1], abc[2], abc[0]);
            vector_normalize(abc[1]);

            for (k = 0; k < 9; k++) target[k] = 0.0;
            if (max_length == MAX_LENGTH_A) {
                 target[1] = 1.0; target[3] = 1.0; target[8] =-1.0;
            } else if (max_length == MAX_LENGTH_B) {
                 target[2] =-1.0; target[4] = 1.0; target[6] = 1.0;
            } else {
                 target[0] = 1.0; target[5] =-1.0; target[7] = 1.0;
            }

            source[0] = abc[0].x;
            source[1] = abc[0].y;
            source[2] = abc[0].z;
            source[3] = abc[1].x;
            source[4] = abc[1].y;
            source[5] = abc[1].z;
            source[6] = abc[2].x;
            source[7] = abc[2].y;
            source[8] = abc[2].z;

            superposition_matrix(3, target, source, rt, xm, ym);
            for (k = 0; k < 3; k++) 
                 for (j = 0; j < 3; j++)
                      m[k][j] = rt[j * 3 + k];

            double tmp_m[4][4];
            get_rotation_matrix(0, tmp_m);
            matrix_concatenate(m, tmp_m);
            return;
       } else if (n == 0) return;

       SYM_CHAIN sym_chain[10];
       int n_chain = 0;
       int exist_protein = 0;
       _CHAIN *chain = mol->GetFirstChain();
       while (chain) {
            if (chain->chain_type() == ATOMP_TOKEN &&
                chain->display()) exist_protein = 1;
            if (chain->chain_type() != ATOMN_TOKEN ||
                chain->display() == 0) {
                 chain = mol->GetNextChain();
                 continue;
            }

            strcpy(sym_chain[n_chain].chainID, chain->ChainID());
            sym_chain[n_chain].op = chain->op();
            sym_chain[n_chain].lx = chain->lx();
            sym_chain[n_chain].ly = chain->ly();
            sym_chain[n_chain].lz = chain->lz();
            n_chain++;
            chain = mol->GetNextChain();
       }
       if (n_chain == 0) return;

       COORD protein;
       protein.x = protein.y = protein.z = 0;
       if (exist_protein) {
            int count = 0;
            chain = mol->GetFirstChain();
            while (chain) {
                 if (chain->chain_type() != ATOMP_TOKEN ||
                     chain->display() == 0) {
                      chain = mol->GetNextChain();
                      continue;
                 }
                 _RESIDUE *residue = chain->GetFirstResidue();
                 while (residue) {
                      _ATOM *atom = residue->find_atom("CA");
                      if (atom) {
                           protein.x += atom->orig().x;
                           protein.y += atom->orig().y;
                           protein.z += atom->orig().z;
                           count++;
                      }
                      residue = chain->GetNextResidue();
                 }
                 chain = mol->GetNextChain();
            }
            protein.x /= (double) count;
            protein.y /= (double) count;
            protein.z /= (double) count;
       }
       if (n_chain == 1) {
            strcpy(sym_chain[1].chainID, sym_chain[0].chainID);
            sym_chain[1].op = sym_chain[0].op;
            sym_chain[1].lx = sym_chain[0].lx;
            sym_chain[1].ly = sym_chain[0].ly;
            sym_chain[1].lz = sym_chain[0].lz;
       }

       int first = 1;
       _RESIDUE *a_f = NULL, *a_e = NULL, *b_f = NULL, *b_e = NULL;
       char symmetry[10];
       strcpy(symmetry, "");
       if (sym_chain[1].op != 1) {
            sprintf(symmetry, "%d_%d%d%d", sym_chain[1].op,
                    sym_chain[1].lx+5, sym_chain[1].ly+5, sym_chain[1].lz+5);
       }
       for (i = 0; i < n; i++) {
            if (!strcmp(base_pair[i]->FstAtom, sym_chain[0].chainID) &&
                !strcmp(base_pair[i]->SndAtom, sym_chain[1].chainID) &&
                !strcmp(base_pair[i]->Asym_or_Gen, symmetry)) {
                 if (first) {
                      first = 0;
                      a_f = mol->Chains(base_pair[i]->FstIndex)->
                                 SeqResIndex(base_pair[i]->FstRes)->Residue;
                      b_f = mol->Chains(base_pair[i]->SndIndex)->
                                 SeqResIndex(base_pair[i]->SndRes)->Residue;
                 } else {
                      a_e = mol->Chains(base_pair[i]->FstIndex)->
                                 SeqResIndex(base_pair[i]->FstRes)->Residue;
                      b_e = mol->Chains(base_pair[i]->SndIndex)->
                                 SeqResIndex(base_pair[i]->SndRes)->Residue;
                 }
            }
       }
       if (a_f == NULL || a_e == NULL || b_f == NULL || b_e == NULL ||
           a_f == b_f || a_e == b_e || a_f == a_e || b_f == b_e) return;

       for (i = 0; i < 9; i++) target[i] = 0;
       target[0] = target[4] = target[8] = 1;

       COORD a, b, c, d, ab, cd, tmp;
       a.x = a.y = a.z = 0;
       int Num_Atom = 0;
       _ATOM *atom = a_f->GetFirstAtom();
       while (atom) {
            a.x += atom->orig().x;
            a.y += atom->orig().y;
            a.z += atom->orig().z;
            Num_Atom++;
            atom = a_f->GetNextAtom();
       }
       if (Num_Atom) {
            a.x /= (float) Num_Atom;
            a.y /= (float) Num_Atom;
            a.z /= (float) Num_Atom;
       }
       b.x = b.y = b.z = 0;
       Num_Atom = 0;
       atom = b_f->GetFirstAtom();
       while (atom) {
            b.x += atom->orig().x;
            b.y += atom->orig().y;
            b.z += atom->orig().z;
            Num_Atom++;
            atom = b_f->GetNextAtom();
       }
       if (Num_Atom) {
            b.x /= (float) Num_Atom;
            b.y /= (float) Num_Atom;
            b.z /= (float) Num_Atom;
       }
       c.x = c.y = c.z = 0;
       Num_Atom = 0;
       atom = a_e->GetFirstAtom();
       while (atom) {
            c.x += atom->orig().x;
            c.y += atom->orig().y;
            c.z += atom->orig().z;
            Num_Atom++;
            atom = a_e->GetNextAtom();
       }
       if (Num_Atom) {
            c.x /= (float) Num_Atom;
            c.y /= (float) Num_Atom;
            c.z /= (float) Num_Atom;
       }
       d.x = d.y = d.z = 0;
       Num_Atom = 0;
       atom = b_e->GetFirstAtom();
       while (atom) {
            d.x += atom->orig().x;
            d.y += atom->orig().y;
            d.z += atom->orig().z;
            Num_Atom++;
            atom = b_e->GetNextAtom();
       }
       if (Num_Atom) {
            d.x /= (float) Num_Atom;
            d.y /= (float) Num_Atom;
            d.z /= (float) Num_Atom;
       }
       vector_middle(ab, a, b);
       vector_middle(cd, c, d);

       if (exist_protein) {
            vector_difference(tmp, cd, ab);
            vector_normalize(tmp);
            source[0] = tmp.x; source[1] = tmp.y; source[2] = tmp.z;

            COORD tmp1;
            vector_middle(tmp1, ab, cd);
            vector_subtract(tmp1, protein);
            vector_cross_product(cd, tmp, tmp1);
            vector_normalize(cd);
            source[6] = cd.x; source[7] = cd.y; source[8] = cd.z;

            vector_cross_product(tmp1, cd, tmp);
            source[3] = tmp1.x; source[4] = tmp1.y; source[5] = tmp1.z;
       } else {
            vector_difference(tmp, ab, cd);
            vector_normalize(tmp);
            source[3] = tmp.x; source[4] = tmp.y; source[5] = tmp.z;
     
            vector_difference(ab, b, a);
            vector_cross_product(cd, ab, tmp);
            vector_normalize(cd);
            source[6] = cd.x; source[7] = cd.y; source[8] = cd.z;
     
            vector_cross_product(ab, tmp, cd);
            source[0] = ab.x; source[1] = ab.y; source[2] = ab.z;
       }

       superposition_matrix(3, target, source, rt, xm, ym);
       for (k = 0; k < 3; k++) 
            for (j = 0; j < 3; j++)
                 m[k][j] = rt[j * 3 + k];
}

static void register_to_html(const char *postfix, const char *legend)
{
       if (num_pic >= MAX_PICTURE) {
            ndb_log_message(NDB_MSG_ERR, "mum_pic (%d) exceeds maximum (%d)\n", num_pic, MAX_PICTURE);
            exit (1);
       }
       strcpy(pic_postfix[num_pic], postfix);
       strcpy(pic_legend[num_pic], legend);
       num_pic++;
}

static void append_to_html(const char *sid, const int public_flag)
{
       int i, first = 1, s1, s2, s3;
       char buff[160], sFile[100], tFile[100], pFile[100];
       FILE *fp = NULL;
       struct stat statbuf;

       memset(buff, 0, 160);
       sprintf(buff, "%s.html", sid);
       fp = fopen(buff, "a");

       if (num_pic) {
            memset(buff, 0, 160);
            sprintf(buff, "%s/bin/%s >& /dev/null", root_dir, CMD_PS2GIF);
            ndb_execute_command(buff);
            memset(buff, 0, 160);
            sprintf(buff, "%s/bin/%s >& /dev/null", root_dir, CMD_PS2PDF);
            ndb_execute_command(buff);
            sleep(5);

            for (i = 0; i < num_pic; i++) {
                 s1 = s2 = s3 = 0;
                 sprintf(sFile, "%s%s.ps", sid, pic_postfix[i]);
                 if (stat(sFile, &statbuf) == 0) s1 = 1;
                 sprintf(tFile, "%s%s.gif", sid, pic_postfix[i]);
                 if (stat(tFile, &statbuf) == 0) s2 = 1;
                 sprintf(pFile, "%s%s.pdf", sid, pic_postfix[i]);
                 if (stat(pFile, &statbuf) == 0) s3 = 1;

                 if (!s1 && !s2 && !s3) continue;
                 if (first) {
                      atlas_head(fp, 2, "Molecular Graphics");
                      atlas_par(fp);
                      first = 0;
                 }
                 if (s1 && s2 && s3) {
                      if (public_flag)
                           fprintf(fp, "%s <a href=\"%s\">(PS)</a> <a href=\"%s\" target=\"_blank\">(GIF)</a> <a href=\"%s\" target=\"_blank\">(PDF)</a> <p>\n",
                              pic_legend[i], sFile, tFile, pFile);
		      else fprintf(fp, "%s <a href=\"%s\">(PS)</a> <a href=\"%s\">(GIF)</a> <a href=\"%s\">(PDF)</a> <p>\n",
                              pic_legend[i], sFile, tFile, pFile);
		 } else if (s1 && s2) {
                      if (public_flag)
                           fprintf(fp, "%s <a href=\"%s\">(PS)</a> <a href=\"%s\" target=\"_blank\">(GIF)</a> <p>\n",
                              pic_legend[i], sFile, tFile);
		      else fprintf(fp, "%s <a href=\"%s\">(PS)</a> <a href=\"%s\">(GIF)</a> <p>\n",
                              pic_legend[i], sFile, tFile);
		 } else if (s1 && s3) {
                      if (public_flag)
                           fprintf(fp, "%s <a href=\"%s\">(PS)</a> <a href=\"%s\" target=\"_blank\">(PDF)</a> <p>\n",
                              pic_legend[i], sFile, pFile);
		      else fprintf(fp, "%s <a href=\"%s\">(PS)</a> <a href=\"%s\">(PDF)</a> <p>\n",
                              pic_legend[i], sFile, pFile);
		 } else if (s1) {
                      fprintf(fp, "%s <a href=\"%s\">(PS)</a> <p>\n", pic_legend[i], sFile);
                 }
            }
       }

       html_atlas_trailer(fp);
       fclose (fp);
}

static void add_color_to_chains(_MOLECULE *mol)
{
       int icolor = 0;
       _CHAIN *chain = mol->GetFirstChain();
       while (chain) {
            if (!chain->display() ||
                 chain->chain_type() != ATOMP_TOKEN) {
                 chain = mol->GetNextChain();
                 continue;
            }
            chain->set_icolor(icolor);
            icolor += 2;
            if (icolor >= MAXCHAINTYPE) icolor -= MAXCHAINTYPE;
            chain = mol->GetNextChain();
       }
}

static void select_biol_unit_to_display(_MOLECULE *mol, const int biol_id)
{
       _CHAIN *chain = mol->GetFirstChain();
       while (chain) {
            if (chain->display() < 0)
                 chain->set_display(0);
            else {
                 chain->set_display(0);
                 if (chain->biol_id() == biol_id || biol_id == 0)
                      chain->set_display(1);
            }
            chain = mol->GetNextChain();
       }
}

