/*
FILE:     ndb_ring.C
*/
/*
VERSION:  1.700
*/
/*
DATE:     7/13/2005
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2005 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

#include "erf_util.h"
#include "ndb_filterexitlib.h"
#include "ndb_misclib.h"
#include "ndb_ring.h"
#include "plot_util.h"

#define ALPHACONVENTION 1
#define OMEGACONVENTION 2
#define REGORDER 1
#define REVORDER 2

#define SYMBOL_FONT_alpha     "141"
#define SYMBOL_FONT_beta      "142"
#define SYMBOL_FONT_chi       "143"
#define SYMBOL_FONT_delta     "144"
#define SYMBOL_FONT_epsilon   "145"
#define SYMBOL_FONT_phi       "146"
#define SYMBOL_FONT_gamma     "147"
#define SYMBOL_FONT_eta       "150"
#define SYMBOL_FONT_iota      "151"
#define SYMBOL_FONT_phi1      "152"
#define SYMBOL_FONT_kappa     "153"
#define SYMBOL_FONT_lambda    "154"
#define SYMBOL_FONT_mu        "155"
#define SYMBOL_FONT_nu        "156"
#define SYMBOL_FONT_omicron   "157"
#define SYMBOL_FONT_pi        "160"
#define SYMBOL_FONT_theta     "161"
#define SYMBOL_FONT_rho       "162"
#define SYMBOL_FONT_sigma     "163"
#define SYMBOL_FONT_tau       "164"
#define SYMBOL_FONT_upsilon   "165"
#define SYMBOL_FONT_omega1    "166"
#define SYMBOL_FONT_omega     "167"
#define SYMBOL_FONT_xi        "170"
#define SYMBOL_FONT_psi       "171"
#define SYMBOL_FONT_zeta      "172"
#define SYMBOL_FONT_zero      "060"
#define SYMBOL_FONT_one       "061"
#define SYMBOL_FONT_two       "062"
#define SYMBOL_FONT_three     "063"
#define SYMBOL_FONT_four      "064"
#define SYMBOL_FONT_five      "065"
#define SYMBOL_FONT_six       "066"
#define SYMBOL_FONT_seven     "067"
#define SYMBOL_FONT_eight     "070"
#define SYMBOL_FONT_nine      "071"
#define SYMBOL_FONT_degree    "260"

static TORSION_STD *torsion_stds[40];
static int n_torsion_stds = 0;
static char date_torsion_stds[BUFFLEN];
static int max_torsion_stds = 40;

static NDBRING *fill_ring_struct(const char *structureid, const int ringorder, 
                 const int ringconvention, const char *type, TblIndexObj &_Index,
                 _TORSION **torsion);
static void ndb_draw_conf_ring_diagram(_plot &plot, const int nhframe, const int nvframe,
              const int nrings, const NDBRING *ring, const int make_title);
static double *get_torsion_values(const char *torsion_name, TblIndexObj &_Index,
                 _TORSION **torsion, int &numdat);
static char *ndb_roman_translation(const char *colname, const int convention);
static char *ndb_greek_translation(const char *colname, const int convention);
static double ndb_convert_angle_pref(const double angle, const double rotate);
static void ndb_draw_conf_rings(_plot& plot, const double xll, const double yll, 
             const double xur, const double yur, const NDBRING *ring, const int make_title);
static int ndb_get_a_torsion_std_ps(const NDBRING *ring, double **std1 , double **std2,
                double **mean, double **std_dev, char ***notes, int **ringn, 
                double *rcol, double *gcol, double *bcol);
static TORSION_STD *get_a_torsion_std(const char *name);
static int load_torsion_stds(void);
static void sort_torsion_std(const NDBRING *ring, TORSION_STD *std);
static TORSION_STD *make_a_torsion_std(void);
static int nullst_or_comment(const char *st);

void write_torsion_part(_plot &plot, const char *sid, const int struct_type,
           TblIndexObj &_Index, _TORSION **torsion)
{
       int ringorder = REVORDER, ringconvention = ALPHACONVENTION;
       int /* i, */ j;
       // char buffer[100];
       double xmin, xmax, ymin, ymax, x, y /* , length */ ;
       // static char conf_type[3][10] = {"A-DNA", "B-DNA", "Z-DNA"};
       // static char label[3][10] = {"(a)", "(b)", "(c)"};
       NDBRING *tring = NULL;
/*
       if (struct_type == ATOMN_TYPE_DNA_ONLY) {
            xmin = 196;
            xmax = 413;
            length = 215;
            for (i = 0; i < 3; i++) {
                 ymax = 760 - (double) i * (length + 10);
                 ymin = ymax - length;
                 plot.draw_view_window(xmin, xmax, ymin, ymax, label[i]);
     
                 tring = fill_ring_struct(sid, ringorder, ringconvention,
                           conf_type[i], _Index, torsion);
                 if (tring) {
                      // Make a PS image: the last argument calls for no title.
                      ndb_draw_conf_ring_diagram(plot, 1, 1, 1, tring, 0);
     
                      for (j = 0; j < tring->nrings; j++) {
                           if (tring->rings[j].name) delete [] tring->rings[j].name;
                           if (tring->rings[j].data) delete [] tring->rings[j].data;
                      }
                      if (tring->rings) delete [] tring->rings;
                      if (tring->angleoverlay) delete [] tring->angleoverlay;
                      if (tring->name) delete [] tring->name;
                      delete tring;
                 }
            }

            x = (plot.page_xmin() + plot.page_xmax()) / 2.0;
            y = plot.page_ymin() + 20.0;
            sprintf(buffer, "Torsion Wheel for Structure %s Overlayed Over", sid);
            strcat(buffer, " Standard Ranges.");
            plot.print_string(buffer, FONT_HELVETICA, 10, x, y+10, CENTERX);
            sprintf(buffer, "(a) A-DNA, (b) B-DNA, (c) Z-DNA.");
            plot.print_string(buffer, FONT_HELVETICA, 10, x, y-2, CENTERX);
       } else {
*/
            x = (plot.page_xmin() + plot.page_xmax()) / 2.0;
            y = plot.page_ymax();
            plot.setfont(FONT_HELVETICA, 30, 0);
            plot.wrt_litstr_xy("Conformation Wheel", x, y-60, CENTERX);
            plot.setfont(FONT_HELVETICA, 20, 0);
            plot.wrt_litstr_xy(sid, x, y-90, CENTERX);

            xmin = 116;
            xmax = 496;
            ymax = 560;
            ymin = 180;
            plot.draw_view_window(xmin, xmax, ymin, ymax, "");

            tring = fill_ring_struct(sid, ringorder, ringconvention,
                      NULL, _Index, torsion);
            if (tring) {
                 ndb_draw_conf_ring_diagram(plot, 1, 1, 1, tring, 0);

                 for (j = 0; j < tring->nrings; j++) {
                      if (tring->rings[j].name) delete [] tring->rings[j].name;
                      if (tring->rings[j].data) delete [] tring->rings[j].data;
                 }
                 if (tring->rings) delete [] tring->rings;
                 if (tring->angleoverlay) delete [] tring->angleoverlay;
                 if (tring->name) delete [] tring->name;
                 delete tring;
            }
/*
       }
*/
}

NDBRING *fill_ring_struct(const char *structureid, const int ringorder, 
                 const int ringconvention, const char *type, TblIndexObj &_Index,
                 _TORSION **torsion)
{
       int i, ithring;
       static char order[7][10] = {"ALPHA", "BETA", "GAMMA", "DELTA", "EPSILON",
                           "ZETA", "CHI"};
       NDBRING *ring = NULL;

       ring = new NDBRING;
       ring->name = new char[strlen(structureid)+1];
       strcpy(ring->name, structureid);
       if (type) {
            ring->angleoverlay = new char[strlen(type) + 1];
            strcpy(ring->angleoverlay, type);
       } else ring->angleoverlay = NULL;
       fflush(stdin);

       ring->nrings = 7;
       ring->rings = new NDBRINGDATA[ring->nrings];
       for (i = 0; i < ring->nrings; i++) {
            ring->rings[i].name = NULL;
            ring->rings[i].data = NULL;
       }
       ring->rcol1 = .70;
       ring->gcol1 = .70;
       ring->bcol1 = .95;
       ring->rcol2 = .85;
       ring->gcol2 = .85;
       ring->bcol2 = .95;
       ring->title_font_index = FONT_HELVETICA_OBL;
       ring->axis_font_index = FONT_HELVETICA_OBL;
       ring->label_font_index = FONT_SYMBOL;
 
       for (i = 0; i < ring->nrings; i++) {
            if (ringorder == REGORDER) ithring = i;
            else ithring = ring->nrings - i - 1;

            if (!ring->rings[ithring].name)
                 ring->rings[ithring].name = new char[80];
            strcpy(ring->rings[ithring].name, 
                 ndb_greek_translation(order[ithring], ringconvention));
            ring->rings[ithring].Rname1[0] = '\0';
            sprintf(ring->rings[ithring].Rname2, "%s",
                 ndb_roman_translation(order[ithring], ringconvention));

            if (ring->rings[ithring].data)
                 delete [] ring->rings[ithring].data;
            ring->rings[ithring].data = get_torsion_values(order[ithring],
                   _Index, torsion, ring->rings[ithring].n_realdat);
            ring->rings[ithring].numdat = ring->rings[ithring].n_realdat;
       }
 
       return ring;
}

static void ndb_draw_conf_ring_diagram(_plot &plot, const int nhframe, const int nvframe,
              const int nrings, const NDBRING *ring, const int make_title)
{
       int i, nrow, ncol;
       double xw_min, xw_max, yw_min, yw_max, frame_width, frame_height;
       double xll, yll, xur, yur;

       xw_min = plot.view_xmin();
       xw_max = plot.view_xmax();
       yw_min = plot.view_ymin();  
       yw_max = plot.view_ymax();
       frame_width  = (xw_max - xw_min) / (double) nhframe;
       frame_height = (yw_max - yw_min) / (double) nvframe;
 
       for (i = 0; i < nrings; i++) {
            nrow = (i/nhframe) + 1;
            ncol = i - (i/nhframe)*nhframe + 1;
            xll = xw_min + (double)(ncol-1)*frame_width;
            yll = yw_max - (double)(nrow)*frame_height;
            xur = xw_min + (double)(ncol)*frame_width;
            yur = yw_max - (double)(nrow-1)*frame_height;

            plot.box(xll,yll,xur,yur,0.0,0.0,0.0,0);  
            
            ndb_draw_conf_rings(plot,xll,yll,xur,yur,&ring[i], make_title);
       }
}

static double *get_torsion_values(const char *torsion_name, TblIndexObj &_Index,
                _TORSION **torsion, int &numdat)
{
       int ret;
       double *data = NULL;
       CifString index;

       numdat = 0;
       index.Copy(torsion_name);
       ret = _Index.Seek(index);
       while (ret >= 0) {
            if (index != _Index.GetCurrData()) break;
            numdat++;
            ret = _Index.GoToNext();
       }
       if (!numdat) return NULL;

       data = new double[numdat];
       numdat = 0;
       ret = _Index.Seek(index);
       while (ret >= 0) {
            if (index != _Index.GetCurrData()) break;
            data[numdat] = torsion[ret]->val;
            numdat++;
            ret = _Index.GoToNext();
       }
       return data;
}

static char *ndb_roman_translation(const char *colname, const int convention)
{
       static char result[10];
       if (convention == ALPHACONVENTION) {
            if (!strcmp("ALPHA", colname))        strcpy(result, "alpha");
            else if (!strcmp("BETA", colname))    strcpy(result, "beta");
            else if (!strcmp("GAMMA", colname))   strcpy(result, "gamma");
            else if (!strcmp("DELTA", colname))   strcpy(result, "delta");
            else if (!strcmp("EPSILON", colname)) strcpy(result, "epsilon");
            else if (!strcmp("ZETA", colname))    strcpy(result, "zeta");
            else if (!strcmp("CHI", colname))     strcpy(result, "chi");
       } else if (convention == OMEGACONVENTION) {
            if (!strcmp("ALPHA", colname))        strcpy(result, "omega");
            else if (!strcmp("BETA", colname))    strcpy(result, "phi");
            else if (!strcmp("GAMMA", colname))   strcpy(result, "psi");
            else if (!strcmp("DELTA", colname))   strcpy(result, "psi1");
            else if (!strcmp("EPSILON", colname)) strcpy(result, "phi1");
            else if (!strcmp("ZETA", colname))    strcpy(result, "omega1");
       }
       return(result);
}

static char *ndb_greek_translation(const char *colname, const int convention)
{
       static char label[80];
       char escape = '\\';
       memset(label, 0, 80);
       strcpy(label,colname);
       if (convention == ALPHACONVENTION) {
            if (!strcmp("ALPHA", colname))
                 sprintf(label, "%c%s", escape, SYMBOL_FONT_alpha);
            else if (!strcmp("BETA", colname))
                 sprintf(label, "%c%s", escape, SYMBOL_FONT_beta);
            else if (!strcmp("GAMMA", colname))
                 sprintf(label, "%c%s", escape, SYMBOL_FONT_gamma);
            else if (!strcmp("DELTA", colname))
                 sprintf(label, "%c%s", escape, SYMBOL_FONT_delta);
            else if (!strcmp("EPSILON", colname))
                 sprintf(label, "%c%s", escape, SYMBOL_FONT_epsilon);
            else if (!strcmp("ZETA", colname))
                 sprintf(label, "%c%s", escape, SYMBOL_FONT_zeta);
       } else if (convention == OMEGACONVENTION) {
            if (!strcmp("ALPHA", colname))
                 sprintf(label, "%c%s", escape, SYMBOL_FONT_omega);
            else if (!strcmp("BETA", colname))
                 sprintf(label, "%c%s", escape, SYMBOL_FONT_phi);
            else if (!strcmp("GAMMA", colname))
                 sprintf(label, "%c%s", escape, SYMBOL_FONT_psi);
            else if (!strcmp("DELTA", colname))
                 sprintf(label, "%c%s%c%s", escape, SYMBOL_FONT_psi, 
                          escape, SYMBOL_FONT_one);
            else if (!strcmp("EPSILON", colname))
                 sprintf(label, "%c%s", escape, SYMBOL_FONT_phi1);
            else if (!strcmp("ZETA", colname))
                 sprintf(label, "%c%s", escape, SYMBOL_FONT_omega1);
       }

       if (!strcmp("CHI", colname))
            sprintf(label, "%c%s", escape, SYMBOL_FONT_chi);
       return(label);
}


static double ndb_convert_angle_pref(const double angle, const double rotate)
{
       double result = rotate - angle;
       if (result > 360.0) result -= 360.0;
       if (result <  0.0)  result += 360.0;
       return(result);
}

static void ndb_draw_conf_rings(_plot &plot, const double xll, const double yll, 
             const double xur, const double yur, const NDBRING *ring, const int make_title)
{
       int i, j, ndigits_per_label=5, lentitle, ntics, *ringn, nstds;
       double frame_width, xc, yc, rinc, r, radius1, radius2, max_radius;
       double max_axis_lbl_font_pts = 20.0, axis_lbl_frac_size=0.20;
       double max_title_font_pts=20.0,title_frac_size=.50; 
       double lbl_angle = 90.0, t, font_pts, title_pts = 0, margin;
       double char_size, torad, x1,y1,x2,y2, costh, sinth, tic;
       double lbl_font_pts, tic_length, tic_interval, rcol, gcol, bcol;
       double *std1, *std2, *mean, *std_dev;
       char lbl[80], escape = '\\', title[80], **notes;

       strcpy(title, "");
       title_frac_size = 1.0; 
       if (make_title) {
            if (ring->angleoverlay)
                 sprintf(title, "%s over expected ranges for %s.",
                      ring->name, ring->angleoverlay);
            else sprintf(title, "%s", ring->name);
       }
 
       torad = acos(-1.0) / 180.0;

       frame_width  = xur - xll;
       font_pts = (frame_width * axis_lbl_frac_size) / (double) ndigits_per_label;
       if (font_pts > max_axis_lbl_font_pts) font_pts = max_axis_lbl_font_pts;
       margin = font_pts / 2.0;
       char_size = font_pts;
 
       lentitle = strlen(title);
       if (lentitle && make_title) {
            title_pts = (frame_width * title_frac_size) / (double) lentitle;
            if (title_pts > max_title_font_pts) title_pts = max_title_font_pts;
       }
 
       radius1 = frame_width / 2.0 - ((double) ndigits_per_label
               * font_pts) - 2.0 * margin;
       radius2 = frame_width / 2.0 - (title_pts + font_pts) - 4.0 * margin;
       max_radius = max(radius1, radius2);
 
       xc = (xur + xll) / 2.0;
       yc = (yur + yll) / 2.0;
       rinc = max_radius / ((double) ring->nrings + 1.0);

       if (ring->angleoverlay) {
            nstds = ndb_get_a_torsion_std_ps(ring, &std1 , &std2, &mean, &std_dev,
                       &notes, &ringn, &rcol, &gcol, &bcol);
            if ((nstds <= 0)) {
                 ndb_log_message(NDB_MSG_ERR,"standard %s not found in library.\n", 
                           ring->angleoverlay);
                 sleep(10);
                 return;	
            } 

            for (i = 0; i < ring->nrings; i++) {
                 r = max_radius - (double) i * rinc;
                 plot.curve(xc, yc, r, 0.0, 360.0, 1.0, 1.0, 1.0, 1);   

                 // Means and standard deviations
                 for (j = 0; j < nstds; j++) {
                      if (i == ringn[j] && mean[j]) {
                           plot.curve(xc, yc, r,
                      	       ndb_convert_angle_pref(mean[j] + std_dev[j], 90.0),
                      	       ndb_convert_angle_pref(mean[j] - std_dev[j], 90.0),
                               0.75, 0.75, 0.75, 1);
                           plot.curve(xc, yc, r,
                               ndb_convert_angle_pref(mean[j] + 0.5 * std_dev[j], 90.0),
                      	       ndb_convert_angle_pref(mean[j] - 0.5 * std_dev[j], 90.0),
                      	       0.6, 0.6, 0.6, 1);
                           plot.curve(xc, yc, r,
                      	       ndb_convert_angle_pref(mean[j] + 2, 90.0),
                      	       ndb_convert_angle_pref(mean[j] - 2, 90.0),
                      	       0.1, 0.1, 0.1, 1);
                      }
                 }
                 r = max_radius - (((double)i + 0.5) * rinc);
                 plot.curve(xc, yc, r, 0.0, 360.0, 1.0, 1.0, 1.0, 1);   
              
                 // Std ranges
                 for (j = 0; j < nstds; j++) {
                      if (i == ringn[j]) {
                           plot.curve(xc, yc, r, ndb_convert_angle_pref(std2[j], 90.0),
                             ndb_convert_angle_pref(std1[j], 90.0), 0.415, 0.934, 1.0, 1); 
                      }
                 }
            }

            /* fill the central ring with white paint. */
            r = max_radius - (double) (ring->nrings) * rinc;
            plot.curve(xc, yc, r, 0.0, 360.0, 1.0, 1.0, 1.0, 1);
  
            /* laydown the concentric rings. */
            for (i = 0; i < (ring->nrings + 1); i++) {
                 r = max_radius - (double)i * rinc;
                 plot.curve(xc, yc, r, 0.0, 360.0, 0.0, 0.0, 0.0, 0);
            }
  
            tic_length = margin/2.0;
            tic_interval = 30.0;
            ntics = (int) (360.0/tic_interval);
            r = max_radius;

            for (i=0; i < ntics; i++) {
                 tic = (double) i * tic_interval;
                 costh = cos(tic * torad);
                 sinth = sin(tic * torad);
                 x1 = xc + r * costh;
                 y1 = yc + r * sinth;
                 x2 = x1 + tic_length * costh;
                 y2 = y1 + tic_length * sinth;
                 plot.line(x1, y1, x2, y2);
            }
       } else { /*  laydown rings with no paint. */
            // laydown the concentric rings.
            for (i = 0; i < (ring->nrings + 1); i++) {
                 r = max_radius - (double) i * rinc;
                 plot.curve(xc, yc, r, 0.0, 360.0, 0.0, 0.0, 0.0, 0);
                 tic_length = margin/2.0;
                 tic_interval = 30.0;
                 ntics = (int) (360.0 / tic_interval);
                 r = max_radius;
                 for (j = 0; j < ntics; j++) {
                      tic = (double) j * tic_interval;
                      costh = cos(tic * torad);
                      sinth = sin(tic * torad);
                      x1 = xc + r * costh;
                      y1 = yc + r * sinth;
                      x2 = x1 + tic_length * costh;
                      y2 = y1 + tic_length * sinth;
                      plot.line(x1, y1, x2, y2);
                 }
            }
       }

       for (i = 0; i < ring->nrings; i++) {
            if (ring->angleoverlay)
                 r = max_radius - ((double) i + 0.5) * rinc;   
            else r = max_radius - ((double) i) * rinc;   

            for (j = 0; j < ring->rings[i].numdat; j++) {
                 if (ring->rings[i].data[j] != NDB_NONSENSE_DOUBLE) {
                      t = torad * ndb_convert_angle_pref(ring->rings[i].data[j], 90.0);
                      costh = cos(t);
                      sinth = sin(t);
                      x1 = xc + r * costh;
                      y1 = yc + r * sinth;
                      if (ring->angleoverlay) {
                           x2 = x1 - 0.5 * rinc*costh;
                           y2 = y1 - 0.5 * rinc*sinth; 
                      } else {
                           x2 = x1 - rinc*costh;
                           y2 = y1 -rinc*sinth;
                      }
                      plot.line(x1,y1,x2,y2);

                 }
            }
       }

       plot.setcolor(0.0,0.0,0.0);
       lbl_font_pts = (/* xpts_per_unit* */ rinc)/1.2;
       plot.setfont(ring->label_font_index,lbl_font_pts,0.0);

       costh = cos(lbl_angle*torad);
       sinth = sin(lbl_angle*torad);  
       for ( i= 0; i < ring->nrings; i++) {
        r = max_radius - (double)i * rinc - rinc / 2.7;
            x1 = xc + r*costh;
            y1 = yc + r*sinth;
            plot.wrt_litstr_xy(ring->rings[i].name, x1, y1, CENTERXY);
       }

       plot.setcolor(0.0, 0.0, 0.0);
       plot.setfont(ring->label_font_index, font_pts, 0.0);

       sprintf(lbl, "%c%s%c%s", escape, SYMBOL_FONT_zero, 
      	 escape, SYMBOL_FONT_degree);
       plot.wrt_litstr_xy(lbl, xc, yc+max_radius+margin, CENTERX);

       sprintf(lbl, "%c%s%c%s%c%s", escape, SYMBOL_FONT_nine, 
      	 escape, SYMBOL_FONT_zero, 
      	 escape, SYMBOL_FONT_degree);
       plot.wrt_litstr_xy(lbl, xc+max_radius+margin, yc, CENTERY);

       sprintf(lbl, "%c%s%c%s%c%s%c%s", escape, SYMBOL_FONT_one, 
      	 escape, SYMBOL_FONT_eight, 
      	 escape, SYMBOL_FONT_zero, 
      	 escape, SYMBOL_FONT_degree);
       plot.wrt_litstr_xy(lbl, xc, yc-max_radius-margin-char_size, CENTERX);

       sprintf(lbl, "%c%s%c%s%c%s%c%s", escape, SYMBOL_FONT_two, 
      	 escape, SYMBOL_FONT_seven, 
      	 escape, SYMBOL_FONT_zero, 
      	 escape, SYMBOL_FONT_degree);
       plot.wrt_litstr_xy(lbl, xc-max_radius-margin, yc, RJUSTIFY_CENTERY);

       if (lentitle && make_title) {
            plot.setfont(ring->title_font_index, title_pts, 0.0);
            plot.wrtstr_xy(title, xc, yll+margin, CENTERX);
       }
}

static int ndb_get_a_torsion_std_ps(const NDBRING *ring, double **std1 , double **std2,
                double **mean, double **std_dev, char ***notes, int **ringn, 
                double *rcol, double *gcol, double *bcol)
{
       TORSION_STD *std = NULL;
       if (ring->angleoverlay)
            std = get_a_torsion_std(ring->angleoverlay);
       else return (0);
       if(std != NULL) {
            sort_torsion_std(ring, std);
            *std1 = std->std1;
            *std2 = std->std2;
            *mean = std->mean;
            *std_dev = std->std_dev;
            *notes = std->notes;
            *ringn = std->ringn;
            *rcol = std->colr;
            *gcol = std->colg;
            *bcol = std->colb;
            return(std->n);
       } else return(0);
}

static TORSION_STD *get_a_torsion_std(const char *name)
{
       int i;
       if (n_torsion_stds <= 0) {
            load_torsion_stds();
            if (n_torsion_stds <= 0) {
                 ndb_log_message(NDB_MSG_ERR,"error: torsion %s not found in library.\n",name);
                 sleep(5);
                 return(NULL);
            }
       }
      
       for (i = 0; i < n_torsion_stds; i++)
            if (!(strcmp(torsion_stds[i]->name, name)))
                 return(torsion_stds[i]);
       return NULL;
}

static int load_torsion_stds(void)
{
       int nstds = 0, i, j;
       char fname[BUFFLEN], buff[BUFFLEN], tmp[BUFFLEN];
       FILE *fp = NULL;
       TORSION_STD *std = NULL;

       sprintf(fname, "%s/data/ascii/torsions", root_dir);
       fp = fopen(fname, "r");
       if (!fp) {
            ndb_log_message(NDB_MSG_ERR,"warning: standard torsion file %s not found.\n",fname);
            filter_unexpctderror(0);
       }
       fgets(buff, BUFFLEN, fp);
       while (!(feof(fp)) && (nullst_or_comment(buff))) fgets(buff, BUFFLEN, fp);
       if (strncmp("NDB torsion file", buff, 16)) {
            ndb_log_message(NDB_MSG_ERR,
                  "warning: standard torsion file %s corrupted.\n",fname);
            filter_unexpctderror(0);
       }

       fgets(date_torsion_stds, BUFFLEN, fp);
       while (!(feof(fp)) && (nullst_or_comment(buff)))
            fgets(date_torsion_stds, BUFFLEN, fp);
      
       fgets(buff, BUFFLEN, fp);
       while(!(feof(fp)) && (nullst_or_comment(buff))) fgets(buff, BUFFLEN, fp);
      
       if (feof(fp)) {
            ndb_log_message(NDB_MSG_ERR,
                    "Error: standard torsion file %s empty.\n",fname);
            filter_unexpctderror(0);
       }
      
       /* Read torsion standards. */
       sscanf(buff, "%d", &nstds);
       if (nstds >= MAXSTANDARDS){
            ndb_log_message(NDB_MSG_ERR,"Torsion file too big.\n");
            filter_unexpctderror(0);
       }

       for (j = 0; j < nstds; j++) {
            fscanf(fp,"%s", buff);
            if (feof(fp)) {
                 ndb_log_message(NDB_MSG_ERR,
                        "standard torsions file %s corrupted.\n",fname);
                 filter_unexpctderror(0);
            }
            std = make_a_torsion_std();
            strcpy(std->name, buff);
            fscanf(fp, "%d %d %lf %lf %lf",&std->nrings, &std->n,
                   &std->colr, &std->colg, &std->colb);
            if (std->n >= MAXSTANDARDSinRINGS) {
                 ndb_log_message(NDB_MSG_ERR,"Torsion standards file too deep.\n");
                 filter_unexpctderror(0);
            }
          
            /* For a one of the standards, i.e., A-DNA, read all of the torsions. */
            for (i = 0; i < std->n; i++) {
                 fscanf(fp,"%s %lf %lf %lf %lf", buff, &std->std1[i],&std->std2[i],
                        &std->mean[i], &std->std_dev[i]);
                 fgets(tmp, BUFFLEN, fp);
                 if (feof(fp)) {
                      ndb_log_message(NDB_MSG_ERR,
                         "error: standard torsions file %s corrupted.\n",fname);
                      filter_unexpctderror(0);
                 }
                 if (strlen(tmp)) {
                      std->notes[i] = (char *)malloc((strlen(tmp) + 1) * sizeof(char));
                      strcpy(std->notes[i], tmp);
                 }
                 else std->notes[i] = NULL;
                
                 std->torsion[i] = (char *)malloc((strlen(buff) + 1) * sizeof(char));
                 strcpy(std->torsion[i], buff);
                
                 if (n_torsion_stds >=  max_torsion_stds) {
                      ndb_log_message(NDB_MSG_ERR,
                           "error: too many torsion standards.\n");
                      filter_unexpctderror(0);
                 }
            }

            /* Add the complete standard to the `list' of standards. */
            torsion_stds[n_torsion_stds] = std;
            n_torsion_stds++;
       }
       fclose(fp);
       return 1;
}

static void sort_torsion_std(const NDBRING *ring, TORSION_STD *std)
{
       int i, ii;
       for (i = 0; i < std->n; i++)  {
            for (ii = 0; ii < ring->nrings; ii++) {
                 if (!(strcmp(ring->rings[ii].Rname2, std->torsion[i]))) {
                      std->ringn[i] = ii;
                      break;
                 }
            }
            if (ii >= ring->nrings) {
                 ndb_log_message(NDB_MSG_ERR,
                      "error: torsion %s not found in library.\n",std->torsion[i]);
                 filter_unexpctderror(0);
            }
       }
}

static TORSION_STD *make_a_torsion_std(void)
{
       int i;
       TORSION_STD *std = new TORSION_STD;
       std->max = MAXTORSIONS;
       std->colr = 0.0;
       std->colg = 0.0;
       std->colb = 0.0;
       for (i = 0; i < 20; i++) {
            std->match[i] = 0;
            std->torsion[i] = NULL;
            std->std1[i] = 0.0;
            std->std2[i] = 0.0;
            std->mean[i] = 0.0;
            std->std_dev[i] = 0.0;
            std->notes[i] = NULL;
       }
       return std;
}

static int nullst_or_comment(const char *st)
{
       while (isspace(*st) && (*st != '\0')) st++;
       return ((*st == '#') || (*st == '\0'));
}

