/*
FILE:     plot_util.C
*/
/*
VERSION:  1.700
*/
/*
DATE:     7/13/2005
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2005 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>

#include "ndb_misclib.h"
#include "plot_util.h"
#include "map.h"

#define ALPHA     "\141"
#define BETA      "\142"
#define CHI       "\143"
#define DELTA     "\144"
#define EPSILON   "\145"
#define GAMMA     "\147"
#define ZETA      "\172"

#define  NUM_OF_BASE_PARAM     8
static const char *base_params[NUM_OF_BASE_PARAM] = {"Buckle", "Propeller", "Opening",
                "Slide", "Rise", "Tilt", "Roll", "Twist"};

static void draw_a_view(_plot& plot, const int n_pair, _PAIR_VALUE **pair,
         const char *x_title, const char *y_title, int map[SIZE][SIZE]);
static void page_head(_plot &plot, const char *id, const int type, const char base);
static void page_tail(_plot &plot, const int type);
static void page_base_morphology_head(_plot &plot, const char *id);
static void page_base_morphology_tail(_plot &plot);
static void ramfil(_plot &plot);
static void ramlin(_plot &plot);
static void ramlet(_plot &plot);
static void ramreg(const float phi, const float psi, char *region, int &regno, int &regtyp);
static void ramput(_plot &plot, const char *str1, const char *str2, const char *str3,
       const double x, const double x1, const double x2, double &y, const double ygap);

void get_torsion_distribution(_plot &plot, const char *sid, const int _n_z_a_1,
            const int _n_z_b_1, const int _n_z_e, const int _n_a_g,
            const int _n_z_x, const int _n_d_x, _PAIR_VALUE **_z_a_1,
            _PAIR_VALUE **_z_b_1, _PAIR_VALUE **_z_e, _PAIR_VALUE **_a_g,
            _PAIR_VALUE **_z_x, _PAIR_VALUE **_d_x)
{
       plot.set_actual_value_xmin_and_max(0.0, 360.0);
       plot.set_actual_value_ymin_and_max(0.0, 360.0);

       char buffer[200];

       double x = (plot.page_xmin() + plot.page_xmax()) / 2.0;
       double y = plot.page_ymax() - 30;
       plot.print_string("Scattergrams of Selected Torsion Angles",
                     FONT_HELVETICA, 25, x, y, CENTERX);
       plot.print_string(sid, FONT_HELVETICA, 15, x, y - 20, CENTERX);

       sprintf(buffer, "%s + 1", ALPHA);
       plot.set_view_xmin_and_max(105, 270);
       plot.set_view_ymin_and_max(540, 705);
       draw_a_view(plot, _n_z_a_1, _z_a_1, ZETA, buffer, Zeta_Alpha);
       plot.set_view_xmin_and_max(345, 510);
       plot.set_view_ymin_and_max(540, 705);
       sprintf(buffer, "%s + 1", BETA);
       draw_a_view(plot, _n_z_b_1, _z_b_1, ZETA, buffer, Zeta_Beta);
       plot.set_view_xmin_and_max(105, 270);
       plot.set_view_ymin_and_max(335, 495);
       draw_a_view(plot, _n_z_e, _z_e, ZETA, EPSILON, Zeta_Epsilon);
       plot.set_view_xmin_and_max(345, 510);
       plot.set_view_ymin_and_max(335, 495);
       draw_a_view(plot, _n_a_g, _a_g, ALPHA, GAMMA, Alpha_Gamma);
       plot.set_view_xmin_and_max(105, 270);
       plot.set_view_ymin_and_max(125, 290);
       draw_a_view(plot, _n_z_x, _z_x, ZETA, CHI, Zeta_Chi);
       plot.set_view_xmin_and_max(345, 510);
       plot.set_view_ymin_and_max(125, 290);
       draw_a_view(plot, _n_d_x, _d_x, DELTA, CHI, Delta_Chi);
       
       x = plot.page_xmin() + 10.0;
       y = plot.page_ymin() + 20.0;

       strcpy(buffer, "Based on analysis of selected structures of resolution of ");
       strcat(buffer, "at least 1.9 Angstroms, A DNA (yellow), B DNA (blue), Z DNA");
       plot.print_string(buffer, FONT_HELVETICA, 10, x, y+15.0, LJUSTIFY);
       strcpy(buffer, "(red), A and B DNA (green), A and Z DNA (orange), B and Z ");
       strcat(buffer, "DNA (purple), A, B and Z DNA (brown). [B. Schneider et al.,");
       plot.print_string(buffer, FONT_HELVETICA, 10, x, y+5.0, LJUSTIFY);
       strcpy(buffer, "Conformations of the Sugar-Phosphate Backbone in Helical ");
       strcat(buffer, "DNA Crystal Structures Biopoly. 1997, 42, 113-124]");
       plot.print_string(buffer, FONT_HELVETICA, 10, x, y-5.0, LJUSTIFY);
}

void get_bond_angle_distribution(_plot &plot, const double xlength, 
     const double ylength, const double l_margin, const double r_margin,
     const double b_margin, const double t_margin, int x_num, int y_num,
     const int num, MEADESD *parm, const int intrlx, const int type,
     const char base, const char *id, TblIndexObj &_Index, _BOND **_values)
{
       int xgap = 4, ygap = 4;
       char buffer[100];
       CifString index;
       int i = 0, j = 0, ibar = 0, nbars = 0, *bcount = NULL, maxfrq = 0;
       int ret = -1, ret1 = -1, x = 1, y = 1;
       double xmin, ymin, xmax, ymax, view_xt, view_yt;
       double view_xmin = 0, view_xmax = 0, view_ymin = 0, view_ymax = 0;
       double value_max = -1000.0, value_min = 1000.0, actual_ymax = 0;
       double actual_value_min = 1000.0, actual_value_max = -1000.0;
       double value_min_esd = 0, value_max_esd = 0, tmp = 0, width  = 0;
       double xmin1 = 0, xmax1 = 0, ymax1 = 0;

       plot.set_axes_label_precision(intrlx, 0);
       page_head(plot, id, type, base);
       xmin = plot.page_xmin() + l_margin;
       ymin = plot.page_ymin() + b_margin;
       xmax = plot.page_xmax() - r_margin;
       ymax = plot.page_ymax() - t_margin;

       if (!x_num || !y_num) {
            x_num = (int) ((xmax - xmin) / xlength);
            y_num = (int) ((ymax - ymin) / ylength);
            if ((xmax - xmin - (double) x_num * xlength) < 
                xlength / 2.0) x_num--;
            if ((ymax - ymin - (double) y_num * ylength) < 
                ylength / 2.0) y_num--;
       }

       if (x_num * y_num >= 20) xgap = 4;

       view_xt = (xmax - xmin - (double) x_num * xlength) / (double) x_num;
       view_yt = (ymax - ymin - (double) y_num * ylength) / (double) y_num;

       for (i = 0; i < num; i++) {
            if (base != ' ')
                 sprintf(buffer, "%c%s", base, parm[i].label);
            else strcpy(buffer, parm[i].label);
            index.Copy(buffer);
            ret = _Index.Seek(index);
            if (ret < 0) continue;

            view_xmin = xmin + (double) x * view_xt + 
                    (double) (x - 1) * xlength;
            view_xmax = view_xmin + xlength;
            if (view_xmax > xmax) {
                 x = 1;
                 y++;
                 view_xmin = xmin + (double) x * view_xt + 
                             (double) (x - 1) * xlength;
                 view_xmax = view_xmin + xlength;
            }

            view_ymin = ymax - (double) y * (view_yt + ylength);
            view_ymax = view_ymin + ylength;
            if (view_ymin < ymin) {
                 page_tail(plot, type);
                 plot.newpage(1);
                 page_head(plot, id, type, base);
                 x = y = 1;
                 view_xmin = xmin + (double) x * view_xt +
                             (double) (x - 1) * xlength;
                 view_xmax = view_xmin + xlength;
                 view_ymin = ymax - (double) y * (view_yt + ylength);
                 view_ymax = view_ymin + ylength;
            }

            value_max = -1000.0;
            value_min = 1000.0;
            ret1 = ret;
            while (ret1 >= 0) {
                 if (index != _Index.GetCurrData()) break;
                 if (_values[ret1]->val > value_max) value_max = _values[ret1]->val;
                 if (_values[ret1]->val < value_min) value_min = _values[ret1]->val;
                 ret1 = _Index.GoToNext();
            }
            actual_value_min = 1000.0;
            actual_value_max = -1000.0;
            if (parm[i].have_meanstd) {
                 for (j = 0; j < parm[i].num; j++) {
                      tmp = parm[i].meanstd[j].mean - 4.0 
                                 * parm[i].meanstd[j].esd;
                      if (tmp < actual_value_min) {
                           actual_value_min = tmp;
                           value_min_esd = parm[i].meanstd[j].esd;
                      }
                      tmp = parm[i].meanstd[j].mean + 4.0
                          * parm[i].meanstd[j].esd;
                      if (tmp > actual_value_max) {
                           actual_value_max = tmp;
                           value_max_esd = parm[i].meanstd[j].esd;
                      }
                 }
                 if (value_min < actual_value_min) 
                      actual_value_min = value_min - value_min_esd;
                 if (value_max > actual_value_max)
                      actual_value_max = value_max + value_max_esd;
            } else {
                 actual_value_min = value_min - (value_max - value_min) / 4.0;
                 actual_value_max = value_max + (value_max - value_min) / 4.0;
            }


            width = (actual_value_max - actual_value_min) / 60.0;
            nbars = (int) ((actual_value_max - actual_value_min) / width + 0.1);
            bcount = new int[nbars];
            for (j = 0; j < nbars; j++) bcount[j] = 0;
            maxfrq = 0;
            ret1 = ret;
            _Index.GoToRecord(ret1);
            while (ret1 >= 0) {
                 if (index != _Index.GetCurrData()) break;
                 ibar = (int) ((_values[ret1]->val - actual_value_min) / width);
                 bcount[ibar]++;
                 if (bcount[ibar] > maxfrq) maxfrq = bcount[ibar];
                 ret1 = _Index.GoToNext();
            }
            
            j = 1; while (j * ygap < maxfrq) j++;
            actual_ymax = (double) ((j + 1) * ygap);

            plot.set_axes_interval(xgap, ygap);
            plot.set_actual_value_xmin_and_max(actual_value_min, actual_value_max);
            plot.set_actual_value_ymin_and_max(0.0, actual_ymax);
            plot.set_picture_title(parm[i].label);
            plot.set_x_axis_title("");
            plot.set_y_axis_title("Frequency");
            plot.set_view_xmin_and_max(view_xmin, view_xmax);
            plot.set_view_ymin_and_max(view_ymin, view_ymax);
            plot.draw_view_window(0, FONT_HELVETICA, 10.0, 8.0, 0.0, 20.0, 1);
            for (j = 0; j < nbars; j++) {
                 if (!bcount[j]) continue;
                 xmin1 = (double) j * width + actual_value_min;
                 xmax1 = xmin1 + width;
                 ymax1 = (double) bcount[j];
                 plot.plot_a_bar(xmin1, xmax1, 0.0, ymax1, 0.5, 0.0, 1.0);
            }
            if (parm[i].have_meanstd) {
                 for (j = 0; j < parm[i].num; j++) {
                      double x_value = parm[i].meanstd[j].mean;
                      plot.plot_a_line(x_value, 0.0, x_value, actual_ymax, LINE_SOLID);
                      x_value = parm[i].meanstd[j].mean 
                              - parm[i].meanstd[j].esd;
                      plot.plot_a_line(x_value, 0.0, x_value, actual_ymax, LINE_DASH2);
                      x_value = parm[i].meanstd[j].mean 
                              + parm[i].meanstd[j].esd;
                      plot.plot_a_line(x_value, 0.0, x_value, actual_ymax, LINE_DASH2);
                 }
            }
            delete [] bcount;
            x++;
       }
       page_tail(plot, type);
}

void get_base_morphology_distribution(_plot &plot, const double xlength, 
          const double ylength, const double l_margin, const double r_margin,
          const double b_margin, const double t_margin, int x_num, int y_num,
          const int intrlx, const char *id, TblIndexObj &_Index, _BOND **_values)
{
       int xgap = 6, ygap = 4, i = 0, j = 0, ibar = 0, nbars = 0;
       int *bcount = NULL, maxfrq = 0, ret = -1, ret1 = -1, x = 1, y = 1;
       double xmin, ymin, xmax, ymax, view_xt, view_yt;
       double view_xmin = 0, view_xmax = 0, view_ymin = 0, view_ymax = 0;
       double value_max = -1000.0, value_min = 1000.0, actual_ymax = 0;
       double actual_value_min = 1000.0, actual_value_max = -1000.0;
       double width  = 0, xmin1 = 0, xmax1 = 0, ymax1 = 0;
       CifString index;

       plot.set_axes_label_precision(intrlx, 0);
       page_base_morphology_head(plot, id);
       xmin = plot.page_xmin() + l_margin;
       ymin = plot.page_ymin() + b_margin;
       xmax = plot.page_xmax() - r_margin;
       ymax = plot.page_ymax() - t_margin;

       if (!x_num || !y_num) {
            x_num = (int) ((xmax - xmin) / xlength);
            y_num = (int) ((ymax - ymin) / ylength);
            if ((xmax - xmin - (double) x_num * xlength) < xlength / 2.0) x_num--;
            if ((ymax - ymin - (double) y_num * ylength) < ylength / 2.0) y_num--;
       }

       view_xt = (xmax - xmin - (double) x_num * xlength) / (double) x_num;
       view_yt = (ymax - ymin - (double) y_num * ylength) / (double) y_num;

       for (i = 0; i < NUM_OF_BASE_PARAM; i++) {
            index.Copy(base_params[i]);
            ret = _Index.Seek(index);
            if (ret < 0) continue;

            view_xmin = xmin + (double) x * view_xt +
                    (double) (x - 1) * xlength;
            view_xmax = view_xmin + xlength;
            if (view_xmax > xmax) {
                 x = 1;
                 y++;
                 view_xmin = xmin + (double) x * view_xt +
                             (double) (x - 1) * xlength;
                 view_xmax = view_xmin + xlength;
            }

            view_ymin = ymax - (double) y * (view_yt + ylength);
            view_ymax = view_ymin + ylength;
            if (view_ymin < ymin) {
                 plot.newpage(1);
                 page_base_morphology_head(plot, id);
                 x = y = 1;
                 view_xmin = xmin + (double) x * view_xt +
                             (double) (x - 1) * xlength;
                 view_xmax = view_xmin + xlength;
                 view_ymin = ymax - (double) y * (view_yt + ylength);
                 view_ymax = view_ymin + ylength;
            }

            value_max = -1000.0;
            value_min = 1000.0;
            ret1 = ret;
            while (ret1 >= 0) {
                 if (index != _Index.GetCurrData()) break;
                 if (_values[ret1]->val > value_max) value_max = _values[ret1]->val;
                 if (_values[ret1]->val < value_min) value_min = _values[ret1]->val;
                 ret1 = _Index.GoToNext();
            }
            actual_value_min = value_min - (value_max - value_min) / 4.0;
            actual_value_max = value_max + (value_max - value_min) / 4.0;

            width = (actual_value_max - actual_value_min) / 60.0;
            nbars = (int) ((actual_value_max - actual_value_min) / width + 0.1);
            bcount = new int[nbars];
            for (j = 0; j < nbars; j++) bcount[j] = 0;
            maxfrq = 0;
            ret1 = ret;
            _Index.GoToRecord(ret1);
            while (ret1 >= 0) {
                 if (index != _Index.GetCurrData()) break;
                 ibar = (int) ((_values[ret1]->val - actual_value_min) / width);
                 bcount[ibar]++;
                 if (bcount[ibar] > maxfrq) maxfrq = bcount[ibar];
                 ret1 = _Index.GoToNext();
            }
            
            j = 1; while (j * ygap < maxfrq) j++;
            actual_ymax = (double) ((j + 1) * ygap);

            plot.set_axes_interval(xgap, ygap);
            plot.set_actual_value_xmin_and_max(actual_value_min, actual_value_max);
            plot.set_actual_value_ymin_and_max(0.0, actual_ymax);
            plot.set_picture_title(base_params[i]);
            plot.set_x_axis_title("");
            plot.set_y_axis_title("Frequency");
            plot.set_view_xmin_and_max(view_xmin, view_xmax);
            plot.set_view_ymin_and_max(view_ymin, view_ymax);
            plot.draw_view_window(0, FONT_HELVETICA, 10.0, 8.0, 0.0, 20.0, 1);
            for (j = 0; j < nbars; j++) {
                 if (!bcount[j]) continue;
                 xmin1 = (double) j * width + actual_value_min;
                 xmax1 = xmin1 + width;
                 ymax1 = (double) bcount[j];
                 plot.plot_a_bar(xmin1, xmax1, 0.0, ymax1, 0.5, 0.0, 1.0);
            }
            delete [] bcount;
            x++;
       }
       page_base_morphology_tail(plot);
}

void ramachandran_plot(_plot &plot, const char *sid, const int _n_value, _TORSION **_values)
{
       int num_end_residues = 0, num_glycine = 0, num_proline = 0;
       int num_out = 0, num_core = 0, num_allow = 0, num_gener = 0;
       int i, regno, regtyp, type, num_sum_total, num_totals;
       double per_out, per_core, per_allow, per_gener, x, xpos1;
       double xstart = 50.0, xfpos1 = 300.0, xfpos2 = 340.0, ygap = 10.0;
       double xpos2, ystart, y, dy;
       char buffer[100], buffer1[100], region[4];
       COLOR rgb;

       plot.set_view_xmin_and_max(XRAMC1, XRAMC2);
       plot.set_view_ymin_and_max(YRAMC1, YRAMC2);

       plot.set_actual_value_xmin_and_max(-180, 180);
       plot.set_actual_value_ymin_and_max(-180, 180);

       float xcenter = (XRAMC1 + XRAMC2) / 2.0;
       float ycenter = (YRAMC1 + YRAMC2) / 2.0;
       plot.setfont(FONT_HELVETICA, 30, 0);
       plot.wrt_litstr_xy("Ramachandran Plot", xcenter, YRAMC2+40, CENTERX);
       plot.setfont(FONT_HELVETICA, 15, 0);
       plot.wrt_litstr_xy(sid, xcenter, YRAMC2+12, CENTERX);
       
       plot.set_picture_title("");
       plot.set_x_axis_title("Phi (degree)");
       plot.set_y_axis_title("Psi (degree)");

       plot.draw_view_window(0, FONT_HELVETICA, 15, 15, 0, 0, 0);

       ramfil(plot);
       ramlin(plot);
       ramlet(plot);

       plot.box(XRAMC1, YRAMC1, XRAMC2, YRAMC2, 0.0, 0.0, 0.0, 0);
       plot.line_style(LINE_DASH3);
       plot.line(xcenter, YRAMC1, xcenter, YRAMC2);
       plot.line(XRAMC1, ycenter, XRAMC2, ycenter);
       plot.line_style(LINE_SOLID);

       for (i = 0; i < _n_value; i++) {
/* NEED TO ADDED
            if (strcmp(_values[i]->ResName, "GLY") &&
                strcmp(_values[i]->ResName, "PRO")) num_end_residues++;
Original code:
                 for (j = 0; j < mol->chains[i].Num_Residue; j++) {
                      if ((j == 0 || j == mol->chains[i].Num_Residue-1) &&
                           strcmp(mol->chains[i].residues[j].ResName, "GLY") &&
                           strcmp(mol->chains[i].residues[j].ResName, "PRO"))
                           num_end_residues++;
*/
            if (!strcmp(_values[i]->ResName, "GLY")) num_glycine++;
            if (!strcmp(_values[i]->ResName, "PRO")) num_proline++;

            ramreg(_values[i]->val, _values[i]->ept, region, regno, regtyp);

            if (strcmp(_values[i]->ResName, "GLY") &&
                strcmp(_values[i]->ResName, "PRO")) {
                 if (!strcmp(region, "XX")) num_out++;
                 else if (!strcmp(region, "B") ||
                          !strcmp(region, "A") ||
                          !strcmp(region, "L")) num_core++;
                 else if (!strcmp(region, "b") ||
                          !strcmp(region, "a") ||
                          !strcmp(region, "l") ||
                          !strcmp(region, "p")) num_allow++;
                 else if (!strcmp(region, "~b") ||
                          !strcmp(region, "~a") ||
                          !strcmp(region, "~l") ||
                          !strcmp(region, "~p")) num_gener++;
            }

            if (strcmp(_values[i]->ResName, "GLY") &&
                regtyp <= 2) {
                 get_rgb_color_from_name("red", rgb);

                 sprintf(buffer, "%s %d (%s)", _values[i]->ResName,
                         _values[i]->ResNum, _values[i]->ChainID);
                 
                 dy = (YRAMC2 - YRAMC1) / 200.0;
                 x = XRAMC1 + (XRAMC2 - XRAMC1)
                          * (_values[i]->val + 180.0) / 360.0;
                 y = YRAMC1 + (YRAMC2 - YRAMC1)
                          * (_values[i]->ept + 180.0) / 360.0;

                 plot.setcolor(rgb.r, rgb.g, rgb.b);
                 plot.setfont(FONT_HELVETICA, 10, 0);
                 plot.wrt_litstr_xy(buffer, x, y + 3.5 * dy, CENTERX);
                 plot.setcolor(0, 0, 0);
            } else get_rgb_color_from_name("black", rgb);

            type = BOX;
            if (!strcmp(_values[i]->ResName, "GLY"))
                 type = TRIANGLE;

            plot.plot_a_point(_values[i]->val, _values[i]->ept, rgb.r, rgb.g,
                  rgb.b, type, 1, NULL);
       }

       num_sum_total = num_core + num_allow + num_gener + num_out;
       num_totals = num_sum_total + num_end_residues + num_glycine
                      + num_proline;
       per_out = (double) num_out / (double) num_sum_total * 100;
       per_core = (double) num_core / (double) num_sum_total * 100;
       per_allow = (double) num_allow / (double) num_sum_total * 100;
       per_gener = (double) num_gener / (double) num_sum_total * 100;

       x = XRAMC1 + xstart;
       xpos1 = xfpos1 + XRAMC1;
       xpos2 = xfpos2 + XRAMC1;
       ystart = YRAMC1 - 52.0;
       y = ystart;

       y -= 2.0 * ygap;
       plot.setfont(FONT_HELVETICA, 12, 0);
       plot.wrt_litstr_xy("Plot statistics", xcenter, y, CENTERX);
       y -= 2.0 * ygap;
  
       plot.setfont(FONT_HELVETICA, 9, 0);
       sprintf(buffer, "%6d", num_core);
       sprintf(buffer1, "%5.1f%%", per_core);
       ramput(plot, "Residues in most favoured regions  [A,B,L]",
               buffer, buffer1, x, xpos1, xpos2, y, ygap);

       sprintf(buffer, "%6d", num_allow);
       sprintf(buffer1, "%5.1f%%", per_allow);
       ramput(plot, "Residues in additional allowed regions  [a,b,l,p]",
               buffer, buffer1, x, xpos1, xpos2, y, ygap);

       sprintf(buffer, "%6d", num_gener);
       sprintf(buffer1, "%5.1f%%", per_gener);
       ramput(plot, "Residues in generously allowed regions  [~a,~b,~l,~p]",
               buffer, buffer1, x, xpos1, xpos2, y, ygap);

       sprintf(buffer, "%6d", num_out);
       sprintf(buffer1, "%5.1f%%", per_out);
       ramput(plot, "Residues in disallowed regions",
               buffer, buffer1, x, xpos1, xpos2, y, ygap);
       
       strcpy(buffer, "----");
       strcpy(buffer1, "------");
       ramput(plot, "", "----", "------", x, xpos1, xpos2, y, ygap);

       sprintf(buffer, "%6d", num_sum_total);
       strcpy(buffer1, "100.0%");
       ramput(plot, "Number of non-glycine and non-proline residues",
               buffer, buffer1, x, xpos1, xpos2, y, ygap);

       sprintf(buffer, "%6d", num_end_residues);
       ramput(plot, "Number of end-residues (excl. Gly and Pro)",
               buffer, "", x, xpos1, xpos2, y, ygap);

       sprintf(buffer, "%6d", num_glycine);
       ramput(plot, "Number of glycine residues (shown as triangles)",
               buffer, "", x, xpos1, xpos2, y, ygap);
 
       sprintf(buffer, "%6d", num_proline);
       ramput(plot, "Number of proline residues",
               buffer, "", x, xpos1, xpos2, y, ygap);

       ramput(plot, "", "----", "", x, xpos1, xpos2, y, ygap);

       sprintf(buffer, "%6d", num_totals);
       ramput(plot, "Total number of residues",
               buffer, "", x, xpos1, xpos2, y, ygap);

       y -= 2.0 * ygap;

       plot.setfont(FONT_HELVETICA, 8, 0);
       plot.wrt_litstr_xy("Based on an analysis of 118 structures of resolution of at least 2.0 Angstroms", 
             xcenter, y, CENTERX);
       y -= ygap;
       plot.wrt_litstr_xy("and R-factor no greater than 20%, a good quality model would be expected", 
             xcenter, y, CENTERX);
       y -= ygap;
       plot.wrt_litstr_xy("to have over 90% in the most favoured regions.",
             xcenter, y, CENTERX);
}

int is_torsion_out_of_regions(const float phi, const float psi)
{
       int regno, regtyp;
       char region[4];
       region[0] = '\0';
       ramreg(phi, psi, region, regno, regtyp);
       if (!strcmp(region, "XX")) return 1;
       return 0;
}

static void draw_a_view(_plot& plot, const int n_pair, _PAIR_VALUE **pair,
         const char *x_title, const char *y_title, int map[SIZE][SIZE])
{
       char buffer[80];
       sprintf(buffer, "%s", x_title);
       plot.set_picture_title("");
       plot.set_x_axis_title(buffer);
       sprintf(buffer, "%s", y_title);
       plot.set_y_axis_title(buffer);
       plot.draw_view_window(0, FONT_SYMBOL, 12.0, 7.0, 25.0, 30.0, 0);
       plot.project_map_on_view_window(map);
       
       for (int i = 0; i < n_pair; i++) {
            sprintf(buffer, "%d", pair[i]->res_num);
            plot.plot_a_point(pair[i]->x, pair[i]->y, 0.0, 0.0, 0.0, BOX, 0, buffer);
       }
}

static void page_head(_plot &plot, const char *id, const int type, const char base)
{
       char buffer[100];
       double xcenter = (plot.page_xmin() + plot.page_xmax()) / 2.0;
       double y_title = plot.page_ymax() - 33.0;
       double y_code  = plot.page_ymax() - 63.0;
       switch (type) {
           case SUGAR_BOND:
                plot.print_string("Sugar and Phosphate Bond Lengths",
                   FONT_HELVETICA, 30.0, xcenter, y_title, 
                   CENTERX);
                break;
           case BASE_BOND:
                switch(base) {
                    case 'A':
                        sprintf(buffer, "Adenine Bond Lengths");
                        break;
                    case 'C':
                        sprintf(buffer, "Cytosine Bond Lengths");
                        break;
                    case 'G':
                        sprintf(buffer, "Guanine Bond Lengths");
                        break;
                    case 'T':
                        sprintf(buffer, "Thymine Bond Lengths");
                        break;
                    case 'U':
                        sprintf(buffer, "Uracil Bond Lengths");
                        break;
                }
                // sprintf(buffer, "Base %c Bond Lengths", base);
                plot.print_string(buffer, // "Base Bond Lengths",
                   FONT_HELVETICA, 30.0, xcenter, y_title, 
                   CENTERX);
                break;
           case SUGAR_ANGLE:
                plot.print_string("Sugar and Phosphate Bond Angles",
                   FONT_HELVETICA, 30.0, xcenter, y_title, 
                   CENTERX);
                break;
           case BASE_ANGLE:
                switch(base) {
                    case 'A':
                        sprintf(buffer, "Adenine Bond Angles");
                        break;
                    case 'C':
                        sprintf(buffer, "Cytosine Bond Angles");
                        break;
                    case 'G':
                        sprintf(buffer, "Guanine Bond Angles");
                        break;
                    case 'T':
                        sprintf(buffer, "Thymine Bond Angles");
                        break;
                    case 'U':
                        sprintf(buffer, "Uracil Bond Angles");
                        break;
                }
                // sprintf(buffer, "Base %c Bond Angles", base);
                plot.print_string(buffer, // "Base Bond Angles",
                   FONT_HELVETICA, 30.0, xcenter, y_title, 
                   CENTERX);
                break;
           case PROTEIN_BOND:
                plot.print_string("Main-Chain Bond Lengths",
                   FONT_HELVETICA, 30.0, xcenter, y_title,
                   CENTERX);
                break;
           case PROTEIN_ANGLE:
                plot.print_string("Main-Chain Bond Angles",
                   FONT_HELVETICA, 30.0, xcenter, y_title,
                   CENTERX);
                break;
       }
       plot.print_string(id, FONT_HELVETICA, 20.0, xcenter,
               y_code, CENTERX);
}

static void page_tail(_plot &plot, const int type)
{
       char buffer[300];
       double x = plot.page_xmin() + 10.0;
       double y = plot.page_ymin() + 20.0;

       switch (type) {
           case SUGAR_BOND:
                strcpy(buffer, "Solid and dashed lines represent the mean and standard ");
                strcat(buffer, "deviation values as compilation of covalent geometries ");
                strcat(buffer, "for small");
                plot.print_string(buffer, FONT_HELVETICA, 10, x, y+15.0, LJUSTIFY);
                strcpy(buffer, "molecule crystal structures. [A.Gelbin et al., Geometric ");
                strcat(buffer, "Parameters in Nucleic Acids: Sugar and Phosphate ");
                strcat(buffer, "Constituents,"); 
                plot.print_string(buffer, FONT_HELVETICA, 10, x, y+5.0, LJUSTIFY);
                plot.print_string("J.Am.Chem.Soc. 1996, 118, 519-529]",
                    FONT_HELVETICA, 10, x, y-5.0, LJUSTIFY);
                break;
           case BASE_BOND:
                strcpy(buffer, "Solid and dashed lines represent the mean and standard ");
                strcat(buffer, "deviation values as compilation of covalent geometries ");
                strcat(buffer, "for small");
                plot.print_string(buffer, FONT_HELVETICA, 10, x, y+15.0, LJUSTIFY);
                strcpy(buffer, "molecule crystal structures. [L.Clowney et al., Geometric ");
                strcat(buffer, "Parameters in Nucleic Acids: Nitrogenous Bases, J.Am.");
                strcat(buffer, "Chem.");
                plot.print_string(buffer, FONT_HELVETICA, 10, x, y+5.0, LJUSTIFY);
                plot.print_string("Soc. 1996, 118, 509-518]", FONT_HELVETICA, 10, x, y-5.0,
                               LJUSTIFY);
                break;
           case SUGAR_ANGLE:
                strcpy(buffer, "Solid and dashed lines represent the mean and standard ");
                strcat(buffer, "deviation values as compilation of covalent geometries ");
                strcat(buffer, "for small");
                plot.print_string(buffer, FONT_HELVETICA, 10, x, y+15.0, LJUSTIFY);
                strcpy(buffer, "molecule crystal structures. [A.Gelbin et al., Geometric ");
                strcat(buffer, "Parameters in Nucleic Acids: Sugar and Phosphate ");
                strcat(buffer, "Constituents,"); 
                plot.print_string(buffer, FONT_HELVETICA, 10, x, y+5.0, LJUSTIFY);
                plot.print_string("J.Am.Chem.Soc. 1996, 118, 519-529]",
                    FONT_HELVETICA, 10, x, y-5.0, LJUSTIFY);
                break;
           case BASE_ANGLE:
                strcpy(buffer, "Solid and dashed lines represent the mean and standard ");
                strcat(buffer, "deviation values as compilation of covalent geometries ");
                strcat(buffer, "for small");
                plot.print_string(buffer, FONT_HELVETICA, 10, x, y+15.0, LJUSTIFY);
                strcpy(buffer, "molecule crystal structures. [L.Clowney et al., Geometric ");
                strcat(buffer, "Parameters in Nucleic Acids: Nitrogenous Bases, J.Am.");
                strcat(buffer, "Chem.");
                plot.print_string(buffer, FONT_HELVETICA, 10, x, y+5.0, LJUSTIFY);
                plot.print_string("Soc. 1996, 118, 509-518]", FONT_HELVETICA, 10, x, y-5.0,
                               LJUSTIFY);
                break;
           case PROTEIN_BOND:
           case PROTEIN_ANGLE:
                plot.print_string("Black bars > 2.0 st. devs. from mean.",
                       FONT_HELVETICA, 10, x, y+15.0, LJUSTIFY);
                strcpy(buffer, "Solid and dashed lines represent the mean and standard ");
                strcat(buffer, "deviation values as per Engh & Huber small-molecule data.");
                plot.print_string(buffer, FONT_HELVETICA, 10, x, y+5.0, LJUSTIFY);
                break;
       }
}

static void page_base_morphology_head(_plot &plot, const char *id)
{
       double xcenter = (plot.page_xmin() + plot.page_xmax()) / 2.0;
       double y_title = plot.page_ymax() - 33.0;
       double y_code  = plot.page_ymax() - 63.0;
       plot.print_string("Base Morphology Parameters", FONT_HELVETICA,
                  30.0, xcenter, y_title, CENTERX);
       plot.print_string(id, FONT_HELVETICA, 20.0, xcenter,
               y_code, CENTERX);
}

static void page_base_morphology_tail(_plot &plot)
{
       char buffer[300];

       double x = plot.page_xmin() + 10.0;
       double y = plot.page_ymin() + 20.0;
       strcpy(buffer, "R. Lavery and H. Sklenar, The Definition of Generalised ");
       strcat(buffer, "Helicoidal Parameters and of Axis Curvature in Irregular ");
       strcat(buffer, "Nucleic");
       plot.print_string(buffer, FONT_HELVETICA, 10, x, y+15.0, LJUSTIFY);
       strcpy(buffer, "Acids. Biomol. Struct. Dynam., 1988, 6, 63-91.");
       plot.print_string(buffer, FONT_HELVETICA, 10, x, y+5.0, LJUSTIFY);
}

static void ramfil(_plot &plot)
{
       static const char *colram[] = {"white", "white", "cream", "yellow",
                      "red", "black", "red"};
       static float factor[12] = {1.0, 0.95, 1.0, 0.95, 0.95, 1.0, 1.0,
                       1.0, 1.0, 0.95, 0.95, 0.95};
       static int rcol[12] = {2, 5, 4, 5, 4, 5, 4, 4, 3, 3, 3, 3};
       int i, j, ireg, regtyp;
       float colour[3][12], pgap, xgap, ygap, x, y, psi, phi;
       char region[4];
       COLOR rgb;

       pgap = 360.0 / MAPSIZ;
       xgap = (plot.view_xmax() - plot.view_xmin()) / MAPSIZ;
       ygap = (plot.view_ymax() - plot.view_ymin()) / MAPSIZ;

       for (i = 0; i < 12; i++) {
            get_rgb_color_from_name(colram[rcol[i]-1], rgb);
            colour[0][i] = factor[i] * rgb.r;
            colour[1][i] = factor[i] * rgb.g;
            colour[2][i] = factor[i] * rgb.b;
       }

       plot.setline_width(0);

       y = plot.view_ymin();
       psi = -180.0 + pgap / 2.0;
       for (i = 0; i < MAPSIZ; i++) {
            x = plot.view_xmin();
            phi = -180.0 + pgap / 2.0;
            for (j = 0; j < MAPSIZ; j++) {
                 ramreg(phi, psi, region, ireg, regtyp);
                 plot.box(x, y, x+xgap, y+ygap, colour[0][ireg-1],
                          colour[1][ireg-1], colour[2][ireg-1], 1);
                 x += xgap;
                 phi += pgap;
            }
            y += ygap;
            psi += pgap;
       }

       plot.setline_width(0.2);
}

static void ramlin(_plot &plot)
{
       int ltype[MAPSIZ], i, j, ireg, regtyp, itype;
       float pgap, xgap, ygap, x, y, phi, psi;
       char region[4];

       for (i = 0; i < MAPSIZ; i++) ltype[i] = 0;

       plot.setline_width(0.2);

       pgap = 360.0 / MAPSIZ;
       xgap = (plot.view_xmax() - plot.view_xmin()) / MAPSIZ;
       ygap = (plot.view_ymax() - plot.view_ymin()) / MAPSIZ;

       y = plot.view_ymin();
       psi = -180.0 + pgap / 2.0;
       for (i = 0; i < MAPSIZ; i++) {
            x = plot.view_xmin();
            phi = -180.0 + pgap / 2.0;
            for (j = 0; j < MAPSIZ; j++) {
                 ramreg(phi, psi, region, ireg, regtyp);
                 itype = regtyp;
                 if (j && itype != ltype[j-1]) {
                      if (itype < ltype[j-1]) {
                           plot.line(x, y, x, y + ygap);
                      }
                 }

                 if (i) {
                      if (itype != ltype[j]) {
                           if (itype > ltype[j]) {
                                plot.line(x, y, x + xgap, y);
                           }
                      }
                 }

                 ltype[j] = itype;
                 x += xgap;
                 phi += pgap;
            }
            y += ygap;
            psi += pgap;
       }
       plot.setline_width(0.2);
}

static void ramlet(_plot &plot)
{
       float xgap = (plot.view_xmax() - plot.view_xmin()) / MAPSIZ;
       float ygap = (plot.view_ymax() - plot.view_ymin()) / MAPSIZ;
       float xmin = plot.view_xmin();
       float ymax = plot.view_ymax();

       plot.setfont(FONT_HELVETICA, 15.0, 0.0);
       plot.wrt_litstr_xy("B", xmin +  2.0 * xgap, ymax -  1.0 * ygap,
                          CENTERX);
       plot.wrt_litstr_xy("A", xmin +  7.0 * xgap, ymax - 17.0 * ygap,
                          CENTERX);
       plot.wrt_litstr_xy("b", xmin +  1.0 * xgap, ymax -  5.0 * ygap,
                          CENTERX);
       plot.wrt_litstr_xy("a", xmin +  2.0 * xgap, ymax - 15.0 * ygap,
                          CENTERX);
       plot.setfont(FONT_HELVETICA, 10.0, 0.0);
       plot.wrt_litstr_xy("L", xmin + 23.5 * xgap, ymax - 14.0 * ygap,
                          CENTERX);
       plot.wrt_litstr_xy("l", xmin + 21.5 * xgap, ymax - 10.0 * ygap,
                          CENTERX);
       plot.wrt_litstr_xy("p", xmin + 23.5 * xgap, ymax - 33.0 * ygap,
                          CENTERX);
       plot.wrt_litstr_xy("~p", xmin + 21.5 * xgap, ymax - 31.0 * ygap,
                          CENTERX);
       plot.wrt_litstr_xy("~b", xmin + 16.5 * xgap, ymax -  5.0 * ygap,
                          CENTERX);
       plot.wrt_litstr_xy("~a", xmin + 16.5 * xgap, ymax - 19.5 * ygap,
                          CENTERX);
       plot.wrt_litstr_xy("~l", xmin + 22.5 * xgap, ymax -  7.0 * ygap,
                          CENTERX);
       plot.wrt_litstr_xy("b", xmin +  2.0 * xgap, ymax - 33.0 * ygap,
                          CENTERX);
       plot.wrt_litstr_xy("~b", xmin +  0.5 * xgap, ymax - 31.0 * ygap,
                          CENTERX);
       plot.wrt_litstr_xy("b", xmin + 35.0 * xgap, ymax -  2.0 * ygap,
                          CENTERX);
       plot.wrt_litstr_xy("~b", xmin + 31.5 * xgap, ymax -  1.0 * ygap,
                          CENTERX);
       plot.wrt_litstr_xy("~b", xmin + 33.5 * xgap, ymax - 34.0 * ygap,
                          CENTERX);
}

static void ramreg(const float phi, const float psi, char *region, int &regno, int &regtyp)
{
       static int first = 1, rammap[MAPSIZ][MAPSIZ];
       static char code[] =  "oBbAaLlexyzw";
       static char newcod[] = "XXB b A a L l p ~b~a~l~p";
       static int rtype[12] = {1, 4, 3, 4, 3, 4, 3, 3, 2, 2, 2, 2};

       static char mapstr[MAPSIZ][MAPSIZ+1] = {
              "bBBBBBBBBBBbbbxxooooowwwwwoooooxxxxb",
              "bBBBBBBBBBBBbbbxxooooooooooooooxxxbb",
              "bBBBBBBBBBBBbbbxxxxooooooooooooxxbbb",
              "bbBBBBBBBBBBBbbbbxxooooooooooooxxxbb",
              "bbBBBBBBBBBBBBbbxxxooooooooooooxxxxb",
              "bbBBBBBBBBBBBbbbxxxoooooooooooooxxxb",
              "bbbBBBBBBBBBbbbbxxxooozzzzzzoooooxxb",
              "bbbbBBBBBBBbbbbbbxxzzzzzzzzzoooooxxb",
              "bbbbbBbbBbbbbbbbbxxzzzzzllzzoooooxxb",
              "bbbbbbbbbbbbbbxxxxxzzllllzzzoooooxxx",
              "bbbbbbbbbbbbxxxxxxxzzllllzzzoooooxxx",
              "xbbbbbbbbbbbbxxoooozzzlllzzzzooooxxx",
              "xxbbbbbbbbbbbxxoooozzllllllzzooooxxx",
              "yyaaaaaaaaaayyyoooozzllLllzzzooooxxx",
              "yaaaaaaaaaaayyyoooozzzlLLlzzzooooxxx",
              "yaaaaaaAaaaaayyyooozzzlllllzzooooxxx",
              "yaaaaaAAAAaaayyyyooozzzlllzzzooooxxx",
              "yaaaaaAAAAAaaayyyooozzzzlllzzooooxxx",
              "yaaaaAAAAAAAaaayyyoozzzlzllzzooooxxx",
              "yaaaaaAAAAAAAaayyyyozzzzzzzzzooooxxx",
              "yyaaaaAAAAAAAAaayyyozzzzzzzzzooooxxx",
              "yyaaaaaAAAAAAAaaayyyoooooooooooooxxx",
              "yyyaaaaaAAAAAAAaayyyoooooooooooooxxx",
              "yaaaaaaaaAAAAAAaaayyyooooooooooooxxx",
              "aayaaaaaaaaAAAAaaayyyooooooooooooxxx",
              "yyyaaaaaaaaaaaaaaaayyooooooooooooxxx",
              "yyyyyaaaaaaaaaaaaayyyooooooooooooxxx",
              "oyyyyaaaaaaaaaayyyyyyooooooooooooooo",
              "oooyyyyyyyyayyyyyyyyoooooooooooooooo",
              "oooxxxxbbxxxxxxxxooooooooooooooooooo",
              "xxxxxbbxxxxxxxooooooowwwwwoooooooooo",
              "xxxxxxbbbbxxxxooooooowwwwwoooooooooo",
              "xbbbxbbbbbxxxxxoooooowwewwwooooooooo",
              "xbbbbbbbbbbxxxxoooooowwewwwooooooxxx",
              "xbbbbbbbbbbbbxxoooooowweewwooooooxxx",
              "bbbbbbbbbbbbbxxoooooowwwwwwooooooxxb"};

       int i, j;

       if (first) {
            first = 0;
            for (i = 0; i < MAPSIZ; i++) {
                 for (j = 0; j < MAPSIZ; j++) {
                      char ch = mapstr[i][j];
                      for (int k = 0; k < 12; k++)
                           if (ch == code[k]) 
                                rammap[MAPSIZ-i-1][j] = k + 1;
                 }
            }
       }
       float gap = 360.0 / MAPSIZ;

       if (phi > 180.0 || psi > 180.0) {
            regno = 1;
            regtyp = 0;
            strcpy(region, "XX");
       } else {
            i = (int) ((psi + 180.0) / gap);
            if (i < 0) i = 0;
            if (i > MAPSIZ - 1) i = MAPSIZ - 1;
            j = (int) ((phi + 180.0) / gap);
            if (j < 0) j = 0;
            if (j > MAPSIZ - 1) j = MAPSIZ - 1;
            regno = rammap[i][j];
            for (int k = 0; k < 2; k++)
                 region[k] = newcod[2 * (regno - 1) + k];
            region[2] = '\0';
            ndb_clean_string(region);
            regtyp = rtype[regno - 1];
       }
}

static void ramput(_plot &plot, const char *str1, const char *str2, const char *str3,
        const double x, const double x1, const double x2, double &y, const double ygap)
{
       if (strcmp(str1, ""))
            plot.wrt_litstr_xy(str1, x, y, LJUSTIFY);
       if (strcmp(str2, ""))
            plot.wrt_litstr_xy(str2, x1, y, CENTERX);
       if (strcmp(str3, ""))
            plot.wrt_litstr_xy(str3, x2, y, CENTERX);
       y -= ygap;
}
