/*
FILE:     ps_util.C
*/
/*
VERSION:  1.700
*/
/*
DATE:     7/13/2005
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2005 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <iostream.h>
#include <assert.h>

#include "_validation.h"
#include "ps_util.h"
#include "misc_util.h"

typedef struct {
        int    code;
        int    slot;
        double depth;
} _DEPTH;

typedef struct {
        COORD  p1, p2, p3, p4;
        int    col;
        double znorm;
} _PLANE;

typedef struct {
        COORD  v1, v2;
        int    col;
        int    style;
} _LINE;

typedef struct {
        double xlo, xhi, ylo, yhi, zlo, zhi;
} EXTENT;

static _DEPTH *depth_array = NULL;
static int depth_alloc = 0;
static int depth_count = 0;

static _PLANE *plane_array = NULL;
static int plane_alloc = 0;
static int plane_count = 0;

static _LINE *line_array = NULL;
static int line_alloc = 0;
static int line_count = 0;

static EXTENT curr_ext;

static double slab;

static void draw_init(void);
static void draw_ribbons(const int _n_ribbons, _RIBBONS **_ribbons, const double m[4][4]);
static void draw_bonds(const int _n_link_bonds, _LINK_BOND **_link_bonds, const double m[4][4]);
static void draw_blocks(const int _n_blocks, _BLOCK **_blocks, const double m[4][4]);
static void draw_plane_1(const COORD p1, const COORD p2, const COORD p3, const COORD p4,
                    const int col);
static void draw_crystal(COORD *base);
static void write_postscripts(_plot &plot, const double area[4], COORD abc[3]);
static void draw_plane(const COORD p1, const COORD p2, const COORD p3, const COORD p4,
                const int c, const double znorm, const int code);
static void draw_line(COORD &v1, COORD &v2, const int colour, const int style);
static void plot_init(void);
static void output_postscript(_plot &plot, const double window, const double area[4],COORD abc[3]);
static void enter_depth(const int code, const int slot, const double depth);

#ifdef SUN_OS
#ifdef __cplusplus
extern "C" {
#endif
#endif
static int depth_compare(const void *de1, const void *de2);
#ifdef SUN_OS
#ifdef __cplusplus
}
#endif
#endif

static void output_line_params(_plot &plot, const int color_index, const double depth);
static double shade_factor(const double znorm);
static double depthcue(const double depth);
static double plane_znorm(COORD &p1, COORD &p2, COORD &p3, COORD &p4,
              double znorm, const int col);

void _VALIDATION::draw_ps_picture(_plot &plot, const double rot[4][4], COORD *base,
               const double area[4], COORD abc[3])
{
       draw_init();
       draw_ribbons(_n_ribbons, _ribbons, rot);
       draw_bonds(_n_link_bonds, _link_bonds, rot);
       if (_n_blocks) draw_blocks(_n_blocks, _blocks, rot);
       if (base) draw_crystal(base);
       write_postscripts(plot, area, abc);
}

void extent_init(void)
{
       curr_ext.xlo = 1.0e20;
       curr_ext.xhi = -1.0e20;
       curr_ext.ylo = 1.0e20;
       curr_ext.yhi = -1.0e20;
       curr_ext.zlo = 1.0e20;
       curr_ext.zhi = -1.0e20;
}

void extent_update(const COORD p, const double radius)
{
       double u = p.x - radius;
       if (u < curr_ext.xlo) curr_ext.xlo = u;
       u = p.x + radius;
       if (u > curr_ext.xhi) curr_ext.xhi = u;
       u = p.y - radius;
       if (u < curr_ext.ylo) curr_ext.ylo = u;
       u = p.y + radius;
       if (u > curr_ext.yhi) curr_ext.yhi = u;
       u = p.z - radius;
       if (u < curr_ext.zlo) curr_ext.zlo = u;
       u = p.z + radius;
       if (u > curr_ext.zhi) curr_ext.zhi = u;
}

void get_extent(COORD &center, COORD &size)
{
       center.x = 0.5 * (curr_ext.xlo + curr_ext.xhi);
       center.y = 0.5 * (curr_ext.ylo + curr_ext.yhi);
       center.z = 0.5 * (curr_ext.zlo + curr_ext.zhi);
       size.x = curr_ext.xhi - curr_ext.xlo;
       size.y = curr_ext.yhi - curr_ext.ylo;
       size.z = curr_ext.zhi - curr_ext.zlo;
}

void rotation_ribbon(_RIBBON &ribbon, const double m[4][4])
{
       COORD pos, normal, side;

       vector_copy(pos, ribbon.c);
       vector_copy(normal, ribbon.n);
       vector_copy(side, ribbon.w);
       matrix_rotate(pos, m);
       matrix_rotate(normal, m);
       matrix_rotate(side, m);

       vector_sum_scaled(ribbon.p1, pos, RIBBON_W, side);
       vector_add_scaled(ribbon.p1, RIBBON_H, normal);
       vector_sum_scaled(ribbon.p2, pos, RIBBON_W, side);
       vector_add_scaled(ribbon.p2, -RIBBON_H, normal);
       vector_sum_scaled(ribbon.p3, pos, -RIBBON_W, side);
       vector_add_scaled(ribbon.p3, -RIBBON_H, normal);
       vector_sum_scaled(ribbon.p4, pos, -RIBBON_W, side);
       vector_add_scaled(ribbon.p4, RIBBON_H, normal);
       vector_copy(ribbon.n1, normal);
       vector_copy(ribbon.n2, side);
       vector_scaled(ribbon.n3, -1.0, normal);
       vector_scaled(ribbon.n4, -1.0, side);

       extent_update(pos, RIBBON_W);
}

static void draw_init(void)
{
       plot_init();
       extent_init();
}

static void draw_ribbons(const int _n_ribbons, _RIBBONS **_ribbons, const double m[4][4])
{
       int i, k, col, code, n_ribbon, thickness;
       COORD normal, dir1, dir2;
       _RIBBON *ribbon = NULL;

       thickness = RIBBON_H > 0.01;
       for (k = 0; k < _n_ribbons; k++) {
            for (i = 0; i < _ribbons[k]->n_ribbon; i++)
                 rotation_ribbon(_ribbons[k]->ribbon[i], m);
     
            n_ribbon = _ribbons[k]->n_ribbon;
            if (n_ribbon <= 0) continue;
            ribbon = _ribbons[k]->ribbon;
            for (i = 0; i < n_ribbon - 1; i++) {
                 col = ribbon[i].index;
                 code = P13_CODE;
                 if (!i) code = P134_CODE;
                 vector_sum(normal, ribbon[i].n1, ribbon[i+1].n1);
                 vector_normalize(normal);
                 if (thickness) {
                      draw_plane(ribbon[i].p1, ribbon[i+1].p1,
                                 ribbon[i+1].p4, ribbon[i].p4,
                            col, plane_znorm(ribbon[i].p1, ribbon[i+1].p1,
                            ribbon[i+1].p4, ribbon[i].p4, normal.z, col), code);
                      draw_plane(ribbon[i].p2, ribbon[i+1].p2,
                                 ribbon[i+1].p3, ribbon[i].p3,
                            col, plane_znorm(ribbon[i].p2, ribbon[i+1].p2,
                            ribbon[i+1].p3, ribbon[i].p3, -normal.z, col), code);
                 } else draw_plane(ribbon[i].p1, ribbon[i+1].p1,
                                   ribbon[i+1].p4, ribbon[i].p4,
                                        col, fabs(normal.z), code);
            }
     
            if (thickness) {
                 vector_difference(dir1, ribbon[0].p3, ribbon[0].p2);
                 vector_difference(dir2, ribbon[0].p1, ribbon[0].p2);
                 vector_cross_product(normal, dir1, dir2);
                 vector_normalize(normal);
     
                 col = ribbon[0].index;
                 draw_plane(ribbon[0].p1, ribbon[0].p2, ribbon[0].p3,
                           ribbon[0].p4, col, normal.z, P1234_CODE);
                 draw_line(ribbon[0].p1, ribbon[0].p2, col, LINE_SOLID);
                 draw_line(ribbon[0].p2, ribbon[0].p3, col, LINE_SOLID);
                 draw_line(ribbon[0].p3, ribbon[0].p4, col, LINE_SOLID);
                 draw_line(ribbon[0].p4, ribbon[0].p1, col, LINE_SOLID);
                 
                 for (i = 0; i < n_ribbon - 1; i++) {
                      col = ribbon[i].index;
                      code = P13_CODE;
                      if (!i) code = P134_CODE;
                      vector_sum(normal, ribbon[i].n2, ribbon[i+1].n2);
                      vector_normalize(normal);
                      draw_plane(ribbon[i].p1, ribbon[i+1].p1,
                                  ribbon[i+1].p2, ribbon[i].p2,
                             col, plane_znorm(ribbon[i].p1, ribbon[i+1].p1,
                             ribbon[i+1].p2, ribbon[i].p2,normal.z, col), code);
                      draw_plane(ribbon[i].p4, ribbon[i+1].p4,
                                 ribbon[i+1].p3, ribbon[i].p3,
                             col, plane_znorm(ribbon[i].p4, ribbon[i+1].p4,
                             ribbon[i+1].p3,ribbon[i].p3,-normal.z,col),code);
                 }
     
                 vector_difference(dir1, ribbon[n_ribbon-1].p3,
                                         ribbon[n_ribbon-1].p2);
                 vector_difference(dir2, ribbon[n_ribbon-1].p1,
                                         ribbon[n_ribbon-1].p2);
                 vector_cross_product(normal, dir2, dir1);
                 vector_normalize(normal);
                 draw_plane(ribbon[n_ribbon-1].p1,
                            ribbon[n_ribbon-1].p2,
                            ribbon[n_ribbon-1].p3,
                            ribbon[n_ribbon-1].p4,
                            col, normal.z, P1234_CODE);
                 draw_line(ribbon[n_ribbon-1].p1,
                           ribbon[n_ribbon-1].p2, col, LINE_SOLID);
                 draw_line(ribbon[n_ribbon-1].p2,
                           ribbon[n_ribbon-1].p3, col, LINE_SOLID);
                 draw_line(ribbon[n_ribbon-1].p3,
                           ribbon[n_ribbon-1].p4, col, LINE_SOLID);
                 draw_line(ribbon[n_ribbon-1].p4,
                           ribbon[n_ribbon-1].p1, col, LINE_SOLID);
            }
       }
}

static void draw_bonds(const int _n_link_bonds, _LINK_BOND **_link_bonds, const double m[4][4])
{
       int i;
       COORD tc1, tc2, middle;
       for (i = 0; i < _n_link_bonds; i++) {
            tc1.x = _link_bonds[i]->a->orig().x;
            tc1.y = _link_bonds[i]->a->orig().y;
            tc1.z = _link_bonds[i]->a->orig().z;
            tc2.x = _link_bonds[i]->b->orig().x;
            tc2.y = _link_bonds[i]->b->orig().y;
            tc2.z = _link_bonds[i]->b->orig().z;
            matrix_rotate(tc1, m);
            extent_update(tc1, 0);
            matrix_rotate(tc2, m);
            extent_update(tc2, 0);
            vector_middle(middle, tc1, tc2);
            draw_line(tc1, middle, _link_bonds[i]->color_a, LINE_SOLID);
            draw_line(middle, tc2, _link_bonds[i]->color_b, LINE_SOLID);
       }
}

static void draw_blocks(const int _n_blocks, _BLOCK **_blocks, const double m[4][4])
{
       int i, k;
       COORD coord[8];
       for (k = 0; k < _n_blocks; k++) {
            for (i = 0; i < 8; i++) {
                 coord[i] = _blocks[k]->coord[i];
                 matrix_rotate(coord[i], m);
            }
            draw_plane_1(coord[0], coord[1], coord[2], coord[3], _blocks[k]->index);
            draw_plane_1(coord[5], coord[4], coord[7], coord[6], _blocks[k]->index);
            draw_plane_1(coord[4], coord[0], coord[3], coord[7], _blocks[k]->index);
            draw_plane_1(coord[1], coord[5], coord[6], coord[2], _blocks[k]->index);
            draw_plane_1(coord[0], coord[1], coord[5], coord[4], _blocks[k]->index);
            draw_plane_1(coord[7], coord[6], coord[2], coord[3], _blocks[k]->index);
       }
}

static void draw_plane_1(const COORD p1, const COORD p2, const COORD p3,
             const  COORD p4, const int col)
{
       COORD dir1, dir2, normal;
       vector_difference(dir1, p1, p2);
       vector_difference(dir2, p3, p2);
       vector_cross_product(normal, dir1, dir2);
       vector_normalize(normal);

       draw_plane(p1, p2, p3, p4, col, fabs(normal.z), P134_CODE);
}

static void draw_crystal(COORD *base)
{
       int i, index[8], style;
       _DEPTH depth[8];

       for (i = 0; i < 8; i++) {
            depth[i].depth = base[i].z;
            depth[i].code = i;
       }
       qsort(depth, 8, sizeof(_DEPTH), depth_compare);
       for (i = 0; i < 8; i++) {
            if (i < 4) index[depth[i].code] = 0;
            else       index[depth[i].code] = 1;
       }

       style = LINE_SOLID;
       if (index[0] && index[1])
            style = LINE_SOLID;
       else style = LINE_DASH3;
       draw_line(base[0], base[1], COLOR_BLACK, style);

       if (index[1] && index[2])
            style = LINE_SOLID;
       else style = LINE_DASH3;
       draw_line(base[1], base[2], COLOR_BLACK, style);

       if (index[2] && index[3])
            style = LINE_SOLID;
       else style = LINE_DASH3;
       draw_line(base[2], base[3], COLOR_BLACK, style);

       if (index[3] && index[0])
            style = LINE_SOLID;
       else style = LINE_DASH3;
       draw_line(base[3], base[0], COLOR_BLACK, style);

       if (index[4] && index[5])
            style = LINE_SOLID;
       else style = LINE_DASH3;
       draw_line(base[4], base[5], COLOR_BLACK, style);

       if (index[5] && index[6])
            style = LINE_SOLID;
       else style = LINE_DASH3;
       draw_line(base[5], base[6], COLOR_BLACK, style);

       if (index[6] && index[7])
            style = LINE_SOLID;
       else style = LINE_DASH3;
       draw_line(base[6], base[7], COLOR_BLACK, style);

       if (index[7] && index[4])
            style = LINE_SOLID;
       else style = LINE_DASH3;
       draw_line(base[7], base[4], COLOR_BLACK, style);

       if (index[0] && index[4])
            style = LINE_SOLID;
       else style = LINE_DASH3;
       draw_line(base[0], base[4], COLOR_BLACK, style);

       if (index[1] && index[5])
            style = LINE_SOLID;
       else style = LINE_DASH3;
       draw_line(base[1], base[5], COLOR_BLACK, style);

       if (index[2] && index[6])
            style = LINE_SOLID;
       else style = LINE_DASH3;
       draw_line(base[2], base[6], COLOR_BLACK, style);

       if (index[3] && index[7])
            style = LINE_SOLID;
       else style = LINE_DASH3;
       draw_line(base[3], base[7], COLOR_BLACK, style);
}

static void write_postscripts(_plot &plot, const double area[4], COORD abc[3])
{
       int i;
       double t_m[4][4], center_x, center_y, center_z, aspect_ratio;
       double size_x, size_y, size_z, window, d;
       _LINE *le = NULL;
       _PLANE *pe = NULL;

       center_x = 0.5 * (curr_ext.xlo + curr_ext.xhi);
       center_y = 0.5 * (curr_ext.ylo + curr_ext.yhi);
       center_z = 0.5 * (curr_ext.zlo + curr_ext.zhi);

       matrix_init(t_m);
       matrix_translation(t_m, -center_x, -center_y, -center_z);

       extent_init();
       le = line_array;
       for (i = 0; i < line_count; i++, le++) {
            matrix_transform(le->v1, t_m);
            extent_update(le->v1, 0);
            matrix_transform(le->v2, t_m);
            extent_update(le->v2, 0);
       }

       pe = plane_array;
       for (i = 0; i < plane_count; i++, pe++) {
            matrix_transform(pe->p1, t_m);
            extent_update(pe->p1, 0);
            matrix_transform(pe->p2, t_m);
            extent_update(pe->p2, 0);
            matrix_transform(pe->p3, t_m);
            extent_update(pe->p3, 0);
            matrix_transform(pe->p4, t_m);
            extent_update(pe->p4, 0);
       }

       center_x = 0.5 * (curr_ext.xlo + curr_ext.xhi);
       center_y = 0.5 * (curr_ext.ylo + curr_ext.yhi);
       center_z = 0.5 * (curr_ext.zlo + curr_ext.zhi);

       size_x = 0.5 * (curr_ext.xhi - curr_ext.xlo);
       size_y = 0.5 * (curr_ext.yhi - curr_ext.ylo);
       size_z = 0.5 * (curr_ext.zhi - curr_ext.zlo);

       aspect_ratio = (area[3] - area[1]) / (area[2] - area[0]);

       if (aspect_ratio < 1.0) size_x *= aspect_ratio;
       else                    size_y /= aspect_ratio;
       window = fabs(center_x - size_x);
       d = fabs(center_x + size_x);
       if (d > window) window = d;
       d = fabs(center_y - size_y);
       if (d > window) window = d;
       d = fabs(center_y + size_y);
       if (d > window) window = d;
       window += 2.0;

       slab = fabs(center_z - size_z);
       d = fabs(center_z + size_z);
       if (d > slab) slab = d;

       output_postscript(plot, window, area, abc);
       plot.setcolor(0.0, 0.0, 0.0);
}

static void draw_plane(const COORD p1, const COORD p2, const COORD p3, const COORD p4,
             const int c, const double znorm, const int code)
{
       double len1, len2, depth;
       COORD vec;
       _PLANE *pe = NULL;
       
       assert(znorm >= -1.0);
       assert(znorm <= 1.0);
       if (znorm < 0.0) return;

       if (plane_count >= plane_alloc) {
            plane_alloc += 1000;
            plane_array = (_PLANE *) realloc(plane_array, 
                         plane_alloc * sizeof(_PLANE));
       }

       pe = plane_array + plane_count;
       pe->znorm = znorm;
       pe->col = c;
       vector_copy(pe->p1, p1);
       vector_copy(pe->p2, p2);

       vector_difference(vec, p1, p4);
       vec.z = 0.0;
       len1 = vector_length(vec);
       vector_difference(vec,  p2, p3);
       vec.z = 0.0;
       len1 += vector_length(vec);

       vector_difference(vec, p1, p3);
       vec.z = 0.0;
       len2 = vector_length(vec);
       vector_difference(vec,  p2, p4);
       vec.z = 0.0;
       len2 += vector_length(vec);

       if (len1 < len2) {
            vector_copy(pe->p3, p3);
            vector_copy(pe->p4, p4);
       } else {
            vector_copy(pe->p3, p4);
            vector_copy(pe->p4, p3);
       }

       depth = (p1.z > p2.z) ? p1.z : p2.z;
       if (p3.z > depth) depth = p3.z;
       if (p4.z > depth) depth = p4.z;

       enter_depth(code, plane_count++, depth);
}

static void draw_line(COORD &v1, COORD &v2, const int colour, const int style)
{

       if (vector_distance(v1, v2) == 0.0) return;

       if (vector_distance(v1, v2) > 2.0) {
            COORD middle;
            vector_middle(middle, v1, v2);
            draw_line(v1, middle, colour, style);
            draw_line(middle, v2, colour, style);
       } else {
            if (line_count >= line_alloc) {
                 line_alloc += 1000;
                 line_array = (_LINE *) realloc(line_array, 
                       line_alloc * sizeof(_LINE));
            }
     
            _LINE *le = line_array + line_count;
            vector_copy(le->v1, v1);
            vector_copy(le->v2, v2);
            le->col = colour;
            le->style = style;
     
            enter_depth(LINE_CODE, line_count++, (v1.z + v2.z) / 2.0);
       }
}

static void plot_init(void)
{
       if (depth_alloc == 0) {
            depth_alloc = 16384;
            depth_array = (_DEPTH *) malloc(depth_alloc * sizeof(_DEPTH));
       }
       depth_count = 0;

       if (plane_alloc == 0) {
            plane_alloc = 4096;
            plane_array = (_PLANE *) malloc(plane_alloc * sizeof(_PLANE));
       }
       plane_count = 0;

       if (line_alloc == 0) {
            line_alloc = 2048;
            line_array = (_LINE *) malloc (line_alloc * sizeof(_LINE));
       }
       line_count = 0;
}

static void output_postscript(_plot &plot, const double window, const double area[4],
             COORD abc[3])
{
       int i;
       double matrix[4][4], temp_matrix[4][4], scale;
       COLOR col;
       _LINE *le = NULL;
       _PLANE *pe = NULL;
       _DEPTH *de = NULL;

       matrix_orthographic_projection(matrix, -window, window,
                 -window, window, -1.0, 1.0);
 
       if (area[2] - area[0] < area[3] - area[1]) 
            scale = (area[2] - area[0]) / 2.0;
       else scale = (area[3] - area[1]) / 2.0;
 
       matrix_scale(temp_matrix, scale, scale, 1.0);
       matrix_concatenate(matrix, temp_matrix);
       matrix_translation(temp_matrix, (area[0] + area[2]) / 2.0,
                             (area[1] + area[3]) / 2.0, 0.0);
       matrix_concatenate(matrix, temp_matrix);

       for (i = 0; i < 3; i++) matrix_transform(abc[i], matrix);

       if (depth_count > 0)
            qsort(depth_array, depth_count, sizeof(_DEPTH), depth_compare);
 
       le = line_array;
       for (i = 0; i < line_count; i++, le++) {
            matrix_transform(le->v1, matrix);
            matrix_transform(le->v2, matrix);
       }
 
       pe = plane_array;
       for (i = 0; i < plane_count; i++, pe++) {
            matrix_transform(pe->p1, matrix);
            matrix_transform(pe->p2, matrix);
            matrix_transform(pe->p3, matrix);
            matrix_transform(pe->p4, matrix);
       }
 
       for (i = 0; i < depth_count; i++) {
            de = depth_array + i;
            switch (de->code) {
                 case LINE_CODE:
                      le = line_array + de->slot;
                      output_line_params(plot, le->col, de->depth);
                      plot.line_style(le->style);
                      plot.line(le->v1.x, le->v1.y, le->v2.x, le->v2.y);
                      break;
                 case P13_CODE:
                      pe = plane_array + de->slot;
                      output_line_params(plot, pe->col, de->depth);
                      get_rgb_color_from_index(pe->col, col);
                      color_darker(col, shade_factor(pe->znorm), col);
                      plot.box(col, pe->p4.x, pe->p4.y, pe->p3.x, pe->p3.y,
                               pe->p2.x, pe->p2.y, pe->p1.x, pe->p1.y, "P13");
                      break;
                 case P134_CODE:
                      pe = plane_array + de->slot;
                      output_line_params(plot, pe->col, de->depth);
                      get_rgb_color_from_index(pe->col, col);
                      color_darker(col, shade_factor(pe->znorm), col);
                      plot.box(col, pe->p4.x, pe->p4.y, pe->p3.x, pe->p3.y,
                               pe->p2.x, pe->p2.y, pe->p1.x, pe->p1.y, "P134");
                      break;
                 case P1234_CODE:
                      pe = plane_array + de->slot;
                      output_line_params(plot, pe->col, de->depth);
                      get_rgb_color_from_index(pe->col, col);
                      color_darker(col, shade_factor(pe->znorm), col);
                      plot.box(col, pe->p4.x, pe->p4.y, pe->p3.x, pe->p3.y,
                               pe->p2.x, pe->p2.y, pe->p1.x, pe->p1.y, "P1234");
                      break;
            }
       }
}

static void enter_depth(const int code, const int slot, const double depth)
{
       if (depth_count >= depth_alloc) {
            depth_alloc += 1000;
            depth_array = (_DEPTH *) realloc(depth_array, 
                        depth_alloc * sizeof(_DEPTH));
       }

       _DEPTH *de = depth_array + depth_count;
       de->code = code;
       de->slot = slot;
       de->depth = depth;

       depth_count++;
}

#ifdef SUN_OS
#ifdef __cplusplus
extern "C" {
#endif
#endif
static int depth_compare(const void *de1, const void *de2)
{
       if (((_DEPTH *) de1)->depth < ((_DEPTH *) de2)->depth) return -1;
       else if (((_DEPTH *) de1)->depth > ((_DEPTH *) de2)->depth) return 1;
       else return 0;
}
#ifdef SUN_OS
#ifdef __cplusplus
}
#endif
#endif

static void output_line_params(_plot &plot, const int color_index, const double depth)
{
       COLOR col;
       get_rgb_color_from_index(color_index, col);
       plot.setcolor(col);
       double width = depthcue(depth);
       plot.setline_width(width);
}

static double shade_factor(const double znorm)
{
       double shading = 0.5;
       double shadingexponent = 1.5;
       return (shading * pow(znorm, shadingexponent) + 1.0 - shading);
}

static double depthcue(const double depth)
{
       double depth_cue = 0.75;

       double factor = depth / slab;
       if (factor < -1.0) factor = -1.0;
       else if (factor > 1.0) factor = 1.0;

       factor = (factor / 2.0 - 0.5) * depth_cue + 1.0;
       return factor;
}

static double plane_znorm(COORD &p1, COORD &p2, COORD &p3, COORD &p4,
        double znorm, const int col)
{
       if (znorm >= 0.0) return znorm;

       COORD dir1, dir2, across, n1, n2;
       vector_difference(dir1, p2, p1);
       vector_difference(dir2, p3, p4);
       vector_sum(across, p1, p2);
       vector_subtract(across, p3);
       vector_subtract(across, p4);
       vector_cross_product(n1, across, dir1);
       vector_cross_product(n2, across, dir2);

       if (n1.z >= 0.0 || n2.z >= 0.0) {
            draw_line(p1, p4, col, LINE_SOLID);
            draw_line(p2, p3, col, LINE_SOLID);
            p1.z -= 0.01;
            p2.z -= 0.01;
            p3.z -= 0.01;
            p4.z -= 0.01;
            return 0.0;
       } else return znorm;
}
