/*
FILE:     topology_global.h
*/
/*
VERSION:  1.700
*/
/*
DATE:     7/13/2005
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2005 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#ifndef _TOPOLOGY_GLOBAL_
#define _TOPOLOGY_GLOBAL_

#include "TblIndexObj.h"

#define NDB_MAX_PARAM_BONDS      25
#define NDB_MAX_PARAM_ANGLES     37

#define MAX_MONOMER              30

typedef struct {
        float val, std;
        const char  *type;
} _MEAN_STD;

typedef struct _param_bond {
        const char *Rep_Name;
        const char *type;
        const char *at1;
        const char *at2;
        int  num;
        _MEAN_STD mean_std[4];
} MONOMER_BOND;

typedef struct _param_angle {
        const char *Rep_Name;
        const char *type;
        const char *at1;
        const char *at2;
        const char *at3;
        int  num;
        _MEAN_STD mean_std[4];
} MONOMER_ANGLE;

typedef struct _param_monomer {
       const char    *name;
       int     nbond,nangle;
       MONOMER_BOND  bond[NDB_MAX_PARAM_BONDS];
       MONOMER_ANGLE angle[NDB_MAX_PARAM_ANGLES];
} MONOMER;

typedef struct _param_db {
       char resname[4];
       int  type;
       MONOMER *mon;
} MONOMER_DB;

typedef struct {
        const char *Rep_Name;
        const char *type;
        const char *at1;
        const char *at2;
        const char *at3;
        const char *at4;
} MONOMER_TORSION;

MONOMER_TORSION ALPHA   = { "ALPHA",   "O3*-P-O5*-C5*",   "O3*", "P",   "O5*", "C5*" };
/*
MONOMER_TORSION BETA    = { "BETA",    "P-O5*-C5*-C4*",   "P",   "O5*", "C5*", "C4*" };
MONOMER_TORSION GAMMA   = { "GAMMA",   "O5*-C5*-C4*-C3*", "O5*", "C5*", "C4*", "C3*" };
MONOMER_TORSION DELTA   = { "DELTA",   "C5*-C4*-C3*-O3*", "C5*", "C4*", "C3*", "O3*" };
*/
MONOMER_TORSION EPSILON = { "EPSILON", "C4*-C3*-O3*-P",   "C4*", "C3*", "O3*", "P" };
MONOMER_TORSION ZETA    = { "ZETA",    "C3*-O3*-P-O5*",   "C3*", "O3*", "P",   "O5*" };
/*
MONOMER_TORSION CHI[2]  = { "CHI",     "O4*-C1*-N9-C4",   "O4*", "C1*", "N9",  "C4",
                            "CHI",     "O4*-C1*-N1-C2",   "O4*", "C1*", "N1",  "C2" };
MONOMER_TORSION CHI1[2] = { "CHI1",    "O4*-C1*-N9-C8",   "O4*", "C1*", "N9",  "C8",
                            "CHI1",    "O4*-C1*-N1-C6",   "O4*", "C1*", "N1",  "C6" };
MONOMER_TORSION NU0     = { "NU0",     "C4*-O4*-C1*-C2*", "C4*", "O4*", "C1*", "C2*" };
MONOMER_TORSION NU1     = { "NU1",     "O4*-C1*-C2*-C3*", "O4*", "C1*", "C2*", "C3*" };
MONOMER_TORSION NU2     = { "NU2",     "C1*-C2*-C3*-C4*", "C1*", "C2*", "C3*", "C4*" };
MONOMER_TORSION NU3     = { "NU3",     "C2*-C3*-C4*-O4*", "C2*", "C3*", "C4*", "O4*" };
MONOMER_TORSION NU4     = { "NU4",     "C3*-C4*-O4*-C1*", "C3*", "C4*", "O4*", "C1*" };
MONOMER_TORSION NU5     = { "NU5",     "C5*-C4*-C3*-C2*", "C5*", "C4*", "C3*", "C2*" };
MONOMER_TORSION NU6     = { "NU6",     "O4*-C4*-C3*-O3*", "O4*", "C4*", "C3*", "O3*" };
MONOMER_TORSION NU7     = { "NU7",     "O3*-C3*-C2*-O2*", "O3*", "C3*", "C2*", "O2*" };
MONOMER_TORSION NU8     = { "NU8",     "O5*-C5*-C4*-O4*", "O5*", "C5*", "C4*", "O4*" };
*/

#define NUM_TOR_NA 5

MONOMER_TORSION tor_na[NUM_TOR_NA] = {
       { "BETA",    "P-O5*-C5*-C4*",   "P",   "O5*", "C5*", "C4*" },
       { "GAMMA",   "O5*-C5*-C4*-C3*", "O5*", "C5*", "C4*", "C3*" },
       { "DELTA",   "C5*-C4*-C3*-O3*", "C5*", "C4*", "C3*", "O3*" },
       { "CHI",     "O4*-C1*-N9-C4",   "O4*", "C1*", "N9",  "C4" },
       { "CHI",     "O4*-C1*-N1-C2",   "O4*", "C1*", "N1",  "C2" }
};

MONOMER_TORSION PHI     = { "PHI",     "C-N-CA-C",        "C",   "N",   "CA",  "C" };
MONOMER_TORSION PSI     = { "PSI",     "N-CA-C-N",        "N",   "CA",  "C",   "N" };
MONOMER_TORSION OMEGA   = { "OMEGA",   "CA-C-N-CA",       "CA",  "C",   "N",   "CA" };

#if 0
MONOMER_TORSION tor_na[8] = {
       "Alpha",   "O3*-P-O5*-C5*",   "O3*", "P",   "O5*", "C5*",
       "Beta",    "P-O5*-C5*-C4*",   "P",   "O5*", "C5*", "C4*",
       "Gamma",   "O5*-C5*-C4*-C3*", "O5*", "C5*", "C4*", "C3*",
       "Delta",   "C5*-C4*-C3*-O3*", "C5*", "C4*", "C3*", "O3*",
       "Epsilon", "C4*-C3*-O3*-P",   "C4*", "C3*", "O3*", "P",
       "Zeta",    "C3*-O3*-P-O5*",   "C3*", "O3*", "P",   "O5*",
       "Chi",     "O4*-C1*-N9-C4",   "O4*", "C1*", "N9",  "C4",
       "Chi",     "O4*-C1*-N1-C2",   "O4*", "C1*", "N1",  "C2"
};

MONOMER_TORSION tor_aa[3] = {
       "Phi",     "N-CA-C",        "C",   "N",   "CA",  "C",       /*  0 */
       "Psi",     "N-CA-C-N",        "N",   "CA",  "C",   "N",       /*  1 */
       "Omega",   "CA-C-N-CA",       "CA",  "C",   "N",   "CA",      /*  2 */
       "Chi-1",   "N-CA-CB-CG",      "N",   "CA",  "CB",  "CG",      /*  3 */
       "Chi-1",   "N-CA-CB-CG1",     "N",   "CA",  "CB",  "CG1",     /*  4 */
       "Chi-1",   "N-CA-CB-OG",      "N",   "CA",  "CB",  "OG",      /*  5 */
       "Chi-1",   "N-CA-CB-OG1",     "N",   "CA",  "CB",  "OG1",     /*  6 */
       "Chi-1",   "N-CA-CB-SG",      "N",   "CA",  "CB",  "SG",      /*  7 */
       "Chi-2",   "CA-CB-CG-CD",     "CA",  "CB",  "CG",  "CD",      /*  8 */
       "Chi-2",   "CA-CB-CG-CD1",    "CA",  "CB",  "CG",  "CD1",     /*  9 */
       "Chi-2",   "CA-CB-CG-ND1",    "CA",  "CB",  "CG",  "ND1",     /* 10 */
       "Chi-2",   "CA-CB-CG-OD1",    "CA",  "CB",  "CG",  "OD1",     /* 11 */
       "Chi-2",   "CA-CB-CG-SD",     "CA",  "CB",  "CG",  "SD",      /* 12 */
       "Chi-2",   "CA-CB-CG1-CD1",   "CA",  "CB",  "CG1", "CD1",     /* 13 */
       "Chi-3",   "CB-CG-CD-CE",     "CB",  "CG",  "CD",  "CE",      /* 14 */
       "Chi-3",   "CB-CG-CD-NE",     "CB",  "CG",  "CD",  "NE",      /* 15 */
       "Chi-3",   "CB-CG-CD-OE1",    "CB",  "CG",  "CD",  "OE1",     /* 16 */
       "Chi-3",   "CB-CG-SD-CE",     "CB",  "CG",  "SD",  "CE",      /* 17 */
       "Chi-4",   "CG-CD-CE-NZ",     "CG",  "CD",  "CE",  "NZ",      /* 18 */
       "Chi-4",   "CG-CD-NE-CZ",     "CG",  "CD",  "NE",  "CZ",      /* 19 */
       "Chi-5",   "CD-NE-CZ-NH1",    "CD",  "NE",  "CZ",  "NH1"      /* 20 */
};
#endif

MONOMER DA = {  "A",  23, 33, // 24,  37,
     {
       { "P-O1P/O2P", "O1P-P",   "P",   "O1P",  1,  { { 1.485, 0.017, "s" } } },
       { "P-O1P/O2P", "O2P-P",   "P",   "O2P",  1,  { { 1.485, 0.017, "s" } } },
       { "P-O5'",     "O5*-P",   "P",   "O5*",  1,  { { 1.593, 0.010, "s" } } },
       // { "P-O3'",     "O3*-P",   "P",   "O3*",  1,  { { 1.607, 0.012, "s" } } },
       { "O5'-C5'",   "C5*-O5*", "O5*", "C5*",  3, {{ 1.440, 0.016, "(P)O5'-C5'" },
                                                    { 1.418, 0.025, "(H)C2'-endo" },
                                                    { 1.423, 0.011, "(H)C3'-endo" }}
       },
       { "C5'-C4'",   "C4*-C5*", "C5*", "C4*",  2, {{ 1.512, 0.007, "C2'-endo" },
                                                    { 1.509, 0.011, "C3'-endo" }}
       },
       { "C3'-C4'",   "C3*-C4*", "C4*", "C3*",  2, {{ 1.529, 0.010, "C2'-endo" },
                                                    { 1.521, 0.010, "C3'-endo" }}
       },
       { "C2'-C3'",   "C2*-C3*", "C3*", "C2*",  2, {{ 1.516, 0.008, "C2'-endo" },
                                                    { 1.518, 0.012, "C3'-endo" }}
       },
       { "C1'-C2'",   "C1*-C2*", "C2*", "C1*",  2, {{ 1.518, 0.010, "C2'-endo" },
                                                    { 1.519, 0.010, "C3'-endo" }}
       },
       { "O4'-C1'",   "C1*-O4*", "O4*", "C1*",  2, {{ 1.420, 0.011, "C2'-endo" },
                                                    { 1.418, 0.012, "C3'-endo" }}
       },
       { "C4'-O4'",   "C4*-O4*", "O4*", "C4*",  2, {{ 1.446, 0.010, "C2'-endo" },
                                                    { 1.449, 0.009, "C3'-endo" }}
       },
       { "C3'-O3'",   "C3*-O3*", "O3*", "C3*",  2, {{ 1.435, 0.013, "C2'-endo" },
                                                    { 1.419, 0.006, "C3'-endo" }}
       },
       { "C1'-N1/N9", "C1*-N9",  "C1*", "N9",   2, {{ 1.468, 0.014, "C2'-endo" },
                                                    { 1.488, 0.013, "C3'-endo" }}
       },
       { "N1-C2",   "C2-N1",   "N1",  "C2",   1,  { { 1.339, 0.009, "s" } } },
       { "C2-N3",   "C2-N3",   "C2",  "N3",   1,  { { 1.331, 0.009, "s" } } },
       { "N3-C4",   "C4-N3",   "N3",  "C4",   1,  { { 1.344, 0.006, "s" } } },
       { "C4-C5",   "C4-C5",   "C4",  "C5",   1,  { { 1.383, 0.007, "s" } } },
       { "C5-C6",   "C5-C6",   "C5",  "C6",   1,  { { 1.406, 0.009, "s" } } },
       { "C6-N1",   "C6-N1",   "C6",  "N1",   1,  { { 1.351, 0.007, "s" } } },
       { "C5-N7",   "C5-N7",   "C5",  "N7",   1,  { { 1.388, 0.006, "s" } } },
       { "N7-C8",   "C8-N7",   "N7",  "C8",   1,  { { 1.311, 0.007, "s" } } },
       { "C8-N9",   "C8-N9",   "C8",  "N9",   1,  { { 1.373, 0.008, "s" } } },
       { "N9-C4",   "C4-N9",   "N9",  "C4",   1,  { { 1.374, 0.006, "s" } } },
       { "C6-N6",   "C6-N6",   "C6",  "N6",   1,  { { 1.335, 0.008, "s" } } }
     },
     {
       { "O1P-P-O2P",     "O1P-P-O2P",   "O1P", "P",   "O2P",  1,
             { { 119.600, 1.5, "s" } }
       },
       { "O5'-P-O1P/O2P", "O1P-P-O5*",   "O5*", "P",   "O1P",  2,
            {{ 110.700, 1.2, "large" },
             { 105.700, 0.9, "small" }}
       },
       { "O5'-P-O1P/O2P", "O2P-P-O5*",   "O5*", "P",   "O2P",  2,
            {{ 110.700, 1.2, "large" },
             { 105.700, 0.9, "small" }}
       },
/*
       { "O3'-P-O5'",     "O3*-P-O5*",   "O3*", "P",   "O5*",  1,
             { { 104.000, 1.9, "s" } }
       },
       { "O3'-P-O2P/O2P", "O2P-P-O3*",   "O2P", "P",   "O3*",  2,
            {{ 110.500, 1.1, "large" },
             { 105.200, 2.2, "small" }}
       },
       { "O3'-P-O1P/O2P", "O1P-P-O3*",   "O1P", "P",   "O3*",  2,
            {{ 110.500, 1.1, "large" },
             { 105.200, 2.2, "small" }}
       },
*/
       { "O5'-C5'-C4'",   "C4*-C5*-O5*", "O5*", "C5*", "C4*",  3,
            {{ 109.400, 0.8, "(P)O5'-C5'-C4'" },
             { 110.900, 1.7, "C2'-endo" },
             { 111.000, 2.5, "C3'-endo" }}
       },
       { "P-O5'-C5'",     "C5*-O5*-P",   "P",   "O5*", "C5*",  1,
             { { 120.900, 1.6, "s" } }
       },
/*
       { "C3'-O3'-P",     "C3*-O3*-P",   "P",   "O3*", "C3*",  1,
             { { 119.700, 1.2, "s" } }
       },
*/
       { "C3'-C4'-O4'",   "C3*-C4*-O4*",  "O4*", "C4*", "C3*",  2,
            {{ 106.000, 0.6, "C2'-endo" },
             { 104.500, 0.4, "C3'-endo" }}
       },
       { "C5'-C4'-C3'",   "C3*-C4*-C5*", "C5*", "C4*", "C3*",  2,
            {{ 114.100, 1.8, "C2'-endo" },
             { 115.700, 1.2, "C3'-endo" }}
       },
       { "C5'-C4'-O4'",   "C5*-C4*-O4*", "C5*", "C4*", "O4*",  2,
            {{ 109.300, 1.9, "C2'-endo" },
             { 109.800, 1.1, "C3'-endo" }}
       },
       { "C4'-O4'-C1'",   "C1*-O4*-C4*", "C1*", "O4*", "C4*",  2,
            {{ 110.100, 1.0, "C2'-endo" },
             { 110.300, 0.7, "C3'-endo" }}
       },
       { "C4'-C3'-O3'",   "C4*-C3*-O3*", "C4*", "C3*", "O3*",  2,
            {{ 109.700, 2.5, "C2'-endo" },
             { 112.300, 2.0, "C3'-endo" }}
       },
       { "C2'-C3'-O3'",   "C2*-C3*-O3*", "C2*", "C3*", "O3*",  2,
            {{ 109.400, 2.5, "C2'-endo" },
             { 112.600, 3.3, "C3'-endo" }}
       },
       { "C2'-C3'-C4'",   "C2*-C3*-C4*", "C4*", "C3*", "C2*",  2,
            {{ 103.100, 0.9, "C2'-endo" },
             { 102.200, 0.7, "C3'-endo" }}
       },
       { "C1'-C2'-C3'",   "C1*-C2*-C3*", "C3*", "C2*", "C1*",  2,
            {{ 102.500, 1.2, "C2'-endo" },
             { 102.400, 0.8, "C3'-endo" }}
       },
       { "O4'-C1'-C2'",   "C2*-C1*-O4*", "O4*", "C1*", "C2*",  2,
            {{ 105.900, 0.8, "C2'-endo" },
             { 106.800, 0.5, "C3'-endo" }}
       },
       { "C2'-C1'-N1/N9", "C2*-C1*-N9",  "N9",  "C1*", "C2*",  2,
            {{ 114.300, 1.4, "C2'-endo" },
             { 112.600, 1.9, "C3'-endo" }}
       },
       { "O4'-C1'-N1/N9", "N9-C1*-O4*",  "O4*", "C1*", "N9",   2,
            {{ 108.000, 0.7, "C2'-endo" },
             { 108.300, 0.3, "C3'-endo" }}
       },
       { "C6-N1-C2",    "C2-N1-C6",    "C6",  "N1",  "C2",   1,
             { { 118.600, 0.6, "s" } }
       },
       { "N1-C2-N3",    "N1-C2-N3",    "N1",  "C2",  "N3",   1,
             { { 129.300, 0.5, "s" } }
       },
       { "C2-N3-C4",    "C2-N3-C4",    "C2",  "N3",  "C4",   1,
             { { 110.600, 0.5, "s" } }
       },
       { "N3-C4-C5",    "C5-C4-N3",    "N3",  "C4",  "C5",   1,
             { { 126.800, 0.7, "s" } }
       },
       { "C4-C5-C6",    "C4-C5-C6",    "C4",  "C5",  "C6",   1,
             { { 117.000, 0.5, "s" } }
       },
       { "C5-C6-N1",    "C5-C6-N1",    "C5",  "C6",  "N1",   1,
             { { 117.700, 0.5, "s" } }
       },
       { "C4-C5-N7",    "C4-C5-N7",    "C4",  "C5",  "N7",   1,
             { { 110.700, 0.5, "s" } }
       },
       { "C5-N7-C8",    "C5-N7-C8",    "C5",  "N7",  "C8",   1,
             { { 103.900, 0.5, "s" } }
       },
       { "N7-C8-N9",    "N7-C8-N9",    "N7",  "C8",  "N9",   1,
             { { 113.800, 0.5, "s" } }
       },
       { "C8-N9-C4",    "C4-N9-C8",    "C8",  "N9",  "C4",   1,
             { { 105.800, 0.4, "s" } }
       },
       { "N9-C4-C5",    "C5-C4-N9",    "N9",  "C4",  "C5",   1,
             { { 105.800, 0.4, "s" } }
       },
       { "N3-C4-N9",    "N3-C4-N9",    "N3",  "C4",  "N9",   1,
             { { 127.400, 0.8, "s" } }
       },
       { "C6-C5-N7",    "C6-C5-N7",    "C6",  "C5",  "N7",   1,
             { { 132.300, 0.7, "s" } }
       },
       { "N1-C6-N6",    "N1-C6-N6",    "N1",  "C6",  "N6",   1,
             { { 118.600, 0.6, "s" } }
       },
       { "C5-C6-N6",    "C5-C6-N6",    "C5",  "C6",  "N6",   1,
             { { 123.700, 0.8, "s" } }
       },
       { "C8-N9-C1'",   "C1*-N9-C8",   "C8",  "N9",  "C1*",  1,
             { { 127.700, 1.8, "s" } }
       },
       { "C4-N9-C1'",   "C1*-N9-C4",   "C4",  "N9",  "C1*",  1,
             { { 126.300, 1.8, "s" } }
       }
     }
};

MONOMER RA = { "A", 24, 35, // 25, 39,
     {
       { "P-O1P/O2P", "O1P-P",   "P",   "O1P",  1,  { { 1.485, 0.017, "s" } } },
       { "P-O1P/O2P", "O2P-P",   "P",   "O2P",  1,  { { 1.485, 0.017, "s" } } },
       { "P-O5'",     "O5*-P",   "P",   "O5*",  1,  { { 1.593, 0.010, "s" } } },
       // { "P-O3'",     "O3*-P",   "P",   "O3*",  1,  { { 1.607, 0.012, "s" } } },
       { "O5'-C5'",   "C5*-O5*", "O5*", "C5*",  3, {{ 1.440, 0.016, "(P)O5'-C5'" },
                                                    { 1.424, 0.016, "(H)C2'-endo" },
                                                    { 1.420, 0.009, "(H)C3'-endo" }}
       },
       { "C5'-C4'",   "C4*-C5*", "C5*", "C4*",  2, {{ 1.509, 0.012, "C2'-endo" },
                                                    { 1.508, 0.007, "C3'-endo" }}
       },
       { "C3'-C4'",   "C3*-C4*", "C4*", "C3*",  2, {{ 1.527, 0.011, "C2'-endo" },
                                                    { 1.521, 0.010, "C3'-endo" }}
       },
       { "C2'-C3'",   "C2*-C3*", "C3*", "C2*",  2, {{ 1.525, 0.011, "C2'-endo" },
                                                    { 1.523, 0.011, "C3'-endo" }}
       },
       { "C1'-C2'",   "C1*-C2*", "C2*", "C1*",  2, {{ 1.526, 0.008, "C2'-endo" },
                                                    { 1.529, 0.011, "C3'-endo" }}
       },
       { "O4'-C1'",   "C1*-O4*", "O4*", "C1*",  2, {{ 1.415, 0.012, "C2'-endo" },
                                                    { 1.412, 0.013, "C3'-endo" }}
       },
       { "C4'-O4'",   "C4*-O4*", "O4*", "C4*",  2, {{ 1.454, 0.010, "C2'-endo" },
                                                    { 1.451, 0.013, "C3'-endo" }}
       },
       { "C3'-O3'",   "C3*-O3*", "O3*", "C3*",  2, {{ 1.427, 0.012, "C2'-endo" },
                                                    { 1.417, 0.014, "C3'-endo" }}
       },
       { "C1'-N1/N9", "C1*-N9",  "C1*", "N9",   2, {{ 1.464, 0.014, "C2'-endo" },
                                                    { 1.483, 0.015, "C3'-endo" }}
       },
       { "C2'-O2'",   "C2*-O2*", "C2*", "O2*",  2, {{ 1.412, 0.013, "C2'-endo" },
                                                    { 1.420, 0.010, "C3'-endo" }}
       },
       { "N1-C2",   "C2-N1",   "N1",  "C2",   1,  { { 1.339, 0.009, "s" } } },
       { "C2-N3",   "C2-N3",   "C2",  "N3",   1,  { { 1.331, 0.009, "s" } } },
       { "N3-C4",   "C4-N3",   "N3",  "C4",   1,  { { 1.344, 0.006, "s" } } },
       { "C4-C5",   "C4-C5",   "C4",  "C5",   1,  { { 1.383, 0.007, "s" } } },
       { "C5-C6",   "C5-C6",   "C5",  "C6",   1,  { { 1.406, 0.009, "s" } } },
       { "C6-N1",   "C6-N1",   "C6",  "N1",   1,  { { 1.351, 0.007, "s" } } },
       { "C5-N7",   "C5-N7",   "C5",  "N7",   1,  { { 1.388, 0.006, "s" } } },
       { "N7-C8",   "C8-N7",   "N7",  "C8",   1,  { { 1.311, 0.007, "s" } } },
       { "C8-N9",   "C8-N9",   "C8",  "N9",   1,  { { 1.373, 0.008, "s" } } },
       { "N9-C4",   "C4-N9",   "N9",  "C4",   1,  { { 1.374, 0.006, "s" } } },
       { "C6-N6",   "C6-N6",   "C6",  "N6",   1,  { { 1.335, 0.008, "s" } } }
     },
     {
       { "O1P-P-O2P",     "O1P-P-O2P",   "O1P", "P",   "O2P",  1,
             { { 119.600, 1.5, "s" } }
       },
       { "O5'-P-O1P/O2P", "O1P-P-O5*",   "O5*", "P",   "O1P",  2,
            {{ 110.700, 1.2, "large" },
             { 105.700, 0.9, "small" }}
       },
       { "O5'-P-O1P/O2P", "O2P-P-O5*",   "O5*", "P",   "O2P",  2,
            {{ 110.700, 1.2, "large" },
             { 105.700, 0.9, "small" }}
       },
/*
       { "O3'-P-O5'",     "O3*-P-O5*",   "O3*", "P",   "O5*",  1,
             { { 104.000, 1.9, "s" } }
       },
       { "O3'-P-O2P/O2P", "O2P-P-O3*",   "O2P", "P",   "O3*",  2,
            {{ 110.500, 1.1, "large" }
             { 105.200, 2.2, "small" }}
       },
       { "O3'-P-O1P/O2P", "O1P-P-O3*",   "O1P", "P",   "O3*",  2,
            {{ 110.500, 1.1, "large" }
             { 105.200, 2.2, "small" }}
       },
*/
       { "O5'-C5'-C4'",   "C4*-C5*-O5*", "O5*", "C5*", "C4*",  3,
            {{ 109.400, 0.8, "(P)O5'-C5'-C4'" },
             { 111.700, 1.9, "C2'-endo" },
             { 111.500, 1.6, "C3'-endo" }}
       },
       { "P-O5'-C5'",     "C5*-O5*-P",   "P",   "O5*", "C5*",  1,
             { { 120.900, 1.6, "s" } }
       },
/*
       { "C3'-O3'-P",     "C3*-O3*-P",   "P",   "O3*", "C3*",  1,
             { { 119.700, 1.2, "s" } }
       },
*/
       { "C3'-C4'-O4'",   "C3*-C4*-O4*",  "O4*", "C4*", "C3*",  2,
            {{ 106.100, 0.8, "C2'-endo" },
             { 104.000, 1.0, "C3'-endo" }}
       },
       { "C5'-C4'-C3'",   "C3*-C4*-C5*", "C5*", "C4*", "C3*",  2,
            {{ 115.200, 1.4, "C2'-endo" },
             { 116.000, 1.6, "C3'-endo" }}
       },
       { "C5'-C4'-O4'",   "C5*-C4*-O4*", "C5*", "C4*", "O4*",  2,
            {{ 109.100, 1.2, "C2'-endo" },
             { 109.800, 0.9, "C3'-endo" }}
       },
       { "C4'-O4'-C1'",   "C1*-O4*-C4*", "C1*", "O4*", "C4*",  2,
            {{ 109.700, 0.7, "C2'-endo" },
             { 109.900, 0.8, "C3'-endo" }}
       },
       { "C4'-C3'-O3'",   "C4*-C3*-O3*", "C4*", "C3*", "O3*",  2,
            {{ 109.400, 2.1, "C2'-endo" },
             { 113.000, 2.0, "C3'-endo" }}
       },
       { "C2'-C3'-O3'",   "C2*-C3*-O3*", "C2*", "C3*", "O3*",  2,
            {{ 109.500, 2.2, "C2'-endo" },
             { 113.700, 1.6, "C3'-endo" }}
       },
       { "C2'-C3'-C4'",   "C2*-C3*-C4*", "C4*", "C3*", "C2*",  2,
            {{ 102.600, 1.0, "C2'-endo" },
             { 102.600, 1.0, "C3'-endo" }}
       },
       { "C1'-C2'-C3'",   "C1*-C2*-C3*", "C3*", "C2*", "C1*",  2,
            {{ 101.500, 0.8, "C2'-endo" },
             { 101.300, 0.7, "C3'-endo" }}
       },
       { "O4'-C1'-C2'",   "C2*-C1*-O4*", "O4*", "C1*", "C2*",  2,
            {{ 105.800, 1.0, "C2'-endo" },
             { 107.600, 0.9, "C3'-endo" }}
       },
       { "C2'-C1'-N1/N9", "C2*-C1*-N9",  "N9",  "C1*", "C2*",  2,
            {{ 114.000, 1.3, "C2'-endo" },
             { 112.000, 1.1, "C3'-endo" }}
       },
       { "O4'-C1'-N1/N9", "N9-C1*-O4*",  "O4*", "C1*", "N9",   2,
            {{ 108.200, 0.8, "C2'-endo" },
             { 108.500, 0.7, "C3'-endo" }}
       },
       { "C1'-C2'-O2'",   "C1*-C2*-O2*", "C1*", "C2*", "O2*",  1,
             { { 110.600, 3.0, "s" } }
       },
       { "C3'-C2'-O2'",   "C3*-C2*-O2*", "C3*", "C2*", "O2*",  1,
             { { 113.300, 2.9, "s" } }
       },
       { "C6-N1-C2",    "C2-N1-C6",    "C6",  "N1",  "C2",   1,
             { { 118.600, 0.6, "s" } }
       },
       { "N1-C2-N3",    "N1-C2-N3",    "N1",  "C2",  "N3",   1,
             { { 129.300, 0.5, "s" } }
       },
       { "C2-N3-C4",    "C2-N3-C4",    "C2",  "N3",  "C4",   1,
             { { 110.600, 0.5, "s" } }
       },
       { "N3-C4-C5",    "C5-C4-N3",    "N3",  "C4",  "C5",   1,
             { { 126.800, 0.7, "s" } }
       },
       { "C4-C5-C6",    "C4-C5-C6",    "C4",  "C5",  "C6",   1,
             { { 117.000, 0.5, "s" } }
       },
       { "C5-C6-N1",    "C5-C6-N1",    "C5",  "C6",  "N1",   1,
             { { 117.700, 0.5, "s" } }
       },
       { "C4-C5-N7",    "C4-C5-N7",    "C4",  "C5",  "N7",   1,
             { { 110.700, 0.5, "s" } }
       },
       { "C5-N7-C8",    "C5-N7-C8",    "C5",  "N7",  "C8",   1,
             { { 103.900, 0.5, "s" } }
       },
       { "N7-C8-N9",    "N7-C8-N9",    "N7",  "C8",  "N9",   1,
             { { 113.800, 0.5, "s" } }
       },
       { "C8-N9-C4",    "C4-N9-C8",    "C8",  "N9",  "C4",   1,
             { { 105.800, 0.4, "s" } }
       },
       { "N9-C4-C5",    "C5-C4-N9",    "N9",  "C4",  "C5",   1,
             { { 105.800, 0.4, "s" } }
       },
       { "N3-C4-N9",    "N3-C4-N9",    "N3",  "C4",  "N9",   1,
             { { 127.400, 0.8, "s" } }
       },
       { "C6-C5-N7",    "C6-C5-N7",    "C6",  "C5",  "N7",   1,
             { { 132.300, 0.7, "s" } }
       },
       { "N1-C6-N6",    "N1-C6-N6",    "N1",  "C6",  "N6",   1,
             { { 118.600, 0.6, "s" } }
       },
       { "C5-C6-N6",    "C5-C6-N6",    "C5",  "C6",  "N6",   1,
             { { 123.700, 0.8, "s" } }
       },
       { "C8-N9-C1'",   "C1*-N9-C8",   "C8",  "N9",  "C1*",  1,
             { { 127.700, 1.8, "s" } }
       },
       { "C4-N9-C1'",   "C1*-N9-C4",   "C4",  "N9",  "C1*",  1,
             { { 126.300, 1.8, "s" } }
       }
     }
};

MONOMER DC = { "C", 20, 28, // 21, 32,
     {
       { "P-O1P/O2P", "O1P-P",   "P",   "O1P",  1,  { { 1.485, 0.017, "s" } } },
       { "P-O1P/O2P", "O2P-P",   "P",   "O2P",  1,  { { 1.485, 0.017, "s" } } },
       { "P-O5'",     "O5*-P",   "P",   "O5*",  1,  { { 1.593, 0.010, "s" } } },
       // { "P-O3'",     "O3*-P",   "P",   "O3*",  1,  { { 1.607, 0.012, "s" } } },
       { "O5'-C5'",   "C5*-O5*", "O5*", "C5*",  3, {{ 1.440, 0.016, "(P)O5'-C5'" },
                                                    { 1.418, 0.025, "(H)C2'-endo" },
                                                    { 1.423, 0.011, "(H)C3'-endo" }}
       },
       { "C5'-C4'",   "C4*-C5*", "C5*", "C4*",  2, {{ 1.512, 0.007, "C2'-endo" },
                                                    { 1.509, 0.011, "C3'-endo" }}
       },
       { "C3'-C4'",   "C3*-C4*", "C4*", "C3*",  2, {{ 1.529, 0.010, "C2'-endo" },
                                                    { 1.521, 0.010, "C3'-endo" }}
       },
       { "C2'-C3'",   "C2*-C3*", "C3*", "C2*",  2, {{ 1.516, 0.008, "C2'-endo" },
                                                    { 1.518, 0.012, "C3'-endo" }}
       },
       { "C1'-C2'",   "C1*-C2*", "C2*", "C1*",  2, {{ 1.518, 0.010, "C2'-endo" },
                                                    { 1.519, 0.010, "C3'-endo" }}
       },
       { "O4'-C1'",   "C1*-O4*", "O4*", "C1*",  2, {{ 1.420, 0.011, "C2'-endo" },
                                                    { 1.418, 0.012, "C3'-endo" }}
       },
       { "C4'-O4'",   "C4*-O4*", "O4*", "C4*",  2, {{ 1.446, 0.010, "C2'-endo" },
                                                    { 1.449, 0.009, "C3'-endo" }}
       },
       { "C3'-O3'",   "C3*-O3*", "O3*", "C3*",  2, {{ 1.435, 0.013, "C2'-endo" },
                                                    { 1.419, 0.006, "C3'-endo" }}
       },
       { "C1'-N1/N9", "C1*-N1",  "C1*", "N1",   2, {{ 1.468, 0.014, "C2'-endo" },
                                                    { 1.488, 0.013, "C3'-endo" }}
       },
       { "C2-O2",   "C2-O2",   "C2",  "O2",   1,  { { 1.240, 0.009, "s" } } },
       { "C4-N4",   "C4-N4",   "C4",  "N4",   1,  { { 1.335, 0.009, "s" } } },
       { "N1-C2",   "C2-N1",   "N1",  "C2",   1,  { { 1.397, 0.010, "s" } } },
       { "C6-N1",   "C6-N1",   "N1",  "C6",   1,  { { 1.367, 0.006, "s" } } },
       { "C2-N3",   "C2-N3",   "C2",  "N3",   1,  { { 1.353, 0.008, "s" } } },
       { "N3-C4",   "C4-N3",   "N3",  "C4",   1,  { { 1.335, 0.007, "s" } } },
       { "C4-C5",   "C4-C5",   "C4",  "C5",   1,  { { 1.425, 0.008, "s" } } },
       { "C5-C6",   "C5-C6",   "C5",  "C6",   1,  { { 1.339, 0.008, "s" } } }
     },
     {
       { "O1P-P-O2P",     "O1P-P-O2P",   "O1P", "P",   "O2P",  1,
             { { 119.600, 1.5, "s" } }
       },
       { "O5'-P-O1P/O2P", "O1P-P-O5*",   "O5*", "P",   "O1P",  2,
            {{ 110.700, 1.2, "large" },
             { 105.700, 0.9, "small" }}
       },
       { "O5'-P-O1P/O2P", "O2P-P-O5*",   "O5*", "P",   "O2P",  2,
            {{ 110.700, 1.2, "large" },
             { 105.700, 0.9, "small" }}
       },
/*
       { "O3'-P-O5'",     "O3*-P-O5*",   "O3*", "P",   "O5*",  1,
             { { 104.000, 1.9, "s" } }
       },
       { "O3'-P-O2P/O2P", "O2P-P-O3*",   "O2P", "P",   "O3*",  2,
            {{ 110.500, 1.1, "large" }
             { 105.200, 2.2, "small" }}
       },
       { "O3'-P-O1P/O2P", "O1P-P-O3*",   "O1P", "P",   "O3*",  2,
            {{ 110.500, 1.1, "large" }
             { 105.200, 2.2, "small" }}
       },
*/
       { "O5'-C5'-C4'",   "C4*-C5*-O5*", "O5*", "C5*", "C4*",  3,
            {{ 109.400, 0.8, "(P)O5'-C5'-C4'" },
             { 110.900, 1.7, "C2'-endo" },
             { 111.000, 2.5, "C3'-endo" }}
       },
       { "P-O5'-C5'",     "C5*-O5*-P",   "P",   "O5*", "C5*",  1,
             { { 120.900, 1.6, "s" } }
       },
/*
       { "C3'-O3'-P",     "C3*-O3*-P",   "P",   "O3*", "C3*",  1,
             { { 119.700, 1.2, "s" } }
       },
*/
       { "C3'-C4'-O4'",   "C3*-C4*-O4*",  "O4*", "C4*", "C3*",  2,
            {{ 106.000, 0.6, "C2'-endo" },
             { 104.500, 0.4, "C3'-endo" }}
       },
       { "C5'-C4'-C3'",   "C3*-C4*-C5*", "C5*", "C4*", "C3*",  2,
            {{ 114.100, 1.8, "C2'-endo" },
             { 115.700, 1.2, "C3'-endo" }}
       },
       { "C5'-C4'-O4'",   "C5*-C4*-O4*", "C5*", "C4*", "O4*",  2,
            {{ 109.300, 1.9, "C2'-endo" },
             { 109.800, 1.1, "C3'-endo" }}
       },
       { "C4'-O4'-C1'",   "C1*-O4*-C4*", "C1*", "O4*", "C4*",  2,
            {{ 110.100, 1.0, "C2'-endo" },
             { 110.300, 0.7, "C3'-endo" }}
       },
       { "C4'-C3'-O3'",   "C4*-C3*-O3*", "C4*", "C3*", "O3*",  2,
            {{ 109.700, 2.5, "C2'-endo" },
             { 112.300, 2.0, "C3'-endo" }}
       },
       { "C2'-C3'-O3'",   "C2*-C3*-O3*", "C2*", "C3*", "O3*",  2,
            {{ 109.400, 2.5, "C2'-endo" },
             { 112.600, 3.3, "C3'-endo" }}
       },
       { "C2'-C3'-C4'",   "C2*-C3*-C4*", "C4*", "C3*", "C2*",  2,
            {{ 103.100, 0.9, "C2'-endo" },
             { 102.200, 0.7, "C3'-endo" }}
       },
       { "C1'-C2'-C3'",   "C1*-C2*-C3*", "C3*", "C2*", "C1*",  2,
            {{ 102.500, 1.2, "C2'-endo" },
             { 102.400, 0.8, "C3'-endo" }}
       },
       { "O4'-C1'-C2'",   "C2*-C1*-O4*", "O4*", "C1*", "C2*",  2,
            {{ 105.900, 0.8, "C2'-endo" },
             { 106.800, 0.5, "C3'-endo" }}
       },
       { "C2'-C1'-N1/N9", "C2*-C1*-N1",  "N1",  "C1*", "C2*",  2,
            {{ 114.300, 1.4, "C2'-endo" },
             { 112.600, 1.9, "C3'-endo" }}
       },
       { "O4'-C1'-N1/N9", "N1-C1*-O4*",  "O4*", "C1*", "N1",   2,
            {{ 108.000, 0.7, "C2'-endo" },
             { 108.300, 0.3, "C3'-endo" }}
       },
       { "C6-N1-C2",    "C2-N1-C6",    "C6",  "N1",  "C2",   1,
             { { 120.300, 0.4, "s" } }
       },
       { "N1-C2-N3",    "N1-C2-N3",    "N1",  "C2",  "N3",   1,
             { { 119.200, 0.7, "s" } }
       },
       { "C2-N3-C4",    "C2-N3-C4",    "C2",  "N3",  "C4",   1,
             { { 119.900, 0.5, "s" } }
       },
       { "N3-C4-C5",    "C5-C4-N3",    "N3",  "C4",  "C5",   1,
             { { 121.900, 0.4, "s" } }
       },
       { "C4-C5-C6",    "C4-C5-C6",    "C4",  "C5",  "C6",   1,
             { { 117.400, 0.5, "s" } }
       },
       { "C5-C6-N1",    "C5-C6-N1",    "C5",  "C6",  "N1",   1,
             { { 121.000, 0.5, "s" } }
       },
       { "N1-C2-O2",    "N1-C2-O2",    "N1",  "C2",  "O2",   1,
             { { 118.900, 0.6, "s" } }
       },
       { "N3-C2-O2",    "N3-C2-O2",    "N3",  "C2",  "O2",   1,
             { { 121.900, 0.7, "s" } }
       },
       { "N3-C4-N4",    "N3-C4-N4",    "N3",  "C4",  "N4",   1,
             { { 118.000, 0.7, "s" } }
       },
       { "C5-C4-N4",    "C5-C4-N4",    "C5",  "C4",  "N4",   1,
             { { 120.200, 0.7, "s" } }
       },
       { "C6-N1-C1'",   "C1*-N1-C6",   "C6",  "N1",  "C1*",  1,
             { { 120.800, 1.2, "s" } }
       },
       { "C2-N1-C1'",   "C1*-N1-C2",   "C2",  "N1",  "C1*",  1,
             { { 118.800, 1.1, "s" } }
       }
     }
};

MONOMER RC = { "C", 21, 30, // 22, 34,
     {
       { "P-O1P/O2P", "O1P-P",   "P",   "O1P",  1,  { { 1.485, 0.017, "s" } } },
       { "P-O1P/O2P", "O2P-P",   "P",   "O2P",  1,  { { 1.485, 0.017, "s" } } },
       { "P-O5'",     "O5*-P",   "P",   "O5*",  1,  { { 1.593, 0.010, "s" } } },
       // { "P-O3'",     "O3*-P",   "P",   "O3*",  1,  { { 1.607, 0.012, "s" } } },
       { "O5'-C5'",   "C5*-O5*", "O5*", "C5*",  3, {{ 1.440, 0.016, "(P)O5'-C5'" },
                                                    { 1.424, 0.016, "(H)C2'-endo" },
                                                    { 1.420, 0.009, "(H)C3'-endo" }}
       },
       { "C5'-C4'",   "C4*-C5*", "C5*", "C4*",  2, {{ 1.509, 0.012, "C2'-endo" },
                                                    { 1.508, 0.007, "C3'-endo" }}
       },
       { "C3'-C4'",   "C3*-C4*", "C4*", "C3*",  2, {{ 1.527, 0.011, "C2'-endo" },
                                                    { 1.521, 0.010, "C3'-endo" }}
       },
       { "C2'-C3'",   "C2*-C3*", "C3*", "C2*",  2, {{ 1.525, 0.011, "C2'-endo" },
                                                    { 1.523, 0.011, "C3'-endo" }}
       },
       { "C1'-C2'",   "C1*-C2*", "C2*", "C1*",  2, {{ 1.526, 0.008, "C2'-endo" },
                                                    { 1.529, 0.011, "C3'-endo" }}
       },
       { "O4'-C1'",   "C1*-O4*", "O4*", "C1*",  2, {{ 1.415, 0.012, "C2'-endo" },
                                                    { 1.412, 0.013, "C3'-endo" }}
       },
       { "C4'-O4'",   "C4*-O4*", "O4*", "C4*",  2, {{ 1.454, 0.010, "C2'-endo" },
                                                    { 1.451, 0.013, "C3'-endo" }}
       },
       { "C3'-O3'",   "C3*-O3*", "O3*", "C3*",  2, {{ 1.427, 0.012, "C2'-endo" },
                                                    { 1.417, 0.014, "C3'-endo" }}
       },
       { "C1'-N1/N9", "C1*-N1",  "C1*", "N1",   2, {{ 1.464, 0.014, "C2'-endo" },
                                                    { 1.483, 0.015, "C3'-endo" }}
       },
       { "C2'-O2'",   "C2*-O2*", "C2*", "O2*",  2, {{ 1.412, 0.013, "C2'-endo" },
                                                    { 1.420, 0.010, "C3'-endo" }}
       },
       { "C2-O2",   "C2-O2",   "C2",  "O2",   1,  { { 1.240, 0.009, "s" } } },
       { "C4-N4",   "C4-N4",   "C4",  "N4",   1,  { { 1.335, 0.009, "s" } } },
       { "N1-C2",   "C2-N1",   "N1",  "C2",   1,  { { 1.397, 0.010, "s" } } },
       { "C6-N1",   "C6-N1",   "N1",  "C6",   1,  { { 1.367, 0.006, "s" } } },
       { "C2-N3",   "C2-N3",   "C2",  "N3",   1,  { { 1.353, 0.008, "s" } } },
       { "N3-C4",   "C4-N3",   "N3",  "C4",   1,  { { 1.335, 0.007, "s" } } },
       { "C4-C5",   "C4-C5",   "C4",  "C5",   1,  { { 1.425, 0.008, "s" } } },
       { "C5-C6",   "C5-C6",   "C5",  "C6",   1,  { { 1.339, 0.008, "s" } } }
     },
     {
       { "O1P-P-O2P",     "O1P-P-O2P",   "O1P", "P",   "O2P",  1,
             { { 119.600, 1.5, "s" } }
       },
       { "O5'-P-O1P/O2P", "O1P-P-O5*",   "O5*", "P",   "O1P",  2,
            {{ 110.700, 1.2, "large" },
             { 105.700, 0.9, "small" }}
       },
       { "O5'-P-O1P/O2P", "O2P-P-O5*",   "O5*", "P",   "O2P",  2,
            {{ 110.700, 1.2, "large" },
             { 105.700, 0.9, "small" }}
       },
/*
       { "O3'-P-O5'",     "O3*-P-O5*",   "O3*", "P",   "O5*",  1,
             { { 104.000, 1.9, "s" } }
       },
       { "O3'-P-O2P/O2P", "O2P-P-O3*",   "O2P", "P",   "O3*",  2,
            {{ 110.500, 1.1, "large" },
             { 105.200, 2.2, "small" }}
       },
       { "O3'-P-O1P/O2P", "O1P-P-O3*",   "O1P", "P",   "O3*",  2,
            {{ 110.500, 1.1, "large" },
             { 105.200, 2.2, "small" }}
       },
*/
       { "O5'-C5'-C4'",   "C4*-C5*-O5*", "O5*", "C5*", "C4*",  3,
            {{ 109.400, 0.8, "(P)O5'-C5'-C4'" },
             { 111.700, 1.9, "C2'-endo" },
             { 111.500, 1.6, "C3'-endo" }}
       },
       { "P-O5'-C5'",     "C5*-O5*-P",   "P",   "O5*", "C5*",  1,
             { { 120.900, 1.6, "s" } }
       },
/*
       { "C3'-O3'-P",     "C3*-O3*-P",   "P",   "O3*", "C3*",  1,
             { { 119.700, 1.2, "s" } }
       },
*/
       { "C3'-C4'-O4'",   "C3*-C4*-O4*",  "O4*", "C4*", "C3*",  2,
            {{ 106.100, 0.8, "C2'-endo" },
             { 104.000, 1.0, "C3'-endo" }}
       },
       { "C5'-C4'-C3'",   "C3*-C4*-C5*", "C5*", "C4*", "C3*",  2,
            {{ 115.200, 1.4, "C2'-endo" },
             { 116.000, 1.6, "C3'-endo" }}
       },
       { "C5'-C4'-O4'",   "C5*-C4*-O4*", "C5*", "C4*", "O4*",  2,
            {{ 109.100, 1.2, "C2'-endo" },
             { 109.800, 0.9, "C3'-endo" }}
       },
       { "C4'-O4'-C1'",   "C1*-O4*-C4*", "C1*", "O4*", "C4*",  2,
            {{ 109.700, 0.7, "C2'-endo" },
             { 109.900, 0.8, "C3'-endo" }}
       },
       { "C4'-C3'-O3'",   "C4*-C3*-O3*", "C4*", "C3*", "O3*",  2,
            {{ 109.400, 2.1, "C2'-endo" },
             { 113.000, 2.0, "C3'-endo" }}
       },
       { "C2'-C3'-O3'",   "C2*-C3*-O3*", "C2*", "C3*", "O3*",  2,
            {{ 109.500, 2.2, "C2'-endo" },
             { 113.700, 1.6, "C3'-endo" }}
       },
       { "C2'-C3'-C4'",   "C2*-C3*-C4*", "C4*", "C3*", "C2*",  2,
            {{ 102.600, 1.0, "C2'-endo" },
             { 102.600, 1.0, "C3'-endo" }}
       },
       { "C1'-C2'-C3'",   "C1*-C2*-C3*", "C3*", "C2*", "C1*",  2,
            {{ 101.500, 0.8, "C2'-endo" },
             { 101.300, 0.7, "C3'-endo" }}
       },
       { "O4'-C1'-C2'",   "C2*-C1*-O4*", "O4*", "C1*", "C2*",  2,
            {{ 105.800, 1.0, "C2'-endo" },
             { 107.600, 0.9, "C3'-endo" }}
       },
       { "C2'-C1'-N1/N9", "C2*-C1*-N1",  "N1",  "C1*", "C2*",  2,
            {{ 114.000, 1.3, "C2'-endo" },
             { 112.000, 1.1, "C3'-endo" }}
       },
       { "O4'-C1'-N1/N9", "N1-C1*-O4*",  "O4*", "C1*", "N1",   2,
            {{ 108.200, 0.8, "C2'-endo" },
             { 108.500, 0.7, "C3'-endo" }}
       },
       { "C1'-C2'-O2'",   "C1*-C2*-O2*", "C1*", "C2*", "O2*",  1,
             { { 110.600, 3.0, "s" } }
       },
       { "C3'-C2'-O2'",   "C3*-C2*-O2*", "C3*", "C2*", "O2*",  1,
             { { 113.300, 2.9, "s" } }
       },
       { "C6-N1-C2",    "C2-N1-C6",    "C6",  "N1",  "C2",   1,
             { { 120.300, 0.4, "s" } }
       },
       { "N1-C2-N3",    "N1-C2-N3",    "N1",  "C2",  "N3",   1,
             { { 119.200, 0.7, "s" } }
       },
       { "C2-N3-C4",    "C2-N3-C4",    "C2",  "N3",  "C4",   1,
             { { 119.900, 0.5, "s" } }
       },
       { "N3-C4-C5",    "C5-C4-N3",    "N3",  "C4",  "C5",   1,
             { { 121.900, 0.4, "s" } }
       },
       { "C4-C5-C6",    "C4-C5-C6",    "C4",  "C5",  "C6",   1,
             { { 117.400, 0.5, "s" } }
       },
       { "C5-C6-N1",    "C5-C6-N1",    "C5",  "C6",  "N1",   1,
             { { 121.000, 0.5, "s" } }
       },
       { "N1-C2-O2",    "N1-C2-O2",    "N1",  "C2",  "O2",   1,
             { { 118.900, 0.6, "s" } }
       },
       { "N3-C2-O2",    "N3-C2-O2",    "N3",  "C2",  "O2",   1,
             { { 121.900, 0.7, "s" } }
       },
       { "N3-C4-N4",    "N3-C4-N4",    "N3",  "C4",  "N4",   1,
             { { 118.000, 0.7, "s" } }
       },
       { "C5-C4-N4",    "C5-C4-N4",    "C5",  "C4",  "N4",   1,
             { { 120.200, 0.7, "s" } }
       },
       { "C6-N1-C1'",   "C1*-N1-C6",   "C6",  "N1",  "C1*",  1,
             { { 120.800, 1.2, "s" } }
       },
       { "C2-N1-C1'",   "C1*-N1-C2",   "C2",  "N1",  "C1*",  1,
             { { 118.800, 1.1, "s" } }
       }
     }
};

MONOMER DG = { "G", 24, 35, // 25, 39,
     {
       { "P-O1P/O2P", "O1P-P",   "P",   "O1P",  1,  { { 1.485, 0.017, "s" } } },
       { "P-O1P/O2P", "O2P-P",   "P",   "O2P",  1,  { { 1.485, 0.017, "s" } } },
       { "P-O5'",     "O5*-P",   "P",   "O5*",  1,  { { 1.593, 0.010, "s" } } },
       // { "P-O3'",     "O3*-P",   "P",   "O3*",  1,  { { 1.607, 0.012, "s" } } },
       { "O5'-C5'",   "C5*-O5*", "O5*", "C5*",  3, {{ 1.440, 0.016, "(P)O5'-C5'" },
                                                    { 1.418, 0.025, "(H)C2'-endo" },
                                                    { 1.423, 0.011, "(H)C3'-endo" }}
       },
       { "C5'-C4'",   "C4*-C5*", "C5*", "C4*",  2, {{ 1.512, 0.007, "C2'-endo" },
                                                    { 1.509, 0.011, "C3'-endo" }}
       },
       { "C3'-C4'",   "C3*-C4*", "C4*", "C3*",  2, {{ 1.529, 0.010, "C2'-endo" },
                                                    { 1.521, 0.010, "C3'-endo" }}
       },
       { "C2'-C3'",   "C2*-C3*", "C3*", "C2*",  2, {{ 1.516, 0.008, "C2'-endo" },
                                                    { 1.518, 0.012, "C3'-endo" }}
       },
       { "C1'-C2'",   "C1*-C2*", "C2*", "C1*",  2, {{ 1.518, 0.010, "C2'-endo" },
                                                    { 1.519, 0.010, "C3'-endo" }}
       },
       { "O4'-C1'",   "C1*-O4*", "O4*", "C1*",  2, {{ 1.420, 0.011, "C2'-endo" },
                                                    { 1.418, 0.012, "C3'-endo" }}
       },
       { "C4'-O4'",   "C4*-O4*", "O4*", "C4*",  2, {{ 1.446, 0.010, "C2'-endo" },
                                                    { 1.449, 0.009, "C3'-endo" }}
       },
       { "C3'-O3'",   "C3*-O3*", "O3*", "C3*",  2, {{ 1.435, 0.013, "C2'-endo" },
                                                    { 1.419, 0.006, "C3'-endo" }}
       },
       { "C1'-N1/N9", "C1*-N9",  "C1*", "N9",   2, {{ 1.468, 0.014, "C2'-endo" },
                                                    { 1.488, 0.013, "C3'-endo" }}
       },
       { "N1-C2",   "C2-N1",   "N1",  "C2",   1,  { { 1.373, 0.008, "s" } } },
       { "C2-N3",   "C2-N3",   "C2",  "N3",   1,  { { 1.323, 0.008, "s" } } },
       { "N3-C4",   "C4-N3",   "N3",  "C4",   1,  { { 1.350, 0.007, "s" } } },
       { "C4-C5",   "C4-C5",   "C4",  "C5",   1,  { { 1.379, 0.007, "s" } } },
       { "C5-C6",   "C5-C6",   "C5",  "C6",   1,  { { 1.419, 0.010, "s" } } },
       { "C6-N1",   "C6-N1",   "C6",  "N1",   1,  { { 1.391, 0.007, "s" } } },
       { "C5-N7",   "C5-N7",   "C5",  "N7",   1,  { { 1.388, 0.006, "s" } } },
       { "N7-C8",   "C8-N7",   "N7",  "C8",   1,  { { 1.305, 0.006, "s" } } },
       { "C8-N9",   "C8-N9",   "C8",  "N9",   1,  { { 1.374, 0.007, "s" } } },
       { "N9-C4",   "C4-N9",   "N9",  "C4",   1,  { { 1.375, 0.008, "s" } } },
       { "C2-N2",   "C2-N2",   "C2",  "N2",   1,  { { 1.341, 0.010, "s" } } },
       { "C6-O6",   "C6-O6",   "C6",  "O6",   1,  { { 1.237, 0.009, "s" } } }
     },
     {
       { "O1P-P-O2P",     "O1P-P-O2P",   "O1P", "P",   "O2P",  1,
             { { 119.600, 1.5, "s" } }
       },
       { "O5'-P-O1P/O2P", "O1P-P-O5*",   "O5*", "P",   "O1P",  2,
            {{ 110.700, 1.2, "large" },
             { 105.700, 0.9, "small" }}
       },
       { "O5'-P-O1P/O2P", "O2P-P-O5*",   "O5*", "P",   "O2P",  2,
            {{ 110.700, 1.2, "large" },
             { 105.700, 0.9, "small" }}
       },
/*
       { "O3'-P-O5'",     "O3*-P-O5*",   "O3*", "P",   "O5*",  1,
             { { 104.000, 1.9, "s" } }
       },
       { "O3'-P-O2P/O2P", "O2P-P-O3*",   "O2P", "P",   "O3*",  2,
            {{ 110.500, 1.1, "large" },
             { 105.200, 2.2, "small" }}
       },
       { "O3'-P-O1P/O2P", "O1P-P-O3*",   "O1P", "P",   "O3*",  2,
            {{ 110.500, 1.1, "large" },
             { 105.200, 2.2, "small" }}
       },
*/
       { "O5'-C5'-C4'",   "C4*-C5*-O5*", "O5*", "C5*", "C4*",  3,
            {{ 109.400, 0.8, "(P)O5'-C5'-C4'" },
             { 110.900, 1.7, "C2'-endo" },
             { 111.000, 2.5, "C3'-endo" }}
       },
       { "P-O5'-C5'",     "C5*-O5*-P",   "P",   "O5*", "C5*",  1,
             { { 120.900, 1.6, "s" } }
       },
/*
       { "C3'-O3'-P",     "C3*-O3*-P",   "P",   "O3*", "C3*",  1,
             { { 119.700, 1.2, "s" } }
       },
*/
       { "C3'-C4'-O4'",   "C3*-C4*-O4*",  "O4*", "C4*", "C3*",  2,
            {{ 106.000, 0.6, "C2'-endo" },
             { 104.500, 0.4, "C3'-endo" }}
       },
       { "C5'-C4'-C3'",   "C3*-C4*-C5*", "C5*", "C4*", "C3*",  2,
            {{ 114.100, 1.8, "C2'-endo" },
             { 115.700, 1.2, "C3'-endo" }}
       },
       { "C5'-C4'-O4'",   "C5*-C4*-O4*", "C5*", "C4*", "O4*",  2,
            {{ 109.300, 1.9, "C2'-endo" },
             { 109.800, 1.1, "C3'-endo" }}
       },
       { "C4'-O4'-C1'",   "C1*-O4*-C4*", "C1*", "O4*", "C4*",  2,
            {{ 110.100, 1.0, "C2'-endo" },
             { 110.300, 0.7, "C3'-endo" }}
       },
       { "C4'-C3'-O3'",   "C4*-C3*-O3*", "C4*", "C3*", "O3*",  2,
            {{ 109.700, 2.5, "C2'-endo" },
             { 112.300, 2.0, "C3'-endo" }}
       },
       { "C2'-C3'-O3'",   "C2*-C3*-O3*", "C2*", "C3*", "O3*",  2,
            {{ 109.400, 2.5, "C2'-endo" },
             { 112.600, 3.3, "C3'-endo" }}
       },
       { "C2'-C3'-C4'",   "C2*-C3*-C4*", "C4*", "C3*", "C2*",  2,
            {{ 103.100, 0.9, "C2'-endo" },
             { 102.200, 0.7, "C3'-endo" }}
       },
       { "C1'-C2'-C3'",   "C1*-C2*-C3*", "C3*", "C2*", "C1*",  2,
            {{ 102.500, 1.2, "C2'-endo" },
             { 102.400, 0.8, "C3'-endo" }}
       },
       { "O4'-C1'-C2'",   "C2*-C1*-O4*", "O4*", "C1*", "C2*",  2,
            {{ 105.900, 0.8, "C2'-endo" },
             { 106.800, 0.5, "C3'-endo" }}
       },
       { "C2'-C1'-N1/N9", "C2*-C1*-N9",  "N9",  "C1*", "C2*",  2,
            {{ 114.300, 1.4, "C2'-endo" },
             { 112.600, 1.9, "C3'-endo" }}
       },
       { "O4'-C1'-N1/N9", "N9-C1*-O4*",  "O4*", "C1*", "N9",   2,
            {{ 108.000, 0.7, "C2'-endo" },
             { 108.300, 0.3, "C3'-endo" }}
       },
       { "C6-N1-C2",    "C2-N1-C6",    "C6",  "N1",  "C2",   1,
             { { 125.100, 0.6, "s" } }
       },
       { "N1-C2-N3",    "N1-C2-N3",    "N1",  "C2",  "N3",   1,
             { { 123.900, 0.6, "s" } }
       },
       { "C2-N3-C4",    "C2-N3-C4",    "C2",  "N3",  "C4",   1,
             { { 111.900, 0.5, "s" } }
       },
       { "N3-C4-C5",    "C5-C4-N3",    "N3",  "C4",  "C5",   1,
             { { 128.600, 0.5, "s" } }
       },
       { "C4-C5-C6",    "C4-C5-C6",    "C4",  "C5",  "C6",   1,
             { { 118.800, 0.6, "s" } }
       },
       { "C5-C6-N1",    "C5-C6-N1",    "C5",  "C6",  "N1",   1,
             { { 111.500, 0.5, "s" } }
       },
       { "C4-C5-N7",    "C4-C5-N7",    "C4",  "C5",  "N7",   1,
             { { 110.800, 0.4, "s" } }
       },
       { "C5-N7-C8",    "C5-N7-C8",    "C5",  "N7",  "C8",   1,
             { { 104.300, 0.5, "s" } }
       },
       { "N7-C8-N9",    "N7-C8-N9",    "N7",  "C8",  "N9",   1,
             { { 113.100, 0.5, "s" } }
       },
       { "C8-N9-C4",    "C4-N9-C8",    "C8",  "N9",  "C4",   1,
             { { 106.400, 0.4, "s" } }
       },
       { "N9-C4-C5",    "C5-C4-N9",    "N9",  "C4",  "C5",   1,
             { { 105.400, 0.4, "s" } }
       },
       { "N3-C4-N9",    "N3-C4-N9",    "N3",  "C4",  "N9",   1,
             { { 126.000, 0.6, "s" } }
       },
       { "C6-C5-N7",    "C6-C5-N7",    "C6",  "C5",  "N7",   1,
             { { 130.400, 0.6, "s" } }
       },
       { "N1-C2-N2",    "N1-C2-N2",    "N1",  "C2",  "N2",   1,
             { { 116.200, 0.9, "s" } }
       },
       { "N3-C2-N2",    "N2-C2-N3",    "N3",  "C2",  "N2",   1,
             { { 119.900, 0.7, "s" } }
       },
       { "N1-C6-O6",    "N1-C6-O6",    "N1",  "C6",  "O6",   1,
             { { 119.900, 0.6, "s" } }
       },
       { "C5-C6-O6",    "C5-C6-O6",    "C5",  "C6",  "O6",   1,
             { { 128.600, 0.6, "s" } }
       },
       { "C8-N9-C1'",   "C1*-N9-C8",   "C8",  "N9",  "C1*",  1,
             { { 127.000, 1.3, "s" } }
       },
       { "C4-N9-C1'",   "C1*-N9-C4",   "C4",  "N9",  "C1*",  1,
             { { 126.500, 1.3, "s" } }
       }
     }
};

MONOMER RG = { "G", 25, 37, // 26, 41,
     {
       { "P-O1P/O2P", "O1P-P",   "P",   "O1P",  1,  { { 1.485, 0.017, "s" } } },
       { "P-O1P/O2P", "O2P-P",   "P",   "O2P",  1,  { { 1.485, 0.017, "s" } } },
       { "P-O5'",     "O5*-P",   "P",   "O5*",  1,  { { 1.593, 0.010, "s" } } },
       // { "P-O3'",     "O3*-P",   "P",   "O3*",  1,  { { 1.607, 0.012, "s" } } },
       { "O5'-C5'",   "C5*-O5*", "O5*", "C5*",  3, {{ 1.440, 0.016, "(P)O5'-C5'" },
                                                    { 1.424, 0.016, "(H)C2'-endo" },
                                                    { 1.420, 0.009, "(H)C3'-endo" }}
       },
       { "C5'-C4'",   "C4*-C5*", "C5*", "C4*",  2, {{ 1.509, 0.012, "C2'-endo" },
                                                    { 1.508, 0.007, "C3'-endo" }}
       },
       { "C3'-C4'",   "C3*-C4*", "C4*", "C3*",  2, {{ 1.527, 0.011, "C2'-endo" },
                                                    { 1.521, 0.010, "C3'-endo" }}
       },
       { "C2'-C3'",   "C2*-C3*", "C3*", "C2*",  2, {{ 1.525, 0.011, "C2'-endo" },
                                                    { 1.523, 0.011, "C3'-endo" }}
       },
       { "C1'-C2'",   "C1*-C2*", "C2*", "C1*",  2, {{ 1.526, 0.008, "C2'-endo" },
                                                    { 1.529, 0.011, "C3'-endo" }}
       },
       { "O4'-C1'",   "C1*-O4*", "O4*", "C1*",  2, {{ 1.415, 0.012, "C2'-endo" },
                                                    { 1.412, 0.013, "C3'-endo" }}
       },
       { "C4'-O4'",   "C4*-O4*", "O4*", "C4*",  2, {{ 1.454, 0.010, "C2'-endo" },
                                                    { 1.451, 0.013, "C3'-endo" }}
       },
       { "C3'-O3'",   "C3*-O3*", "O3*", "C3*",  2, {{ 1.427, 0.012, "C2'-endo" },
                                                    { 1.417, 0.014, "C3'-endo" }}
       },
       { "C1'-N1/N9", "C1*-N9",  "C1*", "N9",   2, {{ 1.464, 0.014, "C2'-endo" },
                                                    { 1.483, 0.015, "C3'-endo" }}
       },
       { "C2'-O2'",   "C2*-O2*", "C2*", "O2*",  2, {{ 1.412, 0.013, "C2'-endo" },
                                                    { 1.420, 0.010, "C3'-endo" }}
       },
       { "N1-C2",   "C2-N1",   "N1",  "C2",   1,  { { 1.373, 0.008, "s" } } },
       { "C2-N3",   "C2-N3",   "C2",  "N3",   1,  { { 1.323, 0.008, "s" } } },
       { "N3-C4",   "C4-N3",   "N3",  "C4",   1,  { { 1.350, 0.007, "s" } } },
       { "C4-C5",   "C4-C5",   "C4",  "C5",   1,  { { 1.379, 0.007, "s" } } },
       { "C5-C6",   "C5-C6",   "C5",  "C6",   1,  { { 1.419, 0.010, "s" } } },
       { "C6-N1",   "C6-N1",   "C6",  "N1",   1,  { { 1.391, 0.007, "s" } } },
       { "C5-N7",   "C5-N7",   "C5",  "N7",   1,  { { 1.388, 0.006, "s" } } },
       { "N7-C8",   "C8-N7",   "N7",  "C8",   1,  { { 1.305, 0.006, "s" } } },
       { "C8-N9",   "C8-N9",   "C8",  "N9",   1,  { { 1.374, 0.007, "s" } } },
       { "N9-C4",   "C4-N9",   "N9",  "C4",   1,  { { 1.375, 0.008, "s" } } },
       { "C2-N2",   "C2-N2",   "C2",  "N2",   1,  { { 1.341, 0.010, "s" } } },
       { "C6-O6",   "C6-O6",   "C6",  "O6",   1,  { { 1.237, 0.009, "s" } } }
     },
     {
       { "O1P-P-O2P",     "O1P-P-O2P",   "O1P", "P",   "O2P",  1,
             { { 119.600, 1.5, "s" } }
       },
       { "O5'-P-O1P/O2P", "O1P-P-O5*",   "O5*", "P",   "O1P",  2,
            {{ 110.700, 1.2, "large" },
             { 105.700, 0.9, "small" }}
       },
       { "O5'-P-O1P/O2P", "O2P-P-O5*",   "O5*", "P",   "O2P",  2,
            {{ 110.700, 1.2, "large" },
             { 105.700, 0.9, "small" }}
       },
/*
       { "O3'-P-O5'",     "O3*-P-O5*",   "O3*", "P",   "O5*",  1,
             { { 104.000, 1.9, "s" } }
       },
       { "O3'-P-O2P/O2P", "O2P-P-O3*",   "O2P", "P",   "O3*",  2,
            {{ 110.500, 1.1, "large" },
             { 105.200, 2.2, "small" }}
       },
       { "O3'-P-O1P/O2P", "O1P-P-O3*",   "O1P", "P",   "O3*",  2,
            {{ 110.500, 1.1, "large" },
             { 105.200, 2.2, "small" }}
       },
*/
       { "O5'-C5'-C4'",   "C4*-C5*-O5*", "O5*", "C5*", "C4*",  3,
            {{ 109.400, 0.8, "(P)O5'-C5'-C4'" },
             { 111.700, 1.9, "C2'-endo" },
             { 111.500, 1.6, "C3'-endo" }}
       },
       { "P-O5'-C5'",     "C5*-O5*-P",   "P",   "O5*", "C5*",  1,
             { { 120.900, 1.6, "s" } }
       },
/*
       { "C3'-O3'-P",     "C3*-O3*-P",   "P",   "O3*", "C3*",  1,
             { { 119.700, 1.2, "s" } }
       },
*/
       { "C3'-C4'-O4'",   "C3*-C4*-O4*",  "O4*", "C4*", "C3*",  2,
            {{ 106.100, 0.8, "C2'-endo" },
             { 104.000, 1.0, "C3'-endo" }}
       },
       { "C5'-C4'-C3'",   "C3*-C4*-C5*", "C5*", "C4*", "C3*",  2,
            {{ 115.200, 1.4, "C2'-endo" },
             { 116.000, 1.6, "C3'-endo" }}
       },
       { "C5'-C4'-O4'",   "C5*-C4*-O4*", "C5*", "C4*", "O4*",  2,
            {{ 109.100, 1.2, "C2'-endo" },
             { 109.800, 0.9, "C3'-endo" }}
       },
       { "C4'-O4'-C1'",   "C1*-O4*-C4*", "C1*", "O4*", "C4*",  2,
            {{ 109.700, 0.7, "C2'-endo" },
             { 109.900, 0.8, "C3'-endo" }}
       },
       { "C4'-C3'-O3'",   "C4*-C3*-O3*", "C4*", "C3*", "O3*",  2,
            {{ 109.400, 2.1, "C2'-endo" },
             { 113.000, 2.0, "C3'-endo" }}
       },
       { "C2'-C3'-O3'",   "C2*-C3*-O3*", "C2*", "C3*", "O3*",  2,
            {{ 109.500, 2.2, "C2'-endo" },
             { 113.700, 1.6, "C3'-endo" }}
       },
       { "C2'-C3'-C4'",   "C2*-C3*-C4*", "C4*", "C3*", "C2*",  2,
            {{ 102.600, 1.0, "C2'-endo" },
             { 102.600, 1.0, "C3'-endo" }}
       },
       { "C1'-C2'-C3'",   "C1*-C2*-C3*", "C3*", "C2*", "C1*",  2,
            {{ 101.500, 0.8, "C2'-endo" },
             { 101.300, 0.7, "C3'-endo" }}
       },
       { "O4'-C1'-C2'",   "C2*-C1*-O4*", "O4*", "C1*", "C2*",  2,
            {{ 105.800, 1.0, "C2'-endo" },
             { 107.600, 0.9, "C3'-endo" }}
       },
       { "C2'-C1'-N1/N9", "C2*-C1*-N9",  "N9",  "C1*", "C2*",  2,
            {{ 114.000, 1.3, "C2'-endo" },
             { 112.000, 1.1, "C3'-endo" }}
       },
       { "O4'-C1'-N1/N9", "N9-C1*-O4*",  "O4*", "C1*", "N9",   2,
            {{ 108.200, 0.8, "C2'-endo" },
             { 108.500, 0.7, "C3'-endo" }}
       },
       { "C1'-C2'-O2'",   "C1*-C2*-O2*", "C1*", "C2*", "O2*",  1,
             { { 110.600, 3.0, "s" } }
       },
       { "C3'-C2'-O2'",   "C3*-C2*-O2*", "C3*", "C2*", "O2*",  1,
             { { 113.300, 2.9, "s" } }
       },
       { "C6-N1-C2",    "C2-N1-C6",    "C6",  "N1",  "C2",   1,
             { { 125.100, 0.6, "s" } }
       },
       { "N1-C2-N3",    "N1-C2-N3",    "N1",  "C2",  "N3",   1,
             { { 123.900, 0.6, "s" } }
       },
       { "C2-N3-C4",    "C2-N3-C4",    "C2",  "N3",  "C4",   1,
             { { 111.900, 0.5, "s" } }
       },
       { "N3-C4-C5",    "C5-C4-N3",    "N3",  "C4",  "C5",   1,
             { { 128.600, 0.5, "s" } }
       },
       { "C4-C5-C6",    "C4-C5-C6",    "C4",  "C5",  "C6",   1,
             { { 118.800, 0.6, "s" } }
       },
       { "C5-C6-N1",    "C5-C6-N1",    "C5",  "C6",  "N1",   1,
             { { 111.500, 0.5, "s" } }
       },
       { "C4-C5-N7",    "C4-C5-N7",    "C4",  "C5",  "N7",   1,
             { { 110.800, 0.4, "s" } }
       },
       { "C5-N7-C8",    "C5-N7-C8",    "C5",  "N7",  "C8",   1,
             { { 104.300, 0.5, "s" } }
       },
       { "N7-C8-N9",    "N7-C8-N9",    "N7",  "C8",  "N9",   1,
             { { 113.100, 0.5, "s" } }
       },
       { "C8-N9-C4",    "C4-N9-C8",    "C8",  "N9",  "C4",   1,
             { { 106.400, 0.4, "s" } }
       },
       { "N9-C4-C5",    "C5-C4-N9",    "N9",  "C4",  "C5",   1,
             { { 105.400, 0.4, "s" } }
       },
       { "N3-C4-N9",    "N3-C4-N9",    "N3",  "C4",  "N9",   1,
             { { 126.000, 0.6, "s" } }
       },
       { "C6-C5-N7",    "C6-C5-N7",    "C6",  "C5",  "N7",   1,
             { { 130.400, 0.6, "s" } }
       },
       { "N1-C2-N2",    "N1-C2-N2",    "N1",  "C2",  "N2",   1,
             { { 116.200, 0.9, "s" } }
       },
       { "N3-C2-N2",    "N2-C2-N3",    "N3",  "C2",  "N2",   1,
             { { 119.900, 0.7, "s" } }
       },
       { "N1-C6-O6",    "N1-C6-O6",    "N1",  "C6",  "O6",   1,
             { { 119.900, 0.6, "s" } }
       },
       { "C5-C6-O6",    "C5-C6-O6",    "C5",  "C6",  "O6",   1,
             { { 128.600, 0.6, "s" } }
       },
       { "C8-N9-C1'",   "C1*-N9-C8",   "C8",  "N9",  "C1*",  1,
             { { 127.000, 1.3, "s" } }
       },
       { "C4-N9-C1'",   "C1*-N9-C4",   "C4",  "N9",  "C1*",  1,
             { { 126.500, 1.3, "s" } }
       }
     }
};

MONOMER DT = { "T", 21, 30, // 22, 34,
     {
       { "P-O1P/O2P", "O1P-P",   "P",   "O1P",  1,  { { 1.485, 0.017, "s" } } },
       { "P-O1P/O2P", "O2P-P",   "P",   "O2P",  1,  { { 1.485, 0.017, "s" } } },
       { "P-O5'",     "O5*-P",   "P",   "O5*",  1,  { { 1.593, 0.010, "s" } } },
       // { "P-O3'",     "O3*-P",   "P",   "O3*",  1,  { { 1.607, 0.012, "s" } } },
       { "O5'-C5'",   "C5*-O5*", "O5*", "C5*",  3, {{ 1.440, 0.016, "(P)O5'-C5'" },
                                                    { 1.418, 0.025, "(H)C2'-endo" },
                                                    { 1.423, 0.011, "(H)C3'-endo" }}
       },
       { "C5'-C4'",   "C4*-C5*", "C5*", "C4*",  2, {{ 1.512, 0.007, "C2'-endo" },
                                                    { 1.509, 0.011, "C3'-endo" }}
       },
       { "C3'-C4'",   "C3*-C4*", "C4*", "C3*",  2, {{ 1.529, 0.010, "C2'-endo" },
                                                    { 1.521, 0.010, "C3'-endo" }}
       },
       { "C2'-C3'",   "C2*-C3*", "C3*", "C2*",  2, {{ 1.516, 0.008, "C2'-endo" },
                                                    { 1.518, 0.012, "C3'-endo" }}
       },
       { "C1'-C2'",   "C1*-C2*", "C2*", "C1*",  2, {{ 1.518, 0.010, "C2'-endo" },
                                                    { 1.519, 0.010, "C3'-endo" }}
       },
       { "O4'-C1'",   "C1*-O4*", "O4*", "C1*",  2, {{ 1.420, 0.011, "C2'-endo" },
                                                    { 1.418, 0.012, "C3'-endo" }}
       },
       { "C4'-O4'",   "C4*-O4*", "O4*", "C4*",  2, {{ 1.446, 0.010, "C2'-endo" },
                                                    { 1.449, 0.009, "C3'-endo" }}
       },
       { "C3'-O3'",   "C3*-O3*", "O3*", "C3*",  2, {{ 1.435, 0.013, "C2'-endo" },
                                                    { 1.419, 0.006, "C3'-endo" }}
       },
       { "C1'-N1/N9", "C1*-N1",  "C1*", "N1",   2, {{ 1.468, 0.014, "C2'-endo" },
                                                    { 1.488, 0.013, "C3'-endo" }}
       },
       { "N1-C2",   "C2-N1",   "N1",  "C2",   1,  { { 1.376, 0.008, "s" } } },
       { "C2-N3",   "C2-N3",   "C2",  "N3",   1,  { { 1.373, 0.008, "s" } } },
       { "N3-C4",   "C4-N3",   "N3",  "C4",   1,  { { 1.382, 0.008, "s" } } },
       { "C4-C5",   "C4-C5",   "C4",  "C5",   1,  { { 1.445, 0.009, "s" } } },
       { "C5-C6",   "C5-C6",   "C5",  "C6",   1,  { { 1.339, 0.007, "s" } } },
       { "C6-N1",   "C6-N1",   "C6",  "N1",   1,  { { 1.378, 0.007, "s" } } },
       { "C2-O2",   "C2-O2",   "C2",  "O2",   1,  { { 1.220, 0.008, "s" } } },
       { "C4-O4",   "C4-O4",   "C4",  "O4",   1,  { { 1.228, 0.009, "s" } } },
       { "C5-C5M",  "C5-C5M",  "C5",  "C5M",  1,  { { 1.496, 0.006, "s" } } }
     },
     {
       { "O1P-P-O2P",     "O1P-P-O2P",   "O1P", "P",   "O2P",  1,
             { { 119.600, 1.5, "s" } }
       },
       { "O5'-P-O1P/O2P", "O1P-P-O5*",   "O5*", "P",   "O1P",  2,
            {{ 110.700, 1.2, "large" },
             { 105.700, 0.9, "small" }}
       },
       { "O5'-P-O1P/O2P", "O2P-P-O5*",   "O5*", "P",   "O2P",  2,
            {{ 110.700, 1.2, "large" },
             { 105.700, 0.9, "small" }}
       },
/*
       { "O3'-P-O5'",     "O3*-P-O5*",   "O3*", "P",   "O5*",  1,
             { { 104.000, 1.9, "s" } }
       },
       { "O3'-P-O2P/O2P", "O2P-P-O3*",   "O2P", "P",   "O3*",  2,
            {{ 110.500, 1.1, "large" },
             { 105.200, 2.2, "small" }}
       },
       { "O3'-P-O1P/O2P", "O1P-P-O3*",   "O1P", "P",   "O3*",  2,
            {{ 110.500, 1.1, "large" },
             { 105.200, 2.2, "small" }}
       },
*/
       { "O5'-C5'-C4'",   "C4*-C5*-O5*", "O5*", "C5*", "C4*",  3,
            {{ 109.400, 0.8, "(P)O5'-C5'-C4'" },
             { 110.900, 1.7, "C2'-endo" },
             { 111.000, 2.5, "C3'-endo" }}
       },
       { "P-O5'-C5'",     "C5*-O5*-P",   "P",   "O5*", "C5*",  1,
             { { 120.900, 1.6, "s" } }
       },
/*
       { "C3'-O3'-P",     "C3*-O3*-P",   "P",   "O3*", "C3*",  1,
             { { 119.700, 1.2, "s" } }
       },
*/
       { "C3'-C4'-O4'",   "C3*-C4*-O4*",  "O4*", "C4*", "C3*",  2,
            {{ 106.000, 0.6, "C2'-endo" },
             { 104.500, 0.4, "C3'-endo" }}
       },
       { "C5'-C4'-C3'",   "C3*-C4*-C5*", "C5*", "C4*", "C3*",  2,
            {{ 114.100, 1.8, "C2'-endo" },
             { 115.700, 1.2, "C3'-endo" }}
       },
       { "C5'-C4'-O4'",   "C5*-C4*-O4*", "C5*", "C4*", "O4*",  2,
            {{ 109.300, 1.9, "C2'-endo" },
             { 109.800, 1.1, "C3'-endo" }}
       },
       { "C4'-O4'-C1'",   "C1*-O4*-C4*", "C1*", "O4*", "C4*",  2,
            {{ 110.100, 1.0, "C2'-endo" },
             { 110.300, 0.7, "C3'-endo" }}
       },
       { "C4'-C3'-O3'",   "C4*-C3*-O3*", "C4*", "C3*", "O3*",  2,
            {{ 109.700, 2.5, "C2'-endo" },
             { 112.300, 2.0, "C3'-endo" }}
       },
       { "C2'-C3'-O3'",   "C2*-C3*-O3*", "C2*", "C3*", "O3*",  2,
            {{ 109.400, 2.5, "C2'-endo" },
             { 112.600, 3.3, "C3'-endo" }}
       },
       { "C2'-C3'-C4'",   "C2*-C3*-C4*", "C4*", "C3*", "C2*",  2,
            {{ 103.100, 0.9, "C2'-endo" },
             { 102.200, 0.7, "C3'-endo" }}
       },
       { "C1'-C2'-C3'",   "C1*-C2*-C3*", "C3*", "C2*", "C1*",  2,
            {{ 102.500, 1.2, "C2'-endo" },
             { 102.400, 0.8, "C3'-endo" }}
       },
       { "O4'-C1'-C2'",   "C2*-C1*-O4*", "O4*", "C1*", "C2*",  2,
            {{ 105.900, 0.8, "C2'-endo" },
             { 106.800, 0.5, "C3'-endo" }}
       },
       { "C2'-C1'-N1/N9", "C2*-C1*-N1",  "N1",  "C1*", "C2*",  2,
            {{ 114.300, 1.4, "C2'-endo" },
             { 112.600, 1.9, "C3'-endo" }}
       },
       { "O4'-C1'-N1/N9", "N1-C1*-O4*",  "O4*", "C1*", "N1",   2,
            {{ 108.000, 0.7, "C2'-endo" },
             { 108.300, 0.3, "C3'-endo" }}
       },
       { "C6-N1-C2",    "C2-N1-C6",    "C6",  "N1",  "C2",   1,
             { { 121.300, 0.5, "s" } }
       },
       { "N1-C2-N3",    "N1-C2-N3",    "N1",  "C2",  "N3",   1,
             { { 114.600, 0.6, "s" } }
       },
       { "C2-N3-C4",    "C2-N3-C4",    "C2",  "N3",  "C4",   1,
             { { 127.200, 0.6, "s" } }
       },
       { "N3-C4-C5",    "C5-C4-N3",    "N3",  "C4",  "C5",   1,
             { { 115.200, 0.6, "s" } }
       },
       { "C4-C5-C6",    "C4-C5-C6",    "C4",  "C5",  "C6",   1,
             { { 118.000, 0.6, "s" } }
       },
       { "C5-C6-N1",    "C5-C6-N1",    "C5",  "C6",  "N1",   1,
             { { 123.700, 0.6, "s" } }
       },
       { "N1-C2-O2",    "N1-C2-O2",    "N1",  "C2",  "O2",   1,
             { { 123.100, 0.8, "s" } }
       },
       { "N3-C2-O2",    "N3-C2-O2",    "N3",  "C2",  "O2",   1,
             { { 122.300, 0.6, "s" } }
       },
       { "N3-C4-O4",    "N3-C4-O4",    "N3",  "C4",  "O4",   1,
             { { 119.900, 0.6, "s" } }
       },
       { "C5-C4-O4",    "C5-C4-O4",    "C5",  "C4",  "O4",   1,
             { { 124.900, 0.7, "s" } }
       },
       { "C4-C5-C5M",   "C4-C5-C5M",   "C4",  "C5",  "C5M",  1,
             { { 119.000, 0.6, "s" } }
       },
       { "C6-C5-C5M",   "C5M-C5-C6",   "C6",  "C5",  "C5M",  1,
             { { 122.900, 0.6, "s" } }
       },
       { "C6-N1-C1'",   "C1*-N1-C6",   "C6",  "N1",  "C1*",  1,
             { { 120.400, 1.5, "s" } }
       },
       { "C2-N1-C1'",   "C1*-N1-C2",   "C2",  "N1",  "C1*",  1,
             { { 118.200, 1.6, "s" } }
       }
     }
};

MONOMER RT = { "T", 22, 32, // 23, 36,
     {
       { "P-O1P/O2P", "O1P-P",   "P",   "O1P",  1,  { { 1.485, 0.017, "s" } } },
       { "P-O1P/O2P", "O2P-P",   "P",   "O2P",  1,  { { 1.485, 0.017, "s" } } },
       { "P-O5'",     "O5*-P",   "P",   "O5*",  1,  { { 1.593, 0.010, "s" } } },
       // { "P-O3'",     "O3*-P",   "P",   "O3*",  1,  { { 1.607, 0.012, "s" } } },
       { "O5'-C5'",   "C5*-O5*", "O5*", "C5*",  3, {{ 1.440, 0.016, "(P)O5'-C5'" },
                                                    { 1.424, 0.016, "(H)C2'-endo" },
                                                    { 1.420, 0.009, "(H)C3'-endo" }}
       },
       { "C5'-C4'",   "C4*-C5*", "C5*", "C4*",  2, {{ 1.509, 0.012, "C2'-endo" },
                                                    { 1.508, 0.007, "C3'-endo" }}
       },
       { "C3'-C4'",   "C3*-C4*", "C4*", "C3*",  2, {{ 1.527, 0.011, "C2'-endo" },
                                                    { 1.521, 0.010, "C3'-endo" }}
       },
       { "C2'-C3'",   "C2*-C3*", "C3*", "C2*",  2, {{ 1.525, 0.011, "C2'-endo" },
                                                    { 1.523, 0.011, "C3'-endo" }}
       },
       { "C1'-C2'",   "C1*-C2*", "C2*", "C1*",  2, {{ 1.526, 0.008, "C2'-endo" },
                                                    { 1.529, 0.011, "C3'-endo" }}
       },
       { "O4'-C1'",   "C1*-O4*", "O4*", "C1*",  2, {{ 1.415, 0.012, "C2'-endo" },
                                                    { 1.412, 0.013, "C3'-endo" }}
       },
       { "C4'-O4'",   "C4*-O4*", "O4*", "C4*",  2, {{ 1.454, 0.010, "C2'-endo" },
                                                    { 1.451, 0.013, "C3'-endo" }}
       },
       { "C3'-O3'",   "C3*-O3*", "O3*", "C3*",  2, {{ 1.427, 0.012, "C2'-endo" },
                                                    { 1.417, 0.014, "C3'-endo" }}
       },
       { "C1'-N1/N9", "C1*-N1",  "C1*", "N1",   2, {{ 1.464, 0.014, "C2'-endo" },
                                                    { 1.483, 0.015, "C3'-endo" }}
       },
       { "C2'-O2'",   "C2*-O2*", "C2*", "O2*",  2, {{ 1.412, 0.013, "C2'-endo" },
                                                    { 1.420, 0.010, "C3'-endo" }}
       },
       { "N1-C2",   "C2-N1",   "N1",  "C2",   1,  { { 1.376, 0.008, "s" } } },
       { "C2-N3",   "C2-N3",   "C2",  "N3",   1,  { { 1.373, 0.008, "s" } } },
       { "N3-C4",   "C4-N3",   "N3",  "C4",   1,  { { 1.382, 0.008, "s" } } },
       { "C4-C5",   "C4-C5",   "C4",  "C5",   1,  { { 1.445, 0.009, "s" } } },
       { "C5-C6",   "C5-C6",   "C5",  "C6",   1,  { { 1.339, 0.007, "s" } } },
       { "C6-N1",   "C6-N1",   "C6",  "N1",   1,  { { 1.378, 0.007, "s" } } },
       { "C2-O2",   "C2-O2",   "C2",  "O2",   1,  { { 1.220, 0.008, "s" } } },
       { "C4-O4",   "C4-O4",   "C4",  "O4",   1,  { { 1.228, 0.009, "s" } } },
       { "C5-C5M",  "C5-C5M",  "C5",  "C5M",  1,  { { 1.496, 0.006, "s" } } }
     },
     {
       { "O1P-P-O2P",     "O1P-P-O2P",   "O1P", "P",   "O2P",  1,
             { { 119.600, 1.5, "s" } }
       },
       { "O5'-P-O1P/O2P", "O1P-P-O5*",   "O5*", "P",   "O1P",  2,
            {{ 110.700, 1.2, "large" },
             { 105.700, 0.9, "small" }}
       },
       { "O5'-P-O1P/O2P", "O2P-P-O5*",   "O5*", "P",   "O2P",  2,
            {{ 110.700, 1.2, "large" },
             { 105.700, 0.9, "small" }}
       },
/*
       { "O3'-P-O5'",     "O3*-P-O5*",   "O3*", "P",   "O5*",  1,
             { { 104.000, 1.9, "s" } }
       },
       { "O3'-P-O2P/O2P", "O2P-P-O3*",   "O2P", "P",   "O3*",  2,
            {{ 110.500, 1.1, "large" },
             { 105.200, 2.2, "small" }}
       },
       { "O3'-P-O1P/O2P", "O1P-P-O3*",   "O1P", "P",   "O3*",  2,
            {{ 110.500, 1.1, "large" },
             { 105.200, 2.2, "small" }}
       },
*/
       { "O5'-C5'-C4'",   "C4*-C5*-O5*", "O5*", "C5*", "C4*",  3,
            {{ 109.400, 0.8, "(P)O5'-C5'-C4'" },
             { 111.700, 1.9, "C2'-endo" },
             { 111.500, 1.6, "C3'-endo" }}
       },
       { "P-O5'-C5'",     "C5*-O5*-P",   "P",   "O5*", "C5*",  1,
             { { 120.900, 1.6, "s" } }
       },
/*
       { "C3'-O3'-P",     "C3*-O3*-P",   "P",   "O3*", "C3*",  1,
             { { 119.700, 1.2, "s" } }
       },
*/
       { "C3'-C4'-O4'",   "C3*-C4*-O4*",  "O4*", "C4*", "C3*",  2,
            {{ 106.100, 0.8, "C2'-endo" },
             { 104.000, 1.0, "C3'-endo" }}
       },
       { "C5'-C4'-C3'",   "C3*-C4*-C5*", "C5*", "C4*", "C3*",  2,
            {{ 115.200, 1.4, "C2'-endo" },
             { 116.000, 1.6, "C3'-endo" }}
       },
       { "C5'-C4'-O4'",   "C5*-C4*-O4*", "C5*", "C4*", "O4*",  2,
            {{ 109.100, 1.2, "C2'-endo" },
             { 109.800, 0.9, "C3'-endo" }}
       },
       { "C4'-O4'-C1'",   "C1*-O4*-C4*", "C1*", "O4*", "C4*",  2,
            {{ 109.700, 0.7, "C2'-endo" },
             { 109.900, 0.8, "C3'-endo" }}
       },
       { "C4'-C3'-O3'",   "C4*-C3*-O3*", "C4*", "C3*", "O3*",  2,
            {{ 109.400, 2.1, "C2'-endo" },
             { 113.000, 2.0, "C3'-endo" }}
       },
       { "C2'-C3'-O3'",   "C2*-C3*-O3*", "C2*", "C3*", "O3*",  2,
            {{ 109.500, 2.2, "C2'-endo" },
             { 113.700, 1.6, "C3'-endo" }}
       },
       { "C2'-C3'-C4'",   "C2*-C3*-C4*", "C4*", "C3*", "C2*",  2,
            {{ 102.600, 1.0, "C2'-endo" },
             { 102.600, 1.0, "C3'-endo" }}
       },
       { "C1'-C2'-C3'",   "C1*-C2*-C3*", "C3*", "C2*", "C1*",  2,
            {{ 101.500, 0.8, "C2'-endo" },
             { 101.300, 0.7, "C3'-endo" }}
       },
       { "O4'-C1'-C2'",   "C2*-C1*-O4*", "O4*", "C1*", "C2*",  2,
            {{ 105.800, 1.0, "C2'-endo" },
             { 107.600, 0.9, "C3'-endo" }}
       },
       { "C2'-C1'-N1/N9", "C2*-C1*-N1",  "N1",  "C1*", "C2*",  2,
            {{ 114.000, 1.3, "C2'-endo" },
             { 112.000, 1.1, "C3'-endo" }}
       },
       { "O4'-C1'-N1/N9", "N1-C1*-O4*",  "O4*", "C1*", "N1",   2,
            {{ 108.200, 0.8, "C2'-endo" },
             { 108.500, 0.7, "C3'-endo" }}
       },
       { "C1'-C2'-O2'",   "C1*-C2*-O2*", "C1*", "C2*", "O2*",  1,
             { { 110.600, 3.0, "s" } }
       },
       { "C3'-C2'-O2'",   "C3*-C2*-O2*", "C3*", "C2*", "O2*",  1,
             { { 113.300, 2.9, "s" } }
       },
       { "C6-N1-C2",    "C2-N1-C6",    "C6",  "N1",  "C2",   1,
             { { 121.300, 0.5, "s" } }
       },
       { "N1-C2-N3",    "N1-C2-N3",    "N1",  "C2",  "N3",   1,
             { { 114.600, 0.6, "s" } }
       },
       { "C2-N3-C4",    "C2-N3-C4",    "C2",  "N3",  "C4",   1,
             { { 127.200, 0.6, "s" } }
       },
       { "N3-C4-C5",    "C5-C4-N3",    "N3",  "C4",  "C5",   1,
             { { 115.200, 0.6, "s" } }
       },
       { "C4-C5-C6",    "C4-C5-C6",    "C4",  "C5",  "C6",   1,
             { { 118.000, 0.6, "s" } }
       },
       { "C5-C6-N1",    "C5-C6-N1",    "C5",  "C6",  "N1",   1,
             { { 123.700, 0.6, "s" } }
       },
       { "N1-C2-O2",    "N1-C2-O2",    "N1",  "C2",  "O2",   1,
             { { 123.100, 0.8, "s" } }
       },
       { "N3-C2-O2",    "N3-C2-O2",    "N3",  "C2",  "O2",   1,
             { { 122.300, 0.6, "s" } }
       },
       { "N3-C4-O4",    "N3-C4-O4",    "N3",  "C4",  "O4",   1,
             { { 119.900, 0.6, "s" } }
       },
       { "C5-C4-O4",    "C5-C4-O4",    "C5",  "C4",  "O4",   1,
             { { 124.900, 0.7, "s" } }
       },
       { "C4-C5-C5M",   "C4-C5-C5M",   "C4",  "C5",  "C5M",  1,
             { { 119.000, 0.6, "s" } }
       },
       { "C6-C5-C5M",   "C5M-C5-C6",   "C6",  "C5",  "C5M",  1,
             { { 122.900, 0.6, "s" } }
       },
       { "C6-N1-C1'",   "C1*-N1-C6",   "C6",  "N1",  "C1*",  1,
             { { 120.400, 1.5, "s" } }
       },
       { "C2-N1-C1'",   "C1*-N1-C2",   "C2",  "N1",  "C1*",  1,
             { { 118.200, 1.6, "s" } }
       }
     }
};

MONOMER DU = { "U", 20, 28, // 21, 32,
     {
       { "P-O1P/O2P", "O1P-P",   "P",   "O1P",  1,  { { 1.485, 0.017, "s" } } },
       { "P-O1P/O2P", "O2P-P",   "P",   "O2P",  1,  { { 1.485, 0.017, "s" } } },
       { "P-O5'",     "O5*-P",   "P",   "O5*",  1,  { { 1.593, 0.010, "s" } } },
       // { "P-O3'",     "O3*-P",   "P",   "O3*",  1,  { { 1.607, 0.012, "s" } } },
       { "O5'-C5'",   "C5*-O5*", "O5*", "C5*",  3, {{ 1.440, 0.016, "(P)O5'-C5'" },
                                                    { 1.418, 0.025, "(H)C2'-endo" },
                                                    { 1.423, 0.011, "(H)C3'-endo" }}
       },
       { "C5'-C4'",   "C4*-C5*", "C5*", "C4*",  2, {{ 1.512, 0.007, "C2'-endo" },
                                                    { 1.509, 0.011, "C3'-endo" }}
       },
       { "C3'-C4'",   "C3*-C4*", "C4*", "C3*",  2, {{ 1.529, 0.010, "C2'-endo" },
                                                    { 1.521, 0.010, "C3'-endo" }}
       },
       { "C2'-C3'",   "C2*-C3*", "C3*", "C2*",  2, {{ 1.516, 0.008, "C2'-endo" },
                                                    { 1.518, 0.012, "C3'-endo" }}
       },
       { "C1'-C2'",   "C1*-C2*", "C2*", "C1*",  2, {{ 1.518, 0.010, "C2'-endo" },
                                                    { 1.519, 0.010, "C3'-endo" }}
       },
       { "O4'-C1'",   "C1*-O4*", "O4*", "C1*",  2, {{ 1.420, 0.011, "C2'-endo" },
                                                    { 1.418, 0.012, "C3'-endo" }}
       },
       { "C4'-O4'",   "C4*-O4*", "O4*", "C4*",  2, {{ 1.446, 0.010, "C2'-endo" },
                                                    { 1.449, 0.009, "C3'-endo" }}
       },
       { "C3'-O3'",   "C3*-O3*", "O3*", "C3*",  2, {{ 1.435, 0.013, "C2'-endo" },
                                                    { 1.419, 0.006, "C3'-endo" }}
       },
       { "C1'-N1/N9", "C1*-N1",  "C1*", "N1",   2, {{ 1.468, 0.014, "C2'-endo" },
                                                    { 1.488, 0.013, "C3'-endo" }}
       },
       { "C2-O2",   "C2-O2",   "C2",  "O2",   1,  { { 1.219, 0.009, "s" } } },
       { "C4-O4",   "C4-O4",   "C4",  "O4",   1,  { { 1.232, 0.008, "s" } } },
       { "N1-C2",   "C2-N1",   "N1",  "C2",   1,  { { 1.381, 0.009, "s" } } },
       { "C6-N1",   "C6-N1",   "N1",  "C6",   1,  { { 1.375, 0.009, "s" } } },
       { "C2-N3",   "C2-N3",   "C2",  "N3",   1,  { { 1.373, 0.007, "s" } } },
       { "N3-C4",   "C4-N3",   "N3",  "C4",   1,  { { 1.380, 0.009, "s" } } },
       { "C4-C5",   "C4-C5",   "C4",  "C5",   1,  { { 1.431, 0.009, "s" } } },
       { "C5-C6",   "C5-C6",   "C5",  "C6",   1,  { { 1.337, 0.009, "s" } } }
     },
     {
       { "O1P-P-O2P",     "O1P-P-O2P",   "O1P", "P",   "O2P",  1,
             { { 119.600, 1.5, "s" } }
       },
       { "O5'-P-O1P/O2P", "O1P-P-O5*",   "O5*", "P",   "O1P",  2,
            {{ 110.700, 1.2, "large" },
             { 105.700, 0.9, "small" }}
       },
       { "O5'-P-O1P/O2P", "O2P-P-O5*",   "O5*", "P",   "O2P",  2,
            {{ 110.700, 1.2, "large" },
             { 105.700, 0.9, "small" }}
       },
/*
       { "O3'-P-O5'",     "O3*-P-O5*",   "O3*", "P",   "O5*",  1,
             { { 104.000, 1.9, "s" } }
       },
       { "O3'-P-O2P/O2P", "O2P-P-O3*",   "O2P", "P",   "O3*",  2,
            {{ 110.500, 1.1, "large" },
             { 105.200, 2.2, "small" }}
       },
       { "O3'-P-O1P/O2P", "O1P-P-O3*",   "O1P", "P",   "O3*",  2,
            {{ 110.500, 1.1, "large" },
             { 105.200, 2.2, "small" }}
       },
*/
       { "O5'-C5'-C4'",   "C4*-C5*-O5*", "O5*", "C5*", "C4*",  3,
            {{ 109.400, 0.8, "(P)O5'-C5'-C4'" },
             { 110.900, 1.7, "C2'-endo" },
             { 111.000, 2.5, "C3'-endo" }}
       },
       { "P-O5'-C5'",     "C5*-O5*-P",   "P",   "O5*", "C5*",  1,
             { { 120.900, 1.6, "s" } }
       },
/*
       { "C3'-O3'-P",     "C3*-O3*-P",   "P",   "O3*", "C3*",  1,
             { { 119.700, 1.2, "s" } }
       },
*/
       { "C3'-C4'-O4'",   "C3*-C4*-O4*",  "O4*", "C4*", "C3*",  2,
            {{ 106.000, 0.6, "C2'-endo" },
             { 104.500, 0.4, "C3'-endo" }}
       },
       { "C5'-C4'-C3'",   "C3*-C4*-C5*", "C5*", "C4*", "C3*",  2,
            {{ 114.100, 1.8, "C2'-endo" },
             { 115.700, 1.2, "C3'-endo" }}
       },
       { "C5'-C4'-O4'",   "C5*-C4*-O4*", "C5*", "C4*", "O4*",  2,
            {{ 109.300, 1.9, "C2'-endo" },
             { 109.800, 1.1, "C3'-endo" }}
       },
       { "C4'-O4'-C1'",   "C1*-O4*-C4*", "C1*", "O4*", "C4*",  2,
            {{ 110.100, 1.0, "C2'-endo" },
             { 110.300, 0.7, "C3'-endo" }}
       },
       { "C4'-C3'-O3'",   "C4*-C3*-O3*", "C4*", "C3*", "O3*",  2,
            {{ 109.700, 2.5, "C2'-endo" },
             { 112.300, 2.0, "C3'-endo" }}
       },
       { "C2'-C3'-O3'",   "C2*-C3*-O3*", "C2*", "C3*", "O3*",  2,
            {{ 109.400, 2.5, "C2'-endo" },
             { 112.600, 3.3, "C3'-endo" }}
       },
       { "C2'-C3'-C4'",   "C2*-C3*-C4*", "C4*", "C3*", "C2*",  2,
            {{ 103.100, 0.9, "C2'-endo" },
             { 102.200, 0.7, "C3'-endo" }}
       },
       { "C1'-C2'-C3'",   "C1*-C2*-C3*", "C3*", "C2*", "C1*",  2,
            {{ 102.500, 1.2, "C2'-endo" },
             { 102.400, 0.8, "C3'-endo" }}
       },
       { "O4'-C1'-C2'",   "C2*-C1*-O4*", "O4*", "C1*", "C2*",  2,
            {{ 105.900, 0.8, "C2'-endo" },
             { 106.800, 0.5, "C3'-endo" }}
       },
       { "C2'-C1'-N1/N9", "C2*-C1*-N1",  "N1",  "C1*", "C2*",  2,
            {{ 114.300, 1.4, "C2'-endo" },
             { 112.600, 1.9, "C3'-endo" }}
       },
       { "O4'-C1'-N1/N9", "N1-C1*-O4*",  "O4*", "C1*", "N1",   2,
            {{ 108.000, 0.7, "C2'-endo" },
             { 108.300, 0.3, "C3'-endo" }}
       },
       { "C6-N1-C2",    "C2-N1-C6",    "C6",  "N1",  "C2",   1,
             { { 121.000, 0.6, "s" } }
       },
       { "N1-C2-N3",    "N1-C2-N3",    "N1",  "C2",  "N3",   1,
             { { 114.900, 0.6, "s" } }
       },
       { "C2-N3-C4",    "C2-N3-C4",    "C2",  "N3",  "C4",   1,
             { { 127.000, 0.6, "s" } }
       },
       { "N3-C4-C5",    "C5-C4-N3",    "N3",  "C4",  "C5",   1,
             { { 114.600, 0.6, "s" } }
       },
       { "C4-C5-C6",    "C4-C5-C6",    "C4",  "C5",  "C6",   1,
             { { 119.700, 0.6, "s" } }
       },
       { "C5-C6-N1",    "C5-C6-N1",    "C5",  "C6",  "N1",   1,
             { { 122.700, 0.5, "s" } }
       },
       { "N1-C2-O2",    "N1-C2-O2",    "N1",  "C2",  "O2",   1,
             { { 122.800, 0.7, "s" } }
       },
       { "N3-C2-O2",    "N3-C2-O2",    "N3",  "C2",  "O2",   1,
             { { 122.200, 0.7, "s" } }
       },
       { "N3-C4-O4",    "N3-C4-O4",    "N3",  "C4",  "O4",   1,
             { { 119.400, 0.7, "s" } }
       },
       { "C5-C4-O4",    "C5-C4-O4",    "C5",  "C4",  "O4",   1,
             { { 125.900, 0.6, "s" } }
       },
       { "C6-N1-C1'",   "C1*-N1-C6",   "C6",  "N1",  "C1*",  1,
             { { 121.200, 1.4, "s" } }
       },
       { "C2-N1-C1'",   "C1*-N1-C2",   "C2",  "N1",  "C1*",  1,
             { { 117.700, 1.2, "s" } }
       }
     }
};

MONOMER RU = { "U", 21, 30, // 22, 34,
     {
       { "P-O1P/O2P", "O1P-P",   "P",   "O1P",  1,  { { 1.485, 0.017, "s" } } },
       { "P-O1P/O2P", "O2P-P",   "P",   "O2P",  1,  { { 1.485, 0.017, "s" } } },
       { "P-O5'",     "O5*-P",   "P",   "O5*",  1,  { { 1.593, 0.010, "s" } } },
       // { "P-O3'",     "O3*-P",   "P",   "O3*",  1,  { { 1.607, 0.012, "s" } } },
       { "O5'-C5'",   "C5*-O5*", "O5*", "C5*",  3, {{ 1.440, 0.016, "(P)O5'-C5'" },
                                                    { 1.424, 0.016, "(H)C2'-endo" },
                                                    { 1.420, 0.009, "(H)C3'-endo" }}
       },
       { "C5'-C4'",   "C4*-C5*", "C5*", "C4*",  2, {{ 1.509, 0.012, "C2'-endo" },
                                                    { 1.508, 0.007, "C3'-endo" }}
       },
       { "C3'-C4'",   "C3*-C4*", "C4*", "C3*",  2, {{ 1.527, 0.011, "C2'-endo" },
                                                    { 1.521, 0.010, "C3'-endo" }}
       },
       { "C2'-C3'",   "C2*-C3*", "C3*", "C2*",  2, {{ 1.525, 0.011, "C2'-endo" },
                                                    { 1.523, 0.011, "C3'-endo" }}
       },
       { "C1'-C2'",   "C1*-C2*", "C2*", "C1*",  2, {{ 1.526, 0.008, "C2'-endo" },
                                                    { 1.529, 0.011, "C3'-endo" }}
       },
       { "O4'-C1'",   "C1*-O4*", "O4*", "C1*",  2, {{ 1.415, 0.012, "C2'-endo" },
                                                    { 1.412, 0.013, "C3'-endo" }}
       },
       { "C4'-O4'",   "C4*-O4*", "O4*", "C4*",  2, {{ 1.454, 0.010, "C2'-endo" },
                                                    { 1.451, 0.013, "C3'-endo" }}
       },
       { "C3'-O3'",   "C3*-O3*", "O3*", "C3*",  2, {{ 1.427, 0.012, "C2'-endo" },
                                                    { 1.417, 0.014, "C3'-endo" }}
       },
       { "C1'-N1/N9", "C1*-N1",  "C1*", "N1",   2, {{ 1.464, 0.014, "C2'-endo" },
                                                    { 1.483, 0.015, "C3'-endo" }}
       },
       { "C2'-O2'",   "C2*-O2*", "C2*", "O2*",  2, {{ 1.412, 0.013, "C2'-endo" },
                                                    { 1.420, 0.010, "C3'-endo" }}
       },
       { "C2-O2",   "C2-O2",   "C2",  "O2",   1,  { { 1.219, 0.009, "s"  } } },
       { "C4-O4",   "C4-O4",   "C4",  "O4",   1,  { { 1.232, 0.008, "s"  } } },
       { "N1-C2",   "C2-N1",   "N1",  "C2",   1,  { { 1.381, 0.009, "s"  } } },
       { "C6-N1",   "C6-N1",   "N1",  "C6",   1,  { { 1.375, 0.009, "s"  } } },
       { "C2-N3",   "C2-N3",   "C2",  "N3",   1,  { { 1.373, 0.007, "s"  } } },
       { "N3-C4",   "C4-N3",   "N3",  "C4",   1,  { { 1.380, 0.009, "s"  } } },
       { "C4-C5",   "C4-C5",   "C4",  "C5",   1,  { { 1.431, 0.009, "s"  } } },
       { "C5-C6",   "C5-C6",   "C5",  "C6",   1,  { { 1.337, 0.009, "s"  } } }
     },
     {
       { "O1P-P-O2P",     "O1P-P-O2P",   "O1P", "P",   "O2P",  1,
             { { 119.600, 1.5, "s" } }
       },
       { "O5'-P-O1P/O2P", "O1P-P-O5*",   "O5*", "P",   "O1P",  2,
            {{ 110.700, 1.2, "large" },
             { 105.700, 0.9, "small" }}
       },
       { "O5'-P-O1P/O2P", "O2P-P-O5*",   "O5*", "P",   "O2P",  2,
            {{ 110.700, 1.2, "large" },
             { 105.700, 0.9, "small" }}
       },
/*
       { "O3'-P-O5'",     "O3*-P-O5*",   "O3*", "P",   "O5*",  1,
             { { 104.000, 1.9, "s" } }
       },
       { "O3'-P-O2P/O2P", "O2P-P-O3*",   "O2P", "P",   "O3*",  2,
            {{ 110.500, 1.1, "large" },
             { 105.200, 2.2, "small" }}
       },
       { "O3'-P-O1P/O2P", "O1P-P-O3*",   "O1P", "P",   "O3*",  2,
            {{ 110.500, 1.1, "large" },
             { 105.200, 2.2, "small" }}
       },
*/
       { "O5'-C5'-C4'",   "C4*-C5*-O5*", "O5*", "C5*", "C4*",  3,
            {{ 109.400, 0.8, "(P)O5'-C5'-C4'" },
             { 111.700, 1.9, "C2'-endo" },
             { 111.500, 1.6, "C3'-endo" }}
       },
       { "P-O5'-C5'",     "C5*-O5*-P",   "P",   "O5*", "C5*",  1,
             { { 120.900, 1.6, "s" } }
       },
/*
       { "C3'-O3'-P",     "C3*-O3*-P",   "P",   "O3*", "C3*",  1,
             { { 119.700, 1.2, "s" } }
       },
*/
       { "C3'-C4'-O4'",   "C3*-C4*-O4*",  "O4*", "C4*", "C3*",  2,
            {{ 106.100, 0.8, "C2'-endo" },
             { 104.000, 1.0, "C3'-endo" }}
       },
       { "C5'-C4'-C3'",   "C3*-C4*-C5*", "C5*", "C4*", "C3*",  2,
            {{ 115.200, 1.4, "C2'-endo" },
             { 116.000, 1.6, "C3'-endo" }}
       },
       { "C5'-C4'-O4'",   "C5*-C4*-O4*", "C5*", "C4*", "O4*",  2,
            {{ 109.100, 1.2, "C2'-endo" },
             { 109.800, 0.9, "C3'-endo" }}
       },
       { "C4'-O4'-C1'",   "C1*-O4*-C4*", "C1*", "O4*", "C4*",  2,
            {{ 109.700, 0.7, "C2'-endo" },
             { 109.900, 0.8, "C3'-endo" }}
       },
       { "C4'-C3'-O3'",   "C4*-C3*-O3*", "C4*", "C3*", "O3*",  2,
            {{ 109.400, 2.1, "C2'-endo" },
             { 113.000, 2.0, "C3'-endo" }}
       },
       { "C2'-C3'-O3'",   "C2*-C3*-O3*", "C2*", "C3*", "O3*",  2,
            {{ 109.500, 2.2, "C2'-endo" },
             { 113.700, 1.6, "C3'-endo" }}
       },
       { "C2'-C3'-C4'",   "C2*-C3*-C4*", "C4*", "C3*", "C2*",  2,
            {{ 102.600, 1.0, "C2'-endo" },
             { 102.600, 1.0, "C3'-endo" }}
       },
       { "C1'-C2'-C3'",   "C1*-C2*-C3*", "C3*", "C2*", "C1*",  2,
            {{ 101.500, 0.8, "C2'-endo" },
             { 101.300, 0.7, "C3'-endo" }}
       },
       { "O4'-C1'-C2'",   "C2*-C1*-O4*", "O4*", "C1*", "C2*",  2,
            {{ 105.800, 1.0, "C2'-endo" },
             { 107.600, 0.9, "C3'-endo" }}
       },
       { "C2'-C1'-N1/N9", "C2*-C1*-N1",  "N1",  "C1*", "C2*",  2,
            {{ 114.000, 1.3, "C2'-endo" },
             { 112.000, 1.1, "C3'-endo" }}
       },
       { "O4'-C1'-N1/N9", "N1-C1*-O4*",  "O4*", "C1*", "N1",   2,
            {{ 108.200, 0.8, "C2'-endo" },
             { 108.500, 0.7, "C3'-endo" }}
       },
       { "C1'-C2'-O2'",   "C1*-C2*-O2*", "C1*", "C2*", "O2*",  1,
             { { 110.600, 3.0, "s" } }
       },
       { "C3'-C2'-O2'",   "C3*-C2*-O2*", "C3*", "C2*", "O2*",  1,
             { { 113.300, 2.9, "s" } }
       },
       { "C6-N1-C2",    "C2-N1-C6",    "C6",  "N1",  "C2",   1,
             { { 121.000, 0.6, "s" } }
       },
       { "N1-C2-N3",    "N1-C2-N3",    "N1",  "C2",  "N3",   1,
             { { 114.900, 0.6, "s" } }
       },
       { "C2-N3-C4",    "C2-N3-C4",    "C2",  "N3",  "C4",   1,
             { { 127.000, 0.6, "s" } }
       },
       { "N3-C4-C5",    "C5-C4-N3",    "N3",  "C4",  "C5",   1,
             { { 114.600, 0.6, "s" } }
       },
       { "C4-C5-C6",    "C4-C5-C6",    "C4",  "C5",  "C6",   1,
             { { 119.700, 0.6, "s" } }
       },
       { "C5-C6-N1",    "C5-C6-N1",    "C5",  "C6",  "N1",   1,
             { { 122.700, 0.5, "s" } }
       },
       { "N1-C2-O2",    "N1-C2-O2",    "N1",  "C2",  "O2",   1,
             { { 122.800, 0.7, "s" } }
       },
       { "N3-C2-O2",    "N3-C2-O2",    "N3",  "C2",  "O2",   1,
             { { 122.200, 0.7, "s" } }
       },
       { "N3-C4-O4",    "N3-C4-O4",    "N3",  "C4",  "O4",   1,
             { { 119.400, 0.7, "s" } }
       },
       { "C5-C4-O4",    "C5-C4-O4",    "C5",  "C4",  "O4",   1,
             { { 125.900, 0.6, "s" } }
       },
       { "C6-N1-C1'",   "C1*-N1-C6",   "C6",  "N1",  "C1*",  1,
             { { 121.200, 1.4, "s" } }
       },
       { "C2-N1-C1'",   "C1*-N1-C2",   "C2",  "N1",  "C1*",  1,
             { { 117.700, 1.2, "s" } }
       }
     }
};

MONOMER ALA = { "ALA", 4, 4, // 5, 7,
     {
       { "N-CA (except Gly,Pro)", "N-CA", "N", "CA", 2, {{1.458, 0.019, "NH1-CH1E"},
                                                         {1.491, 0.021, "NH3-CH1E"}} },
       { "CA-CB (Ala)", "CA-CB", "CA", "CB", 1, { { 1.521, 0.033, "CH1E-CH3E" } } },
       { "CA-C (except Gly)", "CA-C", "CA", "C", 1, { { 1.525, 0.021, "CH1E-C" } } },
       { "C-O", "C-O", "C", "O", 2, {{1.231, 0.020, "C-O" },
                                     {1.249, 0.019, "C-OC"} } }
       // { "C-N (except Pro)", "C-N", "C", "N", 1, { { 1.329, 0.014, "C-NH1" } } },
     },
     {
       { "CB-CA-C (Ala)", "CB-CA-C", "CB", "CA", "C", 1, { { 110.500, 1.500, "CH3E-CH1E-C" } } },
       { "N-CA-CB (Ala)", "N-CA-CB", "N", "CA", "CB", 1, { { 110.400, 1.500, "NH1-CH1E-CH3E" } } },
       { "N-CA-C (except Gly,Pro)", "N-CA-C", "N", "CA", "C", 1, { { 111.200, 2.800, "NH1-CH1E-C" } } },
       { "CA-C-O (except Gly)", "CA-C-O", "CA", "C", "O", 2, {{ 120.800, 1.700, "CH1E-C-O" },
                                                              { 117.000, 2.500, "CH1E-C-OC"} } }
       // { "CA-C-N (except Gly,Pro)", "CA-C-N", "CA", "C", "N", 1, { { 116.200, 2.000, "CH1E-C-NH1" } } },
       // { "O-C-N (except Pro)", "O-C-N", "O", "C", "N", 1, { { 123.000, 1.600, "O-C-NH1" } } },
       // { "C-N-CA (except Gly,Pro)", "C-N-CA", "C", "N", "CA", 1, { { 121.700, 1.800, "C-NH1-CH1E" } } },
     }
};

MONOMER ARG = { "ARG", 10, 11, // 11, 14,
     {
       { "N-CA (except Gly,Pro)", "N-CA", "N", "CA", 2, {{1.458, 0.019, "NH1-CH1E"},
                                                         {1.491, 0.021, "NH3-CH1E"} } },
       { "CA-CB (the rest)", "CA-CB", "CA", "CB", 1, { { 1.530, 0.020, "CH1E-CH2E" } } },
       { "CB-CG", "CB-CG", "CB", "CG", 1, { { 1.520, 0.030, "CH2E-CH2E" } } },
       { "CG-CD", "CG-CD", "CG", "CD", 1, { { 1.520, 0.030, "CH2E-CH2E" } } },
       { "CD-NE", "CD-NE", "CD", "NE", 1, { { 1.460, 0.018, "CH2E-NH1" } } },
       { "NE-CZ", "NE-CZ", "NE", "CZ", 1, { { 1.329, 0.014, "NH1-C" } } },
       { "CZ-NH1", "CZ-NH1", "CZ", "NH1", 1, { { 1.326, 0.018, "C-NC2" } } },
       { "CZ-NH2", "CZ-NH2", "CZ", "NH2", 1, { { 1.326, 0.018, "C-NC2" } } },
       { "CA-C (except Gly)", "CA-C", "CA", "C", 1, { { 1.525, 0.021, "CH1E-C" } } },
       { "C-O", "C-O", "C", "O", 2, {{1.231, 0.020, "C-O" },
                                     {1.249, 0.019, "C-OC"} } }
       // { "C-N (except Pro)", "C-N", "C", "N", 1, { { 1.329, 0.014, "C-NH1" } } },
     },
     {
       { "CB-CA-C (the rest)", "CB-CA-C", "CB", "CA", "C", 1, { { 110.100, 1.900, "CH2E-CH1E-C" } } },
       { "N-CA-CB (the rest)", "N-CA-CB", "N", "CA", "CB", 1, { { 110.500, 1.700, "NH1-CH1E-CH2E" } } },
       { "CA-CB-CG", "CA-CB-CG", "CA", "CB", "CG", 1, { { 114.100, 2.000, "CH1E-CH2E-CH2E" } } },
       { "CB-CG-CD", "CB-CG-CD", "CB", "CG", "CD", 1, { { 111.300, 2.300, "CH2E-CH2E-CH2E" } } },
       { "CG-CD-NE", "CG-CD-NE", "CG", "CD", "NE", 1, { { 112.000, 2.200, "CH2E-CH2E-NH1" } } },
       { "CD-NE-CZ", "CD-NE-CZ", "CD", "NE", "CZ", 1, { { 124.200, 1.500, "CH2E-NH1-C" } } },
       { "NH1-CZ-NH2", "NH1-CZ-NH2", "NH1", "CZ", "NH2", 1, { { 119.700, 1.800, "NC2-C-NC2" } } },
       { "NE-CZ-NH1", "NE-CZ-NH1", "NE", "CZ", "NH1", 1, { { 120.000, 1.900, "NH1-C-NC2" } } },
       { "NE-CZ-NH2", "NE-CZ-NH2", "NE", "CZ", "NH2", 1, { { 120.000, 1.900, "NH1-C-NC2" } } },
       { "N-CA-C (except Gly,Pro)", "N-CA-C", "N", "CA", "C", 1, { { 111.200, 2.800, "NH1-CH1E-C" } } },
       { "CA-C-O (except Gly)", "CA-C-O", "CA", "C", "O", 2, {{ 120.800, 1.700, "CH1E-C-O" },
                                                              { 117.000, 2.500, "CH1E-C-OC"} } }
       // { "CA-C-N (except Gly,Pro)", "CA-C-N", "CA", "C", "N", 1, { { 116.200, 2.000, "CH1E-C-NH1" } } },
       // { "O-C-N (except Pro)", "O-C-N", "O", "C", "N", 1, { { 123.000, 1.600, "O-C-NH1" } } },
       // { "C-N-CA (except Gly,Pro)", "C-N-CA", "C", "N", "CA", 1, { { 121.700, 1.800, "C-NH1-CH1E" } } },
     }
};

MONOMER ASN = { "ASN", 7, 8, // 8, 11,
     {
       { "N-CA (except Gly,Pro)", "N-CA", "N", "CA", 2, {{1.458, 0.019, "NH1-CH1E"},
                                                         {1.491, 0.021, "NH3-CH1E"} } },
       { "CA-CB (the rest)", "CA-CB", "CA", "CB", 1, { { 1.530, 0.020, "CH1E-CH2E" } } },
       { "CB-CG", "CB-CG", "CB", "CG", 1, { { 1.516, 0.025, "CH2E-C" } } },
       { "CG-OD1", "CG-OD1", "CG", "OD1", 1, { { 1.231, 0.020, "C-O" } } },
       { "CG-ND2", "CG-ND2", "CG", "ND2", 1, { { 1.328, 0.021, "C-NH2" } } },
       { "CA-C (except Gly)", "CA-C", "CA", "C", 1, { { 1.525, 0.021, "CH1E-C" } } },
       { "C-O", "C-O", "C", "O", 2, {{1.231, 0.020, "C-O" },
                                     {1.249, 0.019, "C-OC"} } }
       // { "C-N (except Pro)", "C-N", "C", "N", 1, { { 1.329, 0.014, "C-NH1" } } },
     },
     {
       { "CB-CA-C (the rest)", "CB-CA-C", "CB", "CA", "C", 1, { { 110.100, 1.900, "CH2E-CH1E-C" } } },
       { "N-CA-CB (the rest)", "N-CA-CB", "N", "CA", "CB", 1, { { 110.500, 1.700, "NH1-CH1E-CH2E" } } },
       { "CA-CB-CG", "CA-CB-CG", "CA", "CB", "CG", 1, { { 112.600, 1.000, "CH1E-CH2E-C" } } },
       { "OD1-CG-ND2", "OD1-CG-ND2", "OD1", "CG", "ND2", 1, { { 122.600, 1.000, "O-C-NH2" } } },
       { "CB-CG-OD1", "CB-CG-OD1", "CB", "CG", "OD1", 1, { { 120.800, 2.000, "CH2E-C-O" } } },
       { "CB-CG-ND2", "CB-CG-ND2", "CB", "CG", "ND2", 1, { { 116.400, 1.500, "CH2E-C-NH2" } } },
       { "N-CA-C (except Gly,Pro)", "N-CA-C", "N", "CA", "C", 1, { { 111.200, 2.800, "NH1-CH1E-C" } } },
       { "CA-C-O (except Gly)", "CA-C-O", "CA", "C", "O", 2, {{ 120.800, 1.700, "CH1E-C-O" },
                                                              { 117.000, 2.500, "CH1E-C-OC"} } }
       // { "CA-C-N (except Gly,Pro)", "CA-C-N", "CA", "C", "N", 1, { { 116.200, 2.000, "CH1E-C-NH1" } } },
       // { "O-C-N (except Pro)", "O-C-N", "O", "C", "N", 1, { { 123.000, 1.600, "O-C-NH1" } } },
       // { "C-N-CA (except Gly,Pro)", "C-N-CA", "C", "N", "CA", 1, { { 121.700, 1.800, "C-NH1-CH1E" } } },
     }
};

MONOMER ASP = { "ASP", 7, 8, // 8, 11,
     {
       { "N-CA (except Gly,Pro)", "N-CA", "N", "CA", 2, {{1.458, 0.019, "NH1-CH1E"},
                                                         {1.491, 0.021, "NH3-CH1E"} } },
       { "CA-CB (the rest)", "CA-CB", "CA", "CB", 1, { { 1.530, 0.020, "CH1E-CH2E" } } },
       { "CB-CG", "CB-CG", "CB", "CG", 1, { { 1.516, 0.025, "CH2E-C" } } },
       { "CG-OD1", "CG-OD1", "CG", "OD1", 1, { { 1.249, 0.019, "C-OC" } } },
       { "CG-OD2", "CG-OD2", "CG", "OD2", 1, { { 1.249, 0.019, "C-OC" } } },
       { "CA-C (except Gly)", "CA-C", "CA", "C", 1, { { 1.525, 0.021, "CH1E-C" } } },
       { "C-O", "C-O", "C", "O", 2, {{1.231, 0.020, "C-O" },
                                     {1.249, 0.019, "C-OC"} } }
       // { "C-N (except Pro)", "C-N", "C", "N", 1, { { 1.329, 0.014, "C-NH1" } } },
     },
     {
       { "CB-CA-C (the rest)", "CB-CA-C", "CB", "CA", "C", 1, { { 110.100, 1.900, "CH2E-CH1E-C" } } },
       { "N-CA-CB (the rest)", "N-CA-CB", "N", "CA", "CB", 1, { { 110.500, 1.700, "NH1-CH1E-CH2E" } } },
       { "CA-CB-CG", "CA-CB-CG", "CA", "CB", "CG", 1, { { 112.600, 1.000, "CH1E-CH2E-C" } } },
       { "OD1-CG-OD2", "OD1-CG-OD2", "OD1", "CG", "OD2", 1, { { 122.900, 2.400, "OC-C-OC" } } },
       { "CB-CG-OD1", "CB-CG-OD1", "CB", "CG", "OD1", 1, { { 118.400, 2.300, "CH2E-C-OC" } } },
       { "CB-CG-OD2", "CB-CG-OD2", "CB", "CG", "OD2", 1, { { 118.400, 2.300, "CH2E-C-OC" } } },
       { "N-CA-C (except Gly,Pro)", "N-CA-C", "N", "CA", "C", 1, { { 111.200, 2.800, "NH1-CH1E-C" } } },
       { "CA-C-O (except Gly)", "CA-C-O", "CA", "C", "O", 2, {{ 120.800, 1.700, "CH1E-C-O" },
                                                              { 117.000, 2.500, "CH1E-C-OC"} } }
       // { "CA-C-N (except Gly,Pro)", "CA-C-N", "CA", "C", "N", 1, { { 116.200, 2.000, "CH1E-C-NH1" } } },
       // { "O-C-N (except Pro)", "O-C-N", "O", "C", "N", 1, { { 123.000, 1.600, "O-C-NH1" } } },
       // { "C-N-CA (except Gly,Pro)", "C-N-CA", "C", "N", "CA", 1, { { 121.700, 1.800, "C-NH1-CH1E" } } },
     }
};

MONOMER CYS = { "CYS", 5, 5, // 6, 8,
     {
       { "N-CA (except Gly,Pro)", "N-CA", "N", "CA", 2, {{1.458, 0.019, "NH1-CH1E"},
                                                         {1.491, 0.021, "NH3-CH1E"} } },
       { "CA-CB (the rest)", "CA-CB", "CA", "CB", 1, { { 1.530, 0.020, "CH1E-CH2E" } } },
       { "CB-SG", "CB-SG", "CB", "SG", 2, {{1.808, 0.033, "CH2E-SH1E" },
                                           {1.822, 0.020, "CH2E-S" } } },
       { "CA-C (except Gly)", "CA-C", "CA", "C", 1, { { 1.525, 0.021, "CH1E-C" } } },
       { "C-O", "C-O", "C", "O", 2, {{1.231, 0.020, "C-O" },
                                     {1.249, 0.019, "C-OC"} } }
       // { "C-N (except Pro)", "C-N", "C", "N", 1, { { 1.329, 0.014, "C-NH1" } } },
     },
     {
       { "CB-CA-C (the rest)", "CB-CA-C", "CB", "CA", "C", 1, { { 110.100, 1.900, "CH2E-CH1E-C" } } },
       { "N-CA-CB (the rest)", "N-CA-CB", "N", "CA", "CB", 1, { { 110.500, 1.700, "NH1-CH1E-CH2E" } } },
       { "CA-CB-SG", "CA-CB-SG", "CA", "CB", "SG", 1, { { 114.400, 2.300, "CH1E-CH2E-SH1E" } } },
       { "N-CA-C (except Gly,Pro)", "N-CA-C", "N", "CA", "C", 1, { { 111.200, 2.800, "NH1-CH1E-C" } } },
       { "CA-C-O (except Gly)", "CA-C-O", "CA", "C", "O", 2, {{ 120.800, 1.700, "CH1E-C-O" },
                                                              { 117.000, 2.500, "CH1E-C-OC"} } }
       // { "CA-C-N (except Gly,Pro)", "CA-C-N", "CA", "C", "N", 1, { { 116.200, 2.000, "CH1E-C-NH1" } } },
       // { "O-C-N (except Pro)", "O-C-N", "O", "C", "N", 1, { { 123.000, 1.600, "O-C-NH1" } } },
       // { "C-N-CA (except Gly,Pro)", "C-N-CA", "C", "N", "CA", 1, { { 121.700, 1.800, "C-NH1-CH1E" } } },
     }
};

MONOMER GLN = { "GLN", 8, 9, // 9, 12,
     {
       { "N-CA (except Gly,Pro)", "N-CA", "N", "CA", 2, {{1.458, 0.019, "NH1-CH1E"},
                                                         {1.491, 0.021, "NH3-CH1E"} } },
       { "CA-CB (the rest)", "CA-CB", "CA", "CB", 1, { { 1.530, 0.020, "CH1E-CH2E" } } },
       { "CB-CG", "CB-CG", "CB", "CG", 1, { { 1.520, 0.030, "CH2E-CH2E" } } },
       { "CG-CD", "CG-CD", "CG", "CD", 1, { { 1.516, 0.025, "CH2E-C" } } },
       { "CD-OE1", "CD-OE1", "CD", "OE1", 1, { { 1.231, 0.020, "C-O" } } },
       { "CD-NE2", "CD-NE2", "CD", "NE2", 1, { { 1.328, 0.021, "C-NH2" } } },
       { "CA-C (except Gly)", "CA-C", "CA", "C", 1, { { 1.525, 0.021, "CH1E-C" } } },
       { "C-O", "C-O", "C", "O", 2, {{1.231, 0.020, "C-O" },
                                     {1.249, 0.019, "C-OC"} } }
       // { "C-N (except Pro)", "C-N", "C", "N", 1, { { 1.329, 0.014, "C-NH1" } } },
     },
     {
       { "CB-CA-C (the rest)", "CB-CA-C", "CB", "CA", "C", 1, { { 110.100, 1.900, "CH2E-CH1E-C" } } },
       { "N-CA-CB (the rest)", "N-CA-CB", "N", "CA", "CB", 1, { { 110.500, 1.700, "NH1-CH1E-CH2E" } } },
       { "CA-CB-CG", "CA-CB-CG", "CA", "CB", "CG", 1, { { 114.100, 2.000, "CH1E-CH2E-CH2E" } } },
       { "CB-CG-CD", "CB-CG-CD", "CB", "CG", "CD", 1, { { 112.600, 1.700, "CH2E-CH2E-C" } } },
       { "OE1-CD-NE2", "OE1-CD-NE2", "OE1", "CD", "NE2", 1, { { 122.600, 1.000, "O-C-NH2" } } },
       { "CG-CD-OE1", "CG-CD-OE1", "CG", "CD", "OE1", 1, { { 120.800, 2.000, "CH2E-C-O" } } },
       { "CG-CD-NE2", "CG-CD-NE2", "CG", "CD", "NE2", 1, { { 116.400, 1.500, "CH2E-C-NH2" } } },
       { "N-CA-C (except Gly,Pro)", "N-CA-C", "N", "CA", "C", 1, { { 111.200, 2.800, "NH1-CH1E-C" } } },
       { "CA-C-O (except Gly)", "CA-C-O", "CA", "C", "O", 2, {{ 120.800, 1.700, "CH1E-C-O" },
                                                              { 117.000, 2.500, "CH1E-C-OC"} } }
       // { "CA-C-N (except Gly,Pro)", "CA-C-N", "CA", "C", "N", 1, { { 116.200, 2.000, "CH1E-C-NH1" } } },
       // { "O-C-N (except Pro)", "O-C-N", "O", "C", "N", 1, { { 123.000, 1.600, "O-C-NH1" } } },
       // { "C-N-CA (except Gly,Pro)", "C-N-CA", "C", "N", "CA", 1, { { 121.700, 1.800, "C-NH1-CH1E" } } },
     }
};

MONOMER GLU = { "GLU", 8, 9, // 9, 12,
     {
       { "N-CA (except Gly,Pro)", "N-CA", "N", "CA", 2, {{1.458, 0.019, "NH1-CH1E"},
                                                         {1.491, 0.021, "NH3-CH1E"} } },
       { "CA-CB (the rest)", "CA-CB", "CA", "CB", 1, { { 1.530, 0.020, "CH1E-CH2E" } } },
       { "CB-CG", "CB-CG", "CB", "CG", 1, { { 1.520, 0.030, "CH2E-CH2E" } } },
       { "CG-CD", "CG-CD", "CG", "CD", 1, { { 1.516, 0.025, "CH2E-C" } } },
       { "CD-OE1", "CD-OE1", "CD", "OE1", 1, { { 1.249, 0.019, "C-OC" } } },
       { "CD-OE2", "CD-OE2", "CD", "OE2", 1, { { 1.249, 0.019, "C-OC" } } },
       { "CA-C (except Gly)", "CA-C", "CA", "C", 1, { { 1.525, 0.021, "CH1E-C" } } },
       { "C-O", "C-O", "C", "O", 2, {{1.231, 0.020, "C-O" },
                                     {1.249, 0.019, "C-OC"} } }
       // { "C-N (except Pro)", "C-N", "C", "N", 1, { { 1.329, 0.014, "C-NH1" } } },
     },
     {
       { "CB-CA-C (the rest)", "CB-CA-C", "CB", "CA", "C", 1, { { 110.100, 1.900, "CH2E-CH1E-C" } } },
       { "N-CA-CB (the rest)", "N-CA-CB", "N", "CA", "CB", 1, { { 110.500, 1.700, "NH1-CH1E-CH2E" } } },
       { "CA-CB-CG", "CA-CB-CG", "CA", "CB", "CG", 1, { { 114.100, 2.000, "CH1E-CH2E-CH2E" } } },
       { "CB-CG-CD", "CB-CG-CD", "CB", "CG", "CD", 1, { { 112.600, 1.700, "CH2E-CH2E-C" } } },
       { "OE1-CD-OE2", "OE1-CD-OE2", "OE1", "CD", "OE2", 1, { { 122.900, 2.400, "OC-C-OC" } } },
       { "CG-CD-OE1", "CG-CD-OE1", "CG", "CD", "OE1", 1, { { 118.400, 2.300, "CH2E-C-OC" } } },
       { "CG-CD-OE2", "CG-CD-OE2", "CG", "CD", "OE2", 1, { { 118.400, 2.300, "CH2E-C-OC" } } },
       { "N-CA-C (except Gly,Pro)", "N-CA-C", "N", "CA", "C", 1, { { 111.200, 2.800, "NH1-CH1E-C" } } },
       { "CA-C-O (except Gly)", "CA-C-O", "CA", "C", "O", 2, {{ 120.800, 1.700, "CH1E-C-O" },
                                                              { 117.000, 2.500, "CH1E-C-OC"} } }
       // { "CA-C-N (except Gly,Pro)", "CA-C-N", "CA", "C", "N", 1, { { 116.200, 2.000, "CH1E-C-NH1" } } },
       // { "O-C-N (except Pro)", "O-C-N", "O", "C", "N", 1, { { 123.000, 1.600, "O-C-NH1" } } },
       // { "C-N-CA (except Gly,Pro)", "C-N-CA", "C", "N", "CA", 1, { { 121.700, 1.800, "C-NH1-CH1E" } } },
     }
};

MONOMER GLY = { "GLY", 3, 2, // 4, 5,
     {
       { "N-CA (Gly)", "N-CA", "N", "CA", 1, { { 1.451, 0.016, "NH1-CH2G*" } } },
       { "CA-C (Gly)", "CA-C", "CA", "C", 1, { { 1.516, 0.018, "CH2G*-C" } } },
       { "C-O", "C-O", "C", "O", 2, {{1.231, 0.020, "C-O" },
                                     {1.249, 0.019, "C-OC"} } }
       // { "C-N (except Pro)", "C-N", "C", "N", 1, { { 1.329, 0.014, "C-NH1" } } },
     },
     {
       { "N-CA-C (Gly)", "N-CA-C", "N", "CA", "C", 1, { { 112.500, 2.900, "NH1-CH2G*-C" } } },
       { "CA-C-O (Gly)", "CA-C-O", "CA", "C", "O", 2, {{120.800, 2.100, "CH2G*-C-O" },
                                                       {118.400, 2.300, "CH2G*-C-OC"} } }
       // { "CA-C-N (Gly)", "CA-C-N", "CA", "C", "N", 1, { { 116.400, 2.100, "CH2G*-C-NH1" } } },
       // { "O-C-N (except Pro)", "O-C-N", "O", "C", "N", 1, { { 123.000, 1.600, "O-C-NH1" } } },
       // { "C-N-CA (Gly)", "C-N-CA", "C", "N", "CA", 1, { { 120.600, 1.700, "C-NH1-CH2G*" } } },
     }
};

MONOMER HIS = { "HIS", 10, 12, // 11, 15,
     {
       { "N-CA (except Gly,Pro)", "N-CA", "N", "CA", 2, {{1.458, 0.019, "NH1-CH1E"},
                                                         {1.491, 0.021, "NH3-CH1E"} } },
       { "CA-CB (the rest)", "CA-CB", "CA", "CB", 1, { { 1.530, 0.020, "CH1E-CH2E" } } },
       { "CB-CG", "CB-CG", "CB", "CG", 1, { { 1.497, 0.014, "CH2E-C5" } } },
       { "CG-CD2", "CG-CD2", "CG", "CD2", 1, { { 1.354, 0.011, "C5-CR1H" } } },
       { "CG-ND1", "CG-ND1", "CG", "ND1", 1, { { 1.378, 0.011, "C5-NH1" } } },
       { "ND1-CE1", "ND1-CE1", "ND1", "CE1", 1, { { 1.321, 0.010, "NH1-CRHH" } } },
       { "CE1-NE2", "CE1-NE2", "CE1", "NE2", 1, { { 1.321, 0.010, "CRHH-NH1" } } },
       { "NE2-CD2", "NE2-CD2", "NE2", "CD2", 1, { { 1.374, 0.011, "NH1-CR1H" } } },
       { "CA-C (except Gly)", "CA-C", "CA", "C", 1, { { 1.525, 0.021, "CH1E-C" } } },
       { "C-O", "C-O", "C", "O", 2, {{1.231, 0.020, "C-O" },
                                     {1.249, 0.019, "C-OC"} } }
       // { "C-N (except Pro)", "C-N", "C", "N", 1, { { 1.329, 0.014, "C-NH1" } } },
     },
     {
       { "CB-CA-C (the rest)", "CB-CA-C", "CB", "CA", "C", 1, { { 110.100, 1.900, "CH2E-CH1E-C" } } },
       { "N-CA-CB (the rest)", "N-CA-CB", "N", "CA", "CB", 1, { { 110.500, 1.700, "NH1-CH1E-CH2E" } } },
       { "CA-CB-CG", "CA-CB-CG", "CA", "CB", "CG", 1, { { 113.800, 1.000, "CH1E-CH2E-C5" } } },
       { "CB-CG-CD2", "CB-CG-CD2", "CB", "CG", "CD2", 1, { { 131.200, 1.300, "CH2E-C5-CR1H" } } },
       { "ND1-CG-CD2", "ND1-CG-CD2", "ND1", "CG", "CD2", 1, { { 106.100, 1.000, "NH1-C5-CR1H" } } },
       { "CB-CG-ND1", "CB-CG-ND1", "CB", "CG", "ND1", 1, { { 122.700, 1.500, "CH2E-C5-NH1" } } },
       { "CG-ND1-CE1", "CG-ND1-CE1", "CG", "ND1", "CE1", 1, { { 109.300, 1.700, "C5-NH1-CRHH" } } },
       { "ND1-CE1-NE2", "ND1-CE1-NE2", "ND1", "CE1", "NE2", 1, { { 108.400, 1.000, "NH1-CRHH-NH1" } } },
       { "CE1-NE2-CD2", "CE1-NE2-CD2", "CE1", "NE2", "CD2", 1, { { 109.000, 1.000, "CRHH-NH1-CR1H" } } },
       { "CG-CD2-NE2", "CG-CD2-NE2", "CG", "CD2", "NE2", 1, { { 107.200, 1.000, "C5-CR1H-NH1" } } },
       { "N-CA-C (except Gly,Pro)", "N-CA-C", "N", "CA", "C", 1, { { 111.200, 2.800, "NH1-CH1E-C" } } },
       { "CA-C-O (except Gly)", "CA-C-O", "CA", "C", "O", 2, {{ 120.800, 1.700, "CH1E-C-O" },
                                                              { 117.000, 2.500, "CH1E-C-OC"} } }
       // { "CA-C-N (except Gly,Pro)", "CA-C-N", "CA", "C", "N", 1, { { 116.200, 2.000, "CH1E-C-NH1" } } },
       // { "O-C-N (except Pro)", "O-C-N", "O", "C", "N", 1, { { 123.000, 1.600, "O-C-NH1" } } },
       // { "C-N-CA (except Gly,Pro)", "C-N-CA", "C", "N", "CA", 1, { { 121.700, 1.800, "C-NH1-CH1E" } } },
     }
};

MONOMER ILE = { "ILE", 7, 8, // 8, 11,
     {
       { "N-CA (except Gly,Pro)", "N-CA", "N", "CA", 2, {{1.458, 0.019, "NH1-CH1E"},
                                                         {1.491, 0.021, "NH3-CH1E"} } },
       { "CA-CB (Ile,Thr,Val)", "CA-CB", "CA", "CB", 1, { { 1.540, 0.027, "CH1E-CH1E" } } },
       { "CB-CG1", "CB-CG1", "CB", "CG1", 1, { { 1.530, 0.020, "CH1E-CH2E" } } },
       { "CG1-CD", "CG1-CD", "CG1", "CD", 1, { { 1.513, 0.039, "CH2E-CH3E" } } },
       { "CB-CG2", "CB-CG2", "CB", "CG2", 1, { { 1.521, 0.033, "CH1E-CH3E" } } },
       { "CA-C (except Gly)", "CA-C", "CA", "C", 1, { { 1.525, 0.021, "CH1E-C" } } },
       { "C-O", "C-O", "C", "O", 2, {{1.231, 0.020, "C-O" },
                                     {1.249, 0.019, "C-OC"} } }
       // { "C-N (except Pro)", "C-N", "C", "N", 1, { { 1.329, 0.014, "C-NH1" } } },
     },
     {
       { "CB-CA-C (Ile,Thr,Val)", "CB-CA-C", "CB", "CA", "C", 1, { { 109.100, 2.200, "CH1E-CH1E-C" } } },
       { "N-CA-CB (Ile,Thr,Val)", "N-CA-CB", "N", "CA", "CB", 1, { { 111.500, 1.700, "NH1-CH1E-CH1E" } } },
       { "CG1-CB-CG2", "CG1-CB-CG2", "CG1", "CB", "CG2", 1, { { 110.700, 3.000, "CH2E-CH1E-CH3E" } } },
       { "CA-CB-CG1", "CA-CB-CG1", "CA", "CB", "CG1", 1, { { 110.400, 1.700, "CH1E-CH1E-CH2E" } } },
       { "CB-CG1-CD", "CB-CG1-CD", "CB", "CG1", "CD", 1, { { 113.800, 2.100, "CH1E-CH2E-CH3E" } } },
       { "CA-CB-CG2", "CA-CB-CG2", "CA", "CB", "CG2", 1, { { 110.500, 1.700, "CH1E-CH1E-CH3E" } } },
       { "N-CA-C (except Gly,Pro)", "N-CA-C", "N", "CA", "C", 1, { { 111.200, 2.800, "NH1-CH1E-C" } } },
       { "CA-C-O (except Gly)", "CA-C-O", "CA", "C", "O", 2, {{ 120.800, 1.700, "CH1E-C-O" },
                                                              { 117.000, 2.500, "CH1E-C-OC"} } }
       // { "CA-C-N (except Gly,Pro)", "CA-C-N", "CA", "C", "N", 1, { { 116.200, 2.000, "CH1E-C-NH1" } } },
       // { "O-C-N (except Pro)", "O-C-N", "O", "C", "N", 1, { { 123.000, 1.600, "O-C-NH1" } } },
       // { "C-N-CA (except Gly,Pro)", "C-N-CA", "C", "N", "CA", 1, { { 121.700, 1.800, "C-NH1-CH1E" } } },
     }
};

MONOMER LEU = { "LEU", 7, 8, // 8, 11,
     {
       { "N-CA (except Gly,Pro)", "N-CA", "N", "CA", 2, {{1.458, 0.019, "NH1-CH1E"},
                                                         {1.491, 0.021, "NH3-CH1E"} } },
       { "CA-CB (the rest)", "CA-CB", "CA", "CB", 1, { { 1.530, 0.020, "CH1E-CH2E" } } },
       { "CB-CG", "CB-CG", "CB", "CG", 1, { { 1.530, 0.020, "CH2E-CH1E" } } },
       { "CG-CD1", "CG-CD1", "CG", "CD1", 1, { { 1.521, 0.033, "CH1E-CH3E" } } },
       { "CG-CD2", "CG-CD2", "CG", "CD2", 1, { { 1.521, 0.033, "CH1E-CH3E" } } },
       { "CA-C (except Gly)", "CA-C", "CA", "C", 1, { { 1.525, 0.021, "CH1E-C" } } },
       { "C-O", "C-O", "C", "O", 2, {{1.231, 0.020, "C-O" },
                                     {1.249, 0.019, "C-OC"} } }
       // { "C-N (except Pro)", "C-N", "C", "N", 1, { { 1.329, 0.014, "C-NH1" } } },
     },
     {
       { "CB-CA-C (the rest)", "CB-CA-C", "CB", "CA", "C", 1, { { 110.100, 1.900, "CH2E-CH1E-C" } } },
       { "N-CA-CB (the rest)", "N-CA-CB", "N", "CA", "CB", 1, { { 110.500, 1.700, "NH1-CH1E-CH2E" } } },
       { "CA-CB-CG", "CA-CB-CG", "CA", "CB", "CG", 1, { { 116.300, 3.500, "CH1E-CH2E-CH1E" } } },
       { "CD1-CG-CD2", "CD1-CG-CD2", "CD1", "CG", "CD2", 1, { { 110.800, 2.200, "CH3E-CH1E-CH3E" } } },
       { "CB-CG-CD1", "CB-CG-CD1", "CB", "CG", "CD1", 1, { { 110.700, 3.000, "CH2E-CH1E-CH3E" } } },
       { "CB-CG-CD2", "CB-CG-CD2", "CB", "CG", "CD2", 1, { { 110.700, 3.000, "CH2E-CH1E-CH3E" } } },
       { "N-CA-C (except Gly,Pro)", "N-CA-C", "N", "CA", "C", 1, { { 111.200, 2.800, "NH1-CH1E-C" } } },
       { "CA-C-O (except Gly)", "CA-C-O", "CA", "C", "O", 2, {{ 120.800, 1.700, "CH1E-C-O" },
                                                              { 117.000, 2.500, "CH1E-C-OC"} } }
       // { "CA-C-N (except Gly,Pro)", "CA-C-N", "CA", "C", "N", 1, { { 116.200, 2.000, "CH1E-C-NH1" } } },
       // { "O-C-N (except Pro)", "O-C-N", "O", "C", "N", 1, { { 123.000, 1.600, "O-C-NH1" } } },
       // { "C-N-CA (except Gly,Pro)", "C-N-CA", "C", "N", "CA", 1, { { 121.700, 1.800, "C-NH1-CH1E" } } },
     }
};

MONOMER LYS = { "LYS", 8, 8, // 9, 11,
     {
       { "N-CA (except Gly,Pro)", "N-CA", "N", "CA", 2, {{1.458, 0.019, "NH1-CH1E"},
                                                         {1.491, 0.021, "NH3-CH1E"} } },
       { "CA-CB (the rest)", "CA-CB", "CA", "CB", 1, { { 1.530, 0.020, "CH1E-CH2E" } } },
       { "CB-CG", "CB-CG", "CB", "CG", 1, { { 1.520, 0.030, "CH2E-CH2E" } } },
       { "CG-CD", "CG-CD", "CG", "CD", 1, { { 1.520, 0.030, "CH2E-CH2E" } } },
       { "CD-CE", "CD-CE", "CD", "CE", 1, { { 1.520, 0.030, "CH2E-CH2E" } } },
       { "CE-NZ", "CE-NZ", "CE", "NZ", 1, { { 1.489, 0.030, "CH2E-NH3" } } },
       { "CA-C (except Gly)", "CA-C", "CA", "C", 1, { { 1.525, 0.021, "CH1E-C" } } },
       { "C-O", "C-O", "C", "O", 2, {{1.231, 0.020, "C-O" },
                                     {1.249, 0.019, "C-OC"} } }
       // { "C-N (except Pro)", "C-N", "C", "N", 1, { { 1.329, 0.014, "C-NH1" } } },
     },
     {
       { "CB-CA-C (the rest)", "CB-CA-C", "CB", "CA", "C", 1, { { 110.100, 1.900, "CH2E-CH1E-C" } } },
       { "N-CA-CB (the rest)", "N-CA-CB", "N", "CA", "CB", 1, { { 110.500, 1.700, "NH1-CH1E-CH2E" } } },
       { "CA-CB-CG", "CA-CB-CG", "CA", "CB", "CG", 1, { { 114.100, 2.000, "CH1E-CH2E-CH2E" } } },
       { "CB-CG-CD", "CB-CG-CD", "CB", "CG", "CD", 1, { { 111.300, 2.300, "CH2E-CH2E-CH2E" } } },
       { "CG-CD-CE", "CG-CD-CE", "CG", "CD", "CE", 1, { { 111.300, 2.300, "CH2E-CH2E-CH2E" } } },
       { "CD-CE-NZ", "CD-CE-NZ", "CD", "CE", "NZ", 1, { { 111.900, 3.200, "CH2E-CH2E-NH3" } } },
       { "N-CA-C (except Gly,Pro)", "N-CA-C", "N", "CA", "C", 1, { { 111.200, 2.800, "NH1-CH1E-C" } } },
       { "CA-C-O (except Gly)", "CA-C-O", "CA", "C", "O", 2, {{ 120.800, 1.700, "CH1E-C-O" },
                                                              { 117.000, 2.500, "CH1E-C-OC"} } }
       // { "CA-C-N (except Gly,Pro)", "CA-C-N", "CA", "C", "N", 1, { { 116.200, 2.000, "CH1E-C-NH1" } } },
       // { "O-C-N (except Pro)", "O-C-N", "O", "C", "N", 1, { { 123.000, 1.600, "O-C-NH1" } } },
       // { "C-N-CA (except Gly,Pro)", "C-N-CA", "C", "N", "CA", 1, { { 121.700, 1.800, "C-NH1-CH1E" } } },
     }
};

MONOMER MET = { "MET", 7, 7, // 8, 10,
     {
       { "N-CA (except Gly,Pro)", "N-CA", "N", "CA", 2, {{1.458, 0.019, "NH1-CH1E"},
                                                         {1.491, 0.021, "NH3-CH1E"} } },
       { "CA-CB (the rest)", "CA-CB", "CA", "CB", 1, { { 1.530, 0.020, "CH1E-CH2E" } } },
       { "CB-CG", "CB-CG", "CB", "CG", 1, { { 1.520, 0.030, "CH2E-CH2E" } } },
       { "CG-SD", "CG-SD", "CG", "SD", 1, { { 1.803, 0.034, "CH2E-SM" } } },
       { "SD-CE", "SD-CE", "SD", "CE", 1, { { 1.791, 0.059, "SM-CH3E" } } },
       { "CA-C (except Gly)", "CA-C", "CA", "C", 1, { { 1.525, 0.021, "CH1E-C" } } },
       { "C-O", "C-O", "C", "O", 2, {{1.231, 0.020, "C-O" },
                                     {1.249, 0.019, "C-OC"} } }
       // { "C-N (except Pro)", "C-N", "C", "N", 1, { { 1.329, 0.014, "C-NH1" } } },
     },
     {
       { "CB-CA-C (the rest)", "CB-CA-C", "CB", "CA", "C", 1, { { 110.100, 1.900, "CH2E-CH1E-C" } } },
       { "N-CA-CB (the rest)", "N-CA-CB", "N", "CA", "CB", 1, { { 110.500, 1.700, "NH1-CH1E-CH2E" } } },
       { "CA-CB-CG", "CA-CB-CG", "CA", "CB", "CG", 1, { { 114.100, 2.000, "CH1E-CH2E-CH2E" } } },
       { "CB-CG-SD", "CB-CG-SD", "CB", "CG", "SD", 1, { { 112.700, 3.000, "CH2E-CH2E-SM" } } },
       { "CG-SD-CE", "CG-SD-CE", "CG", "SD", "CE", 1, { { 100.900, 2.200, "CH2E-SM-CH3E" } } },
       { "N-CA-C (except Gly,Pro)", "N-CA-C", "N", "CA", "C", 1, { { 111.200, 2.800, "NH1-CH1E-C" } } },
       { "CA-C-O (except Gly)", "CA-C-O", "CA", "C", "O", 2, {{ 120.800, 1.700, "CH1E-C-O" },
                                                              { 117.000, 2.500, "CH1E-C-OC"} } }
       // { "CA-C-N (except Gly,Pro)", "CA-C-N", "CA", "C", "N", 1, { { 116.200, 2.000, "CH1E-C-NH1" } } },
       // { "O-C-N (except Pro)", "O-C-N", "O", "C", "N", 1, { { 123.000, 1.600, "O-C-NH1" } } },
       // { "C-N-CA (except Gly,Pro)", "C-N-CA", "C", "N", "CA", 1, { { 121.700, 1.800, "C-NH1-CH1E" } } },
     }
};

MONOMER PHE = { "PHE", 11, 13, // 12, 16,
     {
       { "N-CA (except Gly,Pro)", "N-CA", "N", "CA", 2, {{1.458, 0.019, "NH1-CH1E"},
                                                         {1.491, 0.021, "NH3-CH1E"} } },
       { "CA-CB (the rest)", "CA-CB", "CA", "CB", 1, { { 1.530, 0.020, "CH1E-CH2E" } } },
       { "CB-CG", "CB-CG", "CB", "CG", 1, { { 1.502, 0.023, "CH2E-CF" } } },
       { "CG-CD2", "CG-CD2", "CG", "CD2", 1, { { 1.384, 0.021, "CF-CR1E" } } },
       { "CG-CD1", "CG-CD1", "CG", "CD1", 1, { { 1.384, 0.021, "CF-CR1E" } } },
       { "CD1-CE1", "CD1-CE1", "CD1", "CE1", 1, { { 1.382, 0.030, "CR1E-CR1E" } } },
       { "CE1-CZ", "CE1-CZ", "CE1", "CZ", 1, { { 1.382, 0.030, "CR1E-CR1E" } } },
       { "CZ-CE2", "CZ-CE2", "CZ", "CE2", 1, { { 1.382, 0.030, "CR1E-CR1E" } } },
       { "CE2-CD2", "CE2-CD2", "CE2", "CD2", 1, { { 1.382, 0.030, "CR1E-CR1E" } } },
       { "CA-C (except Gly)", "CA-C", "CA", "C", 1, { { 1.525, 0.021, "CH1E-C" } } },
       { "C-O", "C-O", "C", "O", 2, {{1.231, 0.020, "C-O" },
                                     {1.249, 0.019, "C-OC"} } }
       // { "C-N (except Pro)", "C-N", "C", "N", 1, { { 1.329, 0.014, "C-NH1" } } },
     },
     {
       { "CB-CA-C (the rest)", "CB-CA-C", "CB", "CA", "C", 1, { { 110.100, 1.900, "CH2E-CH1E-C" } } },
       { "N-CA-CB (the rest)", "N-CA-CB", "N", "CA", "CB", 1, { { 110.500, 1.700, "NH1-CH1E-CH2E" } } },
       { "CA-CB-CG", "CA-CB-CG", "CA", "CB", "CG", 1, { { 113.800, 1.000, "CH1E-CH2E-CF" } } },
       { "CB-CG-CD2", "CB-CG-CD2", "CB", "CG", "CD2", 1, { { 120.700, 1.700, "CH2E-CF-CR1E" } } },
       { "CD1-CG-CD2", "CD1-CG-CD2", "CD1", "CG", "CD2", 1, { { 118.600, 1.500, "CR1E-CF-CR1E" } } },
       { "CB-CG-CD1", "CB-CG-CD1", "CB", "CG", "CD1", 1, { { 120.700, 1.700, "CH2E-CF-CR1E" } } },
       { "CG-CD1-CE1", "CG-CD1-CE1", "CG", "CD1", "CE1", 1, { { 120.700, 1.700, "CF-CR1E-CR1E" } } },
       { "CG-CD2-CE2", "CG-CD2-CE2", "CG", "CD2", "CE2", 1, { { 120.700, 1.700, "CF-CR1E-CR1E" } } },
       { "CD1-CE1-CZ", "CD1-CE1-CZ", "CD1", "CE1", "CZ", 1, { { 120.000, 1.800, "CR1E-CR1E-CR1E" } } },
       { "CE1-CZ-CE2", "CE1-CZ-CE2", "CE1", "CZ", "CE2", 1, { { 120.000, 1.800, "CR1E-CR1E-CR1E" } } },
       { "CZ-CE2-CD2", "CZ-CE2-CD2", "CZ", "CE2", "CD2", 1, { { 120.000, 1.800, "CR1E-CR1E-CR1E" } } },
       { "N-CA-C (except Gly,Pro)", "N-CA-C", "N", "CA", "C", 1, { { 111.200, 2.800, "NH1-CH1E-C" } } },
       { "CA-C-O (except Gly)", "CA-C-O", "CA", "C", "O", 2, {{ 120.800, 1.700, "CH1E-C-O" },
                                                              { 117.000, 2.500, "CH1E-C-OC"} } }
       // { "CA-C-N (except Gly,Pro)", "CA-C-N", "CA", "C", "N", 1, { { 116.200, 2.000, "CH1E-C-NH1" } } },
       // { "O-C-N (except Pro)", "O-C-N", "O", "C", "N", 1, { { 123.000, 1.600, "O-C-NH1" } } },
       // { "C-N-CA (except Gly,Pro)", "C-N-CA", "C", "N", "CA", 1, { { 121.700, 1.800, "C-NH1-CH1E" } } },
     }
};

MONOMER PRO = { "PRO", 7, 8, // 8, 11,
     {
       { "N-CA (Pro)", "N-CA", "N", "CA", 1, { { 1.466, 0.015, "N-CH1E" } } },
       { "CA-CB (the rest)", "CA-CB", "CA", "CB", 1, { { 1.530, 0.020, "CH1E-CH2E" } } },
       { "CB-CG", "CB-CG", "CB", "CG", 1, { { 1.492, 0.050, "CH2E-CH2P" } } },
       { "CG-CD", "CG-CD", "CG", "CD", 1, { { 1.503, 0.034, "CH2P-CH2P" } } },
       { "CD-N", "CD-N", "CD", "N", 1, { { 1.473, 0.014, "CH2P-N" } } },
       { "CA-C (except Gly)", "CA-C", "CA", "C", 1, { { 1.525, 0.021, "CH1E-C" } } },
       { "C-O", "C-O", "C", "O", 2, {{1.231, 0.020, "C-O" },
                                     {1.249, 0.019, "C-OC"} } }
       // { "C-N (Pro)", "C-N", "C", "N", 1, { { 1.341, 0.016, "C-NH1" } } },
     },
     {
       { "CA-N-CD", "CA-N-CD", "CA", "N", "CD", 1, { { 112.000, 1.400, "CH1E-N-CH2P" } } },
       { "CB-CA-C (the rest)", "CB-CA-C", "CB", "CA", "C", 1, { { 110.100, 1.900, "CH2E-CH1E-C" } } },
       { "N-CA-CB (Pro)", "N-CA-CB", "N", "CA", "CB", 1, { { 103.000, 1.100, "N-CH1E-CH2E" } } },
       { "CA-CB-CG", "CA-CB-CG", "CA", "CB", "CG", 1, { { 104.500, 1.900, "CH1E-CH2E-CH2P" } } },
       { "CB-CG-CD", "CB-CG-CD", "CB", "CG", "CD", 1, { { 106.100, 3.200, "CH2E-CH2P-CH2P" } } },
       { "N-CD-CG", "N-CD-CG", "N", "CD", "CG", 1, { { 103.200, 1.500, "N-CH2P-CH2P" } } },
       { "N-CA-C (Pro)", "N-CA-C", "N", "CA", "C", 1, { { 111.800, 2.500, "N-CH1E-C" } } },
       { "CA-C-O (except Gly)", "CA-C-O", "CA", "C", "O", 2, {{ 120.800, 1.700, "CH1E-C-O" },
                                                              { 117.000, 2.500, "CH1E-C-OC"} } }
       // { "CA-C-N (Pro)", "CA-C-N", "CA", "C", "N", 1, { { 116.900, 1.500, "CH1E-C-N" } } },
       // { "O-C-N (Pro)", "O-C-N", "O", "C", "N", 1, { { 122.000, 1.400, "O-C-N" } } },
       // { "C-N-CA (Pro)", "C-N-CA", "C", "N", "CA", 1, { { 122.600, 5.000, "C-N-CH1E" } } },
     }
};

MONOMER SER = { "SER", 5, 5, // 6, 8,
     {
       { "N-CA (except Gly,Pro)", "N-CA", "N", "CA", 2, {{1.458, 0.019, "NH1-CH1E"},
                                                         {1.491, 0.021, "NH3-CH1E"} } },
       { "CA-CB (the rest)", "CA-CB", "CA", "CB", 1, { { 1.530, 0.020, "CH1E-CH2E" } } },
       { "CB-OG", "CB-OG", "CB", "OG", 1, { { 1.417, 0.020, "CH2E-OH1" } } },
       { "CA-C (except Gly)", "CA-C", "CA", "C", 1, { { 1.525, 0.021, "CH1E-C" } } },
       { "C-O", "C-O", "C", "O", 2, {{1.231, 0.020, "C-O" },
                                     {1.249, 0.019, "C-OC"} } }
       // { "C-N (except Pro)", "C-N", "C", "N", 1, { { 1.329, 0.014, "C-NH1" } } },
     },
     {
       { "CB-CA-C (the rest)", "CB-CA-C", "CB", "CA", "C", 1, { { 110.100, 1.900, "CH2E-CH1E-C" } } },
       { "N-CA-CB (the rest)", "N-CA-CB", "N", "CA", "CB", 1, { { 110.500, 1.700, "NH1-CH1E-CH2E" } } },
       { "CA-CB-OG", "CA-CB-OG", "CA", "CB", "OG", 1, { { 111.100, 2.000, "CH1E-CH2E-OH1" } } },
       { "N-CA-C (except Gly,Pro)", "N-CA-C", "N", "CA", "C", 1, { { 111.200, 2.800, "NH1-CH1E-C" } } },
       { "CA-C-O (except Gly)", "CA-C-O", "CA", "C", "O", 2, {{ 120.800, 1.700, "CH1E-C-O" },
                                                              { 117.000, 2.500, "CH1E-C-OC"} } }
       // { "CA-C-N (except Gly,Pro)", "CA-C-N", "CA", "C", "N", 1, { { 116.200, 2.000, "CH1E-C-NH1" } } },
       // { "O-C-N (except Pro)", "O-C-N", "O", "C", "N", 1, { { 123.000, 1.600, "O-C-NH1" } } },
       // { "C-N-CA (except Gly,Pro)", "C-N-CA", "C", "N", "CA", 1, { { 121.700, 1.800, "C-NH1-CH1E" } } },
     }
};

MONOMER THR = { "THR", 6, 7, // 7, 10,
     {
       { "N-CA (except Gly,Pro)", "N-CA", "N", "CA", 2, {{1.458, 0.019, "NH1-CH1E"},
                                                         {1.491, 0.021, "NH3-CH1E"} } },
       { "CA-CB (Ile,Thr,Val)", "CA-CB", "CA", "CB", 1, { { 1.540, 0.027, "CH1E-CH1E" } } },
       { "CB-OG1", "CB-OG1", "CB", "OG1", 1, { { 1.433, 0.016, "CH1E-OH1" } } },
       { "CB-CG2", "CB-CG2", "CB", "CG2", 1, { { 1.521, 0.033, "CH1E-CH3E" } } },
       { "CA-C (except Gly)", "CA-C", "CA", "C", 1, { { 1.525, 0.021, "CH1E-C" } } },
       { "C-O", "C-O", "C", "O", 2, {{1.231, 0.020, "C-O" },
                                     {1.249, 0.019, "C-OC"} } }
       // { "C-N (except Pro)", "C-N", "C", "N", 1, { { 1.329, 0.014, "C-NH1" } } },
     },
     {
       { "CB-CA-C (Ile,Thr,Val)", "CB-CA-C", "CB", "CA", "C", 1, { { 109.100, 2.200, "CH1E-CH1E-C" } } },
       { "N-CA-CB (Ile,Thr,Val)", "N-CA-CB", "N", "CA", "CB", 1, { { 111.500, 1.700, "NH1-CH1E-CH1E" } } },
       { "OG1-CB-CG2", "OG1-CB-CG2", "OG1", "CB", "CG2", 1, { { 109.300, 2.000, "OH1-CH1E-CH3E" } } },
       { "CA-CB-OG1", "CA-CB-OG1", "CA", "CB", "OG1", 1, { { 109.600, 1.500, "CH1E-CH1E-OH1" } } },
       { "CA-CB-CG2", "CA-CB-CG2", "CA", "CB", "CG2", 1, { { 110.500, 1.700, "CH1E-CH1E-CH3E" } } },
       { "N-CA-C (except Gly,Pro)", "N-CA-C", "N", "CA", "C", 1, { { 111.200, 2.800, "NH1-CH1E-C" } } },
       { "CA-C-O (except Gly)", "CA-C-O", "CA", "C", "O", 2, {{ 120.800, 1.700, "CH1E-C-O" },
                                                              { 117.000, 2.500, "CH1E-C-OC"} } }
       // { "CA-C-N (except Gly,Pro)", "CA-C-N", "CA", "C", "N", 1, { { 116.200, 2.000, "CH1E-C-NH1" } } },
       // { "O-C-N (except Pro)", "O-C-N", "O", "C", "N", 1, { { 123.000, 1.600, "O-C-NH1" } } },
       // { "C-N-CA (except Gly,Pro)", "C-N-CA", "C", "N", "CA", 1, { { 121.700, 1.800, "C-NH1-CH1E" } } },
     }
};

MONOMER TRP = { "TRP", 15, 20, // 16, 23,
     {
       { "N-CA (except Gly,Pro)", "N-CA", "N", "CA", 2, {{1.458, 0.019, "NH1-CH1E"},
                                                         {1.491, 0.021, "NH3-CH1E"} } },
       { "CA-CB (the rest)", "CA-CB", "CA", "CB", 1, { { 1.530, 0.020, "CH1E-CH2E" } } },
       { "CB-CG", "CB-CG", "CB", "CG", 1, { { 1.498, 0.031, "CH2E-C5W" } } },
       { "CG-CD2", "CG-CD2", "CG", "CD2", 1, { { 1.433, 0.018, "C5W-CW" } } },
       { "CG-CD1", "CG-CD1", "CG", "CD1", 1, { { 1.365, 0.025, "C5W-CR1E" } } },
       { "CD1-NE1", "CD1-NE1", "CD1", "NE1", 1, { { 1.374, 0.021, "CR1E-NH1" } } },
       { "NE1-CE2", "NE1-CE2", "NE1", "CE2", 1, { { 1.370, 0.011, "NH1-CW" } } },
       { "CE2-CZ2", "CE2-CZ2", "CE2", "CZ2", 1, { { 1.394, 0.021, "CW-CR1W" } } },
       { "CE2-CD2", "CE2-CD2", "CE2", "CD2", 1, { { 1.409, 0.017, "CW-CW" } } },
       { "CD2-CE3", "CD2-CE3", "CD2", "CE3", 1, { { 1.398, 0.016, "CW-CR1E" } } },
       { "CE3-CZ3", "CE3-CZ3", "CE3", "CZ3", 1, { { 1.382, 0.030, "CR1E-CR1E" } } },
       { "CZ3-CH2", "CZ3-CH2", "CZ3", "CH2", 1, { { 1.400, 0.025, "CR1E-CR1W" } } },
       { "CH2-CZ2", "CH2-CZ2", "CH2", "CZ2", 1, { { 1.368, 0.019, "CR1W-CR1W" } } },
       { "CA-C (except Gly)", "CA-C", "CA", "C", 1, { { 1.525, 0.021, "CH1E-C" } } },
       { "C-O", "C-O", "C", "O", 2, {{1.231, 0.020, "C-O" },
                                     {1.249, 0.019, "C-OC"} } }
       // { "C-N (except Pro)", "C-N", "C", "N", 1, { { 1.329, 0.014, "C-NH1" } } },
     },
     {
       { "CB-CA-C (the rest)", "CB-CA-C", "CB", "CA", "C", 1, { { 110.100, 1.900, "CH2E-CH1E-C" } } },
       { "N-CA-CB (the rest)", "N-CA-CB", "N", "CA", "CB", 1, { { 110.500, 1.700, "NH1-CH1E-CH2E" } } },
       { "CA-CB-CG", "CA-CB-CG", "CA", "CB", "CG", 1, { { 113.600, 1.900, "CH1E-CH2E-C5W" } } },
       { "CB-CG-CD2", "CB-CG-CD2", "CB", "CG", "CD2", 1, { { 126.800, 1.400, "CH2E-C5W-CW" } } },
       { "CD1-CG-CD2", "CD1-CG-CD2", "CD1", "CG", "CD2", 1, { { 106.300, 1.600, "CR1E-C5W-CW" } } },
       { "CB-CG-CD1", "CB-CG-CD1", "CB", "CG", "CD1", 1, { { 126.900, 1.500, "CH2E-C5W-CR1E" } } },
       { "CG-CD1-NE1", "CG-CD1-NE1", "CG", "CD1", "NE1", 1, { { 110.200, 1.300, "C5W-CR1E-NH1" } } },
       { "CD1-NE1-CE2", "CD1-NE1-CE2", "CD1", "NE1", "CE2", 1, { { 108.900, 1.800, "CR1E-NH1-CW" } } },
       { "NE1-CE2-CZ2", "NE1-CE2-CZ2", "NE1", "CE2", "CZ2", 1, { { 130.100, 1.500, "NH1-CW-CR1W" } } },
       { "CD2-CE2-CZ2", "CD2-CE2-CZ2", "CD2", "CE2", "CZ2", 1, { { 122.400, 1.000, "CW-CW-CR1W" } } },
       { "NE1-CE2-CD2", "NE1-CE2-CD2", "NE1", "CE2", "CD2", 1, { { 107.400, 1.300, "NH1-CW-CW" } } },
       { "CE2-CD2-CE3", "CE2-CD2-CE3", "CE2", "CD2", "CE3", 1, { { 118.800, 1.000, "CW-CW-CR1E" } } },
       { "CE2-CD2-CG", "CE2-CD2-CG", "CE2", "CD2", "CG", 1, { { 107.200, 1.200, "CW-CW-C5W" } } },
       { "CG-CD2-CE3", "CG-CD2-CE3", "CG", "CD2", "CE3", 1, { { 133.900, 1.000, "C5W-CW-CR1E" } } },
       { "CD2-CE3-CZ3", "CD2-CE3-CZ3", "CD2", "CE3", "CZ3", 1, { { 118.600, 1.300, "CW-CR1E-CR1E" } } },
       { "CE3-CZ3-CH2", "CE3-CZ3-CH2", "CE3", "CZ3", "CH2", 1, { { 121.100, 1.300, "CR1E-CR1E-CR1W" } } },
       { "CZ3-CH2-CZ2", "CZ3-CH2-CZ2", "CZ3", "CH2", "CZ2", 1, { { 121.500, 1.300, "CR1E-CR1W-CR1W" } } },
       { "CH2-CZ2-CE2", "CH2-CZ2-CE2", "CH2", "CZ2", "CE2", 1, { { 117.500, 1.300, "CR1W-CR1W-CW" } } },
       { "N-CA-C (except Gly,Pro)", "N-CA-C", "N", "CA", "C", 1, { { 111.200, 2.800, "NH1-CH1E-C" } } },
       { "CA-C-O (except Gly)", "CA-C-O", "CA", "C", "O", 2, {{ 120.800, 1.700, "CH1E-C-O" },
                                                              { 117.000, 2.500, "CH1E-C-OC"} } }
       // { "CA-C-N (except Gly,Pro)", "CA-C-N", "CA", "C", "N", 1, { { 116.200, 2.000, "CH1E-C-NH1" } } },
       // { "O-C-N (except Pro)", "O-C-N", "O", "C", "N", 1, { { 123.000, 1.600, "O-C-NH1" } } },
       // { "C-N-CA (except Gly,Pro)", "C-N-CA", "C", "N", "CA", 1, { { 121.700, 1.800, "C-NH1-CH1E" } } },
     }
};

MONOMER TYR = { "TYR", 12, 15, // 13, 18,
     {
       { "N-CA (except Gly,Pro)", "N-CA", "N", "CA", 2, {{1.458, 0.019, "NH1-CH1E"},
                                                         {1.491, 0.021, "NH3-CH1E"} } },
       { "CA-CB (the rest)", "CA-CB", "CA", "CB", 1, { { 1.530, 0.020, "CH1E-CH2E" } } },
       { "CB-CG", "CB-CG", "CB", "CG", 1, { { 1.512, 0.022, "CH2E-CY" } } },
       { "CG-CD2", "CG-CD2", "CG", "CD2", 1, { { 1.389, 0.021, "CY-CR1E" } } },
       { "CG-CD1", "CG-CD1", "CG", "CD1", 1, { { 1.389, 0.021, "CY-CR1E" } } },
       { "CD1-CE1", "CD1-CE1", "CD1", "CE1", 1, { { 1.382, 0.030, "CR1E-CR1E" } } },
       { "CE1-CZ", "CE1-CZ", "CE1", "CZ", 1, { { 1.378, 0.024, "CR1E-CY2" } } },
       { "CZ-OH", "CZ-OH", "CZ", "OH", 1, { { 1.376, 0.021, "CY2-OH1" } } },
       { "CZ-CE2", "CZ-CE2", "CZ", "CE2", 1, { { 1.378, 0.024, "CY2-CR1E" } } },
       { "CE2-CD2", "CE2-CD2", "CE2", "CD2", 1, { { 1.382, 0.030, "CR1E-CR1E" } } },
       { "CA-C (except Gly)", "CA-C", "CA", "C", 1, { { 1.525, 0.021, "CH1E-C" } } },
       { "C-O", "C-O", "C", "O", 2, {{1.231, 0.020, "C-O" },
                                     {1.249, 0.019, "C-OC"} } }
       // { "C-N (except Pro)", "C-N", "C", "N", 1, { { 1.329, 0.014, "C-NH1" } } },
     },
     {
       { "CB-CA-C (the rest)", "CB-CA-C", "CB", "CA", "C", 1, { { 110.100, 1.900, "CH2E-CH1E-C" } } },
       { "N-CA-CB (the rest)", "N-CA-CB", "N", "CA", "CB", 1, { { 110.500, 1.700, "NH1-CH1E-CH2E" } } },
       { "CA-CB-CG", "CA-CB-CG", "CA", "CB", "CG", 1, { { 113.900, 1.800, "CH1E-CH2E-CY" } } },
       { "CB-CG-CD2", "CB-CG-CD2", "CB", "CG", "CD2", 1, { { 120.800, 1.500, "CH2E-CY-CR1E" } } },
       { "CD1-CG-CD2", "CD1-CG-CD2", "CD1", "CG", "CD2", 1, { { 118.100, 1.500, "CR1E-CY-CR1E" } } },
       { "CB-CG-CD1", "CB-CG-CD1", "CB", "CG", "CD1", 1, { { 120.800, 1.500, "CH2E-CY-CR1E" } } },
       { "CG-CD1-CE1", "CG-CD1-CE1", "CG", "CD1", "CE1", 1, { { 121.200, 1.500, "CY-CR1E-CR1E" } } },
       { "CG-CD2-CE2", "CG-CD2-CE2", "CG", "CD2", "CE2", 1, { { 121.200, 1.500, "CY-CR1E-CR1E" } } },
       { "CD1-CE1-CZ", "CD1-CE1-CZ", "CD1", "CE1", "CZ", 1, { { 119.600, 1.800, "CR1E-CR1E-CY2" } } },
       { "OH-CZ-CE2", "OH-CZ-CE2", "OH", "CZ", "CE2", 1, { { 119.900, 3.000, "OH1-CY2-CR1E" } } },
       { "CE1-CZ-OH", "CE1-CZ-OH", "CE1", "CZ", "OH", 1, { { 119.900, 3.000, "CR1E-CY2-OH1" } } },
       { "CE1-CZ-CE2", "CE1-CZ-CE2", "CE1", "CZ", "CE2", 1, { { 120.300, 2.000, "CR1E-CY2-CR1E" } } },
       { "CZ-CE2-CD2", "CZ-CE2-CD2", "CZ", "CE2", "CD2", 1, { { 119.600, 1.800, "CY2-CR1E-CR1E" } } },
       { "N-CA-C (except Gly,Pro)", "N-CA-C", "N", "CA", "C", 1, { { 111.200, 2.800, "NH1-CH1E-C" } } },
       { "CA-C-O (except Gly)", "CA-C-O", "CA", "C", "O", 2, {{ 120.800, 1.700, "CH1E-C-O" },
                                                              { 117.000, 2.500, "CH1E-C-OC"} } }
       // { "CA-C-N (except Gly,Pro)", "CA-C-N", "CA", "C", "N", 1, { { 116.200, 2.000, "CH1E-C-NH1" } } },
       // { "O-C-N (except Pro)", "O-C-N", "O", "C", "N", 1, { { 123.000, 1.600, "O-C-NH1" } } },
       // { "C-N-CA (except Gly,Pro)", "C-N-CA", "C", "N", "CA", 1, { { 121.700, 1.800, "C-NH1-CH1E" } } },
     }
};

MONOMER VAL = { "VAL", 6, 7, // 7, 10,
     {
       { "N-CA (except Gly,Pro)", "N-CA", "N", "CA", 2, {{1.458, 0.019, "NH1-CH1E"},
                                                         {1.491, 0.021, "NH3-CH1E"}} },
       { "CA-CB (Ile,Thr,Val)", "CA-CB", "CA", "CB", 1, { { 1.540, 0.027, "CH1E-CH1E" } } },
       { "CB-CG1", "CB-CG1", "CB", "CG1", 1, { { 1.521, 0.033, "CH1E-CH3E" } } },
       { "CB-CG2", "CB-CG2", "CB", "CG2", 1, { { 1.521, 0.033, "CH1E-CH3E" } } },
       { "CA-C (except Gly)", "CA-C", "CA", "C", 1, { { 1.525, 0.021, "CH1E-C" } } },
       { "C-O", "C-O", "C", "O", 2, {{1.231, 0.020, "C-O" },
                                     {1.249, 0.019, "C-OC"} } }
       // { "C-N (except Pro)", "C-N", "C", "N", 1, { { 1.329, 0.014, "C-NH1" } } },
     },
     {
       { "CB-CA-C (Ile,Thr,Val)", "CB-CA-C", "CB", "CA", "C", 1, { { 109.100, 2.200, "CH1E-CH1E-C" } } },
       { "N-CA-CB (Ile,Thr,Val)", "N-CA-CB", "N", "CA", "CB", 1, { { 111.500, 1.700, "NH1-CH1E-CH1E" } } },
       { "CG1-CB-CG2", "CG1-CB-CG2", "CG1", "CB", "CG2", 1, { { 110.800, 2.200, "CH3E-CH1E-CH3E" } } },
       { "CA-CB-CG1", "CA-CB-CG1", "CA", "CB", "CG1", 1, { { 110.500, 1.700, "CH1E-CH1E-CH3E" } } },
       { "CA-CB-CG2", "CA-CB-CG2", "CA", "CB", "CG2", 1, { { 110.500, 1.700, "CH1E-CH1E-CH3E" } } },
       { "N-CA-C (except Gly,Pro)", "N-CA-C", "N", "CA", "C", 1, { { 111.200, 2.800, "NH1-CH1E-C" } } },
       { "CA-C-O (except Gly)", "CA-C-O", "CA", "C", "O", 2, {{ 120.800, 1.700, "CH1E-C-O" },
                                                              { 117.000, 2.500, "CH1E-C-OC"} } }
       // { "CA-C-N (except Gly,Pro)", "CA-C-N", "CA", "C", "N", 1, { { 116.200, 2.000, "CH1E-C-NH1" } } },
       // { "O-C-N (except Pro)", "O-C-N", "O", "C", "N", 1, { { 123.000, 1.600, "O-C-NH1" } } },
       // { "C-N-CA (except Gly,Pro)", "C-N-CA", "C", "N", "CA", 1, { { 121.700, 1.800, "C-NH1-CH1E" } } },
     }
};


MONOMER_DB params[MAX_MONOMER] = {
     { "A",    MONOMER_TYPE_DNA,  &DA },
     { "A",    MONOMER_TYPE_RNA,  &RA },
     { "ALA",  MONOMER_TYPE_A_A,  &ALA },
     { "ARG",  MONOMER_TYPE_A_A,  &ARG },
     { "ASN",  MONOMER_TYPE_A_A,  &ASN },
     { "ASP",  MONOMER_TYPE_A_A,  &ASP },
     { "C",    MONOMER_TYPE_DNA,  &DC },
     { "C",    MONOMER_TYPE_RNA,  &RC },
     { "CYS",  MONOMER_TYPE_A_A,  &CYS },
     { "G",    MONOMER_TYPE_DNA,  &DG },
     { "G",    MONOMER_TYPE_RNA,  &RG },
     { "GLN",  MONOMER_TYPE_A_A,  &GLN },
     { "GLU",  MONOMER_TYPE_A_A,  &GLU },
     { "GLY",  MONOMER_TYPE_A_A,  &GLY },
     { "HIS",  MONOMER_TYPE_A_A,  &HIS },
     { "ILE",  MONOMER_TYPE_A_A,  &ILE },
     { "LEU",  MONOMER_TYPE_A_A,  &LEU },
     { "LYS",  MONOMER_TYPE_A_A,  &LYS },
     { "MET",  MONOMER_TYPE_A_A,  &MET },
     { "PHE",  MONOMER_TYPE_A_A,  &PHE },
     { "PRO",  MONOMER_TYPE_A_A,  &PRO },
     { "SER",  MONOMER_TYPE_A_A,  &SER },
     { "T",    MONOMER_TYPE_DNA,  &DT },
     { "T",    MONOMER_TYPE_RNA,  &RT },
     { "THR",  MONOMER_TYPE_A_A,  &THR },
     { "TRP",  MONOMER_TYPE_A_A,  &TRP },
     { "TYR",  MONOMER_TYPE_A_A,  &TYR },
     { "U",    MONOMER_TYPE_DNA,  &DU },  
     { "U",    MONOMER_TYPE_RNA,  &RU },  
     { "VAL",  MONOMER_TYPE_A_A,  &VAL }
};

MONOMER_BOND A_LINK = { "C-N (except Pro)", "C-N", "C", "N", 1, { { 1.329, 0.014, "C-NH1" } } };

MONOMER_BOND PRO_LINK = { "C-N (Pro)", "C-N", "C", "N", 1, { { 1.341, 0.016, "C-NH1" } } };

MONOMER_BOND N_LINK = { "P-O3'", "O3*-P", "O3*", "P", 1, { { 1.607, 0.012, "s" } } };

MONOMER_ANGLE GLY_ANG1[2] = {
        { "CA-C-N (Gly)", "CA-C-N", "CA", "C", "N", 1, { { 116.400, 2.100, "CH2G*-C-NH1" } } },
        { "O-C-N (except Pro)", "O-C-N", "O", "C", "N", 1, { { 123.000, 1.600, "O-C-NH1" } } }
};

MONOMER_ANGLE GLY_ANG2 = { "C-N-CA (Gly)", "C-N-CA", "C", "N", "CA", 1, { { 120.600, 1.700, "C-NH1-CH2G*" } } };

MONOMER_ANGLE PRO_ANG1[2] = {
        { "CA-C-N (Pro)", "CA-C-N", "CA", "C", "N", 1, { { 116.900, 1.500, "CH1E-C-N" } } },
        { "O-C-N (Pro)", "O-C-N", "O", "C", "N", 1, { { 122.000, 1.400, "O-C-N" } } }
};

MONOMER_ANGLE PRO_ANG2 = { "C-N-CA (Pro)", "C-N-CA", "C", "N", "CA", 1, { { 122.600, 5.000, "C-N-CH1E" } } };

MONOMER_ANGLE AA_ANG1[2] = {
        { "CA-C-N (except Gly,Pro)", "CA-C-N", "CA", "C", "N", 1, { { 116.200, 2.000, "CH1E-C-NH1" } } },
        { "O-C-N (except Pro)", "O-C-N", "O", "C", "N", 1, { { 123.000, 1.600, "O-C-NH1" } } }
};

MONOMER_ANGLE AA_ANG2 = { "C-N-CA (except Gly,Pro)", "C-N-CA", "C", "N", "CA", 1, { { 121.700, 1.800, "C-NH1-CH1E" } } };

MONOMER_ANGLE NN_ANG1 = { "C3'-O3'-P", "C3*-O3*-P", "C3*", "O3*", "P", 1, { { 119.700, 1.2, "s" } } };

MONOMER_ANGLE NN_ANG2[3] = {
        { "O3'-P-O5'", "O3*-P-O5*", "O3*", "P", "O5*", 1, { { 104.000, 1.9, "s" } } },
        { "O3'-P-O1P/O2P", "O2P-P-O3*", "O3*", "P", "O2P", 2,
            {{ 110.500, 1.1, "large" }, { 105.200, 2.2, "small" }}
        },
        { "O3'-P-O1P/O2P", "O1P-P-O3*", "O3*", "P", "O1P", 2,
            {{ 110.500, 1.1, "large" }, { 105.200, 2.2, "small" }}
        }
};

static int _dbIndex_flag = 0;

static TblIndexObj _dbIndex;

static void _dbIndex_init()
{
       CifString index;

       if (_dbIndex_flag) return;

       _dbIndex_flag = 1;
       _dbIndex.Clear();
       for (int i = 0; i < MAX_MONOMER; i++) {
            index.Copy(params[i].resname);
            index += params[i].type;
            _dbIndex.Add(index);
       }
}

static MONOMER_DB *find_monomer_db(CifString &index)
{
       int ret = _dbIndex.Seek(index);
       if (ret >= 0) return (&params[ret]);
       return NULL;
}

#endif
